package weka.classifiers.lazy;

import java.text.DecimalFormat;
import java.util.Enumeration;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.classifiers.UpdateableClassifier;
import weka.classifiers.lazy.tree.LazySplit;
import weka.core.Drawable;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.UnsupportedAttributeTypeException;
import weka.core.UnsupportedClassTypeException;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;

/**
 * Class for generating a lazy decision tree. The Lazy Tree is constructed from the set of labeled instances 
 * and the unlabeled instance that should be classified. The tree has only one leaf. Split criterium is max infogain.
 * Attributes could be nominal and numeric, class attribute nominal.
 * Stop criterion of splitting of nodes is null entropy of instances in the leaf.  
 * For more information, see<p>
 * Jerome H.Friedman, Ron Kohavi, Yeogirl Yun (1996).<i>Lazy Decision Trees</i><p>
 * Ross Quinlan (1993). <i>C4.5: Programs for Machine Learning</i>, 
 * Morgan Kaufmann Publishers, San Mateo, CA. <p>
 *
 * Valid options are: <p>
 *
 * -J <br>
 * Join tree's nodes with split on the same attribute.<p>
 *
 * @author sergey poluchaevskiy (leonsp@cheerful.com)
 * @version 1.12.2005
 */
public class LazyTree extends Classifier implements OptionHandler, UpdateableClassifier, WeightedInstancesHandler {
    /** set of labeled instances */
    Instances instances=null;
    /** the instance that should be classified */
    Instance keyInstance=null;
    /** instances in the leaf of lazy tree */
    Instances leafInstances=null;
   
  /** The decision tree */
  public Vector treeNodes;
  
  /** Unoptimised tree */
  public boolean m_joined = true;
  
  /** Cleanup after the tree has been built. */
  public boolean m_noCleanup = false;
  
  /** Store true if the split on the value of the attribute was allready executed */
  private boolean splitAttributesValues [][];
  /** weights of instances of different classes to provide the same weight of classes in parent knote*/
  double [] classRates;
  /**
   * Returns a string describing classifier
   * @return a description suitable for
   * displaying in the explorer/experimenter gui
   */
  public String globalInfo() {

    return  "Class for generating of lazy decision tree./nThe Lazy Tree is constructed from the set of labeled instances" 
    + "and the unlabeled instance that should be classified./nThe tree has only one leaf. Split criterium is max infogain."
    + "Attributes could be nominal and numeric, class attribute is nominal./n"
    + "Stop criterion of splitting of instances in nodes is 0 entropy of instances in the node (leaf) or no selected split attribute.\n\n"
      + "Jerome H.Friedman, Ron Kohavi, Yeogirl Yun (1996). Lazy Decision Trees\n"
      + "Ross Quinlan (1993). C4.5: Programs for Machine Learning\n\n";
  }
  
  /**
   * Set and preprocess labeled instances.
   *
   * @exception Exception if classifier can't be built successfully
   */
  public void buildClassifier(Instances instances) throws Exception {
      if (instances.classAttribute().isNumeric())
            throw new UnsupportedClassTypeException("Class attribute is numeric! Can classify only nominal values...");
      if (instances.checkForStringAttributes()) {
            throw new UnsupportedAttributeTypeException("Cannot handle string attributes!");
           }
    this.instances=new Instances(instances);
    this.instances.deleteWithMissingClass();
    splitAttributesValues=new boolean[instances.numAttributes()][7];
    keyInstance=null;
    leafInstances=null;
    treeNodes=null;
  }
  /**
     * Adds the supplied instance to the training set
     * 
     * @param instance
     *            the instance to add
     * @exception Exception
     *                if instance could not be incorporated successfully
     */
    public void updateClassifier(Instance instance) throws Exception {
            if (instances.equalHeaders(instance.dataset()) == false)
            throw new Exception("Incompatible instance types");
        if (instance.classIsMissing())
            return;
        instances.add(instance);
    }
  /**
   * create array of size [number of attributes][number of values].  
   */
  private void createSplitAttributesValues(){
      int valuesNumber=0;
      for (int atr=0;atr<instances.numAttributes();atr++){
            valuesNumber=instances.attribute(atr).isNominal()&&!keyInstance.isMissing(atr)?instances.attribute(atr).numValues():1;
            splitAttributesValues[atr]=new boolean[valuesNumber];
            for (int value=0;value<valuesNumber;value++)
                splitAttributesValues[atr][value]=false;
        }
  }
  /** 
   * Returns class probabilities for the instance.
   * @exception Exception if distribution can't be computed successfully
   */
  public double [] distributionForInstance(Instance instance) throws Exception {
      if (instances==null)
          throw new Exception("Training instances are absent.");
      keyInstance=instance;
      buildTree();
    return getDistribution(leafInstances);
  }
  /**
   * get the array of classes frequencies of the instances
   * @return
   */
  public double [] getDistribution(Instances leafInstances){
      double [] distribution=new double[instances.numClasses()];
      Instance inst;
      Enumeration enu = leafInstances.enumerateInstances();
        while (enu.hasMoreElements()) {
            inst=(Instance) enu.nextElement();
            distribution[(int)inst.classValue()]++;
        }
        for (int c=0;c<instances.numClasses();c++)
            distribution[c]=distribution[c]/leafInstances.numInstances();
        return distribution;
  }
  /**
   * build lazy tree (the sequence of nodes with test for an attribute) 
   *
   */
  private void buildTree(){
      createSplitAttributesValues();
      int sortedIndex=-1;
      treeNodes=new Vector();
      LazySplit splitPoint;
      leafInstances=new Instances(instances);
      setWeights(leafInstances);
      for (int i=0;i<10000;i++){
          splitPoint=new LazySplit(leafInstances,keyInstance,splitAttributesValues,classRates,sortedIndex);
          if (splitPoint.getSplitAttribute()!=-1){
              leafInstances=splitPoint.getInstancesInBranch();
              setWeights(leafInstances);
              treeNodes.addElement(new TreeNode(splitPoint.getSplitAttribute(),splitPoint.getSplitValue(),splitPoint.getInfoGain(),splitPoint.getSize(),splitPoint.getEntropy()));
              if (splitPoint.getEntropy()==0||splitPoint.getInfoGain()==0) break;
              sortedIndex=splitPoint.getSortedIndex();
              splitAttributesValues=splitPoint.getSplitAttributesValues();
          } else {
              break;
          }
      }
      if (m_joined)
          collapseTree();
  }
  /**
   * Join tree's nodes with splits on the same attribute.
   */
  private void collapseTree(){
      int splitAttribute;
      int size1=treeNodes.size();
      int size2;
      int i,j;
      double splitValue;
      boolean sign;
      TreeNode current;
    for ( i=0;i<size1;i++){
        if (!(i<(size2=treeNodes.size())))
                break;
        if (instances.attribute(splitAttribute=((TreeNode)treeNodes.get(i)).splitAttribute).isNumeric()){
            splitValue=((TreeNode)treeNodes.get(i)).splitValue;
            sign=LazySplit.getSign(keyInstance,splitAttribute,splitValue);
            for ( j=i+1;j<size2;j++){
              if (!(j<treeNodes.size()))
                  break;
              current=(TreeNode)treeNodes.get(j);
              if (splitAttribute==current.splitAttribute&&sign==LazySplit.getSign(keyInstance,splitAttribute,current.splitValue)){
                  treeNodes.set(j,new TreeNode(splitAttribute,current.splitValue,Double.NEGATIVE_INFINITY,current.nodeSize,current.nodeEntropy));
                  treeNodes.remove(i);
                  i--;
                  break;
              }  
             }
        }
    }
  }
  /**
   * set weights of instances in parent knote (weight=1/number of classes)
   * @param instances
   */
  private void setWeights(Instances instances){
      Instance inst;
      double weight=1.000/instances.classAttribute().numValues();
      double [] classes=new double[instances.classAttribute().numValues()];
      classRates=new double[instances.classAttribute().numValues()];
      for (int i=0;i<classes.length;i++)
          classes[i]=0;
      Enumeration enu=instances.enumerateInstances();
      while (enu.hasMoreElements()){
            classes[(int)((Instance)enu.nextElement()).classValue()]++;  
          }
      for (int i=0;i<classes.length;i++){
          classRates[i]=instances.numInstances()/classes[i];
          //System.out.println(i+".  "+classRate[i]+" "+weight);
      }
      for (int i=0;i<instances.numInstances();i++){
          inst=instances.instance(i);
          instances.instance(i).setWeight(classRates[(int)inst.classValue()]);
      }
  }
  /**
   * print instances (attributes and values for all instances)
   * @param instances
   */
  public static void printInstances(Instances instances){
      System.out.println("Number of instances: "+instances.numInstances());
      Enumeration enu=instances.enumerateInstances();
      Instance inst;
      int i=0;
      while (enu.hasMoreElements()){
          inst=(Instance)enu.nextElement();
          System.out.print(i+". ");
          for (int a=0;a<instances.numAttributes();a++)
              if (a!=instances.classIndex())
              System.out.print(inst.attribute(a).name()+":"+inst.value(a)+" ");
          System.out.println("class: "+inst.classValue());
          i++;
      }
  }
  /**
   * get instances in the leaf of Lazy Tree
   * @return
   */
  public Instances getLeafInstances(){
      return leafInstances;
  }
  /**
   *  Returns the type of graph this classifier
   *  represents.
   *  @return Drawable.TREE
   */   
  public int graphType() {
      return Drawable.TREE;
  }
  /**
   * Returns an enumeration describing the available options.
   *
   * Valid options are: <p>
   *
   * -J <br>
   * Join tree's nodes with split on the same attribute.<p>
   * @return an enumeration of all the available options.
   */
  public Enumeration listOptions() {
    Vector newVector = new Vector();
    newVector.addElement(new Option("\tJoin tree's nodes with split on the same attribute.","J", 0, "-U"));
    return newVector.elements();
  }

  /**
   * Parses a given list of options.
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {
      setJoinedNodes(Utils.getFlag('J', options));
  }
  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {
    String [] options = new String [1];
    int current = 0;
    if (m_joined) {
      options[current++] = "-J";
    } 
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }
  /**
   * get value of attribute as string value
   * @param inst
   * @param index
   * @return
   */
  private String getValue(Instance inst, int index){
      if (instances.attribute(index).isNominal())
          return inst.stringValue(index);
      return inst.value(index)+"";
  }
     
  /** 
   * Returns the description of the classifier.
   */
  public String toString() {
    double distribution [];
    int sorted [];
    String s="";
    DecimalFormat df=new DecimalFormat("#.##");
    DecimalFormat df1=new DecimalFormat("#.####");
    /**if (instances == null) 
      s="Training set is not set. ";
    if (keyInstance==null)
        s+="The to classified instance is not set.";*/
    if (treeNodes==null||keyInstance==null)
        return s+" No Lazy Tree is builded.";
    TreeNode node;
    StringBuffer text = new StringBuffer();
    text.append("Classifier Lazy Tree for the instance:\n");
    for (int i=0;i<instances.numAttributes();i++)
        if (instances.classIndex()!=i)
            text.append(i+". "+keyInstance.attribute(i).name()+": "+(keyInstance.isMissing(i)?"?":getValue(keyInstance,i))+"\n");
    text.append("Class attribute "+keyInstance.classAttribute().name()+"("+instances.numClasses()+" values)\t:"+getValue(keyInstance,instances.classIndex())+"\n");
    text.append("\nLazy tree ("+((treeNodes.size()-1)==0?"":treeNodes.size()-1+"")+" nodes+1 leaf) builded on the base of "+instances.numInstances()+" instances:\n");
    String condition="";
    for (int i=0;i<treeNodes.size();i++){
        node=(TreeNode)treeNodes.get(i);
        if (instances.attribute(node.splitAttribute).isNominal())
            condition="<>"+instances.attribute(node.splitAttribute).value((int)node.splitValue);
        if (instances.attribute(node.splitAttribute).isNumeric()){
            condition=((keyInstance.value(node.splitAttribute)<=node.splitValue)?"<=":">")+df1.format(node.splitValue);
        }
        text.append(i+" "+instances.attribute(node.splitAttribute).name() +condition+"  Size:"+node.nodeSize
            +"  Entropy:"+df.format(node.nodeEntropy)
            +(node.infoGain==Double.NEGATIVE_INFINITY?" ":"  InfoGain:"+df.format(node.infoGain))+"\n" );
    }
    distribution =getDistribution(leafInstances);
    sorted=getSortedIndex(distribution);
    text.append("\nDistribution of class values in the leaf with "+leafInstances.numInstances()+" instances:\n");
    for (int i=0;i<distribution.length;i++)
        if (distribution[sorted[i]]>0)
            text.append(instances.classAttribute().value(sorted[i])+" - "+distribution[sorted[i]]*100+" %\n");  
   return text.toString() +"\n";
}
  /**
   * get sorted descending index of double array
   * @param distribution
   * @return
   */
  public static int [] getSortedIndex(double [] distribution){
      int sorted [];
      int temp;
      sorted=new int[distribution.length];
        for (int i=0;i<distribution.length;i++)
            sorted[i]=i;
        for (int i=0;i<distribution.length-1;i++)
            for (int j=i+1;j<distribution.length;j++){
                if (distribution[(int)sorted[i]]<distribution[(int)sorted[j]]){
                    temp=sorted[i];
                    sorted[i]=sorted[j];
                    sorted[j]=temp;
            }
        }
      return sorted;
  }
  /**
   * Returns the number of instances in the leaf of lazy tree
   */
  public int leafSize() {
      if (leafInstances==null)
          return 0;
    return  leafInstances.numInstances(); 
   }
  /**
   * Returns a number of nodes in lazy tree
   */
  public String toSummaryString() {
      if (treeNodes==null)
          return "The tree is not builded.";
    return "Size of the tree: " + treeNodes.size();
  }

  /**
   * Returns the size of the lazy tree
   * @return the size of the tree
   */
  public double measureTreeSize() {
      if (treeNodes==null)
          return 0;
    return treeNodes.size();
  }

  /**
   * Returns the number of leaves
   * @return the number of leaves
   */
  public double measureNumLeaves() {
    return 1;
  } 
  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String joinedNodesTipText() {
    return "Join nodes of the lazy tree with split on the same attribute.";
  }

  /**
   * Get wether the nodes of the lazy tree with split on the same attribute should be joined.
   *
   * @return Value of unpruned.
   */
  public boolean getJoinedNodes() {
    return m_joined;
  } 
  /**
   * Set wether the nodes of the lazy tree with split on the same attribute should be joined.
   *
   * @return Value of unpruned.
   */
  public void setJoinedNodes(boolean join) {
    m_joined=join;
  } 
  /**
   * contains the description of nodes of tree (split attribut, split value, number of instances, entropy anf infogain of the node) 
   *
   */
  private class TreeNode{
      int splitAttribute;
      double splitValue;
      double infoGain;
      int nodeSize;
      double nodeEntropy;
      Instances nodeInstances=null;
      /**
       * create new tree's node with split information of instances
       * @param attribute split attribute
       * @param value split value
       */
      TreeNode(int attribute, double value ){
          this(attribute,value,Double.NEGATIVE_INFINITY,1,Double.NEGATIVE_INFINITY);
      }
      /**
       * create new tree's node with split information of instances
       * @param attribute
       * @param value
       * @param infogain
       * @param size
       * @param entropy
       */
      TreeNode(int attribute, double value, double infogain,int size, double entropy ){
          splitAttribute=attribute;
          splitValue=value;
          infoGain=infogain;
          nodeSize=size;
          nodeEntropy=entropy;
      }
      /**
       * create new tree's node
       * @param attribute
       * @param value
       * @param infogain
       */
      TreeNode(int attribute, double value, double infogain,int size,double entropy, Instances instances ){
          this(attribute, value, infogain,size,entropy );
          nodeInstances=new Instances(instances);
      }
  }
  /**
   * Main method for testing this class
   *
   * @param String options 
   */
  public static void main(String [] argv){

    try {
      
    } catch (Exception e) {
      System.err.println(e.getMessage());
    }
  }
}
