package weka.classifiers.lazy.tree;
import java.text.DecimalFormat;
import java.util.Enumeration;

import weka.core.Instance;
import weka.core.Instances;
/**
 * class for finding the split point of instances with max info gain
 * @author sergey poluchaevskiy (leonsp@cheerful.com)
 * @version 1.12.2005
 * This class is used by LazyTree to find the best splitting point (attribute and attribute's value with max info gain) on the set of instances.
 * 
 */
public class LazySplit {
    /** the set of labelled classes */
    Instances instances;
    /** the instance that should be classified */
    Instance keyInstance;
    /** rates of classes to provide the same rate of classes in parent knote*/
    double [] classRates;
    /** split attribute index */
    int splitAttribute=-1;
    /** split value of attribute */
    double splitValue;
    /** split index */
    double splitIndex;
    /** entropy of instances after splitting that are relevant for classified instance*/
    double nodeEntropy;
    /** the number of instances after splitting that are relevant for classified instance*/
    int nodeSize;
    /** max info gain that was accessed by splitting */ 
    double maxInfoGain=Double.NEGATIVE_INFINITY;
    /** indicate attribute index by that instances are sorted*/
    int sortedIndex=-1;
    /**distribution of classes for nominal attributes*/
    double [][][] distribution=null;
    /**distribution of classes for nominal attributes*/
    double [][][] distribSize=null;
    /** Array of checked splits. Store true if the split on this value of the nominal attribute was allready executed */
    private boolean splitAttributesValues [][];
    
    protected final double log2=Math.log(2);
    /**
     * 
     * set labelled instances and the instance that should be classified 
     *
     * @param instances
     * @param keyInstance
     * @param splitAttributesValues
     * @param classRates
     * @param sortIndex
     */
    public LazySplit(Instances instances, Instance keyInstance,boolean [][] splitAttributesValues,double [] classRates,int sortIndex){
        this.instances=new Instances(instances);
        this.keyInstance=keyInstance;
        this.classRates=classRates;
        sortedIndex=sortIndex;
        this.splitAttributesValues=splitAttributesValues;
        findSplitting();
        if (splitAttribute!=-1)
            splitAttributesValues[splitAttribute][instances.attribute(splitAttribute).isNominal()?(int)splitValue:0]=true;
    }
    /**
     * get split attribute
     * @return
     */
    public int getSplitAttribute(){
        return splitAttribute;
    }
    /**
     * get split value of split attribute
     */
    public double getSplitValue(){
        return splitValue;
    }
    /**
     * get number of instances in the node after partition
     */
    public int getSize(){
        return nodeSize;
    }
    /**
     * get info gain achieved by splitting
     */
    public double getInfoGain(){
        return maxInfoGain;
    }
    /**
     * get entropy of set of instances after partition
     */
    public double getEntropy(){
        return nodeEntropy;
    }
    /**
     * get index of attribute by that instances are sorted
     */
    public int getSortedIndex(){
        return sortedIndex;
    }
    /**
     * get array of checked splits on values of nominal attributes
     * @return
     */
    public boolean [][] getSplitAttributesValues(){
        return splitAttributesValues;
    }
    /**
     * set index of attribute by that instances are sorted
     */
    public void setSortedIndex(int a){
        sortedIndex=a;
    }
    /**
     * find an attribute and the attribute's value with max information gain
     *
     */
    public void findSplitting(){
        if (!isHomogen(instances)){
            createClassesDistribution();
            getInformationGain();
        }
    }
    /**
     * get the product of value with its log2
     * @param x
     * @return
     */
    private double logFunc(double x){
        if (x==0||x<1e-6)
            return 0;
        return (x*Math.log(x)/log2);
    }
    /**
     * get true if the relevant instances are in right side of splitting (>splitValue) or false else (<=splitValue)
     * @param keyInstance
     * @param splitAttribute
     * @param splitValue
     * @return
     */
    public static boolean getSign(Instance keyInstance,int splitAttribute, double splitValue){
        return keyInstance.value(splitAttribute)>splitValue?true:false;
    }
    /**
     * get the part of instances for the new branch of tree (splitValue in first partition of dataset)
     * @return
     */
    public Instances getInstancesInBranch(){
        if (splitAttribute<0)
            return null;
        Instances branchInstances=new Instances(instances,1);
        Instance inst;
        boolean isNominalAttribute=keyInstance.attribute(splitAttribute).isNominal();
        boolean splitSign=getSign(keyInstance,splitAttribute,splitValue);
        Enumeration enu = instances.enumerateInstances();
        while (enu.hasMoreElements()) {
            inst=(Instance) enu.nextElement();
            if (isNominalAttribute) {
                if (inst.value(splitAttribute)!=splitValue)
                    branchInstances.add(inst);
            } else
                if (getSign(inst,splitAttribute,splitValue)==splitSign)
                    branchInstances.add(inst);
        }
        return branchInstances;
    }
    /**
     * check wether instances are homogen (all instances have the same class value)
     * @param instances
     * @return
     */
    public static boolean isHomogen(Instances instances){
        double c=0;
        Enumeration enu = instances.enumerateInstances();
        if (enu.hasMoreElements())          
            c=((Instance)enu.nextElement()).classValue();
        while (enu.hasMoreElements()) {
            if (((Instance)enu.nextElement()).classValue()!=c)
                return false;
        }
        return true;
    }
    /**
     * create 3d array of classes distribution depends on attribute and the value of attribute. 
     * This array is used to find best splitting point in case of nominal attribute.
     *
     */
    private void createClassesDistribution(){
        Instance inst;
        int valuesNumber;
        distribution=new double [instances.numAttributes()][7][instances.numClasses()];
        distribSize=new double [instances.numAttributes()][7][instances.numClasses()];
        for (int a=0; a<instances.numAttributes();a++){
            valuesNumber=1+((keyInstance.isMissing(a)||instances.attribute(a).isNumeric())?1:keyInstance.attribute(a).numValues());
            distribution[a]=new double[valuesNumber][instances.numClasses()];
            distribSize[a]=new double[valuesNumber][instances.numClasses()];
            for (int v=0;v<valuesNumber;v++){
                distribution[a][v]=new double[instances.numClasses()];
                distribSize[a][v]=new double[instances.numClasses()];
                for (int c=0;c<instances.numClasses();c++){
                    distribution[a][v][c]=0;
                    distribSize[a][v][c]=0;
                }
            }
        }
        for (int i=0; i<instances.numAttributes();i++){
            if (i!=keyInstance.classIndex()&&instances.attribute(i).isNominal()) {
                Enumeration enu = instances.enumerateInstances();
                while (enu.hasMoreElements()) {
                    inst=(Instance) enu.nextElement();
                    if (!inst.isMissing(i)){
                        distribution[i][keyInstance.isMissing(i)?distribution[i].length-1:(int)inst.value(i)][(int)inst.classValue()]+=inst.weight();
                        distribSize[i][keyInstance.isMissing(i)?distribution[i].length-1:(int)inst.value(i)][(int)inst.classValue()]+=inst.weight();
                    }
                } 
            }
        }
    }
    /**
     * find the attribute and the attribute's value with max information gain. For each attribute and attribute value
     * is info gain calculated. The point with max info gain is selected.
     *
     */
    private void getInformationGain(){
        for (int a=0;a<instances.numAttributes();a++)
            if (!keyInstance.isMissing(a)&&a!=instances.classIndex()){
                if (instances.attribute(a).isNominal()){
                    getSplitPointForNominalAttribute(a);
                }
                if (instances.attribute(a).isNumeric()){
                    getSplitPointForNumerAttribute(instances, a);
                }
            }
    }
    
    /**
     * get split point for nominal attribute 
     */
    public void getSplitPointForNominalAttribute(int attrIndex){
        int valuesNumber;
        double infox=0,info,entropy1,entropy2;
        double infoGain=Double.MIN_VALUE;
        double size,size1,size2;
        valuesNumber=distribution[attrIndex].length-1;
        size=getValuesClassesSum(attrIndex);
        for (int value=0;value<valuesNumber;value++){
            if (value!=keyInstance.value(attrIndex)&&splitAttributesValues[attrIndex][value]!=true){
                size1=getClassesSum(attrIndex,value);// number of instances in first branch of tree
                size2=getValuesClassesSum(attrIndex,value);// number of instances in second branch of tree
                entropy1=getAttributeValueEntropy(attrIndex,value,size1);// entropy of first branch
                entropy2=getAttributeValuesEntropy(attrIndex,value,size2);// entropy of second branch
                info=getClassesEntropy(attrIndex,size);// entropy of parent knote
                infox=size1/size*entropy1+size2/size*entropy2;
                infoGain=info-infox;
                if (maxInfoGain<infoGain) {
                    maxInfoGain=infoGain;
                    splitAttribute=attrIndex;
                    splitValue=value;
                    nodeEntropy=entropy2;
                    nodeSize=(int)size2;
                }
            }
        }
    }
    /**
     * get split point for numeric attribute
     * @param instances
     * @param a
     */
    public void getSplitPointForNumerAttribute(Instances instances, int a){
        Enumeration enu;
        double info,infox,infoGain,entropy1,entropy2;
        if (a!=sortedIndex){
            instances.sort(a);
            sortedIndex=a;
        }
        int index;
        double value,value1,middleValue;
        double size=0;
        double size1=0;
        double size2=0;
        int [] classes1 =new int[instances.numClasses()];// first branch of the tree
        int [] classes2 =new int[instances.numClasses()];// second branch of the tree
        Instance inst;
        int max=0;
        enu = instances.enumerateInstances();
        while (enu.hasMoreElements()) {
            inst=(Instance) enu.nextElement();
            if (!inst.isMissing(a)){
                classes2[(int)inst.classValue()]+=inst.weight();
                size+=inst.weight();
                max++;
            }
        }
        info=getAttributeEntropy(classes2,size);
        size2=size;
        size1=0;
        index=0;
        enu = instances.enumerateInstances();
        while (enu.hasMoreElements()&&size2>=0) {
            inst=(Instance) enu.nextElement();
            if (!inst.isMissing(a)){
                classes1[(int)inst.classValue()]+=inst.weight();
                size1+=inst.weight();
                classes2[(int)inst.classValue()]+=-inst.weight();
                size2=size2-inst.weight();
                //System.out.println("s1="+size1+" s2="+size2+" s="+size+" atr="+a+" Index="+index);
                entropy1=getAttributeEntropy(classes1,size1);
                entropy2=getAttributeEntropy(classes2,size2);
                infox=size1/size*entropy1+size2/size*entropy2;
                infoGain=info-infox;
                if (maxInfoGain<infoGain) {
                    value=inst.value(a);
                    //System.out.println("Attribute:"+a+" Value:"+value+" IG:"+infoGain+" next value:"+((index<size-1)?instances.instance(index+1).value(a):-1)+" last:"+(index==size-1));
                    if (index<max-1&&value!=(value1=instances.instance(index+1).value(a))){
                        middleValue=(value+value1)/2;
                        if (keyInstance.value(a)!=middleValue){
                            maxInfoGain=infoGain;
                            splitAttribute=a;
                            splitValue=middleValue;
                            nodeEntropy=(keyInstance.value(a)<=splitValue)?entropy1:entropy2;
                            nodeSize=(keyInstance.value(a)<=splitValue)?(int)index+1:(int)max-index-1;
                            splitIndex=index;   
                        }
                    }
                }
            }
        index++;    
        } 
    }
    
    /**
     * get entropy of array. Used for numeric attribute.
     * @param classes
     * @param size
     * @return
     */
    public double getAttributeEntropy(int [] classes, double size){
        if (size==0){
            //System.out.println(" getAttributeEntropy. size=0"+this);
            return 0;
            }
        double entropy=0;
        double p;
        for (int c=0;c<classes.length;c++){
            p=((double)classes[c])/size;
            entropy+=-logFunc(p);
            }
        return entropy;
    }
    /**
     *  get the entropy of instances with the given value of the attribute (for nominal attribute)
     * @param attribute
     * @param value
     * @return
     */
    private double getAttributeValueEntropy(int attribute, int value, double size){
        double entropy=0;
        double p;
        int classesNumber=distribution[0][0].length;
        for (int c=0;c<classesNumber;c++){
            p=distribution[attribute][value][c]/size;
            entropy+=-logFunc(p);
        }
        return entropy;
    }
    
    /**
     *  get the entropy of instances for all values of the attribute exepted given value (for nominal attribute)
     * @param attribute
     * @param value
     * @return
     */
    private double getAttributeValuesEntropy(int attribute, int value, double size){
        double entropy=0;
        double sum;
        int valuesNumber=distribution[attribute].length-1;
        int classesNumber=distribution[0][0].length;
        for (int c=0;c<classesNumber;c++){
            sum=0;
            for (int v=0;v<valuesNumber;v++)
                if (value!=v)
                    sum+=distribution[attribute][v][c];
            entropy+=-logFunc(sum/size);
        }
        return entropy;
    }
    
    /**
     *  get the entropy of instances for given attribute (for nominal attribute)
     * @return
     */
    private double getClassesEntropy(int attribute, double size){
        double entropy=0;
        double p;
        double sum=0;
        int classesNumber=distribution[0][0].length;
        for (int c=0;c<classesNumber;c++){
            sum=0;
            int valuesNumber=distribution[attribute].length-1;
            for (int v=0;v<valuesNumber;v++)
                sum+=distribution[attribute][v][c];
            p=sum/size;
            entropy+=-logFunc(p);
        }
        return entropy;
    }   
    /**
     * get the number of instances with not missing attribute value, excepted given value
     * @param attribute
     * @param value
     * @return
     */
    private double getValuesClassesSum(int attribute,int value){
        double sum=0;
        int valuesNumber=distribSize[attribute].length-1;
        int classesNumber=distribSize[0][0].length;
        for (int i=0;i<valuesNumber;i++)
            if (i!=value){
                for (int c=0;c<classesNumber;c++)
                    sum=sum+distribSize[attribute][i][c];
            }
        return sum;
    }
    /**
     * get the number of instances with given value of attribute
     * @param attribute
     * @param value
     * @return
     */
    private double getClassesSum(int attribute,int value){
        double sum=0;
        int classesNumber=distribSize[0][0].length;
        for (int c=0;c<classesNumber;c++)
            sum=sum+distribSize[attribute][value][c];
        return sum;
    }
    /**
     * get the number of instances with not missing value of attribute
     * @param attribute
     * @return
     */
    private double getValuesClassesSum(int attribute){
        double sum=0;
        int valuesNumber=distribSize[attribute].length-1;
        int classesNumber=distribSize[0][0].length;
        for (int i=0;i<valuesNumber;i++)
            for (int c=0;c<classesNumber;c++)
                sum=sum+distribSize[attribute][i][c];
        return sum;
    }
    /**
     * get the description of splitting point as String
     */
    public String toString(){
        DecimalFormat df=new DecimalFormat("#.##");
        if (splitAttribute<0)
            return "No splitting builded";
        return "Split Attribute:"+(int)splitAttribute+" Value:"+splitValue+" InfoGain:"+df.format(maxInfoGain)+" Entropy:"+df.format(nodeEntropy)+" NodeSize:"+(int)nodeSize;
    }
}
