/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    CN2.java
 *    Copyright (C) 2004 Jrgen Beringer
 */


package weka.classifiers.rules;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.rules.rise.SimpleConjunctiveRule;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.NoSupportForMissingValuesException;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.UnsupportedAttributeTypeException;
import weka.core.UnsupportedClassTypeException;
import weka.core.Utils;

/**
 * This class implements a rule induction learner, CN2 
 * which is proposed by Peter Clark and Tim Niblett. 
 * For more information, see<p>
 *
 * P. Clark and T. Niblett. <i>The CN2 Induction Algorithm</i>, 
 * Machine Learning, 3(4):261-283, 1989.<br>
 * 
 * P. Clark and R. Boswell. <i>Rule induction with CN2: 
 * Some recent improvements</i>, In Y. Kodratoff, editor, 
 * Machine Learning - EWSL-91, pages 151-163, Berlin, 1991. Springer-Verlag</p>
 * 
 * Valid options are: <p>
 *
 * -B number<br>
 * Set size of the beam array.<p>
 * 
 * -C number<br>
 * Set the confidence probability for the significance test. Only 0.9, 0.95 and 0.99 are supported.<p>
 * 
 * -O<br>
 * Use ordered set of rules.<p>
 * 
 * 
 *
 * @author Jrgen Beringer (beringer@mathematik.uni-marburg.de)
 * @version $Revision: 1.11 $
 */

public class CN2 extends Classifier implements OptionHandler {
	
	/**Confidence bounds for the significance computation : 0.9*/ 
	public static double CONFIDENCE_0_9=2.706;
	/**Confidence bounds for the significance computation : 0.95*/
	public static double CONFIDENCE_0_95=3.841;
	/**Confidence bounds for the significance computation : 0.99*/
	public static double CONFIDENCE_0_99=6.635;
	
	/**Confidence bound (default for p=0.95)*/
	protected double confidence=CONFIDENCE_0_95;
	/**Confidence probability bound*/
	protected double confProb=0.95;
	/** Option if rules are learned ordered.*/
	protected boolean ordered = false;
	/** Beam-value for the beamsearch algorithm.*/
	protected int beamCount = 20;
	/**List of used attributes.*/
	protected List usedAttributeList;
	/**Array to organize used attributes.*/
	protected boolean[] usedAttribute;
	/**Index of the class attribute.*/
	protected int classAttr;
	/**List of all Rules*/
	protected FastVector ruleset;
	/**Default Rule: IF true THEN most frequent class*/
	protected SimpleConjunctiveRule defaultRule;
	/**Frequency of each class*/
	protected double[] orderedClasses;
	/**Store class order (most less class first).*/
	protected int[] classOrder;
	/**Count of data/instances*/
	protected double dataCount;
	/**Instances to learn */
	protected Instances allData;
	/**List of used instances*/
	protected FastVector data;
	/**number of attributes*/
	protected int attrCount;
	/**number of classes*/
	protected int classCount;
	

	/**
	 * Generates a classifier. Must initialize all fields of the classifier
	 * that are not being set via options (ie. multiple calls of buildClassifier
	 * must always lead to the same result). Must not change the dataset
	 * in any way.
	 *
	 * @param dataDFT set of instances serving as training data 
	 * @exception Exception if the classifier has not been 
	 * generated successfully
	 */
	public void buildClassifier(Instances instances) throws Exception {
		if (instances.numInstances() == 0)
				throw new Exception(" No instances with a class value!");

		if (instances.checkForStringAttributes())
				throw new UnsupportedAttributeTypeException(
						" Cannot handle string attributes!");

		if (!instances.classAttribute().isNominal())
				throw new UnsupportedClassTypeException(
						" Only nominal class, please.");

		Enumeration enumAtt = instances.enumerateAttributes();
		while (enumAtt.hasMoreElements()) {
			Attribute attr = (Attribute) enumAtt.nextElement();
			Enumeration enum1 = instances.enumerateInstances();
			while (enum1.hasMoreElements()) {
				if (((Instance) enum1.nextElement()).isMissing(attr)) { throw new NoSupportForMissingValuesException(
						"CN2: no missing values, please."); }
			}
		}

		allData=instances;
		
		defaultRule = getRule(allData);
		defaultRule.computeDistribution(allData);
		defaultRule.computeClass();
		
		orderedClasses=defaultRule.getDistribution();
		
		classOrder=Utils.sort(orderedClasses);		

		dataCount = allData.numInstances();

		if (allData == null)
				throw new Exception(" Unable to randomize the class orders.");

		allData.deleteWithMissingClass();
		if (allData.numInstances() == 0)
				throw new Exception(" No instances with a class value!");

		ruleset = new FastVector();


		classCount = allData.numClasses();
		attrCount = allData.numAttributes();
		classAttr = allData.classIndex();

		computeUsedAttributes(allData);

		//CN2 creates rules beginning with the less frequent class.
		for (int cl = 0; cl < allData.numClasses() - 1; cl++) {
			//if unordered reuse all examples for the next class
			if (!ordered || cl == 0) data = instanceVector(allData);
			addRulesetForOneClass(classOrder[cl]);
		}
		if (!ordered) data = instanceVector(allData);
		if (!ordered) addRulesetForOneClass(classOrder[allData.numClasses() - 1]);
	}
	
	/**
	 * Copy the instances to a FastVector.
	 * @param data Instances to copy
	 * @return FastVector with all instances.
	 */
	protected FastVector instanceVector(Instances data){
		int size=data.numInstances();
		FastVector result=new FastVector(size);
		
		for (int i=size-1;i>=0;i--){
			result.addElement(data.instance(i));
		}
		return result;
	}
	
	/**
	 * Is Attribute in the used attribute list.
	 * @param attribute attribute index
	 * @return if this attribute should be used for learning
	 */
	public boolean isUsedAttribute(int attribute) {
		if (usedAttributeList == null)
			return true;
		else
			return (usedAttributeList.indexOf(new Integer(attribute)) >= 0);
	}
	/**
	 * compute for each attribute, if this attribute can be used for learning.
	 * @param inst data
	 */
	protected void computeUsedAttributes(Instances inst){
		usedAttribute = new boolean[inst.numAttributes()];

		for (int i = 0; i < inst.numAttributes(); i++) {
			if ((i != classAttr) && inst.attribute(i).isNominal()
					&& isUsedAttribute(i)) usedAttribute[i] = true;
		}
	}
	/**
	 * Search the rules for the specified class index.
	 * @param classIndex class
	 */
	protected void addRulesetForOneClass(int classIndex) {
		SimpleConjunctiveRule bestRule = null;
		do {
			//System.out.println(data.size());
			bestRule = findBestRule(classIndex);
			if (bestRule != null) {
				ruleset.addElement(bestRule);
				//Alle untersttzten datenstze lschen
				int size = data.size();
				int pos = 0;
				while (pos < size) {
					Instance inst = (Instance)data.elementAt(pos);
					if (bestRule.covers(inst)) {
						data.removeElementAt(pos);
						size--;
					} else
						pos++;
				}
			}
		} while (bestRule != null && data.size() > 0);
	}
	/**
	 * Get an empty rule.
	 * @param data data-structure
	 * @return empty rule
	 */
	protected SimpleConjunctiveRule getRule(Instances data){
		return new SimpleConjunctiveRule(data,true);
	}
	/**
	 * Get a copy of the rule.
	 * @param rule original rule
	 * @return copied rule
	 */
	protected SimpleConjunctiveRule getRule(SimpleConjunctiveRule rule){
		return new SimpleConjunctiveRule(rule);
	}
	/**
	 * Get a default rule for the specified class
	 * @param classIndex class
	 * @return default rule for the class
	 */
	protected SimpleConjunctiveRule getEmptyRule(int classIndex) {
		SimpleConjunctiveRule empty = getRule(allData);
		empty.setClass(classIndex);
		if (!ordered) {
			for (int i = 0; i < classCount; i++)
				empty.addClass(i, orderedClasses[i]);
			empty.addCovered(data);
		} else {
			empty.computeDistribution(data);
			//if (data.numInstances()==0) empty.addClass(classIndex);
		}
		return empty;
	}

	/**
	 * Find the best rule for the specified class
	 * @param classIndex class
	 * @return best rule or null
	 */
	protected SimpleConjunctiveRule findBestRule(int classIndex) {
		SimpleConjunctiveRule empty = getEmptyRule(classIndex);
		FastVector star = new FastVector();
		star.addElement(empty);
		SimpleConjunctiveRule[] beamArray=new SimpleConjunctiveRule[beamCount];;
		SimpleConjunctiveRule best = null;
		int predCount = 0;
		double bestCriteria = -Double.MAX_VALUE;
		while (star.size() > 0) {
			for (int i = 0; i < beamCount; i++)
				beamArray[i] = null;
			Enumeration enum1 = star.elements();
			while (enum1.hasMoreElements()) {
				SimpleConjunctiveRule rule = (SimpleConjunctiveRule) enum1.nextElement();
				if (rule.getPredicateCount() == predCount) {
					Iterator it = specialize(rule);
					while (it.hasNext()) {
						SimpleConjunctiveRule newRule = (SimpleConjunctiveRule) it.next();
						newRule.computeDistribution(rule.getCovered());
						//newRule.computeDistribution(data);
						double crit = computeCriteria(newRule);
						if (newRule.getCoverage() > 0) {
							if (newRule.frequentClass() == classIndex
									&& crit > bestCriteria) {
								best = newRule;
								bestCriteria = computeCriteria(newRule);
							}
							if (!significance(newRule, classIndex)) {
								int pos = 0;
								while (pos < beamCount
										&& beamArray[pos] != null
										&& computeCriteria(beamArray[pos]) > crit)
									pos++;
								if (pos < beamCount) {
									if (beamArray[pos] != null)
											System.arraycopy(beamArray, pos,
													beamArray, pos + 1,
													beamCount - pos - 1);
									beamArray[pos] = newRule;
								}
							}
						}
					}
				}
			}
			star.removeAllElements();
			int pos = 0;
			while (pos < beamCount && beamArray[pos] != null) {
				star.addElement(beamArray[pos]);
				pos++;
			}
			predCount++;
		}
		return best;
	}
	/**
	 * Compute the rule criteria. Here the Laplace index is used.
	 * @param rule the rule
	 * @return criteria
	 */
	protected double computeCriteria(SimpleConjunctiveRule rule) {
		return (rule.getDistribution()[rule.getClassValue()] + 1)
				/ (rule.getDistributionSum() + classCount);
	}
	/**
	 * Test if the rule is significant better than a default rule.
	 * @param rule the rule
	 * @param classIndex the class of the rule
	 * @return result of the significance test
	 */
	protected boolean significance(SimpleConjunctiveRule rule, int classIndex) {
		double[] distribution = rule.getDistribution();
		double sum = rule.getDistributionSum();
		int cl = classIndex;
		double p = distribution[cl];
		if (p <= 0) return false;
		if (sum - p <= 0) return false;
		if (p<sum/2) return false;
		if (orderedClasses[cl]/dataCount >= p / sum) return false;

		double val = 2 * (p * Math.log(p * dataCount / (orderedClasses[cl] * sum)) + (sum - p)
				* Math.log((sum - p) * dataCount
						/ ((dataCount - orderedClasses[cl]) * sum)));
		return val >= confidence;
		//return true;
	}
	/**
	 * Get an iterator for all specified rules of the input rule
	 * @param rule the rule to specify
	 * @return Iterator of specified rules.
	 */
	protected Iterator specialize(final SimpleConjunctiveRule rule) {
		return new Iterator() {

			SimpleConjunctiveRule next = null;

			int attr = 0;

			int val = 0;

			int valCount = allData.numDistinctValues(attr);

			public void remove() {
			}

			public boolean hasNext() {
				if (next == null)
					return computeNext();
				else
					return true;
			}

			public Object next() {
				if (next == null) computeNext();
				Object val = next;
				next = null;
				return val;
			}

			private boolean computeNext() {
				if (attr >= attrCount) return false;
				next = getRule(rule);
				next.setNominalValue(attr, val);

				val++;
				if (val >= valCount) {
					attr++;
					while (attr < attrCount
							&& ((!usedAttribute[attr])  || rule.getNominalValue(attr) >= 0))
						attr++;
					val = 0;
					if (attr < attrCount)
							valCount = allData.numDistinctValues(attr);
				}
				return true;
			}

		};
	}

	/**
	 * Predicts the class memberships for a given instance. 
	 *
	 * @param instance the instance to be classified
	 * @return an array containing the estimated membership 
	 * probabilities of the test instance in each class 
	 * @exception Exception if distribution could not be 
	 * computed successfully
	 */
	public double[] distributionForInstance(Instance instance) throws Exception {
		double[] distribution = new double[classCount];
		double sum = 0;
		try {
			boolean found = false;
			int lastClass = -1;
			for (int i = 0; i < ruleset.size(); i++) {
				SimpleConjunctiveRule rule = (SimpleConjunctiveRule) ruleset.elementAt(i);
				//if (found && rule.getClassIndex()==lastClass) continue;
				if (rule.covers(instance)) {
					if (ordered) return rule.getRelativeDistribution();
					found = true;
					//lastClass=rule.getClassIndex();
					addDistribution(distribution, rule.getDistribution());
					sum += rule.getDistributionSum();
				}
			}
			if (!found) {
				return defaultRule.getRelativeDistribution();
			} else {
				for (int i = 0; i < classCount; i++) {
					distribution[i] /= sum;
				}
				return distribution;
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}

		return defaultRule.getRelativeDistribution();
	}
	/**
	 * Add an distribution to another.
	 * @param dist the common distribution
	 * @param newDist the new distribution to add
	 */
	public void addDistribution(double[] dist, double[] newDist) {
		for (int i = 0; i < classCount; i++) {
			dist[i] += newDist[i];
		}
	}

	/**
	 * Returns an enumeration of all the available options.
	 *
	 * @return an enumeration of all available options.
	 */
	public Enumeration listOptions() {
		Vector newVector = new Vector(3);
		newVector.addElement(new Option("\tSet the number of rules\n"
				+ "\tused for beam search\n" + "\t(default 2)", "B", 1,
				"-B <number of rules>"));
		
		newVector.addElement(new Option("\tSet the confidence probability\n"
				+ "\tto recognize a significant rule.\n0.9, 0.95 and 0.99 are supported.\n" 
				+ "\t(default 0.95)", "C", 1,
				"-C <confidence probability>"));

		newVector.addElement(new Option("\tSet whether class ordered\n"
				+ "\trules are used (Default: false)", "O", 0, "-O"));

		return newVector.elements();
	}

	/**
	 * Parses a given list of options.
	 *
	 * @param options the list of options as an array of strings
	 * @exception Exception if an option is not supported
	 */
	public void setOptions(String[] options) throws Exception {
		String numBeamString = Utils.getOption('B', options);
		if (numBeamString.length() != 0)
			beamCount = Integer.parseInt(numBeamString);
		else
			beamCount = 2;

		String numConfString = Utils.getOption('C', options);
		if (numConfString.length() != 0){
			double prob = Double.parseDouble(numConfString);
			if (prob>0.899999 && prob <0.900001){
				confProb=0.9;
				confidence=CONFIDENCE_0_9;
			}else if (prob>0.949999 && prob <0.950001){
				confProb=0.95;
				confidence=CONFIDENCE_0_95;
			}else if (prob>0.989999 && prob <0.990001){
				confProb=0.99;
				confidence=CONFIDENCE_0_99;
			}
		}
		
		else{
			confProb=0.95;
			confidence=CONFIDENCE_0_95;
		}

		ordered = Utils.getFlag('O', options);
	}

	/**
	 * Gets the current option settings.
	 *
	 * @return the list of current option settings as an array of strings
	 */
	public String[] getOptions() {
		String[] options = new String[7];
		DecimalFormat format2 = new DecimalFormat("0.0##",
				new DecimalFormatSymbols(Locale.US));
		int current = 0;
		options[current++] = "-B";
		options[current++] = "" + beamCount;
		options[current++] = "-C";
		options[current++] = "" + format2.format(confProb);

		if (!ordered) options[current++] = "-O";

		while (current < options.length)
			options[current++] = "";

		return options;
	}
	
	/**
	 * Return an string representation of the induced rules.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		Enumeration enum1 = ruleset.elements();
		while (enum1.hasMoreElements()) {
			SimpleConjunctiveRule rule = (SimpleConjunctiveRule) enum1.nextElement();
			sb.append(rule.toString());
			sb.append("\n");
		}
		sb.append(defaultRule.toString());
		sb.append("\n");
		return sb.toString();
	}
	
	/**
	 * Main method.
	 *
	 * @param args the options for the classifier
	 */
	public static void main(String[] args) {
		try {
			System.out.println(Evaluation.evaluateModel(new CN2(), args));
		} catch (Exception e) {
			e.printStackTrace();
			System.err.println(e.getMessage());
		}
	}

}