package weka.classifiers.rules;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;
import java.util.Vector;

import utils.indexstructure.Distance;
import utils.indexstructure.DoubleFunction;
import utils.indexstructure.MTree;
import utils.stats.ClassifyTTest;
import weka.classifiers.CaseBasedAlgorithm;
import weka.classifiers.Evaluation;
import weka.classifiers.OnLineClassifier;
import weka.classifiers.RuleBasedAlgorithm;
import weka.classifiers.rules.rise.SimpleConjunctiveRule;
import weka.core.Attribute;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.NormDistance;
import weka.core.Option;
import weka.core.Utils;

/**
 * @author beringer
 * 
 */
// splitstrategie gendert
// Regel nur lschen, wenn beste splitregel besser, sonst nicht splitten.
public class CopyOfRBLDS6 extends OnLineClassifier implements RuleBasedAlgorithm,
        CaseBasedAlgorithm {
    /** Flag if DEBUG output should be produced */
    protected static final boolean DEBUG = false;

    /** Flag if Verify conditions should be proved */
    protected static final boolean VERIFY = false;

    /** Flag if Time should be measured */
    protected static final boolean TIME = false;

    /** double distance which should not be equal */
    protected static final double EPS = Utils.SMALL * 2;

    /** Constant for the next InstanceINFO ID */
    protected static long nextInstID = Long.MIN_VALUE + 1;

    /**
     * Local class to store an Instance with additional intern informations.
     * 
     * @author George
     */
    protected class InstanceInfo implements Serializable {
        /** ID of this Instance */
        public long instID;

        /** wrapped Instance */
        public Instance inst;

        /** initialized Rule which contains only this Instance */
        public Rule ownRule;

        /** List of rules which cover this Instance */
        public List<Rule> rules = new ArrayList();

        /** Flag if this Instance is deleted from model */
        public boolean deleted = false;

        /** ID of the last queried which computed the queryValue. */
        public long queryID;

        /** Variable to execute queries faster. */
        public double queryValue;

        /**
         * Simple Constructor
         * 
         * @param inst
         *            wrapped Instance
         */
        public InstanceInfo(Instance inst) {
            this.inst = inst;
            instID = nextInstID++;
        }

        /**
         * Create new Instance for this dataset
         * 
         * @param attrCount
         *            number of attributes
         * @param header
         *            header of dataset
         */
        public InstanceInfo(int attrCount, Instances header) {
            instID = nextInstID++;
            this.inst = new Instance(1.0, new double[attrCount]);
            this.inst.setDataset(header);
            if (header == null)
                System.out.println("Error null!!!");
        }

        /**
         * Compares the InstanceInfo with the instID.
         */
        public boolean equals(Object ob) {
            InstanceInfo inf = (InstanceInfo) ob;
            return inf.instID == instID;
        }

        /**
         * Get a string representation of this Instance.
         */
        public String toString() {
            if (inst != null)
                return inst.toString();
            else
                return "empty";
        }
    }

    /**
     * Small class to store a rule with a coresponding Instance for rule
     * reduction.
     * 
     * @author George
     * 
     */
    protected static class Combination implements Serializable {
        /** rule to reduce */
        Rule rule;

        /** corresponding instance */
        InstanceInfo instance;

        /** attribute to handle for reduction */
        int attribute;

        /** inner distance of rule and instance */
        double distance;

        /**
         * Create a object.
         * 
         * @param rule
         * @param instance
         * @param attribute
         * @param distance
         */
        public Combination(Rule rule, InstanceInfo instance, int attribute,
                double distance) {
            this.rule = rule;
            this.instance = instance;
            this.attribute = attribute;
            this.distance = distance;
        }
    }

    /**
     * An extended Rule class. This class support the methods to grow, shrink,
     * handles weighted instances.
     */
    public static class Rule extends SimpleConjunctiveRule {

        /** ID for the next Rule */
        public static long nextID = Long.MIN_VALUE + 1;

        /** Array of used attributes to learn. */
        protected int[] usedAttribute;

        /** Flag if induction of rule is finished */
        protected boolean finished = false;

        /** List of covered instances */
        protected List<InstanceInfo> coveredInstInfo;

        /** ID of this rule */
        public long ruleID;

        /** flag if this rule was deleted from model */
        public boolean deleted = false;

        public boolean lastChanged = true;

        public double maxNearest = Double.MAX_VALUE;

        public double maxInnerNearest = Double.MAX_VALUE;

        public ClassifyTTest test;

        /**
         * Create an empty rule.
         * 
         * @param instances
         *            Header of data.
         * @param usedAttribute
         *            Array of used attributes.
         * @param store
         *            If covered elements should be stored.
         */
        public Rule(Instances instances, int[] usedAttribute) {
            super(instances, false);
            this.usedAttribute = usedAttribute;
            ruleID = nextID++;
            coveredInstInfo = new ArrayList();
            test = new ClassifyTTest(3, 4, 50, 10);
        }

        /**
         * Create a simple rule for the instance.
         * 
         * @param instances
         *            Header of data.
         * @param instance
         *            Instance for this simple rule.
         * @param usedAttribute
         *            Array of used attributes.
         * @param store
         *            If covered elements should be stored.
         */
        public Rule(Instances instances, InstanceInfo instance,
                int[] usedAttribute) {
            super(instances, false);
            this.usedAttribute = usedAttribute;
            ruleID = nextID++;
            coveredInstInfo = new ArrayList();
            for (int i = 0; i < attrCount; i++) {
                if (usedAttribute[i] > 0 && !instance.inst.isMissing(i)) {
                    if (usedAttribute[i] == NOMINAL)
                        setNominalValue(i, (int) instance.inst.value(i));
                    else if (usedAttribute[i] == NUMERIC) {
                        double val = instance.inst.value(i);
                        setNumericRange(i, val, val);
                    }
                }
            }
            addDistribution(instance);
            computeClass();
            test = new ClassifyTTest(3, 4, 50, 10);
            test.addClassify(1f);
        }

        /**
         * Copyconstructor
         * 
         * @param rule
         *            Rule to copy
         */
        public Rule(Rule rule) {
            super(rule);
            storeCovered = rule.storeCovered;
            this.usedAttribute = rule.usedAttribute;
            coveredInstInfo = new ArrayList();
            test = new ClassifyTTest(rule.test);
        }

        public boolean equals(Object o) {
            return this == o;
        }

        public void updateWeight(InstanceInfo inf, double old) {
            double w = inf.inst.weight();
            double dif = old - w;
            addClass((int) inf.inst.classValue(), -dif);

        }

        /*
         * public boolean equals(Object ob){ if (ob==this) return true; else
         * return false; }
         */

        /**
         * Compute the distribution with instances of the header object.
         */
        public void computeDistribution() {
            resetDistribution();
            Iterator it = coveredInstInfo.iterator();
            while (it.hasNext()) {
                InstanceInfo instance = (InstanceInfo) it.next();
                addDistribution(instance);
            }
        }

        public void computeDistribution(List coveredInstInfo) {
            computeDistribution(coveredInstInfo.iterator());
        }

        public void computeDistribution(Iterator it) {
            resetDistribution();
            coveredInstInfo.clear();
            while (it.hasNext()) {
                InstanceInfo instance = (InstanceInfo) it.next();
                addDistribution(instance);
            }
        }

        /**
         * Add an instance to the distribution.
         * 
         * @param inst
         *            Instance to add
         */
        public void addDistribution(InstanceInfo inst) {
            if (covers(inst.inst)) {
                if (coveredInstInfo.contains(inst)) {
                    // System.out.println("hm");
                    return;
                }
                addClass((int) inst.inst.classValue(), inst.inst.weight());
                coveredInstInfo.add(inst);
            }
        }

        /**
         * Remove an instance to the distribution.
         * 
         * @param inst
         *            Instance to add
         */
        public void removeDistribution(InstanceInfo inst) {
            if (covers(inst.inst) && remove(coveredInstInfo, inst)) {
                addClass((int) inst.inst.classValue(), -inst.inst.weight());
            }
        }

        /**
         * Expand the rule with this Instance.
         * 
         * @param instance
         *            Instance to expand this rule.
         */
        public void grow(Instance instance) {
            try {
                for (int i = 0; i < attrCount; i++) {
                    if (usedAttribute[i] > 0) {
                        if (usedAttribute[i] == NOMINAL) {
                            if (values[i] >= 0) {
                                if (instance.isMissing(i))
                                    setNominalValue(i, -1);
                                else if (values[i] != (int) instance.value(i))
                                    setNominalValue(i, -1);
                            }
                        } else if (usedAttribute[i] == NUMERIC) {
                            if (bounds[i] != null) {
                                if (instance.isMissing(i))
                                    resetNumericRange(i);
                                else {
                                    double val = instance.value(i);
                                    if (Double.isNaN(val))
                                        continue;
                                    if (val < bounds[i][0])
                                        setNumericRange(i, val, bounds[i][1]);
                                    else if (val > bounds[i][1])
                                        setNumericRange(i, bounds[i][0], val);
                                }
                            }
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        /**
         * Expand the rule with the other Rule.
         * 
         * @param instance
         *            Instance to expand this rule.
         */
        public void grow(Rule rule) {
            try {
                for (int i = 0; i < attrCount; i++) {
                    if (usedAttribute[i] > 0) {
                        if (usedAttribute[i] == NOMINAL) {
                            if (values[i] >= 0) {
                                if (rule.values[i] != values[i])
                                    setNominalValue(i, -1);
                            }
                        } else if (usedAttribute[i] == NUMERIC) {
                            if (bounds[i] != null) {
                                double[] val = rule.bounds[i];
                                if (val == null) {
                                    this.resetNumericRange(i);
                                } else {
                                    if (val[0] < bounds[i][0])
                                        setNumericRange(i, val[0], bounds[i][1]);
                                    if (val != null && val[1] > bounds[i][1])
                                        setNumericRange(i, bounds[i][0], val[1]);
                                }
                            }
                        }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        public boolean computeBoundsFromCovered() {
            int[] newvalues = new int[attrCount];
            double[][] newbounds = new double[attrCount][];
            for (int i = 0; i < attrCount; i++) {
                newvalues[i] = -2;
            }
            Iterator it = coveredInstInfo.iterator();
            int count = 0;
            while (it.hasNext()) {
                InstanceInfo instance = (InstanceInfo) it.next();
                if (getClassValue() == instance.inst.classValue()
                /* && !instance.deleted */) {
                    count++;
                    for (int i = 0; i < attrCount; i++) {
                        if (usedAttribute[i] == NOMINAL) {
                            if (newvalues[i] >= 0) {
                                if (newvalues[i] != (int) instance.inst
                                        .value(i))
                                    newvalues[i] = -1;
                            } else if (newvalues[i] == -2) {
                                newvalues[i] = (int) instance.inst.value(i);
                            }
                        } else if (usedAttribute[i] == NUMERIC) {
                            if (newbounds[i] != null) {
                                double val = instance.inst.value(i);
                                if (Double.isNaN(val))
                                    continue;
                                if (val < newbounds[i][0])
                                    newbounds[i] = new double[] { val,
                                            newbounds[i][1] };
                                else if (val > newbounds[i][1])
                                    newbounds[i] = new double[] {
                                            newbounds[i][0], val };
                            } else {
                                double val = instance.inst.value(i);
                                newbounds[i] = new double[] { val, val };
                            }
                        }
                    }
                }
            }

            for (int i = 0; i < attrCount; i++) {
                if (usedAttribute[i] == NOMINAL) {
                    setNominalValue(i, newvalues[i]);
                } else if (usedAttribute[i] == NUMERIC) {
                    if (newbounds[i] != null) {
                        setNumericRange(i, newbounds[i][0], newbounds[i][1]);
                    } else {
                        resetNumericRange(i);
                    }
                }
            }

            it = coveredInstInfo.iterator();
            while (it.hasNext()) {
                InstanceInfo instance = (InstanceInfo) it.next();
                if (getClassValue() != instance.inst.classValue()) {
                    if (!covers(instance.inst) || count == 0) {
                        addClass((int) instance.inst.classValue(),
                                -instance.inst.weight());
                        it.remove();
                        remove(instance.rules, this);
                    }
                }
            }

            return (count > 0);

        }

        protected boolean isFinished() {
            return finished;
        }

        protected void setFinished(boolean finished) {
            this.finished = finished;
        }
    }

    public static Comparator ruleSizeComparator = new Comparator() {
        public int compare(Object c1, Object c2) {
            double d = ((Rule) c1).getDistributionSum()
                    - ((Rule) c2).getDistributionSum();
            if (d < 0)
                return -1;
            else if (d > 0)
                return +1;
            else
                return 0;
        }
    };

    public static Comparator getInstanceTimeComparator(final int timeIndex) {
        return new Comparator() {
            public int compare(Object c1, Object c2) {
                double d = ((InstanceInfo) c1).inst.value(timeIndex)
                        - ((InstanceInfo) c2).inst.value(timeIndex);
                if (d < 0)
                    return -1;
                else if (d > 0)
                    return +1;
                else {
                    d = ((InstanceInfo) c1).instID - ((InstanceInfo) c2).instID;
                    if (d < 0)
                        return -1;
                    else if (d > 0)
                        return +1;
                    else
                        return 0;
                }
            }
        };
    }

    public Comparator ruleLaPlaceComparator = new Comparator() {
        public int compare(Object c1, Object c2) {
            double d = criterionLaPlace((Rule) c1)
                    - criterionLaPlace((Rule) c2);
            if (d < 0)
                return -1;
            else if (d > 0)
                return +1;
            else
                return 0;
        }
    };

    public static int NOMINAL = NormDistance.NOMINAL;

    public static int NUMERIC = NormDistance.NUMERIC;

    public Comparator ruleCriterionComparator = new Comparator() {

        public int compare(Object candidate1, Object candidate2) {
            Rule rule1 = (Rule) candidate1;
            Rule rule2 = (Rule) candidate2;
            if (criterion(rule1) > criterion(rule2))
                return 1;
            else if (criterion(rule1) < criterion(rule2))
                return -1;
            else
                return 0;
        }
    };

    public Comparator ruleCoveredSizeComparator = new Comparator() {

        public int compare(Object candidate1, Object candidate2) {
            Rule rule1 = (Rule) candidate1;
            Rule rule2 = (Rule) candidate2;
            if (rule1.coveredInstInfo.size() > rule2.coveredInstInfo.size())
                return 1;
            else if (rule1.coveredInstInfo.size() < rule2.coveredInstInfo
                    .size())
                return -1;
            else
                return 0;
        }
    };

    public Distance distanceXXL = new Distance() {

        public double distance(Object object1, Object object2) {
            InstanceInfo inf1 = (InstanceInfo) object1;
            InstanceInfo inf2 = (InstanceInfo) object2;
            return CopyOfRBLDS6.this.distance(inf1.inst, inf2.inst);
        }
    };

    protected DoubleFunction<InstanceInfo> timeFunc = new DoubleFunction<InstanceInfo>() {
        public double invoke(InstanceInfo inf) {
            return inf.inst.value(timeIndex);
        }
    };

    protected Instances header;

    protected int attrCount;

    protected int classIndex;

    protected int classCount;

    protected int timeIndex = 0;

    protected MTree<InstanceInfo> mtree;

    // protected BTree btree;
    protected TreeSet tree;

    protected NormDistance distance;

    protected NormDistance distance2;

    protected int[] usedAttribute;

    protected List usedAttributeList;

    protected List<InstanceInfo> nearest;

    protected double quantile = 0.05;

    protected int numeric = NormDistance.RANGE_NUMERIC;// QUANTILE_NUMERIC;

    protected boolean useVDM = true;

    protected boolean globalStop = false;

    protected int histlen = 100;

    protected int lastHist = 20;

    protected int maxRuleSize = 50;

    /** Choose rule with highest Laplace accuracy */
    public static final int HIGHEST_LAPLACE_TIEBREAK = 0;

    /** Choose rule with highest Laplace accuracy */
    public static final int DISTANCE_TIEBREAK = 3;

    /** Choose rule with highest Laplace accuracy */
    public static final int DISTANCE_WEIGHED_TIEBREAK = 4;

    /** Choose most specific rule */
    public static final int MOST_SPECIFIC_TIEBREAK = 1;

    /** Use frequency-based voting */
    public static final int FREQUENCY_BASED_VOTING_TIEBREAK = 2;

    protected int tieBreak = DISTANCE_TIEBREAK;

    protected boolean simplification = true;

    protected List<Rule> rules;

    protected boolean[] importantRules;

    protected int lastQueryID;

    protected InstanceInfo queryInstance;

    protected int kUpdateValue = 1;

    protected int k_step = 5;

    protected int near = 1;

    protected int nearInner = 1;

    protected int maxSize = 10000;

    protected long updates = 0;

    protected List toDelete = new ArrayList();

    protected List<Rule> updatedRules = new ArrayList();

    protected List<Rule[]> unionCand = new ArrayList();
    
    protected List<Rule> checkCand = new ArrayList();

    /** */
    protected int innerAttribute;

    /** List of new rules */
    protected List newRules = new ArrayList();

    protected List splitRules;

    protected List noRule;

    /** list of new instances to insert */
    protected List newInstances;

    protected boolean init = false;

    protected double minProbDiff = 0.2;

    public ClassifyTTest test;

    protected Comparator instanceTimeComparator;

    protected int maxInstances = 1000;

    protected int maxRules = 100;

    protected boolean improve = false;

    protected int k_redundanz = 10;

    protected int k_noise = 10;

    protected double p_current = 0.5;

    protected double p_young = 0.25;

    public CopyOfRBLDS6() {

    }

    /**
     * 
     * @param maxInstances
     *            maximale Gre der Fallbasis (-1 falls unbegrenzt)
     * @param maxRules
     *            maximale Anzahl Regeln (-1 falls unbegrenzt)
     * @param improve
     *            uninteressant... immer false
     * @param histlen
     *            Lnge der Klassifikationshistorie.
     * @param lastHist
     *            Anzahl Klassifikationen zur Schtzung des aktuellen Fehlers
     *            nach Shift.
     * @param k_step
     *            Anzahl Trainingsdaten fr ein komplettes Update.
     * @param maxRuleSize
     *            Maximale Anzahl von Instanzen in einer Regel (falls mglich).
     * @param k_redundanz
     *            Fr Redundanz notwendige einheitliche Nachbarschaft.
     * @param k_noise
     *            Mindestgre einer Regel, damit berdeckter Noise erkannt
     *            werden kann.
     * @param k_union
     *            Anzahl Regel-Updates bis zum nchsten unionRules()-Aufruf. Bei
     *            -1 wird k_union=Anzahl Attribute gesetzt.
     * @param p_current
     *            Anteil Instanzen einer Regel, die zu den aktuellen gehren.
     *            Wertebereich: [0,1]
     * @param p_young
     *            Anteil jngster Trainingsdaten, die auch ohne Regel
     *            gespeichert werden. Wertebereich: [0,1]
     */
    public CopyOfRBLDS6(int maxInstances, int maxRules, boolean improve, int histlen,
            int lastHist, int k_step, int maxRuleSize, int k_redundanz,
            int k_noise, double p_current, double p_young) {
        this.maxInstances = maxInstances;
        this.maxRules = maxRules;
        this.improve = improve;
        this.histlen = histlen;
        this.lastHist = lastHist;
        this.k_step = k_step;
        this.maxRuleSize = maxRuleSize;
        this.k_redundanz = k_redundanz;
        this.k_noise = k_noise;
        this.p_current = p_current;
        this.p_young = p_young;
    }

    /**
     * Compute model for the initialize training instances.
     */
    public void buildClassifier(Instances data) throws Exception {
        init = true;
        header = new Instances(data, 0);
        attrCount = data.numAttributes();
        classIndex = data.classIndex();
        classCount = data.numClasses();
        computeUsedAttributes();
        newInstances = new ArrayList();
        noRule = new ArrayList();
        instanceTimeComparator = getInstanceTimeComparator(timeIndex);
        if (numeric == NormDistance.QUANTILE_NUMERIC)
            distance = new NormDistance(data, usedAttribute, quantile, useVDM);
        else
            distance = new NormDistance(data, usedAttribute,
                    numeric == NormDistance.RANGE_NUMERIC, useVDM);
        distance2 = (NormDistance) distance.clone();
        mtree = new MTree(distanceXXL, 6, 15);
        // btree=new BTree();
        // btree.initialize(getTimeDescriptor(timeIndex), new
        // MapContainer(),15,50);
        tree = new TreeSet(instanceTimeComparator);
        lastQueryID = Integer.MIN_VALUE;
        queryInstance = new InstanceInfo(null);
        test = new ClassifyTTest(3, 4, histlen, lastHist);

        Enumeration enum1 = data.enumerateInstances();
        while (enum1.hasMoreElements()) {
            Instance inst = (Instance) enum1.nextElement();
            Instance cp = (Instance) inst.copy();
            cp.setDataset(header);
            insert(new InstanceInfo(cp));
        }

        initializeRules();
        induceRules();
        unionRules();
        induceRules();
        if (VERIFY) {
            verifyInstances();
            verifyRules();
            verifyCriterion();
        }
        init = false;
    }

    /**
     * Set the default instance based ruleset and compute the next rule of each
     * element.
     * 
     */
    protected void initializeRules() {
        rules = new ArrayList();
        nearest = new ArrayList();
        Iterator it = tree.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            Instance inst = instinf.inst;
            for (int i = 0; i < attrCount; i++) {
                if (usedAttribute[i] > 0 && inst.isMissing(i))
                    continue;
            }
            boolean found = false;
            for (int i = 0; i < rules.size(); i++) {
                Rule rule = (Rule) rules.get(i);
                if (rule.covers(inst)) {
                    if (instinf.rules.isEmpty())
                        remove(noRule, instinf);
                    instinf.rules.add(rule);
                    rule.addDistribution(instinf);
                    rule.computeClass();
                    found = true;
                    break;
                }
            }
            if (!found) {
                Rule rule = new Rule(header, instinf, usedAttribute);
                instinf.ownRule = rule;
                if (instinf.rules.isEmpty())
                    remove(noRule, instinf);
                instinf.rules.add(rule);
                rules.add(rule);
            }
        }
    }

    /**
     * Compute distribution for the rule.
     * 
     * @param rule
     */
    protected void updateDistribution(final Rule rule) {
        rule.computeDistribution(tree.iterator());

        // final DoubleFunction<MTree<InstanceInfo>.Sphere> dFunc = new
        // DoubleFunction<MTree<InstanceInfo>.Sphere>() {
        // public double invoke(MTree<InstanceInfo>.Sphere s) {
        // return distance(rule, ((InstanceInfo) (s.center)).inst)
        // - s.radius;
        // }
        // };
        // Iterator<InstanceInfo> it = mtree.query(dFunc);
        // List<InstanceInfo> cov = new ArrayList();
        // while (it.hasNext()) {
        // InstanceInfo inf = it.next();
        // if (distance(rule, inf.inst) > 0)
        // break;
        // cov.add(inf);
        // }
        // rule.computeDistribution(cov.iterator());
    }

    /**
     * Compute distribution for the list of nearest Rules.
     * 
     * @param nearestRules
     * @return
     */
    protected double[] getDistribution(List<Rule> nearestRules, InstanceInfo inf) {
        if (nearestRules.size() == 1)
            return nearestRules.get(0).getRelativeDistribution();
        double[] distribution = new double[classCount];
        switch (tieBreak) {
        case DISTANCE_WEIGHED_TIEBREAK:
            double sum = 0;
            for (Rule rule : nearestRules) {
                double val = 1.0 / innerDistance(rule, inf.inst);
                addDistribution(distribution, rule.getRelativeDistribution(),
                        val);
                sum += val;
            }
            for (int i = 0; i < classCount; i++) {
                distribution[i] = distribution[i] / sum;
            }
            break;

        case DISTANCE_TIEBREAK:
            double best = Double.MAX_VALUE;
            for (Rule rule : nearestRules) {
                // double val = ruleRatio(rule, inf.inst);
                double val = innerDistance(rule, inf.inst);

                if (val < best) {
                    best = val;
                    distribution = rule.getRelativeDistribution();
                }
            }
            break;
        case HIGHEST_LAPLACE_TIEBREAK:
            Collections.sort(nearestRules, ruleCriterionComparator);
            int pos = nearestRules.size() - 1;
            if (pos < 0)
                return null;
            Rule rule = (Rule) nearestRules.get(pos);
            double crit = criterionLaPlace(rule);
            distribution = rule.getRelativeDistribution();
            int count = 1;
            double newcrit = 0;
            do {
                pos--;
                if (pos >= 0) {
                    rule = (Rule) nearestRules.get(pos);
                    newcrit = criterion(rule);
                    if (newcrit == crit) {
                        addDistribution(distribution, rule
                                .getRelativeDistribution(), 1);
                        count++;
                    }
                }
            } while (newcrit == crit && pos > 0);
            if (count > 1) {
                for (int i = 0; i < classCount; i++) {
                    distribution[i] /= count;
                }
            }
            break;
        case FREQUENCY_BASED_VOTING_TIEBREAK:
            Iterator it = nearestRules.iterator();
            count = 0;
            while (it.hasNext()) {
                rule = (Rule) it.next();
                addDistribution(distribution, rule.getRelativeDistribution(), 1);
                count++;
            }
            if (count > 1) {
                for (int i = 0; i < classCount; i++) {
                    distribution[i] /= count;
                }
            }
            break;
        case MOST_SPECIFIC_TIEBREAK:
            Collections.sort(nearestRules, ruleCoveredSizeComparator);
            pos = nearestRules.size() - 1;
            if (pos < 0)
                return null;
            rule = (Rule) nearestRules.get(pos);
            crit = rule.coveredInstInfo.size();
            distribution = rule.getRelativeDistribution();
            count = 1;
            newcrit = 0;
            do {
                pos--;
                if (pos >= 0) {
                    rule = (Rule) nearestRules.get(pos);
                    newcrit = rule.coveredInstInfo.size();
                    if (newcrit == crit) {
                        addDistribution(distribution, rule
                                .getRelativeDistribution(), 1);
                        count++;
                    }
                }
            } while (newcrit == crit && pos > 0);
            if (count > 1) {
                for (int i = 0; i < classCount; i++) {
                    distribution[i] /= count;
                }
            }
            break;
        }
        return distribution;
    }

    /**
     * Compute index of the maximal value in the array.
     * 
     * @param doubles
     *            array of values
     * @return
     */
    protected int maxIndex(double[] doubles) {
        double max1 = -1;
        double max2 = -1;
        int maxIndex = 0;

        for (int i = 0; i < doubles.length; i++) {
            if ((i == 0) || (doubles[i] > max1)) {
                maxIndex = i;
                max2 = max1;
                max1 = doubles[i];
            } else if (doubles[i] > max2) {
                max2 = doubles[i];
            }
        }
        if (max1 > max2)
            return maxIndex;
        else
            return -maxIndex - 1;
    }

    /**
     * Compute the most often dlass in the list of the nearest rules.
     * 
     * @param nRules
     *            list of nearest Rules
     * @return
     */
    protected int getMaxClass(List nRules, InstanceInfo inf) {
        double[] distrib = getDistribution(nRules, inf);
        if (distrib == null) {
            System.out.println("ERROR");
            getDistribution(nRules, inf);
        }
        int clVal = maxIndex(distrib);
        return clVal;
    }

    /**
     * Find nearest rules to classifiy the instance
     * 
     * @param instinf
     * @return
     */
    protected List findNearestRules(InstanceInfo instinf) {
        double minDist = Double.MAX_VALUE;
        Iterator it = rules.iterator();
        List nRules = new ArrayList();
        while (it.hasNext()) {
            Rule rule = (Rule) it.next();

            if (instinf.ownRule == rule) {
                if (rule.coveredInstInfo.size() == 1)
                    continue;
                else
                    instinf.ownRule = null;
            }
            if (rule.getDistribution()[rule.getClassValue()] < 3) {
                continue;
            }
            double dist = distance(rule, instinf.inst);
            if (Double.isNaN(dist) && DEBUG) {
                System.out.println("NAN DISTANCE!!!");
            }
            if (minDist > dist) {
                minDist = dist;
                nRules.clear();
                nRules.add(rule);
            } else if (dist == minDist) {
                nRules.add(rule);
            }
        }
        if (nRules.size() == 0) {
            if (DEBUG)
                System.out.println("NO RULE FOUND!!!");
        }
        return nRules;
    }

    /**
     * Try to union rules of same class which overlaps.
     * 
     * @return
     */
    protected void unionRules() {
        for (int i = 0; i < rules.size() - 1; i++) {
            Rule r1 = (Rule) rules.get(i);
            if (r1.deleted)
                continue;
            double r1crit = criterionLaPlace(r1);
            for (int j = i + 1; j < rules.size(); j++) {
                Rule r2 = (Rule) rules.get(j);
                if (r2.deleted)
                    continue;
                if (r1.getClassValue() != r2.getClassValue())
                    continue;
                if (distance(r1, r2) == 0) {
                    double r2crit = criterionLaPlace(r2);
                    Rule r = union(r1, r2);
                    if (criterionLaPlace(r) >= 0.5 * (r1crit + r2crit)) {
                        double imp = 0;
                        // if (improve && rules.size() < 100)
                        // imp = improve(r, r1, r2);
                        if (imp >= 0) {
                            /* && !rules.contains(r) */

                            r2.deleted = true;
                            if (rules.get(j) == r2)
                                rules.remove(j);
                            else
                                remove(rules, r2);

                            r1.deleted = true;
                            if (rules.get(i) == r1)
                                rules.remove(i);
                            else
                                remove(rules, r1);

                            i--;
                            j = rules.size() + 3;
                            rules.add(r);

                            updateNearestRules(r1, r2, r);
                            // if (r.test.getLastError(20) >= 0.75) {
                            // OnLineRuleVisTest.markRule(r);
                            // OnLineRuleVisTest.markedRules.clear();
                            // }

                        }
                    }

                    // OnLineRuleVisTest.ruleData2.clear();
                }
            }
            // r1.lastChanged = false;
        }
        if (VERIFY) {
            verifyInstances();
            verifyRules();
            verifyCriterion();
        }
    }

    /**
     * Try to union rules of same class which overlaps.
     * 
     * @return
     */
    protected void unionRules(Rule r1) {
        // TODO union, induced, new rule in updateRules
        if (r1.deleted)
            return;
        double r1crit = criterionLaPlace(r1);
        for (int j = 0; j < rules.size(); j++) {

            Rule r2 = (Rule) rules.get(j);
            if (r2.deleted || r1 == r2)
                continue;
            if (r1.getClassValue() != r2.getClassValue())
                continue;
            if (distance(r1, r2) == 0) {
                double r2crit = criterionLaPlace(r2);
                Rule r = union(r1, r2);
                if (criterionLaPlace(r) >= 0.5 * (r1crit + r2crit)) {
                    double imp = 0;
                    // if (improve && rules.size() < 100)
                    // imp = improve(r, r1, r2);
                    if (imp >= 0) {
                        /* && !rules.contains(r) */

                        r2.deleted = true;
                        if (rules.get(j) == r2)
                            rules.remove(j);
                        else
                            remove(rules, r2);

                        r1.deleted = true;
                        remove(rules, r1);

                        rules.add(r);

                        updateNearestRules(r1, r2, r);
                        return;
                    }
                }

                // OnLineRuleVisTest.ruleData2.clear();
            }
        }
        // r1.lastChanged = false;

    }

    protected boolean containsRules(Rule r1, Rule r2) {
        for (Rule[] rr : unionCand) {
            if ((rr[0] == r1 && rr[1] == r2) || (rr[0] == r2 && rr[1] == r1))
                return true;
        }
        return false;
    }

    protected void addUnionRules(List<Rule> rulelist) {
        int size = rulelist.size();
        for (int i = 0; i < size - 1; i++) {
            Rule r1 = rulelist.get(i);
            if (r1.deleted)
                continue;
            int cls1=r1.getClassValue();
            for (int j = i + 1; j < size; j++) {
                Rule r2 = rulelist.get(j);
                if (r2.deleted || cls1!=r2.getClassValue())
                    continue;
                if (!containsRules(r1, r2))
                    unionCand.add(new Rule[] { r1, r2 });
            }
        }

    }
    
    

    protected void unionCandRules() {
        for (Rule[] rr : unionCand) {
            Rule r1 = rr[0];
            Rule r2 = rr[1];
            if (r1.deleted || r2.deleted)
                continue;
            double r1crit = criterionLaPlace(r1);
            double r2crit = criterionLaPlace(r2);
            Rule r = union(r1, r2);
            if (criterionLaPlace(r) >= 0.5 * (r1crit + r2crit)) {
                r2.deleted = true;
                remove(this.rules, r2);
                r1.deleted = true;
                remove(this.rules, r1);
                this.rules.add(r);
                this.newRules.add(r);
                updateNearestRules(r1, r2, r);
            }
        }
        unionCand.clear();
        if (VERIFY) {
            verifyInstances();
            verifyRules();
            verifyCriterion();
        }
    }

    /**
     * Union the 2 rules.
     * 
     * @param r1
     * @param r2
     * @return
     */
    protected Rule union(Rule r1, Rule r2) {
        Rule r = new Rule(r1);
        r.grow(r2);
        updateDistribution(r);
        return r;
    }

    // protected double improve(Rule newRule, Rule old1, Rule old2) {
    // int cls = newRule.getClassValue();
    // int all = newRule.coveredInstInfo.size();
    // int better = 0;
    // int worse = 0;
    // int good = 0;
    // int bad = 0;
    // for (InstanceInfo inf : newRule.coveredInstInfo) {
    // int icls = (int) inf.inst.classValue();
    // int ocls = classify(inf);
    // int ncls = -1;
    // if (inf.rules.size() > 0) {
    // List nRules = new ArrayList(inf.rules);
    // if (old1 != null)
    // remove(nRules, old1);
    // if (old2 != null)
    // remove(nRules, old2);
    // if (nRules.size() == 0) {
    // ncls = cls;
    // } else {
    // nRules.add(newRule);
    // double[] dist = getDistribution(nRules, inf);
    //
    // double max = 0;
    // int maxIndex = -1;
    //
    // for (int i = 0; i < dist.length; i++) {
    // if (dist[i] > max) {
    // maxIndex = i;
    // max = dist[i];
    // }
    // }
    // ncls = maxIndex;
    // }
    // } else
    // ncls = cls;
    // if (icls == ocls) {
    // if (icls == ncls)
    // good++;
    // else
    // worse++;
    // } else {
    // if (icls == ncls)
    // better++;
    // else
    // bad++;
    // }
    //
    // }
    // // return 0;
    // return (better - worse) / (double) all;
    // }

    /**
     * Induce rules for the dataset until no better ruleset is found.
     */
    protected void induceRules() {
        boolean changed = false;
        // if (VERIFY) {
        // verifyInstances();
        // verifyRules();
        // verifyCriterion();
        // }

        do {
            changed = false;
            Object[] lastRules = rules.toArray();
            Arrays.sort(lastRules, ruleCoveredSizeComparator);
            int size = lastRules.length;
            for (int i = 0; i < size; i++) {
                Rule rule = (Rule) lastRules[i];
                if (rule.deleted) {
                    remove(rules, rule);
                    continue;
                }
                if (rule.getClassValue() != rule.frequentClass()) {
                    remove(rules, rule);
                    rule.deleted = true;
                    Iterator el = rule.coveredInstInfo.iterator();
                    while (el.hasNext()) {
                        InstanceInfo inf = (InstanceInfo) el.next();
                        if (!inf.rules.isEmpty()) {
                            remove(inf.rules, rule);
                            if (inf.rules.isEmpty())
                                noRule.add(inf);
                        }

                    }
                }
                if (induceRule(rule) != rule)
                    changed = true;
            }
        } while (changed);
    }

    /**
     * Induce and reduce this rule until no improvement is found.
     * 
     * @param rule
     * @return
     */
    public Rule induceRule(Rule rule) {
        if (rule.coveredInstInfo.size() == 0)
            return rule;
        boolean localChanged = false;
        boolean changed = false;
        do {
            localChanged = false;
            if (rule.deleted)
                continue;
            if (!globalStop && rule.isFinished())
                continue;
            Rule bestNew = rule;
            double bestCrit = criterionLaPlace(rule);
            List nearest = findNearest(rule);
            if (nearest != null && nearest.size() > 0) {
                Iterator newRules = generalize(rule, nearest);
                while (newRules.hasNext()) {
                    Rule newRule = (Rule) newRules.next();
                    if (criterionLaPlace(newRule) >= bestCrit) {
                        bestCrit = criterionLaPlace(newRule);
                        bestNew = newRule;
                    }
                }
            }
            nearest = findInnerNearest(rule);
            if (nearest != null && nearest.size() > 0) {
                Iterator newRules = specialize(nearest);
                while (newRules.hasNext()) {
                    Rule newRule = (Rule) newRules.next();
                    if (criterionLaPlace(newRule) > bestCrit) {
                        bestCrit = criterionLaPlace(newRule);
                        if (VERIFY)
                            verifyCriterion(newRule);
                        bestNew = newRule;
                    }
                }
            }
            if (bestNew != rule) {
                changed = true;
                if (!rule.deleted) {
                    rule.deleted = true;
                    remove(rules, rule);
                    Iterator it = rule.coveredInstInfo.iterator();
                    while (it.hasNext()) {
                        InstanceInfo inf = (InstanceInfo) it.next();
                        if (!inf.rules.isEmpty()) {
                            remove(inf.rules, rule);
                            if (inf.rules.isEmpty())
                                noRule.add(inf);
                        }
                    }
                }
                if (!optimizeRule(bestNew)) {
                    localChanged = true;
                    rules.add(bestNew);
                    updateNearestRules(rule, bestNew);
                    rule = bestNew;
                } else {
                    bestNew.deleted = true;
                }
            }
        } while (localChanged);
        if (!rule.deleted)
            rule.setFinished(true);
        return rule;
    }

    /**
     * Get a list of all used training instances.
     */
    public List getTrainInstances() {
        LinkedList list = new LinkedList();
        Iterator it = tree.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            list.add(instinf.inst);
        }
        return list;
    }

    /**
     * Get the size of the current case base.
     */
    public int getTrainInstancesCount() {
        return tree.size();
    }

    /**
     * Remove all objects in the list which are in the remove list.
     * 
     * @param list
     * @param toRemove
     */
    protected void removeAll(List list, List toRemove) {
        Iterator it = toRemove.iterator();
        while (it.hasNext()) {
            remove(list, it.next());
        }
    }

    /**
     * remove this object in the list.
     * 
     * @param list
     * @param ob
     * @return if found and removed
     */
    protected static boolean remove(List list, Object ob) {
        Iterator it = list.iterator();
        while (it.hasNext()) {
            Object o = it.next();
            if (o == ob) {
                it.remove();
                return true;
            }
        }
        return false;
    }

    /**
     * Search the index of this object in the list (== and no equals)
     * 
     * @param list
     *            List
     * @param ob
     *            Object to search
     * @return index or -1 if not found
     */
    protected int indexOf(List list, Object ob) {
        Iterator it = list.iterator();
        int ind = 0;
        while (it.hasNext()) {
            if (it.next() == ob) {
                return ind;
            }
            ind++;
        }
        return -1;
    }

    /**
     * Check if the new rule is already in the global rulelist
     * 
     * @param newRule
     *            new rule to check
     * @return if newRule was found
     */
    protected boolean optimizeRule(Rule newRule) {
        int cl = newRule.getClassValue();
        Iterator it = rules.iterator();
        boolean delete = false;
        while (it.hasNext()) {
            Rule rule = (Rule) it.next();
            if (rule != newRule) {
                if (cl == rule.getClassValue()) {
                    if (rule.equals(newRule)) {
                        delete = true;
                        break;
                    }
                }
            }
        }
        return delete;
    }

    /**
     * Search rules covered by rules, so they not necessary.
     * 
     */
    protected void optimizeAll(List<Rule> rulelist) {
        for (int j = 0; j < rulelist.size() - 1; j++) {
            Rule rule1 = (Rule) rulelist.get(j);
            if (rule1.deleted)
                continue;
            for (int i = j + 1; i < rulelist.size(); i++) {
                Rule rule2 = (Rule) rulelist.get(i);
                if (rule1 == rule2 || rule2.deleted)
                    continue;
                if (rule1.getClassValue() == rule2.getClassValue()) {
                    if (rule1.contains(rule2)) {
                        if (optimize(rule1, rule2)) {
                            i--;
                        }
                    } else if (rule2.contains(rule1)) {
                        if (optimize(rule2, rule1)) {
                            j--;
                            i = rulelist.size();
                        }
                    }
                }
            }
        }
    }

    /**
     * If outer rule is better (LaPlace) than delete inner rule.
     * 
     * @param outer
     * @param inner
     * @param index
     *            index of inner rule in the global rulelist
     * @return if inner rule was deleted
     */
    protected boolean optimize(Rule outer, Rule inner) {
        if (simplification) {
            if (criterionLaPlace(outer) >= criterionLaPlace(inner)) {
                inner.deleted = true;
                remove(rules, inner);
                Iterator it = inner.coveredInstInfo.iterator();
                while (it.hasNext()) {
                    InstanceInfo inf = (InstanceInfo) it.next();
                    if (!inf.rules.isEmpty()) {
                        remove(inf.rules, inner);
                        if (inf.rules.isEmpty())
                            noRule.add(inf);
                    }
                }

                return true;
            } else if (improve) {
                outer.deleted = true;
                remove(rules, outer);
                Iterator it = outer.coveredInstInfo.iterator();
                while (it.hasNext()) {
                    InstanceInfo inf = (InstanceInfo) it.next();
                    if (!inf.rules.isEmpty()) {
                        remove(inf.rules, outer);
                        if (inf.rules.isEmpty())
                            noRule.add(inf);
                    }
                }

                return true;
            }
        } else if ((tieBreak == HIGHEST_LAPLACE_TIEBREAK)
                && criterionLaPlace(outer) > criterionLaPlace(inner)) {
            inner.deleted = true;
            remove(rules, inner);
            Iterator it = inner.coveredInstInfo.iterator();
            while (it.hasNext()) {
                InstanceInfo inf = (InstanceInfo) it.next();
                if (!inf.rules.isEmpty()) {
                    remove(inf.rules, inner);
                    if (inf.rules.isEmpty())
                        noRule.add(inf);
                }
            }
            return true;
        }
        return false;
    }

    /**
     * Update the rule list for the new rule.
     * 
     * @param oldRule
     *            old rule to replace
     * @param newRule
     *            new rule
     */
    protected void updateNearestRules(Rule oldRule, Rule newRule) {
        Iterator it = oldRule.coveredInstInfo.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            if (!instinf.rules.isEmpty()) {
                remove(instinf.rules, oldRule);
                if (instinf.rules.isEmpty())
                    noRule.add(instinf);
            }
        }
        it = newRule.coveredInstInfo.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            if (instinf.rules.size() == 0)
                remove(noRule, instinf);
            instinf.rules.add(newRule);
            if (VERIFY) {
                this.verifyNoRules();
            }
        }
    }

    /**
     * Remove old rules from instance rule lists and insert new rule to this
     * lists after union of this rules.
     * 
     * @param oldRule1
     *            first old rule
     * @param oldRule2
     *            second old rule
     * @param newRule
     *            new union rule
     */
    protected void updateNearestRules(Rule oldRule1, Rule oldRule2, Rule newRule) {
        Iterator it = oldRule1.coveredInstInfo.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            instinf.rules.remove(oldRule1);
        }
        it = oldRule2.coveredInstInfo.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            instinf.rules.remove(oldRule2);
        }
        it = newRule.coveredInstInfo.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            if (instinf.rules.isEmpty()) {
                remove(noRule, instinf);
            }
            instinf.rules.add(newRule);
        }
    }

    /**
     * Create a new Rule which generalize the given Rule with the given Instance
     * 
     * @param rule
     *            Rule to generalize
     * @param nearest
     *            Instance to grow the rule
     * @return new generalized rule
     */
    protected Iterator generalize(final Rule rule, List nearest) {
        final Iterator nextNear = nearest.iterator();
        return new Iterator() {
            Rule next = null;

            public void remove() {
            }

            public boolean hasNext() {
                if (next == null)
                    return computeNext();
                else
                    return true;
            }

            public Object next() {
                if (next == null)
                    computeNext();
                Object val = next;
                next = null;
                return val;
            }

            private boolean computeNext() {
                if (nextNear.hasNext()) {
                    InstanceInfo instinf = (InstanceInfo) nextNear.next();
                    Instance inst = instinf.inst;
                    next = new Rule(rule);
                    next.grow(inst);
                    updateDistribution(next);
                    return true;
                } else
                    return false;
            }
        };
    }

    /**
     * Create a new Rule which generalize the given Rule with each of the given
     * Instances
     * 
     * @param rule
     *            Rule to generalize
     * @param nearest
     *            List of Instances to grow the rule
     * @return Iterator of new specialized rules
     */
    protected Iterator specialize(List nearest) {
        final Iterator nextNear = nearest.iterator();
        return new Iterator() {

            Rule next = null;

            public void remove() {
            }

            public boolean hasNext() {
                if (next == null)
                    return computeNext();
                else
                    return true;
            }

            public Object next() {
                if (next == null)
                    computeNext();
                Object val = next;
                next = null;
                return val;
            }

            /**
             * Compute next Rule
             * 
             * @return
             */
            private boolean computeNext() {
                if (nextNear.hasNext()) {
                    Combination comb = (Combination) nextNear.next();
                    InstanceInfo instinf = comb.instance;
                    Instance inst = instinf.inst;
                    next = new Rule(comb.rule);
                    if (usedAttribute[comb.attribute] == NUMERIC) {
                        double val = inst.value(comb.attribute);
                        double[] range = next.getNumericRange(comb.attribute);
                        if (range != null) {
                            if ((val - range[0]) < (range[1] - val))
                                next.setNumericRange(comb.attribute, val + EPS,
                                        range[1]);
                            else
                                next.setNumericRange(comb.attribute, range[0],
                                        val - EPS);
                        } else
                            System.out.println("prob specialize");
                    } else {
                        next.setNominalValue(comb.attribute, (int) inst
                                .value(comb.attribute));
                    }
                    // updateDistribution(next);
                    next.computeDistribution(comb.rule.coveredInstInfo);
                    // next.computeBoundsFromCovered();
                    return true;
                } else
                    return false;
            }
        };
    }

    /**
     * Find the nearest Instance which is not covered of the given Rule
     * 
     * @param rule
     *            interested Rule
     * @return nearest not covered Instance
     */
    protected List findNearest(Rule rule) {
        nearest.clear();
        int cls = rule.getClassValue();
        Iterator it = tree.iterator();// mtree.query();
        double maxValue = Double.MAX_VALUE;
        Instance last = null;
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            Instance inst = instinf.inst;

            if (rule.covers(inst))
                continue;
            if (last != null && distance(inst, last) == 0.0
                    && inst.classValue() == last.classValue()) {
                continue;
            }
            last = inst;
            double dist = distance(rule, inst);
            if (dist <= maxValue && cls == (int) inst.classValue()) {
                instinf.queryValue = dist;
                maxValue = insert(nearest, instinf, near);
            }
        }
        return nearest;
    }

    /**
     * Insert Instance in the sorted list for search of nearest instances.
     * 
     * @param list
     * @param inf
     * @param maxSize
     * @return
     */
    protected double insert(List list, InstanceInfo inf, int maxSize) {
        double val = inf.queryValue;
        double max = Double.MAX_VALUE;
        if (list.size() >= maxSize) {
            max = ((InstanceInfo) list.get(maxSize - 1)).queryValue;
        }
        if (max < val)
            return max;
        int p = 0;
        while (p < list.size() && ((InstanceInfo) list.get(p)).queryValue < val) {
            p++;
        }
        if ((p < list.size())
                && equals(((InstanceInfo) list.get(p)).inst, inf.inst))
            return max;
        list.add(p, inf);
        if (max < Double.MAX_VALUE) {
            max = ((InstanceInfo) list.get(maxSize - 1)).queryValue;
            while (((InstanceInfo) list.get(list.size() - 1)).queryValue > max) {
                list.remove(list.size() - 1);
            }
        }
        return max;
    }

    /**
     * Find the nearest Instance which is covered of the given Rule
     * 
     * @param rule
     *            interested Rule
     * @return nearest not covered Instance
     */
    protected List findInnerNearest(Rule rule) {
        nearest.clear();
        double maxValue = 0;
        int cls = rule.getClassValue();
        Iterator it = rule.coveredInstInfo.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            Instance inst = instinf.inst;
            if (cls == (int) inst.classValue())
                continue;
            double dist = innerDistance(rule, inst);
            if (dist >= maxValue) {
                instinf.queryValue = dist;
                maxValue = insert(nearest, new Combination(rule, instinf,
                        innerAttribute, dist), nearInner);
            }
        }
        return nearest;
    }

    /**
     * find most similar rule which not covers the instance
     * 
     * @param rule
     *            rule to reduce
     * @param inf
     *            instance
     * @return the reduced rule
     */
    protected Rule reduce(Rule rule, InstanceInfo inf) {
        if (innerAttribute < 0)
            return null;
        Instance inst = inf.inst;
        Rule next = new Rule(rule);
        if (usedAttribute[innerAttribute] == NUMERIC) {
            double val = inst.value(innerAttribute);
            double[] range = next.getNumericRange(innerAttribute);
            if (range != null) {
                if ((val - range[0]) < (range[1] - val))
                    next.setNumericRange(innerAttribute, val + EPS, range[1]);
                else
                    next.setNumericRange(innerAttribute, range[0], val - EPS);
            } else
                System.out.println("prob specialize");
        } else {
            next.setNominalValue(innerAttribute, (int) inst
                    .value(innerAttribute));
        }
        next.computeDistribution(rule.coveredInstInfo);
        if (next.coveredInstInfo.size() == 0
                || rule.getClassValue() != next.frequentClass()) {
            return null;
        }
        if (!next.computeBoundsFromCovered() || next.covers(inst)) {
            return null;
        }
        return next;
    }

    /**
     * Insert Combination in the sorted listvfor search of nearest inner
     * instance.
     * 
     * @param list
     * @param comb
     * @param maxSize
     * @return
     */
    protected double insert(List list, Combination comb, int maxSize) {
        double val = comb.instance.queryValue;
        double max = Double.MAX_VALUE;
        if (list.size() >= maxSize) {
            max = ((InstanceInfo) ((Combination) list.get(maxSize - 1)).instance).queryValue;
        }
        if (max > val)
            return max;
        int p = 0;
        while (p < list.size()
                && ((InstanceInfo) ((Combination) list.get(p)).instance).queryValue > val) {
            list.add(p, comb);
        }
        if (max < Double.MAX_VALUE) {
            max = ((InstanceInfo) ((Combination) list.get(maxSize - 1)).instance).queryValue;
            while (((InstanceInfo) ((Combination) list.get(list.size() - 1)).instance).queryValue < max) {
                list.remove(list.size() - 1);
            }
        }
        return max;
    }

    /**
     * Update classifier with the new Instances.
     * 
     * @param data
     *            new Instance
     */
    public void updateClassifier(Instances data) {
        updateClassifier(data, true);
    }

    /**
     * Update classifier with the new Instances.
     * 
     * @param data
     *            new Instances
     * @param forceUpdate
     *            if update has to perform now.
     */
    public void updateClassifier(Instances data, boolean forceUpdate) {
        Enumeration enum1 = data.enumerateInstances();
        while (enum1.hasMoreElements()) {
            Instance inst = (Instance) ((Instance) enum1.nextElement()).copy();
            inst.setDataset(header);
            int cls = -1;
            try {
                cls = (int) classifyInstance(inst);
                test.addClassify(cls == (int) inst.classValue() ? 1f : 0f);
                // addClassify(lastCls);
            } catch (Exception e) {
                e.printStackTrace();
            }
            InstanceInfo newinfo = new InstanceInfo(inst);
            Iterator it = rules.iterator();
            while (it.hasNext()) {
                Rule rule = (Rule) it.next();
                if (rule.covers(inst)) {
                    newinfo.rules.add(rule);
                }
            }
            newInstances.add(newinfo);
            if (newinfo.rules.isEmpty())
                noRule.add(newinfo);
        }
        if (newInstances.size() >= k_step || forceUpdate)
            updateClassifier();
    }

    protected void significanceTest() {

        if ((1 - test.getCurrentError()) <= 1.0 / classCount
                || (test.isSignificant())) {
            double pDiff = test.getProbDiff();
            double[] mins = test.getMinValues();
            double pmin = mins[0];
            double p = test.getCurrentError();
            if ((p - pmin) > minProbDiff && pDiff > minProbDiff) {
                double pp = pDiff;
                if ((1 - p) <= 1.0 / classCount
                        || ((1 - pmin) - pDiff) <= 1.0 / classCount)
                    pDiff = 1.0;
                int toDel = Math.min((int) (tree.size() * pDiff), tree.size()
                        - rules.size());
                if (toDel > 0) {
                    List removed = mtree.remove(toDel, timeFunc);
                    // tree.removeAll(removed);
                    int rem = removed.size();
                    // Function map = new Function<InstanceInfo, Instance>() {
                    // public Instance invoke(InstanceInfo ob) {
                    // return ob.inst;
                    // }
                    // };
                    // distance2.removeInstances(new Mapper(map, removed
                    // .iterator()));

                    toDelete.addAll(removed);
                    doDelete();

                    if (DEBUG)
                        System.out.println("STATS: " + rem);
                    test.resetStats(pmin + pp);
                    if (pDiff > 0.5) {
                        while (rules.size() > 0) {
                            Rule rule = rules.get(0);
                            rule.deleted = true;
                            remove(rules, rule);
                            Iterator it2 = rule.coveredInstInfo.iterator();
                            while (it2.hasNext()) {
                                InstanceInfo inf = (InstanceInfo) it2.next();
                                remove(inf.rules, rule);
                            }
                        }
                        initializeRules();
                        induceRules();
                        unionRules();
                        induceRules();
                    }
                }
            }
        }

    }

    /** Update classifier with new Instances */
    protected void updateClassifier() {
        updates++;
        // if (VERIFY) {
        // verifyInstances();
        // verifyRules();
        // verifyCriterion();
        // }
        // if (updates % 100 == 0) {
        // for (Rule rule : rules) {
        // checkBigRules(rule);
        // }
        // doDelete();
        // doUpdate();
        // }
        significanceTest();

        // Update used data for all new instances (forgetting)
        // if (VERIFY) {
        // verifyInstances();
        // verifyRules();
        // verifyCriterion();
        // }
        // insert new instances
        updateCaseBase();
        // if (VERIFY) {
        // verifyInstances();
        // verifyRules();
        // verifyCriterion();
        // }

        // update rulelist
        updateRules();

        // checkRules();
        newInstances.clear();

        if (NormDistance.maxDistance(distance, distance2) >= 0.1) {
            distance = (NormDistance) distance2.clone();
            mtree.updateDistance(distanceXXL);
        }

    }

    protected void updateCaseBase() {
        ArrayList<Rule> locRules = new ArrayList();
        for (Iterator it = newInstances.iterator(); it.hasNext();) {
            InstanceInfo inf = (InstanceInfo) it.next();
            insert(inf);
            Iterator it2 = inf.rules.iterator();
            while (it2.hasNext()) {
                Rule rule = (Rule) it2.next();
                rule.addDistribution(inf);

                if (indexOf(locRules, rule) < 0) {
                    locRules.add(rule);
                }
            }

        }
        Iterator<Rule> itRule = locRules.iterator();
        while (itRule.hasNext()) {
            Rule rule = itRule.next();
            checkBigRules(rule);
            if (indexOf(updatedRules, rule) < 0) {
                updatedRules.add(rule);
            }
        }

        updateNoRule();

        if (maxInstances > 0 && tree.size() > maxInstances)
            updateOverInstances();

        doDelete();
        doUpdate();
    }

    protected void checkBigRules(Rule rule) {
        int s = rule.coveredInstInfo.size();
        int removed = 0;
        if (s < k_noise)
            return;
        int cls = rule.getClassValue();
        Collections.sort(rule.coveredInstInfo, instanceTimeComparator);
        boolean eq = true;
        int bord = Math.max((int) (s * p_current), 10);
        if (s < bord)
            return;
        for (int i = 0; i < bord; i++) {
            if ((int) rule.coveredInstInfo.get(s - i - 1).inst.classValue() != cls) {
                eq = false;
                break;
            }
        }
        if (eq) {
            for (int i = 0; i < rule.coveredInstInfo.size() - bord; i++) {
                InstanceInfo inf = rule.coveredInstInfo.get(i);
                if (inf.deleted)
                    continue;
                if ((int) inf.inst.classValue() != cls
                /* && (eq || hasOnlyBadRules(inf) ) */) {
                    toDelete.add(inf);
                    removed++;
                    for (Rule rule2 : inf.rules) {
                        if (indexOf(updatedRules, rule2) < 0) {
                            updatedRules.add(rule2);
                        }
                        rule2.setFinished(false);
                    }
                }
            }
        }

        int max = maxRuleSize;
        s = rule.coveredInstInfo.size() - removed;
        if (s <= max)
            return;
        Collections.sort(rule.coveredInstInfo, instanceTimeComparator);
        Iterator<InstanceInfo> it = rule.coveredInstInfo.iterator();
        while (it.hasNext() && (rule.coveredInstInfo.size() - removed) > max) {
            InstanceInfo inf = it.next();
            if (inf.deleted)
                continue;
            if (((inf.rules.size() == 1 || hasOnlyGoodRules(inf))
                    && !onBorder(rule, inf.inst) && hasOnlyGoodNeighbours(inf))) {
                toDelete.add(inf);
                removed++;
            }
        }
    }

    protected void updateNoRule() {
        int pos = 0;
        int del = 0;
        Iterator<InstanceInfo> it = tree.iterator();
        List<Rule> toUpdate = new ArrayList();
        while (pos - del < ((tree.size() - del) * (1 - p_young))) {
            InstanceInfo inf = it.next();
            if (inf.deleted)
                continue;
            if (inf.rules.isEmpty() || hasOnlyBadRules(inf)) {
                toDelete.add(inf);
                for (Rule rule2 : inf.rules) {
                    toUpdate.add(rule2);
                    rule2.setFinished(false);
                }
                del++;
            }
            pos++;
        }
        for (Rule rule2 : toUpdate) {
            rule2.computeBoundsFromCovered();
        }
    }

    // protected void updateNoRule2() {
    // int pos = 0;
    // int del = 0;
    // Collections.sort(noRule, instanceTimeComparator);
    // Iterator<InstanceInfo> it = noRule.iterator();
    // List<Rule> toUpdate = new ArrayList();
    // while (noRule.size() > (mtree.size() * 0.1) && it.hasNext()) {
    // InstanceInfo inf = it.next();
    // if (inf.deleted) {
    // it.remove();
    // continue;
    // }
    // if (inf.rules.isEmpty()) {
    // toDelete.add(inf);
    // for (Rule rule2 : inf.rules) {
    // toUpdate.add(rule2);
    // rule2.setFinished(false);
    // }
    // del++;
    // }
    // pos++;
    // }
    // for (Rule rule2 : toUpdate) {
    // rule2.computeBoundsFromCovered();
    // }
    // doDelete();
    // }

    protected void updateOverInstances() {
        Iterator<InstanceInfo> it = tree.iterator();
        List<Rule> toUpdate = new ArrayList();
        int del = toDelete.size();
        while (tree.size() - del > maxInstances) {
            InstanceInfo inf = it.next();
            if (inf.deleted)
                continue;
            toDelete.add(inf);
            del++;
            for (Rule rule2 : inf.rules) {
                toUpdate.add(rule2);
                rule2.setFinished(false);
            }
        }
        for (Rule rule2 : toUpdate) {
            rule2.computeBoundsFromCovered();
        }
    }

    protected void insert(InstanceInfo inf) {
        mtree.add(inf);
        if (!init)
            distance2.addInstance(inf.inst);
        // btree.insert(inf);
        tree.add(inf);
    }

    protected Object delete(InstanceInfo inf) {
        inf.deleted = true;
        boolean rem = tree.remove(inf);// btree.remove(inf);
        if (rem) {
            mtree.remove(inf);
            if (inf.rules.isEmpty())
                remove(noRule, inf);
            else {
                addUnionRules(rules);
                List infrules=inf.rules;
                while (!infrules.isEmpty()) {
                    Rule rule=(Rule) infrules.remove(infrules.size()-1);
                    rule.removeDistribution(inf);
                    //if (!checkCand.contains(rule)) checkCand.add(rule);
                }
            }
            distance2.removeInstance(inf.inst);
        }
        return rem;
    }
    
    

    // public void checkRules() {
    // if (rules.size() > tree.size() || rules.size() > 100) {
    // Collections.sort(rules, this.ruleLaPlaceComparator);
    // Iterator it = rules.iterator();
    // while (it.hasNext()
    // && (rules.size() > tree.size() || rules.size() > 100)) {
    // Rule r = (Rule) it.next();
    // it.remove();
    // Iterator it2 = r.coveredInstInfo.iterator();
    // while (it2.hasNext()) {
    // InstanceInfo inf = (InstanceInfo) it2.next();
    // if (inf.rules.isEmpty())
    // remove(noRule, inf);
    // inf.rules.add(r);
    // }
    // }
    // }
    // }

    /**
     * Update the model with this labeled Instance
     */
    public void updateClassifier(Instance newInst) {
        updateClassifier(newInst, false);
    }

    /**
     * Update the model with this labeled Instance and force an update now.
     */
    public void updateClassifier(Instance newInst, boolean forceUpdate) {
        Instance inst = (Instance) newInst.copy();
        inst.setDataset(header);

        int cls = -1;
        try {
            cls = (int) classifyInstance(inst);
            test.addClassify(cls == (int) inst.classValue() ? 1f : 0f);
            // addClassify(lastCls);
        } catch (Exception e) {
            e.printStackTrace();
        }

        InstanceInfo newinfo = new InstanceInfo(inst);
        Iterator it = rules.iterator();
        while (it.hasNext()) {
            Rule rule = (Rule) it.next();
            if (rule.covers(inst)) {
                newinfo.rules.add(rule);
            }
        }
        newInstances.add(newinfo);
        if (newInstances.size() >= k_step || forceUpdate)
            updateClassifier();
    }

    /**
     * Delete small rules and some instances of them.
     * 
     */
    protected void checkSmallRules(Rule rule) {
        // Don't delete rules if less than 3 rules exist.
        if (rules.size() <= classCount)
            return;

        // Don't delete rules if less than 10 instances are in the model
        if (tree.size() < 10)
            return;

        // verifyRules(rule);
        // System.out.println(rule.test.getLastError(20));
        if (rule.getDistribution()[rule.getClassValue()] > 3
                && rule.getClassValue() == rule.frequentClass())
                //&& (rule.test.getLastError(10) < 0.5 || rule.getCoverage() <= 0.5))
            return;

        remove(rules, rule);
        rule.deleted = true;

        Iterator el = rule.coveredInstInfo.iterator();
        while (el.hasNext()) {
            InstanceInfo inf = (InstanceInfo) el.next();
            if (!inf.rules.isEmpty()) {
                remove(inf.rules, rule);
                if (inf.rules.isEmpty())
                    noRule.add(inf);
            }

            // if (inf.rules.size() > 1 && hasOnlyBadRules(inf)) {
            // double old = inf.inst.weight();
            // inf.inst.setWeight(0.5 * (thresh + old));
            // updateWeight(inf, old);
            // // inf.deleted = true;
            // // toDelete.add(inf);
            // }
        }
    }

    protected void checkAllSmallRules() {
        // Don't delete rules if less than 3 rules exist.
        if (rules.size() <= classCount)
            return;

        // Don't delete rules if less than 10 instances are in the model
        if (tree.size() < 10)
            return;

        for (int i = 0; i < rules.size(); i++) {
            Rule rule = rules.get(i);

            // verifyRules(rule);
            // System.out.println(rule.test.getLastError(20));
            if (rule.getDistribution()[rule.getClassValue()] > 3)
                return;

            remove(rules, rule);
            rule.deleted = true;
            i--;

            Iterator el = rule.coveredInstInfo.iterator();
            while (el.hasNext()) {
                InstanceInfo inf = (InstanceInfo) el.next();
                if (!inf.rules.isEmpty()) {
                    remove(inf.rules, rule);
                    if (inf.rules.isEmpty())
                        noRule.add(inf);
                }

                // if (inf.rules.size() > 1 && hasOnlyBadRules(inf)) {
                // double old = inf.inst.weight();
                // inf.inst.setWeight(0.5 * (thresh + old));
                // updateWeight(inf, old);
                // // inf.deleted = true;
                // // toDelete.add(inf);
                // }
            }
            if (rules.size() <= classCount)
                return;
        }
        doDelete();
    }

    /**
     * Test if this Instancde is only covered with rules of other classes
     * 
     * @param inf
     *            Instance
     * @return
     */
    protected boolean hasOnlyBadRules(InstanceInfo inf) {
        if (inf.rules.size() == 0)
            return false;
        int cls = (int) inf.inst.classValue();
        for (Iterator it = inf.rules.iterator(); it.hasNext();) {
            Rule rule = (Rule) it.next();
            if (rule.getClassValue() == cls)
                return false;
        }
        return true;
    }

    /**
     * Test if this Instancde is only covered with rules of its class
     * 
     * @param inf
     *            Instance
     * @return
     */
    protected boolean hasOnlyGoodRules(InstanceInfo inf) {
        if (inf.rules.size() == 0)
            return false;
        int cls = (int) inf.inst.classValue();
        for (Iterator it = inf.rules.iterator(); it.hasNext();) {
            Rule rule = (Rule) it.next();
            if (rule.getClassValue() != cls)
                return false;
        }
        return true;
    }

    /**
     * Test if this Instancde is only covered with rules of its class
     * 
     * @param inf
     *            Instance
     * @return
     */
    protected boolean hasOnlyGoodNeighbours(InstanceInfo inf) {
        int cls = (int) inf.inst.classValue();
        Iterator<InstanceInfo> it = mtree.getNearestNeighbours(inf);
        int max = k_redundanz;
        int pos = 0;
        while (pos < max && it.hasNext()) {
            InstanceInfo ninf = it.next();
            if ((int) ninf.inst.classValue() != cls)
                return false;
            pos++;
        }
        return true;
    }

    // /**
    // * Recompute bounds of rules with covered instances.
    // */
    // private void recomputeBounds() {
    // for (int i = 0; i < rules.size(); i++) {
    // Rule rule = (Rule) rules.get(i);
    // if (!rule.computeBoundsFromCovered()) {
    // rules.remove(i);
    // rule.deleted = true;
    // Iterator it = rule.coveredInstInfo.iterator();
    // while (it.hasNext()) {
    // InstanceInfo inf = (InstanceInfo) it.next();
    // if (!inf.rules.isEmpty()) {
    // remove(inf.rules, rule);
    // if (inf.rules.isEmpty())
    // noRule.add(inf);
    // }
    // }
    // i--;
    // }
    // }
    // }

    /**
     * Update rules with new training instances.
     */
    protected void updateRules() {
        newRules.clear();
        List<Rule> coveredGoodRules = new ArrayList();
        for (Iterator it = newInstances.iterator(); it.hasNext();) {
            InstanceInfo inf = (InstanceInfo) it.next();
            Instance inst = inf.inst;
            // if (VERIFY) {
            // verifyInstances();
            // verifyRules();
            // verifyCriterion();
            // }
            splitRules = new ArrayList();
            coveredGoodRules.clear();
            boolean found = false;
            Rule[] ruleArray = rules.toArray(new Rule[rules.size()]);

            List newR = new ArrayList();
            for (Rule rule : ruleArray) {
                if (rule.covers(inst)) {
                    if (rule.getClassValue() == inst.classValue()) {
                        rule.test.addClassify(1);
                        found = true;
                        rule.setFinished(false);
                        rule = induceRule(rule);
                        checkSmallRules(rule);
                        if (!rule.deleted)
                            coveredGoodRules.add(rule);

                    } else {
                        rule.test.addClassify(0);
                        innerDistance(rule, inf.inst);
                        Rule red = reduce(rule, inf);
                        if (red != null) {
                            double crit = criterionLaPlace(rule);
                            double ncrit = criterionLaPlace(red);
                            if (ncrit > crit) {
                                rule.deleted = true;
                                remove(rules, rule);
                                rules.add(red);
                                updateNearestRules(rule, red);
                                checkSmallRules(red);
                            }
                        }
                    }
                }
            }
            optimizeAll(coveredGoodRules);
            addUnionRules(coveredGoodRules);

            // c=c*2;
            if (!found) {
                Rule newRule = new Rule(header, inf, usedAttribute);
                updateDistribution(newRule);

                if (newRule.frequentClass() == inf.inst.classValue()) {
                    rules.add(newRule);
                    newRules.add(newRule);
                    inf.ownRule = newRule;
                    if (inf.rules.isEmpty()) {
                        remove(noRule, inf);
                    }
                    inf.rules.add(newRule);
                    // newRule=induceRule(newRule);
                    // unionRules(newRule);
                }
            }
            if (VERIFY) {
                verifyInstances();
                verifyRules();
                verifyCriterion();
            }

        }

        unionCandRules();
        //unionRules();
        checkAllSmallRules();

        Iterator it = newRules.iterator();
        coveredGoodRules.clear();
        while (it.hasNext()) {
            Rule rule = (Rule) it.next();
            rule = induceRule(rule);
            if (optimizeRule(rule)) {
                rule.deleted = true;
                it.remove();
                remove(rules, rule);
                Iterator it2 = rule.coveredInstInfo.iterator();
                while (it2.hasNext()) {
                    InstanceInfo inf = (InstanceInfo) it2.next();
                    remove(inf.rules, rule);
                }
            } else
                coveredGoodRules.add(rule);
        }
        newRules.clear();
        optimizeAll(coveredGoodRules);
        // if (VERIFY) {
        // verifyInstances();
        // verifyRules();
        // verifyCriterion();
        // }
        // for (Rule rule : rules) {
        // rule.setFinished(false);
        // }
        // learn rules
        // induceRules();

        if (VERIFY) {
            verifyInstances();
            verifyRules();
            verifyCriterion();
        }

        if (VERIFY) {
            verifyInstances();
            verifyRules();
            verifyCriterion();
        }
        boolean rem = false;

        if (maxRules > 0 && rules.size() > maxRules) {
            while (rules.size() > maxRules) {
                int size = rules.size();
                int min = -1;
                double minSize = Double.MAX_VALUE;
                for (int i = 0; i < size; i++) {
                    Rule r = rules.get(i);
                    if (minSize > criterionLaPlace(r)
                            * r.coveredInstInfo.size()) {
                        minSize = criterionLaPlace(r)
                                * r.coveredInstInfo.size();
                        min = i;
                    }
                    // if (minSize > r.coveredInstInfo.size()) {
                    // minSize = r.coveredInstInfo.size();
                    // min = i;
                    // }
                }
                Rule rule = rules.get(min);
                rules.remove(min);
                rule.deleted = true;
                rem = true;

                Iterator el = rule.coveredInstInfo.iterator();
                while (el.hasNext()) {
                    InstanceInfo inf = (InstanceInfo) el.next();
                    if (!inf.rules.isEmpty()) {
                        remove(inf.rules, rule);
                        if (inf.rules.isEmpty())
                            noRule.add(inf);
                    }

                }
            }
            if (rem)
                doDelete();
        }
    }

    /**
     * Split rule with this instance. For each attribute, new splitted rules are
     * created.
     * 
     * @param rule
     * @param instinfo
     */
    protected void splitRule(Rule rule, InstanceInfo instinfo) {
        Instance inst = instinfo.inst;
        List coverInst = rule.coveredInstInfo;
        int cl = rule.getClassValue();
        if (VERIFY) {
            verifyInstances();
            verifyRules();
            verifyCriterion();
        }

        // System.out.print("split ");
        for (int i = 0; i < attrCount; i++) {
            if (usedAttribute[i] > 0 && !inst.isMissing(i)) {
                if (usedAttribute[i] == NOMINAL) {
                    // TODO testen und korrigieren... Regel in inst.rules
                    // einfgen...
                    int val = (int) inst.value(i);
                    int count = header.attribute(i).numValues();
                    Rule[] newRules = new Rule[count];
                    Iterator it = coverInst.iterator();
                    while (it.hasNext()) {
                        InstanceInfo inf = (InstanceInfo) it.next();
                        if ((int) inf.inst.classValue() != cl)
                            continue;
                        int inval = (int) inf.inst.value(i);
                        if (newRules[inval] != null) {
                            newRules[inval].addDistribution(inf);
                            continue;
                        }
                        if (inval != val && newRules[inval] == null) {
                            Rule newRule = new Rule(rule);
                            newRule.setNominalValue(i, inval);
                            newRule.resetDistribution();
                            newRule.addDistribution(inf);

                            newRules[inval] = newRule;
                        }
                    }
                    for (int j = 0; j < count; j++) {
                        if (newRules[j] != null) {
                            if (newRules[j].coveredInstInfo.size() == 0
                                    || newRules[j].computeClass() != rule
                                            .getClassValue()) {
                                newRules[j] = null;
                            } else {
                                splitRules.add(newRules[j]);
                                // TODO etwas mit newRules anfangen.. in rules
                                // bernehmen? spter?
                            }
                        }
                    }
                } else {
                    double val = inst.value(i);
                    double lower = val - EPS;
                    double larger = val + EPS;

                    if (!Double.isNaN(lower) && rule.getNumericRange(i) != null) {
                        double low = rule.getNumericRange(i)[0];
                        if (low < lower) {
                            Rule newRule = new Rule(rule);
                            newRule.setNumericRange(i, low, lower);
                            newRule.computeDistribution(coverInst);

                            if (newRule.coveredInstInfo.size() > 0
                                    && rule.getClassValue() == newRule
                                            .frequentClass()) {
                                if (newRule.computeBoundsFromCovered()
                                        && !newRule.covers(inst)) {

                                    splitRules.add(newRule);

                                }
                            }
                            if (newRule.covers(inst)) {
                                System.out.println("ERROR 1");
                            }

                        }
                    }
                    if (!Double.isNaN(larger)
                            && rule.getNumericRange(i) != null) {
                        double large = rule.getNumericRange(i)[1];
                        if (large > larger) {
                            Rule newRule = new Rule(rule);
                            newRule.setNumericRange(i, larger, large);
                            newRule.computeDistribution(coverInst);
                            if (newRule.coveredInstInfo.size() > 0
                                    && rule.getClassValue() == newRule
                                            .frequentClass()) {
                                newRule.computeBoundsFromCovered();

                                splitRules.add(newRule);
                            }
                            if (newRule.covers(inst)) {
                                System.out.println("ERROR 2");
                            }
                        }
                    }
                }
            }
        }

        if (VERIFY) {
            verifyInstances();
            verifyRules();
            verifyCriterion();
        }
        // System.out.println("...fertig");
    }

    /**
     * Examine if rulelists have a common rule.
     * 
     * @param rules1
     * @param rules2
     * @return
     */
    protected boolean hasSame(List rules1, List rules2) {
        Iterator it = rules1.iterator();
        while (it.hasNext()) {
            if (rules2.contains(it.next()))
                return true;
        }
        return false;
    }

    /**
     * Delete all Instances of the delete list.
     * 
     */
    protected void doDelete() {
        if (toDelete.size() == 0)
            return;
        // if (tree.size() - toDelete.size() < 15)
        // System.out.println("TOO LESS!!!");
        // else System.out.print(toDelete.size() + " Instanzen gelscht. ");
        while (!toDelete.isEmpty()) {
            InstanceInfo inf = (InstanceInfo) toDelete.remove(toDelete.size()-1);
            if (tree.size()>k_noise) delete(inf);
        }
        toDelete.clear();
//        while(!checkCand.isEmpty()){
//            Rule rule=checkCand.remove(checkCand.size()-1);
//            checkSmallRules(rule);
//        }
    }

    protected void doUpdate() {
        for (Rule rule : updatedRules) {
            if (rule.coveredInstInfo.size() == 0) {
                rule.deleted = true;
                remove(rules, rule);
            } else
                rule.computeBoundsFromCovered();
        }
        updatedRules.clear();
    }

    /**
     * Add one distribution to another.
     * 
     * @param dist
     *            the main distribution
     * @param newDist
     *            distribution to add
     */
    protected void addDistribution(double[] dist, double[] newDist,
            double newWeight) {
        for (int i = 0; i < classCount; i++) {
            dist[i] += newDist[i] * newWeight;
        }
    }

    /**
     * Compute the percentage of the class of all covered examples.
     * 
     * @param rule
     *            the rule
     * @return the criterion.
     */
    protected double criterion(Rule rule) {
        return (rule.getDistribution()[rule.getClassValue()])
                / (rule.getDistributionSum());
    }

    /**
     * Compute the Laplace criterion of an rule.
     * 
     * @param rule
     *            the rule
     * @return the criterion.
     */
    protected double criterionLaPlace(Rule rule) {
        return (rule.getDistribution()[rule.getClassValue()] + 1)
                / (rule.getDistributionSum() + classCount);
    }

    /**
     * Predicts the class memberships for a given instance. If an instance is
     * unclassified, the returned array elements must be all zero.
     * 
     * @param instance
     *            the instance to be classified
     * @return an array containing the estimated membership probabilities of the
     *         test instance in each class
     * @exception Exception
     *                if distribution could not be computed successfully
     */
    public double[] distributionForInstance(Instance instance) {
        InstanceInfo instinf = new InstanceInfo(instance);
        List nRules = findNearestRules(instinf);
        double[] result = getDistribution(nRules, instinf);
        if (result == null) {
            System.out.println("NULL DIST!!!");
            nRules = findNearestRules(instinf);
            result = getDistribution(nRules, instinf);
        }
        return result;
    }

    public double classifyIB(Instance instance, int k) {

        Iterator nearest = mtree
                .getNearestNeighbours(new InstanceInfo(instance));
        double[] res = new double[classCount];
        int count = 0;
        while (nearest.hasNext() && count < k) {
            InstanceInfo inf = (InstanceInfo) nearest.next();
            if (inf.deleted) {
                System.out.println("DELETE ERROR");
                toDelete.add(inf);
                continue;
            }
            res[(int) inf.inst.classValue()] += 1;// inf.inst.weight();
            count += 1;// inf.inst.weight();
        }
        for (int i = 0; i < classCount; i++) {
            res[i] /= count;
        }

        double max = 0;
        int maxIndex = -1;

        for (int i = 0; i < res.length; i++) {
            if (res[i] > max) {
                maxIndex = i;
                max = res[i];
            }
        }
        return maxIndex;
    }

    protected int classify(InstanceInfo inf) {
        List nRules = inf.rules;
        if (nRules.isEmpty())
            nRules = findNearestRules(inf);
        double[] dist = getDistribution(nRules, inf);

        double max = 0;
        int maxIndex = -1;

        for (int i = 0; i < dist.length; i++) {
            if (dist[i] > max) {
                maxIndex = i;
                max = dist[i];
            }
        }
        return maxIndex;
    }

    /**
     * Compute distance of two Instances
     * 
     * @param instance1
     *            first instance
     * @param instance2
     *            second instance
     * @return distance of instances
     */
    protected double distance(Instance instance1, Instance instance2) {
        double dist = 0;
        for (int i = 0; i < attrCount; i++) {
            if (usedAttribute[i] > 0) {
                double val1 = 0, val2 = 0;
                if (instance1.isMissing(i))
                    val1 = Double.NaN;
                else
                    val1 = instance1.value(i);
                if (instance2.isMissing(i))
                    val2 = Double.NaN;
                else
                    val2 = instance2.value(i);
                double d = distance.distance(i, val1, val2);
                // if (d < 0 || d > 1) System.out.println("ERROR!!!");
                dist += d * d;
            }
        }
        return Math.sqrt(dist);
    }

    protected boolean equals(Instance instance1, Instance instance2) {
        for (int i = 0; i < attrCount; i++) {
            if (usedAttribute[i] > 0) {
                double val1 = 0, val2 = 0;
                val1 = instance1.value(i);
                val2 = instance2.value(i);
                if (val1 != val2)
                    return false;
            }
        }
        return true;
    }

    /**
     * Compute the distance of one Rule to an Instance.
     * 
     * @param rule
     *            the rule
     * @param instance
     *            the instance
     * @return distance of instance to rule
     */
    protected double distance(Rule rule, Instance instance) {
        double dist = 0;
        for (int i = 0; i < attrCount; i++) {
            if (usedAttribute[i] > 0) {
                if (instance.isMissing(i))
                    continue;
                if (usedAttribute[i] == NOMINAL) {
                    if (rule.getNominalValue(i) >= 0) {
                        if (rule.getNominalValue(i) != (int) instance.value(i)) {
                            double dif = distance.distance(i, rule
                                    .getNominalValue(i), (int) instance
                                    .value(i));
                            dist += dif * dif;
                        }
                    }
                } else if (usedAttribute[i] == NUMERIC) {
                    if (rule.getNumericRange(i) != null) {
                        double val = instance.value(i);
                        if (val < rule.getNumericRange(i)[0]) {
                            double dif = distance.distance(i, rule
                                    .getNumericRange(i)[0], val);
                            dist += dif * dif;
                        } else if (val > rule.getNumericRange(i)[1]) {
                            double dif = distance.distance(i, val, rule
                                    .getNumericRange(i)[1]);
                            dist += dif * dif;
                        }
                    }
                }
            }
        }
        return Math.sqrt(dist);
    }

    /**
     * Compute the distance of two rules.
     * 
     * @param rule
     *            the first rule
     * @param rule2
     *            the second rule
     * @return distance of the rules
     */
    protected double distance(Rule rule, Rule rule2) {
        double dist = 0;
        for (int i = 0; i < attrCount; i++) {
            if (usedAttribute[i] > 0) {

                if (usedAttribute[i] == NOMINAL) {
                    if (rule.getNominalValue(i) >= 0
                            && rule2.getNominalValue(i) >= 0) {
                        if (rule.getNominalValue(i) != rule2.getNominalValue(i)) {
                            double dif = distance.distance(i, rule
                                    .getNominalValue(i), rule2
                                    .getNominalValue(i));
                            dist += dif * dif;
                        }
                    }
                } else if (usedAttribute[i] == NUMERIC) {
                    if (rule.getNumericRange(i) != null
                            && rule2.getNumericRange(i) != null) {
                        if (rule2.getNumericRange(i)[1] < rule
                                .getNumericRange(i)[0]) {
                            double dif = distance.distance(i, rule
                                    .getNumericRange(i)[0], rule2
                                    .getNumericRange(i)[1]);
                            dist += dif * dif;
                        } else if (rule2.getNumericRange(i)[0] > rule
                                .getNumericRange(i)[1]) {
                            double dif = distance.distance(i, rule2
                                    .getNumericRange(i)[0], rule
                                    .getNumericRange(i)[1]);
                            dist += dif * dif;
                        }
                    }
                }
            }
        }
        return Math.sqrt(dist);
    }

    protected boolean onBorder(Rule rule, Instance instance) {
        for (int i = 0; i < attrCount; i++) {
            if (usedAttribute[i] > 0) {
                if (instance.isMissing(i))
                    continue;
                if (usedAttribute[i] == NUMERIC) {
                    if (rule.getNumericRange(i) != null) {
                        double val = instance.value(i);
                        double[] rVal = rule.getNumericRange(i);
                        if (val == rVal[0])
                            return true;
                        if (val == rVal[1])
                            return true;
                    }
                }
            }
        }
        return false;
    }

    /**
     * Compute the distance of the nearest border of the Rule to an Instance
     * covered by the rule (inner distance).
     * 
     * @param rule
     *            the rule
     * @param instance
     *            the instance
     * @return distance of instance to rule
     */
    protected double innerDistance(Rule rule, Instance instance) {
        double dist = 0.0;
        innerAttribute = -1;
        double maxAtt = -1;
        for (int i = 0; i < attrCount; i++) {
            if (usedAttribute[i] > 0) {
                if (instance.isMissing(i))
                    continue;
                if (usedAttribute[i] == NOMINAL) {
                    if (rule.getNominalValue(i) < 0) {
                        double d = distance.distance(i, instance.value(i),
                                Double.NaN);
                        dist += d * d;
                        // if (d>maxAtt) {
                        // innerAttribute=i;
                        // maxAtt=d;
                        // }
                    }
                } else if (usedAttribute[i] == NUMERIC) {
                    if (rule.getNumericRange(i) != null) {
                        double val = instance.value(i);
                        double[] rVal = rule.getNumericRange(i);
                        double d = Math.abs(distance.distance(i,
                                0.5 * (rVal[0] + rVal[1]), val));

                        dist += d * d;
                        if (d > maxAtt) {
                            innerAttribute = i;
                            maxAtt = d;
                        }
                    }
                }
            }
        }
        return Math.sqrt(dist);
    }

    /**
     * Detect the attributes to use for learning.
     */
    protected void computeUsedAttributes() {
        usedAttribute = new int[attrCount];
        for (int i = 0; i < attrCount; i++) {
            if (i != classIndex && i != timeIndex && isUsedAttribute(i)) {
                Attribute att = header.attribute(i);
                if (att.isNominal()) {
                    usedAttribute[i] = NOMINAL;
                } else if (att.isNumeric())
                    usedAttribute[i] = NUMERIC;
            }
        }
    }

    /**
     * Set the attributes to use for learning.
     */
    public void setOperativeAttributes(List usedAttributeList) {
        this.usedAttributeList = usedAttributeList;
    }

    /**
     * Is Attribute in the used attribute list.
     * 
     * @param attribute
     *            attribute index
     * @return if this attribute should be used for learning
     */
    protected boolean isUsedAttribute(int attribute) {
        if (usedAttributeList == null)
            return true;
        else
            return (usedAttributeList.indexOf(new Integer(attribute)) >= 0);
    }

    /**
     * Get the list of all rules
     * 
     * @return List of rules
     */
    public List getRules() {
        return rules;
    }

    /**
     * Get the number of generated rules
     * 
     * @return number of rules
     */
    public int getRuleCount() {
        return rules.size();
    }

    /**
     * Return an string representation of the induced rules.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        Iterator it = rules.iterator();
        while (it.hasNext()) {
            SimpleConjunctiveRule rule = (SimpleConjunctiveRule) it.next();
            sb.append(rule.toString() + "  (" + rule.getCoverage() + ")");
            sb.append("\n");
        }
        sb.append(rules.size() + "\n\n");
        return sb.toString();
    }

    /**
     * No numeric normalization.
     * 
     * @param flag
     *            normalize flag
     */
    public void setNoNumericNormalization() {
        numeric = NormDistance.NO_NUMERIC;
    }

    /**
     * Use range normalization.
     * 
     * @param flag
     *            normalize flag
     */
    public void setRangeNumericNormalization() {
        numeric = NormDistance.RANGE_NUMERIC;
    }

    /**
     * Use quantile normalization.
     * 
     * @param flag
     *            normalize flag
     */
    public void setQuantileNumericNormalization(int quantile) {
        numeric = NormDistance.QUANTILE_NUMERIC;
        this.quantile = quantile;
    }

    /**
     * Get the flag for numeric normalization.
     * 
     * @return normalize flag
     */
    public int getNumericNormalization() {
        return numeric;
    }

    /**
     * Set the flag for nominal VDM distance.
     * 
     * @param flag
     *            nominal VDM distance flag
     */
    public void setNominalVDMDistance(boolean flag) {
        useVDM = flag;
    }

    /**
     * Get the flag for nominal VDM distance.
     * 
     * @return nominal VDM distance flag
     */
    public boolean getNominalVDMDistance() {
        return useVDM;
    }

    private void verifyInstances() {
        Iterator it = tree.iterator();
        while (it.hasNext()) {
            InstanceInfo inf = (InstanceInfo) it.next();
            verifyInstances(inf);
        }
    }

    private void verifyInstances(InstanceInfo inf) {
        Iterator it = inf.rules.iterator();
        while (it.hasNext()) {
            Rule r = (Rule) it.next();
            if (r.deleted) {
                System.out.println("deleted Rule!!!!");
            }
            if (!r.coveredInstInfo.contains(inf)) {
                System.out
                        .println("missing Instance!!!! " + r.covers(inf.inst));
                r.computeBoundsFromCovered();
                System.out
                        .println("missing Instance!!!! " + r.covers(inf.inst));
                r.computeBoundsFromCovered();
                System.out
                        .println("missing Instance!!!! " + r.covers(inf.inst));
            }
        }
    }

    private void verifyRules() {
        Iterator it = rules.iterator();
        while (it.hasNext()) {
            Rule rule = (Rule) it.next();
            verifyRules(rule);
        }
        verifyNoRules();
    }

    private void verifyNoRules() {

        Iterator it = noRule.iterator();
        while (it.hasNext()) {
            InstanceInfo inf = (InstanceInfo) it.next();
            if (inf.deleted) {
                System.out.println("deleted Instance!!!!");
            }
            if (!inf.rules.isEmpty()) {
                System.out.println("No empty Rules!!!!");
            }
        }
    }

    private void verifyRules(Rule rule) {
        Iterator it = rule.coveredInstInfo.iterator();
        while (it.hasNext()) {
            InstanceInfo inf = (InstanceInfo) it.next();
            if (inf.deleted) {
                System.out.println("deleted Instance!!!!");
            }
            if (!inf.rules.contains(rule)) {
                System.out.println("missing Rule!!!!");
            }

        }
    }

    private void verifyCriterion(Rule rule) {
        double[] d = new double[classCount];
        double sum = 0;
        int sum2 = 0;
        List t = new ArrayList();
        Iterator it = tree.iterator();
        while (it.hasNext()) {
            InstanceInfo instinf = (InstanceInfo) it.next();
            Instance inst = instinf.inst;
            if (rule.covers(inst)) {
                t.add(instinf);
                d[(int) inst.classValue()] += inst.weight();
                sum += inst.weight();
                sum2++;
            }
        }
        double crit1 = (d[rule.getClassValue()] + 1) / (sum + classCount);
        double crit2 = criterionLaPlace(rule);
        if (Math.abs(crit1 - crit2) > 0.00001) {
            System.out.println("wrong crit!!!");
            it = tree.iterator();
            while (it.hasNext()) {
                InstanceInfo instinf = (InstanceInfo) it.next();
                Instance inst = instinf.inst;
                if (rule.covers(inst)) {
                    d[(int) inst.classValue()] += inst.weight();
                    sum += inst.weight();
                }
            }
        }
    }

    private void verifyCriterion() {
        Iterator it = rules.iterator();
        while (it.hasNext()) {
            Rule rule = (Rule) it.next();
            verifyCriterion(rule);
        }
    }

    // private boolean verifyList(List l1, List l2) {
    // if (l1.size() != l2.size()) {
    // System.out.println("WRONG LIST");
    // return false;
    // } else {
    // Iterator it = l1.iterator();
    // while (it.hasNext()) {
    // Object ob = it.next();
    // int id = indexOf(l2, ob);
    // if (id < 0) {
    // System.out.println("WRONG LIST");
    // return false;
    // }
    // }
    // }
    // return true;
    // }

    /**
     * Returns an enumeration of all the available options.
     * 
     * @return an enumeration of all available options.
     */
    public Enumeration listOptions() {
        Vector newVector = new Vector(9);

        newVector
                .addElement(new Option(
                        "\tWhich Normalization of numeric attributes are used.\n"
                                + "\tuse: -N numeric_flag\n"
                                + "\tpossible values for numeric_flag are \n"
                                + "\t    no         use no numeric normalization.\n"
                                + "\t    range      use range normalization (max-min=1.0)\n"
                                + "\t    [0.0-0.5]  use quantile normalization with the given quantile\n\n",
                        "N", 2, "-N"));

        newVector.addElement(new Option(
                "\tDon't use the VDM distance for nominal attributes.", "V", 0,
                "-V"));

        newVector
                .addElement(new Option(
                        "\tWhich tiebreak strategy is used for rules with same distance.\n"
                                + "\tuse: -T method\n"
                                + "\tpossible values for method are \n"
                                + "\t    laplace    Choose rule with highest Laplace accuracy\n"
                                + "\t    specific   Choose most specific rule\n"
                                + "\t    vote       Use frequency-based voting",
                        "T", 2, "-T"));

        newVector.addElement(new Option(
                "\tUse the local stopping strategy for rule induction.", "L",
                0, "-L"));

        newVector.addElement(new Option(
                "\tUse the simplification strategy on the ruleset.", "S", 0,
                "-S"));
        return newVector.elements();
    }

    /**
     * Parses a given list of options.
     * 
     * @param options
     *            the list of options as an array of strings
     * @exception Exception
     *                if an option is not supported
     */
    public void setOptions(String[] options) throws Exception {
        String flag = Utils.getOption('N', options);
        if (flag.equalsIgnoreCase("no")) {
            numeric = NormDistance.NO_NUMERIC;
        } else if (flag.equalsIgnoreCase("range")) {
            numeric = NormDistance.RANGE_NUMERIC;
        } else {
            try {
                double quantile = Double.parseDouble(flag);
                this.quantile = quantile;
                numeric = NormDistance.QUANTILE_NUMERIC;
            } catch (NumberFormatException e) {
                System.err.println("WRONG FORMAT OF THE NORMALIZATION FLAG!");
                System.exit(1);
            }
        }

        flag = Utils.getOption('T', options);
        if (flag.equalsIgnoreCase("laplace")) {
            numeric = HIGHEST_LAPLACE_TIEBREAK;
        } else if (flag.equalsIgnoreCase("specific")) {
            numeric = MOST_SPECIFIC_TIEBREAK;
        } else if (flag.equalsIgnoreCase("vote")) {
            numeric = FREQUENCY_BASED_VOTING_TIEBREAK;
        }

        useVDM = !Utils.getFlag('V', options);
        globalStop = !Utils.getFlag('L', options);
        simplification = Utils.getFlag('S', options);

    }

    /**
     * Gets the current option settings.
     * 
     * @return the list of current option settings as an array of strings
     */
    public String[] getOptions() {
        String[] options = new String[8];
        int current = 0;
        options[current++] = "-N";
        if (numeric == NormDistance.NO_NUMERIC)
            options[current++] = "no";
        else if (numeric == NormDistance.RANGE_NUMERIC)
            options[current++] = "range";
        else if (numeric == NormDistance.QUANTILE_NUMERIC)
            options[current++] = "" + quantile;
        if (!useVDM)
            options[current++] = "-V";

        options[current++] = "-T";
        if (tieBreak == HIGHEST_LAPLACE_TIEBREAK)
            options[current++] = "laplace";
        else if (tieBreak == MOST_SPECIFIC_TIEBREAK)
            options[current++] = "specific";
        else if (tieBreak == FREQUENCY_BASED_VOTING_TIEBREAK)
            options[current++] = "vote";

        if (!globalStop)
            options[current++] = "-L";

        while (current < options.length)
            options[current++] = "";

        return options;
    }

    /**
     * Main method.
     * 
     * @param args
     *            the options for the classifier
     */
    public static void main(String[] args) {
        try {
            System.out.println(Evaluation.evaluateModel(new CopyOfRBLDS6(), args));
        } catch (Exception e) {
            e.printStackTrace();
            System.err.println(e.getMessage());
        }
    }

}