package weka.classifiers.rules.flora;
import java.util.*;
import weka.core.Instance;
import weka.core.Attribute;

public class DescriptionItem implements Cloneable {
    //only positive count is used in ADES
    //only negative count is used in NDES
    //both counters are used in PDES
    private long m_positiveMatchCount;
    private long m_negativeMatchCount;
    private Hashtable<Integer,Integer> m_attValuesPairs = new Hashtable();
    /**
     * Returns a deep copy of this description item
     */
    public Object clone()
    {
        DescriptionItem clone = null;
        try {
          clone = (DescriptionItem)super.clone();
        } catch(CloneNotSupportedException e) {
          // should never happen
        }
        clone.m_attValuesPairs=(Hashtable<Integer,Integer>)m_attValuesPairs.clone();
        
        //deep clone implemented
        clone.setPositiveCount(this.getPositiveCount());
        clone.setNegativeCount(this.getNegativeCount());
        clone.setAttributeTable((Hashtable)this.getAttributeTable().clone());
        return clone;
    }
    /**
     * Outputs the attribute-value pairs for this description item
     * @return A string containing the attribute-value pairs
     */
    public String printItem()
    {
        
        String result = "Covers " + String.valueOf(m_positiveMatchCount) + " positive and " + String.valueOf(m_negativeMatchCount) + " negative instances in the window\n";
        boolean hasAttributes = false;
        Enumeration attributeNames = m_attValuesPairs.keys();
        while(attributeNames.hasMoreElements())
        {
            hasAttributes = true;
            String attributeName = attributeNames.nextElement().toString();
            result += attributeName + ": " + m_attValuesPairs.get(attributeName) + "\n";
        }
        if(!hasAttributes)
            result += "Universal item -no constraints on any attribute - covers all instances";
        return result;
    }
    public DescriptionItem()
    {
        m_positiveMatchCount = 0;
        m_negativeMatchCount = 0;
    }
    /**
     * Initializes a description item to cover one specific instance
     * @param instance
     */
    public DescriptionItem(Instance instance)
    {
        m_negativeMatchCount = 0;
        m_positiveMatchCount = 0;
        if(instance.classValue() > 0)
            m_positiveMatchCount = 1;
        else
            m_negativeMatchCount = 1;
        
        //find the corresponding attribute in the instance
        for(int attrIndex = 0; attrIndex < instance.numAttributes()-1; ++attrIndex)
        {
            int attrValue = (int)instance.value(attrIndex);
            m_attValuesPairs.put(attrIndex, attrValue);
        }
    }
    /**
     * Returns the attribute-value pairs table
     * @return
     */
    public Hashtable<Integer,Integer> getAttributeTable()
    {
        return m_attValuesPairs;
    }
    private void setAttributeTable(Hashtable<Integer,Integer> attributeTable)
    {
        m_attValuesPairs = attributeTable;
    }
    /**
     * Returns the number of positive instances in the window covered by the item
     * @return
     */
    public long getPositiveCount()
    {
        return m_positiveMatchCount;
    }
    /**
     * Returns the number of negative instances in the window covered by the item
     * @return
     */
    public long getNegativeCount()
    {
        return m_negativeMatchCount;
    }
    /**
     * Sets the positive and negative instances counts to zero
     *
     */
    public void resetItemsCoveredCounts()
    {
        m_positiveMatchCount = 0;
        m_negativeMatchCount = 0;
    }
    /**
     * Sets the number of positive instances in the window covered by the item
     * @param count
     */
    public void setPositiveCount(long count)
    {
        m_positiveMatchCount = count;
    }
    /**
     * Sets the number of negative instances in the window covered by the item
     * @param count
     */
    public void setNegativeCount(long count)
    {
        m_negativeMatchCount = count;
    }
    public void setCount(boolean instanceClass, long count)
    {
        if(instanceClass)
            setPositiveCount(count);
        else
            setNegativeCount(count);
    }
    /**
     * Returns the number of positive or negative instances this item covers
     * @param instanceClass - used to signal whether to return positive (true) or negative (false) count
     * @return
     */
    public long getCount(boolean instanceClass)
    {
        if(instanceClass)
            return m_positiveMatchCount;
        else
            return m_negativeMatchCount;        
    }
    
    /**
     * Returns a flag indicating whether this item covers the instance
     * @param instance
     * @return
     */
    public boolean coversInstance(Instance instance)
    {
        boolean result = true;
        //go through all the attributes of the item
        Enumeration attributeNames = m_attValuesPairs.keys();
        while(attributeNames.hasMoreElements())
        {
            Integer aIndex = (Integer)attributeNames.nextElement();
            int attributeValue = (Integer)m_attValuesPairs.get(aIndex);
            boolean found = false;
            //find the corresponding attribute in the instance
            for(int attrIndex = 0; attrIndex < instance.numAttributes()-1; ++attrIndex)
            
            {
                if(attrIndex == aIndex)
                {
                    //found the attribute, comparing the values
                    //right now no generalization performed equality comparison only
                    if(instance.value(attrIndex) == attributeValue)
                        found = true;
                }
            }
            //if the value was not found, the instance was not matched
            if(!found)
            {
                result = false;
                break;
            }
        }
        return result;
    }
    /**
     * Returns the number of dimensions (attributes) where instance is not covered by this description item
     * @param instance
     * @return
     */
    public int getDistanceFromInstance(Instance instance)
    {
        int distance = 0;
        //go through all the attributes of the item
        Enumeration thisItemAttributeNames = m_attValuesPairs.keys();
        while(thisItemAttributeNames.hasMoreElements())
        {
            Integer aIndex = (Integer)thisItemAttributeNames.nextElement();
            int thisItemAttributeValue = (Integer)m_attValuesPairs.get(aIndex);
            Enumeration instanceAttrEnum = instance.enumerateAttributes();
            boolean found = false;
            //find the corresponding attribute in the instance
            for(int attrIndex = 0; attrIndex < instance.numAttributes()-1; ++attrIndex)
                
            {
                if(attrIndex == aIndex)
                {
                    //found the attribute, comparing the values
                    if(instance.value(attrIndex) == thisItemAttributeValue)
                        found = true;
                }
            }
            //if the value was not found, the instance was not matched and a generalization of the attribute is needed
            //the distance is the number of generalizations needed to cover the instance
            if(!found)
                ++distance;
        }
        return distance;
    }
    /**
     * Returns a flag indicating whether this item subsumes (is at least as general as) the input item
     * @param descriptionItem
     * @return
     */
    public boolean subsumesItem(DescriptionItem descriptionItem)
    {
        boolean result = true;
        
        //go through the attributes of THIS item looking for restrictions that are relaxed in the INPUT item
        Enumeration thisItemAttributeNames = this.getAttributeTable().keys();
        while(thisItemAttributeNames.hasMoreElements())
        {
            Integer currentAttributeName = (Integer)thisItemAttributeNames.nextElement();
            int attributeValue = (Integer)this.getAttributeTable().get(currentAttributeName);
            //only compare attribute values if there are restrictions imposed
            if(descriptionItem.getAttributeTable().containsKey(currentAttributeName))
            {
                //compare the value in the input item to the value of the current item for this attribute
                Integer aValue=descriptionItem.getAttributeTable().get(currentAttributeName);
                
                if(aValue==null || aValue != attributeValue)
                {
                    result = false;
                    break;
                }
            }
            //an attribute which appears in THIS item is missing in the INPUT item
            //this means there is a restriction on this attribute in THIS item, which does not appear in the INPUT item
            //this means the INPUT item subsumes the current one with regard to this attribute
            else
            {
                result = false;
                break;
            }
        }
        return result;
    }
    /**
     * Returns a flag indicating whether the description item subsumes some item in the input description set
     * @param DES
     * @return
     */
    public boolean subsumesDESItems(DescriptionSet DES)
    {
        boolean result = false;
        //go through the description items of the description set
        for(int diIndex = 0; diIndex < DES.getDesciptionItems().size(); ++diIndex)
        {
            DescriptionItem di = (DescriptionItem)DES.getDesciptionItems().get(diIndex);
            //if the description item subsumes one of the description items of the description set, exit
            if(this.subsumesItem(di))
            {
                result = true;
                break;
            }
        }
        return result;
    }
    
    /**
     * Returns a flag indicating whether this item overlap the input item
     * @param descriptionItem
     * @return
     */
    public boolean overlapItem(DescriptionItem descriptionItem)
    {
        boolean result = true;
        
        //go through the attributes of THIS item looking for restrictions that are relaxed in the INPUT item
        Enumeration thisItemAttributeNames = this.getAttributeTable().keys();
        while(thisItemAttributeNames.hasMoreElements())
        {
            Integer currentAttributeName = (Integer)thisItemAttributeNames.nextElement();
            int attributeValue = (Integer)this.getAttributeTable().get(currentAttributeName);
            //only compare attribute values if there are restrictions imposed
            if(descriptionItem.getAttributeTable().containsKey(currentAttributeName))
            {
                //compare the value in the input item to the value of the current item for this attribute
                Integer aValue=descriptionItem.getAttributeTable().get(currentAttributeName);
                
                if(aValue!=null && aValue != attributeValue)
                {
                    result = false;
                    break;
                }
            }
            //an attribute which appears in THIS item is missing in the INPUT item
            //this means there is a restriction on this attribute in THIS item, which does not appear in the INPUT item
            //this means the INPUT item subsumes the current one with regard to this attribute
            else
            {
//                result = false;
//                break;
            }
        }
        return result;
    }
    /**
     * Returns a flag indicating whether the description item overlaps some item in the input description set
     * @param DES
     * @return
     */
    public boolean overlapDESItems(DescriptionSet DES)
    {
        boolean result = false;
        //go through the description items of the description set
        for(int diIndex = 0; diIndex < DES.getDesciptionItems().size(); ++diIndex)
        {
            DescriptionItem di = (DescriptionItem)DES.getDesciptionItems().get(diIndex);
            //if the description item subsumes one of the description items of the description set, exit
            if(this.overlapItem(di))
            {
                result = true;
                break;
            }
        }
        return result;
    }
    
    public String toString(){
        return this.m_attValuesPairs.toString(); 
    }
}