package weka.classifiers.rules.flora;
import java.util.*;
import weka.core.Instance;

public class DescriptionSet {
	private ArrayList m_descriptionItems = new ArrayList();
	public DescriptionSet()
	{
		
	}
	/**
	 * Returns a deep copy of this description set
	 */
	public Object clone()
	{
	    Object clone = new DescriptionSet();
	    //deep clone implemented
	    ArrayList newDescriptionItems = new ArrayList();
	    for(int i=0; i < this.getDesciptionItems().size(); ++i)
	    {
	    	newDescriptionItems.add(((DescriptionItem)this.getDesciptionItems().get(i)).clone());
	    }
	    ((DescriptionSet)clone).setDescriptionItems(newDescriptionItems);
	    return clone;
	}
	/**
	 * Removes zero-coverage description items from this description set
	 *
	 */
	public void removeZeroCoverageDescriptionItems()
	{
		//go through all description items in this set 
		for(int i=0; i < m_descriptionItems.size(); ++i)
		{
			DescriptionItem di = ((DescriptionItem)m_descriptionItems.get(i));
			//if the item does not cover any instances, remove it
			if(di.getNegativeCount() == 0 && di.getPositiveCount() == 0)
				m_descriptionItems.remove(i);
		}
	}
	/**
	 * Resets the positive and negative instances covered counters in all description items
	 *
	 */
	public void resetInstancesCoveredCounters()
	{
		for(int i=0; i < m_descriptionItems.size(); ++i)
		{
			DescriptionItem di = ((DescriptionItem)m_descriptionItems.get(i));
			di.resetItemsCoveredCounts();
		}
	}
	private void setDescriptionItems(ArrayList descriptionItems)
	{
		m_descriptionItems = descriptionItems;
	}
	public ArrayList getDesciptionItems()
	{
		return m_descriptionItems;
	}
	public void dropDescriptionItem(DescriptionItem descriptionItem)
	{
		if(m_descriptionItems.contains(descriptionItem))
			m_descriptionItems.remove(descriptionItem);
	}
	public void addDescriptionItem(DescriptionItem descriptionItem)
	{
		//before adding, make sure the new item to be 
		//added is not already subsumed by an existing item
		if(!isSubsumed(descriptionItem))
			m_descriptionItems.add(descriptionItem);
	}
	/**
	 * Updates the counters for the instance in the description set 
	 * instanceClass - class value
	 * updateMode - true:learn, false:forget
	 */
	public boolean updateCoveringItems(Instance instance, boolean instanceClass, boolean updateMode)
	{
		boolean result = false;
		//go through all the description items, verifying that
		//at least one of them covers the instance
		for(int i=0; i < m_descriptionItems.size(); ++i)
		{
			DescriptionItem di = ((DescriptionItem)m_descriptionItems.get(i));
			if(di.coversInstance(instance))
			{
				//learn
				if(updateMode)
					di.setCount(instanceClass, di.getCount(instanceClass)+1);
				//forget
				else
				{
					di.setCount(instanceClass, di.getCount(instanceClass)-1);
				}
				//if the number of instances this item covers falls to 0 delete it
				if(di.getCount(instanceClass) <= 0)
					m_descriptionItems.remove(di);
				result = true;
			}
		}
		return result;
	}
	public boolean coversInstance(Instance instance)
	{
		boolean result = false;
		//go through all the description items, verifying that
		//at least one of them covers the instance
		for(int i=0; i < m_descriptionItems.size(); ++i)
		{
			DescriptionItem di = ((DescriptionItem)m_descriptionItems.get(i));
			if(di.coversInstance(instance))
			{
				result = true;
				break;
			}
		}
		return result;
	}
	/**
	 * Adds a description item with values from the instance 
	 * @param instance
	 * @return
	 */
	public boolean includeItemFromInstance(Instance instance)
	{
		boolean result = true;
		DescriptionItem di = new DescriptionItem(instance);
		if(!isSubsumed(di))
		{
			m_descriptionItems.add(di);
		}
		return result;
	}
	/**
	 * Includes a description item in a description set, if it's not subsumed by another one in the same set
	 * @param di
	 * @param instanceClass
	 */
	public void includeDescriptorItem(DescriptionItem di, boolean instanceClass)
	{
		//add the new description item only if it's not subsumed by 
		//another item in the same description set
		if(!isSubsumed(di))
			m_descriptionItems.add(di);
	}
	/**
	 * Returns a flag indicating whether a description item is subsumed in this description set
	 * @param descriptionItem
	 * @return
	 */
	public boolean isSubsumed(DescriptionItem descriptionItem)
	{
		boolean result = false;
		for(int i=0; i < m_descriptionItems.size(); ++i)
		{
			//an item is considered to be subsumed if if there exists an item at least as general as it
			DescriptionItem di = (DescriptionItem)m_descriptionItems.get(i);
			if(di.subsumesItem(descriptionItem))
			{
				result = true;
				break;
			}
		}
		return result;
	}
	/**
	 * Removes items in a description set which are subsumed by a description item
	 * @param descriptionItem
	 */
	public void removeSubsumedItems(DescriptionItem descriptionItem)
	{
		//keeps the DIs to be deleted
		ArrayList subsumedDIsList = new ArrayList();
		//traverse all the description items of the description set
		for(int i=0; i < m_descriptionItems.size(); ++i)
		{
			//an item is considered to be subsumed if if there exists an item at least as general as it
			DescriptionItem di = (DescriptionItem)m_descriptionItems.get(i);
			if(descriptionItem.subsumesItem(di))
			{
				//adds the number of instances covered from the item to be deleted to the more general item
				//descriptionItem.setPositiveCount(descriptionItem.getPositiveCount() + di.getPositiveCount());
				//descriptionItem.setNegativeCount(descriptionItem.getNegativeCount() + di.getNegativeCount());
				subsumedDIsList.add(di);
			}
		}
		//deletes the DIs from the original DES
		while(subsumedDIsList.size() != 0)
		{
			DescriptionItem di = (DescriptionItem)subsumedDIsList.get(0);
			subsumedDIsList.remove(di);
			m_descriptionItems.remove(di);
		}
	}
	public int getInstancesCoveredCount(boolean instanceClass)
	{
		int result = 0;
		for(int i=0; i < m_descriptionItems.size(); ++i)
		{
			//an item is considered to be subsumed if if there exists an item at least as general as it
			DescriptionItem di = (DescriptionItem)m_descriptionItems.get(i);
			if(instanceClass)
				result += di.getPositiveCount();
			else
				result += di.getNegativeCount();
		}
		return result;
	}
    
    public String toString(){
        return this.m_descriptionItems.toString(); 
    }
}