/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Discretize.java
 *    Copyright (C) 1999 Eibe Frank,Len Trigg
 *
 */

package weka.classifiers.rules.flora;

import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

/**
 * An instance filter that discretizes a range of numeric attributes in the
 * dataset into nominal attributes. Discretization is by simple binning. Skips
 * the class attribute if set.
 * <p>
 * 
 */
public class Discretize {

    /** The number of bins to divide the attribute into */
    protected int numBins = 10;

    protected double[] min;

    protected double[] max;

    protected int attCount = 1;

    protected Instances inputFormat, outputFormat;

   

    /** Constructor - initialises the filter */
    public Discretize(int numBins) {
        this.numBins=numBins;
    }
    

    public Instance convert(Instance inst) {
        double[] values = new double[attCount];
        for (int i = 0; i < inputFormat.numAttributes(); i++) {
            if ((inputFormat.attribute(i).isNumeric())
                    && (inputFormat.classIndex() != i)) {
                values[i] = (int) ((inst.value(i) - min[i]) * numBins / (max[i] - min[i]));
                if (values[i] >= numBins)
                    values[i] = numBins - 1;
                else if (values[i] < 0)
                    values[i] = 0;
            } else {
                values[i] = inst.value(i);
            }

        }
        Instance newInst = new Instance(1.0, values);
        newInst.setDataset(outputFormat);
        return newInst;
    }

    public void setInputFormat(Instances instanceInfo) throws Exception {
        this.inputFormat = instanceInfo;
        attCount = instanceInfo.numAttributes();
        max = new double[attCount];
        min = new double[attCount];
        for (int i = 0; i < attCount; i++) {
            Attribute a = inputFormat.attribute(i);
            if (a.isNumeric()) {
                max[i] = a.getUpperNumericBound();
                min[i] = a.getLowerNumericBound();
            }
        }
        setOutputFormat();
    }

    protected void setOutputFormat() {

        FastVector attributes = new FastVector(attCount);
        for (int i = 0; i < inputFormat.numAttributes(); i++) {
            if ((inputFormat.attribute(i).isNumeric())
                    && (inputFormat.classIndex() != i)) {
                FastVector attribValues = new FastVector(1);
                if (numBins == 1) {
                    attribValues.addElement("'All'");
                } else {
                    for (int j = 0; j < numBins; j++) {

                        attribValues.addElement("'("
                                + Utils.doubleToString(
                                        ((max[i] - min[i]) * j / numBins)
                                                + min[i], 6)
                                + "-"
                                + Utils.doubleToString(((max[i] - min[i])
                                        * (j + 1) / numBins)
                                        + min[i], 6) + "]'");

                    }
                }
                attributes.addElement(new Attribute(inputFormat.attribute(i)
                        .name(), attribValues));
            } else
                attributes.addElement(inputFormat.attribute(i));
        }

        outputFormat = new Instances(inputFormat.relationName(), attributes, 0);
        outputFormat.setClassIndex(inputFormat.classIndex());
    }

}
