package weka.classifiers.rules.flora;

import java.util.*;

import weka.core.Attribute;
import weka.core.*;

public class Hypothesis {
    private DescriptionSet ADES;

    private DescriptionSet PDES;

    private DescriptionSet NDES;

    public Hypothesis() {
        ADES = new DescriptionSet();
        PDES = new DescriptionSet();
        NDES = new DescriptionSet();
    }

    /**
     * Returns a deep copy of this hypothesis
     */
    public Object clone() {
        Object clone = new Hypothesis();

        // deep clone implemented
        ((Hypothesis) clone).setDES((DescriptionSet) ADES.clone(),
                (DescriptionSet) NDES.clone(), (DescriptionSet) PDES.clone());
        return clone;
    }

    public DescriptionSet getADES() {
        return ADES;
    }

    private void setDES(DescriptionSet newADES, DescriptionSet newNDES,
            DescriptionSet newPDES) {
        ADES = newADES;
        NDES = newNDES;
        PDES = newPDES;
    }

    public DescriptionSet getNDES() {
        return NDES;
    }

    public DescriptionSet getPDES() {
        return PDES;
    }

    /**
     * Removes zero-coverage description items from all description sets
     * 
     */
    public void removeZeroCoverageDescriptionItems() {
        ADES.removeZeroCoverageDescriptionItems();
        PDES.removeZeroCoverageDescriptionItems();
        NDES.removeZeroCoverageDescriptionItems();
    }

    /**
     * Resets the positive and negative instances covered counters of
     * description items in all description sets
     */
    public void resetInstancesCoveredCounters() {
        ADES.resetInstancesCoveredCounters();
        NDES.resetInstancesCoveredCounters();
        PDES.resetInstancesCoveredCounters();
    }

    /**
     * Returns a flag indicating whether the hypothesis correctly predicts the
     * instance
     * 
     * @param instance
     * @return
     */
    public boolean correctlyCoversInstance(Instance instance) {
        boolean result = false;
        if ((instance.classValue() > 0 && classifyInstance(instance) > 0)
                || (instance.classValue() == 0 && classifyInstance(instance) == 0))
            result = true;
        return result;
    }

    /**
     * Returns a flag indicating whether the ADES sets of a hypothesis is
     * identical to this one
     * 
     * @param otherHypothesis
     * @return
     */
    public boolean IsIdenticalTo(Hypothesis otherHypothesis) {
        ArrayList otherDescriptionItems = otherHypothesis.getADES()
                .getDesciptionItems();
        ArrayList currentADESDescriptionItems = ((ArrayList) this.getADES()
                .getDesciptionItems().clone());
        // going through the description items of the other hypothesis
        for (int i = 0; i < otherDescriptionItems.size(); ++i) {
            // getting the current description item from the other hypothesis
            DescriptionItem otherDI = ((DescriptionItem) otherDescriptionItems
                    .get(i));
            // assume matching DIs can be found
            boolean isMatchFound = true;
            // going through the description items of this hypothesis
            for (int j = 0; j < currentADESDescriptionItems.size(); ++j) {
                isMatchFound = true;
                // check for equality between the description items, not
                // subsumption
                // need to make sure that all attributes in DI1 have the same
                // values in DI2 and vice versa
                DescriptionItem thisDI = ((DescriptionItem) currentADESDescriptionItems
                        .get(j));
                // go through all the attribute-value pairs in a description
                // item
                Enumeration attributeNames = otherDI.getAttributeTable().keys();
                while (attributeNames.hasMoreElements()) {
                    Integer attributeName = (Integer) attributeNames
                            .nextElement();
                    if (otherDI.getAttributeTable().get(attributeName) != thisDI
                            .getAttributeTable().get(attributeName)) {
                        isMatchFound = false;
                        break;
                    }
                }
                // if one item cannot be matched onto the other, continue with
                // next item
                if (!isMatchFound)
                    continue;

                attributeNames = thisDI.getAttributeTable().keys();
                while (attributeNames.hasMoreElements()) {
                    Integer attributeName = (Integer) attributeNames
                            .nextElement();
                    if (otherDI.getAttributeTable().get(attributeName)== null ||
                            otherDI.getAttributeTable().get(attributeName) != thisDI
                            .getAttributeTable().get(attributeName).intValue()) {
                        isMatchFound = false;
                        break;
                    }
                }
                // if a matching description item was found, continue with next
                // DI
                if (isMatchFound)
                    break;
            }
            // if a match to a DI was not found, hypotheses are not identical
            if (!isMatchFound)
                return false;
        }
        return true;
    }

    /**
     * Adds the current instance to the description sets of a hypothesis
     * 
     * @param instance
     */
    public void LearnFromInstance(Instance instance, Window window) {
        // calls different algorithms for the variants of the FLORA algorithm
        if (FloraClassifier.getFLORAMODE() < 4) {
            if (instance.classValue() > 0)
                learnFromPositiveInstanceFLORA123(instance, window);
            else
                learnFromNegativeInstanceFLORA123(instance, window);
        }
        // FLORA4 introduced probabilistic description item removal
        else {
            if (instance.classValue() > 0)
                learnFromPositiveInstanceFLORA4(instance, window);
            else
                learnFromNegativeInstanceFLORA4(instance, window);
        }
    }

    public void processADESFlora4(Window window) {
        // 0.compute the positive and negative class frequency intervals - to be
        // used for all DIs
        double positiveGammaU = window.getPositiveGammaU();
        double positiveGammaL = window.getPositiveGammaL();

        for (int i = 0; i < ADES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = ((DescriptionItem) ADES.getDesciptionItems()
                    .get(i));
            // 1. get the item's classification accuracy over instances in the
            // window
            int coveredPositiveCount = 0;
            int coveredCount = 0;
            for (int instanceIndex = 0; instanceIndex < window
                    .getInstanceCount(); ++instanceIndex) {
                Instance instance = (Instance) window.getInstances().get(
                        instanceIndex);
                // count correctly classfied positive and negative instances
                if (di.coversInstance(instance)) {
                    coveredCount++;
                    if (instance.classValue() > 0)
                        ++coveredPositiveCount;
                }
            }
            // 2. compute the accuracy confidence interval for this item
            double alphaU = FloraClassifier.getConfidenceInterval(
                    coveredPositiveCount, coveredCount);
            double alphaL = (2 * ((double) coveredPositiveCount / (double) coveredCount))
                    - alphaU;

            // a description item is kept in ADES if the lower endpoint of its
            // accuracy confidence interval
            // is greater than the class freqquency interval's upper endpoint
            // (alpha(l) > gamma(u))
            if (alphaL >= positiveGammaU) {
                // keep item in ADES
            }
            // a description item is moved to PDES if its accuracy confidence
            // interval overlaps with the
            // class frequency interval (alpha(u)>gamma(l))
            else if (alphaU >= positiveGammaL) {
                // move item to PDES
                // 1.remove from ADES
                ADES.getDesciptionItems().remove(di);
                // 2.add to PDES if not subsumed
                if (!PDES.isSubsumed(di))
                    PDES.getDesciptionItems().add(di);
            }
            // a description item is dropped completely if the upper endpoint of
            // its accuracy interval is lower than
            // the class frequency interval's lower endpoint (alpha(u) <
            // gamma(l)
            else if (alphaU < positiveGammaL) {
                // drop item
                ADES.getDesciptionItems().remove(di);
            }
        }
    }

    public void processPDESFlora4(Window window) {
        double positiveClassUpperBound = window.getPositiveGammaU();

        for (int i = 0; i < PDES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = ((DescriptionItem) PDES.getDesciptionItems()
                    .get(i));
            // 1. get the item's classification accuracy over instances in the
            // window
            int coveredPositiveCount = 0;
            int coveredCount = 0;
            for (int instanceIndex = 0; instanceIndex < window
                    .getInstanceCount(); ++instanceIndex) {
                Instance instance = (Instance) window.getInstances().get(
                        instanceIndex);
                // count correctly classfied positive and negative instances
                if (di.coversInstance(instance)) {
                    coveredCount++;
                    if (instance.classValue() > 0)
                        ++coveredPositiveCount;
                }
            }
            // 2. compute the accuracy confidence interval for this item
            double diConfidenceIntervalUpperBound = FloraClassifier
                    .getConfidenceInterval(coveredPositiveCount, coveredCount);
            double diConfidenceIntervalLowerBound = (2 * ((double) coveredPositiveCount / (double) coveredCount))
                    - diConfidenceIntervalUpperBound;

            // a description item is moved to ADES if the lower endpoint of its
            // accuracy confidence
            // interval is greater than the class freqquency interval's upper
            // endpoint (alpha(l) > gamma(u))
            if (diConfidenceIntervalLowerBound >= positiveClassUpperBound) {
                // move to ADES
                PDES.getDesciptionItems().remove(di);
                // adding item to ADES if not subsumed
                if (!ADES.isSubsumed(di))
                    ADES.getDesciptionItems().add(di);
            }
        }
    }

    public void processNDESFlora4(Window window) {
        double negativeClassUpperBound = window.getNegativeGammaU();
        double negativeClassLowerBound = window.getNegativeGammaL();

        for (int i = 0; i < NDES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = (DescriptionItem) NDES.getDesciptionItems()
                    .get(i);
            int coveredNegativeCount = 0;
            int coveredCount = 0;
            for (int instanceIndex = 0; instanceIndex < window
                    .getInstanceCount(); ++instanceIndex) {
                Instance instance = (Instance) window.getInstances().get(
                        instanceIndex);
                // count correctly classfied positive and negative instances
                if (di.coversInstance(instance)){
                    coveredCount++;
                if (instance.classValue() <= 0)
                    ++coveredNegativeCount;
                }
            }
            // 1. compute the accuracy interval for this item over instances in
            // the window - predicting negative instances
            double diConfidenceIntervalUpperBound = FloraClassifier
                    .getConfidenceInterval(coveredNegativeCount, coveredCount);
            // computing the lower boundary by subtracting from the mean
            // the difference between the mean and the upper boundary
            double diConfidenceIntervalLowerBound = (2 * ((double) coveredNegativeCount / (double) coveredCount))
                    - diConfidenceIntervalUpperBound;

            // keep item as long as alpha(l) > gamma(u) computed over negative
            // instances in the window
            if (diConfidenceIntervalLowerBound >= negativeClassUpperBound) {
                // keep item
            }
            // unacceptable DIs are simply dropped
            else {
                NDES.getDesciptionItems().remove(di);
            }
        }
    }

    /**
     * Adds the current instance to the positive description set of a hypothesis
     * 
     * @param positiveInstance
     */
    public void learnFromPositiveInstanceFLORA4(Instance positiveInstance,
            Window instanceWindow) {
        // update counters for pro DES
        if (!ADES.updateCoveringItems(positiveInstance, true, true)) {
            if (!Generalize(positiveInstance, instanceWindow)) {
                // save instance as an item in ADES
                ADES.includeItemFromInstance(positiveInstance);
            }
        }

        processADESFlora4(instanceWindow);

        // update counters for PDES
        PDES.updateCoveringItems(positiveInstance, true, true);
        processPDESFlora4(instanceWindow);

        // check whether item is not covered by the negative DES
        for (int i = 0; i < NDES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = (DescriptionItem) NDES.getDesciptionItems()
                    .get(i);
            // the instance is covered by a description item of the opposite
            // class
            if (di.coversInstance(positiveInstance)) {
                // remove description item from the opposite description set
                di.setPositiveCount(di.getPositiveCount() + 1);
            }
        }
        processNDESFlora4(instanceWindow);
    }

    /**
     * Adds the current instance to the negative description set of a hypothesis
     * 
     * @param negativeInstance
     */
    public void learnFromNegativeInstanceFLORA4(Instance negativeInstance,
            Window instanceWindow) {
        // update counters for pro DES
        if (!NDES.updateCoveringItems(negativeInstance, false, true)) {
            if (!Generalize(negativeInstance, instanceWindow)) {
                // save instance as an item in ADES
                NDES.includeItemFromInstance(negativeInstance);
            }
        }
        processNDESFlora4(instanceWindow);
        // update counters for PDES
        PDES.updateCoveringItems(negativeInstance, false, true);
        processPDESFlora4(instanceWindow);

        // check whether item is not covered by ADES
        for (int i = 0; i < ADES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = (DescriptionItem) ADES.getDesciptionItems()
                    .get(i);
            // the instance is covered by a description item of the positive
            // class
            if (di.coversInstance(negativeInstance))
                di.setNegativeCount(di.getNegativeCount() + 1);
        }
        processADESFlora4(instanceWindow);
    }

    /**
     * Adds the current instance to the positive description set of a hypothesis
     * 
     * @param positiveInstance
     */
    public void learnFromPositiveInstanceFLORA123(Instance positiveInstance,
            Window instanceWindow) {
        // update counters for pro DES
        if (!ADES.updateCoveringItems(positiveInstance, true, true)) {
            if (!Generalize(positiveInstance, instanceWindow)) {
                // save instance as an item in ADES
                ADES.includeItemFromInstance(positiveInstance);
            }
        }
        // update counters for PDES
        PDES.updateCoveringItems(positiveInstance, true, true);

        // check whether item is not covered by the negative DES
        for (int i = 0; i < NDES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = (DescriptionItem) NDES.getDesciptionItems()
                    .get(i);
            // the instance is covered by a description item of the opposite
            // class
            if (di.coversInstance(positiveInstance)) {
                // remove description item from the opposite description set
                NDES.getDesciptionItems().remove(di);
                di.setPositiveCount(1);
                // add a description item for a positive instance
                if (di.getAttributeTable().size() != 0)
                    PDES.includeDescriptorItem(di, true);
            }
        }
    }

    /**
     * Adds the current instance to the negative description set of a hypothesis
     * 
     * @param negativeInstance
     */
    public void learnFromNegativeInstanceFLORA123(Instance negativeInstance,
            Window instanceWindow) {
        // update counters for pro DES
        if (!NDES.updateCoveringItems(negativeInstance, false, true)) {
            if (!Generalize(negativeInstance, instanceWindow)) {
                // save instance as an item in ADES
                NDES.includeItemFromInstance(negativeInstance);
            }
        }
        // update counters for PDES
        PDES.updateCoveringItems(negativeInstance, false, true);

        // check whether item is not covered by ADES
        for (int i = 0; i < ADES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = (DescriptionItem) ADES.getDesciptionItems()
                    .get(i);
            // the instance is covered by a description item of the positive
            // class
            if (di.coversInstance(negativeInstance)) {
                // remove description item from the positive description set
                ADES.getDesciptionItems().remove(di);
                di.setNegativeCount(1);
                // add a description item for a positive instance
                if (di.getAttributeTable().size() != 0)
                    PDES.includeDescriptorItem(di, false);
            }
        }
    }

    /**
     * Removes the current instance from the description sets of a hypothesis
     * 
     * @param instance
     */
    public void ForgetInstance(Instance instance, Window window) {
        if (instance.classValue() > 0)
            ForgetPositiveInstance(instance, window);
        else
            ForgetNegativeInstance(instance, window);
    }

    /**
     * Removes a positive instance from the description sets of a hypothesis
     * 
     * @param positiveInstance
     */
    public void ForgetPositiveInstance(Instance positiveInstance, Window window) {
        // in flora4 items in NDES can cover positive items and will be kept as
        // long as they have good predictive power
        if (FloraClassifier.getFLORAMODE() == 4) {
            NDES.updateCoveringItems(positiveInstance, false, false);
            processNDESFlora4(window);
        }
        // update counters for pro DES
        ADES.updateCoveringItems(positiveInstance, true, false);
        if (FloraClassifier.getFLORAMODE() == 4) {
            processADESFlora4(window);
        }
        // update counters for PDES
        for (int i = 0; i < PDES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = (DescriptionItem) PDES.getDesciptionItems()
                    .get(i);
            if (di.coversInstance(positiveInstance)) {
                di.setCount(true, di.getCount(true) - 1);
                // implements FLORA1-2-3 description item removal logic
                if (FloraClassifier.getFLORAMODE() < 4) {
                    if (di.getCount(true) == 0) {
                        PDES.getDesciptionItems().remove(di);
                        NDES.removeSubsumedItems(di);
                        // instance being forgotten is positive - all the rest
                        // are negative
                        // add item to NDES
                        NDES.getDesciptionItems().add(di);

                    }
                }

            }
        }
        // implements FLORA4 description item removal logic
        if (FloraClassifier.getFLORAMODE() == 4) {
            processPDESFlora4(window);
        }
    }

    /**
     * Removes a negative instance from the description sets of a hypothesis
     * 
     * @param negativeInstance
     */
    public void ForgetNegativeInstance(Instance negativeInstance, Window window) {
        // in flora4 items in ADES can cover negative items and will be kept as
        // long as they have good predictive power
        if (FloraClassifier.getFLORAMODE() == 4) {
            // update counts of covered items in the description items in ADES
            ADES.updateCoveringItems(negativeInstance, false, false);
            // check whether all description items satisfy the conditions to
            // remain in ADES
            processADESFlora4(window);
        }

        // update counters for NDES
        NDES.updateCoveringItems(negativeInstance, false, false);
        if (FloraClassifier.getFLORAMODE() == 4) {
            // check whether all description items satisfy the conditions to
            // remain in NDES
            processNDESFlora4(window);
        }
        // update counters for PDES
        for (int i = 0; i < PDES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = (DescriptionItem) PDES.getDesciptionItems()
                    .get(i);
            if (di.coversInstance(negativeInstance)) {
                di.setCount(false, di.getCount(false) - 1);
                // implements FLORA1-2-3 description item removal logic
                if (FloraClassifier.getFLORAMODE() < 4) {
                    if (di.getCount(false) == 0) {
                        PDES.getDesciptionItems().remove(di);
                        // remove any items that might be subsumed by this one
                        ADES.removeSubsumedItems(di);
                        // the last instance being forgotten is negative
                        // the item covers only positive instances now
                        ADES.getDesciptionItems().add(di);

                    }
                }
            }
        }
        // implements FLORA4 description item removal logic
        if (FloraClassifier.getFLORAMODE() == 4)
            processPDESFlora4(window);
    }

    /**
     * Performs a minimal generalization on an item in a description set to
     * accommodate an instance
     * 
     * @param instance
     * @return
     */
    public boolean Generalize(Instance instance, Window instanceWindow) {
        boolean success = false;
        ArrayList orderedItems;
        // positive class - try to generalize an item in ADES to accommodate it
        if (instance.classValue() > 0)
            orderedItems = GetGeneralizationCandidates(instance, ADES);
        // negative class - try to generalize an item in NDES to accommodate it
        else
            orderedItems = GetGeneralizationCandidates(instance, NDES);

        // go through the list of ordered candidates trying to make them general
        // enough to accommodate the instance
        while (orderedItems.size() != 0) {
            // clone description item, in order to save it in case the
            // generalized item was subsumed
            DescriptionItem originalDI = ((DescriptionItemContainer) orderedItems
                    .get(0)).getDescriptionItem();
            DescriptionItem newDescriptionItem = ((DescriptionItem) originalDI.clone());
            // go through all the attributes, find the restrictive ones and
            // relax them enough to accommodate this instance
            Enumeration thisItemAttributeNames = newDescriptionItem
                    .getAttributeTable().keys();
            while (thisItemAttributeNames.hasMoreElements()) {
                Integer currentAttributeName = (Integer) thisItemAttributeNames
                        .nextElement();
                int thisAValue = (Integer) newDescriptionItem
                        .getAttributeTable().get(currentAttributeName);
                // only compare attribute values if there are restrictions
                // imposed

                // found the attribute, comparing the values
                if (instance.value(currentAttributeName) != thisAValue) {
                    // generalize this attribute in the description set by
                    // removing this condition
                    newDescriptionItem.getAttributeTable().remove(
                            currentAttributeName);
                }
            }

            // if the attribute table is empty, an item does not impose any
            // variable domain restrictions
            // it will subsume every DI and cover every instance
            if (newDescriptionItem.getAttributeTable().size() == 0) {
                // if item does not impose any restrictions, it will subsume
                // every DI and cover every instance
                // remove original DI from candidate seeds
                orderedItems.remove(0);
                continue;
            }

            // then validate the generalized item to make sure it does not
            // subsume items in the other two DES's
            // make sure the new item does not subsume some item in PDES
            if (newDescriptionItem.overlapDESItems(PDES)) {
                // if item is subsumed, remove original DI from candidate seeds
                orderedItems.remove(0);
                continue;
            }

            // make sure the new item does not subsume some item in NDES
            if (instance.classValue() > 0
                    && newDescriptionItem.overlapDESItems(NDES)) {
                // if item is subsumed, remove original DI from candidate seeds
                orderedItems.remove(0);
                continue;
            }

            // make sure the new item does not subsume some item in ADES
            if (instance.classValue() <= 0
                    && newDescriptionItem.overlapDESItems(ADES)) {
                // if item is subsumed, remove original DI from candidate seeds
                orderedItems.remove(0);
                continue;
            }

            // item is not subsumed by any DES - saving it to the appropriate
            // DES
            // 1. remove original description item from list
            // 2.reprocess the window to compute the actual instances covered
            newDescriptionItem.setNegativeCount(0);
            newDescriptionItem.setPositiveCount(0);
            for (int instanceIndex = 0; instanceIndex < instanceWindow
                    .getInstanceCount(); ++instanceIndex) {
                Instance windowInstance = ((Instance) instanceWindow
                        .getInstances().get(instanceIndex));
                boolean instanceClass = instance.classValue() > 0;
                // if the description item covers an instance, increment the
                // corresponding counter
                if (newDescriptionItem.coversInstance(windowInstance))
                    newDescriptionItem.setCount(instanceClass,
                            newDescriptionItem.getCount(instanceClass) + 1);
            }
            // 3. save extended item to list
            if (instance.classValue() > 0) {
                ADES.getDesciptionItems().remove(originalDI);
                // removes items which are subsumed by the new one
                ADES.removeSubsumedItems(newDescriptionItem);
                ADES.getDesciptionItems().add(newDescriptionItem);
            } else {
                NDES.getDesciptionItems().remove(originalDI);
                // removes items which are subsumed by the new one
                NDES.removeSubsumedItems(newDescriptionItem);
                NDES.getDesciptionItems().add(newDescriptionItem);
            }
            success = true;
            break;
        }
        return success;
    }

    /**
     * Returns a list of description items in descending order of fitness to
     * generalize the instance
     */
    public ArrayList GetGeneralizationCandidates(Instance instance,
            DescriptionSet DES) {
        ArrayList candidates;
        Vector unorderedCandidates = new Vector();
        // go through all items in the description set
        for (int i = 0; i < DES.getDesciptionItems().size(); ++i) {
            // count the distances of the items from the instance
            DescriptionItem di = (DescriptionItem) DES.getDesciptionItems()
                    .get(i);
            DescriptionItemContainer dic = new DescriptionItemContainer(di, di
                    .getDistanceFromInstance(instance));
            unorderedCandidates.add(dic);
        }

        candidates = QuickSorter.quickSort(unorderedCandidates);

        return candidates;
    }

    public double getADESLiteralCount() {
        double ADESLiteralCount = 0;
        for (int i = 0; i < ADES.getDesciptionItems().size(); ++i) {
            DescriptionItem descriptionItem = ((DescriptionItem) ADES
                    .getDesciptionItems().get(i));
            ADESLiteralCount += descriptionItem.getAttributeTable().size();
        }
        return ADESLiteralCount;
    }

    public double classifyInstance(Instance instance) {
        double result = 0;
        // go through all the description items, verifying that
        // at least one of them covers the instance
        for (int i = 0; i < ADES.getDesciptionItems().size(); ++i) {
            DescriptionItem di = ((DescriptionItem) ADES.getDesciptionItems()
                    .get(i));
            if (di.coversInstance(instance)) {
                result = 1;
                break;
            }
        }
        return result;
    }
}
