/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Rise3.java
 *    Copyright (C) 2004 Jrgen Beringer
 */
package weka.classifiers.rules.rise;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;

/**
 * This class implements a rule learner, "Rule Induction from a Set of Exemplars" (RISE) 
 * which is called "RISE 3.1 System".
 * For more information see <p>
 * 
 * Pedro Domingos. <i>From Instances to Rules: A Comparison 
 * of Biases</i>, Proceedings of the Third International 
 * Workshop on Multistrategy Learning (pp. 147-154), 1996. 
 * Harpers Ferry, WV: AAAI Press <br>
 * 
 * Pedro Domingos. <i>The RISE System: Conquering Without 
 * Separating</i>, Proceedings of the Sixth IEEE International 
 * Conference on Tools with Artificial Intelligence (pp. 704-707), 
 * 1994. New Orleans, LA: IEEE Computer Society Press.</p>
 * 
 * 
 * Valid options are: <p>
 *
 * <pre>
 * -N numeric_flag    Which Normalization of numeric attributes are used.
 * possible values for numeric_flag are:
 *     no             use no numeric normalization.
 *     range          use range normalization (max-min=1.0). (default)
 *     [0.0-0.5]      use quantile normalization with the given quantile.
 *
 * -V                 Don't use the VDM distance for nominal attributes.
 *
 * -T method          Which tiebreak strategy is used for rules with same distance.
 * possible values for method are:
 *     laplace        Choose rule with highest Laplace accuracy (default)
 *     specific       Choose most specific rule
 *     vote           Use frequency-based voting
 *
 * -L                 Use the local stopping strategy for rule induction.
 *
 * -S                 Use the simplification strategy on the ruleset.
 * 
 * -I                 Only induce new rules which strictly increase the accuracy.
 * 
 * -E                 Induce new rules which not decrease the accuracy.
 *</pre>
 *
 * The default strategy of RISE is to induce rules which not decrease the accuracy but
 * at least one new rule has to increase the accuracy to continue the induction process.
 * 
 * This implementation is tested with some UCI datasets and compared with the origin results.
 *
 * @author Jrgen Beringer (beringer@mathematik.uni-marburg.de)
 * @version $Revision: 1.4 $
 */
public class Rise31 extends Classifier implements OptionHandler,
		WeightedInstancesHandler {

	/**
	 * An extended Rule class for Rise2-Rules. This class support the method
	 * grow(Instance) and simple Rules of one Instance.
	 *  
	 */
	public static class Rule extends SimpleConjunctiveRule {

		/**counter for the rule-ID */
		public static long nextID = Long.MIN_VALUE + 1;

		/** Array of used attributes to learn. */
		protected int[] usedAttribute;

		/** Flag, whether growing of this rule is finished.*/
		protected boolean finished = false;

		/** List of covered InstanceInfo*/
		public List coveredInstInfo;

		/** unique rule-ID*/
		public long ruleID;

		/** Criterion value of this rule*/
		public double criterion;

		/** if this rule is deleted from the rulelist*/
		public boolean deleted = false;

		/**
		 * Create an empty rule.
		 * 
		 * @param instances
		 *            Header of data.
		 * @param usedAttribute
		 *            Array of used attributes.
		 */
		public Rule(Instances instances, int[] usedAttribute) {
			super(instances, false);
			this.usedAttribute = usedAttribute;
			ruleID = nextID++;
			coveredInstInfo = new ArrayList();
		}

		/**
		 * Create a simple rule for the instance.
		 * 
		 * @param instances
		 *            Header of data.
		 * @param instance
		 *            Instance for this simple rule.
		 * @param usedAttribute
		 *            Array of used attributes with type of attribute (numeric/nominal).
		 */
		public Rule(Instances instances, InstanceInfo instance,
				int[] usedAttribute) {
			super(instances, false);
			this.usedAttribute = usedAttribute;
			ruleID = nextID++;
			coveredInstInfo = new ArrayList();
			for (int i = 0; i < attrCount; i++) {
				if (usedAttribute[i] > 0 && !instance.inst.isMissing(i)) {
					if (usedAttribute[i] == NOMINAL)
						setNominalValue(i, (int) instance.inst.value(i));
					else if (usedAttribute[i] == NUMERIC) {
						double val = instance.inst.value(i);
						setNumericRange(i, val, val);
					}
				}
			}
			addDistribution(instance);
			computeClass();
		}

		/**
		 * Copyconstructor
		 * 
		 * @param rule
		 *            Rule to copy
		 */
		public Rule(Rule rule) {
			super(rule);
			storeCovered = rule.storeCovered;
			this.usedAttribute = rule.usedAttribute;
			coveredInstInfo = new ArrayList();
		}

		/*public boolean equals(Object ob){
		 if (ob==this) return true;
		 else return false;
		 }*/

		/**
		 * Compute the distribution with instances of the header object.
		 */
		public void computeDistribution() {
			resetDistribution();
			Iterator it = coveredInstInfo.iterator();
			while (it.hasNext()) {
				InstanceInfo instance = (InstanceInfo) it.next();
				addDistribution(instance);
			}
		}

		/**
		 * Compute the class distribution of this List of InstanceInfo
		 * @param coveredInstInfo List of data
		 */
		public void computeDistribution(List coveredInstInfo) {
			resetDistribution();
			Iterator it = coveredInstInfo.iterator();
			while (it.hasNext()) {
				InstanceInfo instance = (InstanceInfo) it.next();
				addDistribution(instance);
			}
		}

		/**
		 * Add an instance to the distribution.
		 * 
		 * @param inst
		 *            Instance to add
		 */
		public void addDistribution(InstanceInfo inst) {
			if (covers(inst.inst)) {
				addClass((int) inst.inst.classValue(), inst.inst.weight());
				coveredInstInfo.add(inst);
			}
		}

		/**
		 * Remove an instance to the distribution.
		 * 
		 * @param inst
		 *            Instance to add
		 */
		public void removeDistribution(InstanceInfo inst) {
			if (covers(inst.inst)) {
				addClass((int) inst.inst.classValue(), -inst.inst.weight());
				coveredInstInfo.remove(inst);
			}
		}

		/**
		 * Expand the rule with this Instance.
		 * 
		 * @param instance
		 *            Instance to expand this rule.
		 */
		public void grow(Instance instance) {
			try {
				for (int i = 0; i < attrCount; i++) {
					if (usedAttribute[i] > 0) {
						if (usedAttribute[i] == NOMINAL) {
							if (values[i] >= 0) {
								if (instance.isMissing(i))
									setNominalValue(i, -1);
								else if (values[i] != (int) instance.value(i))
										setNominalValue(i, -1);
							}
						} else if (usedAttribute[i] == NUMERIC) {
							if (bounds[i] != null) {
								if (instance.isMissing(i))
									resetNumericRange(i);
								else {
									double val = instance.value(i);
									if (Double.isNaN(val)) continue;
									if (val < bounds[i][0])
										setNumericRange(i, val, bounds[i][1]);
									else if (val > bounds[i][1])
											setNumericRange(i, bounds[i][0],
													val);
								}
							}
						}
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		/**
		 * Whether growing this rule is finished.
		 * @return finished-flag
		 */
		protected boolean isFinished() {
			return finished;
		}

		/**
		 * Set the finish-flag of this rule.
		 * @param finished flag value
		 */
		protected void setFinished(boolean finished) {
			this.finished = finished;
		}
	}

	/**ID for the next Instance */
	protected long nextInstID = Long.MIN_VALUE + 1;

	/**Wrapper class for an Instance to store some more informations. */
	protected class InstanceInfo {

		/**ID of this Instance */
		public long instID;

		/**The Instance*/
		public Instance inst;

		/**The rule which was produced by this instance.*/
		public Rule ownRule;

		/**List of nearest Rules to this Instance.*/
		public List rules = new ArrayList();

		/**Distance to the nearest rules.*/
		public double ruledist = Double.MAX_VALUE;

		/**Flag, if this Instance is classified correctly.*/
		public boolean correct;

		/**
		 * Simple Constructor of an Instance.
		 * @param inst the Instance
		 */
		public InstanceInfo(Instance inst) {
			this.inst = inst;
			instID = nextInstID++;
		}

		/**
		 * Compare this Instance with another (only ID will considered).
		 */
		public boolean equals(Object ob) {
			InstanceInfo inf = (InstanceInfo) ob;
			return inf.instID == instID;
		}

		/**
		 * Return a string representation of this instance.
		 */
		public String toString() {
			if (inst != null)
				return inst.toString();
			else
				return "empty";
		}
	}

	/** Constant for nominal attributes*/
	public static int NOMINAL = NormDistance.NOMINAL;

	/** Constant for numeric attributes*/
	public static int NUMERIC = NormDistance.NUMERIC;

	/** Comparator which compare rules with their criterion value.*/
	protected Comparator ruleCriterionComparator = new Comparator() {

		public int compare(Object candidate1, Object candidate2) {
			Rule rule1 = (Rule) candidate1;
			Rule rule2 = (Rule) candidate2;
			if (rule1.criterion > rule2.criterion)
				return 1;
			else if (rule1.criterion < rule2.criterion)
				return -1;
			else
				return 0;
		}
	};

	/** Comparator which compare rules with their number of covered Instances.*/
	protected Comparator ruleCoveredSizeComparator = new Comparator() {

		public int compare(Object candidate1, Object candidate2) {
			Rule rule1 = (Rule) candidate1;
			Rule rule2 = (Rule) candidate2;
			if (rule1.coveredInstInfo.size() > rule2.coveredInstInfo.size())
				return 1;
			else if (rule1.coveredInstInfo.size() < rule2.coveredInstInfo
					.size())
				return -1;
			else
				return 0;
		}
	};

	/** Metadata of the used instances*/
	protected Instances header;

	/** Number of attributes */
	protected int attrCount;

	/** Index of the class attribute */
	protected int classIndex;

	/** Number of classes. */
	protected int classCount;

	/** List of instances. */
	protected List dataList;

	/** Distance object to compute the distances. */
	protected NormDistance distance;

	/** List of used attributes to generate rules and to compute the distances. */
	protected int[] usedAttribute;

	/** List of the indices of used attributes.*/
	protected List usedAttributeList;

	/** List of the nearest Data for a rule*/
	protected List nearest;

	/** Used quantile for quantile range normalization (default 0.05)*/
	protected double quantile = 0.05;

	/** Flag of used normalization for numeric attributes.*/
	protected int numeric = NormDistance.RANGE_NUMERIC;//QUANTILE_NUMERIC;

	/** Flag of used distance for nominal attributes.*/
	protected boolean useVDM = true;

	/** Flag for global stopping (default true).*/
	protected boolean globalStop = true;

	/** Choose rule with highest Laplace accuracy. */
	public static final int HIGHEST_LAPLACE_TIEBREAK = 0;

	/** Choose most specific rule.*/
	public static final int MOST_SPECIFIC_TIEBREAK = 1;

	/** Use frequency-based voting.*/
	public static final int FREQUENCY_BASED_VOTING_TIEBREAK = 2;

	/** Flag for used tiebreak strategy.*/
	protected int tieBreak = HIGHEST_LAPLACE_TIEBREAK;

	/** Flag if simplification is used (default false)*/
	protected boolean simplification = false;

	protected boolean increase = false;
	
	protected boolean rise3Stop = true;

	/** List of rules*/
	protected List rules;

	/** Array of the important rules.*/
	protected boolean[] importantRules;

	/** current Accuracy */
	protected double ruleAccuracy;

	/**
	 * List of instances, which has to be updated if the new rule will be inserted.
	 */
	protected List updateList = new ArrayList();

	/**
	 * Liste of removed rules for updating in instances.
	 */
	protected List removedRules = new ArrayList();

	/** List of new rules to insert into the rulelist.*/
	protected List newRules = new ArrayList();

	/**
	 * Build the rules from the instances.
	 */
	public void buildClassifier(Instances data) throws Exception {

		header = new Instances(data, 0);
		attrCount = data.numAttributes();
		classIndex = data.classIndex();
		classCount = data.numClasses();
		computeUsedAttributes();
		if (numeric == NormDistance.QUANTILE_NUMERIC)
			distance = new NormDistance(data, usedAttribute, quantile, useVDM);
		else
			distance = new NormDistance(data, usedAttribute,
					numeric == NormDistance.RANGE_NUMERIC, useVDM);
		dataList = new ArrayList();

		Enumeration enum1 = data.enumerateInstances();
		while (enum1.hasMoreElements()) {
			Instance inst = (Instance) enum1.nextElement();
			Instance cp = (Instance) (inst).copy();
			cp.setDataset(header);
			dataList.add(new InstanceInfo(cp));
		}
		initializeRules();
		induceRules();
		if (simplification) {
			unimportant();
			if (tieBreak == HIGHEST_LAPLACE_TIEBREAK) optimizeAll();
		}
	}

	/**
	 * Set the default instance based ruleset and compute the next rule of each
	 * element.
	 *  
	 */
	protected void initializeRules() {
		rules = new ArrayList();
		nearest = new ArrayList();
		Iterator it = dataList.iterator();

		while (it.hasNext()) {
			InstanceInfo instinf = (InstanceInfo) it.next();
			Instance inst = instinf.inst;
			for (int i = 0; i < attrCount; i++) {
				if (usedAttribute[i] > 0 && inst.isMissing(i)) continue;
			}
			boolean found = false;
			for (int i = 0; i < rules.size(); i++) {
				Rule rule = (Rule) rules.get(i);
				if (rule.covers(inst)) {
					instinf.rules.add(rule);
					instinf.ruledist = 0;
					rule.addDistribution(instinf);
					rule.computeClass();
					found = true;
					break;
				}
			}
			if (!found) {
				Rule rule = new Rule(header, instinf, usedAttribute);
				instinf.ownRule = rule;
				rules.add(rule);
			}
		}
		ruleAccuracy = accuracy();
	}

	/**
	 * Compute complete Accuracy of the ruleset.
	 * 
	 * @param rules
	 *            List of rules
	 * @return Accuracy of the rules
	 */
	protected double accuracy() {
		int correct = 0;
		int countAll = 0;

		Iterator it = rules.iterator();
		while (it.hasNext()) {
			Rule rule = (Rule) it.next();
			rule.criterion = criterion(rule);
		}

		it = dataList.iterator();
		while (it.hasNext()) {
			InstanceInfo instinf = (InstanceInfo) it.next();
			Instance inst = instinf.inst;
			countAll++;
			if (instinf.rules.size() == 0) findNearestRules(instinf);
			int clVal = getMaxClass(instinf);

			if (clVal == (int) inst.classValue()) {
				correct++;
				instinf.correct = true;
			} else {
				instinf.correct = false;
			}

		}
		return correct / (double) countAll;
	}
	/**
	 * Get the Distribution of this instance.
	 * @param instinf instance
	 * @return distribution
	 */
	protected double[] getDistribution(InstanceInfo instinf) {
		return getDistribution(instinf.rules);
	}
	/**
	 * Compute the class distribution for this list of rules.
	 * @param nearestRules list of nearest rules
	 * @return sum distribution of this rulelist
	 */
	protected double[] getDistribution(List nearestRules) {
		double[] distribution = new double[classCount];
		switch (tieBreak) {
		case HIGHEST_LAPLACE_TIEBREAK:
			Collections.sort(nearestRules, ruleCriterionComparator);
			int pos = nearestRules.size() - 1;
			if (pos < 0) return null;
			Rule rule = (Rule) nearestRules.get(pos);
			double crit = rule.criterion;
			distribution = rule.getRelativeDistribution();
			int count = 1;
			double newcrit = 0;
			do {
				pos--;
				if (pos >= 0) {
					rule = (Rule) nearestRules.get(pos);
					newcrit = rule.criterion;
					if (newcrit == crit) {
						addDistribution(distribution, rule
								.getRelativeDistribution());
						count++;
					}
				}
			} while (newcrit == crit && pos > 0);
			if (count > 1) {
				for (int i = 0; i < classCount; i++) {
					distribution[i] /= count;
				}
			}
			break;
		case FREQUENCY_BASED_VOTING_TIEBREAK:
			Iterator it = nearestRules.iterator();
			count = 0;
			while (it.hasNext()) {
				rule = (Rule) it.next();
				addDistribution(distribution, rule.getRelativeDistribution());
				count++;
			}
			if (count > 1) {
				for (int i = 0; i < classCount; i++) {
					distribution[i] /= count;
				}
			}
			break;
		case MOST_SPECIFIC_TIEBREAK:
			Collections.sort(nearestRules, ruleCoveredSizeComparator);
			pos = nearestRules.size() - 1;
			if (pos < 0) return null;
			rule = (Rule) nearestRules.get(pos);
			crit = rule.coveredInstInfo.size();
			distribution = rule.getRelativeDistribution();
			count = 1;
			newcrit = 0;
			do {
				pos--;
				if (pos >= 0) {
					rule = (Rule) nearestRules.get(pos);
					newcrit = rule.coveredInstInfo.size();
					if (newcrit == crit) {
						addDistribution(distribution, rule
								.getRelativeDistribution());
						count++;
					}
				}
			} while (newcrit == crit && pos > 0);
			if (count > 1) {
				for (int i = 0; i < classCount; i++) {
					distribution[i] /= count;
				}
			}
			break;
		}

		return distribution;
	}
	/**
	 * Get the index of the biggest value in this array.
	 * @param doubles array
	 * @return Index of the biggest value.
	 */
	protected int maxIndex(double[] doubles) {

		double max1 = -1;
		double max2 = -1;

		int maxIndex = 0;

		for (int i = 0; i < doubles.length; i++) {
			if ((i == 0) || (doubles[i] > max1)) {
				maxIndex = i;
				max2 = max1;
				max1 = doubles[i];
			} else if (doubles[i] > max2) {
				max2 = doubles[i];
			}
		}
		if (max1 > max2)
			return maxIndex;
		else
			return -maxIndex - 1;
	}
	/**
	 * Find the class from the distribution of the nearest rules of this Instance.
	 * @param instinf Instance
	 * @return
	 */
	protected int getMaxClass(InstanceInfo instinf) {
		double[] distrib = getDistribution(instinf);
		if (distrib == null) {
			System.out.println("ERROR");
		}
		int clVal = maxIndex(distrib);
		return clVal;
	}
	/**
	 * Search the nearest Rule of this Instance
	 * @param instinf Instance
	 */
	protected void findNearestRules(InstanceInfo instinf) {
		double minDist = Double.MAX_VALUE;
		Iterator it = rules.iterator();
		while (it.hasNext()) {
			Rule rule = (Rule) it.next();
			if (instinf.ownRule == rule) {
				if (rule.coveredInstInfo.size() == 1)
					continue;
				else
					instinf.ownRule = null;
			}
			double dist = distance(rule, instinf.inst);
			if (minDist > dist) {
				minDist = dist;
				instinf.rules.clear();
				instinf.rules.add(rule);
			} else if (dist == minDist) {
				instinf.rules.add(rule);
			}
		}
		instinf.ruledist = minDist;
	}

	/**
	 * Induce rules for the dataset until no better ruleset is found.
	 */
	protected void induceRules() {
		boolean changed = false;
		do {
			changed = false;
			Object[] lastRules = rules.toArray();
			int size = lastRules.length;
			for (int i = 0; i < size; i++) {
				Rule rule = (Rule) lastRules[i];
				if (rule.deleted) continue;
				if (!globalStop && rule.isFinished()) continue;
				List nearest = findNearest(rule);
				if (nearest == null) continue;
				boolean more = false;
				if (nearest.size() > 1) {
					more = true;
				}
				Iterator newRules = generalize(rule, nearest);
				double bestAcc = 0.0;
				Rule bestNew = rule;
				List bestUp = null;
				while (newRules.hasNext()) {
					Rule newRule = (Rule) newRules.next();
					double newAccuracy = updateAccuracy(rule, newRule);
					if (newAccuracy > bestAcc
							|| (newAccuracy == bestAcc && newRule.criterion > bestNew.criterion)) {
						bestAcc = newAccuracy;
						bestNew = newRule;
						if (more) {
							bestUp = new ArrayList(updateList);
						} else
							bestUp = updateList;
					}
				}
				boolean update = false;
				if (increase)
					update = (bestAcc > ruleAccuracy);
				else
					update = (bestAcc >= ruleAccuracy);
				if (update) {
					if (increase || !rise3Stop) changed = true;
					else {
						if (bestAcc > ruleAccuracy) changed = true;
					}
					updateList = bestUp;
					if (!rule.deleted) {
						rule.deleted = true;
						remove(rules, rule);
					}
					rules.add(bestNew);
					optimize(bestNew, rules.size() - 1);
					updateNearestRules(rule, bestNew);
					ruleAccuracy = bestAcc;

				}
				if (!rule.deleted) rule.setFinished(true);
			}
		} while (changed);
		//System.out.println("fertig");
		//induceRules();
	}

	protected void removeAll(List list, List toRemove) {
		Iterator it = toRemove.iterator();
		while (it.hasNext()) {
			remove(list, it.next());
		}
	}

	protected boolean remove(List list, Object ob) {
		Iterator it = list.iterator();
		while (it.hasNext()) {
			if (it.next() == ob) {
				it.remove();
				return true;
			}
		}
		return false;
	}

	protected int indexOf(List list, Object ob) {
		Iterator it = list.iterator();
		int ind = 0;
		while (it.hasNext()) {
			if (it.next() == ob) { return ind; }
			ind++;
		}
		return -1;
	}

	/**
	 * Optimize ruleset by removing rules which are covered complete by the new
	 * rule.
	 * 
	 * @param rules
	 *            List of rules
	 * @param pos
	 *            index of the new rule
	 */
	protected void optimize(Rule newRule, int index) {
		int cl = newRule.getClassValue();
		Iterator it = rules.iterator();
		boolean delete = false;
		while (it.hasNext()) {
			Rule rule = (Rule) it.next();
			if (rule != newRule) {
				if (cl == rule.getClassValue()) {
					if (rule.equals(newRule)) {
						delete = true;
						break;
					}
				}
			}
		}
		if (delete) {
			newRule.deleted = true;
			rules.remove(index);
			//if (rules.contains(newRule)) System.out.println("error deleting");
		}
	}

	/**
	 * Remove each rule, which is covered by another rule of same class.
	 * @param rules List of rules
	 */
	protected void optimizeAll() {
		for (int j = 0; j < rules.size(); j++) {
			Rule newRule = (Rule) rules.get(j);
			for (int i = 0; i < rules.size(); i++) {
				if (i == j) continue;
				Rule rule = (Rule) rules.get(i);
				if (newRule.contains(rule)) {
					if ((tieBreak == HIGHEST_LAPLACE_TIEBREAK)
							&& criterion(newRule) > criterion(rule)) {
						rules.remove(i);
						if (i < j) j--;
						i--;
					}
				}
			}
		}
	}

	/**
	 * Update the rule list for the new rule.
	 * @param oldRule old rule to replace
	 * @param newRule new rule
	 */
	protected void updateNearestRules(Rule oldRule, Rule newRule) {
		boolean toremove = removedRules.size() > 0;
		Iterator it = updateList.iterator();
		while (it.hasNext()) {
			InstanceInfo instinf = (InstanceInfo) it.next();
			Instance inst = instinf.inst;
			double dist = distance(newRule, inst);
			boolean old = (indexOf(instinf.rules, oldRule) >= 0);
			if (dist < instinf.ruledist) {
				if (!newRule.deleted) {
					instinf.rules.clear();
					instinf.rules.add(newRule);
					instinf.ruledist = dist;
					if (newRule.getClassValue() == (int) inst.classValue()) {
						instinf.correct = true;
					} else
						instinf.correct = false;
				} else {
					if (toremove) removeAll(instinf.rules, removedRules);
					if (old) remove(instinf.rules, oldRule);
					if (getMaxClass(instinf) == (int) inst.classValue()) {
						instinf.correct = true;
					} else
						instinf.correct = false;
				}
			} else {
				if (toremove) removeAll(instinf.rules, removedRules);
				if (old) remove(instinf.rules, oldRule);
				if (dist == instinf.ruledist && !newRule.deleted) {
					instinf.rules.add(newRule);
				}
				if (getMaxClass(instinf) == (int) inst.classValue()) {
					instinf.correct = true;
				} else
					instinf.correct = false;
			}
		}
		removedRules.clear();
	}

	/**
	 * Delete unimportant rules. A rule is unimportant, if it doesn't
	 * lead to a classification of an instance.
	 */
	protected void unimportant() {
		importantRules = new boolean[rules.size()];

		Iterator it = dataList.iterator();
		while (it.hasNext()) {
			InstanceInfo instinf = (InstanceInfo) it.next();
			if (!instinf.correct) continue;
			if (instinf.rules.size() == 1) {
				important((Rule) instinf.rules.get(0));
			} else {
				important(instinf.rules, (int) instinf.inst.classValue());
			}

		}
		it = rules.iterator();
		for (int i = 0; i < rules.size(); i++) {
			Rule rule = (Rule) it.next();
			if (!importantRules[i]) {
				rule.deleted = true;
				it.remove();
			}
		}
	}

	/**
	 * Set the Rule important.
	 * @param rule important rule
	 */
	protected void important(Rule rule) {
		int ind = indexOf(rules, rule);
		if (ind >= 0) importantRules[ind] = true;
	}

	/**
	 * Find the important rules of this nearest rule list.
	 * @param nearestRules List of nearest rules
	 * @param cls correct classification
	 */
	protected void important(List nearestRules, int cls) {
		switch (tieBreak) {
		case HIGHEST_LAPLACE_TIEBREAK:
			Collections.sort(nearestRules, ruleCriterionComparator);
			int pos = nearestRules.size() - 1;
			if (pos < 0) return;
			Rule rule = (Rule) nearestRules.get(pos);
			double crit = rule.criterion;
			double newcrit = 0;
			if (rule.getClassValue() == cls) important(rule);
			do {
				pos--;
				if (pos >= 0) {
					rule = (Rule) nearestRules.get(pos);
					newcrit = rule.criterion;
					if (newcrit == crit) {
						if (rule.getClassValue() == cls) important(rule);
					}
				}
			} while (newcrit == crit && pos > 0);
			break;
		case FREQUENCY_BASED_VOTING_TIEBREAK:
			Iterator it = nearestRules.iterator();
			while (it.hasNext()) {
				rule = (Rule) it.next();
				if (rule.getClassValue() == cls) important(rule);
			}
			break;
		case MOST_SPECIFIC_TIEBREAK:
			Collections.sort(nearestRules, ruleCoveredSizeComparator);
			pos = nearestRules.size() - 1;
			if (pos < 0) return;
			rule = (Rule) nearestRules.get(pos);
			crit = rule.coveredInstInfo.size();
			if (rule.getClassValue() == cls) important(rule);
			newcrit = 0;
			do {
				pos--;
				if (pos >= 0) {
					rule = (Rule) nearestRules.get(pos);
					newcrit = rule.coveredInstInfo.size();
					if (newcrit == crit) {
						if (rule.getClassValue() == cls) important(rule);
					}
				}
			} while (newcrit == crit && pos > 0);
			break;
		}
	}

	/**
	 * Compute the accuracy, if the oldRule is replaced by the newRule
	 * @param oldRule old Rule
	 * @param newRule new Rule
	 * @return new accuracy
	 */
	protected double updateAccuracy(Rule oldRule, Rule newRule) {
		int correct = 0;
		int countAll = 0;
		Iterator it = dataList.iterator();
		updateList.clear();
		newRule.resetDistribution();
		newRule.coveredInstInfo.clear();
		while (it.hasNext()) {
			InstanceInfo instinf = (InstanceInfo) it.next();
			Instance inst = instinf.inst;
			newRule.addDistribution(instinf);
			double dist = distance(newRule, inst);
			boolean old = (indexOf(instinf.rules, oldRule) >= 0);
			if (Utils.gr(dist, instinf.ruledist) && !old) {
				countAll++;
				if (instinf.correct) correct++;
				continue;
			}
			updateList.add(instinf);
		}

		newRule.criterion = criterion(newRule);

		it = updateList.iterator();
		while (it.hasNext()) {
			InstanceInfo instinf = (InstanceInfo) it.next();
			Instance inst = instinf.inst;
			countAll++;
			double dist = distance(newRule, inst);
			boolean old = (indexOf(instinf.rules, oldRule) >= 0);
			if (Utils.gr(instinf.ruledist, dist)) {
				if (newRule.getClassValue() == (int) inst.classValue()) {
					correct++;
				}
			} else if (Utils.eq(dist, instinf.ruledist)) {
				List newRules = new ArrayList(instinf.rules);
				if (old) remove(newRules, oldRule);
				newRules.add(newRule);
				if (maxIndex(getDistribution(newRules)) == (int) inst
						.classValue()) correct++;

			}
		}
		return correct / (double) countAll;
	}

	/**
	 * Create a new Rule which generalize the given Rule with the given Instance
	 * 
	 * @param rule
	 *            Rule to generalize
	 * @param nearest
	 *            Instance to grow the rule
	 * @return new generalized rule
	 */
	protected Iterator generalize(final Rule rule, List nearest) {
		final Iterator nextNear = nearest.iterator();
		return new Iterator() {

			Rule next = null;

			public void remove() {
			}

			public boolean hasNext() {
				if (next == null)
					return computeNext();
				else
					return true;
			}

			public Object next() {
				if (next == null) computeNext();
				Object val = next;
				next = null;
				return val;
			}

			private boolean computeNext() {
				if (nextNear.hasNext()) {
					InstanceInfo instinf = (InstanceInfo) nextNear.next();
					Instance inst = instinf.inst;
					next = new Rule(rule);
					next.grow(inst);
					return true;
				} else
					return false;
			}
		};
	}

	/**
	 * Find the nearest Instance which is not covered of the given Rule
	 * 
	 * @param rule
	 *            interested Rule
	 * @return nearest not covered Instance
	 */
	protected List findNearest(Rule rule) {
		nearest.clear();
		double minDist = Double.MAX_VALUE;
		int cls = rule.getClassValue();
		Iterator it = dataList.iterator();
		while (it.hasNext()) {
			InstanceInfo instinf = (InstanceInfo) it.next();
			Instance inst = instinf.inst;
			if (rule.covers(inst) || cls != (int) inst.classValue()) continue;
			double dist = distance(rule, inst);
			if (Utils.gr(minDist, dist)) {
				nearest.clear();
				minDist = dist;
				nearest.add(instinf);
			} else if (Utils.eq(minDist, dist)) {
				nearest.add(instinf);
			}
		}
		return nearest;
	}

	/**
	 * Add one distribution to another.
	 * 
	 * @param dist
	 *            the main distribution
	 * @param newDist
	 *            distribution to add
	 */
	protected void addDistribution(double[] dist, double[] newDist) {
		for (int i = 0; i < classCount; i++) {
			dist[i] += newDist[i];
		}
	}

	/**
	 * Compute criterion of an rule. Here the Laplace index is used.
	 * 
	 * @param rule
	 *            the rule
	 * @return the criterion.
	 */
	protected double criterion(Rule rule) {
		return (rule.getDistribution()[rule.getClassValue()] + 1)
				/ (rule.getDistributionSum() + classCount);
	}

	/**
	 * Predicts the class memberships for a given instance. If an instance is
	 * unclassified, the returned array elements must be all zero.
	 * 
	 * @param instance
	 *            the instance to be classified
	 * @return an array containing the estimated membership probabilities of the
	 *         test instance in each class
	 * @exception Exception
	 *                if distribution could not be computed successfully
	 */
	public double[] distributionForInstance(Instance instance) {
		InstanceInfo instinf = new InstanceInfo(instance);
		findNearestRules(instinf);
		double[] result = getDistribution(instinf);
		if (result == null) {
			System.out.println("NULL DIST!!!");
			findNearestRules(instinf);
			result = getDistribution(instinf);
		}
		return result;
	}

	/**
	 * Compute distance of two Instances
	 * 
	 * @param instance1
	 *            first instance
	 * @param instance2
	 *            second instance
	 * @return distance of instances
	 */
	protected double distance(Instance instance1, Instance instance2) {
		double dist = 0;
		for (int i = 0; i < attrCount; i++) {
			if (usedAttribute[i] > 0) {
				double val1 = 0, val2 = 0;
				if (instance1.isMissing(i))
					val1 = Double.NaN;
				else
					val1 = instance1.value(i);
				if (instance2.isMissing(i))
					val2 = Double.NaN;
				else
					val2 = instance2.value(i);
				double d = distance.distance(i, val1, val2);
				//if (d < 0 || d > 1) System.out.println("ERROR!!!");
				dist += d * d;
			}
		}
		return dist;
	}

	/**
	 * Compute the distance of one Rule to an Instance.
	 * 
	 * @param rule
	 *            the rule
	 * @param instance
	 *            the instance
	 * @return distance of instance to rule
	 */
	protected double distance(Rule rule, Instance instance) {
		double dist = 0;
		for (int i = 0; i < attrCount; i++) {
			if (usedAttribute[i] > 0) {
				if (instance.isMissing(i)) continue;
				if (usedAttribute[i] == NOMINAL) {
					if (rule.getNominalValue(i) >= 0) {
						if (rule.getNominalValue(i) != (int) instance.value(i)) {
							double dif = distance.distance(i, rule.getNominalValue(i),
									(int) instance.value(i));
							dist += dif * dif;
						}
					}
				} else if (usedAttribute[i] == NUMERIC) {
					if (rule.getNumericRange(i) != null) {
						double val = instance.value(i);
						if (val < rule.getNumericRange(i)[0]) {
							double dif = distance.distance(i,
									rule.getNumericRange(i)[0], val);
							dist += dif * dif;
						} else if (val > rule.getNumericRange(i)[1]) {
							double dif = distance.distance(i, val,
									rule.getNumericRange(i)[1]);
							dist += dif * dif;
						}
					}
				}
			}
		}
		return dist;
	}

	/**
	 * Detect the attributes to use for learning.
	 */
	protected void computeUsedAttributes() {
		usedAttribute = new int[attrCount];
		for (int i = 0; i < attrCount; i++) {
			if (i != classIndex && isUsedAttribute(i)) {
				Attribute att = header.attribute(i);
				if (att.isNominal()) {
					usedAttribute[i] = NOMINAL;
				} else if (att.isNumeric()) usedAttribute[i] = NUMERIC;
			}
		}
	}

	/**
	 * Set the list of used attributes.
	 * @param usedAttributeList List of used attributes
	 */
	public void setUsedAttributes(List usedAttributeList) {
		this.usedAttributeList = usedAttributeList;
	}

	/**
	 * Is Attribute in the used attribute list.
	 * 
	 * @param attribute
	 *            attribute index
	 * @return if this attribute should be used for learning
	 */
	protected boolean isUsedAttribute(int attribute) {
		if (usedAttributeList == null)
			return true;
		else
			return (usedAttributeList.indexOf(new Integer(attribute)) >= 0);
	}

	/**
	 * Set all rules to be not finished.
	 */
	protected void unsetRules() {
		int size = rules.size();
		for (int i = 0; i < size; i++) {
			Rule rule = (Rule) rules.get(i);
			rule.setFinished(false);
		}
	}

	/**
	 * Get the rulelist.
	 * @return rulelist
	 */
	public List getRules() {
		return rules;
	}

	/**
	 * Return an string representation of the induced rules.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		Iterator it = rules.iterator();
		while (it.hasNext()) {
			SimpleConjunctiveRule rule = (SimpleConjunctiveRule) it.next();
			sb.append(rule.toString());
			sb.append("\n");
		}
		sb.append(rules.size() + "\n\n");
		return sb.toString();
	}

	/**
	 * No numeric normalization.
	 */
	public void setNoNumericNormalization() {
		numeric = NormDistance.NO_NUMERIC;
	}

	/**
	 * Use range normalization.
	 */
	public void setRangeNumericNormalization() {
		numeric = NormDistance.RANGE_NUMERIC;
	}

	/**
	 * Use quantile normalization.
	 * @param quantile quantile value between 0 and 1.
	 */
	public void setQuantileNumericNormalization(int quantile) {
		numeric = NormDistance.QUANTILE_NUMERIC;
		this.quantile = quantile;
	}

	/**
	 * Get the flag for numeric normalization.
	 * @return normalize flag
	 */
	public int getNumericNormalization() {
		return numeric;
	}

	/**
	 * Set the flag for nominal VDM distance.
	 * @param flag nominal VDM distance flag
	 */
	public void setNominalVDMDistance(boolean flag) {
		useVDM = flag;
	}

	/**
	 * Get the flag for nominal VDM distance.
	 * @return nominal VDM distance flag
	 */
	public boolean getNominalVDMDistance() {
		return useVDM;
	}

	/**
	 * Set the flag for simplification of rules.
	 * @param flag simplify flag
	 */
	public void setSimplification(boolean flag) {
		simplification = flag;
	}

	/**
	 * Get the flag for simplification of rules.
	 * @return simplify flag
	 */
	public boolean getSimplification() {
		return simplification;
	}

	/**
	 * Set the flag for only use new rules, which strictly increase the accuracy.
	 * @param flag strict increase flag
	 */
	public void setStrictIncrease(boolean flag) {
		increase = flag;
	}

	/**
	 * Get the flag for simplification of rules.
	 * @return strict increase flag
	 */
	public boolean getStrictIncrease() {
		return increase;
	}

	/**
	 * Returns an enumeration of all the available options.
	 *
	 * @return an enumeration of all available options.
	 */
	public Enumeration listOptions() {
		FastVector newVector = new FastVector(9);

		newVector
				.addElement(new Option(
						"\tWhich Normalization of numeric attributes are used.\n"
								+ "\tuse: -N numeric_flag\n"
								+ "\tpossible values for numeric_flag are \n"
								+ "\t    no         use no numeric normalization.\n"
								+ "\t    range      use range normalization (max-min=1.0)\n"
								+ "\t    [0.0-0.5]  use quantile normalization with the given quantile\n\n",
						"N", 2, "-N"));

		newVector.addElement(new Option(
				"\tDon't use the VDM distance for nominal attributes.", "V", 0,
				"-V"));

		newVector
				.addElement(new Option(
						"\tWhich tiebreak strategy is used for rules with same distance.\n"
								+ "\tuse: -T method\n"
								+ "\tpossible values for method are \n"
								+ "\t    laplace    Choose rule with highest Laplace accuracy\n"
								+ "\t    specific   Choose most specific rule\n"
								+ "\t    vote       Use frequency-based voting",
						"T", 2, "-T"));

		newVector.addElement(new Option(
				"\tUse the local stopping strategy for rule induction.", "L",
				0, "-L"));

		newVector.addElement(new Option(
				"\tUse the simplification strategy on the ruleset.", "S", 0,
				"-S"));

		newVector.addElement(new Option(
				"\tOnly induce new rules, which strictly increase the accuracy.",
				"I", 0, "-I"));
		
		newVector.addElement(new Option(
				"\tInduce new rules, which not decrease the accuracy.",
				"E", 0, "-E"));
		return newVector.elements();
	}

	/**
	 * Parses a given list of options.
	 *
	 * @param options the list of options as an array of strings
	 * @exception Exception if an option is not supported
	 */
	public void setOptions(String[] options) throws Exception {
		String flag = Utils.getOption('N', options);
		if (flag.equalsIgnoreCase("no")) {
			numeric = NormDistance.NO_NUMERIC;
		} else if (flag.equalsIgnoreCase("range")) {
			numeric = NormDistance.RANGE_NUMERIC;
		} else {
			try {
				double quantile = Double.parseDouble(flag);
				this.quantile = quantile;
				numeric = NormDistance.QUANTILE_NUMERIC;
			} catch (NumberFormatException e) {
				System.err.println("WRONG FORMAT OF THE NORMALIZATION FLAG!");
				System.exit(1);
			}
		}

		flag = Utils.getOption('T', options);
		if (flag.equalsIgnoreCase("laplace")) {
			numeric = HIGHEST_LAPLACE_TIEBREAK;
		} else if (flag.equalsIgnoreCase("specific")) {
			numeric = MOST_SPECIFIC_TIEBREAK;
		} else if (flag.equalsIgnoreCase("vote")) {
			numeric = FREQUENCY_BASED_VOTING_TIEBREAK;
		}

		useVDM = !Utils.getFlag('V', options);
		globalStop = !Utils.getFlag('L', options);
		simplification = Utils.getFlag('S', options);
		increase = Utils.getFlag('I', options);
		if (increase) rise3Stop=false;
		else rise3Stop = !Utils.getFlag('E', options);

	}

	/**
	 * Gets the current option settings.
	 *
	 * @return the list of current option settings as an array of strings
	 */
	public String[] getOptions() {
		String[] options = new String[9];
		int current = 0;
		options[current++] = "-N";
		if (numeric == NormDistance.NO_NUMERIC)
			options[current++] = "no";
		else if (numeric == NormDistance.RANGE_NUMERIC)
			options[current++] = "range";
		else if (numeric == NormDistance.QUANTILE_NUMERIC)
				options[current++] = "" + quantile;
		if (!useVDM) options[current++] = "-V";

		options[current++] = "-T";
		if (tieBreak == HIGHEST_LAPLACE_TIEBREAK)
			options[current++] = "laplace";
		else if (tieBreak == MOST_SPECIFIC_TIEBREAK)
			options[current++] = "specific";
		else if (tieBreak == FREQUENCY_BASED_VOTING_TIEBREAK)
				options[current++] = "vote";

		if (!globalStop) options[current++] = "-L";

		if (simplification) options[current++] = "-S";

		if (increase) options[current++] = "-I";
		else if (!rise3Stop) options[current++] = "-E";

		while (current < options.length)
			options[current++] = "";

		return options;
	}

	/**
	 * Main method.
	 *
	 * @param args the options for the classifier
	 */
	public static void main(String[] args) {
		try {
			System.out.println(Evaluation.evaluateModel(new Rise31(), args));
		} catch (Exception e) {
			e.printStackTrace();
			System.err.println(e.getMessage());
		}
	}

}