/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    SimpleConjunctiveRule.java
 *    Copyright (C) 2004 Jrgen Beringer
 */
package weka.classifiers.rules.rise;

import weka.classifiers.rules.Rule;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

/**
 * Class of a simple conjunctive rule for numeric and nominal attributes. This class
 * should be extended for a special algorithm.
 * 
 * @author Jrgen Beringer (beringer@mathematik.uni-marburg.de)
 * @version $Revision: 1.3 $
 */
public class SimpleConjunctiveRule extends Rule {

	/**Value for nominal attributes. */
	protected int[] values;

	/**Bounds for numeric attributes. */
	protected double[][] bounds;

	/**Absolut distribution of classes for this rule. */
	protected double[] distribution;

	/**Sum of all weights added to distribution. */
	protected double distSum;

	/**Index of class this rule will predict.*/
	protected int classValue;

	/**Number of attributes */
	protected int attrCount;

	/** Number of classes */
	protected int classCount;

	/** List of all used Attributes in the rule. */
	protected int[] usedAttr;

	/**Number of simple prdeicates. */
	protected int predCount;

	/**Header of the data. */
	protected Instances instances;

	/**By this rule covered instances. */
	protected FastVector coveredExamples;

	/**Flag if covered instances will be stored. */
	protected boolean storeCovered = true;

	/** 
	 * Default constructor.
	 */
	protected SimpleConjunctiveRule() {
	}

	/**
	 * Create an empty rule.
	 * @param instances Header of used instances
	 * @param store flag to store covered data 
	 */
	public SimpleConjunctiveRule(Instances instances, boolean store) {
		this.instances = instances;
		this.attrCount = instances.numAttributes();
		this.classCount = instances.numClasses();
		values = new int[attrCount];
		bounds = new double[attrCount][];
		distribution = new double[classCount];
		predCount = 0;
		usedAttr = new int[attrCount];
		for (int i = 0; i < attrCount; i++)
			values[i] = -1;
		storeCovered = store;
		resetCoveredExamples(0);
	}

	/**
	 * Copyconstructor.
	 * @param rule Rule to copy
	 */
	public SimpleConjunctiveRule(SimpleConjunctiveRule rule) {
		this.instances = rule.instances;
		this.attrCount = rule.attrCount;
		this.classCount = rule.classCount;
		values = new int[attrCount];
		bounds = new double[attrCount][];
		distribution = new double[classCount];
		usedAttr = new int[attrCount];
		predCount = rule.predCount;
		System.arraycopy(rule.usedAttr, 0, usedAttr, 0, Math.min(predCount,attrCount));
		classValue = rule.classValue;
		for (int i = 0; i < attrCount; i++) {
			values[i] = rule.values[i];
			if (rule.bounds[i] != null)
					bounds[i] = new double[] { rule.bounds[i][0],
							rule.bounds[i][1]};
		}
		storeCovered = rule.storeCovered;
		resetCoveredExamples(rule.coveredExamples.size());
	}

	/**
	 * Reset list of covered data.
	 * @param capacity initial capacity of the list
	 */
	public void resetCoveredExamples(int capacity) {
		coveredExamples = new FastVector(capacity);
	}

	/**
	 * Set the flag for storing covered data.
	 * @param val flag value.
	 */
	public void setStoreCovered(boolean val) {
		storeCovered = val;
		if (!val) coveredExamples.removeAllElements();
	}

	/**
	 * Get flag for storing covered examples.
	 * @return the store flag
	 */
	public boolean getStoreCovered() {
		return storeCovered;
	}

	/**
	 * Get list of covered data.
	 * @return list
	 */
	public FastVector getCovered() {
		if (!storeCovered) return null;
		return coveredExamples;
	}

	/**
	 * Set list of covered data.
	 * @param covered List of data, which are covered of this rule.
	 */
	public void setCovered(FastVector covered) {
		if (!storeCovered)
				throw new IllegalStateException(
						"Store covered elements is off!");
		coveredExamples = covered;
	}

	/**
	 * Add instances to the list of covered data. 
	 * @param covered Instances which are covered of this rule.
	 */
	public void addCovered(Instances covered) {
		if (!storeCovered)
				throw new IllegalStateException(
						"Store covered elements is off!");
		int size = covered.numInstances();
		for (int i = 0; i < size; i++)
			coveredExamples.addElement(covered.instance(i));
	}

	/**
	 * Add instances to the list of covered data. 
	 * @param covered List of data, which are covered of this rule.
	 */
	public void addCovered(FastVector covered) {
		if (!storeCovered)
				throw new IllegalStateException(
						"Store covered elements is off!");
		coveredExamples.appendElements(covered);
	}

	/**
	 * Compare this rule to another one.
	 * @param rule 
	 * @return if rules are equal
	 */
	public boolean equals(Object rule) {
		return equals((SimpleConjunctiveRule) rule);
	}

	/**
	 * Compare this rule to another one.
	 * @param rule 
	 * @return if rules are equal
	 */
	public boolean equals(SimpleConjunctiveRule rule) {
		if (rule == null) return false;
		if (rule.attrCount != attrCount) return false;
		if (rule.classValue != classValue) return false;
		if (rule.classCount != classCount) return false;
		for (int i = 0; i < attrCount; i++) {
			if (rule.values[i] != values[i]) return false;
			if ((rule.bounds[i] != null) && (bounds[i] != null)) {
				//if (rule.borders[i].length!=2) return false;
				//if (borders[i].length!=2) return false;
				if ((rule.bounds[i][0] != bounds[i][0])
						|| (rule.bounds[i][1] != bounds[i][1])) return false;
			} else if ((rule.bounds[i] != null) || (bounds[i] != null))
					return false;
		}
		return true;
	}

	/**
	 * Test if this rule contains the other rule.
	 * @param rule Rule to test.
	 * @return If this rules complete covers the other rule.
	 */
	public boolean contains(SimpleConjunctiveRule rule) {
		for (int i = 0; i < attrCount; i++) {
			if (values[i] >= 0 && rule.values[i] != values[i]) return false;
			if (bounds[i] != null && rule.bounds[i] != null) {
				if ((rule.bounds[i][0] < bounds[i][0])
						|| (rule.bounds[i][1] > bounds[i][1])) return false;
			} else if (bounds[i] != null) return false;
		}
		return true;
	}

	/**
	 * Test if this rule overlaps the other rule.
	 * @param rule Rule to test.
	 * @return If this rules overlaps the other rule.
	 */
	public boolean overlaps(SimpleConjunctiveRule rule) {
		for (int i = 0; i < attrCount; i++) {
			if (values[i] >= 0 && rule.values[i] >= 0
					&& rule.values[i] != values[i]) return false;
			if (bounds[i] != null && rule.bounds[i] != null) {
				if ((rule.bounds[i][0] > bounds[i][1])
						|| (rule.bounds[i][1] < bounds[i][0])) return false;
			}
		}
		return true;
	}

	/**
	 * Set the nominal value for the attribute. A predicate attribute==value will be added or 
	 * exchanged. Use value=-1 to unset this predicate.
	 * @param attribut Index of attribute.
	 * @param value Index of the value for this attribute or -1 if not set.
	 */
	public void setNominalValue(int attribut, int value) {
		if (bounds[attribut] == null) {
			if (values[attribut] == -1 && value>=0) {
				usedAttr[predCount] = attribut;
                values[attribut] = value;
				predCount++;
			}else 
			if (value == -1 && values[attribut] >= 0) {
                values[attribut] =-1;
				predCount--;
				for (int i = predCount-1; i >= 0; i--) {
					if (usedAttr[i] == attribut) {
						System.arraycopy(usedAttr, i + 1, usedAttr, i,
								predCount - i);
                        break;
					}
				}
			}
			
		} else
			throw new IllegalStateException(
					"Borders are set, so this attribut should be numerical!");
	}

	/**
	 * Get the nominal value of this attribute or -1 if no value is set.
	 * @param attribut Index of attribute.
	 * @return Index of the value for this attribute or -1.
	 */
	public int getNominalValue(int attribut) {
		return values[attribut];
	}

	/**
	 * Get the numeric bounds of this attribute or null if no bounds are set.
	 * @param attribut Index of the attribute.
	 * @return the bounds or null
	 */
	public double[] getNumericRange(int attribut) {
		return bounds[attribut];
	}

	/**
	 * Test if this rule covers the instance.
	 * @param instance Instance to cover.
	 * @return if the instance is covered
	 */
	public boolean covers(Instance instance) {
		for (int i = predCount - 1; i >= 0; i--) {
			int attr = usedAttr[i];
			if (values[attr] != -1) {
				if (instance.isMissing(attr)) continue;
				if (((int) instance.value(attr)) != values[attr]) return false;
			} else if (bounds[attr] != null) {
				if (instance.isMissing(attr)) continue;
				if (Utils.gr(bounds[attr][0], instance.value(attr))
						|| Utils.gr(instance.value(attr), bounds[attr][1]))
						return false;
			}
		}
		return true;
	}

	/**
	 * Set the numeric bounds of this attribute. The predicate 
	 * min<=attribute<=max is added to the rule.
	 * @param attribut Index of the attribute
	 * @param min Minimum value of this attribute.
	 * @param max Maximum value of this attribute.
	 */
	public void setNumericRange(int attribut, double min, double max) {
		if (values[attribut] == -1) {
			if (bounds[attribut] == null) {
				usedAttr[predCount] = attribut;
				predCount++;
			}
			bounds[attribut] = new double[] { min, max};
		} else
			throw new IllegalStateException(
					"Value is set, so this attribut should be nominal!");
	}

	/**
	 * Delete the bounds for this attribute.
	 * @param attribut Index of attribute
	 */
	public void resetNumericRange(int attribut) {
		if (values[attribut] == -1) {
			if (bounds[attribut] != null) {
                bounds[attribut] = null;
				predCount--;
				for (int i = predCount - 1; i >= 0; i--) {
					if (usedAttr[i] == attribut) {
						System.arraycopy(usedAttr, i + 1, usedAttr, i,
								predCount - i);
					}
				}
			}
		} else
			throw new IllegalStateException(
					"Value is set, so this attribut should be nominal!");
	}

	/**
	 * Get the number of predicates in this rule.
	 * @return Number of predicates in this rule
	 */
	public int getPredicateCount() {
		return predCount;
	}

	/**
	 * Get a list of used attributes in the predicates.
	 * @return list of attributes
	 */
	public int[] getUsedAttributs() {
		return usedAttr;
	}

	/**
	 * Add the class with weight 1 to the distribution.
	 * @param classIndex Index of class
	 */
	public void addClass(int classIndex) {
		distribution[classIndex]++;
		distSum++;
	}

	/**
	 * Add the class with specified weight to the distribution.
	 * @param classIndex Index of class
	 * @param count weight of class
	 */
	public void addClass(int classIndex, double count) {
		distribution[classIndex] += count;
		distSum += count;
	}

	/**
	 * Delete distribution.
	 */
	public void resetDistribution() {
		for (int i = 0; i < classCount; i++) {
			distribution[i] = 0;
		}
		distSum = 0;
	}

	/**
	 * Set predicted class of this rule.
	 * @param classIndex Index of class to predict.
	 */
	public void setClass(int classIndex) {
		this.classValue = classIndex;
	}

	/**
	 * Compute the distribution with instances of the header object.
	 */
	public void computeDistribution() {
		computeDistribution(instances);
	}

	/**
	 * Compute the distribution with this Instances.
	 * @param data Instances to compute distribution
	 */
	public void computeDistribution(Instances data) {
		resetDistribution();
		int size = data.numInstances();
		for (int i = size - 1; i >= 0; i--) {
			Instance inst = data.instance(i);
			if (covers(inst)) {
				addClass((int) inst.classValue(), inst.weight());
				if (storeCovered) coveredExamples.addElement(inst);
			}
		}
	}

	/**
	 * Compute the distribution with this list of instances.
	 * @param data List of instances to compute distribution
	 */
	public void computeDistribution(FastVector data) {
		resetDistribution();
		int size = data.size();
		for (int i = size - 1; i >= 0; i--) {
			Instance inst = (Instance) data.elementAt(i);
			if (covers(inst)) {
				addClass((int) inst.classValue(), inst.weight());
				if (storeCovered) coveredExamples.addElement(inst);
			}
		}
	}

	/**
	 * Compute the predicted class as the most frequent class of the distribution.
	 * @return selected class index
	 */
	public int computeClass() {
		classValue = frequentClass();
		return classValue;
	}

	/**
	 * Get the most frequent class.
	 * @return most frequent class.
	 */
	public int frequentClass() {
		int max = 0;
		double maxVal = distribution[0];
		for (int i = 1; i < classCount; i++) {
			if (distribution[i] > maxVal) {
				max = i;
				maxVal = distribution[i];
			}
		}
		return max;
	}

	/**
	 * Get the absolut distribution of this rule.
	 * @return the distribution
	 */
	public double[] getDistribution() {
		return distribution;
	}

	/**
	 * Get the relative distribution of this rule. 
	 * @return relative distribution
	 */
	public double[] getRelativeDistribution() {
		double[] relDist = new double[classCount];
		if (distSum > 0) {
			for (int i = 0; i < classCount; i++) {
				relDist[i] = distribution[i] / distSum;
			}
		}
		return relDist;
	}

	/**
	 * Set the absolut distribution.
	 * @param distribution new distribution
	 */
	public void setDistribution(double[] distribution) {
		this.distribution = distribution;
	}

	/**
	 * Get the sum of all weights in the distribution.
	 * @return sum of distribution
	 */
	public double getDistributionSum() {
		return distSum;
	}

	/**
	 * Set the sum of all weights in the distribution.
	 * @param distSum sum of distribution
	 */
	public void setDistributionSum(double distSum) {
		this.distSum = distSum;
	}

	/**
	 * Get the Index of the predicted class.
	 * @return Index of predicted class
	 */
	public int getClassValue() {
		return classValue;
	}

	/**
	 * Get the coverage of the predicted class (frequency of the predicted class).
	 * @return coverage of the predicted class
	 */
	public double getCoverage() {
		return distribution[classValue] / (double) distSum;
	}

	/**
	 * Get a string representation of this rule.
	 * @return rule 
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		int pos = 0;
		for (int i = 0; i < predCount; i++) {
			while (values[pos] == -1 && bounds[pos] == null)
				pos++;
			if (values[pos] != -1) {
				/*sb.append(instances.attribute(pos).name()
				 + "="
				 + instances.attribute(pos).value(
				 values[pos]) + " ");*/

				sb.append(instances.attribute(pos).name() + "=" + values[pos]
						+ " ");
				if (i + 1 < predCount) sb.append("& ");
				pos++;
			} else {
				if (bounds[pos][0] == bounds[pos][1]) {
					sb.append(instances.attribute(pos).name() + "="
							+ bounds[pos][0] + " ");
					if (i + 1 < predCount) sb.append("& ");
					pos++;
				} else {
					sb.append(instances.attribute(pos).name() + " in ["
							+ bounds[pos][0] + "," + bounds[pos][1] + "] ");
					if (i + 1 < predCount) sb.append("& ");
					pos++;
				}
			}
		}
		if (predCount == 0) {
			sb.append("true ");
		}
		sb.append("=>" + instances.classAttribute().value(classValue));
		return sb.toString();
	}

	/**
	 * Grow this rule with this instances. This method is not supported here!
	 */
	public void grow(Instances data) {
		throw new UnsupportedOperationException();

	}

	/**
	 * Whether this rule has antecedents. This Method is extended from Rule and
	 * returns true if any predicate exist in this rule. 
	 */
	public boolean hasAntds() {
		return predCount > 0;
	}

	/**
	 * Returns the predicted class of this rule.
	 */
	public double getConsequent() {
		return classValue;
	}

	/**
	 * Returns the count of predicates in this rule.
	 */
	public double size() {
		return predCount;
	}

}