package weka.classifiers.trees;

import java.io.Serializable;

import weka.classifiers.Classifier;
import weka.core.Instance;
import weka.core.Instances;
import weka.filters.Filter;
import weka.filters.supervised.attribute.Discretize;

public class CopyOfLazyForestDT extends Classifier implements Serializable {

    protected static class Split implements Serializable {
        public int attribut;

        public int value;

        public Instances data;

        public double[] instDistrib;

        public int[] usedAtts;

        public double gain;
    }

    private static final long serialVersionUID = 695872586476417759L;

    private static final double log2 = Math.log(2);

    protected Instances data;

    protected int attrCount;

    protected int classCount;

    protected int classIndex;

    protected Split split;

    protected int[] attrValCount;
    
    protected int treeCount=5;
    
    protected int kValue;
    
    protected Discretize disc;

    public void buildClassifier(Instances builddata) throws Exception {
        disc = new Discretize();
        disc.setInputFormat(builddata);
        this.data = Filter.useFilter(builddata, disc);
        classIndex = data.classIndex();
        classCount = data.attribute(data.classIndex()).numValues();
        attrCount = data.numAttributes();
        attrValCount = new int[attrCount];
        kValue=attrCount/2;
        for (int i = 0; i < attrCount; i++) {
            attrValCount[i] = data.attribute(i).numValues();
        }
        split = new Split();
        split.data = data;
        split.attribut = -1;
        split.usedAtts = new int[attrCount];
        split.instDistrib = getClassDistribution(data);
    }

    public double[] distributionForInstance(Instance inst) {
        disc.input(inst);
        inst=disc.output();
        double[] dist = new double[classCount];
        for (int i = 0; i < treeCount; i++) {
            double[] subdist = buildNode(split, inst,0);
            for (int c = 0; c < classCount; c++) {
                dist[c] += subdist[c] / treeCount;
            }
        }
        return dist;
    }

    protected double[] buildNode(Split split, Instance inst, int dep) {
        Split newSplit = getSplits(split, inst);
        if (newSplit == null) {
            //System.out.println(dep);
            double[] dist = split.instDistrib;
            return dist;
        } else if (newSplit.gain == 0.0) {
//            System.out.println("  attribut=" + newSplit.attribut + "  value="
//                    + newSplit.value);
            //System.out.println(dep);
            return newSplit.instDistrib;
        }
//        System.out.println("  attribut=" + newSplit.attribut + "  value="
//                + newSplit.value);
        double[] subDist = buildNode(newSplit, inst,dep+1);
        return subDist;
    }

    protected double[] getClassDistribution(Instances data) {
        double[] dist = new double[classCount];
        for (int i = 0; i < data.numInstances(); i++) {
            dist[(int) data.instance(i).classValue()]++;
        }
        for (int i = 0; i < classCount; i++) {
            dist[i] = dist[i] / data.numInstances();
        }
        return dist;
    }

    protected double[] getWeighedClassDistribution(Instances data,
            double[] preDist) {
        double[] dist = new double[classCount];
        double sum = 0;
        for (int i = 0; i < data.numInstances(); i++) {
            int cl = (int) data.instance(i).classValue();
            dist[cl] += 1.0 / preDist[cl];
            sum += 1.0 / preDist[cl];
        }
        for (int i = 0; i < classCount; i++) {
            dist[i] = dist[i] / sum;
        }
        return dist;
    }
    
    protected boolean[] getSelected(int count){
        boolean[] select=new boolean[attrCount];
        for(int i=0;i<count;i++){
            int sel=-1;
            do{
                sel=(int)(Math.random()*attrCount);
            }while (sel==classIndex || select[sel]);
            select[sel]=true;
        }
        return select;
    }

    protected Split getSplits(Split lastSplit, Instance inst) {
        boolean[] select=getSelected(kValue);
        Split newSplit = null;
        Instances data = lastSplit.data;
        double[] weightdist = lastSplit.instDistrib;
        for (int i = 0; i < attrCount; i++) {
            if (!select[i] || lastSplit.usedAtts[i] >= attrValCount[i] - 1 || i == classIndex
                    || inst.isMissing(i))
                continue;

            int n = attrValCount[i];
            int[][] distrib = new int[n + 1][classCount];
            int[] sum = new int[n + 1];
            int in = data.numInstances();
            int iValue = (int) inst.value(i);

            for (int nr = 0; nr < in; nr++) {
                Instance instance = data.instance(nr);
                if (instance.isMissing(i)) {
                    distrib[n][(int) instance.classValue()]++;
                    sum[n]++;
                } else {
                    int val = (int) instance.value(i);
                    distrib[val][(int) instance.classValue()]++;
                    sum[val]++;
                }
            }
            for (int v = 0; v < n; v++) {
                if (v == iValue || sum[v] == 0 || sum[v] == data.numInstances())
                    continue;
                Split sp = new Split();

                sp.attribut = i;
                sp.value = v;
                double gain = 0;
                boolean insert = true;

                double entropy2 = 0;
                double[] dv2 = new double[classCount];
                double sum2 = 0;
                for (int a = 0; a <= n; a++) {
                    if (a != v) {
                        for (int c = 0; c < classCount; c++) {
                            if (weightdist[c] > 0) {
                                dv2[c] += distrib[a][c] * 1 / weightdist[c];
                                sum2 += distrib[a][c] * 1 / weightdist[c];
                            }
                        }
                    }
                }
                for (int c = 0; c < classCount; c++) {
                    dv2[c] /= sum2;
                }
                for (int c = 0; c < classCount; c++) {
                    double p = dv2[c];
                    if (p > 0.0)
                        entropy2 += -p * Math.log(p) / log2;
                }
                gain = entropy2;
                sp.gain = gain;

                if (insert && newSplit != null && pref(newSplit, sp)) {
                    insert = false;
                }
                if (insert) {
                    sp.usedAtts = lastSplit.usedAtts.clone();
                    sp.usedAtts[i]++;
                    newSplit = sp;
                }
            }
        }
        if (newSplit != null) {
            newSplit.data = new Instances(data, 0);
            int in = data.numInstances();
            for (int nr = 0; nr < in; nr++) {
                Instance instance = data.instance(nr);
                if (instance.isMissing(newSplit.attribut)
                        || (int) instance.value(newSplit.attribut) != newSplit.value) {
                    newSplit.data.add(instance);
                }
            }
            newSplit.instDistrib = getClassDistribution(newSplit.data);
        } 
        return newSplit;
    }

    public void weight(double[] dist1, double[] normdist) {
        double sum = 0;
        for (int i = 0; i < classCount; i++) {
            double val = dist1[i] / normdist[i];
            sum += val;
            dist1[i] = val;
        }
        for (int i = 0; i < classCount; i++) {
            dist1[i] /= sum;
        }
    }

    public boolean pref(Split s1, Split s2) {
        return s1.gain < s2.gain;

    }

}
