package weka.classifiers.trees;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.classifiers.trees.LazyEnsembleForestDT.Split;
import weka.core.AdditionalMeasureProducer;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.OptionHandler;
import weka.filters.Filter;
import weka.filters.supervised.attribute.Discretize;

public class EnsembleForestDT extends Classifier implements Serializable,
        OptionHandler, AdditionalMeasureProducer {

    protected static class Split implements Serializable {
        public int attribut;

        public Instances data;

        public double[] instDistrib;

        public double[] sortDist;

        public boolean[] usedAtts;
    }

    private static final long serialVersionUID = 695872586476417759L;

    private static final double log2 = Math.log(2);

    protected Instances data;

    protected int attrCount;

    protected int classCount;

    protected int classIndex;

    protected Split split;

    protected int[] attrValCount;

    protected int treeCount = 5;

    protected int kValue;

    protected Discretize disc;

    protected int comb = 1;

    /**
     * Returns a string describing classifier
     * 
     * @return a description suitable for displaying in the
     *         explorer/experimenter gui
     */
    public String globalInfo() {

        return "LazyEnsembleForestDT";
    }

    public void buildClassifier(Instances builddata) throws Exception {
        disc = new Discretize();
        disc.setInputFormat(builddata);
        this.data = Filter.useFilter(builddata, disc);
        classIndex = data.classIndex();
        classCount = data.attribute(data.classIndex()).numValues();
        attrCount = data.numAttributes();
        attrValCount = new int[attrCount];
        kValue = attrCount / 2;
        for (int i = 0; i < attrCount; i++) {
            attrValCount[i] = data.attribute(i).numValues();
        }
        split = new Split();
        split.data = data;
        split.attribut = -1;
        split.usedAtts = new boolean[attrCount];
        split.instDistrib = getClassDistribution(data);
    }

    public double[] distributionForInstance(Instance inst) {
        disc.input(inst);
        inst = disc.output();
        double[] dist = new double[classCount];
        int sum = 0;
        for (int i = 0; i < 10; i++) {
            comb = 1;
            double[] subdist = buildNode(split, inst, 0);
            sum += comb;

            for (int c = 0; c < classCount; c++) {
                dist[c] += subdist[c];
            }
        }
        treeCount = Math.max(10, (int) Math.round(5 * sum / 10.0));
        treeCount = Math.min(treeCount, 100);
        // System.out.println("Mean Combinations: "+(treeCount));
        for (int i = 10; i < treeCount; i++) {
            comb = 1;
            double[] subdist = buildNode(split, inst, 0);
            sum += comb;

            for (int c = 0; c < classCount; c++) {
                dist[c] += subdist[c];
            }
        }
        for (int c = 0; c < classCount; c++) {
            dist[c] /= treeCount;
        }
        return dist;
    }

    protected double[] buildNode(Split split, Instance inst, int dep) {
        Split newSplit = getSplits(split, inst);
        if (newSplit == null || newSplit.data.numInstances()==0|| inst.isMissing(newSplit.attribut)) {
            // System.out.println(dep);
            double[] dist = split.instDistrib;
            return dist;
        } else if (newSplit.sortDist[classCount - 1] == 1.0) {
            // System.out.println(" attribut=" + newSplit.attribut + " value="
            // + newSplit.value);
            // System.out.println(dep);
            return newSplit.instDistrib;
        }
        // System.out.println(" attribut=" + newSplit.attribut + " value="
        // + newSplit.value);
        double[] subDist = buildNode(newSplit, inst, dep + 1);
        return subDist;
    }

    protected double[] getClassDistribution(Instances data) {
        double[] dist = new double[classCount];
        for (int i = 0; i < data.numInstances(); i++) {
            dist[(int) data.instance(i).classValue()]++;
        }
        for (int i = 0; i < classCount; i++) {
            dist[i] = dist[i] / data.numInstances();
        }
        return dist;
    }

    protected double[] getWeighedClassDistribution(Instances data,
            double[] preDist) {
        double[] dist = new double[classCount];
        double sum = 0;
        for (int i = 0; i < data.numInstances(); i++) {
            int cl = (int) data.instance(i).classValue();
            dist[cl] += 1.0 / preDist[cl];
            sum += 1.0 / preDist[cl];
        }
        for (int i = 0; i < classCount; i++) {
            dist[i] = dist[i] / sum;
        }
        return dist;
    }

    protected boolean[] getSelected(int count) {
        boolean[] select = new boolean[attrCount];
        for (int i = 0; i < count; i++) {
            int sel = -1;
            do {
                sel = (int) (Math.random() * attrCount);
            } while (sel == classIndex || select[sel]);
            select[sel] = true;
        }
        return select;
    }

    protected Split getSplits(Split lastSplit, Instance inst) {
        boolean[] select = getSelected(kValue);
        List<Split> list = new LinkedList();
        Instances data = lastSplit.data;
        double[] weightdist = lastSplit.instDistrib;
        for (int i = 0; i < attrCount; i++) {
            if (!select[i] || lastSplit.usedAtts[i] || i == classIndex)
                continue;

            int n = attrValCount[i];
            int[][] distrib = new int[n + 1][classCount];
            int[] sum = new int[n + 1];
            int in = data.numInstances();

            for (int nr = 0; nr < in; nr++) {
                Instance instance = data.instance(nr);
                if (instance.isMissing(i)) {
                    distrib[n][(int) instance.classValue()]++;
                    sum[n]++;
                } else {
                    int val = (int) instance.value(i);
                    distrib[val][(int) instance.classValue()]++;
                    sum[val]++;
                }
            }
            in -= sum[n];
            Split sp = new Split();
            sp.attribut = i;
            boolean insert = true;
            double[] dv2 = new double[classCount];
            for (int v = 0; v < n; v++) {

                if (sum[v] == data.numInstances()) {
                    insert = false;
                    break;
                }
                int[] p = distrib[v].clone();
                Arrays.sort(p);
                for (int c = 0; c < classCount; c++) {
                    dv2[c] += p[c]/*  /sum[v]*sum[v] */ / (double) in;
                }
            }

            sp.sortDist = dv2;

            Iterator<Split> it = list.iterator();
            while (insert && it.hasNext()) {
                Split s = it.next();
                if (pref(s, sp)) {
                    insert = false;
                    break;
                }
                if (pref(sp, s)) {
                    it.remove();
                }
            }
            if (insert) {
                sp.usedAtts = lastSplit.usedAtts.clone();
                sp.usedAtts[i]=true;
                list.add(sp);
            }
        }
        int sc = list.size();

        if (sc == 0)
            return null;
        comb *= sc;
        int sel = (int) (Math.random() * sc);

        Split newSplit = list.get(sel);
        int iValue = (int) inst.value(newSplit.attribut);
        if (newSplit != null) {
            newSplit.data = new Instances(data, 0);
            int in = data.numInstances();
            for (int nr = 0; nr < in; nr++) {
                Instance instance = data.instance(nr);
                if (!instance.isMissing(newSplit.attribut)
                        && (int) instance.value(newSplit.attribut) == iValue) {
                    newSplit.data.add(instance);
                }
            }
            newSplit.instDistrib = getClassDistribution(newSplit.data);
        }
        return newSplit;
    }

    public void weight(double[] dist1, double[] normdist) {
        double sum = 0;
        for (int i = 0; i < classCount; i++) {
            double val = dist1[i] / normdist[i];
            sum += val;
            dist1[i] = val;
        }
        for (int i = 0; i < classCount; i++) {
            dist1[i] /= sum;
        }
    }

    public boolean pref(Split split1, Split split2) {
        double sum1 = 0;
        double sum2 = 0;
        for (int c = classCount-1; c >=0; c--) {
            sum1 += split1.sortDist[c];
            sum2 += split2.sortDist[c];
            if (sum2 > sum1)
                return false;
        }
        return true;
    }

    public Enumeration enumerateMeasures() {
        Vector newVector = new Vector(1);
        newVector.addElement("measureTreeCount");
        return newVector.elements();
    }

    public double getMeasure(String additionalMeasureName) {
        if (additionalMeasureName.compareToIgnoreCase("measureTreeCount") == 0) {
            return treeCount;
        } else {
            throw new IllegalArgumentException(additionalMeasureName
                    + " not supported (LazyEnsembleForestDT)");
        }
    }

}
