package weka.classifiers.trees;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.core.AdditionalMeasureProducer;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.Utils;
import weka.filters.Filter;
import weka.filters.supervised.attribute.Discretize;

public class LazyForestDT extends Classifier implements Serializable, OptionHandler, AdditionalMeasureProducer {

    protected static class Split implements Serializable {
        public int attribut;

        public int value;

        public Instances data;

        public double[] instDistrib;

        public boolean[] usedAtts;

        public double gain;
    }

    private static final long serialVersionUID = 695872586476417759L;

    private static final double log2 = Math.log(2);

    protected Instances data;

    protected int attrCount;

    protected int classCount;

    protected int classIndex;

    protected Split split;

    protected int[] attrValCount;
    
    protected int treeCount=15;
    
    protected int selectedAttributCount;
    
    protected double aVal=1;
    
    protected Discretize disc;

    public void buildClassifier(Instances builddata) throws Exception {
        disc = new Discretize();
        disc.setInputFormat(builddata);
        this.data = Filter.useFilter(builddata, disc);
        classIndex = data.classIndex();
        classCount = data.attribute(data.classIndex()).numValues();
        attrCount = data.numAttributes();
        attrValCount = new int[attrCount];
        if (aVal<1)  selectedAttributCount=(int)((attrCount-1)*aVal);
        else selectedAttributCount=Math.min(attrCount-1,(int)(aVal));
        for (int i = 0; i < attrCount; i++) {
            attrValCount[i] = data.attribute(i).numValues();
        }
        split = new Split();
        split.data = data;
        split.attribut = -1;
        split.usedAtts = new boolean[attrCount];
        split.instDistrib = getClassDistribution(data);
    }

    public double[] distributionForInstance(Instance inst) {
        disc.input(inst);
        inst=disc.output();
        double[] dist = new double[classCount];
        for (int i = 0; i < treeCount; i++) {
            double[] subdist = buildNode(split, inst,0);
            for (int c = 0; c < classCount; c++) {
                dist[c] += subdist[c] / treeCount;
            }
        }
        return dist;
    }

    protected double[] buildNode(Split split, Instance inst, int dep) {
        Split newSplit = getSplits(split, inst);
        if (newSplit == null) {
            //System.out.println(dep);
            double[] dist = split.instDistrib;
            return dist;
        } else if (newSplit.gain == 0.0) {
//            System.out.println("  attribut=" + newSplit.attribut + "  value="
//                    + newSplit.value);
            //System.out.println(dep);
            return newSplit.instDistrib;
        }
//        System.out.println("  attribut=" + newSplit.attribut + "  value="
//                + newSplit.value);
        double[] subDist = buildNode(newSplit, inst,dep+1);
        return subDist;
    }

    protected double[] getClassDistribution(Instances data) {
        double[] dist = new double[classCount];
        for (int i = 0; i < data.numInstances(); i++) {
            dist[(int) data.instance(i).classValue()]++;
        }
        for (int i = 0; i < classCount; i++) {
            dist[i] = dist[i] / data.numInstances();
        }
        return dist;
    }

    protected double[] getWeighedClassDistribution(Instances data,
            double[] preDist) {
        double[] dist = new double[classCount];
        double sum = 0;
        for (int i = 0; i < data.numInstances(); i++) {
            int cl = (int) data.instance(i).classValue();
            dist[cl] += 1.0 / preDist[cl];
            sum += 1.0 / preDist[cl];
        }
        for (int i = 0; i < classCount; i++) {
            dist[i] = dist[i] / sum;
        }
        return dist;
    }
    
    protected boolean[] getSelected(int count, boolean[] used){
        int a=0;
        for (int c=0;c<attrCount;c++){
            if (c!=classIndex && !used[c]) a++;
        }
        count=Math.min(a,count);
        boolean[] select=new boolean[attrCount];
        for(int i=0;i<count;i++){
            int sel=-1;
            do{
                sel=(int)(Math.random()*attrCount);
            }while (sel==classIndex || select[sel]|| used[sel]);
            select[sel]=true;
        }
        return select;
    }

    protected Split getSplits(Split lastSplit, Instance inst) {
        boolean[] select=getSelected(selectedAttributCount,lastSplit.usedAtts);
        Split newSplit = null;
        Instances data = lastSplit.data;
        double[] weightdist = lastSplit.instDistrib;
        for (int i = 0; i < attrCount; i++) {
            if (!select[i] || lastSplit.usedAtts[i] || i == classIndex
                    || inst.isMissing(i))
                continue;

            int n = attrValCount[i];
            int[][] distrib = new int[n + 1][classCount];
            int[] sum = new int[n + 1];
            int in = data.numInstances();
            int iValue = (int) inst.value(i);

            for (int nr = 0; nr < in; nr++) {
                Instance instance = data.instance(nr);
                if (instance.isMissing(i)) {
                    distrib[n][(int) instance.classValue()]++;
                    sum[n]++;
                } else {
                    int val = (int) instance.value(i);
                    distrib[val][(int) instance.classValue()]++;
                    sum[val]++;
                }
            }
            for (int v = 0; v < n; v++) {
                if (v != iValue || sum[v] == 0 || sum[v] == data.numInstances())
                    continue;
                Split sp = new Split();

                sp.attribut = i;
                sp.value = v;
                double gain = 0;
                boolean insert = true;

                double entropy2 = 0;
                double[] dv2 = new double[classCount];
                double sum2 = 0;
                for (int a = 0; a <= n; a++) {
                    if (a == v) {
                        for (int c = 0; c < classCount; c++) {
                            if (weightdist[c] > 0) {
                                dv2[c] += distrib[a][c] * 1 / weightdist[c];
                                sum2 += distrib[a][c] * 1 / weightdist[c];
                            }
                        }
                    }
                }
                for (int c = 0; c < classCount; c++) {
                    dv2[c] /= sum2;
                }
                for (int c = 0; c < classCount; c++) {
                    double p = dv2[c];
                    if (p > 0.0)
                        entropy2 += -p * Math.log(p) / log2;
                }
                gain = entropy2;
                sp.gain = gain;

                if (insert && newSplit != null && pref(newSplit, sp)) {
                    insert = false;
                }
                if (insert) {
                    sp.usedAtts = lastSplit.usedAtts.clone();
                    sp.usedAtts[i]=true;
                    newSplit = sp;
                }
            }
        }
        if (newSplit != null) {
            newSplit.data = new Instances(data, 0);
            int in = data.numInstances();
            for (int nr = 0; nr < in; nr++) {
                Instance instance = data.instance(nr);
                if (!instance.isMissing(newSplit.attribut)
                        && (int) instance.value(newSplit.attribut) == newSplit.value) {
                    newSplit.data.add(instance);
                }
            }
            newSplit.instDistrib = getClassDistribution(newSplit.data);
        } 
        return newSplit;
    }

    public void weight(double[] dist1, double[] normdist) {
        double sum = 0;
        for (int i = 0; i < classCount; i++) {
            double val = dist1[i] / normdist[i];
            sum += val;
            dist1[i] = val;
        }
        for (int i = 0; i < classCount; i++) {
            dist1[i] /= sum;
        }
    }

    public boolean pref(Split s1, Split s2) {
        return s1.gain < s2.gain;

    }
    
    
    /**
     * Returns an enumeration describing the available options.
     *
     * @return an enumeration of all the available options.
     */
    public Enumeration listOptions() {

      Vector newVector = new Vector(2);
      

      newVector.addElement(new Option(
          "\tSet the number of trees",
          "K", 1, "-K <number of trees>"));
      newVector.addElement(new Option(
              "\tSet the number of selected attributes",
              "A", 1, "-A <number of selected attributes or part of number of attributs>"));
      return newVector.elements();
    }

    /**
     * Parses a given list of options. Valid options are:<p>
     *
     * -D  <br>
     * If set, classifier is run in debug mode and 
     * may output additional info to the console.<p>
     *
     * @param options the list of options as an array of strings
     * @exception Exception if an option is not supported
     */
    public void setOptions(String[] options) throws Exception {

      String val=Utils.getOption('K', options);
      if (val!=null && !val.equals("")){
          treeCount=Integer.parseInt(val);
      }
      val=Utils.getOption('A', options);
      if (val!=null && !val.equals("")){
          aVal=Double.parseDouble(val);
      }
    }

    /**
     * Gets the current settings of the Classifier.
     *
     * @return an array of strings suitable for passing to setOptions
     */
    public String [] getOptions() {

      String [] options=new String[]{"-K",""+treeCount,"-A",""+aVal};
      return options;
    }
    
    public Enumeration enumerateMeasures() {
        Vector newVector = new Vector(1);
        
        return newVector.elements();
    }

    public double getMeasure(String measureName) {
        // TODO Auto-generated method stub
        return 0;
    }
    
    public void setTreeCount(int treeCount){
        this.treeCount=treeCount;
    }
    
    public int getTreeCount(){
        return treeCount;
    }
    
    public void setAttributCount(double aVal){
        this.aVal=aVal;
    }
    
    public double getAttributCount(){
        return aVal;
    }
    
    


}
