/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Partitioning
 *    Copyright (C) 2004 Jrgen Beringer
 *
 */
package weka.filters.supervised.attribute;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;

import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.SparseInstance;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;
import weka.filters.Filter;
import weka.filters.SupervisedFilter;

/** 
 * An instance filter that discretizes a range of numeric attributes in 
 * the dataset into nominal attributes. Discretization is by 
 * Fayyad & Irani's MDL method (the default).<p>
 *
 * 
 * @author Jrgen Beringer (beringer@mathematik.uni-marburg.de)
 * @version $Revision: 1.1 $
 */
public class Partitioning extends Filter implements SupervisedFilter,WeightedInstancesHandler {

	protected int classCount;

	protected int attCount;

	protected static double log2 = Math.log(2);

	protected List[] splits;

	protected double[][] splitValues;

	protected Instances input;

	protected double[] min;

	protected double[] max;

	public boolean batchFinished() throws Exception {

		input = getInputFormat();

		if (input == null) { throw new IllegalStateException(
				"No input instance format defined"); }

		initialize();
		findSplits();
		setOutputFormat();
		discretize();
		return (numPendingOutput() != 0);
	}

	protected void initialize() {
		this.classCount = input.classAttribute().numValues();
		this.attCount = input.numAttributes();
		splits = new List[attCount + 1];
		splitValues = new double[attCount + 1][];
		min = new double[attCount + 1];
		max = new double[attCount + 1];
		for (int i = 0; i <= attCount; i++) {
			splits[i] = new ArrayList();
		}
	}

	protected void findSplits() {
		for (int i = 0; i < attCount; i++) {
			if (i == input.classIndex()) continue;
			if (!input.attribute(i).isNumeric()) continue;
			findSplits(i);
		}
	}

	protected void findSplits(int attribute) {
		double[] values = extractValues(attribute);
		min[attribute] = values[0];
		max[attribute] = values[values.length - 1];
		findSplits(attribute, values, 0, values.length - 1, input);
		Collections.sort(splits[attribute]);
		splitValues[attribute] = new double[splits[attribute].size()];
		Iterator it = splits[attribute].iterator();
		int pos = 0;
		while (it.hasNext()) {
			splitValues[attribute][pos] = ((Double) it.next()).doubleValue();
			pos++;
		}
	}

	protected void findSplits(int attribute, double[] values, int start,
			int end, Instances input) {
		double last = values[start];
		int pos = start + 1;
		double[] ent = entropy(input);
		double entS = ent[0];
		int classCountS = (int) ent[1];
		double minValue = Double.NaN;
		int minPos = -1;
		while (pos < end) {
			while (values[pos] <= last && pos < end) {
				pos++;
			}
			if (pos < end) {
				double val = compute(attribute, values[pos], input, entS,
						classCountS);
				if (!Double.isNaN(val)) {
					if (Double.isNaN(minValue) || (minValue > val)) {
						minValue = val;
						minPos = pos;
					}
				}
				last = values[pos];
				pos++;
			}
		}
		if (minPos >= 0) {
			splits[attribute].add(new Double(
					(values[minPos] + values[minPos - 1]) / 2.0));
			Instances[] lists = split(attribute, values[minPos], input);
			Instances s1 = lists[0];
			Instances s2 = lists[1];
			findSplits(attribute, values, start, minPos - 1, s1);
			findSplits(attribute, values, minPos, end, s2);
		}
	}

	protected double[] extractValues(int attribute) {
		int size = input.numInstances();
		double[] data = new double[size];
		Enumeration it = input.enumerateInstances();
		int pos = 0;
		while (it.hasMoreElements()) {
			Instance value = (Instance) it.nextElement();
			data[pos] = value.value(attribute);
			pos++;
		}
		Arrays.sort(data);
		int dupl = 0;
		for (int i = 1; i < size; i++) {
			if (data[i] == data[i - 1]) dupl++;
		}

		double[] result = new double[size - dupl - 1];
		result[0] = data[0];
		pos = 0;
		for (int i = 1; i < size; i++) {
			if (data[i] != data[i - 1]) {
				result[pos] = data[i];
				pos++;
			}
		}
		return result;
	}

	protected double compute(int attribute, double t, Instances sample,
			double entS, int classCountS) {
		Instances[] lists = split(attribute, t, sample);
		Instances s1 = lists[0];
		Instances s2 = lists[1];

		double[] res1 = entropy(s1);
		double entS1 = res1[0];
		int ccS1 = (int) res1[1];
		double sizeS1 = res1[2];
		double[] res2 = entropy(s2);
		double entS2 = res2[0];
		int ccS2 = (int) res2[1];
		double sizeS2 = res2[2];
		double size = sizeS1 + sizeS2;
		double val = sizeS1 / size * entS1 + sizeS2 / size * entS2;
		double delta = log2(Math.pow(3, classCountS) - 2)
				- (classCountS * entS - ccS1*entS1 - ccS2*entS2);
		if (entS - val < (log2(size - 1) + delta)
				/ size) return Double.NaN;
		return val;
	}

	protected Instances[] split(int attribute, double t, Instances sample) {
		Instances s1 = new Instances(sample, 0);
		Instances s2 = new Instances(sample, 0);
		int pos = 0;
		while (pos < sample.numInstances()) {
			Instance value = sample.instance(pos);
			if (value.value(attribute) >= t) {
				s2.add(value);
			} else {
				s1.add(value);
			}
			pos++;
		}
		return new Instances[] { s1, s2};
	}

	protected double[] entropy(Instances list) {
		double[] counter = new double[classCount];
		Enumeration it = list.enumerateInstances();
		double size = 0;
		while (it.hasMoreElements()) {
			Instance value = (Instance) it.nextElement();
			size += value.weight();
			counter[(int) value.classValue()] += value.weight();
		}
		double entropy = 0.0;
		int diffClasses = 0;
		for (int i = 0; i < classCount; i++) {
			if (counter[i] > 0) diffClasses++;
			counter[i] /= size;
			if (counter[i] > 0) entropy -= counter[i] * log2(counter[i]);
		}
		return new double[] { entropy, diffClasses, size};
	}

	private double log2(double d) {
		return Math.log(d) / log2;
	}

	/**
	 * Set the output format. Takes the currently defined cutpoints and 
	 * m_InputFormat and calls setOutputFormat(Instances) appropriately.
	 */
	protected void setOutputFormat() {

		FastVector attributes = new FastVector(getInputFormat().numAttributes());
		int classIndex = getInputFormat().classIndex();
		for (int i = 0; i < getInputFormat().numAttributes(); i++) {
			if ((getInputFormat().attribute(i).isNumeric())
					&& (getInputFormat().classIndex() != i)) {

				FastVector attribValues = new FastVector(1);
				if (splits[i].size() == 0) {
					attribValues.addElement("'All'");
				} else {
					for (int j = 0; j <= splits[i].size(); j++) {
						if (j == 0) {
							attribValues.addElement("'(-inf-"
									+ Utils.doubleToString(((Number) splits[i]
											.get(j)).doubleValue(), 6) + "]'");
						} else if (j == splits[i].size()) {
							attribValues.addElement("'("
									+ Utils.doubleToString(((Number) splits[i]
											.get(j - 1)).doubleValue(), 6)
									+ "-inf)'");
						} else {
							String att = "'("
									+ Utils.doubleToString(((Number) splits[i]
											.get(j - 1)).doubleValue(), 6)
									+ "-"
									+ Utils.doubleToString(((Number) splits[i]
											.get(j)).doubleValue(), 6) + "]'";
							while (attribValues.indexOf(att) >= 0)
								att = att + "b";
							attribValues.addElement(att);
						}
					}
				}
				attributes.addElement(new Attribute(getInputFormat().attribute(
						i).name(), attribValues));

			} else {
				attributes.addElement(getInputFormat().attribute(i).copy());
			}
		}
		Instances outputFormat = new Instances(getInputFormat().relationName(),
				attributes, 0);
		outputFormat.setClassIndex(input.classIndex());
		setOutputFormat(outputFormat);
	}

	protected void discretize() {
		Enumeration it = input.enumerateInstances();
		Instances output = getOutputFormat();
		while (it.hasMoreElements()) {
			Instance instance = (Instance) it.nextElement();
			int index = 0;
			double[] vals = new double[outputFormatPeek().numAttributes()];
			// Copy and convert the values
			for (int i = 0; i < getInputFormat().numAttributes(); i++) {
				if (getInputFormat().attribute(i).isNumeric()
						&& (getInputFormat().classIndex() != i)) {

					double currentVal = instance.value(i);
					if (instance.isMissing(i)) {
						vals[index] = Instance.missingValue();
					} else {
						vals[index] = discretize(i, currentVal);
					}
					index++;
				} else {
					vals[index] = instance.value(i);
					index++;
				}
			}
			Instance newVal = null;
			if (instance instanceof SparseInstance) {
				newVal = new SparseInstance(instance.weight(), vals);
			} else {
				newVal = new Instance(instance.weight(), vals);
			}
			newVal.setDataset(output);
			push(newVal);
		}
	}

	public int discretize(int attribute, double value) {
		int val = 0;
		while (val < splitValues[attribute].length) {
			double t = splitValues[attribute][val];
			if (value > t) {
				val++;
			} else
				break;
		}
		return val;
	}

	public double[] getSplitValues(int attribute) {
		return splitValues[attribute];
	}

	public double[] getMinValues() {
		return min;
	}

	public double[] getMaxValues() {
		return max;
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < attCount; i++) {
			if (splitValues[i] == null) continue;
			sb.append("Splits of Attribute " + i + "\n");
			for (int j = 0; j < splitValues[i].length; j++) {
				sb.append(splitValues[i][j] + " ");
			}
			sb.append("\n");

		}
		return sb.toString();
	}
}