/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Normalize.java
 *    Copyright (C) 1999 Eibe Frank
 *
 */

package weka.filters.unsupervised.attribute;

import weka.filters.*;
import java.io.*;
import java.util.*;
import weka.core.*;

/**
 * Normalizes all numeric values in the given dataset. The resulting values are
 * in [0,1] for the data used to compute the normalization intervals.
 * 
 * @author Eibe Frank (eibe@cs.waikato.ac.nz)
 * @version $Revision: 1.1 $
 */
public class NormalNormalize extends PotentialClassIgnorer implements
        UnsupervisedFilter {

    private double[] sum;

    private double[] sum2;

    private double[] count;

    private double[] mean;

    private double[] std;

    /**
     * Returns a string describing this filter
     * 
     * @return a description of the filter suitable for displaying in the
     *         explorer/experimenter gui
     */
    public String globalInfo() {

        return "Normalizes all numeric values in the given dataset (apart from the "
                + "class attribute, if set). The resulting attributes "
                + "have mean 0 and standard deviation 1 for the data used to compute the normalization intervals. ";
    }

    /**
     * Sets the format of the input instances.
     * 
     * @param instanceInfo
     *            an Instances object containing the input instance structure
     *            (any instances contained in the object are ignored - only the
     *            structure is required).
     * @return true if the outputFormat may be collected immediately
     * @exception Exception
     *                if the input format can't be set successfully
     */
    public boolean setInputFormat(Instances instanceInfo) throws Exception {

        super.setInputFormat(instanceInfo);
        setOutputFormat(instanceInfo);
        sum = sum2 = count = mean = std = null;
        return true;
    }

    /**
     * Input an instance for filtering. Filter requires all training instances
     * be read before producing output.
     * 
     * @param instance
     *            the input instance
     * @return true if the filtered instance may now be collected with output().
     * @exception IllegalStateException
     *                if no input format has been set.
     */
    public boolean input(Instance instance) {

        if (getInputFormat() == null) {
            throw new IllegalStateException("No input instance format defined");
        }
        if (m_NewBatch) {
            resetQueue();
            m_NewBatch = false;
        }
        if (sum == null) {
            bufferInput(instance);
            return false;
        } else {
            convertInstance(instance);
            return true;
        }
    }

    /**
     * Signify that this batch of input to the filter is finished. If the filter
     * requires all instances prior to filtering, output() may now be called to
     * retrieve the filtered instances.
     * 
     * @return true if there are instances pending output
     * @exception IllegalStateException
     *                if no input structure has been defined
     */
    public boolean batchFinished() {

        if (getInputFormat() == null) {
            throw new IllegalStateException("No input instance format defined");
        }
        if (sum == null) {
            Instances input = getInputFormat();
            // Compute minimums and maximums
            sum = new double[input.numAttributes()];
            sum2 = new double[input.numAttributes()];
            count = new double[input.numAttributes()];
            mean = new double[input.numAttributes()];
            std = new double[input.numAttributes()];

            for (int j = 0; j < input.numInstances(); j++) {
                double[] value = input.instance(j).toDoubleArray();
                for (int i = 0; i < input.numAttributes(); i++) {
                    if (input.attribute(i).isNumeric()
                            && (input.classIndex() != i)) {
                        if (!Instance.isMissingValue(value[i])) {
                            sum[i] += value[i];
                            sum2[i] += value[i] * value[i];
                            count[i]++;
                        }
                    }
                }
            }
            for (int i = 0; i < input.numAttributes(); i++) {
                if (input.attribute(i).isNumeric() && (input.classIndex() != i)) {
                    mean[i] = sum[i] / count[i];
                    std[i] = Math.sqrt((sum2[i] / count[i]) - (mean[i] * mean[i]));
                }
            }
            // Convert pending input instances
            for (int i = 0; i < input.numInstances(); i++) {
                convertInstance(input.instance(i));
            }
        }
        // Free memory
        flushInput();

        m_NewBatch = true;
        return (numPendingOutput() != 0);
    }

    /**
     * Convert a single instance over. The converted instance is added to the
     * end of the output queue.
     * 
     * @param instance
     *            the instance to convert
     */
    private void convertInstance(Instance instance) {

        Instance inst = null;
        if (instance instanceof SparseInstance) {
            double[] newVals = new double[instance.numAttributes()];
            int[] newIndices = new int[instance.numAttributes()];
            double[] vals = instance.toDoubleArray();
            int ind = 0;
            for (int j = 0; j < instance.numAttributes(); j++) {
                double value;
                if (instance.attribute(j).isNumeric()
                        && (!Instance.isMissingValue(vals[j]))
                        && (getInputFormat().classIndex() != j)) {
                    if (Double.isNaN(mean[j]) || (std[j] == 0)) {
                        value = 0;
                    } else {
                        value = (vals[j] - mean[j]) / (std[j]);
                    }
                    if (value != 0.0) {
                        newVals[ind] = value;
                        newIndices[ind] = j;
                        ind++;
                    }
                } else {
                    value = vals[j];
                    if (value != 0.0) {
                        newVals[ind] = value;
                        newIndices[ind] = j;
                        ind++;
                    }
                }
            }
            double[] tempVals = new double[ind];
            int[] tempInd = new int[ind];
            System.arraycopy(newVals, 0, tempVals, 0, ind);
            System.arraycopy(newIndices, 0, tempInd, 0, ind);
            inst = new SparseInstance(instance.weight(), tempVals, tempInd,
                    instance.numAttributes());
        } else {
            double[] vals = instance.toDoubleArray();
            for (int j = 0; j < getInputFormat().numAttributes(); j++) {
                if (instance.attribute(j).isNumeric()
                        && (!Instance.isMissingValue(vals[j]))
                        && (getInputFormat().classIndex() != j)) {
                    if (Double.isNaN(mean[j]) || (std[j] == 0)) {
                        vals[j] = 0;
                    } else {
                        vals[j] = (vals[j] - mean[j]) / (std[j]);
                    }
                }
            }
            inst = new Instance(instance.weight(), vals);
        }
        inst.setDataset(instance.dataset());
        push(inst);
    }

    /**
     * Main method for testing this class.
     * 
     * @param argv
     *            should contain arguments to the filter: use -h for help
     */
    public static void main(String[] argv) {

        try {
            if (Utils.getFlag('b', argv)) {
                Filter.batchFilterFile(new NormalNormalize(), argv);
            } else {
                Filter.filterFile(new NormalNormalize(), argv);
            }
        } catch (Exception ex) {
            System.out.println(ex.getMessage());
        }
    }
}
