/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.awt;

import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.MouseEvent;

import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;
import javax.swing.event.EventListenerList;

import xxl.applications.gui.actions.AddNewAction;
import xxl.applications.gui.actions.MainActionManager;
import xxl.applications.gui.graph.Graph;
import xxl.applications.gui.graph.Vertex;
import xxl.applications.gui.graph.event.ConnectionEvent;
import xxl.applications.gui.graph.event.ConnectionListener;
import xxl.applications.gui.graph.viewer2d.SimpleArrow;
import xxl.applications.gui.windows.GraphFrame;

/**
 * Tool used for connecting graph nodes
 * @author maxim
 */
public class ConnectingTool extends AbstractTool {

	/** singleton */
	private static ConnectingTool sharedInstance;

	/** getter for tool instance */
	public static ConnectingTool getInstance() {
		if (sharedInstance == null)
			sharedInstance = new ConnectingTool();
		return sharedInstance;
	}

	/** source vertex, caller */
	private transient Vertex source;
	/** start point of the connecting line */
	private transient Point start;
	/** end point of the connection line */
	private transient Point end;
	/** representation of the connecting line */
	private transient SimpleArrow arrow;
	/** popup menu bounded to this tool */
	private transient ContextMenu contextMenu;
	/** event listeners*/
	private transient EventListenerList eventListeners = new EventListenerList();

	/** creates new ConnectionTool instance operating on <code>panel</code>  */
	private ConnectingTool() {
		super();
		arrow = new SimpleArrow(new Point(), new Point());
	}

	//TODO this is an ugly hack too...
	public void setPane(DefaultGraphPanel pane) {
		Container parent =
			getPane() == null ? null : SwingUtilities.getAncestorOfClass(GraphFrame.class, getPane());
		if (parent != null) {
			removeConnectionListener((GraphFrame) parent);
		}
		if (pane != null) {
			parent = SwingUtilities.getAncestorOfClass(GraphFrame.class, pane);
			if (parent != null) {
				addConnectionListener((GraphFrame) parent);
			}
		}
		super.setPane(pane);
	}

	public void addConnectionListener(ConnectionListener listener) {
		eventListeners.add(ConnectionListener.class, listener);
	}

	public void removeConnectionListener(ConnectionListener listener) {
		eventListeners.remove(ConnectionListener.class, listener);
	}

	protected void fireConnectionEvent(ConnectionEvent ce) {
		Object[] listeners = eventListeners.getListenerList();
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == ConnectionListener.class)
				 ((ConnectionListener) listeners[i + 1]).handleConnection(ce);
		}
	}

	/** is called by mouse press on the panel, sets the source vertex */
	public void mousePressed(MouseEvent e) {
		super.mousePressed(e);
		if (e.getButton() == MouseEvent.BUTTON1) {
			source = getPane().getDrawer().vertexAtPoint(e.getPoint());
		}
	}

	/** performs drawing during mouse dragging on the panel */
	public void mouseDragged(MouseEvent e) {
		if (source != null) {
			end = e.getPoint();
			double angle = angle(source.getLocation(), end);
			start = getPane().getDrawer().pointOnVertex(source, angle);
			getPane().repaint();
		}
	}

	public void mouseReleased(MouseEvent e) {
		super.mouseReleased(e);
		Vertex destination = getPane().getDrawer().vertexAtPoint(e.getPoint());
		if (destination != null && source != null) {
			Graph g = getPane().getDrawer().getGraph();
			fireConnectionEvent(new ConnectionEvent(g, source, destination));
		}
		source = null;
		start = end = null;
		getPane().repaint();
	}

	/** is called for reflection tool operations of the panel */
	public void draw(Graphics g) {
		if (start != null && end != null) {
			//transform points to view
			arrow.setStart(getPane().transformFromGraphToView(start));
			arrow.setEnd(getPane().transformFromGraphToView(end));
			arrow.draw((Graphics2D) g);
		}
	}

	private static double angle(Point p1, Point p2) {
		double x = (double) (p2.x - p1.x);
		double y = (double) (p2.y - p1.y);
		return Math.atan2(y, x);
	}

	protected JPopupMenu getContexMenu() {
		if (contextMenu == null)
			contextMenu = new ContextMenu();
		return contextMenu;
	}

	private class ContextMenu extends AbstractPopupMenu {
		public ContextMenu() {
			super(MainActionManager.sharedInstance());
		}

		protected void configureMenu() {
			addMenuItem(getManager().getAction(AddNewAction.ACTION_COMMAND_VALUE));			
		}
	}
}
