/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.awt;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.border.Border;

/**
 * @author maxim
 * @version 1.0
 */

public final class StatusBar extends JPanel {

	private static StatusBar sharedInstance;

	static final String MAIN_CELL = "main";
	private static final String FILE_CELL = "file";
	private static final String MOUSE_CELL = "mouse";
	private static final String[] FILESTRINGS =
		{ SimpleDateFormat.getDateTimeInstance().format(new Date())};
	private static final String[] MOUSESTRINGS = { "X:9999, Y:9999" };

	private static final Insets NULL_INSETS = new Insets(0, 0, 0, 0);
	private static final Border INNER_BORDER =
		BorderFactory.createCompoundBorder(
			BorderFactory.createMatteBorder(
				1,
				1,
				1,
				1,
				UIManager.getDefaults().getColor("controlDkShadow")),
			BorderFactory.createLoweredBevelBorder());

	private static final Border CELL_BORDER_LEFT =
		BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(2, 0, 0, 2), INNER_BORDER);

	private static final Border CELL_BORDER =
		BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(2, 2, 0, 2), INNER_BORDER);

	private List cellList = new Vector();

	public static StatusBar sharedInstance() {
		if (sharedInstance == null)
			sharedInstance = new StatusBar();
		return sharedInstance;
	}

	public StatusBar() {
		super(new GridBagLayout());
		initPanel();
		setVisible(true);
	}

	public void setVisible(boolean visible) {
		if (visible)
			refreshBar();
		super.setVisible(visible);

	}

	private void initPanel() {
		addCell(-1, MAIN_CELL, null);
		addCell(-1, FILE_CELL, FILESTRINGS).setHorizontalAlignment(SwingConstants.CENTER);
		addCell(-1, MOUSE_CELL, MOUSESTRINGS).setHorizontalAlignment(SwingConstants.CENTER);

	}

	private Cell addCell(int pos, String name, String[] widestString) {
		Cell cell = new Cell(name, widestString);
		addCell(pos, cell);
		return cell;
	}

	public void notify(String msg) {
		getCellByName(MAIN_CELL).setText(msg);
	}

	private void addCell(int pos, Cell cell) {
		if (pos < 0 || pos > cellList.size()) {
			cellList.add(cell);
		}
		else
			cellList.add(pos, cell);
		refreshBar();
	}

	public JLabel getCellByName(String name) {
		Iterator iter = cellList.iterator();
		while (iter.hasNext()) {
			Cell cell = (Cell) iter.next();
			if (cell.getName().equals(name))
				return cell;
		}
		return null;
	}

	private void refreshBar() {
		
		if (!SwingUtilities.isEventDispatchThread()) {
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					refreshBar();
				}
			});
			return;
		}
		Iterator iter = cellList.iterator();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridx = GridBagConstraints.RELATIVE;
		gbc.gridwidth = 1;
		gbc.gridheight = 1;
		gbc.weighty = GridBagConstraints.VERTICAL;
		while (iter.hasNext()) {
			Cell cell = (Cell) iter.next();
			cell.setPreferredSize(cell.getPreferedSize());
			cell.setMinimumSize(cell.getMinimumSize());
			if (MAIN_CELL.equals(cell.getName())) {
				gbc.fill = GridBagConstraints.BOTH;
				gbc.weightx = 1.0;
			}
			else {
				gbc.fill = GridBagConstraints.VERTICAL;
				gbc.weightx = 0.0;
			}
			StatusBar.this.add(cell, gbc);
		}
	}

	static class Cell extends JLabel {
		private String[] widestStrings;
		private Dimension dimension;

		public Cell(String name, String[] widestStrings) {
			super();
			setName(name);
			setOpaque(true);
			this.widestStrings = widestStrings;
			setBorder(CELL_BORDER);
			updateSize();
		}

		private void updateSize() {
			Font font = getFont();
			if (dimension == null)
				dimension = new Dimension();
			if (font != null) {
				Border b = getBorder();
				Insets ins = (b != null) ? b.getBorderInsets(this) : NULL_INSETS;
				FontMetrics fm = getFontMetrics(font);
				int mw = fm.stringWidth(this.getText());
				dimension.height = fm.getHeight() + ins.top + ins.bottom;
				if (widestStrings != null) {
					for (int i = 0; i < widestStrings.length; i++) {
						mw = Math.max(mw, fm.stringWidth(widestStrings[i]));
					}
				}
				dimension.width = mw + ins.left + ins.right;
			}
		}

		public Dimension getPreferedSize() {
			if (dimension == null)
				updateSize();
			return new Dimension(dimension);
		}

		public Dimension getMinimumSize() {
			return getPreferedSize();
		}

		public void setFont(Font font) {
			super.setFont(font);
			updateSize();
		}
	}

}
