/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.awt.tree;

import java.util.TreeSet;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import xxl.applications.gui.actions.CallableAction;

/**
 * @author maxim
 * 
 */
public abstract class TreeObject extends DefaultMutableTreeNode {

	public static final int PATH_ENTRY_OBJECT = 0x0000;
	public static final int JAR_FILE_OBJECT = 0x0001;
	public static final int PACKAGE_OBJECT = 0x0002;
	public static final int CLASS_OBJECT = 0x0004;
	public static final int CATEGORY_OBJECT = 0x0008;

	static final CallableAction[] EMPTY_ACTIONS = new CallableAction[0];

	/** node type */
	private int type = 0;
	/** node data */
	protected String objectEntry;
	/** model to inform about node changes */
	protected DefaultTreeModel model;
	/** invalidate behaviour on close */
	protected boolean removeChildrenOnClose = true;

	public TreeObject() {
		super();
	}

	public TreeObject(Object userObject) {
		super(userObject);
	}

	public void setModel(DefaultTreeModel model) {
		this.model = model;
	}

	public void setRemoveChildrenOnClose(boolean removeOnClose) {
		this.removeChildrenOnClose = removeOnClose;
	}

	public boolean shouldRemoveChildrenOnClose() {
		return removeChildrenOnClose;
	}

	protected void fireModelChange() {
		if (model != null)
			model.nodeStructureChanged(this);
	}

	protected void setObjectEntry(String objectEntry) {
		if (objectEntry == null)
			throw new IllegalArgumentException();
		this.objectEntry = objectEntry;
	}

	public boolean isPathObject() {
		return getType() == PATH_ENTRY_OBJECT;
	}

	public boolean isJarFileObject() {
		return getType() == JAR_FILE_OBJECT;
	}

	public boolean isPackageObject() {
		return getType() == PACKAGE_OBJECT;
	}

	public boolean isClassObject() {
		return getType() == CLASS_OBJECT;
	}

	public void setType(int type) {
		if (type < 0)
			throw new IllegalArgumentException();
		this.type = type;
	}

	public int getType() {
		return this.type;
	}

	public boolean getAllowsChildren() {
		return getType() != CLASS_OBJECT;
	}

	public boolean isLeaf() {
		return false;
	}

	public void invalidate() {
		if (shouldRemoveChildrenOnClose())
			removeAllChildren();
	}

	public void removeAllChildren() {
		super.removeAllChildren();
		fireModelChange();
	}

	public abstract CallableAction[] getActions();

	/**
	 * Node icon resource as string, could be passed to Utilities.loadIcon(String)
	 * @return icon resource as a string
	 */
	public abstract String getIconResource();
	/**
	 * Name of the represented resource e.g. name of a Class or a Jarfile
	 * @return object name
	 */
	public abstract String getObjectName();
	/**
	 * Path to represented resource e.g. package name for a Class or location path for a Jarfile
	 * Some resources have no path (or we are not intreseted in) e.g packages and libraries 	 
	 * @return
	 */
	public abstract String getObjectPath();
	/**
	 * Add children to the node, type of child node is determined by application means, 
	 * hier it is hierarchy of java resources: library(path)->jarfiles->packages->classes
	 * @param children TreeSet containig children data
	 */
	protected abstract void addChildren(TreeSet children);

}
