/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.code;

import java.awt.Point;

import xxl.applications.gui.graph.ComplexVertex;
import xxl.applications.gui.graph.Edge;
import xxl.applications.gui.graph.Vertex;

/**
 * Basic Class representing Java Items (Classes, Objects) in ClassGraph.
 * 
 * @author maxim
 */
public abstract class AbstractItem extends ComplexVertex {
	
	static final long serialVersionUID = -8949833337796472300L;
	
	/** array item */
	public static final int ARRAY_ITEM = 3;
	/** class, access only to static members */
	public static final int CLASS_ITEM = 1;
	/** type descriptions */
	private static final String[] DESC = { "Object", "Class", "Primitive", "Array", "Unknown" };
	/** object, need to be instantiated before use */
	public static final int OBJECT_ITEM = 0;
	/** one of java primitive types enclosing strings */
	public static final int PRIMITIVE_ITEM = 2;
	/** type of the by vertex enclosed type */
	protected Type enclosedType;
	/** last inserted call in the item, used for ordering calls */
	private transient CallVertex lastInserted;
	/** type of the vertex as java item */
	protected int type;

	/** provides default constructor for subclasses */
	protected AbstractItem() {
	}

	/** Copy constructor */
	public AbstractItem(AbstractItem item) {
		super(item);
		try {
			enclosedType = Type.createFromClass(item.enclosedType.toClass());
		}
		//should not happen since item is valid instance
		catch (ClassNotFoundException cnfe) {
			cnfe.printStackTrace();
		}
	}

	/** Creates an AbstractItem instance from class name */
	protected AbstractItem(Point location, String className) throws IllegalArgumentException {
		this();
		enclosedType = Type.parse(className);
		setLocation(location);
	}

	/** overrides the method from super class, since we need connect item's calls */
	protected Vertex addChildVertex(int position, Vertex child) {
		if (!(child instanceof CallVertex))
			throw new IllegalArgumentException("Add supports only CallVertex");
		Vertex ret = super.addChildVertex(position, child);
		getGraph().addVertex(child);
		if (lastInserted != null) {
			Edge e = new Edge(lastInserted, child);
			getGraph().addEdge(e);
			System.out.println("AbstractItem.addChildVertex() connecting " + e);
		}
		lastInserted = (CallVertex) child;
		return ret;
	}

	public void addChildVertex(Vertex child) {
		addChildVertex(-1, child);
	}

	/** Item description to show in the proeprty pane */
	public String getDescription() {
		int i = type;
		if (i < 0 || i >= DESC.length - 1)
			i = DESC.length - 1;
		return DESC[i];
	}

	/** Gets the type of the enclosed java item */
	public Type getEnclosedType() {
		return enclosedType;
	}

	//TODO there is a little problem with names...
	public String getName() {
		String name = "v" + getID();
		if (name.equals(super.getName()))
			return null;
		return super.getName();
	}

	/** Gets the type of the vertex as java item */
	public int getType() {
		return type;
	}

	/** Proves the item for validity */
	public abstract boolean isValid();

	/**
	 * Removes the <code>child</code> vertex from the children-list and also
	 * from the graph
	 */
	public void removeChildVertex(Vertex child) {
		if (child == null)
			return;
		super.removeChildVertex(child);
		graph.removeVertex(child);
	}

	public String toString() {
		StringBuffer str_buf = new StringBuffer(super.toString());
		str_buf.append('\n');
		for (int i = 0; i < children.size(); i++) {
			str_buf.append('\t');
			str_buf.append(children.get(i));
		}
		return str_buf.toString();
	}
}
