/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.code;

import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.util.StringTokenizer;

import xxl.applications.gui.graph.Edge;
import xxl.applications.gui.graph.Graph;

/**
 * Code element representing a call of a constructor.
 * @author maxim
 *
 */
public class ConstructorCall extends ElementCall {
	private static Type[] NULL_TYPES = new Type[0];
	static final long serialVersionUID = 7113883169601685846L;
	
	public final static ConstructorCall create(
	CallVertex caller,
	Constructor con,
	ObjectItem instance) {
		ConstructorCall call = new ConstructorCall(caller);
		Class[] params = con.getParameterTypes();
		call.setCallType(Type.createFromClass(con.getDeclaringClass()));
		Type[] types = new Type[params.length];
		for (int i = 0; i < params.length; i++) {
			types[i] = Type.createFromClass(params[i]);
		}
		call.setParamTypes(types);
		call.setIdentifier(instance.getName());
		return call;
	}
	/** Body of the anonymous class used by this call, should be null if creating regular instance */
	private String body;
	private Type[] parameterTypes;
	private ElementCall[] parameterValues;
	
	/**
	 * Creates new instance of ConstructorCall
	 * @param caller
	 * @param declaringClass
	 * @param paramTypes
	 * @param instanceName
	 */
	public ConstructorCall(CallVertex caller) {
		super(caller);
		setType(CONSTRUCTOR_CALL);
	}
	
	public void generateCode(PrintWriter writer, int indentation) {
		//		if (!isValid())
		//			throw new IllegalStateException("Constructor call is not valid");
		Identifier typeId = callType.getTypeIdentifier();
		indent(writer, indentation);
		writer.print(typeId.getName());
		writer.print(' ');
		writer.print(getIdentifier().getName());
		writer.print(" = new ");
		writer.print(typeId.getName());
		writer.print('(');
		for (int i = 0; i < getParamCount(); i++) {
			writer.print(i > 0 ? ", " : "");
			writer.print(
			getParamValues()[i] == null
			? "?"
			: getParamValues()[i].getIdentifier().getName());
		}
		if (body != null && !body.equals("")) {
			writer.println("){");
			StringTokenizer str_tok =
			new StringTokenizer(body, System.getProperty("line.separator"));
			while (str_tok.hasMoreTokens()) {
				indent(writer, indentation + 1);
				writer.println(str_tok.nextToken());
				
			}
			indent(writer, indentation);
			writer.print("};");
		} else
			writer.print(");");
	}
	
	public String getBody() {
		return body;
	}
	
	/** Number of parameters of the call.
	 * @return -1 if parameter part of signature is not set	 */
	public final int getParamCount() {
		return parameterTypes == null ? -1 : parameterTypes.length;
	}
	
	/** Gets the <code>Type</code> of the i-th call parameter */
	public final Type getParamType(int paramId) {
		return parameterTypes[paramId];
	}
	
	public ElementCall getParamValue(int param) {
		return getParamValues()[param];
	}
	
	ElementCall[] getParamValues() {
		//lazy init
		if (parameterValues == null){
			if (getParamCount() >= 0)
				//try to create values array
				parameterValues = new ElementCall[parameterTypes.length];
		}
		return parameterValues;
	}
	
	public final boolean isParameterSet(int i) {
		if (getParamValues().length == 0)
			return true;
		return getParamValues()[i] != null;
	}
	
	public boolean isValid() {
		boolean valid = true;
		if (getParamCount() < 0){
			return false;
		}
		for (int i = 0; i < getParamCount(); i++)
			valid &= isParameterSet(i);
		return valid && callType != null && getIdentifier() != null;
	}
	
	public void setBody(String body) {
		this.body = body;
	}
	
	/**
	 * Sets the i-th parameter of this call
	 * @param id number of parameter to be set
	 * @param param <code>MethodParameter</code> to be set
	 * @throws CodeException if the type of the <code>param</code> argument is not assignable to the
	 * call's i-th parameter type
	 */
	public final void setParameter(int id, ElementCall param)
	throws CodeException {
		if (param != null
		&& !parameterTypes[id].isAssignableFrom(param.getCallType())) {
			throw new CodeException(
			"Unassignable parameter type "
			+ param.getCallType()
			+ " for "
			+ parameterTypes[id]);
		}
		ElementCall old = getParamValues()[id];
		Graph g = getCallVertex().getGraph();
		//trenne old und this im graphen
		if (old != null) {
			Edge[] edges = g.getEdgesFromVertexToVertex(old.caller, caller);
			if (param instanceof MethodCall)
				((MethodCall) param).setUseReturn(false);
			for (int i = 0; i < edges.length; i++) {
				g.removeEdge(edges[i]);
			}
		}
		if (param != null) {
			g.addEdge(new Edge(param.caller, caller));
			if (param instanceof MethodCall)
				((MethodCall) param).setUseReturn(true);
		}
		getParamValues()[id] = param;
		firePropertyChange(PARAMETERS_PROPERTY, old, param);
		
	}
	
	public final void setParamTypes(Type[] params) {
		if (parameterValues != null) {
			//unset parameters
			for (int i = 0; i < parameterValues.length; i++) {
				try {
					setParameter(i, null);
				} catch (CodeException ce) {
					//this should not happen
					ce.printStackTrace();
				}
			}
		}
		parameterTypes = params;
		//signature changed, invalidate values
		parameterValues = null;
	}
}
