/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.code;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Set;

import xxl.applications.gui.graph.ClassGraph;

/**
 * @author maxim
 *
 */
public abstract class ElementCall extends Element {
	
	static final long serialVersionUID = 1028588730522019897L;

	/** constructor call type */
	public static final int CONSTRUCTOR_CALL = 1;

	/** method call type */
	public static final int METHOD_CALL = 2;

	/** primitive call type */
	public static final int PRIMITIVE_CALL = 3;

	/** The <code>CallVertex</code> object this call is represented by in <code>ClassGraph</code> */
	protected CallVertex caller;

	/** Type of object generated by this call, e.g. the type of declaring class by constructor calls
	 *  or the type of the return value by method calls
	 *  (<code>null</code> is <code>void</code> return type).
	 */
	protected Type callType;

	/** Name for an object used in this call.
	 * This is the name of the new class instance, if the call is constructor call or
	 * the name of a class or a instance of invoked method by method calls.
	 */
	protected Identifier identifier;

	/** type of the call instance*/
	private int type;

	public ElementCall(CallVertex caller) {
		super();
		if (caller == null)
			throw new IllegalArgumentException();
		this.caller = caller;
		caller.setCall(this);
	}

	/** Getter for <code>callType</code> property of the class/object this call is generated for */
	public Type getCallType() {
		return callType;
	}

	/** Returns the Call Vertex representing this call in the graph */
	public CallVertex getCallVertex() {
		return caller;
	}

	/**
	 * Getter for <code>identifier</code> property of the by this call generated object
	 * @return <code>Identifier</code> of the generated object
	 */
	public final Identifier getIdentifier() {
		//TODO name by method calls generated objects somewhere else...
		if (identifier == null) {
			ClassGraph g = (ClassGraph) caller.getGraph();
			identifier = Identifier.create(g.getNameFor(callType));
		}
		return identifier;
	}

	/**
	 * Add class of the call to import set
	 * @param imports <code>Set</code> add an import to
	 */
	public void getImport(Set imports) {
		//TODO also imports types returned objects, even if they are not used (MethodCall)
		Identifier id = getImportIdentifier(callType);
		if (id == null || id.getQualifier().equals("java.lang")) {
			return;
		}
		imports.add(new Import(id, false));
	}

	private static Identifier getImportIdentifier(Type type) {
		if (type.isPrimitive())
			return null;
		if (type.isArray())
			return getImportIdentifier(type.getElementType());
		return type.getTypeIdentifier();
	}

	public int getType() {
		return type;
	}

	/**	Determines if this call is valid */
	public abstract boolean isValid();

	public static String getTypeIdentifier(Type type) {	
		if (type.isPrimitive())
			return type.toString();
		String str = type.getFullString();
		int i = str.lastIndexOf('.');
		return i < 0 ? str : str.substring(i + 1, str.length());
	}

	public void setCallType(Type type) {
		if (type != null) {
			Type old = callType;
			callType = type;
			firePropertyChange(TYPE_PROPERTY, old, type);
		}
	}

	protected static final String getTypeString(int type) {
		if (type == CONSTRUCTOR_CALL)
			return "Constructor";
		if (type == METHOD_CALL)
			return "Method";
		if (type == PRIMITIVE_CALL)
			return "Primitive";
		return "Unknown";
	}

	public final void setIdentifier(String id) {
		Identifier old = this.identifier;
		identifier = Identifier.create(id);
		firePropertyChange(IDENTIFIER_PROPERTY, old, identifier);
	}

	protected void setType(int type) {
		this.type = type;
		caller.setName(getTypeString(type));
	}

	//TODO crappy toString()
	public String toString() {
		StringWriter stw = new StringWriter();
		PrintWriter writer = new PrintWriter(stw);
		generateCode(writer);
		return stw.getBuffer().toString();
	}
}
