/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.code;

import java.awt.Point;
import java.util.HashSet;
import java.util.Set;

import xxl.applications.gui.graph.Vertex;

/**
 * This item represents an instantiated Class. All public members could be
 * accessed.
 * 
 * @author maxim
 */
public final class ObjectItem extends AbstractItem {
	static final long serialVersionUID = 1027320368691234634L;

	/** the super class property for the item */
	private Type superClass;
	/** the implemented interfaces property for the item */
	private Type[] interfaces;

	/** default constructor, creates an java.lang.Object item */
	public ObjectItem() {
		this("java.lang.Object");
	}

	/** creates an item from classname and puts it on (0, 0) */
	public ObjectItem(String className) {
		this(new Point(), className);
	}

	/** creates an item from classname and puts it on the specified location */
	public ObjectItem(Point location, String className) {
		super(location, className);
		type = OBJECT_ITEM;
	}

	/**
	 * Getter for the superClass property
	 * 
	 * @return type of the super class
	 */
	public Type getSuperClass() {
		if (superClass == null) {
			try {
				Class clazz = getEnclosedType().toClass().getSuperclass();
				if(clazz == null) clazz = java.lang.Object.class;
				superClass = Type.createFromClass(clazz);
				//superClass = Type.createFromClass(getEnclosedType().toClass().getSuperclass());
			}
			catch (ClassNotFoundException e) {
			}
		}
		return superClass;
	}

	/**
	 * Setter for the interfaces property
	 * 
	 * @return an array of <code>Type</code> type, containig interfaces
	 *         implemented by the item's class
	 */
	public Object[] getInterfaces() {
		if (interfaces == null) {
			try {
				Set set = new HashSet();
				getInterfacesOf(getEnclosedType().toClass(), set);
				Class[] inter = (Class[]) set.toArray(new Class[0]);
				//Class[] inter = getEnclosedType().toClass().getInterfaces();				
				interfaces = new Type[inter.length];
				for (int i = 0; i < inter.length; i++) {
					interfaces[i] = Type.createFromClass(inter[i]);
					System.out.println("ObjectItem.getInterfaces() " + inter[i].getName());
				}
			}
			catch (ClassNotFoundException cnfe) {
			}
		}
		return interfaces;
	}

	private static void getInterfacesOf(Class clazz, Set set) {
		Class[] iter = clazz.getInterfaces();
		for (int i = 0; i < iter.length; i++) {
			set.add(iter[i]);
			System.out.println("ObjectItem.getInterfacesOf() " + iter[i]);
			getInterfacesOf(iter[i], set);
		}

	}

	/**
	 * Fake method to enable property editor and show the list of implemented
	 * interfaces
	 */
	public void setInterfaces(Object[] interfaces) {
	}

	/**
	 * Sets the Items constructor.
	 * 
	 * @param call
	 *            ConstructorCall to set, may be <code>null</code>
	 */
	public void setConstructor(ConstructorCall call) {
		CallVertex callVertex = call.getCallVertex();
		ConstructorCall old = getConstructor();
		if (old != null) {
			Type[] types = new Type[call.getParamCount()];
			for (int i = 0; i < call.getParamCount(); i++) {
				types[i] = call.getParamType(i);
			}
			old.setParamTypes(types);
			return;
		}
		addChildVertex(0, callVertex);
	}

	/** Getter for the items constructor */
	public ConstructorCall getConstructor() {
		if (children.size() < 1)
			return null;
		CallVertex callVertex = (CallVertex) children.get(0);
		return (ConstructorCall)
			(callVertex.getCall() instanceof ConstructorCall ? callVertex.getCall() : null);
	}

	/**
	 * Adds a new CallVertex to the item. <code>ConstructorCall</code> s will
	 * not be added by this method, use <code>setConstructor()</code> instead.
	 */
	public void addChildVertex(Vertex child) {
		ElementCall call = ((CallVertex) child).getCall();
		if (call.getType() == ElementCall.PRIMITIVE_CALL)
			return;
		if (call.getType() == ElementCall.CONSTRUCTOR_CALL) {
			setConstructor((ConstructorCall) call);
			return;
		}
		addChildVertex(-1, child);
	}

	/**
	 * proves the item for validity. the item is valid iff it has a constructor
	 * and all contained calls (constructor and method calls) are valid.
	 */
	public boolean isValid() {
		CallVertex[] calls = (CallVertex[]) children.toArray(new CallVertex[0]);
		// empty vertex is not valid
		if (calls.length < 1) {
			return false;
		}
		// if the vertex contains only one call, it is either a constructor or a
		// primitive
		if (calls.length == 1) {
			ElementCall call = calls[0].getCall();
			return (call instanceof ConstructorCall && call.isValid());
		}
		ElementCall call = calls[0].getCall();
		for (int i = 1; i < calls.length; i++) {
			ElementCall vertexCall = calls[i].getCall();
			if (!vertexCall.isValid()) {
				return false;
			}
		}
		if (call.getType() == ElementCall.CONSTRUCTOR_CALL) {
			return (call != null && call.isValid());
		}
		return false;
	}
}
