/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.graph;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.event.EventListenerList;

import xxl.applications.gui.graph.event.EdgeEvent;
import xxl.applications.gui.graph.event.EdgeListener;
import xxl.applications.gui.graph.event.VertexEvent;
import xxl.applications.gui.graph.event.VertexListener;

/**
 * The <code>Graph</code> class is a basic class for all Graphs in the
 * application. <code>ClassGraph</code> is only one known sublclass of it for
 * now. <code>Graph</code> implements basic functionality for manipulation
 * Graph data structure.
 * <p>
 * The graph is implemented as follows: there is a set containig all vertices of
 * graph and maps are used to attach lists of incoming and outgoing edges to
 * vertices. This results in some sort of adjacency list: every element of
 * <code>vertices</code> set has corresponding linked list mapped to it in
 * <code>edgesFromVertex</code> and <code>edgesToVertex</code> maps.
 * </p>
 * 
 * @author maxim
 *  
 */
public class Graph extends GraphElement {

	/** Name of the graph */
	private String name;
	/** Counter of unique ids for graph vertices */
	private int ids;
	/** java.util.Set object containig edges of the graph */
	protected transient Set edges;
	/** Map used to attach list of edges starting in the key vertex */
	protected transient Map edgesFromVertex;
	/** Map used to attach list of edges ending in the key vertex */
	protected transient Map edgesToVertex;
	/**
	 * List of listeners for <code>EdgeEvent</code> and
	 * <code>VertexEvent</code>
	 */
	private transient EventListenerList eventListeners;
	/**
	 * PropertyChangeListener used for delegating PropertyChangeEvents from
	 * vertices to VertexListeners
	 */
	private transient PropertyChangeListener propertyHandler;
	/** java.util.Set object containig vertices of the graph */
	protected transient Set vertices;
	/** Boolean array containig transitive hull of the graph */
	private transient boolean[][] transHull;

	static final long serialVersionUID = 6101054556636903166L;

	/** Creates new empty Graph */
	public Graph() {
		init();
	}

	/**
	 * Adds a new Edge to the graph. Adding succeeds iff the <code>edge</code>
	 * is not <code>null</code> and vertices-set doesn't allready contains
	 * this edge. 
	 * 
	 * @param edge
	 *            edge to be added to the graph
	 */
	public void addEdge(Edge edge) {
		if (edge == null)
			throw new IllegalArgumentException("NULL Edge can not be added!");
		if (edges.contains(edge)) {
			return;
		}
		EdgeEvent edgeEvent = new EdgeEvent(this, edge, EdgeEvent.EDGE_ADDED);
		edges.add(edge);
		/*
		 * Integer i = new Integer(edge.getFromVertex().getID()); List list =
		 * (List) edgesFromVertex.get(i); list.add(edge); i = new
		 * Integer(edge.getToVertex().getID()); list = (List)
		 * edgesToVertex.get(i); list.add(edge);
		 */
		List list = (List) edgesFromVertex.get(edge.getFromVertex());
		list.add(edge);
		list = (List) edgesToVertex.get(edge.getToVertex());
		list.add(edge);
		fireEdgeEvent(edgeEvent);
	}
	
	public Edge[] getEdges() {
		return (Edge[]) edges.toArray(new Edge[0]);
	}

	public void addEdgeListener(EdgeListener listener) {
		eventListeners.add(EdgeListener.class, listener);
	}

	public void addVertex(Vertex vertex) {
		if (vertex == null)
			throw new NullPointerException("Null Vertex can not be added!");
		if (!vertices.contains(vertex)) {
			vertex.setID(ids++);
			vertices.add(vertex);
			vertex.setGraph(this);
			vertex.addPropertyChangeListener(getPropertyChangeListener());
			/*
			 * edgesFromVertex.put(new Integer(vertex.getID()), new
			 * LinkedList()); edgesToVertex.put(new Integer(vertex.getID()), new
			 * LinkedList());
			 */
			edgesFromVertex.put(vertex, new LinkedList());
			edgesToVertex.put(vertex, new LinkedList());
			fireVertexEvent(new VertexEvent(this, vertex, VertexEvent.VERTEX_ADDED));
		}
	}

	public void addVertexListener(VertexListener listener) {
		eventListeners.add(VertexListener.class, listener);
	}

	public void fireEdgeEvent(EdgeEvent ee) {
		Object[] listenerList = eventListeners.getListenerList();
		for (int i = listenerList.length - 2; i >= 0; i -= 2) {
			if (listenerList[i] == EdgeListener.class) {
				((EdgeListener) listenerList[i + 1]).edgeEventInvoked(ee);
			}
		}
	}

	protected void fireVertexEvent(VertexEvent ve) {
		Object[] listenerList = eventListeners.getListenerList();
		for (int i = listenerList.length - 2; i >= 0; i -= 2) {
			if (listenerList[i] == VertexListener.class) {
				((VertexListener) listenerList[i + 1]).vertexEventInvoked(ve);
			}
		}
	}

	public Edge[] getEdgesFromVertex(Vertex vertex) {
		//List list = (List) edgesFromVertex.get(new Integer(vertex.getID()));
		List list = (List) edgesFromVertex.get(vertex);
		return (Edge[]) list.toArray(new Edge[0]);
	}

	public Edge[] getEdgesToVertex(Vertex vertex) {
		//List list = (List) edgesToVertex.get(new Integer(vertex.getID()));
		List list = (List) edgesToVertex.get(vertex);
		return (Edge[]) list.toArray(new Edge[0]);
	}

	public Edge[] getEdgesFromVertexToVertex(Vertex from, Vertex to) {
		Edge[] edges = getEdgesFromVertex(from);
		List toEdges = new ArrayList();
		for (int i = 0; i < edges.length; i++) {
			if (edges[i].getToVertex().getID() == to.getID()) {
				//System.out.println("Graph.getEdgesFromVertexToVertex() " +
				// from + " " + edges[i]);
				toEdges.add(edges[i]);
			}
		}
		return (Edge[]) toEdges.toArray(new Edge[toEdges.size()]);
	}

	public Vertex[] getVertices() {
		return (Vertex[]) vertices.toArray(new Vertex[0]);
	}

	public void setName(String name) {
		String old = this.name;
		this.name = name;
		firePropertyChange(NAME_PROPERTY, old, name);
	}

	public String getName() {
		return name;
	}

	private void init() {
		ids = 0;
		vertices = new HashSet();
		edges = new HashSet();
		edgesFromVertex = new HashMap();
		edgesToVertex = new HashMap();
		eventListeners = new EventListenerList();
	}

	PropertyChangeListener getPropertyChangeListener() {
		if (propertyHandler == null)
			propertyHandler = new PropertyChangeHandler(this);
		return propertyHandler;
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		init();
		ids = in.readInt();
		Set s = (Set) in.readObject();
		Iterator iter = s.iterator();
		while (iter.hasNext())
			addVertex((Vertex) iter.next());
		s = (Set) in.readObject();
		iter = s.iterator();
		while (iter.hasNext())
			addEdge((Edge) iter.next());
	}

	private void writeObject(ObjectOutputStream out) throws IOException {
		out.writeInt(ids);
		out.writeObject(vertices);
		out.writeObject(edges);
	}

	public void removeEdge(Edge edge) {
		edges.remove(edge);
		/*
		 * List list = (List) edgesFromVertex.get(new
		 * Integer(edge.getFromVertex().getID())); list.remove(edge); list =
		 * (List) edgesToVertex.get(new Integer(edge.getToVertex().getID()));
		 * list.remove(edge);
		 */
		List list = (List) edgesFromVertex.get(edge.getFromVertex());
		list.remove(edge);
		list = (List) edgesToVertex.get(edge.getToVertex());
		list.remove(edge);
		fireEdgeEvent(new EdgeEvent(this, edge, EdgeEvent.EDGE_REMOVED));
	}

	public void removeAllEdges() {
		init();
	}

	public void removeEdgeListener(EdgeListener listener) {
		eventListeners.remove(EdgeListener.class, listener);
	}

	public void removeVertex(Vertex vertex) {
		Edge[] edges = getEdgesFromVertex(vertex);
		for (int i = 0; i < edges.length; i++)
			removeEdge(edges[i]);
		edges = getEdgesToVertex(vertex);
		for (int i = 0; i < edges.length; i++)
			removeEdge(edges[i]);
		vertices.remove(vertex);
		//TODO remove complex vertex
		if (vertex instanceof ComplexVertex) {
			Vertex[] v = ((ComplexVertex) vertex).getChildren();
			for (int i = 0; i < v.length; i++)
				removeVertex(v[i]);
			((ComplexVertex) vertex).removeAllChildren();
		}
		vertex.removePropertyChangeListener(getPropertyChangeListener());
		vertex.setGraph(null);
		fireVertexEvent(new VertexEvent(this, vertex, VertexEvent.VERTEX_REMOVED));
	}

	public void removeAllVertices() {
		vertices = new HashSet();
	}

	public void removeVertexListener(VertexListener listener) {
		eventListeners.remove(VertexListener.class, listener);
	}

	// �berpr�ft ob der Graph Zyklen hat
	public boolean hasCycle() {
		boolean[][] transHull = getTransHull();
		if (transHull != null) {
			for (int i = 0; i < transHull.length; i++) {
				if (transHull[i][i])
					return true;
			}
			return false;
		}
		else
			return false;
	}

	public String toString() {
		StringBuffer str_buf = new StringBuffer();
		str_buf.append(getClass().getName());
		str_buf.append('\n');
		Vertex[] vertices = getVertices();
		str_buf.append("VERTEXES:\n");
		for (int i = 0; i < vertices.length; i++) {
			str_buf.append(vertices[i]);
			str_buf.append('\n');
		}
		Edge[] edges = getEdges();
		str_buf.append("EDGES\n");
		for (int i = 0; i < edges.length; i++) {
			str_buf.append('\t');
			str_buf.append(edges[i]);
			str_buf.append('\n');
		}
		return str_buf.toString();
	}

	private transient int time;	
	protected Map dfs() {
		Map color = new HashMap();
		Map d_time = new HashMap();
		Map f_time = new HashMap();
		time = 0;
		Vertex[] vertices = getVertices();
		for (int i = vertices.length - 1; i >= 0; i--) {
			color.put(vertices[i], new Integer(0));
		}
		for (int i = 0; i < vertices.length; i++) {
			int clr = ((Integer) color.get(vertices[i])).intValue();
			if (clr == 0)
				dfs_visit(vertices[i], color, d_time, f_time);
		}
		return f_time;
	}

	private void dfs_visit(final Vertex v, Map color, Map d_time, Map f_time) {
		color.put(v, new Integer(1));
		time++;
		d_time.put(v, new Integer(time));
		Edge[] from = getEdgesFromVertex(v);
		for (int i = 0; i < from.length; i++) {
			Vertex to = from[i].getToVertex();
			if (((Integer) color.get(to)).intValue() == 0)
				dfs_visit(to, color, d_time, f_time);
		}
		color.put(v, new Integer(2));
		f_time.put(v, new Integer(++time));
	}

	private void setTransHull() {
		Vertex[] vertexes = getVertices();
		transHull = new boolean[vertexes.length][vertexes.length];
		Edge[] edgeArr = getEdges();
		for (int i = 0; i < vertexes.length; i++)
			for (int j = 0; j < vertexes.length; j++)
				if (getEdgesFromVertexToVertex(vertexes[i], vertexes[j]).length > 0)
					transHull[i][j] = true;
		for (int j = 0; j < vertexes.length; j++)
			for (int i = 0; i < vertexes.length; i++)
				if (transHull[i][j])
					for (int k = 0; k < vertexes.length; k++)
						if (transHull[j][k])
							transHull[i][k] = true;
	}

	boolean[][] getTransHull() {
		setTransHull();
		return transHull;
	}

	final static class PropertyChangeHandler implements PropertyChangeListener {

		private Graph graph;

		public PropertyChangeHandler(Graph graph) {
			this.graph = graph;
		}

		public void propertyChange(PropertyChangeEvent evt) {
			if (evt.getPropertyName() == GraphElement.LOCATION_PROPERTY) {
				graph.fireVertexEvent(new VertexEvent(graph, (Vertex) evt.getSource(),
						VertexEvent.VERTEX_MOVED));
			}
			else {
				graph.fireVertexEvent(new VertexEvent(graph, (Vertex) evt.getSource(),
						VertexEvent.VERTEX_CHANGED));
			}
		}
	}
}
