/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.graph;

import java.awt.Point;

/** This class defines a node of the graph and provides a basic functionality for
 *  all Vertex classes
 * @author maxim
 */
public class Vertex extends GraphElement {	
	
	static final long serialVersionUID = -5531106692274306391L;

	/** 
	 * The graph this vertex belongs to, maybe <code>null</code> for "parentless" nodes.
	 *  This field marked as transient, since the <code>Graph</code>-Property will be set
	 *  every time vertex is added to a graph.
	 */
	protected transient Graph graph;
	private int id;
	/** Location property of the vertex, is the position of the vertex on the drawing pane*/
	private Point location;
	/** The name property of the vertex, used label the vertex when drawn on the drawing pane */
	private String name;
	/** The parent vertex of this vertex, it any */
	private ComplexVertex parent;
	/** Selected property of the vertex, such ones should be drawn differently */
	private boolean selected;
	/** The visible property of the vertex */
	private boolean visible;	

	public Vertex() {
		this(null);
	}

	public Vertex(String name) {
		this(name, new Point());
	}

	public Vertex(String name, Point location) {
		setName(name);
		setLocation(location);
		setVisible(true);
		id = -1;	
	}

	public Graph getGraph() {
		return graph;
	}

	public int getID() {
		return id;
	}

	public Point getLocation() {
		return location;
	}

	public String getName() {
		if (name == null)
			setName('v' + Integer.toString(id));
		return name;
	}

	/**
	 * Getter for the parent property
	 * @return <code>ComplexVertex</code> beeing the first top-level node.
	 */
	public ComplexVertex getParent() {
		if (parent != null) {
			return parent.getParent() == null ? parent : parent.getParent();
		}
		return null;
	}

	public boolean hasPathTo(Vertex dest) {
		int start = 0;
		int end = 0;
		Vertex[] vertexes = getGraph().getVertices();
		boolean[][] tHull = getGraph().getTransHull();
		for (int i = 0; i < vertexes.length; i++) {
			if (vertexes[i].equals(this))
				start = i;
			if (vertexes[i].equals(dest))
				end = i;
		}
		return tHull[start][end];
	}

	public boolean isSelected() {
		return selected;
	}

	public boolean isVisible() {
		return visible;
	}

	protected void setGraph(Graph graph) {
		this.graph = graph;
	}

	protected void setID(int id) {
		this.id = id;
	}

	public void setLocation(Point l) {
		if (location == null)
			location = new Point(l);
		else
			location.setLocation(l.x, l.y);
		firePropertyChange(LOCATION_PROPERTY, null, l);

	}

	public void setName(String name) {
		String old = this.name;
		this.name = name;
		firePropertyChange(NAME_PROPERTY, old, name);
	}

	public void setParent(ComplexVertex parent) {
		Vertex old = this.parent;
		this.parent = parent;
		firePropertyChange(PARENT_PROPERTY, old, parent);
		setVisible(parent == null);
	}

	public void setSelected(boolean selected) {
		if (this.selected != selected) {
			boolean old = this.selected;
			this.selected = selected;
			firePropertyChange(SELECTED_PROPERTY, old, selected);
		}
	}

	public void setVisible(boolean visible) {
		if (this.visible != visible) {
			boolean old = this.visible;
			this.visible = visible;
			firePropertyChange(VISIBLE_PROPERTY, old, visible);
		}
	}

	public String toString() {
		return getName();
	}
}
