/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.graph.viewer2d;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xxl.applications.gui.graph.Edge;
import xxl.applications.gui.graph.Graph;
import xxl.applications.gui.graph.Vertex;
import xxl.applications.gui.graph.event.EdgeEvent;
import xxl.applications.gui.graph.event.GraphListener;
import xxl.applications.gui.graph.event.VertexEvent;

/**
 * @author maxim
 *
 */
public class GraphDrawer {

	private static Vertex[] EMPTY_VERTEXES = new Vertex[0];
	private static Edge[] EMPTY_EDGES = new Edge[0];
	protected static final VertexDrawer DEFAULT_DRAWER = new VertexDrawer();
	protected Graph graph;
	private boolean valid;
	protected Map vertexDrawerMap = new Hashtable();

	private GraphListener listener = new GraphListener() {
		public void edgeEventInvoked(EdgeEvent e) {
			processEdgeEvent(e);
		}
		public void vertexEventInvoked(VertexEvent e) {
			processVertexEvent(e);
		}
	};
	private Map arrowToEdgeMap = new HashMap();
	private Map edgeToArrowMap = new HashMap();
	private Rectangle bounds;

	public GraphDrawer() {

	}

	public GraphDrawer(Graph graph) {
		setGraph(graph);
	}

	public void setGraph(Graph graph) {
		this.graph = graph;
		getGraph().addEdgeListener(listener);
		getGraph().addVertexListener(listener);
		invalidate();
	}

	public void registerVertexDrawer(VertexDrawer drawer, Class vertexClass) {
		vertexDrawerMap.put(vertexClass, drawer);
	}

	public Rectangle getVertexBounds(Vertex vertex) {
		Rectangle2D bounds =
			getVertexDrawer(vertex.getClass()).getBoundsAtPoint(vertex.getLocation());
		return new Rectangle(bounds.getBounds());
	}

	public Rectangle getBounds() {
		if (bounds == null) {
			if (!valid)
				refreshArrowMap();
			Vertex[] vertexes = getGraph().getVertices();
			if (vertexes.length == 0)
				return null;
			Rectangle br = getVertexBounds(vertexes[0]);
			for (int i = 1; i < vertexes.length; i++) {
				br.add(getVertexBounds(vertexes[i]));
			}
			Iterator iter = arrowToEdgeMap.keySet().iterator();
			while (iter.hasNext()) {
				SimpleArrow arrow = (SimpleArrow) iter.next();
				br.add(arrow.getBounds());
			}
			bounds = br;
		}
		return bounds;
	}

	public void invalidate() {
		valid = false;
		bounds = null;
	}

	public Graph getGraph() {
		return this.graph;
	}

	public void drawGraph(Graphics graphics) {
		if (getGraph() == null)
			return;
		if (!valid)
			refreshArrowMap();
		Graphics2D g = (Graphics2D) graphics.create();
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g.setFont(g.getFont().deriveFont(10.0f));
		g.setColor(Color.black);
		drawEdges(g);
		Vertex[] vertexes = getGraph().getVertices();
		for (int i = 0; i < vertexes.length; i++) {
			drawVertex(g, vertexes[i]);
		}
		g.dispose();
	}

	protected void drawEdges(Graphics2D g) {
		Set arrows = arrowToEdgeMap.keySet();
		Iterator iter = arrows.iterator();
		while (iter.hasNext()) {
			((SimpleArrow) iter.next()).draw(g);

		}
	}

	protected void drawVertex(Graphics graphics, Vertex vertex) {
		VertexDrawer drawer = getVertexDrawer(vertex.getClass());
		drawer.drawVertex(graphics, vertex);
	}

	public VertexDrawer getVertexDrawer(Class vertexClass) {
		VertexDrawer drawer = (VertexDrawer) vertexDrawerMap.get(vertexClass);
		if (drawer == null)
			drawer = DEFAULT_DRAWER;
		return drawer;

	}

	protected SimpleArrow arrowForEdge(Edge edge) {
		return (SimpleArrow) edgeToArrowMap.get(edge);
	}

	public Vertex vertexAtPoint(Point point) {
		Vertex[] vertexes = getGraph().getVertices();
		for (int i = vertexes.length - 1; i >= 0; i--) {
			VertexDrawer drawer = getVertexDrawer(vertexes[i].getClass());
			//we found the vertex bounding the point
			if (drawer.getBoundsAtPoint(vertexes[i].getLocation()).contains(point)) {
				//TODO quickfix for invisible vertexes
				if (!vertexes[i].isVisible()) {
					continue;
				}
				return vertexes[i];
			}
		}
		return null;
	}

	/**
	 * Get an array of vertexes enclosed by <code>bounds</code> Rectangle. The method 
	 * never returns <code>null</code>, if no vertexes are enclosed by <code>bounds</code>
	 * an array of length 0 will be returned.
	 * @param bounds enclosing Rectangle
	 * @return array of enclosed vertexes
	 */

	public Vertex[] getEclosedVertexes(Rectangle bounds) {
		if (bounds == null)
			return EMPTY_VERTEXES;
		Vertex[] vertexes = getGraph().getVertices();
		List vertexList = new ArrayList();
		for (int i = vertexes.length - 1; i >= 0; i--) {
			VertexDrawer drawer = getVertexDrawer(vertexes[i].getClass());
			if (vertexes[i].isVisible()
				&& bounds.contains(drawer.getBoundsAtPoint(vertexes[i].getLocation())))
				vertexList.add(vertexes[i]);
		}
		return (Vertex[]) vertexList.toArray(new Vertex[0]);

	}

	public Edge edgeAtPoint(Point point) {
		if (!valid)
			refreshArrowMap();
		Set arrows = arrowToEdgeMap.keySet();
		Iterator iter = arrows.iterator();
		while (iter.hasNext()) {
			SimpleArrow arrow = (SimpleArrow) iter.next();
			if (arrow.isOnArrow(point, 2))
				return (Edge) arrowToEdgeMap.get(arrow);
		}
		return null;

	}
	private static double angle(Vertex v1, Vertex v2) {
		Point p1 = v1.getLocation();
		Point p2 = v2.getLocation();
		double x = (double) (p2.x - p1.x);
		double y = (double) (p2.y - p1.y);
		return Math.atan2(y, x);
	}

	public Point pointOnVertex(Vertex vertex, double angle) {
		Point point = new Point(vertex.getLocation());
		int x, y;
		while (angle >= 2 * Math.PI)
			angle -= 2 * Math.PI;
		if (angle < 0)
			angle += Math.PI * 2;
		//Ost
		if (angle <= Math.PI / 4 || angle >= 7 * Math.PI / 4) {
			x = getVertexDrawer(vertex.getClass()).getWidth();
			y = 0;
		}
		//	Nord
		else if (angle > Math.PI / 4 && angle <= 3 * Math.PI / 4) {
			x = 0;
			y = getVertexDrawer(vertex.getClass()).getHeight();
		}
		// West
		else if (angle > 3 * Math.PI / 4 && angle <= 5 * Math.PI / 4) {
			x = -getVertexDrawer(vertex.getClass()).getWidth();
			y = 0;
		}
		//Sued
		else {
			x = 0;
			y = -getVertexDrawer(vertex.getClass()).getHeight();
		}
		point.translate((int) x, (int) y);
		return point;
	}

	private void refreshArrowMap() {
		Vertex[] vertexes = getGraph().getVertices();			
		arrowToEdgeMap.clear();
		edgeToArrowMap.clear();		
		for (int i = 0; i < vertexes.length; i++) {
			for (int j = i + 1; j < vertexes.length; j++) {		
				Edge[] i2j = getGraph().getEdgesFromVertexToVertex(vertexes[i], vertexes[j]);
				Edge[] j2i = getGraph().getEdgesFromVertexToVertex(vertexes[j], vertexes[i]);
				if (i2j.length + j2i.length == 0) {		
					continue;
				}		
				//TODO testing items connections
				//Vertex vi = vertexes[i];
				//Vertex vj = vertexes[j];
				Vertex vi = vertexes[i].getParent() == null ? vertexes[i] : vertexes[i].getParent();
				Vertex vj = vertexes[j].getParent() == null ? vertexes[j] : vertexes[j].getParent();
				//TODO keine pfeile zu sich selbst...
				if (vi.getID() == vj.getID()) {		
					continue;
				}
				else {
					double angle = angle(vi, vj);
					Point pi = pointOnVertex(vi, angle);
					Point pj = pointOnVertex(vj, angle + Math.PI);
					//TODO mehrfache verbindungen zwichen 2 knoten erkennen und nur einfach zeichnen
					for (int n = 0; n < i2j.length; n++) {
						if (edgeToArrowMap.containsKey(i2j[n]))
							continue;
						SimpleArrow arrow = new SimpleArrow(pi, pj);					
						arrowToEdgeMap.put(arrow, i2j[n]);
						edgeToArrowMap.put(i2j[n], arrow);
					}
					pi = pointOnVertex(vi, angle);
					pj = pointOnVertex(vj, angle + Math.PI);
					for (int n = 0; n < j2i.length; n++) {
						if (edgeToArrowMap.containsKey(j2i[n]))
							continue;
						SimpleArrow arrow = new SimpleArrow(pj, pi);					
						arrowToEdgeMap.put(arrow, j2i[n]);
						edgeToArrowMap.put(j2i[n], arrow);
					}
				}
			}
		}
		valid = true;
	}

	protected void processEdgeEvent(EdgeEvent e) {
		invalidate();
	}

	protected void processVertexEvent(VertexEvent e) {
		if (e.getType() == VertexEvent.VERTEX_CHANGED)
			return;
		invalidate();
	}
}
