/* -- JFLAP 4.0 --
 *
 * Copyright information:
 *
 * Susan H. Rodger, Thomas Finley
 * Computer Science Department
 * Duke University
 * April 24, 2003
 * Supported by National Science Foundation DUE-9752583.
 *
 * Copyright (c) 2003
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the author.  The name of the author may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

package xxl.applications.gui.graph.viewer2d;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;

import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import xxl.applications.gui.graph.ComplexVertex;
import xxl.applications.gui.graph.Graph;
import xxl.applications.gui.graph.Vertex;
import xxl.applications.gui.graph.event.EdgeEvent;
import xxl.applications.gui.graph.event.EdgeListener;
import xxl.applications.gui.graph.event.VertexEvent;
import xxl.applications.gui.graph.event.VertexListener;

/**
 * A simple view that draws a graph.
 * 
 * Based on AutomatonPane by
 * @author Thomas Finley
 */

public class GraphPane extends JPanel {

	/** Whether or not we should adapt the view by transforming it. */
	private boolean adapt;
	/** The drawer. */
	protected GraphDrawer drawer;
	/** The transform. */
	protected AffineTransform transform = new AffineTransform();
	/** Whether the transform needs to be reformed. */
	private transient boolean transformNeedsReform = true;
	/** Custom scale factor */
	private double scale = 1.0d;
	/** Crappy grid */
	private Grid grid;	

	/**
	 * Instantiates an GraphPane with the default graph drawer
	 * as the drawer for the passed in graph.
	 * @param graph the graph to draw
	 */
	public GraphPane(Graph graph) {
		this(new GraphDrawer(graph));
	}
	/**
	 * Instantiates an GraphPane.
	 * @param drawer the graph drawer
	 */
	public GraphPane(GraphDrawer drawer) {
		this(drawer, true);
		this.setLayout(null);
	}

	/**
	 * Instantiates an GraphPane.
	 * @param drawer the graph drawer
	 * @param adapt whether or not to adapt the size of the view
	 */
	public GraphPane(GraphDrawer drawer, boolean adapt) {
		super();
		setDrawer(drawer);
		this.adapt = adapt;
		Listener listener = new Listener();
		drawer.getGraph().addEdgeListener(listener);
		drawer.getGraph().addVertexListener(listener);
		drawer.registerVertexDrawer(new ComplexVertexDrawer(), ComplexVertex.class);
		addComponentListener(listener);
	}

	public void setGrid(Grid grid) {
		if (this.grid != grid) {
			this.grid = grid;
			repaint();
		}
	}

	public Grid getGrid() {
		return this.grid;
	}

	public void setAdapt(boolean adapt) {
		if (this.adapt != adapt) {
			this.adapt = adapt;
			transformNeedsReform = true;
			repaint();
		}
	}

	public boolean isAdapting() {
		return adapt;
	}

	/**
	 * This will edit the underlying graph so that it fits within
	 * the bounds of the view (at 1 to 1 scale).
	 * @param padding the amount of padding
	 */
	public void fitToBounds(int padding) {
		Rectangle viewBounds = getVisibleRect();
		Rectangle graphBounds = new Rectangle(drawer.getBounds());
		graphBounds.grow(padding, padding);
		viewBounds.x = viewBounds.y = 0; // Just pretend...
		Vertex[] vertexes = drawer.getGraph().getVertices();

		// Translate and scale each point.
		for (int i = 0; i < vertexes.length; i++) {
			Point p = vertexes[i].getLocation();
			p.setLocation(
				(p.getX() - graphBounds.getX()) * (viewBounds.getWidth() / graphBounds.getWidth()),
				(p.getY() - graphBounds.getY()) * (viewBounds.getHeight() / graphBounds.getHeight()));
			vertexes[i].setLocation(p);
		}
	}

	/**
	 * Returns what the point (0,0) in this component is in graph
	 * space.
	 */
	public Point getGraphOrigin() {
		return new Point(- (int) transform.getTranslateX(), - (int) transform.getTranslateY());
	}

	/**
	 * Returns the graph drawer.
	 * @return the graph drawer
	 */
	public GraphDrawer getDrawer() {
		return drawer;
	}

	/**
	 * Returns a rectangle that roughly bounds the graph, or the
	 * section of the graph that should be shown if not all should
	 * be shown.
	 * @return a rectangle that bounds the graph
	 */
	protected Rectangle getGraphBounds() {
		Rectangle rect = drawer.getBounds();
		if (rect == null)
			return new Rectangle(getSize());
		return rect;
	}

	/**
	 * Returns the location for a tool tip to display.
	 * @return the location for a tool tip
	 */
	public Point getToolTipLocation(MouseEvent event) {
		try {
			return transformFromGraphToView(drawer.vertexAtPoint(event.getPoint()).getLocation());
		}
		catch (NullPointerException e) {
			return null;
		}
	}

	/**
	 * Return the text for a tool tip.
	 * @return text for a tool tip
	 */
	public String getToolTipText(MouseEvent event) {

		Vertex s = drawer.vertexAtPoint(event.getPoint());
		if (s == null)
			return null;
		return s.getName();

	}

	protected void drawGrid(Graphics g) {
		if (getGrid() != null) {
			getGrid().draw(g, SwingUtilities.getRoot(this).getBounds());
		}
	}

	/**
	 * Draws itself, and the graph.
	 * @param g the graphics object to draw upon
	 */
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		if (transformNeedsReform)
			reformTransform(new Rectangle(getSize()));
		g.setColor(java.awt.Color.white);
		g.fillRect(0, 0, getSize().width, getSize().height);
		Graphics2D g2 = (Graphics2D) g;
		g2.transform(transform);
		//drawGrid(g2);
		drawer.drawGraph(g);
		//		g2.setColor(Color.RED);
		//		g2.draw(getGraphBounds());
	}

	/**
	 * Prints itself, with the graph drawn so that it fills the
	 * space.
	 * @param g the graphics interface for the printer device
	 */
	public void printComponent(Graphics g) {
		boolean oldAdapt = adapt;
		adapt = true;
		reformTransform(g.getClipBounds());
		Graphics2D g2 = (Graphics2D) g.create();
		g2.transform(transform);
		drawer.drawGraph(g2);
		adapt = oldAdapt;
		reformTransform(new Rectangle(getSize()));
	}

	/**
	 * Processes mouse events with a transformed mouse event.
	 * @param event the mouse event
	 */
	public void processMouseEvent(MouseEvent event) {
		transformMouseEvent(event);
		super.processMouseEvent(event);
	}

	/**
	 * Processes mouse motion events with a transformed mouse event.
	 * @param event the mouse event
	 */
	public void processMouseMotionEvent(MouseEvent event) {
		transformMouseEvent(event);
		super.processMouseMotionEvent(event);
	}

	/**
	 * Reforms the transform.
	 * @param bounds the current bounds of the drawing area, which may
	 * be used if this component adapts it's size
	 */
	private void reformTransform(Rectangle viewBounds) {
		transformNeedsReform = false;
		if (!adapt) {
			adjustVisibleSpace();
		}
		else {
			fitIntoSpace(viewBounds);
		}

	}

	private void adjustVisibleSpace() {
		Rectangle bounds = new Rectangle(getGraphBounds());
		bounds.add(new Rectangle(0, 0, 0, 0));
		double x1 = bounds.getX() * getScale();
		double y1 = bounds.getY() * getScale();
		double width = bounds.getWidth() * getScale();
		double height = bounds.getHeight() * getScale();

		bounds.setRect(x1, y1, width, height);
		// Much of this is to make sure that this component
		// displays properly in a scroll box.  (It must work
		// either in a scroll box, or not!)
		Rectangle visible = getVisibleRect();
		// What is the point of the upper left corner of the
		// viewport in graph drawer space?
		Point viewportUpperLeft =
			new Point(
				(int) (visible.x - transform.getTranslateX()),
				(int) (visible.y - transform.getTranslateY()));
		Point componentUpperLeft = new Point(bounds.x, bounds.y);
		// Make the transform draw the graph in the
		// appropriate location.
		transform = new AffineTransform();
		transform.translate(-componentUpperLeft.x, -componentUpperLeft.y);
		transform.scale(getScale(), getScale());
		// 		Set the size of the thing appropriately.
		//		int x = Math.max(bounds.width + bounds.x, viewportUpperLeft.x + visible.width) - componentUpperLeft.x;
		//		int y =
		//		Math.max(bounds.height + bounds.y, viewportUpperLeft.y + visible.height) - componentUpperLeft.y;
		Dimension newSize = new Dimension(bounds.width, bounds.height);
		if (newSize.equals(getPreferredSize()))
			return;
		setPreferredSize(newSize);
		// Scroll to make viewportUpperLeft an "isopoint".
		Rectangle scrollRect =
			new Rectangle(
				viewportUpperLeft.x - componentUpperLeft.x,
				viewportUpperLeft.y - componentUpperLeft.y,
				visible.width,
				visible.height);
		revalidate();
		scrollRectToVisible(visible);
	}

	private void fitIntoSpace(Rectangle viewBounds) {
		Rectangle bounds = new Rectangle(getGraphBounds());
		//We do make the graph fit in the visible space!
		transform = new AffineTransform();
		bounds.grow(20, 20);
		Rectangle ourBounds = viewBounds;
		double aRatio = bounds.getWidth() / bounds.getHeight();
		double vRatio = ourBounds.getWidth() / ourBounds.getHeight();
		if (aRatio > vRatio) {
			// The graph is wider than the view.
			double targetHeight = bounds.getWidth() / vRatio;
			targetHeight -= bounds.getHeight();
			// Must extend by targetHeight.
			bounds.setRect(
				bounds.getX(),
				bounds.getY() - targetHeight / 2.0,
				bounds.getWidth(),
				bounds.getHeight() + targetHeight);
		}
		else {
			// The graph is taller than the view.
			double targetWidth = bounds.getHeight() * vRatio;
			targetWidth -= bounds.getWidth();
			// Extend by targetWidth.
			bounds.setRect(
				bounds.getX() - targetWidth / 2.0,
				bounds.getY(),
				bounds.getWidth() + targetWidth,
				bounds.getHeight());
		}
		double scale = ourBounds.getWidth() / bounds.getWidth();
		transform.scale(scale, scale);
		transform.translate(ourBounds.getX() - bounds.getX(), ourBounds.getY() - bounds.getY());
	}

	public void setScale(double scale) {
		if (scale > 0) {
			this.scale = scale;
			transformNeedsReform = true;
			repaint();
		}
	}

	public double getScale() {
		return this.scale;
	}

	/**
	 * Transforms a point from graph space to view space.
	 * @param point the point, which will be modified
	 */
	public Point transformFromGraphToView(Point point) {
		return (Point) transform.transform(point, new Point());
	}

	/**
	 * Transforms a point from view space to graph space.
	 * @param point the point, which will be modified
	 */
	public Point transformFromViewToGraph(Point point) {
		try {
			return (Point) transform.inverseTransform(point, new Point());
		}
		catch (NoninvertibleTransformException e) {
			// Nothing to do.
		}
		return new Point(point);
	}

	/**
	 * Transforms a mouse event to have its mouse click location
	 * reflect the graph space, not the view space.
	 * @param event the mouse event to transform
	 */
	public void transformMouseEvent(MouseEvent event) {
		if (transformNeedsReform)
			reformTransform(new Rectangle(getSize()));
		Point point = new Point(), ePoint = event.getPoint();
		try {
			//transform.transform(ePoint, point);
			transform.inverseTransform(ePoint, point);
			event.translatePoint(point.x - ePoint.x, point.y - ePoint.y);
		}
		catch (NoninvertibleTransformException e) {
			// Well, what CAN we do?			
		}
	}

	//	public Dimension getPreferredScrollableViewportSize() {
	//		return getPreferredSize();
	//	}
	//
	//	public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction) {
	//		return orientation == SwingConstants.VERTICAL ? visibleRect.height : visibleRect.width;
	//	}
	//
	//	public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {
	//		return 5;
	//	}
	//
	//	public boolean getScrollableTracksViewportWidth() {
	//		return adapt;
	//	}
	//
	//	public boolean getScrollableTracksViewportHeight() {
	//		return adapt;
	//	}
	
	/**
	 * If an graph changes, repaint this object.
	 */
	private class Listener extends ComponentAdapter implements VertexListener, EdgeListener {

		public void componentResized(ComponentEvent e) {
			if (adapt) {
				transformNeedsReform = true;
				repaint();
			}
		}

		public void edgeEventInvoked(EdgeEvent e) {
			transformNeedsReform = true;
			repaint();
		}

		public void vertexEventInvoked(VertexEvent e) {			
			transformNeedsReform = true;
			repaint();

		}
	}

	public void setDrawer(GraphDrawer drawer) {
		if (drawer == null)
			throw new IllegalArgumentException("Drawer is null");
		if (this.drawer == null)
			this.drawer = drawer;
		else {
			Graph graph = this.drawer.graph;
			this.drawer = drawer;
			drawer.setGraph(graph);
		}
	}
}
