/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.util;

import java.io.File;
import java.io.Serializable;
import java.util.Locale;
import java.util.StringTokenizer;

/** 
 * @author maxim
 */

public class ClassPath implements Serializable {

	private Object[] items;
	private String classpath;

	public ClassPath(String[] items) {
		this.items = items;
	}

	public ClassPath(File[] items) {
		this.items = items;
	}

	public ClassPath(String path) {
		this.classpath = path;
		this.items = new Object[0];
	}

	public static ClassPath createLibraryPath() {
		return new ClassPath("");
	}

	private static ClassPath createPath(String property) {
		String cp = System.getProperty(property);
		if (cp == null || cp.length() == 0)
			return new ClassPath(""); // NOI18N
		StringBuffer buf = new StringBuffer(cp.length());
		StringTokenizer tok = new StringTokenizer(cp, File.pathSeparator);
		boolean appended = false;
		while (tok.hasMoreTokens()) {
			String piece = tok.nextToken();
			if (appended) {
				buf.append(File.pathSeparatorChar);
			}
			else {
				appended = true;
			}
			buf.append(piece);
		}
		return new ClassPath(buf.toString());

	}

	public static ClassPath createClassPath() {
		return createPath("java.class.path");
	}

	public static ClassPath createBootClassPath() {
		// boot
		String boot = System.getProperty("sun.boot.class.path"); // NOI18N
		StringBuffer sb = (boot != null ? new StringBuffer(boot) : new StringBuffer());

		// std extensions
		String extensions = System.getProperty("java.ext.dirs"); // NOI18N
		if (extensions != null) {
			for (StringTokenizer st = new StringTokenizer(extensions, File.pathSeparator);
				st.hasMoreTokens();
				) {
				File dir = new File(st.nextToken());
				File[] entries = dir.listFiles();
				if (entries != null) {
					for (int i = 0; i < entries.length; i++) {
						String name = entries[i].getName().toLowerCase(Locale.US);
						if (name.endsWith(".zip") || name.endsWith(".jar")) { // NOI18N
							if (sb.length() > 0) {
								sb.append(File.pathSeparatorChar);
							}
							sb.append(entries[i].getPath());
						}
					}
				}
			}
		}

		return new ClassPath(sb.toString());
	}

	/** Create class path representation. The implementation <i>will return the string quoted 
	 *  (using doublequotes)</i>, if it contains a space character.
	 * @return string representing the classpath items separated by File.separatorChar, possibly quoted.
	 */
	public String getClassPath() {
		if (classpath != null)
			return classpath;
		synchronized (this) {
			if (classpath != null)
				return classpath;

			if (items.length == 0) {
				return classpath = ""; // NOI18N
			}
			else {
				StringBuffer sb = new StringBuffer();
				boolean haveone = false;
				for (int i = 0; i < items.length; i++) {
					Object o = items[i];
					if (o == null || (!(o instanceof String) && !(o instanceof File))) {
						// we accept only strings/files
						continue;
					}

					if (haveone) {
						sb.append(File.pathSeparatorChar);
					}
					else {
						haveone = true;
					}
					sb.append(o.toString());
					items[i] = null;
				}
				String clsPth;
				if ((clsPth = sb.toString()).indexOf(' ') >= 0) {
					sb.insert(0, '"');
					sb.append('"');
					classpath = sb.toString();
				}
				else {
					classpath = clsPth;
				}
				return classpath;
			}
		}
	}

	/* equals */
	public boolean equals(Object o) {
		if (!(o instanceof ClassPath))
			return false;
		ClassPath him = (ClassPath) o;
		return getClassPath().equals(him.getClassPath());
	}
}
