/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.util;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

public class DecoratorGenerator {

	protected static StringBuffer generateSourceCode(Class sourceClass, final String destinationPackage) {
		StringBuffer packageDeclaration = generatePackageDeclaration(destinationPackage);
		TreeSet importedClasses = new TreeSet(new Comparator() {
			public int compare(Object o1, Object o2) {
				return ((Class) o1).getPackage().getName().compareTo(((Class) o2).getPackage().getName())
					!= 0
					? ((Class) o1).getPackage().getName().compareTo(((Class) o2).getPackage().getName())
					: ((Class) o1).getName().compareTo(((Class) o2).getName());
			}

			public boolean equals(Object o) throws UnsupportedOperationException {
				throw new UnsupportedOperationException();
			}
		}) {
			public boolean add(Object o) {
				while (((Class) o).isArray())
					o = ((Class) o).getComponentType();
				if (!(((Class) o).isPrimitive() || ((Class) o).getPackage() == null)) {
					String packageName = ((Class) o).getPackage().getName();
					if (!(packageName.equals("java.lang") || packageName.equals(destinationPackage)))
						return super.add(o);
				}
				return false;
			}
		};
		StringBuffer classDeclaration = generateClassDeclaration(sourceClass, importedClasses);
		StringBuffer importDeclaration = generateImportDeclaration(importedClasses.iterator());

		return packageDeclaration.append(importDeclaration).append(classDeclaration);
	}

	protected static StringBuffer generatePackageDeclaration(String destinationPackage) {
		StringBuffer packageDeclaration = new StringBuffer();

		if (destinationPackage != null)
			packageDeclaration.append("package ").append(destinationPackage).append(";\n\n");
		return packageDeclaration;
	}

	protected static StringBuffer generateImportDeclaration(Iterator importedClasses) {
		StringBuffer importDeclaration = new StringBuffer();

		while (importedClasses.hasNext())
			importDeclaration.append("import ").append(((Class) importedClasses.next()).getName()).append(
				";\n");
		return importDeclaration.length() == 0 ? importDeclaration : importDeclaration.append('\n');
	}

	public static StringBuffer generateClassDeclaration(Class sourceClass, Set importedClasses) {
		StringBuffer classDeclaration = new StringBuffer();
		String sourceClassName = getName(sourceClass).toString();
		String sourceClassIdentifier = getIdentifier(sourceClass).toString();

		importedClasses.add(sourceClass);
		classDeclaration
			.append("public abstract class Decorator")
			.append(sourceClassName)
			.append(sourceClass.isInterface() ? " implements " : " extends ")
			.append(sourceClassName)
			.append(" {\n\n\tprotected ")
			.append(sourceClassName)
			.append(' ')
			.append(sourceClassIdentifier)
			.append(";\n\n\tpublic Decorator")
			.append(sourceClassName)
			.append(" (")
			.append(sourceClassName)
			.append(' ')
			.append(sourceClassIdentifier)
			.append(") {\n\t\tthis.")
			.append(sourceClassIdentifier)
			.append(" = ")
			.append(sourceClassIdentifier)
			.append(";\n\t}\n");

		Method[] methods = sourceClass.getDeclaredMethods();
		for (int i = 0; i < methods.length; i++)
			if (!(Modifier.isFinal(methods[i].getModifiers())
				|| Modifier.isPrivate(methods[i].getModifiers())))
				classDeclaration.append(
					generateMethodDeclaration(methods[i], sourceClassIdentifier, importedClasses));
		return classDeclaration.append("}\n");
	}

	protected static StringBuffer generateMethodDeclaration(
		Method sourceMethod,
		String sourceClassIdentifier,
		Set importedClasses) {
		StringBuffer methodDeclaration = new StringBuffer();
		Class returnType = sourceMethod.getReturnType();
		Class[] parameterTypes = sourceMethod.getParameterTypes();
		Class[] exceptionTypes = sourceMethod.getExceptionTypes();
		StringBuffer[] parameterIdentifiers = getUniqueIdentifiers(parameterTypes);

		methodDeclaration.append("\n\t");
		if (Modifier.isProtected(sourceMethod.getModifiers()))
			methodDeclaration.append("protected ");
		if (Modifier.isPublic(sourceMethod.getModifiers()))
			methodDeclaration.append("public ");
		importedClasses.add(returnType);
		methodDeclaration.append(getName(returnType)).append(' ').append(sourceMethod.getName()).append(
			' ').append(
			generateParameterList(parameterTypes, parameterIdentifiers, importedClasses));
		if (exceptionTypes.length > 0)
			methodDeclaration.append(" throws ");
		for (int i = 0; i < exceptionTypes.length; i++) {
			if (i > 0)
				methodDeclaration.append(", ");
			importedClasses.add(exceptionTypes[i]);
			methodDeclaration.append(getName(exceptionTypes[i]));
		}
		methodDeclaration.append(" {\n\t\t");
		if (!returnType.equals(void.class))
			methodDeclaration.append("return ");
		return methodDeclaration
			.append(sourceClassIdentifier)
			.append('.')
			.append(sourceMethod.getName())
			.append(generateParameterList(null, parameterIdentifiers, importedClasses))
			.append(";\n\t}\n");
	}

	protected static StringBuffer generateParameterList(
		Class[] parameterTypes,
		StringBuffer[] parameterIdentifiers,
		Set importedClasses) {
		StringBuffer parameterList = new StringBuffer();

		parameterList.append('(');
		for (int i = 0; i < parameterIdentifiers.length; i++) {
			if (i > 0)
				parameterList.append(", ");
			if (parameterTypes != null) {
				importedClasses.add(parameterTypes[i]);
				parameterList.append(getName(parameterTypes[i])).append(" ");
			}
			parameterList.append(parameterIdentifiers[i]);
		}
		return parameterList.append(')');
	}

	protected static StringBuffer getName(Class type) {
		int array = 0;
		while (type.isArray()) {
			type = type.getComponentType();
			array++;
		}
		StringBuffer name = new StringBuffer(type.getName());
		name.delete(0, name.lastIndexOf(".") + 1);
		for (int i = 0; i < array; i++)
			name.append("[]");
		return name;
	}

	protected static StringBuffer getIdentifier(Class type) {
		boolean array = type.isArray();
		while (type.isArray())
			type = type.getComponentType();
		StringBuffer identifier = new StringBuffer(type.getName());
		if (type.isPrimitive())
			identifier.append(array ? "Array" : "Value");
		else {
			identifier.delete(0, identifier.toString().lastIndexOf('.') + 1);
			identifier.setCharAt(0, Character.toLowerCase(identifier.charAt(0)));
			if (array)
				identifier.append("Array");
		}
		return identifier;
	}

	protected static StringBuffer[] getUniqueIdentifiers(Class[] types) {

		class Usage {
			protected int amount;
			protected int used;

			public Usage() {
				amount = 1;
				used = 0;
			}

			public void increase() {
				amount++;
			}

			public boolean multiple() {
				return amount > 1;
			}

			public boolean usable() {
				return used < amount;
			}

			public int use() {
				return ++used;
			}
		}

		HashMap mapping = new HashMap();
		for (int i = 0; i < types.length; i++) {
			Class type = types[i];
			while (type.isArray() && type.getComponentType().isArray())
				type = type.getComponentType();
			Usage usage = (Usage) mapping.get(type);
			if (usage == null)
				mapping.put(type, new Usage());
			else
				usage.increase();
		}

		StringBuffer[] identifiers = new StringBuffer[types.length];
		for (int i = 0; i < identifiers.length; i++) {
			identifiers[i] = getIdentifier(types[i]);
			Class type = types[i];
			while (type.isArray() && type.getComponentType().isArray())
				type = type.getComponentType();
			Usage usage = (Usage) mapping.get(type);
			if (usage.multiple() && usage.usable())
				identifiers[i].append(usage.use());
		}
		return identifiers;
	}

	public static void main(String[] args) {
		try {
			Class sourceClass = Class.forName(args[0]);
			StringBuffer decorationName = getName(sourceClass).insert(0, "Decorator").append(".java");

			File sourceCodeFile = new File(decorationName.toString());
			if (sourceCodeFile.exists()) {
				new File(decorationName.append(".bak").toString()).delete();
				sourceCodeFile.renameTo(new File(decorationName.toString()));
				sourceCodeFile =
					new File(
						decorationName
							.delete(decorationName.length() - 3, decorationName.length())
							.toString());
			}
			FileWriter sourceCodeWriter = new FileWriter(sourceCodeFile);
			sourceCodeWriter.write(
				generateSourceCode(
					sourceClass,
					args.length > 1
						? args[1].equals("none")
						? null
						: args[1] : sourceClass.getPackage() == null
						? null
						: sourceClass.getPackage().getName())
					.toString());

			sourceCodeWriter.close();
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("No class or interface is specified.");
		}
		catch (ClassNotFoundException e) {
			System.out.println("The specified class or interface cannot be found.");
		}
		catch (IOException e) {
			System.out.println("The source code of the decoration cannot be written.");
		}
	}
}
