/*
 *                 Sun Public License Notice
 * 
 * The contents of this file are subject to the Sun Public License
 * Version 1.0 (the "License"). You may not use this file except in
 * compliance with the License. A copy of the License is available at
 * http://www.sun.com/
 * 
 * The Original Code is NetBeans. The Initial Developer of the Original
 * Code is Sun Microsystems, Inc. Portions Copyright 1997-2003 Sun
 * Microsystems, Inc. All Rights Reserved.
 */

package xxl.applications.gui.util;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Vector;

import javax.swing.Icon;
import javax.swing.ImageIcon;

/** Source code origin is org.netbeans.openide.util.Utilities 
 * Otherwise uncategorized useful static methods.
*
* @author Jan Palka, Ian Formanek, Jaroslav Tulach
*/

public final class Utilities {

	/** Implementation of the active queue.
	 */
	private static final class ActiveQueue extends ReferenceQueue implements Runnable {
		private boolean deprecated;
		private boolean running;

		public ActiveQueue(boolean deprecated) {
			this.deprecated = deprecated;

			Thread t = new Thread(this, "Active Reference Queue Daemon"); // NOI18N
			t.setPriority(Thread.MIN_PRIORITY);
			t.setDaemon(true); // to not prevent exit of VM
			t.start();
		}

		public Reference poll() {
			throw new java.lang.UnsupportedOperationException();
		}

		public Reference remove() throws InterruptedException {
			throw new java.lang.InterruptedException();
		}

		public Reference remove(long timeout) throws IllegalArgumentException, InterruptedException {
			throw new java.lang.InterruptedException();
		}

		/** Called either from Thread.run or RequestProcessor.post. In first case
		 * calls scanTheQueue (only once) in the second and nexts calls cleanTheQueue
		 */
		public void run() {
			synchronized (this) {
				if (running) {
					return;
				}
				running = true;
			}

			for (;;) {
				try {
					Reference ref = super.remove(0);
					if (!(ref instanceof Runnable)) {
						System
							.err
							.println(
								"A reference not implementing runnable has been added to the Utilities.activeReferenceQueue (): "
								+ ref.getClass()
						// NOI18N
						);
						continue;
					}

					if (deprecated) {
					}

					// do the cleanup
					try {
						((Runnable) ref).run();
					}
					catch (ThreadDeath td) {
						throw td;
					}
					catch (Throwable t) {
						// Should not happen.
						// If it happens, it is a bug in client code, notify!
						t.printStackTrace();
					}
				}
				catch (InterruptedException ex) {
					ex.printStackTrace();
				}
			}
		}
	}

	/** variable holding the activeReferenceQueue */
	private static ReferenceQueue activeReferenceQueue;

	public static final Dimension BUTTCON_PREF_SIZE = new Dimension(28, 28);
	public static final Dimension BUTTON_PREF_SIZE = new Dimension(100, 26);

	/** Hashtable contains keywords. It is forbidden to use this
		keywords as a java identifier */
	private static Reference keywords;

	/** The operating system on which NetBeans runs*/
	private static int operatingSystem = -1;
	/** Operating system is IBM AIX. */
	public static final int OS_AIX = 64;
	/** Operating system is Compaq TRU64 Unix */
	public static final int OS_TRU64 = 512;
	/** @deprecated please use OS_TRU64 instead */
	@Deprecated
	public static final int OS_DEC = OS_TRU64;
	/** Operating system is HP-UX. */
	public static final int OS_HP = 32;
	/** Operating system is SGI IRIX. */
	public static final int OS_IRIX = 128;
	/** Operating system is Linux. */
	public static final int OS_LINUX = 16;
	/** Operating system is Mac. */
	public static final int OS_MAC = 2048;
	/** Operating system is OS/2. */
	public static final int OS_OS2 = 1024;
	/** Operating system is unknown. */
	public static final int OS_OTHER = 65536;
	/** Operating system is Solaris. */
	public static final int OS_SOLARIS = 8;
	/** Operating system is Sun OS. */
	public static final int OS_SUNOS = 256;
	/** A mask for Unix platforms. */
	public static final int OS_UNIX_MASK =
		OS_SOLARIS | OS_LINUX | OS_HP | OS_AIX | OS_IRIX | OS_SUNOS | OS_TRU64 | OS_MAC;
	/** Operating system is Compaq OpenVMS */
	public static final int OS_VMS = 8192;
	/**
	 *Operating system is one of the Windows variants but we don't know which
	 *one it is
	 */
	public static final int OS_WIN_OTHER = 16384;
	/** Operating system is Windows 2000. */
	public static final int OS_WIN2000 = 4096;
	/** Operating system is Windows 95. */
	public static final int OS_WIN95 = 2;
	/** Operating system is Windows 98. */
	public static final int OS_WIN98 = 4;
	/** Operating system is Windows NT. */
	public static final int OS_WINNT = 1;
	/** A mask for Windows platforms. */
	public static final int OS_WINDOWS_MASK = OS_WINNT | OS_WIN95 | OS_WIN98 | OS_WIN2000 | OS_WIN_OTHER;
	/** A height of the Mac OS X's menu */
	private static final int TYPICAL_MACOSX_MENU_HEIGHT = 24;
	/** A height of the windows's taskbar */
	public static final int TYPICAL_WINDOWS_TASKBAR_HEIGHT = 27;

	/** Useful queue for all parts of system that use <code>java.lang.ref.Reference</code>s
		* together with some <code>ReferenceQueue</code> and need to do some clean up
		* when the reference is enqued. Usually, in order to be notified about that, one 
		* needs to either create a dedicated thread that blocks on the queue and is 
		* <code>Object.notify</code>-ed, which is the right approach but consumes 
		* valuable system resources (threads) or one can peridically check the content
		* of the queue by <code>RequestProcessor.Task.schedule</code> which is 
		* completelly wrong, because it wakes up the system every (say) 15 seconds.
		* In order to provide useful support for this problem, this queue has been
		* provided.
		* <P>
		* If you have a reference that needs clean up, make it implement <link>Runnable</link>
		* inteface and register it with the <code>activeReferenceQueue</code>:
		* <PRE>
		* class MyReference extends WeakReference implements Runnable {
		*   private Object dataToCleanUp;
		*
		*   public MyReference (Object ref, Object data) {
		*     super (ref, Utilities.activeReferenceQueue ()); // here you specify the queue
		*     dataToCleanUp = data;
		*   }
		*
		*   public void run () {
		*     // clean up your data
		*   }
		* }
		* </PRE>
		* When the <code>ref</code> object is garbage collected, your run method
		* will be invoked by calling 
		* <code>((Runnable)reference).run ()</code>
		* and you can perform what ever cleanup is necessary. Be sure not to block
		* in such cleanup for a long time as this prevents other waiting references 
		* to cleanup themselves.
		* <P>
		* Please do not call any methods of the ReferenceQueue yourself. They
		* will throw exceptions.
		*
		* @since 3.11
		*/
	public static synchronized ReferenceQueue activeReferenceQueue() {
		if (activeReferenceQueue == null) {
			activeReferenceQueue = new ActiveQueue(false);
		}
		return activeReferenceQueue;
	}

	/**
	 * Returns cursor created from given icon.
	*/
	public static Cursor createCustomCursor(Image icon, String name) {
		Toolkit t = Toolkit.getDefaultToolkit();
		Dimension d = t.getBestCursorSize(16, 16);
		Image i = icon;
		if (d.width != icon.getWidth(null)) {
			// need to resize the icon
			Image empty = IconManager.createBufferedImage(d.width, d.height);
			i = Utilities.mergeImages(icon, empty, 0, 0);
		}
		return t.createCustomCursor(i, new Point(1, 1), name);
	}

	public static GraphicsConfiguration getCurrentGraphicsConfiguration() {
		Frame[] frames = Frame.getFrames();

		for (int i = 0; i < frames.length; i++) {
			if (javax.swing.SwingUtilities.findFocusOwner(frames[i]) != null) {
				return frames[i].getGraphicsConfiguration();
			}
		}
		return GraphicsEnvironment
			.getLocalGraphicsEnvironment()
			.getDefaultScreenDevice()
			.getDefaultConfiguration();
	}
	/** Get the operating system on which the IDE is running.
	* @return one of the <code>OS_*</code> constants (such as {@link #OS_WINNT})
	*/
	public static final int getOperatingSystem() {
		if (operatingSystem == -1) {
			String osName = System.getProperty("os.name");
			if ("Windows NT".equals(osName)) // NOI18N
				operatingSystem = OS_WINNT;
			else if ("Windows 95".equals(osName)) // NOI18N
				operatingSystem = OS_WIN95;
			else if ("Windows 98".equals(osName)) // NOI18N
				operatingSystem = OS_WIN98;
			else if ("Windows 2000".equals(osName)) // NOI18N
				operatingSystem = OS_WIN2000;
			else if (osName.startsWith("Windows ")) // NOI18N
				operatingSystem = OS_WIN_OTHER;
			else if ("Solaris".equals(osName)) // NOI18N
				operatingSystem = OS_SOLARIS;
			else if (osName.startsWith("SunOS")) // NOI18N
				operatingSystem = OS_SOLARIS;
			// JDK 1.4 b2 defines os.name for me as "Redhat Linux" -jglick
			else if (osName.endsWith("Linux")) // NOI18N
				operatingSystem = OS_LINUX;
			else if ("HP-UX".equals(osName)) // NOI18N
				operatingSystem = OS_HP;
			else if ("AIX".equals(osName)) // NOI18N
				operatingSystem = OS_AIX;
			else if ("Irix".equals(osName)) // NOI18N
				operatingSystem = OS_IRIX;
			else if ("SunOS".equals(osName)) // NOI18N
				operatingSystem = OS_SUNOS;
			else if ("Digital UNIX".equals(osName)) // NOI18N
				operatingSystem = OS_TRU64;
			else if ("OS/2".equals(osName)) // NOI18N
				operatingSystem = OS_OS2;
			else if ("OpenVMS".equals(osName)) // NOI18N
				operatingSystem = OS_VMS;
			else if (osName.equals("Mac OS X")) // NOI18N
				operatingSystem = OS_MAC;
			else if (osName.startsWith("Darwin")) // NOI18N
				operatingSystem = OS_MAC;
			else
				operatingSystem = OS_OTHER;
		}
		return operatingSystem;
	}

	public static Rectangle getUsableScreenBounds(GraphicsConfiguration gconfig) {
		Rectangle bounds = new Rectangle(gconfig.getBounds());
		try {
			Toolkit toolkit = Toolkit.getDefaultToolkit();
			Method m =
				Toolkit.class.getMethod("getScreenInsets", new Class[] { GraphicsConfiguration.class });
			if (m == null)
				return bounds;
			Insets insets = (Insets) m.invoke(toolkit, new Object[] { gconfig });
			bounds.y += insets.top;
			bounds.x += insets.left;
			bounds.height -= insets.top + insets.bottom;
			bounds.width -= insets.left + insets.right;
		}
		catch (Exception e) {

		}
		return bounds;
	}

	/** Test whether a given string is a valid Java identifier.
	* @param id string which should be checked
	* @return <code>true</code> if a valid identifier
	*/
	public static final boolean isJavaIdentifier(String id) {
		if (id == null)
			return false;
		if (id.equals(""))
			return false; // NOI18N
		if (!(java.lang.Character.isJavaIdentifierStart(id.charAt(0))))
			return false;
		for (int i = 1; i < id.length(); i++) {
			if (!(java.lang.Character.isJavaIdentifierPart(id.charAt(i))))
				return false;
		}
		// test if id is a keyword
		if (keywords().containsKey(id))
			return false;

		return true;
	}
	/** Test whether the IDE is running on some variant of Unix.
	* Linux is included as well as the commercial vendors.
	* @return <code>true</code> some sort of Unix, <code>false</code> if some other manner of operating system
	*/
	public static final boolean isUnix() {
		return (getOperatingSystem() & OS_UNIX_MASK) != 0;
	}
	/** Test whether the IDE is running on some variant of Windows.
	* @return <code>true</code> if Windows, <code>false</code> if some other manner of operating system
	*/
	public static final boolean isWindows() {
		return (getOperatingSystem() & OS_WINDOWS_MASK) != 0;
	}

	private static synchronized HashMap keywords() {
		if (keywords != null) {
			HashMap map = (HashMap) keywords.get();
			if (map != null) {
				return map;
			}
		}

		HashMap keywords = new HashMap(71);
		keywords.put("abstract", "abstract");
		keywords.put("default", "default"); // NOI18N
		keywords.put("if", "if");
		keywords.put("private", "private"); // NOI18N
		keywords.put("throw", "throw");
		keywords.put("boolean", "boolean"); // NOI18N
		keywords.put("do", "do");
		keywords.put("implements", "implements"); // NOI18N
		keywords.put("protected", "protected");
		keywords.put("throws", "throws"); // NOI18N
		keywords.put("break", "break");
		keywords.put("double", "double"); // NOI18N
		keywords.put("import", "import");
		keywords.put("public", "public"); // NOI18N
		keywords.put("transient", "transient");
		keywords.put("byte", "byte"); // NOI18N
		keywords.put("else", "else");
		keywords.put("instanceof", "instanceof"); // NOI18N
		keywords.put("return", "return");
		keywords.put("try", "try"); // NOI18N
		keywords.put("case", "case");
		keywords.put("extends", "extends"); // NOI18N
		keywords.put("int", "int");
		keywords.put("short", "short"); // NOI18N
		keywords.put("void", "void");
		keywords.put("catch", "catch"); // NOI18N
		keywords.put("final", "final");
		keywords.put("interface", "interface"); // NOI18N
		keywords.put("static", "static");
		keywords.put("volatile", "volatile"); // NOI18N
		keywords.put("char", "char");
		keywords.put("finally", "finally"); // NOI18N
		keywords.put("long", "long");
		keywords.put("class", "class"); // NOI18N
		keywords.put("while", "while");
		keywords.put("super", "super"); // NOI18N
		keywords.put("float", "float");
		keywords.put("native", "native"); // NOI18N
		keywords.put("switch", "switch");
		keywords.put("const", "const"); // NOI18N
		keywords.put("for", "for");
		keywords.put("new", "new"); // NOI18N
		keywords.put("synchronized", "synchronized");
		keywords.put("continue", "continue"); // NOI18N
		keywords.put("continue", "continue");
		keywords.put("goto", "goto"); // NOI18N
		keywords.put("package", "package");
		keywords.put("this", "this"); // NOI18N
		keywords.put("null", "null");
		keywords.put("true", "true"); // NOI18N
		keywords.put("false", "false"); // NOI18N
		keywords.put("assert", "assert"); //JDK 1.4 NOI18N

		Utilities.keywords = new SoftReference(keywords);

		return keywords;
	}

	public static final Icon loadIcon(String resourceID) {
		Image img = loadImage(resourceID);
		if (img == null)
			return null;
		return new ImageIcon(loadImage(resourceID));
	}

	public static final Image loadImage(String resourceID) {
		return IconManager.getIcon(resourceID);
	}

	/** This method merges two images into the new one. The second image is drawn
		* over the first one with its top-left corner at x, y. Images mustn't be of the same size.
		* New image will have a size of max(second image size + top-left corner, first image size).
		* Method is used mostly when second image contains transparent pixels (e.g. for badging).
		* If both images are <code>null</code>, it makes default transparent 16x16 image.
		* @param image1 underlying image
		* @param image2 second image
		* @param x x position of top-left corner
		* @param y y position of top-left corner
		* @return new merged image
		*/
	public static final Image mergeImages(Image image1, Image image2, int x, int y) {
		return IconManager.mergeImages(image1, image2, x, y);
	}

	/** Parses parameters from a given string in shell-like manner.
	* Users of the Bourne shell (e.g. on Unix) will already be familiar with the behavior.
	* For example, when using <code>org.openide.execution.NbProcessDescriptor</code> (Execution API)
	* you should be able to:
	* <ul>
	* <li>Include command names with embedded spaces, such as <code>c:\Program Files\jdk\bin\javac</code>.
	* <li>Include extra command arguments, such as <code>-Dname=value</code>.
	* <li>Do anything else which might require unusual characters or processing. For example:
	* <p><code><pre>
	* "c:\program files\jdk\bin\java" -Dmessage="Hello /\\/\\ there!" -Xmx128m
	* </pre></code>
	* <p>This example would create the following executable name and arguments:
	* <ol>
	* <li> <code>c:\program files\jdk\bin\java</code>
	* <li> <code>-Dmessage=Hello /\/\ there!</code>
	* <li> <code>-Xmx128m</code>
	* </ol>
	* Note that the command string does not escape its backslashes--under the assumption
	* that Windows users will not think to do this, meaningless escapes are just left
	* as backslashes plus following character.
	* </ul>
	* <em>Caveat</em>: even after parsing, Windows programs (such as the Java launcher)
	* may not fully honor certain
	* characters, such as quotes, in command names or arguments. This is because programs
	* under Windows frequently perform their own parsing and unescaping (since the shell
	* cannot be relied on to do this). On Unix, this problem should not occur.
	* @param s a string to parse
	* @return an array of parameters
	*/
	public static String[] parseParameters(String s) {
		int NULL = 0x0; // STICK + whitespace or NULL + non_"
		int INPARAM = 0x1; // NULL + " or STICK + " or INPARAMPENDING + "\ // NOI18N
		int INPARAMPENDING = 0x2; // INPARAM + \
		int STICK = 0x4; // INPARAM + " or STICK + non_" // NOI18N
		int STICKPENDING = 0x8; // STICK + \
		Vector params = new Vector(5, 5);
		char c;

		int state = NULL;
		StringBuffer buff = new StringBuffer(20);
		int slength = s.length();
		for (int i = 0; i < slength; i++) {
			c = s.charAt(i);
			if (Character.isWhitespace(c)) {
				if (state == NULL) {
					if (buff.length() > 0) {
						params.addElement(buff.toString());
						buff.setLength(0);
					}
				}
				else if (state == STICK) {
					params.addElement(buff.toString());
					buff.setLength(0);
					state = NULL;
				}
				else if (state == STICKPENDING) {
					buff.append('\\');
					params.addElement(buff.toString());
					buff.setLength(0);
					state = NULL;
				}
				else if (state == INPARAMPENDING) {
					state = INPARAM;
					buff.append('\\');
					buff.append(c);
				}
				else { // INPARAM
					buff.append(c);
				}
				continue;
			}

			if (c == '\\') {
				if (state == NULL) {
					++i;
					if (i < slength) {
						char cc = s.charAt(i);
						if (cc == '"' || cc == '\\') {
							buff.append(cc);
						}
						else if (Character.isWhitespace(cc)) {
							buff.append(c);
							--i;
						}
						else {
							buff.append(c);
							buff.append(cc);
						}
					}
					else {
						buff.append('\\');
						break;
					}
					continue;
				}
				else if (state == INPARAM) {
					state = INPARAMPENDING;
				}
				else if (state == INPARAMPENDING) {
					buff.append('\\');
					state = INPARAM;
				}
				else if (state == STICK) {
					state = STICKPENDING;
				}
				else if (state == STICKPENDING) {
					buff.append('\\');
					state = STICK;
				}
				continue;
			}

			if (c == '"') {
				if (state == NULL) {
					state = INPARAM;
				}
				else if (state == INPARAM) {
					state = STICK;
				}
				else if (state == STICK) {
					state = INPARAM;
				}
				else if (state == STICKPENDING) {
					buff.append('"');
					state = STICK;
				}
				else { // INPARAMPENDING
					buff.append('"');
					state = INPARAM;
				}
				continue;
			}

			if (state == INPARAMPENDING) {
				buff.append('\\');
				state = INPARAM;
			}
			else if (state == STICKPENDING) {
				buff.append('\\');
				state = STICK;
			}
			buff.append(c);
		}
		// collect
		if (state == INPARAM) {
			params.addElement(buff.toString());
		}
		else if ((state & (INPARAMPENDING | STICKPENDING)) != 0) {
			buff.append('\\');
			params.addElement(buff.toString());
		}
		else { // NULL or STICK
			if (buff.length() != 0) {
				params.addElement(buff.toString());
			}
		}
		String[] ret = new String[params.size()];
		params.copyInto(ret);
		return ret;
	}

	private Utilities() {
	}

}
