/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.windows;

import java.awt.Component;
import java.awt.Container;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;

import javax.swing.JFileChooser;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.FileFilter;

import xxl.applications.gui.actions.ActionManager;
import xxl.applications.gui.actions.AddNewAction;
import xxl.applications.gui.actions.ConnectingToolAction;
import xxl.applications.gui.actions.DeleteAction;
import xxl.applications.gui.actions.FoldSelectionAction;
import xxl.applications.gui.actions.MovingToolAction;
import xxl.applications.gui.actions.SaveAction;
import xxl.applications.gui.actions.SelectionToolAction;
import xxl.applications.gui.actions.UnfoldSelectionAction;
import xxl.applications.gui.actions.ZoomInAction;
import xxl.applications.gui.actions.ZoomOutAction;
import xxl.applications.gui.awt.AbstractPopupMenu;
import xxl.applications.gui.awt.AbstractTool;
import xxl.applications.gui.awt.ConnectingTool;
import xxl.applications.gui.awt.DefaultGraphPanel;
import xxl.applications.gui.awt.MovingTool;
import xxl.applications.gui.awt.SelectionTool;
import xxl.applications.gui.awt.filter.FileFilterSerializedObject;
import xxl.applications.gui.awt.filter.FileFilterXML;
import xxl.applications.gui.code.AbstractItem;
import xxl.applications.gui.graph.ComplexVertex;
import xxl.applications.gui.graph.Edge;
import xxl.applications.gui.graph.Graph;
import xxl.applications.gui.graph.Vertex;
import xxl.applications.gui.graph.event.ConnectionEvent;
import xxl.applications.gui.graph.event.ConnectionListener;
import xxl.applications.gui.serializer.XMLDataOutput;
import xxl.applications.gui.serializer.XMLSerializeConverter;
import xxl.applications.gui.serializer.XMLSerializer;

/**
 * @author maxim
 */
public class GraphFrame extends InternalFrame implements ConnectionListener {

	protected DefaultGraphPanel graphPane;
	private double zoom = 1D;

	protected GraphFrame() {
		super();
	}

	public GraphFrame(String title, Graph graph) {
		super();
		graphPane = new DefaultGraphPanel(graph == null ? createNew() : graph);
		getContentPane().add(new JScrollPane(graphPane));
		setTitle(title);
	}

	protected Graph createNew() {
		return new Graph();
	}

	protected void registerActions() {
		ActionListener toolListener = new ToolActionImpl();
		ActionListener zoomListener = new ZoomActionImpl();
		registerAction(ZoomInAction.ACTION_COMMAND_VALUE, zoomListener);
		registerAction(ZoomOutAction.ACTION_COMMAND_VALUE, zoomListener);
		registerAction(MovingToolAction.ACTION_COMMAND_VALUE, toolListener);
		registerAction(ConnectingToolAction.ACTION_COMMAND_VALUE, toolListener);
		registerAction(SelectionToolAction.ACTION_COMMAND_VALUE, toolListener);
		registerAction(SaveAction.ACTION_COMMAND_VALUE, new SaveActionImpl());
		registerAction(AddNewAction.ACTION_COMMAND_VALUE, new AddNewActionImpl());
		registerAction(DeleteAction.ACTION_COMMAND_VALUE, new DeleteActionImpl());
		registerAction(FoldSelectionAction.ACTION_COMMAND_VALUE, new FoldSelectionActionImpl());
		registerAction(UnfoldSelectionAction.ACTION_COMMAND_VALUE, new UnfoldSelectionActionImpl());
	}

	private void save(String fileName) throws FileNotFoundException, IOException {
		Graph graph = graphPane.getDrawer().getGraph();
		if (graph != null) {
			ObjectOutputStream ous = new ObjectOutputStream(new FileOutputStream(fileName));
			ous.writeObject(graph);
			ous.close();
		}
	}

	private void saveXML(String fileName) throws FileNotFoundException, IOException {
		Graph graph = graphPane.getDrawer().getGraph();
		if (graph != null) {
			FileOutputStream fos = new FileOutputStream(fileName);
			XMLDataOutput xmlOut = new XMLDataOutput(fos);			
			//XMLSerializeConverter.DEFAULT_INSTANCE.write(xmlOut, graph);
			XMLSerializeConverter.DEFAULT_INSTANCE.write(xmlOut, graph.getVertices());
			XMLSerializeConverter.DEFAULT_INSTANCE.write(xmlOut, graph.getEdges());
			xmlOut.close();
			fos.close();			
		}
	}

	/**
	 * Saves the frame content to xml, works only for simple graphs for now
	 * @param fileName name of the file to save the graph to
	 * @throws FileNotFoundException delegated from OutputStream();
	 * @throws IOException delegated from OutputStream.close();
	 */
	private void saveXmlNEW(String fileName) throws FileNotFoundException, IOException {
		Graph graph = graphPane.getDrawer().getGraph();
		if (graph != null) {
			FileOutputStream fos = new FileOutputStream(fileName);
			XMLDataOutput xmlOut = new XMLDataOutput(fos);
			XMLSerializer.getInstance().write(xmlOut, graph);
			xmlOut.flush();
			xmlOut.close();
			fos.close();
		}
	}

	public void doActivate() {
		super.doActivate();
		activateTool(true);
	}

	public void doDeactivate() {
		super.doDeactivate();
		activateTool(false);
	}

	private void activateTool(boolean active) {
		if (graphPane.getTool() != null) {
			String action = null;
			AbstractTool tool = graphPane.getTool();
			if (tool instanceof MovingTool) {
				action = MovingToolAction.ACTION_COMMAND_VALUE;
			}
			else if (tool instanceof SelectionTool) {
				action = SelectionToolAction.ACTION_COMMAND_VALUE;
			}
			else if (tool instanceof ConnectingTool) {
				action = ConnectingToolAction.ACTION_COMMAND_VALUE;
			}
			if (action != null) {
				System.out.println("GraphFrame.activateTool() " + action);
				ActionManager.sharedInstance().getToggleAction(action).toggle(active);
			}
		}
	}

	private void doSave() {
		JFileChooser fileChooser = new JFileChooser();
		// Only allow our formats (not "All files")
		fileChooser.setAcceptAllFileFilterUsed(false);
		FileFilter xmlFilter = new FileFilterXML();
		fileChooser.addChoosableFileFilter(xmlFilter);
		fileChooser.addChoosableFileFilter(new FileFilterSerializedObject());
		// XML is the default
		fileChooser.setFileFilter(xmlFilter);
		String fileName;
		int result = fileChooser.showSaveDialog(this);
		if (result == JFileChooser.APPROVE_OPTION) {
			fileName = fileChooser.getSelectedFile().getAbsolutePath();
			FileFilter filter = fileChooser.getFileFilter();
			try {
				// No file extension given?
				if (fileName.indexOf(".") == -1) {
					//	append ".xml" to the filename
					if (filter instanceof FileFilterXML)
						fileName += ".xml";
					// append ".xso" to the filename
					else if (filter instanceof FileFilterSerializedObject)
						fileName += ".xso";
				}
				// save as XML
				if (filter instanceof FileFilterXML){
					
					saveXmlNEW(fileName);					
					//saveXML(fileName);
				}
				// save using Java serialization
				else
					save(fileName);
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	public void handleConnection(ConnectionEvent ce) {
		Edge e = new Edge(ce.getOrigin(), ce.getDestination());
		System.out.println("GraphFrame.handleConnection() " + e);
		ce.getGrap().addEdge(e);
	}

	public void setTool(AbstractTool tool) {
		graphPane.setTool(tool);
	}

	private final class AddNewActionImpl implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			Container popup = SwingUtilities
					.getAncestorOfClass(JPopupMenu.class, (Component) e.getSource());
			Vertex v = new Vertex();
			if (popup instanceof AbstractPopupMenu) {
				v.setLocation(((AbstractPopupMenu) popup).getLastScreenLocation());
			}
			graphPane.getDrawer().getGraph().addVertex(v);
		}
	}

	private final class SaveActionImpl implements ActionListener {

		public void actionPerformed(ActionEvent e) {			
			doSave();
		}
	}

	private final class ToolActionImpl implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			String ac = e.getActionCommand().intern();
			if (ac == MovingToolAction.ACTION_COMMAND_VALUE) {
				setTool(MovingTool.getInstance());
			}
			else if (ac == SelectionToolAction.ACTION_COMMAND_VALUE) {
				setTool(SelectionTool.getInstance());
			}
			else if (ac == ConnectingToolAction.ACTION_COMMAND_VALUE) {
				setTool(ConnectingTool.getInstance());
			}
		}
	}

	private final class ZoomActionImpl implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			String ac = e.getActionCommand().intern();
			if (ac == ZoomInAction.ACTION_COMMAND_VALUE) {
				if (zoom < 10) {
					zoom += 0.2d;
				}
			}
			else if (ac == ZoomOutAction.ACTION_COMMAND_VALUE) {
				if (zoom >= 0.4) {
					zoom -= 0.2d;
				}
			}
			graphPane.setScale(zoom);
		}
	}

	private final class DeleteActionImpl implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			Vertex[] v = graphPane.getSelection();
			Graph g = graphPane.getDrawer().getGraph();
			if (g != null) {
				for (int i = 0; i < v.length; i++)
					g.removeVertex(v[i]);
			}
		}
	}

	private final class FoldSelectionActionImpl implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			if (graphPane != null) {
				Vertex[] selection = graphPane.getSelection();
				if (selection.length > 0) {
					Rectangle bounds = null;
					ComplexVertex cv = new ComplexVertex();
					for (int i = 0; i < selection.length; i++) {
						Rectangle tmp = graphPane.getDrawer().getVertexBounds(selection[i]);
						cv.addChildVertex(selection[i]);
						if (i == 0)
							bounds = new Rectangle(tmp);
						else
							bounds.add(tmp);
					}
					Point location = new Point((bounds.x + bounds.width) / 2,
							(bounds.y + bounds.height) / 2);
					cv.setLocation(location);
					graphPane.getDrawer().getGraph().addVertex(cv);
				}
			}
		}
	}

	private final class UnfoldSelectionActionImpl implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			if (graphPane != null) {
				Vertex[] selection = graphPane.getSelection();
				if (selection.length > 0) {
					for (int i = 0; i < selection.length; i++) {
						if (selection[i] instanceof AbstractItem)
							continue;
						else if (selection[i] instanceof ComplexVertex) {
							ComplexVertex cv = (ComplexVertex) selection[i];
							cv.removeAllChildren();
							graphPane.getDrawer().getGraph().removeVertex(cv);
						}
					}
				}
			}
		}
	}
}
