/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.windows;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import javax.swing.DefaultDesktopManager;
import javax.swing.JDesktopPane;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JSplitPane;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.FileFilter;

import xxl.applications.gui.actions.AboutAction;
import xxl.applications.gui.actions.ActionManager;
import xxl.applications.gui.actions.CallableAction;
import xxl.applications.gui.actions.ExitAction;
import xxl.applications.gui.actions.NewAction;
import xxl.applications.gui.actions.OpenAction;
import xxl.applications.gui.actions.TileHorizontalAction;
import xxl.applications.gui.actions.TileVerticalAction;
import xxl.applications.gui.awt.ClassTreePanel;
import xxl.applications.gui.awt.MainMenu;
import xxl.applications.gui.awt.StatusBar;
import xxl.applications.gui.awt.filter.FileFilterSerializedObject;
import xxl.applications.gui.awt.filter.FileFilterXML;
import xxl.applications.gui.awt.toolbars.FileToolBar;
import xxl.applications.gui.awt.toolbars.GraphEditToolBar;
import xxl.applications.gui.awt.toolbars.ToolBarPanel;
import xxl.applications.gui.awt.toolbars.ZoomToolBar;
import xxl.applications.gui.graph.ClassGraph;
import xxl.applications.gui.graph.Edge;
import xxl.applications.gui.graph.Graph;
import xxl.applications.gui.graph.Vertex;
import xxl.applications.gui.property.PropertyPane;
import xxl.applications.gui.serializer.XMLDataInput;
import xxl.applications.gui.serializer.XMLSerializeConverter;
import xxl.applications.gui.serializer.XMLSerializer;
import xxl.applications.gui.util.Utilities;

/**
 * @author maxim
 * @version 1.0
 */
public final class MainFrame extends JFrame {

	//	constants specifying resize behavior
	public final static int RESIZE_VERTICAL = 0;
	public final static int RESIZE_HORIZONTAL = 1;
	public final static int RESIZE_NONE = -1;
	static final int NEW_GRAPH = 0;
	static final int NEW_CLASS_GRAPH = 1;
	static final String[] NEW_DESCRIPTION = { "Graph", "Class Graph" };
	private static MainFrame sharedInstance;
	JDesktopPane desktopPane;
	private ActionListener actionListener = new ActionAdapter();
	private DialogDisplayer dialog;
	private int resizeMode;

	public static MainFrame getInstance() {
		if (sharedInstance == null)
			sharedInstance = new MainFrame();
		return sharedInstance;
	}

	private MainFrame() {
		initialize();
		setTitle(" XXLViewer v. 2.0");
		registerActions();
		addComponentListener(new ComponentAdapter() {

			public void componentResized(ComponentEvent e) {
				doResize(resizeMode);
			}
		});
	}

	private void initialize() {
		if (!SwingUtilities.isEventDispatchThread()) {
			SwingUtilities.invokeLater(new Runnable() {

				public void run() {
					initialize();
				}
			});
			return;
		}
		desktopPane = new JDesktopPane();
		desktopPane.setDesktopManager(new DefaultDesktopManager());
		desktopPane.setDragMode(JDesktopPane.OUTLINE_DRAG_MODE);
		getContentPane().setLayout(new BorderLayout());
		JSplitPane split2Pane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		split2Pane.setTopComponent(new ClassTreePanel());
		split2Pane.setRightComponent(PropertyPane.getInstance());
		split2Pane.setResizeWeight(0.2);
		JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
		splitPane.setOneTouchExpandable(true);
		splitPane.setTopComponent(split2Pane);
		splitPane.setRightComponent(desktopPane);
		getContentPane().add(splitPane, BorderLayout.CENTER);
		ToolBarPanel panel = new ToolBarPanel();
		panel.add(new FileToolBar());
		panel.add(new GraphEditToolBar());
		panel.add(new ZoomToolBar());
		getContentPane().add(panel, BorderLayout.NORTH);
		getContentPane().add(StatusBar.sharedInstance(), BorderLayout.SOUTH);
		setJMenuBar(new MainMenu());
		addWindowListener(new WindowHandler());
		setIconImage(Utilities.loadImage("xxl/applications/gui/windows/resources/xxl.gif"));
	}

	private void registerActions() {
		registerAction(NewAction.ACTION_COMMAND_VALUE, actionListener);
		registerAction(ExitAction.ACTION_COMMAND_VALUE, actionListener);
		registerAction(OpenAction.ACTION_COMMAND_VALUE, actionListener);
		registerAction(TileVerticalAction.ACTION_COMMAND_VALUE, actionListener);
		registerAction(TileHorizontalAction.ACTION_COMMAND_VALUE, actionListener);
		registerAction(AboutAction.ACTION_COMMAND_VALUE, actionListener);
	}

	private void registerAction(String actionCommand, ActionListener l) {
		CallableAction action = ActionManager.sharedInstance().getCallableAction(actionCommand);
		action.setEnabled(true);
		action.setActionListener(l);
	}

	private void doExit() {
		System.exit(0);
	}

	private void doNew() {
		if (dialog == null) {
			dialog = new DialogDisplayer(this, "New");
		}
		dialog.setVisible(true);
		Object value = dialog.getValue();
		int index = value == null ? -1 : ((Integer) value).intValue();
		InternalFrame innerFrame = null;
		if (index == 0) {
			innerFrame = new GraphFrame("<untitled>", null);
		}
		else if (index == 1) {
			innerFrame = new ClassGraphFrame("<untitled>", null);
		}
		if (innerFrame != null) {
			desktopPane.add(innerFrame);
			innerFrame.setSize(desktopPane.getSize());
			innerFrame.setVisible(true);
			innerFrame.toFront();
		}
	}

	private void doOpen() {
		JFileChooser fcoTmp = new JFileChooser();
		//fcoTmp.setAcceptAllFileFilterUsed(false);
		FileFilter xmlFilter = new FileFilterXML();
		fcoTmp.addChoosableFileFilter(xmlFilter);
		fcoTmp.addChoosableFileFilter(new FileFilterSerializedObject());
		// XML is the default
		fcoTmp.setFileFilter(xmlFilter);
		int intResult;
		String strFileName = null;
		// object we deserialize
		Object o = null;
		Object o2 = null;
		Object o3 = null;
		intResult = fcoTmp.showOpenDialog(this);
		if (intResult == JFileChooser.APPROVE_OPTION) {
			strFileName = fcoTmp.getSelectedFile().getAbsolutePath();
			FileFilter filter = fcoTmp.getFileFilter();
			try {
				if (filter instanceof FileFilterXML || strFileName.endsWith(".xml")) {
					FileInputStream fis = new FileInputStream(strFileName);
					XMLDataInput xmlIn = new XMLDataInput(fis);
					// TODO EXPERIMENTELLER CODE... Sollte mit einem Konverter
					// "schöner" werden
					//o = XMLSerializeConverter.DEFAULT_INSTANCE.read(xmlIn);
					o2 = XMLSerializeConverter.DEFAULT_INSTANCE.read(xmlIn);
					o3 = XMLSerializeConverter.DEFAULT_INSTANCE.read(xmlIn);
					//Graph g = (Graph) o;
					Graph g = new Graph();
					//g.removeAllEdges();
					//g.removeAllVertices();
					Vertex[] v = (Vertex[]) o2;
					for (int i = 0; i < v.length; i++) {
						g.addVertex(v[i]);
						//System.out.println(	"MainFrame.doOpen()  Added vertex "	+ v[i] + " " + v[i].getID() + " (#"	+ (i + 1)+ "/" + v.length + ")");
					}

					Edge[] edges = (Edge[]) o3;
					for (int i = 0; i < edges.length; i++) {
						//System.out.println("Edges[" + i + "]: " + edges[i]);
						//System.out.println("Added edge (#" + (i + 1) + "/" + edges.length + ").");
						Edge edge = edges[i];
						g.addEdge(edge);

					}
					xmlIn.close();
					fis.close();
					o = g;
				}
				//Java object deserialization
				else {
					ObjectInputStream ois = new ObjectInputStream(new FileInputStream(strFileName));
					o = ois.readObject();
				}
			}
			catch (IOException ioex) {
				ioex.printStackTrace();
			}
			catch (ClassNotFoundException cnfe) {
				cnfe.printStackTrace();
			}
		}
		if (o instanceof Graph) {
			GraphFrame frame =
				(o instanceof ClassGraph)
					? new ClassGraphFrame(strFileName, (ClassGraph) o)
					: new GraphFrame(strFileName, (Graph) o);
			desktopPane.add(frame);
			frame.setSize(desktopPane.getSize());
			frame.setVisible(true);
			frame.toFront();
		}
	}

	/**
	 * Opens a graph saved using XMLSerializer and Converter classes
	 */
	private void doOpenNEW() {
		JFileChooser fcoTmp = new JFileChooser();
		//fcoTmp.setAcceptAllFileFilterUsed(false);
		FileFilter xmlFilter = new FileFilterXML();
		fcoTmp.addChoosableFileFilter(xmlFilter);
		fcoTmp.addChoosableFileFilter(new FileFilterSerializedObject());
		// XML is the default
		fcoTmp.setFileFilter(xmlFilter);
		int intResult;
		String strFileName = null;
		// object we deserialize
		Object o = null;
		intResult = fcoTmp.showOpenDialog(this);
		if (intResult == JFileChooser.APPROVE_OPTION) {
			strFileName = fcoTmp.getSelectedFile().getAbsolutePath();
			FileFilter filter = fcoTmp.getFileFilter();
			try {
				if (filter instanceof FileFilterXML || strFileName.endsWith(".xml")) {
					FileInputStream fis = new FileInputStream(strFileName);
					XMLDataInput xmlIn = new XMLDataInput(fis);
					o = XMLSerializer.getInstance().read(xmlIn);
					xmlIn.close();
					fis.close();
				}
				//Java object deserialization
				else {
					ObjectInputStream ois = new ObjectInputStream(new FileInputStream(strFileName));
					o = ois.readObject();
				}
			}
			catch (IOException ioex) {
				ioex.printStackTrace();
			}
			catch (ClassNotFoundException cnfe) {
				cnfe.printStackTrace();
			}
		}
		if (o instanceof Graph) {
			GraphFrame frame =
				(o instanceof ClassGraph)
					? new ClassGraphFrame(strFileName, (ClassGraph) o)
					: new GraphFrame(strFileName, (Graph) o);
			desktopPane.add(frame);
			frame.setSize(desktopPane.getSize());
			frame.setVisible(true);
			frame.toFront();
		}
	}

	private void doResize(int resizeMode) {
		switch (resizeMode) {
			case RESIZE_VERTICAL :
				tileVertical();
				break;
			case RESIZE_HORIZONTAL :
				tileHorizontal();
				break;
		}
	}

	private void tileVertical() {
		resizeMode = RESIZE_VERTICAL;
		if (desktopPane != null) {
			JInternalFrame[] frames = desktopPane.getAllFrames();
			Dimension dim = desktopPane.getSize();
			Dimension location = new Dimension(0, 0);
			for (int i = 0; i < frames.length; i++) {
				frames[i].setLocation(location.width, location.height);
				location.setSize((int) (dim.width * 1.0 / frames.length), 0);
				frames[i].setSize(location.width, dim.height);
			}
		}
	}

	private void tileHorizontal() {
		resizeMode = RESIZE_HORIZONTAL;
		if (desktopPane != null) {
			JInternalFrame[] frames = desktopPane.getAllFrames();
			Dimension dim = desktopPane.getSize();
			Dimension location = new Dimension(0, 0);
			for (int i = 0; i < frames.length; i++) {
				frames[i].setLocation(location.width, location.height);
				location.setSize(0, (int) (dim.height * 1.0 / frames.length));
				frames[i].setSize(dim.width, location.height);
			}
		}
	}

	public void validate() {
		super.validate();
		Dimension curSize = getSize();
		Dimension minSize = getMinimumSize();
		Dimension maxSize = getMaximumSize();
		final Dimension newSize =
			new Dimension(
				Math.min(maxSize.width, Math.max(curSize.width, minSize.width)),
				Math.min(maxSize.height, Math.max(curSize.height, minSize.height)));
		if (!newSize.equals(curSize)) {
			SwingUtilities.invokeLater(new Runnable() {

				public void run() {
					setSize(newSize);
				}
			});
		}
	}

	public Dimension getMinimumSize() {
		return new Dimension(320, super.getPreferredSize().height);
	}

	public Dimension getPreferredSize() {
		return Utilities
			.getUsableScreenBounds(Utilities.getCurrentGraphicsConfiguration())
			.getSize();
	}

	private class WindowHandler extends WindowAdapter {

		public void windowClosing(WindowEvent e) {
			doExit();
		}
	}

	private class ActionAdapter implements ActionListener {

		public void actionPerformed(ActionEvent e) {
			String ac = e.getActionCommand().intern();
			System.out.println(ac);
			if (ac == ExitAction.ACTION_COMMAND_VALUE) {
				doExit();
			}
			else if (ac == OpenAction.ACTION_COMMAND_VALUE) {
				//XXX  
				doOpenNEW();
				//doOpen();
			}
			else if (ac == NewAction.ACTION_COMMAND_VALUE) {
				doNew();
			}
			else if (ac == TileHorizontalAction.ACTION_COMMAND_VALUE) {
				tileHorizontal();
			}
			else if (ac == TileVerticalAction.ACTION_COMMAND_VALUE) {
				tileVertical();
			}
			else if(ac == AboutAction.ACTION_COMMAND_VALUE) {
				AboutDialog.showAboutDlg();
			}
		}
	}
}
