/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.gui.windows.wizards;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableCellRenderer;

import xxl.applications.gui.code.CallVertex;
import xxl.applications.gui.code.CodeException;
import xxl.applications.gui.code.ConstructorCall;
import xxl.applications.gui.code.ElementCall;
import xxl.applications.gui.code.Type;
import xxl.applications.gui.graph.Vertex;

/**
 *
 * @author  maxim
 */
public class ConstructorSecondStepPanel extends JPanel {

	private ConstructorSecondStep step;
	private JTable paramTable;
	private JTable valueTable;
	private int paramTableSelection = -1;
	private int valueTableSelection = -1;

	/** Creates new form SecondStepPanel */
	public ConstructorSecondStepPanel(ConstructorSecondStep step) {
		this.step = step;
		initComponents();
		paramTable = new JTable();
		valueTable = new JTable();
		callComboBox.setRenderer(new CallListCellRenderer());
		TableCellRenderer renderer = new TypeTableCellRenderer();
		paramTable.setDefaultRenderer(Type.class, renderer);
		paramTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		paramTable.getSelectionModel().addListSelectionListener(new ParamSelectionListener());
		valueTable.setDefaultRenderer(Type.class, renderer);
		valueTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		valueTable.getSelectionModel().addListSelectionListener(new ValueSelectionListener());
		valuesScrollPane.setViewportView(valueTable);
		paramsScrollPane.setViewportView(paramTable);
		ActionListener listener = new ButtonActionListener();
		setButton.addActionListener(listener);
		unsetButton.addActionListener(listener);
		deleteCallButton.addActionListener(listener);
		setBorder(BorderFactory.createEmptyBorder(0, 11, 11, 11));
	}

	/** initialize the rest of the panel, is called by second step when it is also initialized */
	void performInit() {
		callComboBox.setModel(itemCallsModel());
		callComboBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				performSelection();
			}
		});
		valueTable.setModel(new ValuesTableModel(step.getSourceItem()));
	}

	/** Crates ComboBoxModel from avaliable calls on target item */
	private DefaultComboBoxModel itemCallsModel() {
		DefaultComboBoxModel model = new DefaultComboBoxModel();
		Vertex[] vertexes = step.getTargetItem().getChildren();
		for (int i = 0; i < vertexes.length; i++) {
			if (vertexes[i] instanceof CallVertex) {
				model.addElement(((CallVertex) vertexes[i]).getCall());
			}
		}
		return model;
	}

	/** is called when ComboBox selection changes */
	private void performSelection() {
		ElementCall item = (ElementCall) callComboBox.getSelectedItem();
		if (item.getType() != ElementCall.PRIMITIVE_CALL) {
			paramTable.setModel(new ParamsTableModel((ConstructorCall) item));
		}
	}

	/** helper class managing actions on set and unset buttons */
	private class ButtonActionListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			ElementCall call = (ElementCall) callComboBox.getSelectedItem();
			//1. there is no call selected 
			if (call == null)
				return;
			if (e.getSource() == deleteCallButton) {
				step.getTargetItem().removeChildVertex(call.getCallVertex());
				return;
			}
			if (call.getType() == ElementCall.PRIMITIVE_CALL)
				return;
			ConstructorCall cCall = (ConstructorCall) call;
			if (cCall.getParamCount() == 0)
				return;
			//the call has one or more parameters
			int row = paramTable.getSelectedRow();
			try {
				if (e.getSource() == unsetButton) {
					//TODO should i remove calls, if it was the last parameter unset?                    
					cCall.setParameter(row, null);
					return;
				}
				int vRow = valueTable.getSelectedRow();
				ElementCall sCall = ((ValuesTableModel) valueTable.getModel()).getCallAt(vRow);
				if (!cCall.isParameterSet(row)) {
					cCall.setParameter(row, sCall);
				}
			}
			//this should not happen since only valid selections enable buttons
			catch (CodeException ce) {
				ce.printStackTrace();
			}
			clearSelection();
		}
	}

	private void clearSelection() {
		paramTable.getSelectionModel().clearSelection();
		valueTable.getSelectionModel().clearSelection();
		valueTableSelection = -1;
		paramTableSelection = -1;
		setButton.setEnabled(false);
		unsetButton.setEnabled(false);
		performSelection();
	}

	private boolean canEnableSet(int paramSelection, int valueSelection) {
		if (paramSelection < 0 || valueSelection < 0)
			return false;
		Type v = (Type) valueTable.getValueAt(valueSelection, 0);
		Type p = (Type) paramTable.getValueAt(paramSelection, 0);
		System.out.println("ConstructorSecondStepPanel.canEnableSet(): value " + v + " param " + p);
		boolean  b = p.isAssignableFrom(v);
		System.out.println("IS ASSIGNABLE : " + b);
		return b;
	}

	private class ParamSelectionListener implements ListSelectionListener {
		public void valueChanged(ListSelectionEvent e) {
			if (e.getValueIsAdjusting())
				return;
			ListSelectionModel model = (ListSelectionModel) e.getSource();
			if (model.isSelectionEmpty())
				return;
			paramTableSelection = model.getMinSelectionIndex();
			if (canEnableSet(paramTableSelection, valueTableSelection)) {
				setButton.setEnabled(true);
			}
			Boolean isSet = (Boolean) paramTable.getValueAt(paramTableSelection, 1);
			if (isSet.booleanValue() == true) {
				unsetButton.setEnabled(true);
			}
		}
	}

	private class ValueSelectionListener implements ListSelectionListener {
		public void valueChanged(ListSelectionEvent e) {
			if (e.getValueIsAdjusting())
				return;
			ListSelectionModel model = (ListSelectionModel) e.getSource();
			if (model.isSelectionEmpty())
				return;
			valueTableSelection = model.getMinSelectionIndex();
			if (canEnableSet(paramTableSelection, valueTableSelection)) {
				setButton.setEnabled(true);
			}
		}
	}

	/** This method is called from within the constructor to
	 * initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is
	 * always regenerated by the Form Editor.
	 */
	private void initComponents() { //GEN-BEGIN:initComponents
		java.awt.GridBagConstraints gridBagConstraints;

		headLabel = new javax.swing.JLabel();
		valuesScrollPane = new javax.swing.JScrollPane();
		paramsScrollPane = new javax.swing.JScrollPane();
		jPanel1 = new javax.swing.JPanel();
		setButton = new javax.swing.JButton();
		unsetButton = new javax.swing.JButton();
		deleteCallButton = new javax.swing.JButton();
		valuesLabel = new javax.swing.JLabel();
		paramLabel = new javax.swing.JLabel();
		callComboBox = new javax.swing.JComboBox();

		setLayout(new java.awt.GridBagLayout());

		headLabel.setText("Select Call to Edit:");
		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridwidth = 3;
		gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
		gridBagConstraints.insets = new java.awt.Insets(0, 0, 2, 0);
		add(headLabel, gridBagConstraints);

		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 0;
		gridBagConstraints.gridy = 3;
		gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
		gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
		gridBagConstraints.weightx = 0.7;
		gridBagConstraints.weighty = 1.0;
		add(valuesScrollPane, gridBagConstraints);

		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 2;
		gridBagConstraints.gridy = 3;
		gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
		gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
		gridBagConstraints.weightx = 0.5;
		gridBagConstraints.weighty = 1.0;
		add(paramsScrollPane, gridBagConstraints);

		jPanel1.setLayout(new java.awt.GridBagLayout());

		setButton.setText("Set >");
		setButton.setMaximumSize(new java.awt.Dimension(108, 26));
		setButton.setMinimumSize(new java.awt.Dimension(108, 26));
		setButton.setPreferredSize(new java.awt.Dimension(108, 26));
		setButton.setEnabled(false);
		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.insets = new java.awt.Insets(0, 0, 2, 0);
		jPanel1.add(setButton, gridBagConstraints);

		unsetButton.setText("< Unset");
		unsetButton.setMaximumSize(new java.awt.Dimension(108, 26));
		unsetButton.setMinimumSize(new java.awt.Dimension(108, 26));
		unsetButton.setPreferredSize(new java.awt.Dimension(108, 26));
		unsetButton.setEnabled(false);
		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 0;
		gridBagConstraints.gridy = 1;
		gridBagConstraints.insets = new java.awt.Insets(0, 0, 6, 0);
		jPanel1.add(unsetButton, gridBagConstraints);

		deleteCallButton.setText("Delete Call");
		deleteCallButton.setMaximumSize(new java.awt.Dimension(108, 26));
		deleteCallButton.setMinimumSize(new java.awt.Dimension(108, 26));
		deleteCallButton.setPreferredSize(new java.awt.Dimension(108, 26));
		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 0;
		gridBagConstraints.gridy = 2;
		jPanel1.add(deleteCallButton, gridBagConstraints);

		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 1;
		gridBagConstraints.gridy = 3;
		gridBagConstraints.anchor = java.awt.GridBagConstraints.NORTH;
		gridBagConstraints.insets = new java.awt.Insets(0, 6, 0, 6);
		add(jPanel1, gridBagConstraints);

		valuesLabel.setText("Select value:");
		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 0;
		gridBagConstraints.gridy = 2;
		gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
		add(valuesLabel, gridBagConstraints);

		paramLabel.setText("Select Parameter:");
		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 2;
		gridBagConstraints.gridy = 2;
		gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
		add(paramLabel, gridBagConstraints);

		callComboBox.setFont(new java.awt.Font("Dialog", 0, 12));
		gridBagConstraints = new java.awt.GridBagConstraints();
		gridBagConstraints.gridx = 0;
		gridBagConstraints.gridy = 1;
		gridBagConstraints.gridwidth = 3;
		gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
		gridBagConstraints.weightx = 1.0;
		add(callComboBox, gridBagConstraints);

	} //GEN-END:initComponents

	// Variables declaration - do not modify//GEN-BEGIN:variables
	private javax.swing.JComboBox callComboBox;
	private javax.swing.JButton deleteCallButton;
	private javax.swing.JLabel headLabel;
	private javax.swing.JPanel jPanel1;
	private javax.swing.JLabel paramLabel;
	private javax.swing.JScrollPane paramsScrollPane;
	private javax.swing.JButton setButton;
	private javax.swing.JButton unsetButton;
	private javax.swing.JLabel valuesLabel;
	private javax.swing.JScrollPane valuesScrollPane;
	// End of variables declaration//GEN-END:variables

}
