/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.indexStructures;

import java.io.File;
import java.io.IOException;

import xxl.applications.indexStructures.IndexTests.BlockCounter;
import xxl.applications.indexStructures.IndexTests.TestDataElement;
import xxl.applications.indexStructures.IndexTests.TestDataHelper;
import xxl.applications.indexStructures.MVBTreeHelper.LongMVRegion;
import xxl.applications.indexStructures.MVBTreeHelper.LongMVSeparator;
import xxl.applications.indexStructures.MVBTreeHelper.LongVersion;
import xxl.core.collections.containers.Container;
import xxl.core.collections.containers.CounterContainer;
import xxl.core.collections.containers.io.BlockFileContainer;
import xxl.core.collections.containers.io.BufferedContainer;
import xxl.core.collections.containers.io.ConverterContainer;
import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.queues.Heap;
import xxl.core.cursors.Cursor;
import xxl.core.cursors.Cursors;
import xxl.core.indexStructures.AutoDeleteMVBTree;
import xxl.core.indexStructures.MVBTree;
import xxl.core.io.Buffer;
import xxl.core.io.LRUBuffer;
import xxl.core.pipes.elements.TemporalObject;

/**
 */
public class AutoDeleteMVBTreeTests {
	private static String path = xxl.core.util.XXLSystem.getOutPath(new String[] { "InputData" }) + File.separator;

	private static final float MIN_RATIO=0.25f;
	private static CounterContainer counterBuf, counterUnbuf;
	private static CounterContainer counterRootsBuf, counterRootsUnbuf;

	public static AutoDeleteMVBTree createTree(String name, int blockSize, int bufferCapacity) {	
		String path = xxl.core.util.XXLSystem.getOutPath(new String[]{ "AutoDeleteMVBTree" }) + File.separator + name;
	
		AutoDeleteMVBTree tree = new AutoDeleteMVBTree(blockSize, MIN_RATIO, 0.5f);
		Buffer buffer = new LRUBuffer(bufferCapacity);

		counterRootsBuf =
			new CounterContainer(
				new ConverterContainer(
					new BlockFileContainer(
						path + ".Roots", blockSize),
					tree.rootsTree().nodeConverter()));
		Container rootsTreeContainer = counterRootsUnbuf =
			new CounterContainer( 
				new BufferedContainer(counterRootsBuf, buffer, true));

		counterBuf = new CounterContainer(new ConverterContainer(new BlockFileContainer(path, blockSize), tree.nodeConverter()));
		Container treeContainer = counterUnbuf = new CounterContainer( 
			new BufferedContainer(counterBuf, buffer, true));

		tree.initialize(TestDataElement.GET_KEY,
						rootsTreeContainer,
						treeContainer, 
						LongVersion.DEFAULT_CONVERTER,
						TestDataElement.KEY_CONVERTER, 
						TestDataElement.DEFAULT_CONVERTER, 
						LongMVSeparator.FACTORY_FUNCTION,
						LongMVRegion.FACTORY_FUNCTION);
		return tree;
	}

	private static BlockCounter queryCounterBuf, queryCounterUnbuf;
	private static String snapshotFile;
	
	private static void queryTest(MVBTree tree, long ts, int window) {
		System.out.println("Query test at "+ts);
		queryCounterBuf.sub(counterBuf);
		queryCounterUnbuf.sub(counterUnbuf);
		
		String file = snapshotFile + ts + ".txt";
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
		int count = 0, missing = 0;
		while (it.hasNext() && count < window / 32) {
			TemporalObject obj = it.next();
			int key = ((TestDataElement)obj.getObject()).getKey();
			if (tree.exactMatchQuery(new Integer(key), tree.currentVersion()) == null) {
				missing++;
				//System.out.println("ERROR: didn't find "+obj);
			}
			if (++count%5000==0)System.out.println("query at "+count);
		}
		it.close();
		System.out.println("queried: "+count+", missing: "+missing);
		
		queryCounterBuf.add(counterBuf);
		queryCounterUnbuf.add(counterUnbuf);
		System.out.println("Query buf: "+queryCounterBuf);
		System.out.println("Query unbuf: "+queryCounterUnbuf);
	}

	private static void rangeQueryTest(MVBTree tree, long ts, int window) {
		System.out.println("Range query test at "+ts);
		queryCounterBuf.sub(counterBuf);
		queryCounterUnbuf.sub(counterUnbuf);

		int range = 2500;
		String file = snapshotFile + ts + "_range" + range + ".txt";
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);

		while (it.hasNext()) {
			TemporalObject from = it.next();
			TemporalObject to = it.next();

			Cursor result = tree.keyRangeQuery(
//					Integer.MIN_VALUE,
//					Integer.MAX_VALUE,
					((TestDataElement)from.getObject()).getKey(),
					((TestDataElement)to.getObject()).getKey(),
					tree.currentVersion()
			);
			int n = Cursors.count(result); 
			if (n != range)
				System.out.println("ERRROR. Found "+n+" elements.");
			result.close();
		}
		it.close();

		queryCounterBuf.add(counterBuf);
		queryCounterUnbuf.add(counterUnbuf);
		System.out.println("Query buf: "+queryCounterBuf);
		System.out.println("Query unbuf: "+queryCounterUnbuf);
	}

	private static void performanceTest() throws IOException {
		int number = 1500000;
		int window = 500000;
//		int window = 250000;

		int queryTest = 0;
		int r = 31250;
		
		int blockSize = 4096;
		int elementSize = 100;
		TestDataElement.setDataSize(elementSize - 20);
		int bufferSize = window / 2 / (blockSize / elementSize) / 10;
//		int bufferSize = window / (blockSize / elementSize) / 10;
		
		if (queryTest != 0) {
			queryCounterBuf = new BlockCounter();
			queryCounterUnbuf = new BlockCounter();
			snapshotFile = path + "elem_" + number + "_" + window + "_snapshot";
		}
			

		System.out.println("Number = "+number+", window = "+window);
		System.out.println("element size = "+elementSize+", blockSize = "+blockSize+", bufferSize = "+bufferSize);

		System.out.println("Creating and initializing the AutoDeleteMVBTree.");
		AutoDeleteMVBTree tree = createTree("mvbtree", blockSize, bufferSize);
		System.out.println("MVBTree: Leaves: B="+tree.getLeafNodeB()+", D="+tree.getLeafNodeD()+"   Innner: B="+tree.getIndexNodeB()+", D="+tree.getIndexNodeD());
		System.out.println("Root-Tree:Leaves: B="+tree.rootsTree().getLeafNodeB()+", D="+tree.rootsTree().getLeafNodeD()+"   Innner: B="+tree.rootsTree().getIndexNodeB()+", D="+tree.rootsTree().getIndexNodeD());
		System.out.println();

		String file = path + "elem_" + number + "_" + window + ".txt";
//		TestDataHelper.createKeyFile_RandomExpTime(file, number, window);
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);

		int count = 0;
		int maxSize = 0;
		long avgSize = 0;

		while (it.hasNext()) {
			TemporalObject obj = it.next();
			tree.insert(new LongVersion(obj.getStart()), new LongVersion(obj.getEnd()), obj.getObject());
//			BufferedContainer cont = (BufferedContainer)mvbTree.getContainer.invoke();
//			System.out.println(cont.fixedElements());
			tree.setCutoffVersion(tree.currentVersion());
			if (obj.getStart() == window) {
				counterRootsBuf.reset();
				counterRootsUnbuf.reset();
				counterBuf.reset();
				counterUnbuf.reset();
				maxSize = 0;
				avgSize = 0;
			}
			if (counterBuf.size() > maxSize)
				maxSize = counterBuf.size();
			avgSize += counterBuf.size();
			if (++count % 50000 == 0)
				System.out.println(count);
			if (queryTest == 1 && obj.getStart() > window && obj.getStart()%r == 0) {
				queryTest(tree, obj.getStart(), window);
			}
			if (queryTest == 2 && obj.getStart() > window && obj.getStart()%r == 0) {
				rangeQueryTest(tree, obj.getStart(), window);
			}
		}

//		mvbTree.printInfo();
//		queryTest(file, mvbTree);
		
		printCounter();
		System.out.println("\nmax. size: "+maxSize+", avg. size: "+(avgSize/(count-window))+", roots' size: "+counterRootsBuf.size());
	}

	private static void printCounter() {
		System.out.println("Tree:");		
		System.out.println("Buffered: "+new BlockCounter(counterBuf));		
		System.out.println("Unbuffered: "+new BlockCounter(counterUnbuf));
		System.out.println("Roots-Tree:");		
		System.out.println("Buffered: "+new BlockCounter(counterRootsBuf));		
		System.out.println("Unbuffered: "+new BlockCounter(counterRootsUnbuf));
		BlockCounter allBuf = new BlockCounter(counterBuf);
		allBuf.add(counterRootsBuf);
		BlockCounter allUnbuf = new BlockCounter(counterUnbuf);
		allUnbuf.add(counterRootsUnbuf);
		System.out.println("Total:");		
		System.out.println("Buffered: "+allBuf);		
		System.out.println("Unbuffered: "+allUnbuf);
	}

	private static AutoDeleteMVBTree test(String file, int blockSize, int bufferCapacity) {
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
		AutoDeleteMVBTree tree = createTree("tree", blockSize, bufferCapacity);
		int count = 0;
		while (it.hasNext()) {
			TemporalObject obj = it.next();
			tree.insert(new LongVersion(obj.getStart()),
						new LongVersion(obj.getEnd()),
						obj.getObject());
			if (++count % 10000 == 0)
				System.out.println(count);
		}
		it.close();

		printCounter();
		System.out.println("Size = "+((Container)tree.determineContainer.invoke()).size());
		System.out.println("Index node capacity = "+tree.getIndexNodeB()+", leaf node capacity = "+tree.getLeafNodeB());

		//TreePrinter.showMVBTree(tree, true);

		return tree;
	}

	private static MVBTree testRegular(String file, int blockSize, int bufferCapacity) {
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
		MVBTree tree = MVBTreeTests.createTree("tree", blockSize, bufferCapacity);
		Heap<TemporalObject> heap = new DynamicHeap<TemporalObject>(TemporalObject.END_TIMESTAMP_COMPARATOR);

		BlockCounter inserts = new BlockCounter();
		BlockCounter removes = new BlockCounter();
		
		int count = 0;
		while (it.hasNext()) {
			TemporalObject obj = it.next();
			while (!heap.isEmpty() && heap.peek().getEnd() <= obj.getStart()) {
				TemporalObject rem = heap.dequeue();
				removes.sub(MVBTreeTests.counterUnbuf);
				MVBTree.LeafEntry del = (MVBTree.LeafEntry)tree.remove(new LongVersion(rem.getEnd()), rem.getObject());
				if (!(del != null && del.data().equals(rem.getObject()))) {
					System.out.println("DELETE FAILED");
				}
				removes.add(MVBTreeTests.counterUnbuf);
			}
			inserts.sub(MVBTreeTests.counterUnbuf);
			tree.insert(new LongVersion(obj.getStart()), obj.getObject());
			inserts.add(MVBTreeTests.counterUnbuf);
			heap.enqueue(obj);
			if (++count % 10000 == 0)
				System.out.println(count);
		}
		it.close();

		MVBTreeTests.printCounter();
		System.out.println("Size = "+((Container)tree.determineContainer.invoke()).size());
		System.out.println("Index node capacity = "+tree.getIndexNodeB()+", leaf node capacity = "+tree.getLeafNodeB());
		System.out.println("inserts "+inserts);
		System.out.println("removes "+removes);
		
		return tree;
	}

	private static void compareToStandardMVBTree() {
		int number = 250000;
		int window = 10000;
		int blockSize = 4096;
		int bufferCapacity = 15;
		TestDataElement.setDataSize(80);

		String file = path + "elem_" + number + "_" + window + ".txt";
		TestDataHelper.createKeyFile_RandomExpTime(file, number, window);

		test(file, blockSize, bufferCapacity);
		
		testRegular(file, blockSize, bufferCapacity);
	}

	private static void queryPerformanceTest(String file, MVBTree tree) {
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);

		System.out.println("Starting query test");
		
		MVBTree.Version version = tree.currentVersion();
		long time = ((LongVersion)version).time;

		CounterContainer cont = (CounterContainer)tree.getContainer.invoke();
		BlockCounter elem;
		BlockCounter total = new BlockCounter(cont);
		total.neg();
		int count = 0;
		while(it.hasNext()) {
			TemporalObject obj = it.next();
			if (obj.getTimeInterval().contains(time)) {
				int key = ((TestDataElement)obj.getObject()).getKey();
				elem = new BlockCounter(cont);
				elem.neg();
				if (tree.exactMatchQuery(new Integer(key), version) == null)
					System.out.println("Missing "+key+", ["+obj.getStart()+","+obj.getEnd()+")");
				else {
					elem.add(cont);
					System.out.println(elem+"  "+obj);
				}
			}
		}
		it.close();
		total.add(cont);
		System.out.println("TOTAL: "+total);
		
		System.out.println("Finished query test");
	}

	private static void queryTest(String file, MVBTree mvbTree) {
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);

		System.out.println("Starting query test");

		int count = 0;
		while(it.hasNext()) {
			TemporalObject obj = it.next();
			int key = ((TestDataElement)obj.getObject()).getKey();
			if (mvbTree.exactMatchQuery(new Integer(key), new LongVersion(obj.getStart())) == null)
				System.out.println("Missing "+key+", ["+obj.getStart()+","+obj.getEnd()+")");
			if (++count % 10000 == 0)
				System.out.println(count);
		}
		
		it.close();
		
		System.out.println("Finished query test");
	}

	private static void test1() {
		int number = 10000;
		int window = 1000;
		int blockSize = 512;
		int bufferCapacity = 10;
		TestDataElement.setDataSize(0);

		String file = path + "elem_" + number + "_" + window + ".txt";
		//TestDataHelper.createKeyFile_RandomExpTime(file, number, window);
		MVBTree tree = testRegular(file, blockSize, bufferCapacity);
		//TreePrinter.showMVBTree(tree, true);
		queryPerformanceTest(file, tree);
//		MVBTreeTests.counterBuffered.DEBUG = true;
		//System.out.println("\n"+tree.exactMatchQuery(new Integer(440), new LongVersion(1)));
	}
	
	private static void createSnapshots() {
		int number = 1500000;
		int window = 500000;
		int r = 31250;
		String file = path + "elem_" + number + "_" + window + ".txt";

		for (int time = window; time <= number; time += r) {
			String outFile = path + "elem_" + number + "_" + window + "_snapshot" + time + ".txt";
			TestDataHelper.createSnapshot(file, outFile, time);
		}
	}

	private static void createRanges() {
		int number = 1500000;
		int window = 500000;
		int r = 31250;
		int count = 5;
		int size = (int)(window / 2 * 0.01);

		for (int time = window; time <= number; time += r) {
			String snapshotFile = path + "elem_" + number + "_" + window + "_snapshot" + time + ".txt";
			String outFile = path + "elem_" + number + "_" + window + "_snapshot" + time + "_range" + size + ".txt";
			TestDataHelper.createRangeFromSnapshot(snapshotFile, outFile, count, size, 249000);
		}
	}
	
	public static void main(String[] args) throws Exception {
//		createSnapshots();
//		createRanges();

//		compareToStandardMVBTree();
		performanceTest();
	}
}
