/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.indexStructures;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Iterator;
import java.util.Random;
import java.util.StringTokenizer;

import xxl.applications.indexStructures.Tests.Employee;
import xxl.core.collections.containers.io.BlockFileContainer;
import xxl.core.collections.containers.io.BufferedContainer;
import xxl.core.collections.containers.io.ConverterContainer;
import xxl.core.indexStructures.BPlusTree;
import xxl.core.io.LRUBuffer;

/**
 * @author Husain Aljazzar
 */
public class BPlusTreeTests {
	
	public static final int BLOCK_SIZE=512;
	public static final float MIN_RATIO=0.5f;
	public static final int P=6;
	public static final int BUFFER_SIZE=10;
	public static final String TEST_FILE_EXTENTION="_TEST.BPT";
		
	private static BPlusTree createBPlusTree(String name) {	
		System.out.println("Initialization of the B+ tree.");
		BPlusTree tree=new BPlusTree(BLOCK_SIZE, MIN_RATIO);
		final BufferedContainer treeContainer = 
			new BufferedContainer(
				new ConverterContainer(
					new BlockFileContainer(name, BLOCK_SIZE),
					tree.nodeConverter()
				),
				new LRUBuffer(BUFFER_SIZE),
				true);
		tree.initialize(Employee.getKey, treeContainer, Employee.KEY_CONVERTER, 
						Employee.DEFAULT_CONVERTER, IntSeparator.FACTORY_FUNCTION, 
						IntKeyRange.FACTORY_FUNCTION); 
		System.out.println("Initialization of the B+ tree finished.");
		return tree;
	}

	public static void testBPlusTree(int high) throws IOException {
		String[] keyFiles=Tests.getKeysFiles(high);
		String testFile=Tests.WORK_DIR_NAME+"/BPlusTree"+high+".test";
		Tests.makeInsertTestFile(keyFiles, testFile);
		testBPlusTree(testFile);
	}
	
	
	public static void testBPlusTree(String testFile) throws IOException {
		BPlusTree bpTree= createBPlusTree(testFile);
		System.out.println("BPlusTree has been created.");
//		System.out.println("MVBTree: B="+bpTree.getB()+" D="+bpTree.getD());
		
		//Read test data
		System.out.println();
		System.out.println("Reading test data from file: "+ testFile);
		BufferedReader in= new BufferedReader( new FileReader(testFile));
		String s;
		while((s=in.readLine())!=null) {
			StringTokenizer tok=new StringTokenizer(s, " ");
			String s1=tok.nextToken();
			String s2=tok.nextToken();
			int persNr=Integer.parseInt(s1);
			double incom=Double.parseDouble(s2);
			Employee d=new Employee(persNr, incom);
			bpTree.insert(d);
		}
		bpTree.wasReorg();
		TreePrinter.showBPlusTree(bpTree);
		BPlusTree.KeyRange region=(BPlusTree.KeyRange)bpTree.rootDescriptor();
		int min=((Integer)region.minBound()).intValue();
		int max=((Integer)region.maxBound()).intValue();
		Random r= new Random();

		System.out.println("Test: Delete & Remove!");
		for(int i=0; i<1000; i++) {
			int key= r.nextInt(max-min)+min;
			int op=r.nextInt(2);
			if(op==0) {
				System.out.print("Remove: "+key);
				Object rmEntry=bpTree.remove(new Employee(key, 0));
				System.out.println(" \t-->"+rmEntry);
			}
			if(op==1) {
				System.out.println("Update: "+key);
				bpTree.update(new Employee(key, 0), new Employee(key, -1));
			}
			/*
			if(bpTree.wasReorg()) {
				TreePrinter.showBPlusTree(bpTree);
				Tests.waitForUser();
			}
			*/	
		}
		//Query Test
		System.out.println("Query Test:");
		int maxQR= max+10;//r.nextInt(max+100);
		int minQR= 0;//r.nextInt(maxQR);
		System.out.println("Query: ["+minQR+", "+maxQR+"]");
		Iterator results=bpTree.rangeQuery(new Integer(minQR), new Integer(maxQR));
		System.out.println("Result: ");
		int count=0;
		while(results.hasNext()) {
			if(count!=0) {
				System.out.print(" \t");
				if(count%5==0) System.out.println();
			}
			System.out.print(results.next());
			count++;
		}
		System.out.println();
		System.out.println("Responses number: "+count);
		System.out.println("End.");
	}
		
	public static void main(String[] args) throws Exception{
		System.out.println("START");
		if(args.length<2) {
			System.out.println("BPlusTreeTests: a test program for the index structure xxl.core.indexStructures.BPlusTree");
			System.out.println();
			System.out.println("This applications has to be called on two wise:");
			System.out.println("1. > java BPlusTreeTests [-o] testFileName");
			System.out.println("2. > java BPlusTreeTests [-n] testSize");
			return;
		}
		if(args[0].equals("-o")) {
			testBPlusTree(Tests.WORK_DIR_NAME+"/"+args[1]);
			return;
		}
		if(args[0].equals("-n")) {
			int high=Integer.parseInt(args[1]);
			testBPlusTree(high);
			return;
		}
		throw new IllegalArgumentException(args[0]);
	}
}
