/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.indexStructures;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import xxl.applications.indexStructures.IndexTests.TestDataHelper;
import xxl.core.collections.containers.Container;
import xxl.core.collections.containers.io.BlockFileContainer;
import xxl.core.collections.containers.io.BufferedContainer;
import xxl.core.collections.containers.io.ConverterContainer;
import xxl.core.cursors.Cursor;
import xxl.core.functions.Function;
import xxl.core.indexStructures.BPlusTree;
import xxl.core.indexStructures.BPlusTreeBulkLoading;
import xxl.core.io.LRUBuffer;
import xxl.core.io.converters.MeasuredConverter;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.predicates.Predicate;

public class BTreeRemoveTest {
	public static BPlusTree createBPlusTree(
			final String treeName,
			final Function getDescriptor,
			final Function getKey,
			final Function separatorFactory,
			final Function keyRangeFactory,
			final MeasuredConverter objectConverter,
			final MeasuredConverter keyConverter) {
		final int blockSize = 512;
		
		BPlusTree tree = new BPlusTree(blockSize, 0.5f);
		final Container treeContainer = 
			new BufferedContainer(
				new ConverterContainer(
					new BlockFileContainer(treeName, blockSize), tree.nodeConverter()
				),
				new LRUBuffer(30),
				true);
		
		tree.initialize(new Function<TemporalObject,Object>() {
							public Object invoke(TemporalObject obj) {
								return getKey.invoke(obj.getObject());
							}
						},
						treeContainer,
						keyConverter, 
						TemporalObject.getConverter(objectConverter),
						separatorFactory, 
						keyRangeFactory);
		return tree;
	}

	public static void test1(BPlusTree tree) {
		String keyFile = "d:\\Benutzer\\btreetest.txt";
		int n = 100000;
		double p = 0.999;

		// erzeuge n zufllige Schlsselwerte
		TestDataHelper.createKeyFile(keyFile, n);

		// fge Objekte in Baum ein
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(keyFile);
		while(it.hasNext())
			tree.insert(it.next());
		it.close();
		
/*
		// mache n exakte Anfragen, um jedes Objekt wiederzufinden
		it = KeyFileHelper.keyFileCursor(keyFile);
		while(it.hasNext()) {
			TemporalObject obj = it.next();
			Cursor result = tree.query(obj);
			if (!result.hasNext())
				System.out.println("Missing "+obj);
			result.close();
			//System.out.println(cont.fixedElements());
		}
		it.close();
*/		
		
		List<TemporalObject> removed = new ArrayList<TemporalObject>();
		List<TemporalObject> remaining = new ArrayList<TemporalObject>();
		
		it = tree.query();
		while (it.hasNext()) {
			TemporalObject obj = it.next();
			if (Math.random() < p) {
				it.remove();
				removed.add(obj);
			}
			else
				remaining.add(obj);
		}
		it.close();

		System.out.println("Remaining elements in tree: "+remaining.size());

		System.out.println("Checking remaining elements");
		Iterator<TemporalObject> it1 = remaining.iterator();
		while (it1.hasNext()) {
			TemporalObject obj = it1.next();
			Cursor result = tree.query(obj);
			if (!result.hasNext())
				System.out.println("Error: missing "+obj);
			Object obj1 = result.next();
			if (!obj1.equals(obj))
				System.out.println("Error: expected "+obj+", found "+obj1);
			result.close();
		}

		System.out.println("Checking removed elements");
		Iterator<TemporalObject> it2 = removed.iterator();
		while (it2.hasNext()) {
			TemporalObject obj = it2.next();
			Cursor result = tree.query(obj);
			if (result.hasNext())
				System.out.println("Error: found "+obj);
			result.close();
		}
	}

	public static void test2(final BPlusTree tree) {
		List list = new ArrayList();
		for (int i = 0; i < 100000; i++)
			list.add(new TemporalObject(new Tests.Employee(i, 0), new TimeInterval(1, 2)));
		
		final double fill = 1;
		
		Predicate<BPlusTree.Node> overflows = new Predicate<BPlusTree.Node>() {
			public boolean invoke(BPlusTree.Node node) {
				return node.number() > (node.level() == 0 ? tree.getLeafNodeB() * fill : tree.getIndexNodeB() * fill);
			}
		};
		
		new BPlusTreeBulkLoading(tree, list.iterator(), tree.determineContainer, overflows);
		
		System.out.println("Container size: "+((Container)tree.getContainer.invoke()).size());
		System.out.println("Leaf: B="+tree.getLeafNodeB()+", D="+tree.getLeafNodeD());
		System.out.println("Inner: B="+tree.getIndexNodeB()+", D="+tree.getIndexNodeD());
		
		for (int i = 0; i < list.size(); i++)
			if (tree.exactMatchQuery(new Integer(i)) == null)
				System.out.println("missing "+i);
		
		Cursor<TemporalObject<Tests.Employee>> it = tree.query();
		for (int i = 0; i < list.size(); i++) {
			if (it.next().getObject().getPersNr() != i)
				System.out.println("error "+i);
		}
		it.close();
	}
	
	public static void main(String[] args) {
		BPlusTree tree;
		
		final String treeName =
			xxl.core.util.XXLSystem.getOutPath(new String[] { "BPlusTreeTests" })
			+ File.separator + "btree";

		tree = createBPlusTree(
				treeName,
				Tests.Employee.getDescriptor,
				Tests.Employee.getKey,
				IntSeparator.FACTORY_FUNCTION, 
				IntKeyRange.FACTORY_FUNCTION,
				Tests.Employee.DEFAULT_CONVERTER,
				Tests.Employee.KEY_CONVERTER);
		
		test2(tree);
	}		
}
