/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.indexStructures;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;

import xxl.applications.indexStructures.IndexTests.BlockCounter;
import xxl.applications.indexStructures.IndexTests.TestDataElement;
import xxl.applications.indexStructures.IndexTests.TestDataHelper;
import xxl.applications.indexStructures.MVBTreeHelper.LongMVRegion;
import xxl.applications.indexStructures.MVBTreeHelper.LongMVSeparator;
import xxl.applications.indexStructures.MVBTreeHelper.LongVersion;
import xxl.applications.indexStructures.Tests.Employee;
import xxl.core.collections.containers.Container;
import xxl.core.collections.containers.CounterContainer;
import xxl.core.collections.containers.io.BlockFileContainer;
import xxl.core.collections.containers.io.BufferedContainer;
import xxl.core.collections.containers.io.ConverterContainer;
import xxl.core.cursors.Cursor;
import xxl.core.indexStructures.MVBTree;
import xxl.core.io.Buffer;
import xxl.core.io.LRUBuffer;
import xxl.core.pipes.elements.TemporalObject;

/**
 * @author Husain Aljazzar
 */
public class MVBTreeTests {
	public static final int BLOCK_SIZE=256;
	public static final float MIN_RATIO=0.25f;
	public static final int P=6;
	public static final int BUFFER_SIZE=30;
	public static final String TEST_FILE_EXTENTION="_TEST.MVBT";
	public static CounterContainer counterBuf, counterUnbuf;
	public static CounterContainer counterRootsBuf, counterRootsUnbuf;

	public static MVBTree createTree(String name, int blockSize, int bufferCapacity) {	
		String path = xxl.core.util.XXLSystem.getOutPath(new String[]{ "MVBTree" }) + File.separator + name;
	
		MVBTree tree = new MVBTree(blockSize, MIN_RATIO, 0.5f);
		Buffer buffer = new LRUBuffer(bufferCapacity);

		counterRootsBuf =
			new CounterContainer(
				new ConverterContainer(
					new BlockFileContainer(
						path + ".Roots", blockSize),
					tree.rootsTree().nodeConverter()));
		Container rootsTreeContainer = counterRootsUnbuf =
			new CounterContainer( 
				new BufferedContainer(counterRootsBuf, buffer, true));

		counterBuf = new CounterContainer(new ConverterContainer(new BlockFileContainer(path, blockSize), tree.nodeConverter()));
		Container treeContainer = counterUnbuf = new CounterContainer( 
			new BufferedContainer(counterBuf, buffer, true));

		tree.initialize(TestDataElement.GET_KEY,
						rootsTreeContainer,
						treeContainer, 
						LongVersion.DEFAULT_CONVERTER,
						TestDataElement.KEY_CONVERTER, 
						TestDataElement.DEFAULT_CONVERTER, 
						LongMVSeparator.FACTORY_FUNCTION,
						LongMVRegion.FACTORY_FUNCTION);

		return tree;
	}

	public static void printCounter() {
		System.out.println("Tree:");		
		System.out.println("External: "+new BlockCounter(counterBuf));		
		System.out.println("Buffered: "+new BlockCounter(counterUnbuf));
		System.out.println("Roots-Tree:");		
		System.out.println("External: "+new BlockCounter(counterRootsBuf));		
		System.out.println("Buffered: "+new BlockCounter(counterRootsUnbuf));
		BlockCounter allBuf = new BlockCounter(counterBuf);
		allBuf.add(counterRootsBuf);
		BlockCounter allUnbuf = new BlockCounter(counterUnbuf);
		allUnbuf.add(counterRootsUnbuf);
		System.out.println("Total:");		
		System.out.println("Buffered: "+allBuf);		
		System.out.println("Unbuffered: "+allUnbuf);
	}

	private static MVBTree createMVBTree(String name) {	
		String path = xxl.core.util.XXLSystem.getOutPath(new String[]{ "MVBTree" }) + File.separator + name;
		System.out.println("Tree path: " + path);

		MVBTree tree= new MVBTree(BLOCK_SIZE, MIN_RATIO, 0.5f);
		final Container rootsTreeContainer= 
			new BufferedContainer(
				new ConverterContainer(
					new BlockFileContainer(path + ".Roots", BLOCK_SIZE),
					tree.rootsTree().nodeConverter()
				),
				new LRUBuffer(BUFFER_SIZE),
				true);
					
		final Container treeContainer = 
			new BufferedContainer(
				new ConverterContainer(
					new BlockFileContainer(path, BLOCK_SIZE),
					tree.nodeConverter()
				),
				new LRUBuffer(BUFFER_SIZE),
				true);
					
		tree.initialize(	Employee.getKey,
							rootsTreeContainer, treeContainer, 
							LongVersion.DEFAULT_CONVERTER,
							Employee.KEY_CONVERTER, 
							Employee.DEFAULT_CONVERTER, 
							IntMVSeparator.FACTORY_FUNCTION,
							IntMVRegion.FACTORY_FUNCTION);
		return tree;
	}

	public static void testMVBTree(int high) throws IOException {
		String[] keyFiles=Tests.getKeysFiles(high);
		String testFile=Tests.WORK_DIR_NAME+"/MVBT"+high+".test";
		Tests.makeInsertTestFile(keyFiles, testFile);
		testMVBTree(testFile);
	}
	
	public static void simpleTest1(String testFile) throws IOException {
		System.out.println("Creating and Initializing the BufferedMVBTree.");
		MVBTree mvbTree= createMVBTree(testFile);
		System.out.println("BufferedMVBTree has been created.");
		System.out.println("MVBTree: Leaves: B="+mvbTree.getLeafNodeB()+", D="+mvbTree.getLeafNodeD()+"   Innner: B="+mvbTree.getIndexNodeB()+", D="+mvbTree.getIndexNodeD());
		System.out.println("Root-Tree:Leaves: B="+mvbTree.rootsTree().getLeafNodeB()+", D="+mvbTree.rootsTree().getLeafNodeD()+"   Innner: B="+mvbTree.rootsTree().getIndexNodeB()+", D="+mvbTree.rootsTree().getIndexNodeD());
		
		int version=0;

		// fge 10 Elemente mit absteigenden Schlsselwerten ein
		
		for(int i = 10; i > 0; i--) {
			Employee d = new Employee(i, 0);
			mvbTree.insert(new LongVersion(version), d);
			version++;
		}
		
		TreePrinter.showMVBTree(mvbTree);

		// frage das Element 1 in der aktuellen Version ab
		
		System.out.println(mvbTree.exactMatchQuery(new Integer(1), new LongVersion(version)));
	}
	
	public static Iterator diff(Iterator it1, Iterator it2) {
		List result = new ArrayList();
		while (it1.hasNext())
			result.add(it1.next());
		
		while (it2.hasNext())
			result.remove(it2.next());
		
		return result.iterator();
	}

	private static MVBTree fill(Cursor<TemporalObject> it, String testFile) throws IOException {
		System.out.println("Creating MVBTree.");
		MVBTree mvbTree = createTree(testFile, 512, 20);
		System.out.println("MVBTree: Leaves: B="+mvbTree.getLeafNodeB()+", D="+mvbTree.getLeafNodeD()+"   Innner: B="+mvbTree.getIndexNodeB()+", D="+mvbTree.getIndexNodeD());
		System.out.println("Root-Tree:Leaves: B="+mvbTree.rootsTree().getLeafNodeB()+", D="+mvbTree.rootsTree().getLeafNodeD()+"   Innner: B="+mvbTree.rootsTree().getIndexNodeB()+", D="+mvbTree.rootsTree().getIndexNodeD());

		while (it.hasNext()) {
			TemporalObject obj = it.next();
			mvbTree.insert(new LongVersion((int)obj.getStart()), obj.getObject());
		}

		TreePrinter.showMVBTree(mvbTree);
		
		return mvbTree;
	}
	
	private static void queryTest(MVBTree tree, Cursor<TemporalObject> it) {
		MVBTree.Version version = tree.currentVersion();
		int count = 0;
		while (it.hasNext()) {
			TestDataElement obj = (TestDataElement)it.next().getObject();
			MVBTree.LeafEntry le = (MVBTree.LeafEntry)tree.exactMatchQuery(obj.getKey(), version);
			if (le == null)
				System.out.println("Missing " + obj);
			if (++count%1000 == 0)
				System.out.println(count);
//			else
//				System.out.println(le.data());
//			BufferedContainer cont = (BufferedContainer)tree.getContainer.invoke();
//			System.out.println(cont.fixedElements());
		}
	}

	public static void simpleTest3(String testFile) throws IOException {
		System.out.println("Creating and Initializing the BufferedMVBTree.");
		MVBTree mvbTree= createMVBTree(testFile);
		System.out.println("BufferedMVBTree has been created.");
		System.out.println("MVBTree: Leaves: B="+mvbTree.getLeafNodeB()+", D="+mvbTree.getLeafNodeD()+"   Innner: B="+mvbTree.getIndexNodeB()+", D="+mvbTree.getIndexNodeD());
		System.out.println("Root-Tree:Leaves: B="+mvbTree.rootsTree().getLeafNodeB()+", D="+mvbTree.rootsTree().getLeafNodeD()+"   Innner: B="+mvbTree.rootsTree().getIndexNodeB()+", D="+mvbTree.rootsTree().getIndexNodeD());

		int version=0;

		for(int i = 0; i < 1000; i++) {
			Employee d=new Employee(i, 0);

			mvbTree.insert(new LongVersion(version), d);
			version++;
		}
		
		for(int i = 490; i < 750; i++) {
			Employee d=new Employee(i, 0);
			mvbTree.remove(new LongVersion(version), d);
			version++;
		}
		System.out.println("Aktuelle Version: " + version);
		
//		System.out.println(mvbTree.exactMatchQuery(new Integer(490), new LongVersion(900)));
		
//		TreePrinter.showMVBTree(mvbTree);
		System.out.println("\n");
		
		Iterator it1 = mvbTree.rangePeriodQuery(new Integer(480), new Integer(500), new LongVersion(1), new LongVersion(2000));
		
//		while (it1.hasNext())
//			System.out.println(it1.next());

		System.out.println("\nFinished.");
	}

	public static void testMVBTree(String testFile) throws IOException {
		System.out.println("Creating and Initializing the BufferedMVBTree.");
		MVBTree mvbTree= createMVBTree(testFile);
		System.out.println("BufferedMVBTree has been created.");
		System.out.println("MVBTree: Leaves: B="+mvbTree.getLeafNodeB()+", D="+mvbTree.getLeafNodeD()+"   Innner: B="+mvbTree.getIndexNodeB()+", D="+mvbTree.getIndexNodeD());
		System.out.println("Root-Tree:Leaves: B="+mvbTree.rootsTree().getLeafNodeB()+", D="+mvbTree.rootsTree().getLeafNodeD()+"   Innner: B="+mvbTree.rootsTree().getIndexNodeB()+", D="+mvbTree.rootsTree().getIndexNodeD());
		
		//Read test data
		System.out.println();
		System.out.println("Reading test data from file: "+ testFile);
		BufferedReader in= new BufferedReader( new FileReader(testFile));
		String s;
		int version=1;
		while((s=in.readLine())!=null) {
			StringTokenizer tok=new StringTokenizer(s, " ");
			String s1=tok.nextToken();
			String s2=tok.nextToken();
			int persNr=Integer.parseInt(s1);
			double incom=Double.parseDouble(s2);
			Employee d=new Employee(persNr, incom);
			mvbTree.insert(new LongVersion(version++), d);
		}
		
		mvbTree.wasReorg();
		TreePrinter.showMVBTree(mvbTree);
		IntMVRegion region=(IntMVRegion)mvbTree.rootDescriptor();
		int min=((Integer)region.minBound()).intValue();
		int max=((Integer)region.maxBound()).intValue();
		Random r= new Random();

		System.out.println("Test: Delete & Remove!");
		for(int i=0; i<1000; i++) {
			int key= r.nextInt(max-min)+min;
			int op=r.nextInt(2);
			if(op==0) {
				System.out.print("Remove: "+key+" at version: "+version);
				MVBTree.LeafEntry rmEntry=(MVBTree.LeafEntry)mvbTree.remove(new LongVersion(version++), new Employee(key, 0));
				System.out.println(" \t-->"+rmEntry);
			}
			if(op==1) {
				System.out.println("Update: "+key+" at version: "+version);
				mvbTree.update(new LongVersion(version++), new Employee(key, 0), new Employee(key, -1));
			}	
			/*
			if(mvbTree.wasReorg()) {
				TreePrinter.showMVBTree(mvbTree);
				Tests.waitForUser();
			}
			*/
		}
		
		//Query Test
		System.out.println("Query Test:");
		int maxQR= max+10;//r.nextInt(max+100);
		int minQR= 0;//r.nextInt(maxQR);
		LongVersion maxV=new LongVersion(/*r.nextInt(version)*/version);
		LongVersion minV=new LongVersion(0/*r.nextInt(maxV.time)*/);
		System.out.println("Query: ["+minQR+", "+maxQR+", <"+minV+", "+maxV+">]");
		Iterator results=mvbTree.rangePeriodQuery(new Integer(minQR), new Integer(maxQR), minV, maxV);
		System.out.println("Result: ");
		int count=0;
		while(results.hasNext()) {
			if(count!=0) {
				System.out.print(" \t");
				if(count%5==0) System.out.println();
			}
			System.out.print(results.next());
			count++;
		}
		System.out.println();
		System.out.println("Responses number: "+count);
		System.out.println("End.");
	}
	
	private static void rangeQueryTest(MVBTree tree, Comparable min, Comparable max) {
		Cursor it = tree.keyRangeQuery(min, max, tree.currentVersion());
		while (it.hasNext()) {
			System.out.println(it.next());
		}
		it.close();				
	}

	public static void main(String[] args) throws Exception {
		String file = "d:\\Benutzer\\testkeys.txt";
		//TestDataHelper.createKeyFile_RandomExpTime(file, 50, 10);

		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
		MVBTree tree = fill(it, "mvbt");
		it.close();
		
		rangeQueryTest(tree, new Integer(Integer.MIN_VALUE), new Integer(Integer.MAX_VALUE));
	}
}
