/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.indexStructures;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import xxl.core.collections.containers.Container;
import xxl.core.cursors.mappers.Mapper;
import xxl.core.functions.Function;
import xxl.core.indexStructures.BPlusTree;
import xxl.core.indexStructures.MVBTree;
import xxl.core.indexStructures.Tree;
import xxl.core.indexStructures.MVBTree.Lifespan;
import xxl.core.indexStructures.MVBTree.Root;
import xxl.core.indexStructures.Tree.IndexEntry;
import xxl.core.predicates.Predicate;

/**
 * @author Husain Aljazzar
 */
public class TreePrinter {

	public static void showTree(BPlusTree tree) {
		if(tree.isEmpty()) {
			System.out.println("The Tree is empty.");
			return;
		} 
		if(tree instanceof MVBTree) showMVBTree((MVBTree)tree);
		else showBPlusTree(tree);
	}
	
	public static void showBPlusTree(BPlusTree tree) {
		System.out.println("BPlusTree");
		System.out.println("Leaf: B="+tree.getLeafNodeB()+", D="+tree.getLeafNodeD());
		System.out.println("Inner: B="+tree.getIndexNodeB()+", D="+tree.getIndexNodeD());
		showNodeHierarchy(tree.rootEntry().id(), tree.container(), new LinkedList());
	}
	
	public static void showMVBTree(MVBTree tree) {
		showMVBTree(tree, false);
	}

	public static void showMVBTree(MVBTree tree, boolean onlyCurrentRoot) {
		System.out.println("MVBTree******");
		List roots=new LinkedList();
		System.out.println("Current Root: "+tree.rootEntry());
		if (!onlyCurrentRoot && !tree.rootsTree().isEmpty()) {
			System.out.println("Roots' BTree:");
			showTree(tree.rootsTree());
			Lifespan life= (Lifespan)tree.rootsTree().rootDescriptor();
			Iterator rootsIter= tree.rootsTree().rangeQuery(life.beginVersion(), life.endVersion());
			while(rootsIter.hasNext()) roots.add(((Root)rootsIter.next()).toIndexEntry());
		}
		roots.add(tree.rootEntry());
		final List markedIds=new LinkedList();
		Iterator rootsIter= roots.iterator();
		while(rootsIter.hasNext()) 	{
			IndexEntry root= (IndexEntry) rootsIter.next();
			System.out.println("******************************************");
			showNodeHierarchy(root.id(), tree.container(), markedIds);
		}
	}

	public static void showCurrentMVBTree(MVBTree tree) {
		System.out.println("Current Version of the MVBTree****");
		System.out.println("Current Root:"+tree.rootEntry());
		showBPlusTree(tree);
	}
	
	private static void showNodeHierarchy(Object id, Container container, final List markedIds) {
		Predicate filter= new Predicate() {
			public boolean invoke(Object o) {
				return ! markedIds.contains(o);
			}
		};
		showNodeHierarchy(id, container, markedIds, filter);
	} 
	
	private static void showNodeHierarchy(Object id, Container container, final List markedIds, Predicate filter) {
		List levelIds=new LinkedList();
		levelIds.add(id);
		do {
			levelIds=showLevel(levelIds, container, markedIds, filter);
			System.out.println();
		}while(!levelIds.isEmpty());
	}
	
	private static List showLevel(List levelIds, Container container, List markedIds, Predicate filter) {
		List nextLevelIds=new LinkedList();
		for(int i=0; i<levelIds.size(); i++) 
			if(filter.invoke(levelIds.get(i))) 
				showNode(levelIds.get(i), container, nextLevelIds, markedIds); 
		return nextLevelIds;
	}
	
	private static List showNode(Object id, Container container, List nextLevelIds, List markedIds) {
		Tree.Node node= (Tree.Node)container.get(id);
		printNode(id,node);
		markedIds.add(id);
		if(node.level()>0) {
			Iterator ids= new Mapper(
									new Function() {
										public Object invoke(Object entry) {
											return ((Tree.IndexEntry)entry).id();
										}
									},node.entries());
			while(ids.hasNext()) nextLevelIds.add(ids.next());
		}
		return nextLevelIds;
	}
	
	private static void printNode(Object id, Tree.Node node) {
		Iterator entries=node.entries();
		System.out.print("Node: ID="+id+" level="+node.level()+" number="+node.number());
		if(node instanceof MVBTree.Node)
			System.out.println(" predecessors=" + ((MVBTree.Node)node).predecessors());
		else
			System.out.println();
		int i=0;
		while(entries.hasNext()) {
			i++;
			System.out.print(entries.next());
			if(i%5!=0) System.out.print("\t");
			else System.out.println();
		} 
		if(i%5!=0) System.out.println();
		System.out.print("_____________________________________________________________");
		System.out.println();	
		
	}
}
