/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.applications.indexStructures;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import xxl.applications.indexStructures.IndexTests.BlockCounter;
import xxl.applications.indexStructures.IndexTests.TestDataElement;
import xxl.applications.indexStructures.IndexTests.TestDataHelper;
import xxl.core.collections.containers.Container;
import xxl.core.collections.containers.CounterContainer;
import xxl.core.collections.containers.io.BlockFileContainer;
import xxl.core.collections.containers.io.BufferedContainer;
import xxl.core.collections.containers.io.ConverterContainer;
import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.queues.Heap;
import xxl.core.comparators.ComparableComparator;
import xxl.core.cursors.Cursor;
import xxl.core.cursors.Cursors;
import xxl.core.cursors.filters.Filter;
import xxl.core.cursors.sources.EmptyCursor;
import xxl.core.cursors.sources.SingleObjectCursor;
import xxl.core.cursors.unions.Merger;
import xxl.core.cursors.wrappers.QueueCursor;
import xxl.core.functions.Function;
import xxl.core.indexStructures.BPlusTree;
import xxl.core.indexStructures.BPlusTreeBulkLoading;
import xxl.core.indexStructures.WaveIndex;
import xxl.core.indexStructures.WaveIndex.ConstituentIndex;
import xxl.core.io.Buffer;
import xxl.core.io.FilesystemOperations;
import xxl.core.io.LRUBuffer;
import xxl.core.io.converters.MeasuredConverter;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.predicates.Predicate;
import xxl.core.util.WrappingRuntimeException;

public class WaveIndexTests {
	public static class LazyClearBlockFileContainer extends BlockFileContainer {
		public LazyClearBlockFileContainer(String prefix, FilesystemOperations fso) {
			super(prefix, fso);
		}

		public LazyClearBlockFileContainer(String prefix, int blockSize, FilesystemOperations fso) {
			super(prefix, blockSize, fso);
		}

		public LazyClearBlockFileContainer(String prefix, int blockSize) {
			super(prefix, blockSize);
		}

		public LazyClearBlockFileContainer(String prefix) {
			super(prefix);
		}
		
		@Override
		public void clear() {
			try {
				reservedBitMap.seek(0);
				for (int i = 0; i < reservedBitMap.length(); i++)
					reservedBitMap.write(0);

				updatedBitMap.seek(0);
				for (int i = 0; i < updatedBitMap.length(); i++)
					updatedBitMap.write(0);
				
				int n = (int)(container.length() / blockSize);
				freeList.setLength(8 * n);
				freeList.seek(0);
				for (int i = 0; i < n; i++)
					freeList.writeLong((long)i * blockSize);
				
				size = 0;
			}
			catch (IOException ie) {
				throw new WrappingRuntimeException(ie);
			}
		}
	}

	private static String path = xxl.core.util.XXLSystem.getOutPath(new String[] { "InputData" }) + File.separator;
	
	/**
	 * Predicate that returns true for a TemporalObject NOT expiring in the
	 * given time interval.
	 */
	private static Predicate<TemporalObject> notDeletedPredicate(final TimeInterval interval) {
		return new Predicate<TemporalObject>() {
			public boolean invoke(TemporalObject obj) {
				return !interval.contains(obj.getEnd());
			}
		}; 
	}

	/**
	 * Implements a very simple ConstituentIndex for the DoublyPartitionedWaveIndex
	 * based on a linked list.
	 */
	private static class SimpleIndex implements ConstituentIndex {
		List<TemporalObject> list = new LinkedList<TemporalObject>();

		public void insert(Iterator<TemporalObject> objects) {
			while (objects.hasNext())
				list.add(objects.next());
		}

		public void remove(TimeInterval interval) {
			Iterator<TemporalObject> it = list.iterator();
			while (it.hasNext())
				if (interval.contains(it.next().getEnd()))
					it.remove();
		}

		public Cursor<TemporalObject> query(final Comparable data) {
			return new Filter<TemporalObject>(list.iterator(),
					new Predicate<TemporalObject>() {
						public boolean invoke(TemporalObject obj) {
							return obj.getObject().equals(data);
						}
					});
		}

		public Cursor<TemporalObject> rangeQuery(final Comparable dataFrom,
				final Comparable dataTo) {
			return new Filter<TemporalObject>(list.iterator(),
					new Predicate<TemporalObject>() {
						public boolean invoke(TemporalObject obj) {
							return dataFrom.compareTo(obj.getObject()) <= 0
									&& dataTo.compareTo(obj.getObject()) >= 0;
						}
					});
		}

		public ConstituentIndex rebuild(boolean keep, final TimeInterval interval, Iterator<TemporalObject> newElements) {
			List<TemporalObject> newList;
			
			if (keep && interval == null) {
				newList = list;
			}
			else {
				newList = new LinkedList<TemporalObject>();
				if (interval != null) {
					Iterator<TemporalObject> it = new Filter<TemporalObject>(list.iterator(), notDeletedPredicate(interval)); 
					while (it.hasNext())
						newList.add(it.next());
				}
			}

			//System.out.print("REBUILD: ");
			
			while (newElements.hasNext()) {
				TemporalObject o = newElements.next();
				//System.out.print(o);
				newList.add(o);
			}
			//System.out.println();
			
			list = newList;
			
			return this;
		}

		public String toString() {
			return list.toString();
		}
	}

	/**
	 * Function for creating a new SimpleIndex.
	 */
	private static Function<Integer, ConstituentIndex> createSimpleIndex = new Function<Integer, ConstituentIndex>() {
		public ConstituentIndex invoke(Integer index) {
			return new SimpleIndex();
		}
	};

	/**
	 * Implements a ConstituentIndex for the DoublyPartitionedWaveIndex
	 * based on a BPlusTree.
	 */
	private static class BPlusTreeIndex implements ConstituentIndex {
		protected BPlusTree tree, tree2;
		protected CounterContainer counterBuf, counter2Buf;
		protected CounterContainer counterUnbuf, counter2Unbuf;
		protected boolean nameFlag = false;

		protected final String treeName;
		protected final Function getDescriptor;
		protected final Function getKey;
		protected final Function separatorFactory;
		protected final Function keyRangeFactory;
		protected final MeasuredConverter objectConverter;
		protected final MeasuredConverter keyConverter;
		protected final Buffer buffer;
		protected final int blockSize;

		protected final Comparator<TemporalObject> comp;

		public BlockCounter insertsBuf = new BlockCounter();
		public BlockCounter insertsUnbuf = new BlockCounter();
		public BlockCounter removesBuf = new BlockCounter();
		public BlockCounter removesUnbuf = new BlockCounter();

		public BPlusTreeIndex(
				final String treeName,
				final Function getDescriptor,
				final Function getKey,
				final Function separatorFactory,
				final Function keyRangeFactory,
				final MeasuredConverter objectConverter,
				final MeasuredConverter keyConverter,
				final Buffer buffer,
				final int blockSize) throws IOException {
			this.treeName = treeName;
			this.getDescriptor = getDescriptor;
			this.getKey = getKey;
			this.separatorFactory = separatorFactory;
			this.keyRangeFactory = keyRangeFactory;
			this.objectConverter = objectConverter;
			this.keyConverter = keyConverter;
			this.buffer = buffer;
			this.blockSize = blockSize;
			
			comp = new Comparator<TemporalObject>() {
				public int compare(TemporalObject obj1, TemporalObject obj2) {
					return ((Comparable)getKey.invoke(obj1.getObject())).compareTo(getKey.invoke(obj2.getObject()));
				}
			};
			
			initialize();
		}
		
		protected void initialize() {
			tree = new BPlusTree(blockSize, 0.5f);

			counterBuf = new CounterContainer(new ConverterContainer(new LazyClearBlockFileContainer(treeName + (nameFlag ? "X" : ""), blockSize), tree.nodeConverter()));
			final Container treeContainer = counterUnbuf = new CounterContainer( 
				new BufferedContainer(
					counterBuf,
					buffer,
					true));

			tree.initialize(new Function<TemporalObject,Object>() {
								public Object invoke(TemporalObject obj) {
									return getKey.invoke(obj.getObject());
								}
							},
							treeContainer,
							keyConverter, 
							TemporalObject.getConverter(objectConverter),
							separatorFactory, 
							keyRangeFactory);
/*
			System.out.println("Created B+Tree:");
			System.out.println("Leaf: B="+tree.getLeafNodeB()+", D="+tree.getLeafNodeD());
			System.out.println("Inner: B="+tree.getIndexNodeB()+", D="+tree.getIndexNodeD());
			System.out.println("Block size: "+blockSize+", buffer capacity: "+buffer.capacity()+", buffer: "+buffer);
			*/
		}

		public void insert(Iterator<TemporalObject> objects) {
			//System.out.println("before insert: external: "+new BlockCounter(counter)+", buffered: "+new BlockCounter(counterBuffered));
			insertsBuf.sub(counterBuf);
			insertsUnbuf.sub(counterUnbuf);
			while (objects.hasNext()) {
				//System.out.println("# external: "+new BlockCounter(counter)+", buffered: "+new BlockCounter(counterBuffered));
				tree.insert(objects.next());
			}
			insertsBuf.add(counterBuf);
			insertsUnbuf.add(counterUnbuf);
			//System.out.println("after  insert: external: "+new BlockCounter(counter)+", buffered: "+new BlockCounter(counterBuffered));
			//System.out.println("size "+counter.size());
		}

		public void remove(TimeInterval interval) {
			//System.out.println("before remove: external: "+new BlockCounter(counter)+", buffered: "+new BlockCounter(counterBuffered));
			removesBuf.sub(counterBuf);
			removesUnbuf.sub(counterUnbuf);
			Cursor<TemporalObject> it = tree.query();
			while (it.hasNext())
				if (interval.contains(it.next().getEnd()))
					it.remove();
			it.close();
			removesBuf.add(counterBuf);
			removesUnbuf.add(counterUnbuf);
			//System.out.println("after remove: external: "+new BlockCounter(counter)+", buffered: "+new BlockCounter(counterBuffered));
			//System.out.println("size "+counter.size());
		}

		public Cursor<TemporalObject> query(final Comparable data) {
			Comparable key = (Comparable)getKey.invoke(data);
			TemporalObject obj = (TemporalObject)tree.exactMatchQuery(key);
			if (obj != null)
				return new SingleObjectCursor<TemporalObject>(obj);
			return new EmptyCursor<TemporalObject>();
		}

		public Cursor<TemporalObject> rangeQuery(final Comparable dataFrom, final Comparable dataTo) {
			Comparable from = (Comparable)getKey.invoke(dataFrom);
			Comparable to = (Comparable)getKey.invoke(dataTo);
			return tree.rangeQuery(from, to);
		}

		/* (non-Javadoc)
		 * @see xxl.core.indexStructures.DoublyPartitionedWaveIndex.ConstituentIndex#rebuild(xxl.core.pipes.TimeInterval, java.util.Iterator)
		 * 
		 * The <tt>newElements</tt> have to be sorted in ascending order for this method to work correctly.
		 */
		public ConstituentIndex rebuild(boolean keep, TimeInterval interval, Iterator<TemporalObject> newElements) {
//			System.out.println("REBUILD "+treeName+" "+keep);
			BPlusTree oldTree = tree;
			CounterContainer oldCounter = counterBuf;
			CounterContainer oldCounterBuffered = counterUnbuf;
			
			//System.out.println("alt: "+oldCounter);

			if (tree2 == null) {
				nameFlag = !nameFlag;
				initialize();
			} else {
				tree = tree2;
				counterBuf = counter2Buf;
				counterUnbuf = counter2Unbuf;
			}

			Cursor<TemporalObject> oldElements;
			if (keep)
				oldElements = interval == null ? 
					oldTree.query() : 
					new Filter<TemporalObject>(oldTree.query(), notDeletedPredicate(interval));
			else
				oldElements = new EmptyCursor<TemporalObject>();
			
			final double fillFactor = 1;
			new BPlusTreeBulkLoading(tree, new Merger<TemporalObject>(comp, oldElements, newElements), tree.determineContainer, getOverflowPredicate(fillFactor));
			oldElements.close();

//			System.out.println("nach bulk-loading, alt: "+oldCounter);
//			System.out.println("nach bulk-loading, neu: "+counter);

			addCounter(counterBuf, oldCounter);
			addCounter(counterUnbuf, oldCounterBuffered);

//			System.out.println("neu: "+counter);

			if (!oldTree.isEmpty())
				oldTree.clear();
			oldCounter.reset();
			oldCounterBuffered.reset();

			tree2 = oldTree;
			counter2Buf = oldCounter;
			counter2Unbuf = oldCounterBuffered;

			return this;
		}
		
		private Predicate<BPlusTree.Node> getOverflowPredicate(final double fillFactor) {
			return new Predicate<BPlusTree.Node>() {
				public boolean invoke(BPlusTree.Node node) {
					return node.number() >
						(node.level() == 0 ? tree.getLeafNodeB() * fillFactor : tree.getIndexNodeB() * fillFactor);
				}
			};
		}
		
		public String toString() {
			StringBuffer str = new StringBuffer();
			Iterator it = tree.query();
			while (it.hasNext())
				str.append(it.next() + " ");
			return str.toString();
		}
	}

	private static Buffer gbuffer;

	private static Function<Integer, ConstituentIndex> createBPlusTreeIndex(final boolean globalBuffer, final int bufferCapacity, final int blockSize) {
		return new Function<Integer, ConstituentIndex>() {
			public ConstituentIndex invoke(Integer index) {
				final String treeName =
					xxl.core.util.XXLSystem.getOutPath(new String[] { "WaveIndex" })
					+ File.separator + "btree" + index;

				try {
					if (globalBuffer && gbuffer == null)
						gbuffer = new LRUBuffer(bufferCapacity);
					Buffer buffer = globalBuffer ? gbuffer : new LRUBuffer(bufferCapacity);
					return new BPlusTreeIndex(
						treeName,
						TestDataElement.GET_DESCRIPTOR,
						TestDataElement.GET_KEY,
						IntSeparator.FACTORY_FUNCTION, 
						IntKeyRange.FACTORY_FUNCTION,
						TestDataElement.DEFAULT_CONVERTER,
						TestDataElement.KEY_CONVERTER,
						buffer,
						blockSize);
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			}
		};
	}
	
	/**
	 * An in-memory buffer for the DoublyPartitionedWaveIndex that returns its
	 * elements in ascending order.
	 */
	private static class SortingMemoryBuffer implements WaveIndex.ElementBuffer {
		protected Heap<TemporalObject>[] buffer;
		
		public SortingMemoryBuffer(int numberOfPartitions, Comparator comp) {
			buffer = new Heap[numberOfPartitions];
			for (int i = 0; i < numberOfPartitions; i++)
				buffer[i] = new DynamicHeap<TemporalObject>(comp);
		}

		public void add(int partition, TemporalObject obj) {
			buffer[partition].enqueue(obj);
		}

		public Iterator<TemporalObject> get(int partition) {
			return new QueueCursor<TemporalObject>(buffer[partition]);
		}

		public void clear() {
			for (int i = 0; i < buffer.length; i++)
				buffer[i].clear();		
		}
		
		public String toString() {
			StringBuffer sb = new StringBuffer();
			for (int i = 0; i < buffer.length; i++) {
				sb.append("#"+i+":");
				Iterator it = get(i);
				while (it.hasNext())
					sb.append(it.next()+" ");
				sb.append("\n");
			}
			return sb.toString();
		}
	}
	
	private static void addCounter(CounterContainer base, CounterContainer add) {
		base.gets += add.gets;
		base.inserts += add.inserts;
		base.removes += add.removes;
		base.reserves += add.reserves;
		base.updates += add.updates;
	}

	private static void addWaveIndexCounters(WaveIndex index, BlockCounter counterBuf, BlockCounter counterUnbuf) {
		ConstituentIndex[] indices = index.getIndices();
		for (int i = 0; i < indices.length; i++) {
			BPlusTreeIndex ind = (BPlusTreeIndex)indices[i];
			counterBuf.add(ind.counterBuf);
			counterUnbuf.add(ind.counterUnbuf);
		}
	}

	private static BlockCounter queryCounterBuf, queryCounterUnbuf;
	private static String snapshotFile;
	
	private static void queryTest(WaveIndex index, long ts, int window) {
		System.out.println("Query test at "+ts);
		BlockCounter beforeBuf = new BlockCounter();
		BlockCounter beforeUnbuf = new BlockCounter();
		addWaveIndexCounters(index, beforeBuf, beforeUnbuf);
		queryCounterBuf.sub(beforeBuf);
		queryCounterUnbuf.sub(beforeUnbuf);
		
		String file = snapshotFile + ts + ".txt";
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
		int count = 0;
		while (it.hasNext() && count < window / 32) {
			TemporalObject obj = it.next();
			Cursor<TemporalObject> result = index.query((Comparable)obj.getObject());
			if (!result.hasNext())
				WaveIndex.DEBUG_OUTPUT.println("ERROR: didn't find " + obj);
			if (result.hasNext() && !obj.equals(result.next()))
				WaveIndex.DEBUG_OUTPUT.println("ERROR: found " + result + ", expected " + obj);
			result.close();
			count++;
			if (count%5000==0)System.out.println("query at "+count);
		}
		it.close();
		
		addWaveIndexCounters(index, queryCounterBuf, queryCounterUnbuf);
		System.out.println("Query buf: "+queryCounterBuf);
		System.out.println("Query unbuf: "+queryCounterUnbuf);
	}

	private static void rangeQueryTest(WaveIndex index, long ts, int window) {
		System.out.println("Range query test at "+ts);
		BlockCounter beforeBuf = new BlockCounter();
		BlockCounter beforeUnbuf = new BlockCounter();
		addWaveIndexCounters(index, beforeBuf, beforeUnbuf);
		queryCounterBuf.sub(beforeBuf);
		queryCounterUnbuf.sub(beforeUnbuf);

		int range = 2500;
		String file = snapshotFile + ts + "_range" + range + ".txt";
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);

		while (it.hasNext()) {
			TemporalObject from = it.next();
			TemporalObject to = it.next();

			Cursor<TemporalObject> result = index.rangeQuery(
//					new TestDataElement(Integer.MIN_VALUE),
//					new TestDataElement(Integer.MAX_VALUE)
					(TestDataElement)from.getObject(),
					(TestDataElement)to.getObject()
			);
			int n = Cursors.count(result); 
			if (n != range)
				System.out.println("ERRROR. Found "+n+" elements.");
			result.close();
		}
		it.close();
		
		addWaveIndexCounters(index, queryCounterBuf, queryCounterUnbuf);
		System.out.println("Query buf: "+queryCounterBuf);
		System.out.println("Query unbuf: "+queryCounterUnbuf);
	}

	private static void testWaveIndex(Cursor<TemporalObject> it, int window, int r, int g, int e, WaveIndex.Strategy strategy, boolean lazy, boolean fixed, int bufferSize, boolean globalBuffer, int blockSize, int queryTest) {
		if (!globalBuffer)
			bufferSize /= g*e;

		WaveIndex.DEBUG_ON_UPDATE = collectSize;
		
		queryCounterBuf = new BlockCounter();
		queryCounterUnbuf = new BlockCounter();
		
		WaveIndex index = new WaveIndex(
				window, r, g, e,
				strategy,
				lazy,
				fixed,
				createBPlusTreeIndex(globalBuffer, bufferSize, blockSize),
				new Function<Integer,WaveIndex.ElementBuffer>() {
					public WaveIndex.ElementBuffer invoke(Integer partitions) {
						return new SortingMemoryBuffer(partitions, TemporalObject.getValueBasedComparator(new ComparableComparator()));						
					}
				});

		int count = 0;
		long t1 = System.currentTimeMillis();
		while (it.hasNext()) {
			TemporalObject obj = it.next();
			index.insert(obj);
			if (++count % 100000 == 0)
				System.out.println(count);
			if (obj.getStart() == window) {
				// reset counters after initialization stage
				for (ConstituentIndex ind : index.getIndices()) {
					BPlusTreeIndex tree = (BPlusTreeIndex)ind;
					tree.counterBuf.reset();
					if (tree.counter2Buf != null) tree.counter2Buf.reset();
					tree.counterUnbuf.reset();
					if (tree.counter2Unbuf != null) tree.counter2Unbuf.reset();
					tree.insertsBuf.reset();
					tree.insertsUnbuf.reset();
					tree.removesBuf.reset();
					tree.removesUnbuf.reset();
					maxSize = 0;
					avgSize = 0;
				}
			}
			if (queryTest == 1 && obj.getStart() > window && obj.getStart()%r == 0) {
				queryTest(index, obj.getStart(), window);
			}
			if (queryTest == 2 && obj.getStart() > window && obj.getStart()%r == 0) {
				rangeQueryTest(index, obj.getStart(), window);
			}
		}
		long t2 = System.currentTimeMillis();
		
		WaveIndex.ConstituentIndex[] indices = index.getIndices();
/*
		if (!lazy) {
		for (int i = 0; i < indices.length; i++) {
			BPlusTreeIndex ind = (BPlusTreeIndex)indices[i];
			Cursor<TemporalObject> xx = ind.tree.query();
			while (xx.hasNext()) {
				TemporalObject obj = xx.next();
				if (obj.getEnd() < index.getLastUpdateTimestamp())
					System.out.println("ERROR: "+obj);
			}
			xx.close();
		}
		}
*/		
		BlockCounter allBuf = new BlockCounter();
		BlockCounter allUnbuf = new BlockCounter();
		BlockCounter insertsBuf = new BlockCounter();
		BlockCounter insertsUnbuf = new BlockCounter();
		BlockCounter removesBuf = new BlockCounter();
		BlockCounter removesUnbuf = new BlockCounter();
		int size = 0;

		for (int i = 0; i < indices.length; i++) {
			BPlusTreeIndex ind = (BPlusTreeIndex)indices[i];
			CounterContainer counterBuf = ind.counterBuf;
			CounterContainer counterUnbuf = ind.counterUnbuf;
//			counterBuffered.flush();
			System.out.println("Index "+i+":\nbuffered: "+new BlockCounter(counterBuf)+"\nunbuffered: "+new BlockCounter(counterUnbuf)+"\nsize: "+ind.counterBuf.size()+"\n");
			allBuf.add(counterBuf);
			allUnbuf.add(counterUnbuf);
			insertsBuf.add(ind.insertsBuf);
			insertsUnbuf.add(ind.insertsUnbuf);
			removesBuf.add(ind.removesBuf);
			removesUnbuf.add(ind.removesUnbuf);
			size += counterBuf.size();
		}

		WaveIndex.DEBUG_OUTPUT.println("\nbuffered: "+allBuf+"\nunbuffered: "+allUnbuf+"\nsize: "+size);
		WaveIndex.DEBUG_OUTPUT.println("\ninserts buffered: "+insertsBuf);
		WaveIndex.DEBUG_OUTPUT.println("inserts unbuffered: "+insertsUnbuf);
		WaveIndex.DEBUG_OUTPUT.println("removes buffered: "+removesBuf);
		WaveIndex.DEBUG_OUTPUT.println("removes unbuffered: "+removesUnbuf);
		if (queryTest != 0) {
			WaveIndex.DEBUG_OUTPUT.println("\nqueries buffered: "+queryCounterBuf);
			WaveIndex.DEBUG_OUTPUT.println("queries unbuffered: "+queryCounterUnbuf);
		}
		WaveIndex.DEBUG_OUTPUT.println("\nTime: "+(t2-t1));

		WaveIndex.DEBUG_OUTPUT.println("\nMax size: "+maxSize + " avg size: "+(avgSize/((count-window)/r)));
//		index.dump();
	}
	
	private static void functionTest() {
		WaveIndex.DEBUG_PRINT_UPDATE_INFO = true;

		int number = 200;
		int window = 20;
		int r = 2;
		int g = 3;
		int e = 3;
		WaveIndex.Strategy strategy = WaveIndex.Strategy.NoRebuild;
		boolean lazy = true;
		TestDataElement.setDataSize(100);

		int bufferSize = 20;
		boolean globalBuffer = false;
		int blockSize = 4096;
		
		String file = path + "elem_" + number + "_" + window + ".txt";
		TestDataHelper.createKeyFile_RandomExpTime(file, number, window);
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
		testWaveIndex(it, window, r, g, e, strategy, lazy, false, bufferSize, globalBuffer, blockSize, 0);
	}

	private static void performanceTest() {
		WaveIndex.DEBUG_PRINT_UPDATE_INFO = true;

		int number = 2000000;
		int window = 500000;
		String file = path + "elem_" + number + "_" + window + ".txt";
//		TestDataHelper.createKeyFile_RandomExpTime(file, number, window);

		int r = 62500;
		int elementSize = 100;
		TestDataElement.setDataSize(elementSize - 20);

		int g = 3;
		int e = 3;
		WaveIndex.Strategy strategy = WaveIndex.Strategy.NoRebuild;
		boolean lazy = false;

		boolean globalBuffer = false;
		int blockSize = 4096;
		int bufferSize = window / 2 / (blockSize / elementSize) / 10;
		
//		TestFileHelper.createKeyFile_RandomExpTime(file, number, window);
		Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
		testWaveIndex(it, window, r, g, e, strategy, lazy, false, bufferSize, globalBuffer, blockSize, 0);
	}
	
	private static PrintStream createFile(String name) {
		try {
			return new PrintStream(new BufferedOutputStream(new FileOutputStream(name)));
		} catch (IOException e) {
			throw new RuntimeException();			
		}
	}
	
	private static int maxSize, avgSize;
	
	private static Function collectSize = new Function() {
		public Object invoke(Object obj) {
			WaveIndex index = (WaveIndex)obj;
			ConstituentIndex[] indices = index.getIndices();
			int size = 0;
			for (int i = 0; i < indices.length; i++)
				size += ((BPlusTreeIndex)indices[i]).counterBuf.size();
			if (size > maxSize)
				maxSize = size;
			avgSize += size;
			return null;
		}
	};
	
	private static void testAll() {
		WaveIndex.DEBUG_PRINT_UPDATE_INFO = false;
		String logfile = "d:\\Benutzer\\log";

		int number = 1500000;
//		int window = 500000;
		int window = 250000;

		boolean fixed = true;
		//int[] update = { 31250, 4000, 500 };
		int[] update = { 31250,15625,10000,2000,1000,250 };
		int[] part = { 2,16 };
		WaveIndex.Strategy[] strategies = {
				WaveIndex.Strategy.NoRebuild,
				WaveIndex.Strategy.Rebuild
		};
		boolean[] lazyArr = { false,true };
		
		int queryTest = 0;
		
		String file = path + "elem_" + number + "_" + window + ".txt";
//		TestDataHelper.createKeyFile_RandomExpTime(file, number, window);
//		TestDataHelper.createKeyFile_FixedInterval(file, number, window);

		int elementSize = 100;
		TestDataElement.setDataSize(elementSize - 20);

		boolean globalBuffer = false;
		int blockSize = 4096;
		int bufferSize = window / (blockSize / elementSize) / 10;
//		int bufferSize = window / 2 / (blockSize / elementSize) / 10;

		if (queryTest != 0)
			snapshotFile = path + "elem_" + number + "_" + window + "_snapshot";

		WaveIndex.DEBUG_OUTPUT.println("Number = "+number+", window = "+window);
		WaveIndex.DEBUG_OUTPUT.println("element size = "+elementSize+", blockSize = "+blockSize+", bufferSize = "+bufferSize);

		int count = 0;
		for (int r : update) {
			WaveIndex.DEBUG_OUTPUT = createFile(logfile + r + ".txt"); 
			for (WaveIndex.Strategy strategy : strategies) {
				for (boolean lazy : lazyArr) {
					for (int g : part) {
						int e = fixed ? 1 : g;

						System.out.println("Test "+count++);
						WaveIndex.DEBUG_OUTPUT.println("\n");
						Cursor<TemporalObject> it = TestDataHelper.keyFileCursor(file);
						try {
							testWaveIndex(it, window, r, g, e, strategy, lazy, fixed, bufferSize, globalBuffer, blockSize, queryTest);
						} catch (Exception ex) {
							WaveIndex.DEBUG_OUTPUT.println(ex);
						}
						WaveIndex.DEBUG_OUTPUT.flush();
					}
				}
			}
			WaveIndex.DEBUG_OUTPUT.close();
		}
	}

	public static void main(String[] args) {
//		performanceTest();
//		functionTest();
		testAll();
	}
}
