/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can 
subscribe to our mailing-list. Send an email to 
    
    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body. 
*/
package xxl.connectivity.geotools;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.geotools.feature.FeatureType;

import xxl.connectivity.jts.Geometry2DAdapter;
import xxl.connectivity.jts.io.Geometry2DConverter;
import xxl.core.collections.containers.Container;
import xxl.core.collections.containers.io.ConverterContainer;
import xxl.core.collections.containers.io.MultiBlockContainer;
import xxl.core.cursors.Cursor;
import xxl.core.io.converters.BooleanConverter;
import xxl.core.io.converters.Converter;
import xxl.core.io.converters.Converters;
import xxl.core.relational.tuples.TupleConverter;
import xxl.core.util.Arrays;
import xxl.core.util.XXLSystem;

/** A <code>FeatureTupleConverter</code> combines several base-type-converters to
 *  seriialize the content of a {@link FeatureTuple}.
 *
 */
public class FeatureTupleConverter extends TupleConverter{
	
	/** Maps a converter to each <code>FeatureType</code> so that some computation can be saved if 
	 *  there already exists a converter for a requested FeatureType.
	 */
	final static protected Map<FeatureType, FeatureTupleConverter> featureTupleConverters = new HashMap<FeatureType, FeatureTupleConverter>();	
	
	
	/** Creates a new <code>FeatureTupleConverter</code> by searching base-type-converters for each
	 *  attribute of the given {@link FeatureTuple}.
	 *  <br><br> The constructor is protected to enforce the use of <code>getConverter(FeatureTuple t)</code> in
	 *  order to get an appropriate converter. 
	 * @param t The tuple whose attribute-information is used to retrieve appropriate base-type-converters 
	 */
	protected FeatureTupleConverter(FeatureTuple t) {
		super(true, Converters.getObjectConverter(BooleanConverter.DEFAULT_INSTANCE));
		
		ArrayList<Converter<Object>> converters = new ArrayList<Converter<Object>>(t.getColumnCount());
		
		for(int i=0; i< t.getColumnCount(); i++){
			 if( Geometry2DAdapter.class.isAssignableFrom(t.getAttributeType(i+1)) )					 
				converters.add(i, Converters.getObjectConverter(Geometry2DConverter.DEFAULT_INSTANCE));
			 else converters.add(i, Converters.getConverterForJavaType( t.getAttributeType(i+1).getCanonicalName()));
		}
		this.converters = converters;
	}	
	
	
	/** Returns a converter-object for the given feature. First a map is queried for an appropriate
	 *  converter and a new converter is created only if the map doesn't already contain one for the given
	 *  feature-type.  
	 * @param featureTuple The tuple whose attribute-information is used to retrieve appropriate base-type-converters
	 * @return the converter-object for the given <code>FeatureTuple</code>
	 */
	public static FeatureTupleConverter getConverter(FeatureTuple featureTuple){
		FeatureTupleConverter converter = featureTupleConverters.get(featureTuple.featureType); 
		if( converter == null){
			converter = new FeatureTupleConverter(featureTuple);
			featureTupleConverters.put(featureTuple.featureType, converter);
		}			
		return converter;
	}
	
	
	/** This main-method contains a usage example for the serialization of FeatureTuples. First
	 *  a ShapeFile is read, before the <code>FeatureTuples</code> contained in this file are written to a {@link MultiBlockContainer}.
	 * @param args you can specify a Shape-file to read by giving the name of the *.shp-file
	 *             If no parameter is given, the program reads a default-file.
	 * @throws IOException if the file cannot be read or the container cannot be written
	 */
	public static void main(String [] args) throws IOException{

		File input = new File( args.length> 0 
				? args[0]
				: XXLSystem.getDataPath(new String[]{"geo","shp","houston_rails.shp"})
			);		

		// ffnet Shapefile und speichert die Features als Tupel in einem Container
		Cursor<FeatureTuple> cursor = new ShapefileFeatureImport(input).getFeatureTuples();			
			
		FeatureTupleConverter converter = FeatureTupleConverter.getConverter( cursor.peek() );
		String[] attributeNames = cursor.peek().getAttributeNames();
		
		List<Long> featureIds = new ArrayList<Long>();
		MultiBlockContainer mContainer = new MultiBlockContainer(XXLSystem.getOutPath()+"/FeatureTupleConverterTest", 32); 
		Container container = new ConverterContainer(
					mContainer,
					converter
				);
		
		while(cursor.hasNext())
			featureIds.add( (Long) container.insert(cursor.next()) );
		
		Arrays.print(attributeNames, System.out);
		System.out.println();
		
		for(Long id : featureIds){
			Object t = container.get(id);
			System.out.println(t);
		}
		
		mContainer.delete();
	}
}