/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can 
subscribe to our mailing-list. Send an email to 
    
    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body. 
*/
package xxl.connectivity.geotools;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;

import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.feature.Feature;

import xxl.connectivity.jts.Geometry2DFactory;
import xxl.connectivity.jts.visual.VisualGeometry2DCursor;
import xxl.connectivity.jts.visual.VisualOutput;
import xxl.core.cursors.Cursor;
import xxl.core.cursors.Cursors;
import xxl.core.cursors.mappers.Mapper;
import xxl.core.functions.Function;
import xxl.core.spatial.geometries.Geometry2D;
import xxl.core.spatial.rectangles.Rectangle;
import xxl.core.util.WrappingRuntimeException;
import xxl.core.util.XXLSystem;

import com.vividsolutions.jts.geom.Envelope;

/** This class provides an import-function for ESRI-Shape-files, a common GIS-file-format.
 */
public class ShapefileFeatureImport{

	private ShapefileDataStore dataStore = null;

	/** Creates a new {@link ShapefileFeatureImport}-object and opens the 
	 *  Shapefile-dataset specified by the given name of the *.shp- file.
	 * @param shapeFile the name of the *.shp-file to read
	 */
	public ShapefileFeatureImport(File shapeFile){
		try {
			dataStore = new ShapefileDataStore(shapeFile.toURL());
		} catch (MalformedURLException e) {
			throw new WrappingRuntimeException(e);
		}				
	}

	/** Returns the bounding rectangle of the feature-set 
	 * @return the bounding rectangle of the feature-set 
	 */
	public Rectangle getUniverse(){
		Envelope e;
		try {
			e = dataStore.getFeatureSource().getBounds();
			if(e== null)
				e = dataStore.getFeatureSource().getFeatures().getBounds();
			
			return Geometry2DFactory.envelopeToXXLRectangle(e);
		} catch (IOException e1) {
			throw new WrappingRuntimeException(e1);
		}		
	}

	/** Returns a cursor over the features contained in the shape-file.
	 * @return a cursor over the features contained in the shape-file 
	 */
	public Cursor<FeatureTuple> getFeatureTuples(){
		try {
			return 	new Mapper<Feature, FeatureTuple>(
						new Function<Feature, FeatureTuple>(){
							public FeatureTuple invoke(Feature f){
								return new FeatureTuple(f);
							}
						},				
						dataStore.getFeatureSource().getFeatures().iterator() 
				);
		} catch (IOException e) {
			throw new WrappingRuntimeException(e);
		}
	}
	
	
	
	
	/** A use case to demonstrate how to open a Shape-file-dataset. 
	 * @param args you can specify a Shape-file to read by giving the name of the *.shp-file
	 *             If no parameter is given, the program reads a default-file.
	 * @throws IOException if ther occurs any error reading the file
	 */
	public static void main(String [] args) throws IOException{

		File input = new File( args.length> 0 
									? args[0]
									: XXLSystem.getDataPath(new String[]{"geo","shp","houston_roads.shp"})
								);		
		
		ShapefileFeatureImport sfi = new ShapefileFeatureImport(input);
								
		
		VisualOutput out = new VisualOutput(input.getName(), sfi.getUniverse(),700);
		
		System.out.println(	"Features contained in this shapefile: " +			
			Cursors.count(
				new VisualGeometry2DCursor(
					new Mapper<FeatureTuple, Geometry2D>(
							new Function<FeatureTuple, Geometry2D>(){
								public Geometry2D invoke(FeatureTuple f){
									return f.getDefaultGeometry();
								}							
							},
							sfi.getFeatureTuples()
					),  
					out,
					Color.red
				)
			)
		);
		out.repaint();
		System.out.println("MBR: "+sfi.getUniverse());		
	}	
}
