/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can 
subscribe to our mailing-list. Send an email to 
    
    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body. 
*/
package xxl.connectivity.geotools;

import java.awt.Color;
import java.io.File;
import java.io.IOException;

import org.geotools.data.tiger.TigerDataStore;
import org.geotools.feature.Feature;

import xxl.connectivity.jts.Geometry2DFactory;
import xxl.connectivity.jts.visual.VisualGeometry2DCursor;
import xxl.connectivity.jts.visual.VisualOutput;
import xxl.core.cursors.Cursor;
import xxl.core.cursors.Cursors;
import xxl.core.cursors.mappers.Mapper;
import xxl.core.functions.Function;
import xxl.core.spatial.geometries.Geometry2D;
import xxl.core.spatial.rectangles.Rectangle;
import xxl.core.util.WrappingRuntimeException;
import xxl.core.util.XXLSystem;

import com.vividsolutions.jts.geom.Envelope;

/** A class to read Tiger/Line-files:   <br>
 *  Each Tiger/Line-dataset contains one or more feature-sets, e.g. roads, railways,... .
 *  You can access the names of the feature-sets via {@link TigerFeatureImport#getFeatureNames()}
 *  and retrieve the features of these sets by {@link TigerFeatureImport#getFeatureTuples(String)}, which
 *  returns a cursor of {@link FeatureTuple}-objects.
 *
 */
public class TigerFeatureImport{
	
	private TigerDataStore dataStore = null;
	
	
	/** Creates a new {@link TigerFeatureImport}-object and opens the 
	 *  Tiger/Line-dataset specified by one of the Tiger/Line-files 
	 * @param tigerFile one of the files of the Tiger/Line-dataset to read
	 */
	public TigerFeatureImport(File tigerFile){	
		dataStore = new TigerDataStore(tigerFile.getParent());		
	}
	
	
	/** Returns the names of the Feature-sets contained in the current dataset
	 * @return the names of the Feature-sets contained in the current dataset
	 */
	public String[] getFeatureNames(){
		return dataStore.getTypeNames();
	}
	
	
	/** Returns the bounding rectangle of the feature-set specified by the given name
	 * @param featureName the name of the feature-set of which the bounding rectangle is requested
	 * @return the bounding rectangle of the feature-set specified by the given name
	 */
	public Rectangle getUniverse(String featureName){		
		Envelope e = null; 
		try {	
			e= dataStore.getFeatureSource(featureName).getBounds();		
			if(e== null)
				e = dataStore.getFeatureSource(featureName).getFeatures().getBounds();
		
			return Geometry2DFactory.envelopeToXXLRectangle(e);
		} catch (IOException e1) {
			throw new WrappingRuntimeException(e1);
		}	
	}
	
	
	/** Returns a cursor over the features contained in the feature-set specified by the given name.
	 * @param featureName the name of the feature-set of which the elements are requested
	 * @return a cursor over the features contained in the feature-set specified by the given name. 
	 */
	public Cursor<FeatureTuple> getFeatureTuples(String featureName){
		try {
			return	new Mapper<Feature, FeatureTuple>(
				new Function<Feature, FeatureTuple>(){
					public FeatureTuple invoke(Feature f){
						return new FeatureTuple(f);
					}
				},				
				dataStore.getFeatureSource(featureName).getFeatures().iterator() 
			);
		} catch (IOException e) {
			throw new WrappingRuntimeException(e);
		}
	}
		
	
	/** A use case to demonstrate how to open a Tiger-Line-dataset. 
	 * @param args you can specify a Tiger/Line-dataset to read by giving the name of one of
	 *        the tiger-files. If no parameter is given, the program reads a default-dataset.
	 * @throws IOException if there occurs any error opening the data-set 
	 */
	public static void main(String [] args) throws IOException{

		File input = new File( args.length> 0 
									? args[0]
									: XXLSystem.getDataPath(new String[]{"geo","tgr","la_tgr06001.rt1"})
								);		
		
		TigerFeatureImport tfi = new TigerFeatureImport(input);
		
		String feature = tfi.getFeatureNames()[0];					
		
		VisualOutput out = new VisualOutput(feature, tfi.getUniverse(feature),700);
		
		System.out.println(	"Features contained in this Tiger/Line feature-set: " +			
			Cursors.count(
				new VisualGeometry2DCursor(
					new Mapper<FeatureTuple, Geometry2D>(
							new Function<FeatureTuple, Geometry2D>(){
								public Geometry2D invoke(FeatureTuple f){
									return f.getDefaultGeometry();
								}							
							},
							tfi.getFeatureTuples(feature)
					),  
					out,
					Color.red
				)
			)
		);
		out.repaint();
		System.out.println("MBR: "+tfi.getUniverse(feature));		
	}	
}