/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can 
subscribe to our mailing-list. Send an email to 
    
    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body. 
*/
package xxl.connectivity.jts;

import java.awt.Shape;
import java.awt.geom.NoninvertibleTransformException;

import xxl.core.spatial.geometries.Geometry2D;
import xxl.core.spatial.geometries.Geometry2DException;
import xxl.core.spatial.rectangles.Rectangle;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.PrecisionModel;
import com.vividsolutions.jts.geom.TopologyException;
import com.vividsolutions.jts.io.WKBWriter;
import com.vividsolutions.jts.io.WKTWriter;
import com.vividsolutions.jts.operation.buffer.BufferOp;
import com.vividsolutions.jts.precision.EnhancedPrecisionOp;
import com.vividsolutions.jump.workbench.ui.renderer.java2D.Java2DConverter;

/** This is the base class of the wrapped JTS geometry model in XXL. It implements
 *  the interface {@link Geometry2D} which is the base interface of the abstract XXL 
 *  geometry model. 
 *  This implementation provides an <a href="http://www.opengis.org/">OGC</a> 
 *  <i>Simple Feature Specification</i> compliant geometry model based on the model 
 *  of the Java Topology Suite (<a href="http://jump-project.org">http://jump-project.org</a>).
 *  Objects of the JTS model are wrapped inside <code>Geometry2DAdapter</code> objects and 
 *  most function calls are delegated to those wrapped objects wherever possible or usefull.
 *  <br><br>
 *  <code>Geometry2DAdapter</code> does not only implement the required methods of the {@link Geometry2D} 
 *  interface, but already implements some methods of its subinterfaces, wherever the underlying
 *  JTS <code>Geometry</code> provides the required functionality. Because of that some
 *  subclasses of <code>Geometry2DAdapter</code> form empty classes which are only needed to 
 *  represent the complete model.
 *    
 * @see Geometry2DFactory
 * @see Geometry2D
 * @see Geometry
 */
public abstract class Geometry2DAdapter implements Geometry2D{        
	
 ////////////////////
 // static fields //
 ////////////////////	
	
	private static final WKBWriter WKB_WRITER_2D = new WKBWriter(2);
	private static final WKBWriter WKB_WRITER_3D = new WKBWriter(3);
	private static final WKTWriter WKT_WRITER_2D = new WKTWriter();	

 ////////////////////
 //  member fields //
 ////////////////////	
	
	/** the encpasulatet JTS-<code>Geometry</code> */
	protected final Geometry geometry; 	
	
 ////////////////////
 //  constructor   //
 ////////////////////		
   
    /** Creates a new two-dimensional <code>Geometry2DAdapter</code>- object specified by the given JTS-<code>Geometry</code>
     *  The constructor can only be called from within this class or its subclasses. Objects of these classes will only be instantiated by calling the
     *  static functions {@link Geometry2DFactory#wrap(Geometry)} or by using a {@link Geometry2DFactory} function.
     * @param geometry the JTS-<code>Geometry</code> wrapped inside this object
     */
    protected Geometry2DAdapter(Geometry geometry){          
        if(geometry == null)
        	throw new Geometry2DException("Cannot handle null value!");        
        this.geometry = geometry;                
    }
    
 //////////////
 // methods  //
 //////////////		

    /** Returns the wrapped JTS-<code>Geometry</code> of this object 
     * @return the wrapped JTS-<code>Geometry</code>
     */
    public Geometry getJTSGeometry(){ 
    	return (Geometry) geometry.clone();
    }
     
    
    /** Creates and returns a physical copy of this <code>Geometry2DAdapter</code> object.
	 * @return a copy of this <code>Geometry2DAdapter</code>
	*/    	
	public Geometry2DAdapter clone() { 
		return Geometry2DFactory.wrap((Geometry) geometry.clone()); 
	}

	
	/**
	 *  Returns whether this <code>Geometry2DAdapter</code> is greater than, equal to,
	 *  or less than another <code>Geometry2DAdapter</code>. <P>
	 *
	 *  If their classes are different, they are compared using the following
	 *  ordering:
	 *  <UL>
	 *    <LI> Point2DAdapter (lowest)
	 *    <LI> MultiPoint2DAdapter
	 *    <LI> LineString2DAdapter
	 *    <LI> LinearRing2DAdapter
	 *    <LI> MultiLineString2DAdapter
	 *    <LI> Polygon2DAdapter
	 *    <LI> MultiPolygon2DAdapter
	 *    <LI> GeometryCollection2DAdapter (highest)
	 *  </UL>
	 *  If the two <code>Geometry2DAdapter</code>s have the same class, their first
	 *  elements are compared. If those are the same, the second elements are
	 *  compared, etc.
	 *
	 *@param  other  a <code>Geometry2DAdapter</code> with which to compare this <code>Geometry</code>
	 *@return a positive number, 0, or a negative number, depending on whether
	 *      this object is greater than, equal to, or less than <code>other</code>
	 */
    public int compareTo(Geometry2D other) { 
    	return(geometry.compareTo(((Geometry2DAdapter)other).geometry)); 
    }
    
    
  /* spatial predicates */
    
    /** Returns <code>true</code> if this geometry is equal to the
     *  specified geometry.
     *  <p>
     *  The <code>equals</code> predicate has the following equivalent definitions:
     *  <ul>
     *  <li>The two geometries have at least one point in common,
     *  and no point of either geometry lies in the exterior of the other geometry.
     *  <li>The DE-9IM Intersection Matrix for the two geometries is T*F**FFF*
     *  </ul>
     *
     *  @param  other  the <code>Geometry2DAdapter</code> with which to compare this <code>Geometry2DAdapter</code>
     *  @return <tt>true</tt> if the given objects equals this geometry, <tt>false</tt> otherwise 
	 */
	public boolean equals(Geometry2D other){ 
		return geometry.equals( ((Geometry2DAdapter)other).geometry ); 
	}

	/** @inheritDoc
	 * <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
     *  @see Geometry#contains(Geometry)
     */
    public boolean contains(Geometry2D other) {    	
    	return geometry.contains( ((Geometry2DAdapter)other).geometry );
    }
    
    /** @inheritDoc
 	 * <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
 	 * @see Geometry#within(Geometry)
	 */
	public boolean isWithin(Geometry2D other) { 
		return geometry.within(((Geometry2DAdapter)other).geometry); 
	}

	/** @inheritDoc
	 * <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
     *  @see Geometry#contains(Geometry)
     */
    public boolean covers(Geometry2D other){
    	return geometry.covers(((Geometry2DAdapter)other).geometry);
    }
        
    /** @inheritDoc
     *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
     *  @see Geometry#coveredBy(Geometry)
     */
    public boolean isCoveredBy(Geometry2D other){
    	return geometry.coveredBy(((Geometry2DAdapter)other).geometry);
    }
    
    /** @inheritDoc
     *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
     *  @see Geometry#overlaps(Geometry)
     */
	public boolean overlaps(Geometry2D other) { 
		return geometry.overlaps(((Geometry2DAdapter)other).geometry); 
	}

    /** @inheritDoc
     *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
     *  @see Geometry#intersects(Geometry)
     */
	public boolean intersects(Geometry2D other){ 
		return geometry.intersects(((Geometry2DAdapter)other).geometry); 
	}

    /** @inheritDoc
     *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 * @see Geometry#crosses(Geometry)
	 */
	public boolean crosses(Geometry2D other) { 
		return geometry.crosses(((Geometry2DAdapter)other).geometry); 	
	}
               
    /** @inheritDoc
     *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 * @see Geometry#touches(Geometry)
	 */
    public boolean touches(Geometry2D other) { 
    	return geometry.touches(((Geometry2DAdapter)other).geometry); 
    }

    /** @inheritDoc
     *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 * @see Geometry#disjoint(Geometry)
	 */
	public boolean isDisjoint(Geometry2D other) { 
		return geometry.disjoint(((Geometry2DAdapter)other).geometry); 
	}
    
    /** @inheritDoc
     *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 * @see Geometry#relate(Geometry)
	 */    
	public boolean relate(Geometry2D other, String intersectionPatternMatrix) {
		return geometry.relate(((Geometry2DAdapter)other).geometry, intersectionPatternMatrix); 				
	}  
	
	/** @inheritDoc
	 *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
     * @see Geometry#distance(Geometry)
     */
    public double distance(Geometry2D other){ 
    	return geometry.distance(((Geometry2DAdapter)other).geometry ); 
    }
  
	/** Returns the area of this <code> Geometry2DAdapter</code>-object. 
	 *  The area of a Point is 0, the area of a LineString equals its length.
	 *  For GeometryCollections this method returns the sum of all contained
	 *  Geometries and for Polygons the actual area is returned.
	 * @return the area of this <code> Geometry2DAdapter</code>
     * @see Geometry#getArea()
     */
    public double getArea() { return geometry.getArea(); } 

    /** @inheritDoc 
     * @see Geometry#getBoundary()
     */
    public Geometry2DAdapter getBoundary() { 
    	return Geometry2DFactory.wrap(geometry.getBoundary()); 
    } 

    /** Returns the dimension of this objects' boundary 
     * @return the dimension of this objects' boundary
     * @see Geometry#getBoundaryDimension()
     */
    public int getBoundaryDimension() { 
    	return geometry.getBoundaryDimension(); 
    }
    
    /** Returns the centroid of this <code>Geometry2DAdapter</code>.
     *  <b>Note:</b> This method is required for Surfaces only but can be
     *  extended to arbitrary geometries   
     * @return the centroid of this <code>Geometry2DAdapter</code>
     * @see Geometry#getCentroid()
     */
    public Point2DAdapter getCentroid(){ 
    	return new Point2DAdapter( geometry.getCentroid()); 
    } 
    
    /** Returns a Point of this <code>Geometry2DAdapter</code> which is guaranteed to lie
     *  on the geometries interior.
     *  <b>Note:</b> This method is required for Surfaces only but can be
     *  extended to arbitrary geometries   
     * @return a Point of this <code>Geometry2DAdapter</code> which is guaranteed to lie on the geometries interior
      * @see Geometry#getInteriorPoint()
     */
    public Point2DAdapter getInteriorPoint(){ 
    	return new Point2DAdapter( geometry.getInteriorPoint()); 
    }   
    
    /** @inheritDoc 
     * @see Geometry#getDimension()
     */    
    public int getDimensions() { 
    	return geometry.getDimension(); 
    }    
    
    /** Returns the type of this <code>Geometry2DAdapter</code> as a String
     * @return any of <code>POINT, MULTIPOINT, LINESTRING, MULTILINESTRING, 
     * 		   POLYGON</code> or <code>MULTIPOLYGON</code>
     * @see Geometry#getGeometryType()
     */
    public String getGeometryType(){ 
    	return geometry.getGeometryType(); 
    }

	/** Returns the minimal bounding rectangle (MBR) of the encapsulated geometry 
     * @return the minimal bounding rectangle of the encapsulated geometry using the geometrys precision model
     */
	public Geometry2DAdapter getEnvelope() { 
		return Geometry2DFactory.envelopeToPolygon(
				geometry.getEnvelopeInternal(), geometry.getPrecisionModel()
			); 
	}
	
	/** Returns the minimal bounding rectangle (MBR) of the encapsulated geometry as a XXL- <code>DoublePointRectangle</code>.
     * @return the minimal bounding rectangle of the encapsulated geometry using the geometrys precision model
     */
    public Rectangle getMBR(){
    	return getMBR(true);      		    	
    }
	    
	/** Returns the minimal bounding rectangle (MBR) of the encapsulated geometry. The MBR is returned as a DoublePointRectangle
	 *  or as a FloatPointRectangle, depending on the parameter <code>returnDoublePointRectangle</code>
	 * @param returnDoublePointRectangle indicates, whether to return the mbr as a {@link xxl.core.spatial.rectangles.DoublePointRectangle DoublePointRectangle} 
     * 						  			(<tt>true</tt>) or as a {@link xxl.core.spatial.rectangles.FloatPointRectangle FloatPointRectangle} (<tt>false</tt>).
     * @return the minimal bounding rectangle of the encapsulated geometry
     */
    public Rectangle getMBR(boolean returnDoublePointRectangle){
    	return Geometry2DFactory.envelopeToXXLRectangle(
    				geometry.getEnvelopeInternal(),
    				returnDoublePointRectangle
    			);           
    }
	     
    /** Returns the number of this objects' vertices
     *  @return the number of vertices      
     */
    public int getNumberOfPoints(){    	
    	return geometry.getNumPoints(); 
    }
	    
    /** Returns the length of linear geometries and the perimeter of
     *  areal geometries respectively.
     * @return the length of linear geometries and the perimeter of areal geometries respectively
     * @see Geometry#getLength()
     */
    public double getLength() { 
    	return geometry.getLength(); 
    }
        
    /** @inheritDoc 
     * @see Geometry#getSRID()
     */
	public int getSRID() { 
		return geometry.getSRID(); 
	}
    
	 /** Sets the ID of the Spatial Reference System used by this geometry 
	 * @param SRID the ID of the Spatial Reference System to set
     * @see Geometry#setSRID(int)
     */
	public void setSRID(int SRID) { 
		geometry.setSRID(SRID); 
	}
	
    /** @inheritDoc
     *  @see Geometry#isEmpty()
	 */
	public boolean isEmpty(){ 
		return geometry== null || geometry.isEmpty(); 
	}

	/** determines this <code>Geometry2DAdapter</code>'s precision model
     * @return the PrecisionModel of the underlying JTS- Geometry
     */
    public PrecisionModel getPrecisionModel(){
    	return geometry.getPrecisionModel();
    }    
        
    /** @inheritDoc
     *  @see Geometry#isSimple()
     */
    public boolean isSimple(){ 
    	return geometry.isSimple(); 
    }
        
    /** Tests the validity of this Geometry2DAdapter.
     * @return <tt>true</tt> if the Geometry2DAdapter object is valid, <tt>false</tt> otherwise
     * @see Geometry#isValid()
     */
    public boolean isValid(){ 
    	return geometry.isValid(); 
    }  
    
    /** @inheritDoc 
     *  @see Geometry#buffer(double)
	 */
	public Geometry2DAdapter buffer(double width){
	    return Geometry2DFactory.wrap( geometry.buffer(width)); 
	}

    /** Computes a buffer around this <code>Geometry2DAdapter</code> having the given width and approximation accuracy.  
     * 
     *  If the computed buffer contains circular arcs, these are approximated by the given number of line-segments.
     *   
     *  @param width the maximum distance between a point inside the buffer and this <code>Geometry2DAdapter</code> 
     *  @param quadrantSegments the number of line-segments to approximate circular buffer shapes 
     *  @return a new <code>Geometry2DAdapter</code>-object representing the buffer
     *  @see Geometry#buffer(double, int)
	 */
	public Geometry2DAdapter buffer(double width, int quadrantSegments){
	    return Geometry2DFactory.wrap( geometry.buffer(width, quadrantSegments)); 
	}
	
	/** The end cap style specifies the buffer geometry that will be created at the ends of linestrings. The styles provided are: */
	public static enum EndCapStyle {
		/**	a semi circle (default) */
		ROUND, 
		/**	a straight line perpendicular to the end segment */
		BUTT, 
		/**	a half-square */
		SQUARE
	};
	
    /** Computes a buffer around this <code>Geometry2DAdapter</code> having the given width and approximation accuracy and
     *  using the given end cap style.
     * 
     *  If the computed buffer contains circular arcs, these are approximated by the given number of line-segments.
     * 
     *  @param width the maximum distance between a point inside the buffer and this <code>Geometry2DAdapter</code> 
     *  @param quadrantSegments the number of line-segments to approximate circular buffer shapes 
     *  @param endCapStyle the end cap style to use {@link EndCapStyle}
     *  @return a new <code>Geometry2DAdapter</code>-object representing the buffer
     *  @see Geometry#buffer(double, int)
	 */
	public Geometry2DAdapter buffer(double width, int quadrantSegments, EndCapStyle endCapStyle){
		switch(endCapStyle){
			case BUTT 	: return Geometry2DFactory.wrap(geometry.buffer(width, quadrantSegments, BufferOp.CAP_BUTT));
			case SQUARE : return Geometry2DFactory.wrap(geometry.buffer(width, quadrantSegments, BufferOp.CAP_SQUARE));
			default   	: return Geometry2DFactory.wrap(geometry.buffer(width, quadrantSegments, BufferOp.CAP_ROUND));
		}		
	}

	/** @inheritDoc
	 *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 *  <br><br>
	 *  This implementation uses an <code>EnhancedPrecisionOp</code> which uses 
	 *  enhanced precision techniques, if the call to {@link Geometry#difference(Geometry)}
	 *  causes a TopologyException due to robusteness errors.
	 *  @throws TopologyException is thrown, if the <code>EnhancedPrecisionOp</code> encounters a robustness error too.  
	 *  @see EnhancedPrecisionOp#difference(Geometry, Geometry)
	 *  @see TopologyException
	 */
	public Geometry2DAdapter difference(Geometry2D other){
		// the EnhancedPrecisionOp decides whether to use simple or advanced 
		// but more expensive algorithms for the computation		
        return Geometry2DFactory.wrap( EnhancedPrecisionOp.difference(geometry, ((Geometry2DAdapter)other).geometry));  
	}

	/** @inheritDoc
	 *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 *  <br><br>
	 *  This implementation uses an <code>EnhancedPrecisionOp</code> which uses 
	 *  enhanced precision techniques, if the call to {@link Geometry#intersection(Geometry)}
	 *  causes a TopologyException due to robusteness errors.
	 *  @throws TopologyException is thrown, if the <code>EnhancedPrecisionOp</code> encounters a robustness error too.  
	 *  @see EnhancedPrecisionOp#intersection(Geometry, Geometry)
	 *  @see TopologyException
	 */
	public Geometry2DAdapter intersection(Geometry2D other) {
        return Geometry2DFactory.wrap( EnhancedPrecisionOp.intersection(geometry, ((Geometry2DAdapter)other).geometry));  
	}
    
    
	/** @inheritDoc
	 *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 *  <br><br>
	 *  This implementation uses an <code>EnhancedPrecisionOp</code> which uses 
	 *  enhanced precision techniques, if the call to {@link Geometry#union(Geometry)}
	 *  causes a TopologyException due to robusteness errors.
	 *  @throws TopologyException is thrown, if the <code>EnhancedPrecisionOp</code> encounters a robustness error too.  
	 *  @see EnhancedPrecisionOp#union(Geometry, Geometry)
	 *  @see TopologyException
	 */
	public Geometry2D union(Geometry2D other) {
        return Geometry2DFactory.wrap( EnhancedPrecisionOp.union(geometry, ((Geometry2DAdapter)other).geometry));  
	}

	/** @inheritDoc
	 *  <code>other</code> is expected to be of type <code>Geometry2DAdapter</code>
	 *  <br><br>
	 *  This implementation uses an <code>EnhancedPrecisionOp</code> which uses 
	 *  enhanced precision techniques, if the call to {@link Geometry#symDifference(Geometry)}
	 *  causes a TopologyException due to robusteness errors.
	 *  @throws TopologyException is thrown, if the <code>EnhancedPrecisionOp</code> encounters a robustness error too.  
	 *  @see EnhancedPrecisionOp#symDifference(Geometry, Geometry)
	 *  @see TopologyException
	 */
	public Geometry2D symDifference(Geometry2D other){
        return Geometry2DFactory.wrap( EnhancedPrecisionOp.symDifference(geometry, ((Geometry2DAdapter)other).geometry));  
	}

	/** @inheritDoc
	 *  @see Geometry#convexHull() 
	 */	
	public Geometry2D convexHull() {
		return Geometry2DFactory.wrap( geometry.convexHull());    	
	}

			
	/** Returns a string representation of this geometry.
	 * @return returns the string representation of this geometry
	*/	
	public String toString(){
//	    StringBuffer bf = new StringBuffer("\n"+this.getClass().getSimpleName()+" [ \n\tCOORDINATES={ ");
//	    
//	    Coordinate[] c = geometry.getCoordinates();
//	    for(int i=0;i< c.length; i++) bf.append("("+c[i].x+", "+c[i].y+"),");        
//	    bf.setCharAt(bf.length()-1,'}');
//	    bf.append("\n\tTYPE="+geometry.getGeometryType());
//	    bf.append("\n\tPRECISION="+getPrecisionModel());        
//	    bf.append("\n\tSRID="+getSRID());
//	    bf.append("\n\tMBR="+geometry.getEnvelopeInternal());
//	    bf.append("\n\t]");
//	    return bf.toString();
		return toWKT();
	}
	
	/** @inheritDoc
	 *  @see WKBWriter#write(Geometry)
     */
    public byte[] toWKB() {	
    	Coordinate c = this.geometry.getInteriorPoint().getCoordinate();    	
		return Double.isNaN(c.z) 
				? WKB_WRITER_2D.write(geometry) 
				: WKB_WRITER_3D.write(geometry);
	}


	/** @inheritDoc
	 *  @see WKTWriter#write(Geometry)
     */
    public String toWKT(){ 
    	return WKT_WRITER_2D.write(geometry); 
    }	

	/** Generates a {@link Shape}-representation of this <code>Geometry2DAdapter</code>
	 * 
	 * @param java2DConverter needed to compute the shape of the given geometrie
	 * @return the {@Link Shape}-object representing this geometries shape 
	 * @throws NoninvertibleTransformException
	 */
	public Shape toShape(Java2DConverter java2DConverter) throws NoninvertibleTransformException {
		return java2DConverter.toShape(geometry);
	}


    /** USE CASE: Runs various tests on Geometry2DAdapter- objects like 
     * 	evaluating spatial predicates and distance operators, conversion from xxl to jts 
     *  and a robustness test.  
	 * @param args Use Case Parameter: obsolete 
     * @throws com.vividsolutions.jts.io.ParseException 
	 */   
    public static void main(String[] args) throws com.vividsolutions.jts.io.ParseException{
        
        Geometry2DAdapter g1= null, g2=null, g3=null;
        
        System.out.println("\n\nCreate some Geometry2DAdapter objects" +
        					 "\n=======================================");
        
        g1 = Geometry2DFactory.createFromWKT("POLYGON ((0.6 0.4, 10.3 10.1, 10.6 0, 0.6 0.4))", new PrecisionModel(1));
        g2 = Geometry2DFactory.createFromWKT("POLYGON ((5 0, 5 5, 15 5, 15 0, 5 0))");                    
        g3 = Geometry2DFactory.createRectangle(Geometry2DFactory.createPoint(5,0),15,-5);
        System.out.println( "\n  g1= "+g1+
        					"\n  g2= "+g2+
        					"\n  g3= "+g3);
                
        
        System.out.println("\n\nTesting spatial predicates" +
		 					 "\n===========================\n"+
		 					 "\n g1.overlaps(g2)= "+g1.overlaps(g2)+                        
        					 "\n g1.distance(g2)= "+g1.distance(g2)+
        					 "\n g2.equals(g3)  = "+g2.equals(g3));
        g3.union(g1);
        System.out.println(  "\n g3.union(g1)= "+g3+        
        					 "\n g3.contains(g2)= "+g3.contains(g2));
        
        
        System.out.println("\n\nTesting graphical output" +
	 			 "\n===========================\n" +
				 "\n red   -> g1 "+                        
				 "\n green -> g2 "+
				 "\n blue  -> g3 "+
				 "\n gray  -> g1.intersection(g2)");

		xxl.connectivity.jts.visual.VisualOutput output = new xxl.connectivity.jts.visual.VisualOutput("Test::Geometry2DAdapter", g1.union(g2).union(g3).getMBR(), 400);
		output.fill(g1.intersection(g2), java.awt.Color.lightGray);
		output.draw(g1, java.awt.Color.red);
		output.draw(g2, java.awt.Color.green);
		output.draw(g3, java.awt.Color.blue);
		output.repaint();
		
        System.out.println("\n\nTesting conversion from XXL to Geometry2DAdapter" +
        					 "\n==================================================\n");
        
        Rectangle d1 = new xxl.core.spatial.rectangles.DoublePointRectangle(new double[]{0,0},new double[]{2,2});
        Rectangle d2 = new xxl.core.spatial.rectangles.FloatPointRectangle(new float[]{5,5},new float[]{10,10});
        System.out.println("  d1= "+d1+"\n  d2= "+d2);
        
        Geometry2DAdapter x1 = null, x2 = null;

        x1 = Geometry2DFactory.xxlRectangleToPolygon2D(d1);
        x2 = Geometry2DFactory.xxlRectangleToPolygon2D(d2);
        	        
        System.out.println("  d1 ~> x1 : "+ x1 +
        		 		 "\n  d2 ~> x2 : "+ x2 +
        				 "\nForce Geometry2DException:");
        
        Geometry2DAdapter x3 = null;        
        try{
            Rectangle d3 = new xxl.core.spatial.rectangles.DoublePointRectangle(new double[]{0,},new double[]{1});
            System.out.println("\n  d3= "+d3+"\n  d3 ~> x3 :");
            x3= Geometry2DFactory.xxlRectangleToPolygon2D(d3);
            System.out.println(x3);
        } catch(Geometry2DException e){ e.printStackTrace(System.out);} 
        
        System.out.println("\n\nTesting Distance- Operator" +
        		             "\n===========================\n"+
        		             "\n x1.distance(g3)= "+x1.distance(g3)+",  closest Points:"+java.util.Arrays.asList(ExtendedDistanceOp.closestPoints(x1,g3))+
        					 "\n x1.maxDistance(g3)= "+ExtendedDistanceOp.maxDistance(x1,g3)+", furthest Points:"+java.util.Arrays.asList(ExtendedDistanceOp.furthestPoints(x1,g3)));
        	        

        String wkt1 = "POLYGON ((708653.498611049 2402311.54647056, 708708.895756966 2402203.47250014, 708280.326454234 2402089.6337791, 708247.896591321 2402252.48269854, 708367.379593851 2402324.00761653, 708248.882609455 2402253.07294874, 708249.523621829 2402244.3124463, 708261.854734465 2402182.39086576, 708262.818392579 2402183.35452387, 708653.498611049 2402311.54647056))";
        String wkt2 = "POLYGON ((708258.754920656 2402197.91172757, 708257.029447455 2402206.56901508, 708652.961095455 2402312.65463437, 708657.068786251 2402304.6356364, 708258.754920656 2402197.91172757))";
        
        Geometry2DAdapter wg1 = Geometry2DFactory.createFromWKT(wkt1);
	    Geometry2DAdapter wg2 = Geometry2DFactory.createFromWKT(wkt2);
	    
        System.out.println("\n\nTesting Robustness" +
        		             "\n=====================\n"+
        		             "\nThe intersection of these two polygons would cause a TopologyException, if the EnhancedPrecisionOp wasn't used!+" +
        		             "\n Polygon 1: "+wg1+
        		             "\n Polygon 2: "+wg2+
        		             "\n Intersection: "+wg1.intersection(wg2));
	    
        Geometry2DAdapter empty = Geometry2DFactory.createGeometryCollection2D(null);
        System.out.println("\n\nTesting empty geometry semantics" +
	    		             "\n=================================\n"+
	    		             "\n empty = "+empty+
	    		             "\n empty.intersects(g3)="+empty.intersects(g3));
        
   }

}
