/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.indexStructures;

import java.io.PrintStream;
import java.util.Iterator;

import xxl.core.cursors.Cursor;
import xxl.core.cursors.sources.EmptyCursor;
import xxl.core.cursors.unions.Sequentializer;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;

/**
 * Implements a wave index with either single or double partitioning as described in
 * [1] "Shivakumar/Garcia-Molina: Wave-Indices: Indexing Evolving Databases (1997)"
 * [2] "Golab/Prahladka/szu: Indexing the Results of Sliding Window Queries (2005)"
 * A wave index is used to index elements from a sliding time window.
 */
public class WaveIndex {
	/**
	 * This interface defines the operations needed for a constituent index
	 * in the wave index.
	 */
	public interface ConstituentIndex {
		/**
		 * Inserts all elements given by an iterator into the index.
		 * 
		 * @param objects objects to be inserted
		 */
		public void insert(Iterator<TemporalObject> objects);

		/**
		 * Removes all elements from the index that have an expiration
		 * timestamp in the given interval.
		 * 
		 * @param interval half-open time interval
		 */
		public void remove(TimeInterval interval);

		/**
		 * Runs an exact match query on the index. Returns all elements that
		 * are equal to the given element.
		 * 
		 * @param data data element to query for
		 */
		public Cursor<TemporalObject> query(Comparable data);

		/**
		 * Runs a range query on the index. Returns all elements from the closed
		 * key interval defined by the two arguments. If the index does not support
		 * range queries it may throw an <tt>UnsupportedOperationException</tt>.
		 * In that case the wave index will not support range queries as well.
		 * 
		 * @param dataFrom data element that defines lower bound
		 * @param dataFrom data element that defines higher bound
		 */
		public Cursor<TemporalObject> rangeQuery(Comparable dataFrom, Comparable dataTo);
		
		/**
		 * Rebuilds this index, deleting all elements that have an expiration
		 * timestamp in the specified interval and inserting all new elements.
		 * 
		 * @param keep keep the contents of the old index?
		 * @param interval half-open time interval of elements to be deleted
		 * @param newElements new data elements to be inserted. Some constituent
		 * indices might require an ascending order on the elements.
		 * @return newly rebuilt index
		 */
		public ConstituentIndex rebuild(boolean keep, TimeInterval interval, Iterator<TemporalObject> newElements);
	}

	/**
	 * Abstract base class for implementing partition strategies. A partition
	 * strategy determines into which partition elements are inserted and from
	 * which partition elements are deleted at any specific point in time.
	 */
	protected abstract class PartitionStrategy {
		protected int numberOfPartitions;
		
		public PartitionStrategy(int numberOfPartitions) {
			this.numberOfPartitions = numberOfPartitions;			
		}

		/**
		 * Returns the partition for the given timestamp.
		 * 
		 * @param ts timestamp
		 * @return correct partition for the given timestamp
		 */
		public abstract int getPartitionNumber(long ts);
	}

	/**
	 * Implements the chronological partitioning strategy from [2].
	 * For a time window of size <tt>s</tt>, each partition will cover
	 * <tt>s / numerOfPartitions</tt> consecutive timestamps at any
	 * point in time.
	 */
	protected class Chronological extends PartitionStrategy {
		protected int rollover, lower, upper, sepTs, mod;

		public Chronological(int numberOfPartitions) {
			super(numberOfPartitions);
			initialize();
			if (DEBUG_PRINT_GENERAL)
				DEBUG_OUTPUT.println("rollover="+rollover+", lower="+lower+", upper="+upper+", sepTs="+sepTs);
		}
		
		protected void initialize() {
			final int p = s / r; // number of updates before window rolls over
			mod = p % numberOfPartitions;
			lower = p / numberOfPartitions;
			upper = mod != 0 ? lower + 1 : lower;
			rollover = s;
			lower *= r;
			upper *= r;
			sepTs = mod * upper;
		}
		
		public int getPartitionNumber(long ts) {
			int tsSinceRollover = (int)((ts - tsStart) % rollover);
			int part;
			if (tsSinceRollover <= sepTs)
				part = tsSinceRollover / upper;
			else
				part = (tsSinceRollover - sepTs) / lower + mod;
			assert part >= 0 && part < numberOfPartitions;
			return part;
		}
	}

	/**
	 * Implements the chronological partitioning strategy, as used for the lazy
	 * strategy, from [2].
	 * For a time window of size <tt>s</tt> and <tt>s / r</tt> updates before a window
	 * rollover, each partition will cover <tt>(p-1) / (numerOfPartitions-1)</tt>
	 * consecutive timestamps at any point in time.
	 */
	protected class LazyChronological extends PartitionStrategy {
		protected int rollover, lower, upper, sepTs, mod;

		public LazyChronological(int numberOfPartitions) {
			super(numberOfPartitions);
			initialize();
			if (DEBUG_PRINT_GENERAL)
				DEBUG_OUTPUT.println("rollover="+rollover+", lower="+lower+", upper="+upper+", sepTs="+sepTs);
		}
		
		protected void initialize() {
			final int p = s / r; // number of updates before window rolls over
			mod = (p-1) % (numberOfPartitions-1);
			lower = (p-1) / (numberOfPartitions-1);
			upper = mod != 0 ? lower + 1 : lower;
			rollover = ((p-1) + upper) * r;
			lower *= r;
			upper *= r;
			sepTs = (mod+1) * upper;
		}

		public int getPartitionNumber(long ts) {
			int tsSinceRollover = (int)((ts - tsStart) % rollover);
			int part;
			if (tsSinceRollover <= sepTs)
				part = tsSinceRollover / upper;
			else
				part = (tsSinceRollover - sepTs) / lower + mod + 1;
			assert part >= 0 && part < numberOfPartitions;
			return part;
		}
	}

	/**
	 * Implements the round-robin partitioning strategy from [2].
	 * Partitions will be switched at every index update, i.e. every
	 * <tt>r</tt> time units. 
	 */
	protected class RoundRobin extends PartitionStrategy {
		public RoundRobin(int numberOfPartitions) {
			super(numberOfPartitions);
		}

		public int getPartitionNumber(long ts) {
			int p = (int)((ts - tsStart) / r % numberOfPartitions);
			assert p >= 0 && p < numberOfPartitions;
			return p;
		}
	}
	
	/**
	 * Abstract base class for implementing update strategies. An update strategy
	 * determines how insertions and deletions are handled at each update.
	 */
	protected abstract class UpdateStrategy {
		protected int insertPartition, deletePartition, lastInsertPartition = -1;
		protected TimeInterval interval;

		public void update() {
			// determine partitions for update timestamp
			insertPartition = insertStrategy.getPartitionNumber(tsUpdate - 1);
			deletePartition = deleteStrategy.getPartitionNumber(tsUpdate - 1);

			// all buffered elements have a generation timestamp in this interval.
			// all elements in the index which have an expiration timestamp in this
			// interval will be deleted.
			interval = new TimeInterval(tsUpdate - r, tsUpdate);
			
			if (DEBUG_PRINT_UPDATE_INFO)
				DEBUG_OUTPUT.println("UPDATE: time " + tsUpdate + ", " + "insertPartition " + insertPartition + ", " + "deletePartition " + deletePartition + ", interval " + interval);

			// delete expired elements
			if (fixedTimeInterval) {
				if (!lazy)
					delete(insertPartition, getIndexNumber(insertPartition, deletePartition), false);
			}
			else 
				for (int ins = 0; ins < g; ins++) {
					// if lazy, we skip deletion from the index belonging
					// to the oldest insertion partition
					if (!(lazy && ins == ((insertPartition + 1) % g)))
						delete(ins, getIndexNumber(ins, deletePartition), lazy && insertPartition != lastInsertPartition && ins == insertPartition);
				}
/*
			for (int i = 0; i < n; i++) {
				System.out.println("Index " + i + ": " + indices[i]);
			}
*/

			// insert buffered elements
			for (int i = 0; i < e; i++) {
				// if lazy and we moved to a new insertion partition,
				// the old index is dropped and recreated
				insert(i, getIndexNumber(insertPartition, i), lazy && insertPartition != lastInsertPartition);
			}

			buffer.clear();

			if (DEBUG_PRINT_UPDATE_CONTENTS)
				for (int i = 0; i < n; i++)
					DEBUG_OUTPUT.println("Index " + i + ": " + indices[i]);

			lastInsertPartition = insertPartition;
		}
		
		protected abstract void delete(int currentInsertPartition, int index, boolean drop);
		
		protected abstract void insert(int currentDeletePartition, int index, boolean drop);
	}
	
	/**
	 * Implements the NoRebuild strategy from [2].
	 */
	protected class NoRebuild extends UpdateStrategy {
		@Override
		protected void delete(int currentInsertPartition, int index, boolean drop) {
			if (!drop) {
				if (DEBUG_PRINT_UPDATE_INFO)
					DEBUG_OUTPUT.println("Deleting from index "+index);
				indices[index].remove(interval);
			}
		}

		@Override
		protected void insert(int currentDeletePartition, int index, boolean drop) {
			if (DEBUG_PRINT_UPDATE_INFO)
				DEBUG_OUTPUT.println("Inserting into index "+index+", drop old: "+drop);
			if (drop)
				//indices[index] = createIndex.invoke(index);
				indices[index] = indices[index].rebuild(false, null, new EmptyCursor<TemporalObject>());
			indices[index].insert(buffer.get(currentDeletePartition));
		}
	}
	
	/**
	 * Implements the Rebuild strategy from [2].
	 */
	protected class Rebuild extends UpdateStrategy {
		@Override
		protected void delete(int currentInsertPartition, int index, boolean drop) {
			if (DEBUG_PRINT_UPDATE_INFO)
				DEBUG_OUTPUT.println("Rebuilding (delete) index "+index+", drop old: "+drop);
			indices[index] = indices[index].rebuild(
					!drop,					
					interval,
					currentInsertPartition == insertPartition ? // elements to be inserted
						buffer.get(deletePartition) :
						new EmptyCursor<TemporalObject>());
		}

		@Override
		protected void insert(int currentDeletePartition, int index, boolean drop) {
			// if (currentDeletePartition == deletePartition), insertions have
			// already been handled when rebuilding the index during deletions
			if (currentDeletePartition != deletePartition) {
				if (DEBUG_PRINT_UPDATE_INFO)
					DEBUG_OUTPUT.println("Rebuilding (insert) index "+index+", drop old: "+drop);
				indices[index] = indices[index].rebuild(
						!drop,
						null,
						buffer.get(currentDeletePartition)
						);
			}
		}
	}
	
	/**
	 * RebuildOnDelete: an index is rebuilt when deletions are necessary; otherwise
	 * insertions are handled sequentially.
	 */
	protected class RebuildOnDelete extends Rebuild {
		@Override
		protected void insert(int currentDeletePartition, int index, boolean drop) {
			// if (currentDeletePartition == deletePartition), insertions have
			// already been handled when rebuilding the index during deletions
			if (currentDeletePartition != deletePartition) {
				if (DEBUG_PRINT_UPDATE_INFO)
					DEBUG_OUTPUT.println("Inserting into index "+index+", drop old: "+drop);
				if (drop)
					//indices[index] = createIndex.invoke(index);
					indices[index] = indices[index].rebuild(false, null, new EmptyCursor<TemporalObject>());
				indices[index].insert(buffer.get(currentDeletePartition));
			}
		}
	}

	/**
	 * Defines the operations needed by the element buffer used in the wave index.
	 * The buffer is used to store incoming elements until the next update. Elements
	 * are stored with respect to their deletion partition.
	 */
	public interface ElementBuffer {
		/**
		 * Adds an element to the buffer.
		 * 
		 * @param deletePartition deletion partition of the element.
		 * @param obj data element to be added.
		 */
		public void add(int deletePartition, TemporalObject obj);

		/**
		 * Returns all buffered elements which belong to the given deletion
		 * partition.
		 * 
		 * @param deletePartition desired deletion partition.
		 * @return iterator over data elements.
		 */
		public Iterator<TemporalObject> get(int deletePartition);
		
		/**
		 * Removes all elements from the buffer.
		 */
		public void clear();
	}

	protected final int s; // window size
	protected final int r; // time interval between two consecutive index updates
	protected final int g; // number of partitions of generation times
	protected final int e; // number of partitions of expiration times
	protected final int n; // number of constituent indexes

	protected final Function<Integer,ConstituentIndex> createIndex;
	protected final ElementBuffer buffer;
	protected final ConstituentIndex[] indices;
	protected final UpdateStrategy strategy;
	protected final PartitionStrategy insertStrategy; // strategy for choosing an insertion partition
	protected final PartitionStrategy deleteStrategy; // strategy for choosing a deletion partition
	protected final boolean lazy;
	protected final boolean fixedTimeInterval;

	protected long tsStart = -1; // generation timestamp of the first ever received element
	protected long tsUpdate; // timestamp at which the next index update occurs
	protected TimeInterval lastTimeInterval;

	public enum Strategy {
		NoRebuild,
		Rebuild
	}
	
	/**
	 * Creates a new DoublyPartitionedWaveIndex.
	 * 
	 * @param s window size
	 * @param r time interval between two consecutive index updates
	 * @param g number of partitions of generation times
	 * @param e number of partitions of expiration times
	 * @param rebuild rebuild constituent index at each update?
	 * @param lazy use lazy deletion? 
	 * @param fixedTimeInterval do all elements have a time interval of the same length? If true, a single partitioned index will be created. 
	 * @param createIndex function that creates a new constituent index: int -> ConstituentIndex, int being the index number.
	 * @param createBuffer function that creates the element buffer: int -> ElementBuffer, int being the number of partitions in the buffer.
	 */
	public WaveIndex(int s, int r, int g, int e, Strategy strategy, boolean lazy, boolean fixedTimeInterval, Function<Integer,ConstituentIndex> createIndex, Function<Integer,ElementBuffer> createBuffer) {
		if (s < 1) throw new IllegalArgumentException("S must be 1 or greater");
		if (r < 1) throw new IllegalArgumentException("r must be 1 or greater");
		if (g < 1) throw new IllegalArgumentException("G must be 1 or greater");
		if (e < 1) throw new IllegalArgumentException("E must be 1 or greater");
		if ((s % r) != 0) throw new IllegalArgumentException("r must divide s");
/*
		if ((s % e) != 0) throw new IllegalArgumentException("E must divide s");
		if ((s/e % r) != 0) throw new IllegalArgumentException("r must divide s/E"); 
		if (!lazy) {
			if ((s % g) != 0) throw new IllegalArgumentException("G must divide s");
			if ((s/g % r) != 0) throw new IllegalArgumentException("r must divide s/G");
		}
*/
		if (fixedTimeInterval && e != 1) throw new IllegalArgumentException("For a fixed time interval E must be 1");
		if (lazy && g < 2) throw new IllegalArgumentException("For the lazy strategy, G must be 2 or greater.");
		if (strategy.equals(Strategy.Rebuild) && lazy && fixedTimeInterval) throw new IllegalArgumentException("For a single partitioned wave index, rebuild and lazy can't be combined.");

		this.s = s;
		this.r = r;
		this.g = g;
		this.e = e;
		this.n = g * e;
		
		this.createIndex = createIndex;
		
		this.buffer = createBuffer.invoke(e);

		switch(strategy) {
			case NoRebuild:
				this.strategy = new NoRebuild();
				break;
			case Rebuild:
				this.strategy = new Rebuild();
				break;
			default:
				throw new IllegalArgumentException();
		}

		this.insertStrategy = lazy ? new LazyChronological(g) : (fixedTimeInterval ? new Chronological(g) : new RoundRobin(g));
		this.deleteStrategy = new RoundRobin(e);
		this.lazy = lazy;
		this.fixedTimeInterval = fixedTimeInterval;
		
		indices = new ConstituentIndex[n];
		for (int i = 0; i < n; i++)
			indices[i] = createIndex.invoke(i);
		
		if (DEBUG_PRINT_GENERAL) {
			DEBUG_OUTPUT.println("Wave Index created:");
			DEBUG_OUTPUT.println("S="+s+", r="+r+", "+g+"x"+e+", "+(lazy?"Lazy ":"")+strategy);
			DEBUG_OUTPUT.print("# of updated sub-indices: ");
			if (fixedTimeInterval)
				DEBUG_OUTPUT.println("1");
			else
				DEBUG_OUTPUT.println(g+e-1-(lazy?1:0));
		}
	}

	/**
	 * Inserts the given element into the index. The inserted objects need to be
	 * chronologically ordered. The time interval of the object must not be
	 * larger than the window size of this wave index.
	 * 
	 * @param obj data element to be inserted.
	 */
	public void insert(TemporalObject obj) {
		if (lastTimeInterval != null && 
			TimeInterval.TIME_INTERVAL_COMPARATOR.compare(lastTimeInterval, obj.getTimeInterval()) > 0)
			throw new IllegalArgumentException("Object's time interval is not chronologically greater than or equal to the last object's time interval.");
			
		long length = obj.getEnd() - obj.getStart();
		if (fixedTimeInterval && length != s)
			throw new IllegalArgumentException("Length of object's time interval must be equal to the window size.");
		if (length > s)
			throw new IllegalArgumentException("Length of object's time interval cannot be larger than the window size.");

		// one-time initialization
		if (tsStart == -1) {
			tsStart = obj.getStart(); // remember first timestamp ever received
			tsUpdate = obj.getStart() + r; // schedule first update
		}

		while (obj.getStart() >= tsUpdate) { // update wave index now?
//			DEBUG_OUTPUT.println("Update");
			strategy.update();
			tsUpdate += r;
			if (DEBUG_ON_UPDATE != null)
				DEBUG_ON_UPDATE.invoke(this);
		}

		// if expiration timestamp of object is <= tsUpdate, the object has to be both
		// inserted and deleted at the next update, so it is dropped
		if (obj.getEnd() >= tsUpdate) {
			// buffer the element
			final int deletePartition = deleteStrategy.getPartitionNumber(obj.getEnd());
			buffer.add(deletePartition, obj);
		}
	}

	/**
	 * Runs an exact match query on the index. Returns all elements that
	 * are equal to the given element.
	 * 
	 * @param data data element to query for
	 */
	public Cursor<TemporalObject> query(Comparable data) {
		Iterator<TemporalObject>[] results = new Iterator[n];
		
		for (int i = 0; i < n; i++)
			results[i] = indices[i].query(data);
		
		return new Sequentializer<TemporalObject>(results);
	}

	/**
	 * Runs a range query on the index. Returns all elements from the closed
	 * key interval given by the two arguments. If the underlying constituent
	 * index does not support range queries, an <tt>UnsupportedOperationException</tt>
	 * is thrown.
	 * 
	 * @param dataFrom data element that defines lower bound
	 * @param dataFrom data element that defines higher bound
	 */
	public Cursor<TemporalObject> rangeQuery(Comparable dataFrom, Comparable dataTo) {
		Iterator<TemporalObject>[] results = new Iterator[n];
		
		for (int i = 0; i < n; i++)
			results[i] = indices[i].rangeQuery(dataFrom, dataTo);
		
		return new Sequentializer<TemporalObject>(results);
	}

	protected int getIndexNumber(int insertPartition, int deletePartition) {
		assert insertPartition >= 0 && insertPartition < g;
		assert deletePartition >= 0 && deletePartition < e;

		return insertPartition * e + deletePartition;
	}
	
	// DEBUGGING
	
	public ConstituentIndex[] getIndices() {
		return indices;
	}
	
	public long getLastUpdateTimestamp() {
		return tsUpdate - r;
	}
	
	public static PrintStream DEBUG_OUTPUT = System.out;
	public static boolean DEBUG_PRINT_GENERAL = true;
	public static boolean DEBUG_PRINT_UPDATE_INFO = false;
	public static boolean DEBUG_PRINT_UPDATE_CONTENTS = false;
	public static Function DEBUG_ON_UPDATE;
}
