/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.metaData;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import xxl.core.functions.Function;
import xxl.core.math.functions.AggregationFunction;
import xxl.core.pipes.queryGraph.AbstractNode.AbstractNodeMetaDataManagement;
import xxl.core.util.metaData.CompositeMetaData;
import xxl.core.util.metaData.MetaDataComposition;
import xxl.core.util.metaData.MetaDataHandler;
import xxl.core.util.metaData.MetaDataManagement;
import static xxl.core.pipes.queryGraph.AbstractNode.AbstractNodeMetaDataManagement.META_DATA_UPDATE_RATE;;

public class AggregationMetaDataIdentifier<P,A> extends ComplexMetaDataIdentifier {
	
	protected Class<? extends AggregationFunction<P, A>> aggregationFunctionClass;
	protected Object [] parameters;
	
	public AggregationMetaDataIdentifier (Class<? extends AggregationFunction<P, A>> aggregationFunctionClass, Object [] parameters, Object basisMetaDataIdentifier) {
		super(basisMetaDataIdentifier);
		this.aggregationFunctionClass = aggregationFunctionClass;
		this.parameters = parameters;
	}

	public AggregationMetaDataIdentifier (Class<? extends AggregationFunction<P, A>> aggregationFunctionClass, Object basisMetaDataIdentifier) {
		this(aggregationFunctionClass, null, basisMetaDataIdentifier);
	}

	
	@Override
	public boolean equals(Object o) {
		AggregationMetaDataIdentifier c;
		try {
			c = (AggregationMetaDataIdentifier) o;
		}
		catch(ClassCastException e) {
			return false;
		}
		if (parameters != null) {
			if (c.parameters == null)
				return false;
			if (parameters.length != c.parameters.length)
				return false;
			for (int i=0; i<parameters.length; i++)
				if (!parameters[i].equals(c.parameters[i]))
						return false;
		}
		else if (c.parameters != null)
			return false;
		return basisMetaDataIdentifier.equals(c.basisMetaDataIdentifier)
			&& aggregationFunctionClass.equals(c.aggregationFunctionClass);
	}

	@Override 
	public int hashCode() {
		return basisMetaDataIdentifier.hashCode() ^ aggregationFunctionClass.toString().hashCode();
	}
	
	@Override
	public String toString() {
		String fs = aggregationFunctionClass.toString();
		if (fs.contains("."))
			fs = fs.substring(fs.lastIndexOf('.')+1);
		if (fs.contains("$"))
			fs = fs.substring(fs.lastIndexOf('$')+1);
		if (parameters!=null)
			for (int i=0; i<parameters.length; i++)
				fs+="{"+parameters[i]+"}";
		return fs+"("+basisMetaDataIdentifier+")";
	}

	public Class getAggregationFunctionClass() {
		return aggregationFunctionClass;
	}

	public Object getBasisMetaDataIdentifier() {
		return basisMetaDataIdentifier;
	}

	public Object newInstance(final MetaDataManagement mdm) {
		final AbstractNodeMetaDataManagement metaDataManagement;
		final AggregationFunction<P, A> aggregationFunction;
		try {
			metaDataManagement = (AbstractNodeMetaDataManagement)mdm;
			if (parameters==null || parameters.length==0)
				aggregationFunction = aggregationFunctionClass.newInstance();
			else {
				Class[] classes = new Class[parameters.length];
				for (int i=0; i<parameters.length; i++)
					classes[i] = parameters[i].getClass();
				Constructor<? extends AggregationFunction<P,A>> con = aggregationFunctionClass.getConstructor(classes);
				aggregationFunction = con.newInstance(parameters);
			}
		} catch (ClassCastException e) {
			return null;
		} catch (InstantiationException e) {
			return null;
		} catch (IllegalAccessException e) {
			return null;
		} catch (SecurityException e) {
			return null;
		} catch (NoSuchMethodException e) {
			return null;
		} catch (IllegalArgumentException e) {
			return null;
		} catch (InvocationTargetException e) {
			return null;
		}
		final CompositeMetaData metaData = metaDataManagement.getMetaData();
		Function<Object, A> metaDataFunction = new Function<Object, A>() {				
			protected A aggregate = null;				
			@SuppressWarnings("unchecked")
			@Override
			public A invoke() {
				synchronized(metaDataManagement) {
					P value = ((MetaDataHandler<P>)metaData.get(basisMetaDataIdentifier)).getMetaData();
					if (value == null) {
						return null;
					}
					aggregate = aggregationFunction.invoke(aggregate, value);
					return aggregate;
				}									
			}
		};
		MetaDataHandler mdh =  (MetaDataHandler<P>)metaData.get(basisMetaDataIdentifier);
		if (mdh instanceof DefaultMetaDataHandler) {
			mdm.include(META_DATA_UPDATE_RATE);
			long updatePeriod = (Long)((MetaDataHandler)((CompositeMetaData)(mdm.getMetaData())).get(META_DATA_UPDATE_RATE)).getMetaData();
			mdm.exclude(META_DATA_UPDATE_RATE);
			return new PeriodicEvaluationMetaDataHandler<A>(metaDataManagement, this, metaDataFunction, updatePeriod);
		}
		else
			return new TriggeredEvaluationMetaDataHandler<A>(metaDataManagement, this, metaDataFunction); 
	}
}
