/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.distincts;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.collections.sweepAreas.PNSweepArea;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.DecoratorPNObject;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.PNTemporalPipe;
import xxl.core.pipes.operators.mappers.TemporalPNPipe;
import xxl.core.pipes.operators.windows.PNTemporalWindow;
import xxl.core.pipes.operators.windows.TemporalWindow;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

public class PNDistinct<I> extends AbstractPNPipe<I,I> implements MemoryMonitorable {
	
	public static class PNDistinctSA<I> extends PNSweepArea<I> {
    	
		public PNDistinctSA(Function<? super PNObject<I>, Integer> hashFunction, Function<Object, List<PNObject<I>>> newList) {
			super(new HashSAImplementor<PNObject<I>>(hashFunction, newList, 1, PNObject.VALUE_EQUIVALENCE_PREDICATE), 0, PNObject.VALUE_EQUIVALENCE_PREDICATE, 1);
		}
	
		public PNDistinctSA() {
			this(
				new Function<PNObject<I>, Integer>() {
					@Override
					public Integer invoke(PNObject<I> pno) {
						return pno.getObject().hashCode();
					}
				},
				new Function<Object, List<PNObject<I>>>() {
					@Override
					public List<PNObject<I>> invoke() {
						return new ArrayList<PNObject<I>>();
					}
				}
			);
		}
		
		public void incrementCounter(PNObject<I> pno) {
			DecoratorPNObject<I,Integer> dpno = (DecoratorPNObject<I,Integer>)pno;
			impl.update(dpno, new DecoratorPNObject<I,Integer>(new PNObject<I>(dpno), dpno.getDecoration()+1));
		}
		
		public void decrementCounter(PNObject<I> pno) {
			DecoratorPNObject<I,Integer> dpno = (DecoratorPNObject<I,Integer>)pno;
			impl.update(dpno, new DecoratorPNObject<I,Integer>(new PNObject<I>(dpno), dpno.getDecoration()-1));
		}
			
	}
	
	protected PNDistinctSA<I> sweepArea;
	
	public PNDistinct(PNDistinctSA<I> sweepArea) {
		super();
		this.sweepArea = sweepArea;
	}
	
	public PNDistinct() {
		this(new PNDistinctSA<I>());
	}
	
	public PNDistinct(Source<? extends PNObject<I>> source, int sourceID) {
		this();
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");		
	}
	
	public PNDistinct(Source<? extends PNObject<I>> source) {
		this(source, DEFAULT_ID);
	}
	
	@Override
	public void processObject(PNObject<I> pno, int sourceID) throws IllegalArgumentException {
		Iterator<PNObject<I>> matches = sweepArea.query(pno, sourceID);
		if(pno.isPositive()) {
			if (!matches.hasNext()) 
				sweepArea.insert(new DecoratorPNObject<I,Integer>(new PNObject<I>(pno), 1));
			else {
				sweepArea.incrementCounter(matches.next());
				return;
			}
		}
		else { // pno is negative
			if (matches.hasNext()) {
				DecoratorPNObject<I,Integer> dpno = (DecoratorPNObject<I,Integer>)matches.next();
				if (dpno.getDecoration() == 1)
					sweepArea.reorganize(pno, sourceID);
				else {
					sweepArea.decrementCounter(dpno);
					return;
				}
			}
		}
		transfer(pno);
	}
	
	@Override
	public void close() {
		super.close();
		if (isClosed) {
			sweepArea.close();
		}
	}
		
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	public static void main(String[] args) {
			
			/*********************************************************************/
			/*                            Example 1                              */
			/*********************************************************************/
			
			final int noOfElements = 1000;
			final int maxValue = 20;
			final int maxTSInc = 50;
			int seed = 17;
			
			TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
					0,
					new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxValue), noOfElements),
					new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxTSInc), noOfElements)
			);
			
			final long windowSize = 2000;
				
			// time interval approach
			TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s1, windowSize);
			TemporalDistinct<Integer> res1 = new TemporalDistinct<Integer>(w1);
			
			// positive-negative approach
			TemporalPNPipe<Integer> pnSource = new TemporalPNPipe<Integer>(s1);
			PNTemporalWindow<Integer> pnWindow = new PNTemporalWindow<Integer>(pnSource, windowSize);
			PNDistinct<Integer> pnDistinct = new PNDistinct<Integer>(pnWindow);
			PNTemporalPipe<Integer> res2 = new PNTemporalPipe<Integer>(pnDistinct);
			
			QueryExecutor exec = new QueryExecutor();
	//		exec.registerQuery(new Printer(w1));
	//		exec.registerQuery(new Printer(pnDistinct));
			exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res1));
			exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res2));
			exec.registerQuery(Pipes.compareSnapshots(res1, res2, false, false));
			exec.startAllQueries();
	
		}

}
