/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.filters;

import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;

import java.lang.reflect.GenericArrayType;
import java.lang.reflect.Type;

import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.elements.TimeStampedObject;
import xxl.core.pipes.metaData.DefaultMetaDataHandler;
import xxl.core.pipes.metaData.MetaDataDependencies;
import xxl.core.pipes.metaData.TriggeredEvaluationMetaDataHandler;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.AbstractPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.filters.TemporalFilter.TemporalFilterMetaDataManagement;
import xxl.core.pipes.operators.mappers.PNTemporalPipe;
import xxl.core.pipes.operators.mappers.TemporalPNPipe;
import xxl.core.pipes.operators.windows.PNTemporalWindow;
import xxl.core.pipes.operators.windows.TemporalWindow;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.predicates.CountingPredicate;
import xxl.core.predicates.Predicate;
import xxl.core.util.metaData.MetaDataHandler;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

public class PNFilter<I> extends AbstractPNPipe<I,I> {

	/**
	 * Inner class for metadata management. 
	 */
	public class PNFilterMetaDataManagement extends AbstractPipeMetaDataManagement {
			
		
		public static final String SINGLE_PREDICATE_COSTS = "SINGLE_PREDICATE_COSTS";
		protected static final String PREDICATE_DECORATION = "PREDICATE_DECORATION";
				
		protected volatile double singlePredicateEvaluationCosts;
			
		public PNFilterMetaDataManagement() {
			super();
		}
		
		
		@Override
		protected boolean addMetaData(Object metaDataIdentifier) {			
			if (metaDataIdentifier.equals(PREDICATE_DECORATION)) {
				processingWLock.lock();
				try {
					predicate = new CountingPredicate<PNObject<I>>(predicate);
				}
				finally {
					processingWLock.unlock();
				}
				metaData.add(metaDataIdentifier, new TriggeredEvaluationMetaDataHandler<long[]>(this, metaDataIdentifier,
					new GenericArrayType() {
						public Type getGenericComponentType() {
							return Long.class;
						}
					},
					new Function<Object,long[]>() {			
						CountingPredicate<PNObject<I>> cp = (CountingPredicate<PNObject<I>>)predicate;

						@Override
						public long[] invoke() {
							synchronized(metaDataManagement) {
								if (isClosed || isDone) 
									return null;
								long [] res = new long[2];
								res[0]=cp.getNoOfHits();
								res[1]=cp.getNoOfCalls();
								cp.resetCounters();
								return res;
							}
						}
					})
				);
				return true;
			}
			if (metaDataIdentifier.equals(SELECTIVITY_MEASUREMENT)) {
				metaData.add(metaDataIdentifier, new TriggeredEvaluationMetaDataHandler<Double>(this, metaDataIdentifier, Double.class, 
					new Function<Object,Double>() {			
						@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								if (isClosed || isDone) 
									return Double.NaN;								
								long[] counters = (long[])((MetaDataHandler)metaData.get(PREDICATE_DECORATION)).getMetaData();
								return ((double)counters[0])/counters[1];
							}
						}
					})
				);
				return true;
			}
			if (metaDataIdentifier.equals(SINGLE_PREDICATE_COSTS)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Double>(this, metaDataIdentifier, Double.class,
					new Function<Object,Double>() {			
						@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								return singlePredicateEvaluationCosts;
							}
						}
					})
				);
				return true;
			}
			if (metaDataIdentifier.equals(COST_MEASUREMENT)) {
				metaData.add(metaDataIdentifier, new TriggeredEvaluationMetaDataHandler<Double>(this, metaDataIdentifier, Double.class,
					new Function<Object,Double>() {			
						@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								if (isClosed || isDone) 
									return Double.NaN;								
								long[] counters = (long[])((MetaDataHandler)metaData.get(PREDICATE_DECORATION)).getMetaData();
								double singlePredicateCosts = (Double)((MetaDataHandler)metaData.get(SINGLE_PREDICATE_COSTS)).getMetaData();
								return counters[1]*singlePredicateCosts/updatePeriod;
							}
						}
					})
				);
				return true;
			}			
				
			if (super.addMetaData(metaDataIdentifier))
				return true;
			return false;
		}
		
		@Override
		protected boolean removeMetaData(Object metaDataIdentifier) {			
			if (metaDataIdentifier.equals(PREDICATE_DECORATION)) {
				if (!(isClosed || isDone)) {
					CountingPredicate<PNObject<I>> cp = (CountingPredicate<PNObject<I>>)predicate;
					processingWLock.lock();
					try {
						predicate = cp.getPredicate();
					}
					finally {
						processingWLock.unlock();
					}					
				}
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (metaDataIdentifier.equals(SINGLE_PREDICATE_COSTS) ||
				metaDataIdentifier.equals(COST_MEASUREMENT) ||
				metaDataIdentifier.equals(SELECTIVITY_MEASUREMENT)) {
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (super.removeMetaData(metaDataIdentifier))
				return true;
			return false;
		}
		
		/**
		 * @param costs
		 */
		public void setSinglePredicateEvaluationCosts(double costs) {
			synchronized(metaDataManagement) {
				this.singlePredicateEvaluationCosts = costs;
			}
			refresh(SINGLE_PREDICATE_COSTS);
		}

	}
	
	/**
	 * Unary selection predicate.
	 */
	protected Predicate<? super PNObject<I>> predicate;

	public PNFilter(Predicate<? super PNObject<I>> predicate) {
		this.predicate = predicate;
	}
	
	public PNFilter(Source<? extends PNObject<I>> source, int sourceID, Predicate<? super PNObject<I>> predicate) {
		this(predicate);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");		
	}
	
	public PNFilter(Source<? extends PNObject<I>> source, Predicate<? super PNObject<I>> predicate) {
		this(source, DEFAULT_ID, predicate);
	}
	
	static {
		MetaDataDependencies.addLocalMetaDataDependencies(PNFilter.class, COST_MEASUREMENT, PNFilter.PNFilterMetaDataManagement.SINGLE_PREDICATE_COSTS, PNFilter.PNFilterMetaDataManagement.PREDICATE_DECORATION);
		MetaDataDependencies.addLocalMetaDataDependencies(PNFilter.class, PNFilter.PNFilterMetaDataManagement.PREDICATE_DECORATION, AbstractPipe.AbstractPipeMetaDataManagement.SYNC_TRIGGER+"_ABSTRACTTIMESTAMPPIPE");
		MetaDataDependencies.addLocalMetaDataDependencies(PNFilter.class, PNFilter.PNFilterMetaDataManagement.SELECTIVITY_MEASUREMENT, PNFilter.PNFilterMetaDataManagement.PREDICATE_DECORATION);
	}
	
	@Override
	public void processObject(PNObject<I> o, int ID) throws IllegalArgumentException {
		if (predicate.invoke(o))
			super.transfer(o);
		else if (activateHeartbeats && lastHB < minTimeStamp && heartbeatPredicate.invoke()) {			
			transferHeartbeat(minTimeStamp);
			lastHB = minTimeStamp;
		}
	}
	
	@Override
	public void createMetaDataManagement() {
		if (metaDataManagement != null)
			throw new IllegalStateException("An instance of MetaDataManagement already exists.");
		metaDataManagement = new PNFilterMetaDataManagement();
	}

	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 1000;
		final int maxValue = 20;
		final int maxTSInc = 50;
		
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(maxTSInc), noOfElements)
		);
		
		Predicate<TimeStampedObject<Integer>> predicate = new Predicate<TimeStampedObject<Integer>>() {
			@Override
			public boolean invoke(TimeStampedObject<Integer> o) {
				return o.getObject()%7 == 0;
			}
		};
		final long windowSize = 10;
			
		// time interval approach
		TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s1, windowSize);
		TemporalFilter<Integer> res1 = new TemporalFilter<Integer>(w1, predicate);
		
		// positive-negative approach
		TemporalPNPipe<Integer> pnSource = new TemporalPNPipe<Integer>(s1);
		PNTemporalWindow<Integer> pnWindow = new PNTemporalWindow<Integer>(pnSource, windowSize);
		PNFilter<Integer> pnFilter = new PNFilter<Integer>(pnWindow, predicate);
		PNTemporalPipe<Integer> res2 = new PNTemporalPipe<Integer>(pnFilter);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res1));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res2));
		exec.registerQuery(Pipes.compareSnapshots(res1, res2, true, true));
		exec.startAllQueries();

	}
}