/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.groupers;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;

import xxl.core.collections.MapEntry;
import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.collections.sweepAreas.PNSweepArea;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.functions.Function;
import xxl.core.math.functions.AggregationFunction;
import xxl.core.math.statistics.parametric.aggregates.Sum;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.PNTemporalPipe;
import xxl.core.pipes.operators.mappers.TemporalPNPipe;
import xxl.core.pipes.operators.windows.PNTemporalWindow;
import xxl.core.pipes.operators.windows.TemporalWindow;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

/**
 * 
 */
public class PNGroupAndAggregator<I,K,V> extends AbstractPNPipe<I,Entry<K,V>> implements MemoryMonitorable {
	
	protected Hashtable<K,PNSweepArea<I>> groups;
	protected Function<? super I, ? extends K> determineGroup;
	protected Function<?,? extends PNSweepArea<I>> newSweepArea;
	protected Hashtable<K,V> previousAggs;
	protected Hashtable<K,V> intermediateResults;
	protected long lastTimeStamp;
	protected HashSet<K> changedGroups;
	protected HashSet<K> negativeInputs;
	protected AggregationFunction<? super I,V> aggFunction;
		
	public PNGroupAndAggregator(Function<? super I,? extends K> determineGroup, AggregationFunction<? super I,V> aggFunction) {
		this.groups = new Hashtable<K,PNSweepArea<I>>();
		this.determineGroup = determineGroup;
		this.newSweepArea = new Function<Object,PNSweepArea<I>>() {
			public PNSweepArea<I> invoke() {
				return new PNSweepArea<I>(
					new HashSAImplementor<PNObject<I>>(
						new Function<PNObject<I>, Integer>() {
							@Override
							public Integer invoke(PNObject<I> pno) {
								return pno.getObject().hashCode();
							}
						},
						new Function<Object, List<PNObject<I>>>() {
							@Override
							public List<PNObject<I>> invoke() {
								return new LinkedList<PNObject<I>>();
							}
						},
							2, PNObject.VALUE_EQUIVALENCE_PREDICATE
					) {
							
						public Iterator<PNObject<I>> query(PNObject<I> o, int ID) throws IllegalArgumentException {
							return filter(iterator(), o, ID);	
						}
							
					},
					0,
					PNObject.VALUE_EQUIVALENCE_PREDICATE,
					1
				);
			}
		};
		this.aggFunction = aggFunction;
		this.previousAggs = new Hashtable<K,V>();
		this.intermediateResults = new Hashtable<K,V>();
		this.negativeInputs = new HashSet<K>();		
		this.lastTimeStamp = Long.MIN_VALUE;
		this.changedGroups = new HashSet<K>();
	}	
	
	public PNGroupAndAggregator(Source<? extends PNObject<I>> source, int sourceID, Function<? super I,? extends K> determineGroup, AggregationFunction<? super I,V> aggFunction) {
		this(determineGroup, aggFunction);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
	
	public PNGroupAndAggregator(Source<? extends PNObject<I>> source, Function<? super I,? extends K> determineGroup, AggregationFunction<? super I,V> aggFunction) {
		this(source, DEFAULT_ID, determineGroup, aggFunction);
	}
		
	@Override
	public void processObject(PNObject<I> pno, int sourceID) throws IllegalArgumentException {
		// the timestamp of the object to process
		long currentTimeStamp = pno.getTimeStamp();
		// the object to process
		I currentObject = pno.getObject();
		//	the group of the object to process
		K currentGroup = determineGroup.invoke(currentObject);
				
		// check for progress in time
		if (currentTimeStamp > lastTimeStamp) {
			processLastTimeStamp();
			lastTimeStamp = currentTimeStamp;
		}
	
		// create SweepArea
		if (!groups.containsKey(currentGroup)) 
			groups.put(currentGroup, newSweepArea.invoke());
		// update intermediate results for current tuple
		PNSweepArea<I> sweepArea = groups.get(currentGroup);
		if (pno.isPositive()) {
			// if only positive inputs: incremental calculation
			if (!negativeInputs.contains(currentGroup)) {
				// check if incremental calculation in progress ...
				V intermediateResult = intermediateResults.get(currentGroup);
				// ... or start one with last aggregate
				if (intermediateResult == null)					
					intermediateResult = previousAggs.get(currentGroup);
				intermediateResults.put(currentGroup, aggFunction.invoke(intermediateResult, currentObject));
			}
			sweepArea.insert(pno);
		}
		else {
			sweepArea.reorganize(pno, sourceID);
			negativeInputs.add(currentGroup);
		}
		
		// mark groups as changed
		changedGroups.add(currentGroup);		
	}
	
	protected void processLastTimeStamp() {
		// progress in time, all tuples for previous timestamp can be released
		// iterate over all groups which changed for the previous timestamp
		for (K group : changedGroups) {
			// produce negative tuples for old aggregate value of the group
			if (previousAggs.containsKey(group))
				transfer(
						new PNObject<Entry<K,V>>(
							new MapEntry<K,V>(group, previousAggs.get(group)), 
							lastTimeStamp,
							false
						)
					); 				
			// determine new aggregate for the group
			V finalResult = null;
			if (negativeInputs.contains(group)) {
				// negative tuple arrived for the group => recalculate aggregate
				Iterator<PNObject<I>> it = groups.get(group).iterator();
				while (it.hasNext()) 
					finalResult = aggFunction.invoke(finalResult, it.next().getObject());	
			}		
			else
				finalResult = intermediateResults.get(group);
			// produce positive tuple for the new aggregate if any
			if (finalResult != null) {
				transfer(
						new PNObject<Entry<K,V>>(
							new MapEntry<K,V>(group, finalResult), 
							lastTimeStamp,
							true
						)
					);
				previousAggs.put(group, finalResult);
			}
			else { 
				groups.remove(group);
				previousAggs.remove(group);
			}
		}
		changedGroups.clear();
		negativeInputs.clear();
		intermediateResults.clear();		
	}

		
	@Override
	public void close() {
		super.close();
		if (isClosed()) {
			Enumeration<PNSweepArea<I>> groupEnum = groups.elements();
			while(groupEnum.hasMoreElements())
				groupEnum.nextElement().close();
		}
	}
		
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
	    	if (isDone()) {
	    		processLastTimeStamp();
				signalDone();
	    	}
	    }
	    finally {
	    	processingWLock.unlock();
	    }
	}

	public int getCurrentMemUsage() {
		Enumeration<PNSweepArea<I>> groupEnum = groups.elements();
		int size = 0;
		while(groupEnum.hasMoreElements()) {
			PNSweepArea<I> next = groupEnum.nextElement();
			int mem = next.getCurrentMemUsage();
			if (mem == SIZE_UNKNOWN)
				return SIZE_UNKNOWN;
			size += mem;
		}
		return size;
	}
		
	/**
	 * @param args
	 */
	@SuppressWarnings("unchecked")
	public static void main(String[] args) {
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
	
		final int noOfElements = 5000;
		final int maxValue = 25;
		final int maxTSInc = 20;
		int seed = 19;
		
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxTSInc), noOfElements)
		);
		
		final long windowSize = 1500;
		
		// time interval approach
		TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s1, windowSize);
		final Sum agg = new Sum();
		final int buckets = 1117;
		final Function<Integer,Integer> hashFunction1 = new Function<Integer,Integer>() {
			@Override
			public Integer invoke(Integer o) {
				return o % buckets;
			}
		};
		TemporalGroupAndAggregator<Integer,Integer,Number> res1 = new TemporalGroupAndAggregator<Integer,Integer,Number>(
			w1, hashFunction1, agg
		);
		
		// positive-negative approach
		TemporalPNPipe<Integer> pnSource = new TemporalPNPipe<Integer>(s1);
		PNTemporalWindow<Integer> pnWindow = new PNTemporalWindow<Integer>(pnSource, windowSize);
		PNGroupAndAggregator<Integer,Integer,Number> pnAgg = new PNGroupAndAggregator<Integer,Integer,Number>(
		   pnWindow, hashFunction1, agg
		);
		PNTemporalPipe<Entry<Integer,Number>> res2 = new PNTemporalPipe<Entry<Integer,Number>>(pnAgg);
		
		QueryExecutor exec = new QueryExecutor();
//		exec.registerQuery(new Printer(res1));
//		exec.registerQuery(new Printer(pnAgg));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res1));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(pnAgg));
		exec.registerQuery(Pipes.compareSnapshots(res1, res2, false, false));
		exec.startAllQueries();
		
	}

}

