/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.groupers;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;
import java.util.Map.Entry;

import xxl.core.collections.MapEntry;
import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.queues.Heap;
import xxl.core.comparators.FeatureComparator;
import xxl.core.functions.Function;
import xxl.core.math.functions.AggregationFunction;
import xxl.core.math.statistics.parametric.aggregates.Sum;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.Mapper;
import xxl.core.pipes.operators.mappers.TemporalAggregator;
import xxl.core.pipes.operators.mappers.TemporalAggregator.TemporalAggregatorSA;
import xxl.core.pipes.operators.unions.TemporalUnion;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.AbstractSink;
import xxl.core.pipes.sinks.Tester;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;

/**
 *
 */
public class TemporalGroupAndAggregator<I,K,V> extends TemporalGrouper<I,K,Entry<K,V>> implements MemoryMonitorable {
	
	protected Heap<K> groupsOrderedByStartTimestamps;
	
	public TemporalGroupAndAggregator(Function<? super I,? extends K> determineGroup, Function<?,? extends TemporalAggregatorSA<I,V>> newTemporalGroupAndAggregatorSA) {
		super(determineGroup, newTemporalGroupAndAggregatorSA);
		this.groupsOrderedByStartTimestamps = new DynamicHeap<K>(
			new FeatureComparator<TemporalObject,K>(
				TemporalObject.START_TIMESTAMP_COMPARATOR, 
				new Function<K,TemporalObject>() {
					@Override
					public TemporalObject invoke(K key) {
						return groups.get(key).iterator().next();
					}
				}
			)
		);
	}	
	
	public TemporalGroupAndAggregator(Source<? extends TemporalObject<I>> source, int sourceID, Function<? super I,? extends K> determineGroup, Function<?,? extends TemporalAggregatorSA<I,V>> newTemporalGroupAndAggregatorSA) {
		this(determineGroup, newTemporalGroupAndAggregatorSA);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
	
	public TemporalGroupAndAggregator(Source<? extends TemporalObject<I>> source, Function<? super I,? extends K> determineGroup, final AggregationFunction<? super I,V> aggFunction) {
		this(source, DEFAULT_ID, determineGroup, new Function<Object,TemporalAggregatorSA<I,V>>() {
			@Override
			public TemporalAggregatorSA<I,V> invoke() {
				return new TemporalAggregatorSA<I,V>(aggFunction);
			}
		});
	}
	
	@Override
	public void processObject(TemporalObject<I> o, int sourceID) throws IllegalArgumentException {
		K groupID = determineGroup.invoke(o.getObject());
		if (!groups.containsKey(groupID)) 
			groups.put(groupID, newSweepArea.invoke());
		boolean emptySA = groups.get(groupID).size() > 0 ? false : true;
		groups.get(groupID).insert(o);
		if (emptySA)
			groupsOrderedByStartTimestamps.enqueue(groupID);
		transferHeap(addExpiredElementsToHeap(o));
	}
	
	protected long addExpiredElementsToHeap(TemporalObject<I> o) {
		long minTS = Long.MIN_VALUE;
		while (!groupsOrderedByStartTimestamps.isEmpty()) {	
			K groupID = groupsOrderedByStartTimestamps.peek();
			long nextTS = groups.get(groupID).iterator().next().getStart();
			if (minTS != Long.MIN_VALUE && minTS == nextTS)
				break;
			minTS = nextTS;
			if (minTS <= o.getStart()) {
				Iterator<TemporalObject> it = groups.get(groupsOrderedByStartTimestamps.dequeue()).expire(o, 0);
				while(it.hasNext()) {
					TemporalObject<V> next = (TemporalObject<V>)it.next();
					TemporalObject<Entry<K,V>> newTSO = new TemporalObject<Entry<K,V>>(
						new MapEntry<K,V>(groupID, next.getObject()), 
						next.getTimeInterval()
					);
					heap.enqueue(newTSO);
				}
				if (groups.get(groupID).size() > 0)
					groupsOrderedByStartTimestamps.enqueue(groupID);
			}
			else break;
		}
		return minTS;
	}
	
	protected void addRemainingElementsToHeap() {
		while (!groupsOrderedByStartTimestamps.isEmpty()) {
			K groupID = groupsOrderedByStartTimestamps.dequeue();
			Iterator<TemporalObject> it = groups.get(groupID).iterator();
			while(it.hasNext()) {
				TemporalObject<V> next = (TemporalObject<V>)it.next();
				TemporalObject<Entry<K,V>> newTSO = new TemporalObject<Entry<K,V>>(
					new MapEntry<K,V>(groupID, next.getObject()), 
					next.getTimeInterval()
				);
				heap.enqueue(newTSO);
			}
		}
	}
	
	@Override
	public void close() {
		super.close();
		if (isClosed()) {
			groupsOrderedByStartTimestamps.clear();
		}
	}
	
	/**
	 * @param args
	 */
	@SuppressWarnings("unchecked")
	public static void main(String[] args) {
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 5000;
		final int buckets = 1117;
		final int startInc = 50;
		final int intervalSize = 750;
		final long seed = 42;
		
		RandomNumber<Integer> r = new RandomNumber<Integer>(
			RandomNumber.DISCRETE, noOfElements, 0
		);		
		
		Source<TemporalObject<Integer>> s = new Mapper<Integer,TemporalObject<Integer>>(r,
			new Function<Integer,TemporalObject<Integer>>() {
				Random random = new Random(seed);
				long start, end;
				long newStart, newEnd;
				@Override
				public TemporalObject<Integer> invoke(Integer o) {
					newStart = start+random.nextInt(startInc);
					newEnd   = newStart+1+random.nextInt(intervalSize-1);
					if (start == newStart) newEnd = end+random.nextInt(intervalSize-1);
					TemporalObject<Integer> object = new TemporalObject<Integer>(o, 
						new TimeInterval(start = newStart, end = newEnd)
					);
					return object;
				}
			}
		);
	
		final Sum agg = new Sum();
		
		final Function<Integer,Integer> hashFunction1 = new Function<Integer,Integer>() {
			@Override
			public Integer invoke(Integer o) {
				return o % buckets;
			}
		};
		final Function<TemporalObject<Integer>,Integer> hashFunction2 = new Function<TemporalObject<Integer>,Integer>() {
			@Override
			public Integer invoke(TemporalObject<Integer> o) {
				return o.getObject()%buckets;
			}
		};
		
		// direct implementation
		TemporalGroupAndAggregator<Integer,Integer,Number> a = new TemporalGroupAndAggregator<Integer,Integer,Number>(s,
			hashFunction1, agg
		);
	
		// indirect implementation
		HashGrouper<TemporalObject<Integer>> grouper = new HashGrouper<TemporalObject<Integer>>(s, 0, hashFunction2, buckets);
		
		TemporalAggregator<Integer,Number>[] aggs = new TemporalAggregator[buckets];
		Mapper<TemporalObject<Number>,TemporalObject<Entry<Integer,Number>>>[] mappers = new Mapper[buckets];
		
		Function<?,Function<TemporalObject<Number>,TemporalObject<Entry<Integer,Number>>>> assignGroupFunction = new Function<Object,Function<TemporalObject<Number>,TemporalObject<Entry<Integer,Number>>>>() {
			protected int groupID = 0;
			
			@Override
			public Function<TemporalObject<Number>,TemporalObject<Entry<Integer,Number>>> invoke() {
				
				Function<TemporalObject<Number>,TemporalObject<Entry<Integer,Number>>> assignGroup = new Function<TemporalObject<Number>,TemporalObject<Entry<Integer,Number>>>() {
					
					protected final int id = groupID;
					
					@Override
					public TemporalObject<Entry<Integer,Number>> invoke(TemporalObject<Number> tso) {
						return new TemporalObject<Entry<Integer,Number>>(
								new MapEntry<Integer,Number>(new Integer(id), tso.getObject()), 
								tso.getTimeInterval()
						);
					}
				};
				groupID++;
				return assignGroup;
			}
		};		
		for (int i = 0; i < buckets; i++) {
			aggs[i] = new TemporalAggregator<Integer,Number>(grouper.getReferenceToGroup(i), new TemporalAggregatorSA<Integer,Number>(agg));
			mappers[i] = new Mapper<TemporalObject<Number>,TemporalObject<Entry<Integer,Number>>>(aggs[i], assignGroupFunction.invoke());
		}
		TemporalUnion<Entry<Integer,Number>> u = new TemporalUnion<Entry<Integer,Number>>(mappers);
	
		// check sorting
		Pipes.verifyStartTimeStampOrdering(a);
		Pipes.verifyStartTimeStampOrdering(u);
				
		AbstractSink sink = new AbstractSink<TemporalObject<MapEntry<Integer,Double>>>(new Source[]{a,u}, new int[]{1,2}) {
			LinkedList<TemporalObject<MapEntry<Integer,Double>>> list1 = new LinkedList<TemporalObject<MapEntry<Integer,Double>>>();
			LinkedList<TemporalObject<MapEntry<Integer,Double>>> list2 = new LinkedList<TemporalObject<MapEntry<Integer,Double>>>();
			
			@Override
			public void processObject(TemporalObject<MapEntry<Integer,Double>> o, int sourceID) {				
				if (sourceID == 1) {
					//System.out.println(o);
					list1.add(o);
					return;
				}
				if (sourceID == 2) {
					list2.add(o);
					return;
				}
			}
			
			@Override
			public void done(int sourceID) {
				super.done(sourceID);
				if (isDone()) {
					processingWLock.lock();
					try {
						System.out.println("list1: "+list1);
						System.out.println("list2: "+list2);
						Iterator it1 = list1.iterator();
						while (it1.hasNext()) {
							Object next1 = it1.next();
							if (!list2.contains(next1)) 
								System.err.println("ERROR: Missing results in indirect implementation: "+next1);
							else 
								list2.remove(next1);
						}
						Iterator it2 = list2.iterator();
						while (it2.hasNext()) 
							System.err.println("ERROR: Missing results in direct implementation: "+it2.next());
						System.out.println("CHECKS FINISHED.");
					}
					finally {
						processingWLock.unlock();
					}
				}
			}
		};

		Tester t1 = new Tester<TemporalObject<Entry<Integer,Number>>>(a);
		Tester t2 = new Tester<TemporalObject<Entry<Integer,Number>>>(u);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(sink);
		exec.registerQuery(t1);
		exec.registerQuery(t2);
		exec.startAllQueries();		
	}

}
