/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.groupers;

import java.util.Enumeration;
import java.util.Hashtable;

import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.queues.Heap;
import xxl.core.collections.sweepAreas.SweepArea;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.sources.Source;
import xxl.core.util.XXLSystem;

/**
 *
 */
public abstract class TemporalGrouper<I,K,O> extends AbstractTemporalPipe<I,O> implements MemoryMonitorable {
		
	protected Hashtable<K,SweepArea<TemporalObject>> groups;
	protected Function<? super I, ? extends K> determineGroup;
	protected Function<?,? extends SweepArea<TemporalObject>> newSweepArea;
	protected Heap<TemporalObject<O>> heap;
	protected boolean checkedHeapObjectSize;
	protected int heapObjectSize;
	

	public TemporalGrouper(Function<? super I,? extends K> determineGroup, Function<?,? extends SweepArea<TemporalObject>> newSweepArea) {
		this.groups = new Hashtable<K,SweepArea<TemporalObject>>();
		this.determineGroup = determineGroup;
		this.newSweepArea = newSweepArea;
		this.heap = new DynamicHeap<TemporalObject<O>>(TemporalObject.START_TIMESTAMP_COMPARATOR);
		this.checkedHeapObjectSize = false;
		this.heapObjectSize = SIZE_UNKNOWN;
	}	
	
	public TemporalGrouper(Source<? extends TemporalObject<I>> source, int sourceID, Function<? super I,? extends K> determineGroup, Function<?,? extends SweepArea<TemporalObject>> newSweepArea) {
		this(determineGroup, newSweepArea);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
	
	protected abstract long addExpiredElementsToHeap(TemporalObject<I> o);
	
	protected abstract void addRemainingElementsToHeap();
	
	@Override
	public void close() {
		super.close();
		if (isClosed()) {
			Enumeration<SweepArea<TemporalObject>> groupEnum = groups.elements();
			while(groupEnum.hasMoreElements())
				groupEnum.nextElement().close();
			heap.close();
		}
	}
		
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
	    	if (isDone()) {
	    		addRemainingElementsToHeap();
		    	while(!isClosed() && !heap.isEmpty())
					super.transfer(heap.dequeue());
		    	signalDone();
	    	}
	    }
	    finally {
	    	processingWLock.unlock();
	    }
	}

	public int getCurrentMemUsage() {
		Enumeration<SweepArea<TemporalObject>> groupEnum = groups.elements();
		int size = 0;
		while(groupEnum.hasMoreElements()) {
			SweepArea<TemporalObject> next = groupEnum.nextElement();
			int mem = next.getCurrentMemUsage();
			if (mem == SIZE_UNKNOWN)
				return SIZE_UNKNOWN;
			size += mem;
		}
		if (!checkedHeapObjectSize && !heap.isEmpty()) {
			try {	
				heapObjectSize = XXLSystem.getObjectSize(heap.peek());
			} catch (IllegalAccessException e) {
				heapObjectSize = MemoryMonitorable.SIZE_UNKNOWN;
			}
			checkedHeapObjectSize = true;
		}
		size += heap.size()*heapObjectSize;
		return size;
	}
	
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {		
		long ret = addExpiredElementsToHeap(new TemporalObject<I>(null, new TimeInterval(minTS, Long.MAX_VALUE)));
		if (transferHeap(ret))
			return NO_HEARTBEAT;
		return ret;
	}
	
	protected boolean transferHeap(long min) {
		boolean ret = false;
		while (!heap.isEmpty() && heap.peek().getStart() <= min) {
			transfer(heap.dequeue());
			ret = true; 
		}
		return ret;
	}
	
}

