/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.identities;

import java.util.Arrays;

import xxl.core.collections.queues.ListQueue;
import xxl.core.collections.queues.Queue;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.AbstractSink;
import xxl.core.pipes.sources.Enumerator;
import xxl.core.pipes.sources.Source;


/**
 *
 */
public class BlockBasedBufferPipe<I> extends AbstractBufferPipe<I,I[]>{

	protected I[] next;
	
	public BlockBasedBufferPipe(Queue<I> buffer, int batchSize, int objectSize) {
		super(buffer, batchSize, objectSize);
	}
	
	public BlockBasedBufferPipe(Source<? extends I>[] sources, int[] sourceIDs, Queue<I> buffer, int batchSize, int objectSize) {
		super(sources, sourceIDs, buffer, batchSize, objectSize);
	}

	@SuppressWarnings("unchecked")
	public BlockBasedBufferPipe(Source<? extends I> source, int sourceID, Queue<I> buffer, int batchSize, int objectSize) {
		this(new Source[]{source}, new int[]{sourceID}, buffer, batchSize, objectSize);
	}
	
	@SuppressWarnings("unchecked")
	public BlockBasedBufferPipe(Source<? extends I> source, int sourceID, int batchSize, int objectSize) {
		this(new Source[]{source}, new int[]{sourceID}, new ListQueue<I>(), batchSize, objectSize);
	}
	
	@SuppressWarnings("unchecked")
	public BlockBasedBufferPipe(Source<? extends I> source, int sourceID, int batchSize) {
		this(new Source[]{source}, new int[]{sourceID}, new ListQueue<I>(), batchSize, MemoryMonitorable.SIZE_UNKNOWN);
	}
	
	@SuppressWarnings("unchecked")
	public BlockBasedBufferPipe(Source<? extends I> source, int batchSize) {
		this(new Source[]{source}, new int[]{DEFAULT_ID}, new ListQueue<I>(), batchSize, MemoryMonitorable.SIZE_UNKNOWN);
	}
	
	@SuppressWarnings("unchecked")
	public BlockBasedBufferPipe(Source<? extends I> source) {
		this(new Source[]{source}, new int[]{DEFAULT_ID}, new ListQueue<I>(), 1, MemoryMonitorable.SIZE_UNKNOWN);
	}	
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.identities.AbstractBufferPipe#execute()
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void execute() {
		if (isClosed || signalledDone) return;
		synchronized(buffer) {
			int size;
			if (buffer.size() >= batchSize)
				size = batchSize;
			else 
				if (isDone) size = buffer.size();
				else return;
			if (size > 0) {
				next = (I[])new Object[size];
				for (int i = 0; i < size; i++) 
					next[i] = buffer.dequeue();
				transfer(next);
			}
			if (isDone && buffer.size() == 0 && !signalledDone) {
				signalDone();
				signalledDone = true;
			}
		}
	}
	
	protected int transferBuffer() {
		int size = 0;
		if (buffer.size() >= batchSize)
			size = batchSize;
		else 
			if (isDone) size = buffer.size();
			else return size;
		if (size > 0) {
			next = (I[])new Object[size];
			for (int i = 0; i < size; i++) 
				next[i] = buffer.dequeue();
			transfer(next);
		}
		return size;
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.identities.AbstractBufferPipe#isFinished()
	 */
	@Override
	public boolean isFinished() {
		processingRLock.lock();
		try {
			return isClosed || signalledDone;
		}
		finally {
			processingRLock.unlock();
		}
	}
	
	@SuppressWarnings("unchecked")
	public static void main(String[] args) {
		final BlockBasedBufferPipe<Double> buffer = new BlockBasedBufferPipe(
				new Enumerator(100, 0), 4);

		AbstractSink print = new AbstractSink(buffer) {
			@Override
			public void processObject(Object o, int sourceID)
					throws IllegalArgumentException {
				System.out.println(Arrays.toString((Object[]) o));
			}
		};

		Processor p = new Processor(0) {
			@Override
			public void process() {
				buffer.execute();
				if (buffer.isFinished())
					terminate();
			}

		};
		p.start();

		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(print);
		exec.startQuery(print);
	}
	
}
