/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.identities;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.sweepAreas.ListSA;
import xxl.core.collections.sweepAreas.ListSAImplementor;
import xxl.core.collections.sweepAreas.SweepAreaImplementor;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.AbstractSink;
import xxl.core.pipes.sinks.AbstractTemporalSink;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;
import xxl.core.predicates.Or;
import xxl.core.predicates.Predicate;
import xxl.core.predicates.Predicates;


public class SplitNEW<E> extends AbstractTemporalPipe<E, E> implements MemoryMonitorable {
	 
	
	public static class SplitSA<E> extends ListSA<TemporalObject<E>> {
		
		public SplitSA(SweepAreaImplementor<TemporalObject<E>> impl, Predicate<? super TemporalObject<E>> queryPredicate, Predicate<? super TemporalObject<E>> removePredicate, int objectSize, List<TemporalObject<E>> list) {
    		super(impl, DEFAULT_ID, true, ReorganizeModes.LAZY_REORGANIZE, Predicates.TRUE, new Or<TemporalObject<E>>(TemporalObject.INTERVAL_OVERLAP_REORGANIZE, removePredicate), 1, list, objectSize);
    	}
    	
    	public SplitSA(Function<? super TemporalObject<E>,Integer> hashFunction) {
    		super(new ListSAImplementor<TemporalObject<E>>(),
    			DEFAULT_ID, true, ReorganizeModes.LAZY_REORGANIZE,
    			Predicates.TRUE, TemporalObject.INTERVAL_OVERLAP_REORGANIZE, 
    			1, new ArrayList<TemporalObject<E>>(), SIZE_UNKNOWN
        	);
    	}
    	
    	public SplitSA() {
    		this(new Function<TemporalObject<E>,Integer>() {
    				@Override
					public Integer invoke(TemporalObject<E> o) {
    					return o.getObject().hashCode();
    				}
    			}
    		);
    	}
    	
    	@Override
		public void insert(TemporalObject<E> o) throws IllegalArgumentException {
    		if (checkObjectSize)
    			computeObjectSize(o);
    		impl.insert(o);
    		// insert into list ordered by end timestamps
    		int index = Collections.binarySearch(list, o, TemporalObject.END_TIMESTAMP_COMPARATOR);
    		list.add(index < 0 ? -index-1 : index, o);
    	}
    	   	
	}
	
	protected SplitSA<E> sweepArea;
	protected long l;
	protected long latestTS;
	protected DynamicHeap<TemporalObject<E>> heap;
	    
	    
    public SplitNEW(SplitSA<E> sweepArea, long l) {
    	this.sweepArea = sweepArea;
    	this.l = l;
    	this.latestTS = Long.MAX_VALUE;
    	heap = new DynamicHeap<TemporalObject<E>>(TemporalObject.START_TIMESTAMP_COMPARATOR);
    }
    
    public SplitNEW(Source<? extends TemporalObject<E>> source, int sourceID, SplitSA<E> sweepArea, long l) {
        this(sweepArea, l);
        if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
    }
    
    public SplitNEW(Source<? extends TemporalObject<E>> source, long l) {
        this(source, DEFAULT_ID, new SplitSA<E>(), l);
    }
	
	@Override
	public void processObject(TemporalObject<E> o, int sourceID) throws IllegalArgumentException {	
		if (latestTS != Long.MAX_VALUE)
			splitUp(sweepArea.query(o, sourceID), o.getStart());
		if (o.getEnd() - o.getStart() > l) {
			transfer(new TemporalObject<E>(o.getObject(), new TimeInterval(o.getStart(), o.getStart()+l)));
			sweepArea.insert(o);
		}
		else 
			transfer(o);
		
		latestTS = o.getStart();
		sweepArea.reorganize(o, sourceID);
	}
		
	protected boolean splitUp(Iterator<TemporalObject<E>> qualifies, long start) {
		boolean transferred = false;
		while (qualifies.hasNext()) {
			TemporalObject<E> next = qualifies.next();
			long lastSplit = next.getStart() + ((latestTS - next.getStart())/l)*l;	
			long newSplit = lastSplit + l;
			while (newSplit+l < next.getEnd() && newSplit <= start) {
				heap.enqueue(new TemporalObject<E>(next.getObject(), new TimeInterval(newSplit, newSplit+l)));
				newSplit += l;
			}
			if (newSplit < next.getEnd() && newSplit <= start) {
				heap.enqueue(new TemporalObject<E>(next.getObject(), new TimeInterval(newSplit, next.getEnd())));
			}
		}
		if (heap.size() > 0) transferred = true;
		while (!heap.isEmpty()) {
			transfer(heap.dequeue());
		}
		return transferred;
	}
	
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {		
				splitUp(sweepArea.iterator(), Long.MAX_VALUE);
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	@Override
	public void close() {
		super.close();
		if (isClosed()) {
			sweepArea.close();
		}
	}
	
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {
		boolean transferred = splitUp(sweepArea.iterator(), minTS);
		sweepArea.reorganize(new TemporalObject<E>(null, new TimeInterval(minTS, TimeInterval.INFINITY)), sourceID);
		return transferred ? NO_HEARTBEAT : minTS;
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 2500;
		final int buckets = 10;
		final int startInc = 35;
		final int intervalSize = 125;
		final long seed = 42;
		final HashMap<Integer,List<Long>> input = new HashMap<Integer,List<Long>>();
		final Integer[] hashCodes = new Integer[buckets];
		for (int i = 0; i < buckets; i++)
			hashCodes[i] = new Integer(i);
	
		RandomNumber<Integer> r = new RandomNumber<Integer>(
			RandomNumber.DISCRETE, noOfElements, 0
		);			
		Source<TemporalObject<Integer>> s = Pipes.decorateWithRandomTimeIntervals(r, input, hashCodes, startInc, intervalSize, seed);
		
		final long l = 15;
		SplitNEW<Integer> split = new SplitNEW<Integer>(s, l);
		
		// check sort order
		Pipes.verifyStartTimeStampOrdering(split);

		AbstractSink sink = new AbstractSink<TemporalObject<Integer>>(split) {
			LinkedList<TemporalObject<Integer>> list = new LinkedList<TemporalObject<Integer>>();
			
			@Override
			public void processObject(TemporalObject<Integer> o, int ID) {
				list.add(o);
			}
			
			@Override
			public void done(int sourceID) {
				super.done(sourceID);
				if (isDone()) {
					Iterator<TemporalObject<Integer>> listIt = list.iterator();
					HashMap<Integer,List<Long>> result = new HashMap<Integer,List<Long>>();
					while (listIt.hasNext()) {
						TemporalObject<Integer> t = listIt.next();
						int key = t.getObject().intValue();
						List<Long> l = result.get(hashCodes[key]);
						if (l == null) l = new LinkedList<Long>();
						l.addAll(TemporalObject.getAllSnapshots(t));					
						result.put(hashCodes[key], l);
					}
					// comparison 
					for (int i = 0; i < buckets; i++) {
						List<Long> in = input.get(hashCodes[i]);
						List<Long> out = result.get(hashCodes[i]);
						if (in != null && out != null) {
							Iterator<Long> it = in.iterator();					
							while (it.hasNext()) {
								Long next = it.next();
								if (!out.remove(next)) 
									System.err.println("ERROR: element lost: value "+i+" at time "+next);						
							}
							System.out.println("CHECK 1 finished for bucket "+i+".");
							
							it = in.iterator();
							while(it.hasNext()) {
								Long next = it.next();
								if (out.contains(next)) 
									System.err.println("ERROR: more elements than expected: value "+i+" at time "+next);
							}
							System.out.println("CHECK 2 finished for bucket "+i+".");
						}
						else {
							if (in != null && out == null)
								System.err.println("ERROR: Input, but no output. (i = "+i+")");
							if (in == null && out != null)
								System.err.println("ERROR: Output, but no input. (i = "+i+")");
						}
					}
				}
			}
		};
		Printer printer = new Printer<TemporalObject<Integer>>(split);
		
		AbstractTemporalSink<Integer> testLength = new AbstractTemporalSink<Integer>(split) {
						
			@Override
			public void processObject(TemporalObject<Integer> o, int sourceID) throws IllegalArgumentException {
				if (o.getEnd() - o.getStart() > l)
					System.err.println("ERROR: Interval too long: "+o);
			}
		};
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(sink);
		exec.registerQuery(printer);
		exec.registerQuery(testLength);
		exec.startAllQueries();
		
    }

}
