/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.joins;

import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import xxl.core.collections.containers.io.BlockFileContainer;
import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.collections.sweepAreas.ImplementorBasedSweepArea;
import xxl.core.collections.sweepAreas.MemoryManageableSA;
import xxl.core.cursors.sources.EmptyCursor;
import xxl.core.functions.Function;
import xxl.core.io.Block;
import xxl.core.io.converters.Converter;
import xxl.core.io.converters.IntegerConverter;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.GlobalMemoryManager;
import xxl.core.pipes.memoryManager.UniformStrategy;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.processors.SourceProcessor;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sinks.Tester;
import xxl.core.pipes.sources.Enumerator;
import xxl.core.pipes.sources.Source;
import xxl.core.predicates.Equal;
import xxl.core.predicates.Predicate;
import xxl.core.predicates.Predicates;
import xxl.core.util.metaData.CompositeMetaData;
import xxl.core.util.timers.TimerUtils;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;


/** 
 * Operator component in a query graph that performs the MJoin operation over two
 * or more input streams.
 * 
 * @since 1.1
 */ 
public class MJoin<I,O> extends MultiWayEquiJoin<I,O> {

	// TODO: Generics, Design, Busy waiting
	
    /** 
     * One data tuple with arrival and departure timestamps for the MJoin.
     */ 
    private static class MJoinObject<T> extends TemporalObject<T> {
	    /** 
     	* Size overhead of an MJoinObject over the contained object.
     	*/ 
        private static final int OBJECT_SIZE_OVERHEAD = 16;
        
	    /** 
     	* Constructs an MJoinObject.
     	*
     	* @param obj data tuple.
     	* @param ats arrival timestamp.
     	*/ 
        public MJoinObject(T obj, Long ats) {
    		this(obj, ats.longValue());
    	}

	    /** 
     	* Constructs an MJoinObject.
     	*
     	* @param obj data tuple.
     	* @param ats arrival timestamp.
     	*/ 
    	public MJoinObject(T obj, long ats) {
    		super(obj, new TimeInterval(ats, Long.MAX_VALUE));
    	}

	    /** 
     	* Returns the arrival timestamp (ATS).
     	*
     	* @return the arrival timestamp (ATS).
     	*/ 
    	public long getATS() {
    		return getStart();
    	}

	    /** 
     	* Returns the departure timestamp (DTS).
     	*
     	* @return the departure timestamp (DTS).
     	*/ 
    	public long getDTS() {
    		return getEnd();
    	}

	    /** 
     	* Sets the departure timestamp (DTS).
     	*
     	* @param dts departure timestamp (DTS).
     	*/ 
    	public void setDTS(Long dts) {
    		setDTS(dts.longValue());
    	}

	    /** 
     	* Sets the departure timestamp (DTS).
     	*
     	* @param dts departure timestamp (DTS).
     	*/ 
    	public void setDTS(long dts) {
    		if (getEnd() != Long.MAX_VALUE)
    			throw new IllegalStateException("The DTS may only be set once for an object.");
    		setTimeInterval(new TimeInterval(getATS(), dts));
    	}

	    /** 
     	* Checks if this object's time interval contains the specified timestamp.
     	*
     	* @param ts timestamp.
     	*
     	* @return true if the interval contains ts, false otherwise.
     	*/ 
    	public boolean timeIntervalContains(long ts) {
    		return getTimeInterval().contains(ts);
    	}
    	
	    /** 
     	* Returns a string representation of the data tuple and its timestamps.
     	*
     	* @return a string.
     	*/ 
    	@Override
		public String toString() {
    		return "[" + getObject().toString() + "," + getStart() + "," + (getEnd() == Long.MAX_VALUE ? "-" : getEnd() + "") + "]";
    	}

    	// Serialization

    	private MJoinObject(T obj, long ats, long dts) {
    		super(obj, new TimeInterval(ats, dts));
    	}

	    /** 
     	* Serializes this MJoinObject to dataOutput using a specified
     	* converter.
     	*
     	* @param dataOutput stream to write to.
     	* @param converter converter to use.
     	*/ 
    	public void write(DataOutput dataOutput, Converter converter) throws IOException {
    		dataOutput.writeLong(getStart());
    		dataOutput.writeLong(getEnd());
    		converter.write(dataOutput, getObject());
    	}

	    /** 
     	* Deserializes an MJoinObject from dataInput using a specified
     	* converter.
     	*
     	* @param dataInput stream to read from.
     	* @param converter converter to use.
     	*/ 
    	public static MJoinObject read(DataInput dataInput, Converter converter) throws IOException {
    		long ats = dataInput.readLong();
    		long dts = dataInput.readLong();
    		Object obj = converter.read(dataInput);

    		return new MJoinObject(obj, ats, dts);
    	}

    }
    
    /** 
     * Sweep area for the the MJoin. This sweep area stores MJoinObjects.
     */ 
    /**
     * @author heinzch
     *
     */
    public static class MJoinSweepArea extends MemoryManageableSA {
    	/**
    	 * Constructs an MJoinSweepArea.
    	 *
    	 * @param ID the ID used to distinguish multiple SweepAreas if necessary
    	 * @param predicates binary predicates used for search operations.
    	 * @param hashFunctions the hash-functions: f: Object --> Integer.
    	 * @param objectSize the size of the objects to be stored.
    	 * @param memSize initial memory size for the sweep area.
    	 * @param blockSize size of the blocks for the external part of the sweep area.
    	 */
    	public MJoinSweepArea(int ID, final Predicate[] predicates, final Function[] hashFunctions, int objectSize, int memSize, int blockSize) {
    		super(	new ImplementorBasedSweepArea(
    					new MJoinSAImplementor(hashFunctions, new MJoinDiskResidentPortion("temp" + ID, blockSize, objectSize)),
    					ID,
    					false,
    					predicates),
    				objectSize + MJoinObject.OBJECT_SIZE_OVERHEAD,
    				memSize);
    		assignMemSize(memSize);
    	}

    	/**
    	 * Sets the converter used to write and read objects to/from disk.
    	 *
    	 * @param converter converter to use.
    	 */
    	public void setConverter(Converter converter) {
    		((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).setConverter(converter);
    	}	

    	/**
    	 * Sets the function that returns the next timestamp.
    	 *
    	 * @param function a function that returns the next timestamp (a Long object).
    	 */
    	public void setTimestampFunction(Function getNextTimestamp) {
    	    ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).setTimestampFunction(getNextTimestamp);
    	}	

    	/**
    	 * Returns the number of partitions in this sweep area.
    	 *
    	 * @return the number of partitions in this sweep area.
    	 */
    	public int getNumberOfPartitions() {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getNumberOfPartitions();
    	}

    	/**
    	 * Returns the index of the partition to which the specified object hashes.
    	 *
    	 * @param object object to check.
    	 *
    	 * @return index of the partition to which the object hashes.
    	 */
    	public int getPartition(Object object) {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getPartition(object);
    	}

    	/**
    	 * Loads the specified partition into memory, i.e. appends all objects
    	 * from the disk resident portion to the memory resident portion.
    	 *
    	 * @param partition index of the desired partition.
    	 */
    	public void loadPartition(int partition) {
    	    ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).loadPartition(partition);
    	}

    	/**
    	 * Deletes all objects from the memory resident portion of the specified
    	 * partition.
    	 *
    	 * @param partition index of the desired partition.
    	 */
    	public void clearMemPartition(int partition) {
    	    ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).clearMemPartition(partition);
    	}
    	
    	/**
    	 * Returns the size of the memory resident portion
    	 * of the specified partition.
    	 *
    	 * @param partition index of the desired partition.
    	 *
    	 * @return number of objects in the specified partition.
    	 */
    	public int getMemPartitionSize(int partition) {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getMemPartitionSize(partition);
    	}

    	/**
    	 * Returns the size of the disk resident portion
    	 * of the specified partition.
    	 *
    	 * @param partition index of the desired partition.
    	 *
    	 * @return number of objects in the specified partition.
    	 */
    	public int getDiskPartitionSize(int partition) {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getDiskPartitionSize(partition);
    	}

    	/**
    	 * Returns the size of the largest disk resident portion of any partition.
    	 *
    	 * @return number of objects in the largest partition.
    	 */
    	public int getLargestDiskPartitionSize() {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getLargestDiskPartitionSize();
    	}

    	/**
    	 * Returns the index of the partition with the largest disk resident portion.
    	 *
    	 * @return index of the partition with the largest disk resident portion.
    	 */
    	public int getLargestDiskPartition() {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getLargestDiskPartition();
    	}

    	/**
    	 * Returns an iterator over the objects in the memory resident portion of the
    	 * specified partition.
    	 *
    	 * @param partition index of the desired partition.
    	 *
    	 * @return iterator over the objects.
    	 */
    	public Iterator getMemPartition(int partition) {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getMemPartition(partition);
    	}

    	/**
    	 * Returns an iterator over the objects in the disk resident portion of the
    	 * specified partition.
    	 *
    	 * @param partition index of the desired partition.
    	 *
    	 * @return iterator over the objects.
    	 */
    	public Iterator getDiskPartition(int partition) {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getDiskPartition(partition);
    	}

    	/**
    	 * Adds an entry to the probe list of the specified partition.
    	 *
    	 * @param partition index of desired partition.
    	 * @param entry array containing two timestamps: Probe-TS, DTS-Last.
    	 */
    	public void addProbeListEntry(int partition, Object[] entry) {
    	    ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).addProbeListEntry(partition, entry);
    	}

    	/**
    	 * Flushes the memory resident portion of the specified partition, i.e.
    	 * writes all objects to the disk resident portion and then clears
    	 * the memory resident portion.
    	 *
    	 * @param partition index of desired partition; 0 to choose the largest
    	 * partition
    	 */
    	public void flushToDisk(int partition) {
    		try {
    		    ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).flushToDisk(partition);
    		} catch (java.io.IOException e) {
    			System.out.println(e);
    			System.exit(0);
    		}
    	}

    	/**
    	 * Returns the probe list of the specified partition.
    	 *
     	 * @param partition index of desired partition.
     	 *
     	 * @return iterator over Object[]-arrays. Each array contains two
     	 *  timestamps: Probe-TS, DTS-Last.
    	 */
    	public Iterator getProbeList(int partition) {
    		return ((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).getProbeList(partition);
    	}
    	
		/**
	 	 * This method is called if a memory overflow has occured.
	 	 */
    	@Override
		public void handleOverflow() {
			while (getCurrentMemUsage() > assignedMemSize)
	    		flushToDisk(0);
		}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.DecoratorSweepArea#close()
    	 */
    	@Override
		public void close() {
    		super.close();
    		((MJoinSAImplementor)((ImplementorBasedSweepArea)sweepArea).getImplementor()).close();
    	}
		
		/* (non-Javadoc)
		 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#getObjectSize()
		 */
		public int getObjectSize() {
			return objectSize;
		}
    	
    }
    
    /** 
     * Implementation of the MJoin sweep area.
     */ 
    public static class MJoinSAImplementor extends HashSAImplementor {
    	// disk resident portion of this sweep area  	
        private MJoinDiskResidentPortion diskPortion;
        // function that gets the next timestamp
    	private Function getNextTimestamp;
    	// probe list for this sweep area
    	private ArrayList probeList = new ArrayList();
    	
    	/**
    	 * Decorator for the SA predicate that checks whether any argument
    	 * is an MJoinObject and if so, passes the actual contained object
    	 * on to the real predicate.
    	 */
    	private Predicate predicateDecorator(final Predicate predicate) {
    		return new Predicate() {
    			@Override
				public boolean invoke(Object o1, Object o2) {
    				if (o1 instanceof TemporalObject)
    					o1 = ((TemporalObject)o1).getObject();
    				if (o2 instanceof TemporalObject)
    					o2 = ((TemporalObject)o2).getObject();
    				return predicate.invoke(o1, o2);
    			}
    		};
    	}

    	/**
    	 * Decorator for the SA hash function that checks whether the argument
    	 * is an MJoinObject and if so, passes the actual contained object
    	 * on to the real hash function.
    	 */
    	private Function hashFunctionDecorator(final Function hashFunction) {
    		return new Function() {
    			@Override
				public Object invoke(Object o) {
    				if (o instanceof TemporalObject)
    					o = ((TemporalObject)o).getObject();
    				return hashFunction.invoke(o);
    			}
    		};
    	}

    	/**
    	 * Constructs an MJoinSAImplementor.
    	 *
    	 * @param hashFunctions array of hash functions
    	 * @param diskPortion disk resident portion of this sweep area
    	 */
    	public MJoinSAImplementor(final Function[] hashFunctions, MJoinDiskResidentPortion diskPortion) {
    		super(	null,
    				new Function() {
    					@Override
						public Object invoke() {
    						return new LinkedList();
    					}
    				});
    		
    		// decorate hash functions so that they can process MJoinObjects
    		this.hashFunctions = new Function[hashFunctions.length];
    		for (int i = 0; i < hashFunctions.length; i++)
    			this.hashFunctions[i] = hashFunctionDecorator(hashFunctions[i]);

    		this.diskPortion = diskPortion;
    	}
    	
    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSAImplementor#setQueryPredicates(int, xxl.core.predicates.Predicate<? super E>[])
    	 */
    	@Override
		public void setQueryPredicates(int ID, Predicate[] predicates) {
    	    // decorate predicates so they can process MJoinObjects
    		Predicate[] predicatesDec = new Predicate[predicates.length];
    		for (int i = 0; i < predicates.length; i++)
    			predicatesDec[i] = predicateDecorator(predicates[i]);
    	    
    	    this.predicates = predicatesDec;
    		this.ID = ID;
    	}
    	
    	/**
    	 * @param converter
    	 */
    	public void setConverter(Converter converter) {
    		diskPortion.setConverter(converter);
    	}	

    	/**
    	 * @param getNextTimestamp
    	 */
    	public void setTimestampFunction(Function getNextTimestamp) {
    		this.getNextTimestamp = getNextTimestamp;
    	}	
    	
    	/**
    	 * @param partition
    	 * @param entry
    	 */
    	public void addProbeListEntry(int partition, Object[] entry) {
    		if (partition >= probeList.size()) {
    			for (int i = probeList.size(); i <= partition; i++)
    				probeList.add(new LinkedList());
    		}
    		((List)probeList.get(partition)).add(entry);
    	}

    	/**
    	 * @param partition
    	 * @return
    	 */
    	public Iterator getProbeList(int partition) {
    		if (partition >= probeList.size())
    			return EmptyCursor.DEFAULT_INSTANCE;
    		
    		return ((List)probeList.get(partition)).iterator();
    	}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.HashListSAImplementor#insert(java.lang.Object)
    	 */
    	@Override
		public void insert(Object object) {
    		// insert object along with arrival timestamp (ATS)
    		super.insert(new MJoinObject(object, (Long)getNextTimestamp.invoke()));
    	}

    	public int getPartition(Object object) {
    		return ((Integer)hashFunctions[ID].invoke(object)).intValue();
    	}

    	public int getNumberOfPartitions() {
    		return hashTable.size();
    	}

    	public int getMemPartitionSize(int partition) {
    		return ((List)hashTable.get(partition)).size();
    	}

    	public int getDiskPartitionSize(int partition) {
    		return diskPortion.getPartitionSize(partition);
    	}

    	public int getLargestDiskPartitionSize() {
    		return diskPortion.getLargestPartitionSize();
    	}
    	
    	public int getLargestDiskPartition() {
    		return diskPortion.getLargestPartition();
    	}

    	public void loadPartition(int partition) {
    		List list = (List)hashTable.get(partition);
    		Iterator it = diskPortion.readPartition(partition);
    		while (it.hasNext())
    			list.add(it.next());
    	}

    	public void clearMemPartition(int partition) {
    		((List)hashTable.get(partition)).clear();
    	}

    	public Iterator getMemPartition(int partition) {
    		return ((List)hashTable.get(partition)).iterator();
    	}

    	public Iterator getDiskPartition(int partition) {
    		return diskPortion.readPartition(partition);
    	}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.HashListSAImplementor#close()
    	 */
    	@Override
		public void close() {
    		super.close();
    		diskPortion.close();
    	}

    	public void flushToDisk(int partition) throws java.io.IOException {
    		if (partition == 0)
    			partition = choosePartition();
    		List list = (List)hashTable.get(partition);
//    		System.out.print("Writing partition " + partitionNo + " of " + hashTable.size() + ", " + list.size() + " elements: ");

    		Iterator it = list.iterator();
    		while (it.hasNext()) {
    			MJoinObject elem = (MJoinObject)it.next();
    			elem.setDTS((Long)getNextTimestamp.invoke());
//    			System.out.print(elem.getElem()+" ");
    		}
//    		System.out.println();

    		diskPortion.writePartition(partition, list.iterator());
    		
    		list.clear();
    	}
    	
    	protected int choosePartition() {
    		int maxSize = 0, index = 0;
    		for (int i = 0; i < getNumberOfPartitions(); i++) {
    			int s = getMemPartitionSize(i);
    			if (s > maxSize) {
    				maxSize = s;
    				index = i;
    			}
    		}
    		return index;
    	}
    }
    
    /**
     * Implements the disk resident portion of a sweep area for the MJoin by using
     * a BlockFileContainer.
     */
   public static class MJoinDiskResidentPortion {
	   	
       // the BlockFileContainer used to store the partitions
	   	protected BlockFileContainer bfc;
	
	   	// size of each block in the BlockFileContainer
	   	private int blockSize;		
	
	   	// max. number of objects that can be stored in one block				
	   	private int objectsPerBlock;
	
	   	// contains a List for each partition; each list contains one entry for each block in the BFC.
	   	// A list entry is an array with two entries { blockID, noOfObjects }, where blockID is the
	   	// ID of the block in the BFC and noOfObjects is the number of objects stored in that block.
	   	private ArrayList blockTable = new ArrayList();
	
	   	// size (= number of blocks) of each partition
	   	private ArrayList partSize = new ArrayList();
	
	   	// number of largest partition and its size
	   	private int largestPartitionNo, largestPartitionSize;
	
	   	// converter used to write and read the objects into/from the BFC
	   	protected Converter converter;
	
	   	/**
	   	 * Constructs an MJoinDiskResidentPortion.
	   	 *
	   	 * @param prefix prefix of the file name of the BlockFileContainer.
	   	 * @param blockSize block size of the BlockFileContainer.
	   	 * @param objectSize size of the objects to be stored.
	   	 */
	   	public MJoinDiskResidentPortion(String prefix, int blockSize, int objectSize) {
	   		this.blockSize = blockSize;
	   		objectsPerBlock = blockSize / (objectSize + MJoinObject.OBJECT_SIZE_OVERHEAD);
	   		if (objectsPerBlock < 1)
	   			throw new IllegalArgumentException("blockSize too small. blockSize must be large enough for a block to contain at least one object.");
	   		bfc = new BlockFileContainer(prefix, blockSize);
	   	}
	
	   	/*
	   	 * Sets the converter used to write and read the objects into/from the BFC.
	   	 *
	   	 * @param converter converter to use.
	   	 */
	   	public void setConverter(Converter converter) {
	   		if (this.converter != null)
	   			throw new IllegalStateException("The converter cannot be changed after it was set.");
	   		this.converter = converter;
	   	}	
	
	   	public void close() {
	   		bfc.close();
	   		bfc.delete();
	   	}
	
	   	/**
	   	 * Returns the size of the specified partition.
	   	 *
	   	 * @param partition index of the desired partition.
	   	 *
	   	 * @return the number of objects stored in the specified partition.
	   	 */
	   	public int getPartitionSize(int partition) {
	   		if (partition >= partSize.size())
	   			return 0;
	   		return ((Integer)partSize.get(partition)).intValue();
	   	}
	
	   	private void incPartitionSize(int partition, int inc) {
	   		if (partition >= partSize.size())
	   			for (int i = partSize.size(); i <= partition; i++)
	   				partSize.add(new Integer(0));
	
	   		int newSize = ((Integer)partSize.get(partition)).intValue() + inc;
	   		partSize.set(partition, new Integer(newSize));
	
	   		if (newSize > largestPartitionSize) {
	   			largestPartitionSize = newSize;
	   			largestPartitionNo = partition;
	   		}
	   	}
	
	   	/**
	   	 * Returns the index of the largest partition.
	   	 *
	   	 * @return the index of the largest partition.
	   	 */
	   	public int getLargestPartitionSize() {
	   		return largestPartitionSize;
	   	}
	
	   	/**
	   	 * Returns the size of the largest partition.
	   	 *
	   	 * @return the number of objects stored in the largest partition.
	   	 */
	   	public int getLargestPartition() {
	   		return largestPartitionNo;
	   	}
	
	   	private List getBlockList(int partition) {
	   		if (blockTable.size() - 1 < partition) {
	   			for (int i = blockTable.size(); i <= partition; i++)
	   				blockTable.add(i, new LinkedList());
	   		}
	   		return (List)blockTable.get(partition);
	   	}
	
	   	/**
	   	 * Appends objects to a partition.
	   	 *
	   	 * @param partition index of desired partition.
	   	 * @param it Iterator over MJoinObjects to append.
	   	 */
	   	public void writePartition(int partition, Iterator it) throws java.io.IOException {
	   		List blockList = getBlockList(partition);
	
	   		while (it.hasNext()) {
	   			Block block = new Block(blockSize);
	   			DataOutputStream out = block.dataOutputStream();
	   			int i;
	
	   			for (i = 0; i < objectsPerBlock && it.hasNext(); i++)
	   				((MJoinObject)it.next()).write(out, converter);
	
	   			if (i != 0) {
	   				blockList.add(new Object[]{ bfc.insert(block), new Integer(i) });
	   				incPartitionSize(partition, i);
	   			}
	//   			System.out.println("wrote "+i+" elements");
	   		}
	   	}
	
	   	/**
	   	 * Returns an iterator over the objects in the specified partition. The
	   	 * objects are read from the hard disk on demand, i.e. only one block
	   	 * at a time is read into memory.
	   	 *
	   	 * @param partition index of desired partition.
	   	 *
	   	 * @return an iterator over the objects in the specified partition.
	   	 */
	   	public Iterator readPartition(int partition) {
	   		Iterator it = getBlockList(partition).iterator();
	
	   		if (!it.hasNext())
	   			return EmptyCursor.DEFAULT_INSTANCE;
	   			
	   		return new PartitionIterator(it);
	   	}
	
	   	private class PartitionIterator implements Iterator {
	   		private Iterator blockListIt;
	   		private Block block;
	   		private int noOfObjects;
	   		private int nextObject;
	   		private DataInputStream in;
	
	   		public PartitionIterator(Iterator blockListIt) {
	   			this.blockListIt = blockListIt;
	   			readNextBlock();
	   		}
	
	   		private void readNextBlock() {
	   			Object[] blockListEntry = (Object[])blockListIt.next();
	   			block = (Block)bfc.get(blockListEntry[0]);
	   			in = block.dataInputStream();
	   			noOfObjects = ((Integer)blockListEntry[1]).intValue();
	   			nextObject = 0;
	   		}
	
	   		public boolean hasNext() {
	   			return (nextObject < noOfObjects || blockListIt.hasNext());
	   		}
	
	   		public Object next() {
	   			if (nextObject == noOfObjects)
	   				readNextBlock();
	   			nextObject++;
	   			MJoinObject obj = null;
	   			try {
	   				obj = MJoinObject.read(in, converter);
	   			} catch (IOException e) {
	   				System.out.println(e);
	   				System.exit(0);
	   			}
	   			return obj;
	   		}
	
	   		public void remove() {
	   			throw new UnsupportedOperationException();
	   		}
	   	}
   }
    
    
    // scheduler that periodically tries to run stage two
	protected Processor stageTwoScheduler;

	protected volatile boolean stageTwoRunning = false;

	protected double lowInputThreshold;
	
	// queue for incoming objects during stage two
	protected LinkedList<Object[]> elementQueue = new LinkedList<Object[]>();

	protected long nextTimestamp = 1;

	protected Function getNextTimestamp = new Function() {
		@Override
		public Object invoke() {
			return new Long(nextTimestamp++);
		}
	};

	/**
	 * Constructs a new MJoin.
	 *
	 * @param sources the join sources in the query graph
	 * @param IDs IDs of the sources
	 * @param converter converter used to write/read elements to/from external storage
	 * @param newResult function that produces a result tuple
	 * @param predicates join predicates for each sweep area
	 * @param hashFunctions hash functions for each sweep area
	 * @param objectSize size of each object
	 * @param memSize memory size for each sweep area
	 * @param blockSize block size for external storage of objects
	 * @param stageTwoInterval time interval (in milliseconds) at which stage two is tried to run
	 * @param lowInputThreshold input rate below which stage two will be started
	 */
	public MJoin(Source<? extends I>[] sources, int[] IDs, Converter[] converter, final Function newResult,
				 final Predicate[] predicates, final Function[] hashFunctions,
				 final int objectSize, final int memSize, final int blockSize,
				 long stageTwoInterval, double lowInputThreshold) {
	    
	    super(sources, IDs, new MJoinSweepArea[sources.length],
			new Function<MJoinObject, Object>() {
				public Object invoke(List<? extends MJoinObject> results) {
					List<Object> tuple =  new ArrayList<Object>(results.size());
					for (int i = 0; i < results.size(); i++)
						if (results.get(i) instanceof MJoinObject)
							tuple.add(((MJoinObject)results.get(i)).getObject());
						else 
							tuple.add(results.get(i));
					return newResult.invoke(tuple);
				}
			}
	   );
	    MJoinSweepArea sa;
	    for (int i = 0; i < sweepAreas.length; i++) {
		    this.sweepAreas[i] = new MJoinSweepArea(i, predicates, hashFunctions, objectSize, memSize, blockSize);
			sa = getSweepArea(i);
			sa.setConverter(converter[i]);
			sa.setTimestampFunction(getNextTimestamp);
		}
	
	    metaDataManagement.include(INPUT_RATE);
	    ((AbstractPipeMetaDataManagement)metaDataManagement).setUpdatePeriod(stageTwoInterval / 3);
	    
		stageTwoScheduler = new Processor(stageTwoInterval) {
			@Override
			public void process() {
				tryStageTwo();
			}
		};
		this.lowInputThreshold = lowInputThreshold;
	}

	protected MJoinSweepArea getSweepArea(int j) {
		return (MJoinSweepArea)sweepAreas[j];
	}
	
	/**
	 * Opens this MJoin.
	 */
	@Override
	public void open() {
		if (isOpened) return;
		super.open();
		stageTwoScheduler.start();
	}


	/** 
	 * Inserts the specified element into the specified sweep area.
	 * If stage two is running, the element is added to a queue instead for
	 * later processing.
	 *
	 * @param o element to be inserted.
	 * @param j index (not ID!) of sweep area to be inserted into.
	 */
	@Override
	protected void insert(I o, int j) {
		// if we are in stage two, save object in queue for later processing
		if (stageTwoRunning) {
			elementQueue.add(new Object[]{o, new Integer(j) });
		}
		// otherwise, insert normally
		else
			super.insert(o, j);
	}

	/** 
	 * Generates the join results for the specified incoming element.
	 * If stage two is running, does nothing.
	 *
	 * @param o element to generate results with.
	 * @param j index (not ID!) of input stream the element comes from.
	 * @param predicate predicate to filter the results with.
	 */
	@Override
	protected void computeResults(Object o, int j, Predicate predicate) {
		// if we are in stage two, do nothing
		if (stageTwoRunning)
			return;
		// otherwise, compute results normally
		super.computeResults(o, j, predicate);
	}

	/** 
	 * Tries to run stage two. This method is called from the stage two scheduler.
	 * Calls <CODE>isInputAvailable()</CODE> to check the input rate.
	 * If no input is available, calls </CODE>runStageTwo</CODE>, afterwards processes
	 * all elements from the queue that arrived during stage two.
	 */
	protected void tryStageTwo() {
		if (!(isDone() || isInputAvailable())) {
			System.out.println("=> low input rate, running stage two");
			stageTwoRunning = true;
			
			runStageTwo();

/*
			try {
				java.lang.Thread.sleep(300);
			} catch (Exception e) {};
*/
			System.out.println("ending stage 2, " + elementQueue.size() + " elements in queue, starting cleanup");

			// process elements in queue
			Object[] arr;
			while (true) {
				processingWLock.lock();
				try {
					arr = (Object[])elementQueue.get(0);
					elementQueue.remove(0);
				} catch (IndexOutOfBoundsException e) {
					stageTwoRunning = false;
					break;
				}
				finally {
					processingWLock.unlock();
				}
				super.insert((I)arr[0], ((Integer)arr[1]).intValue());
				super.computeResults((I)arr[0], ((Integer)arr[1]).intValue(), Predicates.TRUE);
			}
			
			System.out.println("ending cleanup");
		}
		else
			System.out.println("=> high input rate or sources done, not running stage two");
	}

	/**
	 * Stage two of the MJoin.
	 */
	private void runStageTwo() {
		// choose largest disk resident partition across all sweep areas
		int max = 0, j = 0;
		for (int i = 0; i < sweepAreas.length; i++)
			if (getSweepArea(i).getLargestDiskPartitionSize() > max) {
				max = getSweepArea(i).getLargestDiskPartitionSize();
				j = i;
			}
		int partition = getSweepArea(j).getLargestDiskPartition();
//		System.out.println("SA "+j+", partition "+partition);

		Iterator itDisk = getSweepArea(j).getDiskPartition(partition);
		long dtsLast = 0;

		if (itDisk.hasNext()) {
			while (itDisk.hasNext()) {
				MJoinObject elem = (MJoinObject)itDisk.next();
				dtsLast = elem.getDTS();
//				System.out.println("probing with " + elem);
				super.computeResults(elem, j, duplCheck);
			}

			Long probeTS = (Long)getNextTimestamp.invoke();

			getSweepArea(j).addProbeListEntry(partition, new Object[]{ new Long(dtsLast), probeTS });
//			System.out.println("probeTS = " + probeTS);
		}
	}

	/**
	 * Stage three of the MJoin. Produces all results missed during
	 * stages one and two.
	 */
	private void runStageThree() {
		processingWLock.lock();
		try {
			System.out.println("starting stage 3:");

			int n = getNumberOfPartitions();
/*
			Iterator itDisk, itMem;
			for (int i = 0; i < sweepAreas.length; i++) {
				MJoinSweepArea sa = getSweepArea(i);
				System.out.println("SweepArea " + i);
				for (int p = 0; p < n; p++) {
					System.out.println("Partition " + p);
					System.out.print("memory: "+sa.getMemPartitionSize(p)+" ");
					itMem = sa.getMemPartition(p);
					while (itMem.hasNext())
						System.out.print(itMem.next()+" ");
					System.out.println();

					System.out.print("disk:   "+sa.getDiskPartitionSize(p)+" ");
					itDisk = sa.getDiskPartition(p);
					while (itDisk.hasNext())
						System.out.print(itDisk.next()+" ");
					System.out.println();
				}
			}
*/
			for (int i = 0; i < n; i++)
				computeAllResultsFromPartition(i);
		}
		finally {
			processingWLock.unlock();
		}
	}
	
	private int getNumberOfPartitions() {
		int n = 0;
		for (int i = 0; i < sweepAreas.length; i++)
			if (getSweepArea(i).getNumberOfPartitions() > n)
				 n = getSweepArea(i).getNumberOfPartitions();
		return n;
	}

	private void computeAllResultsFromPartition(int partition) {
//		System.out.println("stage 3 partition "+partition);
		// choose sweep area where partition is largest
		int max = 0, j = 0;
		for (int i = 0; i < sweepAreas.length; i++)
			if (getSweepArea(i).getDiskPartitionSize(partition) > max) {
				max = getSweepArea(i).getDiskPartitionSize(partition);
				j = i;
			}

		// load disk resident portion of partition to all SAs except j
		for (int k = 0; k < sweepAreas.length; k++) {
			if (j == k) continue;
			MJoinSweepArea sa = getSweepArea(k);

			int neededMem = sa.getDiskPartitionSize(partition) * sa.getObjectSize();
//			System.out.println("SA "+k+", mem "+neededMem +", free "+(sa.getAssignedMemSize() - sa.getActualMemUsage()));

			// if not enough memory available, flush partitions beginning with "partition+1"
			// until enough memory is freed up
			int clearPartition = partition + 1;
			while (sa.getAssignedMemSize() - sa.getCurrentMemUsage() < neededMem) {
				if (clearPartition >= getNumberOfPartitions()) {
					System.out.println("memory overflow in stage 3, sweep area " + k);
					System.exit(0);
				}
//				System.out.println("flushing partition "+clearPartition);
				sa.flushToDisk(clearPartition);
				clearPartition++;
			}
			sa.loadPartition(partition);
		}

		// generate results
		Iterator itMem = getSweepArea(j).getMemPartition(partition);
		while (itMem.hasNext())
			computeResults(itMem.next(), j, duplCheck);

		Iterator itDisk = getSweepArea(j).getDiskPartition(partition);
		while (itDisk.hasNext())
			computeResults(itDisk.next(), j, duplCheck);

		// clear partitions to free up some memory
		for (int k = 0; k < sweepAreas.length; k++) {
			getSweepArea(k).clearMemPartition(partition);
		}
	}

	protected boolean isInputAvailable() {
		CompositeMetaData metaData = getMetaData();
		Function f = (Function)metaData.get(INPUT_RATE);
		double inputRate = ((Double)f.invoke()).doubleValue();

		System.out.print("input rate " + inputRate + " ");

		return inputRate > lowInputThreshold;
	}

	/**
	 * Predicate used in stages two and three to eliminate duplicates.
	 *
	 * @param arr result tuple to check
	 *
	 * @return true if result tuple is new, false if it was produced before.
	 */
	private Predicate duplCheck = new Predicate<MJoinObject>() {
		public boolean invoke(List<? extends MJoinObject> list) {
			// determine in which partition the input tuples are stored
			int partition = getSweepArea(0).getPartition(list.get(0));
			int n = list.size();

			for (int i = 0; i < n; i++) {
				int j;
				
				// elem_i is the probing tuple
				// elem_j (below) is the tuple being probed
				MJoinObject elem_i = list.get(i);
				long ats = elem_i.getATS();
				long dts = elem_i.getDTS();

				// check if all elem_j's were in memory when elem_i arrived
				// if so, we found a duplicate
				for (j = 0; j < n; j++) {
					if (i == j) continue;
					MJoinObject elem_j = list.get(j);
					if (!elem_j.timeIntervalContains(ats))
						break;
				}
				if (j == n)
					return false;

				// if elem_i is still in memory, skip the following
				if (dts < java.lang.Long.MAX_VALUE) {
					// loop over all runs of stage 2 using elem_i's disk partition
					Iterator it = getSweepArea(i).getProbeList(partition);
					while (it.hasNext()) {
						Object[] entry = (Object[])it.next();
						long dtsLast = ((Long)entry[0]).longValue();

						// was elem_i used in this run?
						if (dts <= dtsLast) {
							long probeTS = ((Long)entry[1]).longValue();
							// check if all elem_j's were in memory at probeTS
							// if so, we found a duplicate
							for (j = 0; j < n; j++) {
								if (i == j) continue;
								MJoinObject elem_j = list.get(j);
								if (!elem_j.timeIntervalContains(probeTS))
									break;
							}
							if (j == n)
								return false;
						}
					}
				}
			}
			return true;
		}
	};

	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.joins.MultiWayEquiJoin#close()
	 */
	@Override
	public void close() {
		stageTwoScheduler.terminate();
		while(stageTwoRunning); // TODO: busy-waiting?
		runStageThree();
	/*		
		for (int i = 0; i < getSweepArea(0).getNumberOfPartitions(); i++) {
			System.out.print("probe list partition "+i+": ");
			Iterator it = getSweepArea(0).getProbeList(i);
			while (it.hasNext()) {
				Object[] arr = (Object[])it.next();
				System.out.print("["+arr[0]+","+arr[1]+"] ");
			}
			System.out.println();
		}
	*/	
		super.close();
	}

	public static void main(String[] args) {
		SourceProcessor proc1 = new SourceProcessor(Processor.getRandomDelayManager(100, 42), TimerUtils.FACTORY_METHOD.invoke(), false);

		Enumerator e1 = new Enumerator(proc1, 0, 1000);
		Enumerator e2 = new Enumerator(0, 1000, 10);
		Enumerator e3 = new Enumerator(0, 1000, 10);
/*		Enumerator e1 = new Enumerator(1, 40, 40);
		Enumerator e2 = new Enumerator(1, 40, 80);
		Enumerator e3 = new Enumerator(1, 40, 40);*/

		final Function hashFunction = new Function() {
			@Override
			public Object invoke(Object o) {
				return new Integer(((Integer)o).intValue()%13);
			}
		};

		MultiWayEquiJoin<Integer,List<Object>> join = new MJoin<Integer,List<Object>>(
			new Source[]{ e1, e2, e3},
			new int[]{ 0, 1, 2 },
			new Converter[]{ IntegerConverter.DEFAULT_INSTANCE, IntegerConverter.DEFAULT_INSTANCE, IntegerConverter.DEFAULT_INSTANCE },
			new Function<Object, Object>() {
				@Override
				public Object invoke(List<? extends Object> result) {
					return result;
				}
			},
			new Predicate[]{ new Equal<Integer>(), new Equal<Integer>(), new Equal<Integer>() },
			new Function[]{ hashFunction, hashFunction, hashFunction },
			4,		// objectSize
			1024,	// memSize
			50,		// blockSize
			500,	// stageTwoInterval
			0.05	// lowInputThreshold
		);
		
		 //  assigning enough memory
		GlobalMemoryManager memManager = new GlobalMemoryManager( 
			new UniformStrategy(2*1024, join)
		);
		
		memManager.requestForMemory(join, 2*512);

		Printer printer = new Printer<List<Object>>(join);
		Tester tester = new Tester<List<Object>>(join);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(tester);
		exec.registerQuery(printer);
		exec.startAllQueries();
		
	}
}
