/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.joins;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.collections.sweepAreas.PNSweepArea;
import xxl.core.collections.sweepAreas.SweepAreaImplementor;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.DecoratorPNObject;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.PNTemporalPipe;
import xxl.core.pipes.operators.mappers.TemporalPNPipe;
import xxl.core.pipes.operators.unions.PNSync;
import xxl.core.pipes.operators.unions.TemporalSync;
import xxl.core.pipes.operators.windows.PNTemporalWindow;
import xxl.core.pipes.operators.windows.TemporalWindow;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.predicates.Predicate;
import xxl.core.predicates.Predicates;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

/**
 *
 * @param <I>
 * @param <O>
 */
public class PNJoin<I,O> extends AbstractPNPipe<I,O> implements MemoryMonitorable {

	public static <I, O> PNJoin<I,O> SNJ(PNSync<I> syncInputs, final Predicate<? super PNObject<I>> theta, Function<? super I, ? extends O> resultFactory) {
		return new PNJoin<I,O>(
			syncInputs,
			new PNNestedLoopsJoinSA<I>(0, 2, theta),
			new PNNestedLoopsJoinSA<I>(1, 2, Predicates.swapArguments(theta)),
			resultFactory
		);
	}
	
	public static <I, O> PNJoin<I,O> SHJ(PNSync<I> syncInputs, final Function<? super PNObject<I>, Integer> hashFunction, Function<? super I, ? extends O> resultFactory) {
		Function<Object, List<PNObject<I>>> newList = new Function<Object, List<PNObject<I>>>() {
			@Override
			public List<PNObject<I>> invoke() {
				return new LinkedList<PNObject<I>>();
			}
		};
		return new PNJoin<I,O>(
			syncInputs,
			new PNJoinSA<I>(
				new HashSAImplementor<PNObject<I>>(hashFunction, newList, 2, PNObject.VALUE_EQUIVALENCE_PREDICATE),
				0,
				PNObject.VALUE_EQUIVALENCE_PREDICATE,
				2
			),
			new PNJoinSA<I>(
				new HashSAImplementor<PNObject<I>>(hashFunction, newList, 2, PNObject.VALUE_EQUIVALENCE_PREDICATE),
				1,
				PNObject.VALUE_EQUIVALENCE_PREDICATE,
				2
			),
			resultFactory
		);
	}
	
	
	public static class PNJoinSA<I> extends PNSweepArea<I> {
		
		public PNJoinSA(SweepAreaImplementor<PNObject<I>> impl, int ID, Predicate<? super PNObject<I>> joinPredicate, int dim) {
			super(impl, ID, joinPredicate, dim);
		}
		
		public PNJoinSA(int ID, int dim, Predicate<? super PNObject<I>> joinPredicate) {
			this(new HashSAImplementor<PNObject<I>>(
					new Function<PNObject<I>, Integer>() {
						@Override
						public Integer invoke(PNObject<I> pno) {
							return pno.getObject().hashCode();
						}
					},
					new Function<Object, List<PNObject<I>>>() {
						@Override
						public List<PNObject<I>> invoke() {
							return new LinkedList<PNObject<I>>();
						}
					},
					2, PNObject.VALUE_EQUIVALENCE_PREDICATE
				), ID, joinPredicate, dim
			);
		}
	}
	
	public static class PNNestedLoopsJoinSA<I> extends PNJoinSA<I> {
		
		public PNNestedLoopsJoinSA(int ID, int dim, Predicate<? super PNObject<I>> joinPredicate, final Function<? super PNObject<I>, Integer> hashFunction) {
			super(
				new HashSAImplementor<PNObject<I>>(
				hashFunction,
				new Function<Object, List<PNObject<I>>>() {
					@Override
					public List<PNObject<I>> invoke() {
						return new LinkedList<PNObject<I>>();
					}
				},
				2, PNObject.VALUE_EQUIVALENCE_PREDICATE
				) {
					
					public Iterator<PNObject<I>> query(PNObject<I> o, int ID) throws IllegalArgumentException {
						return filter(iterator(), o, ID);	
					}
						
				}, ID, joinPredicate, dim
			);
		}	
		
		public PNNestedLoopsJoinSA(int ID, int dim, Predicate<? super PNObject<I>> joinPredicate) {
			this(ID, dim, joinPredicate, 
				new Function<PNObject<I>, Integer>() {
					@Override
					public Integer invoke(PNObject<I> pno) {
						return pno.getObject().hashCode();
					}
				}
			);
		}		
	}
	
	protected PNSweepArea<I>[] sweepAreas;
	protected Function<? super I,? extends O> newResult;
		
	public PNJoin(final PNSweepArea<I> sweepArea0, final PNSweepArea<I> sweepArea1, final Function<? super I, ? extends O> newResult) {
		this.sweepAreas = new PNSweepArea[2];
		this.sweepAreas[0] = sweepArea0;
		this.sweepAreas[1] = sweepArea1;
		this.newResult = newResult; 
	}
	    
    public PNJoin(PNSync<I> syncInputs, final PNSweepArea<I> sweepArea0, final PNSweepArea<I> sweepArea1, final Function<? super I, ? extends O> newResult) {
		this(sweepArea0, sweepArea1, newResult);
		if (!(Pipes.connect(syncInputs, this, DEFAULT_ID)))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");
	}
	    
	@Override
	public void processObject(PNObject<I> pno, int sourceID) throws IllegalArgumentException {
		DecoratorPNObject<I,Integer> dpno = (DecoratorPNObject<I,Integer>)pno;
		sourceID = dpno.getDecoration();
		
		int j = sourceID == sourceIDs[0] ? 0 : 1;
		int k = 1 - j;	
		if (pno.isNegative())
			sweepAreas[j].reorganize(pno, j);
		else
			sweepAreas[j].insert(pno);
		
		Iterator<? extends PNObject<I>> results = sweepAreas[k].query(pno, j);
		while(results.hasNext()) {
			PNObject<O> next = j == 0 ? 
				new PNObject<O>(newResult.invoke(pno.getObject(), results.next().getObject()), pno.getTimeStamp(), pno.isPositive()) : 
				new PNObject<O>(newResult.invoke(results.next().getObject(), pno.getObject()), pno.getTimeStamp(), pno.isPositive());
			transfer(next);
		}
	}

	/**
	 * Calls <CODE>super.close()</CODE> and closes
	 * both sweep areas.
	 */
	@Override
	public void close() {
		super.close();
		if (isClosed()) {
			sweepAreas[0].close();
			sweepAreas[1].close();
		}
	}
	
	public int getCurrentMemUsage() {
		int sa0Mem = sweepAreas[0].getCurrentMemUsage();
		int sa1Mem = sweepAreas[1].getCurrentMemUsage();
		if (sa0Mem == SIZE_UNKNOWN || sa1Mem == SIZE_UNKNOWN)
			return SIZE_UNKNOWN;
		return sa0Mem + sa1Mem;		
	}
	
	public PNSync<I> getSyncInputs() {
		return (PNSync<I>)this.getSource(0);
	}
	
		
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 2500;
		final int maxValue = 20;
		final int maxTSInc = 50;
		final int noOfBuckets = 10;
		
		int seed1 = 42;
		int seed2 = 17;
		
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed1, maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed1, maxTSInc), noOfElements)
		);
		TimeStampCursorSource<Integer> s2 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed2, maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed2, maxTSInc), noOfElements)
		);
		final long windowSize1 = 750;
		final long windowSize2 = 500;
				
		// time interval approach
		TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s1, windowSize1);
		TemporalWindow<Integer> w2 = new TemporalWindow<Integer>(s2, windowSize2);
		TemporalJoin<Integer,Integer> res1 = Joins.TemporalIntegerSHJProject(w1, w2, noOfBuckets);
		
		// positive-negative approach
		TemporalPNPipe<Integer> pnSource1 = new TemporalPNPipe<Integer>(s1);
		TemporalPNPipe<Integer> pnSource2 = new TemporalPNPipe<Integer>(s2);
		PNTemporalWindow<Integer> pnw1 = new PNTemporalWindow<Integer>(pnSource1, windowSize1);
		PNTemporalWindow<Integer> pnw2 = new PNTemporalWindow<Integer>(pnSource2, windowSize2);
		PNJoin<Integer,Integer> pnJoin = SHJ(new PNSync<Integer>(pnw1, pnw2), Joins.hashFunctionFactory(noOfBuckets), Joins.projectResultFunctionFactory(pnw1, pnw2, 0));
		PNTemporalPipe<Integer> res2 = new PNTemporalPipe<Integer>(pnJoin);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res1));
//		exec.registerQuery(new Printer(res1, System.out, true));
//		exec.registerQuery(new Printer(pnJoin, System.out, true));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(pnJoin));
		exec.registerQuery(Pipes.compareSnapshots(res1, res2, false, false));
		exec.startAllQueries();
	}

}
