/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.joins;

import java.util.Iterator;

import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.DecoratorTemporalObject;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.metaData.DefaultMetaDataHandler;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.AbstractTemporalPipe.AbstractTemporalPipeMetaDataManagement;
import xxl.core.pipes.operators.joins.TemporalJoin.TemporalJoinHeapSA;
import xxl.core.pipes.operators.joins.TemporalJoin.TemporalJoinSA;
import xxl.core.pipes.operators.unions.TemporalSync;
import xxl.core.pipes.operators.windows.TemporalWindow;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

/**
 *
 * @param <I>
 * @param <O>
 */
public class SyncTemporalJoin<I,O> extends AbstractTemporalPipe<I,O> implements MemoryMonitorable {
	
	protected TemporalJoinSA<I>[] sweepAreas;
	protected Function<? super TemporalObject<I>,? extends TemporalObject<O>> newResult;
	 
		
	public SyncTemporalJoin(final TemporalJoinSA<I> sweepArea0, final TemporalJoinSA<I> sweepArea1, final Function<? super I, ? extends O> newResult) {
		this.sweepAreas = new TemporalJoinSA[2];
		this.sweepAreas[0] = sweepArea0;
		this.sweepAreas[1] = sweepArea1;
		this.newResult = TemporalJoin.wrapResultFunction(newResult); 
	}
	    
    public SyncTemporalJoin(TemporalSync<I> syncInputs, final TemporalJoinSA<I> sweepArea0, final TemporalJoinSA<I> sweepArea1, final Function<? super I, ? extends O> newResult) {
		this(sweepArea0, sweepArea1, newResult);
		if (!(Pipes.connect(syncInputs, this, DEFAULT_ID)))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");
	}
	    
	@Override
	public void processObject(TemporalObject<I> to, int sourceID) throws IllegalArgumentException {
		DecoratorTemporalObject<I,Integer> dto = (DecoratorTemporalObject<I,Integer>)to;
		sourceID = dto.getDecoration();
		
		int j = sourceID == sourceIDs[0] ? 0 : 1;
		int k = 1 - j;	
		
		sweepAreas[k].reorganize(to, j);
		Iterator<? extends TemporalObject<I>> results = sweepAreas[k].query(to, j);
		while(results.hasNext()) {
			transfer( j == 0 ? 
				newResult.invoke(to, results.next()) 
				: newResult.invoke(results.next(), to)
			);
		}
		sweepAreas[j].insert(to);
	}

	/**
	 * Calls <CODE>super.close()</CODE> and closes
	 * both sweep areas.
	 */
	@Override
	public void close() {
		super.close();
		if (isClosed()) {
			sweepAreas[0].close();
			sweepAreas[1].close();
		}
	}
	
	public int getCurrentMemUsage() {
		int sa0Mem = sweepAreas[0].getCurrentMemUsage();
		int sa1Mem = sweepAreas[1].getCurrentMemUsage();
		if (sa0Mem == SIZE_UNKNOWN || sa1Mem == SIZE_UNKNOWN)
			return SIZE_UNKNOWN;
		return sa0Mem + sa1Mem;		
	}
	
	public TemporalSync<I> getSyncInputs() {
		return (TemporalSync<I>)this.getSource(0);
	}
	
		
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 2500;
		final int maxValue = 20;
		final int maxTSInc = 50;
		final int noOfBuckets = 10;
		
		int seed1 = 42;
		int seed2 = 17;
		
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed1, maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed1, maxTSInc), noOfElements)
		);
		TimeStampCursorSource<Integer> s2 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed2, maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed2, maxTSInc), noOfElements)
		);
		final long windowSize1 = 750;
		final long windowSize2 = 500;
				
		// time interval approach
		TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s1, windowSize1);
		TemporalWindow<Integer> w2 = new TemporalWindow<Integer>(s2, windowSize2);
		TemporalJoin<Integer,Integer> res1 = Joins.TemporalIntegerSHJProject(w1, w2, noOfBuckets);
		
		// sync approach
		SyncTemporalJoin<Integer,Integer> res2 = new SyncTemporalJoin<Integer,Integer>(new TemporalSync(w1, w2),
			new TemporalJoinHeapSA<Integer>(
				new HashSAImplementor<TemporalObject<Integer>>(Joins.hashFunctionFactory(1117), 2),
				0,
				TemporalObject.VALUE_EQUIVALENCE_PREDICATE,
				2
			),
			new TemporalJoinHeapSA<Integer>(
				new HashSAImplementor<TemporalObject<Integer>>(Joins.hashFunctionFactory(1117), 2),
				1,
				TemporalObject.VALUE_EQUIVALENCE_PREDICATE,
				2
			),
			new Function<Integer,Integer>() {
            	public Integer invoke(Integer o1, Integer o2) {
            		return o1;
            	}
        	}
		);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res1));
//		exec.registerQuery(new Printer(res1, System.out, true));
//		exec.registerQuery(new Printer(res2, System.out, true));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res2));
		exec.registerQuery(Pipes.compareSnapshots(res1, res2, false, false));
		exec.startAllQueries();
	}

}

