/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.mappers;

import static xxl.core.collections.sweepAreas.SweepArea.SWEEPAREA_COSTS;
import static xxl.core.collections.sweepAreas.SweepArea.SWEEPAREA_PERIODIC_TRIGGER;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.D_ESTIMATION;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.G_ESTIMATION;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.L_ESTIMATION;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.MEM_USAGE_ESTIMATION;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeSet;

import xxl.core.collections.CountingComparator;
import xxl.core.collections.DecoratorList;
import xxl.core.collections.DecoratorSortedSet;
import xxl.core.collections.sweepAreas.AbstractSweepArea;
import xxl.core.cursors.AbstractCursor;
import xxl.core.cursors.Cursors;
import xxl.core.cursors.sources.EmptyCursor;
import xxl.core.functions.Function;
import xxl.core.math.functions.AggregationFunction;
import xxl.core.math.functions.CountingAggregationFunction;
import xxl.core.math.statistics.parametric.aggregates.Sum;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.elements.TimeStampedObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.metaData.MetaDataDependencies;
import xxl.core.pipes.metaData.TriggeredEvaluationMetaDataHandler;
import xxl.core.pipes.operators.AbstractPipe;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.FastTemporalAggregator.TemporalAggregatorSA.AggEntry;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.AbstractSink;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sinks.Tester;
import xxl.core.pipes.sources.AbstractSource;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;
import xxl.core.predicates.CountingPredicate;
import xxl.core.predicates.Less;
import xxl.core.predicates.Predicate;
import xxl.core.util.metaData.ExternalTriggeredPeriodicMetaData;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * The temporal aggregation operator.
 * 
 * @since 1.1
 */
public class FastTemporalAggregator<I,O> extends AbstractTemporalPipe<I, O> implements MemoryMonitorable {

	
	public class TemporalAggregatorMetaDataManagement extends AbstractTemporalPipeMetaDataManagement {
		
		public TemporalAggregatorMetaDataManagement() {
			super();
		}

		@Override
		protected boolean addMetaData(Object metaDataIdentifier) {
			if (metaDataIdentifier.equals(COST_MEASUREMENT)) {
				metaData.add(metaDataIdentifier, new TriggeredEvaluationMetaDataHandler<Double>(this, metaDataIdentifier,
					new Function<Object,Double>() {
						@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								if (isClosed || isDone) 
									return Double.NaN;
								return (Double)((MetaDataHandler)metaData.get(SWEEPAREA_COSTS)).getMetaData();
							}
						}
					})
				);
				return true;
			}				
			if (metaDataIdentifier.equals(SWEEPAREA_COSTS)) {
				if (sweepArea.getMetaDataManagement().include(COST_MEASUREMENT)) {	
					
					Cursors.println(sweepArea.getMetaData().identifiers());
					
					metaData.add(metaDataIdentifier, new TriggeredEvaluationMetaDataHandler<Double>(this, metaDataIdentifier, Double.class,
						new Function<Object,Double>() {
							@Override
							public Double invoke() {
								synchronized(metaDataManagement) {
									if (isClosed || isDone) 
										return Double.NaN;
									return ((Function<?,Double>)sweepArea.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT)).invoke();
								}
							}
						})
					);
					return true;
				}
				return false;
			}		
			if (metaDataIdentifier.equals(SWEEPAREA_PERIODIC_TRIGGER)) {
				metaData.add(metaDataIdentifier, new TriggeredEvaluationMetaDataHandler<Object>(this, metaDataIdentifier,
					new Function<Object,Object>() {			
						@Override
						public Object invoke() {
							synchronized(metaDataManagement) {
								if (isClosed || isDone) 
									return null;
								try {
									((ExternalTriggeredPeriodicMetaData)sweepArea.getMetaDataManagement()).updatePeriodicMetaData(updatePeriod);
								}
								catch (ClassCastException e) {}
								return null;
							}
						}
					})
				);
				return true;
			}						
			return super.addMetaData(metaDataIdentifier);
		}
		
		@Override
		protected boolean removeMetaData(Object metaDataIdentifier) {
			if (metaDataIdentifier.equals(COST_MEASUREMENT)) {
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (metaDataIdentifier.equals(SWEEPAREA_PERIODIC_TRIGGER)) {
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (metaDataIdentifier.equals(SWEEPAREA_COSTS)) {
				sweepArea.getMetaDataManagement().exclude(COST_MEASUREMENT);
				metaData.remove(metaDataIdentifier);
				return true;
			}
			return super.removeMetaData(metaDataIdentifier);
		}
		
	}
	
    public static class TemporalAggregatorSA<I,O> extends AbstractSweepArea<TemporalObject> {
    	
    	public class TemporalAggregatorSAMetaDataManagement extends AbstractSAMetaDataManagement implements ExternalTriggeredPeriodicMetaData {
    		
    		public class CountingTree<T> extends DecoratorSortedSet<T> {

				protected int getCount = 0;
    			protected int addCount = 0;
    			protected int setCount = 0;    			
    			protected int removeCount = 0;

    			public CountingTree(SortedSet<T> sortedSet) {
					super(sortedSet);
				}

    			public T first() {
    				getCount++;
    				return sortedSet.first();
    			}

    			public T last() {
    				getCount++;
    				return sortedSet.last();
    			}

    			public boolean add(T o) {
    				addCount++;
    				return sortedSet.add(o);
    			}

    			public boolean addAll(Collection<? extends T> c) {
    				addCount += c.size();
    				return sortedSet.addAll(c);
    			}

    			public boolean remove(Object o) {
    				removeCount++;
    				return sortedSet.remove(o);
    			}

    			public boolean removeAll(Collection<?> c) {
    				removeCount += c.size();
    				return sortedSet.removeAll(c);
    			}
    			
				public int getAddCount() {
					return addCount;
				}

				public int getGetCount() {
					return getCount;
				}

				public int getRemoveCount() {
					return removeCount;
				}

				public int getSetCount() {
					return setCount;
				}
    			    			
				public void resetCounters() {
	    			getCount = 0;
	    			addCount = 0;
	    			setCount = 0;    			
	    			removeCount = 0;
				}
    		}
    		
    		public static final String TREE_DECORATION = "TREE_DECORATION";        	
        	protected boolean countTreeOperations = false;
    		protected long treeOperationCount;

    		public static final String SINGLE_TREE_OPERATION_COSTS = "SINGLE_TREE_OPERATION_COSTS";
    		protected double singleTreeOperationCosts;
    		public static final String TREE_COSTS = "TREE_COSTS";
    		protected volatile boolean measureTreeCosts = false;
    		protected double treeCosts = 0.0;

    		public static final String PREDICATE_DECORATION = "PREDICATE_DECORATION";        	
        	protected boolean countPredicateCalls = false;
    		protected long predicateCallCount;

    		public static final String SINGLE_PREDICATE_COSTS = "SINGLE_PREDICATE_COSTS";
    		protected double singlePredicateCosts;
    		public static final String PREDICATE_COSTS = "PREDICATE_COSTS";
    		protected volatile boolean measurePredicateCosts = false;
    		protected double predicateCosts = 0.0;

    		public static final String AGGREGATION_FUNCTION_DECORATION = "AGGREGATION_FUNCTION_DECORATION";        	
        	protected boolean countAggregationFunctionCalls = false;
    		protected long aggregationFunctionCallCount;
    		protected long aggregationFunctionInitCount;

    		public static final String SINGLE_AGGREGATION_FUNCTION_COSTS = "SINGLE_AGGREGATION_FUNCTION_COSTS";
    		protected double singleAggregationFunctionCosts;
    		public static final String AGGREGATION_FUNCTION_COSTS = "AGGREGATION_FUNCTION_COSTS";
    		protected volatile boolean measureAggregationFunctionCosts = false;
    		protected double aggregationFunctionCosts = 0.0;

       		public static final String COMPARATOR_DECORATION = "COMPARATOR_DECORATION";        	
        	protected boolean countComparatorCalls = false;
    		protected long comparatorCallCount;

    		public static final String SINGLE_COMPARATOR_COSTS = "SINGLE_COMPARATOR_COSTS";
    		protected double singleComparatorCosts;
    		public static final String COMPARATOR_COSTS = "COMPARATOR_COSTS";
    		protected volatile boolean measureComparatorCosts = false;
    		protected double comparatorCosts = 0.0;

    		protected volatile boolean measureCosts = false;
    		protected double costs = 0.0;
    		
    		public TemporalAggregatorSAMetaDataManagement() {
    			super();
    		}

			public boolean needsPeriodicUpdate(Object metaDataIdentifier) {
    			if (metaDataIdentifier.equals(TREE_DECORATION) ||
        			metaDataIdentifier.equals(TREE_COSTS) ||    				
        			metaDataIdentifier.equals(PREDICATE_DECORATION) ||
        			metaDataIdentifier.equals(PREDICATE_COSTS) ||    				
        			metaDataIdentifier.equals(AGGREGATION_FUNCTION_DECORATION) ||
        			metaDataIdentifier.equals(AGGREGATION_FUNCTION_COSTS) ||    				
        			metaDataIdentifier.equals(COMPARATOR_DECORATION) ||
        			metaDataIdentifier.equals(COMPARATOR_COSTS) ||    				
    				metaDataIdentifier.equals(COST_MEASUREMENT)) {
    					return true;
        		}
				return false;
			}

			public void updatePeriodicMetaData(long period) {
    			if (countTreeOperations) {
    				CountingTree<AggEntry<O>> t = (CountingTree<AggEntry<O>>) aggs;
    				treeOperationCount = 0;
    				treeOperationCount += t.getAddCount();
    				// listOperationCount += l.getGetCount();
    				treeOperationCount += t.getRemoveCount();
    				treeOperationCount += t.getSetCount();
    				t.resetCounters();
    			}
    			if (countPredicateCalls) {
    				CountingPredicate<? super TemporalObject<O>> p = (CountingPredicate<? super TemporalObject<O>>)removePredicate;
    				predicateCallCount = p.getNoOfCalls();
    				p.resetCounters();
    			}
    			if (countAggregationFunctionCalls) {
    				CountingAggregationFunction<? super I,O> f = (CountingAggregationFunction<? super I,O>)aggFunction;
    				aggregationFunctionCallCount = f.getCalls();
    				aggregationFunctionInitCount = f.getInitCalls();
    				f.resetCounters();
    			}
    			if (countComparatorCalls) {
    				CountingComparator<TemporalObject> c = (CountingComparator<TemporalObject>) comparator;
    				comparatorCallCount = c.getNoOfCalls();
    				c.resetCounter();
    			}
    			if (measureTreeCosts) {
    				treeCosts = (treeOperationCount * singleTreeOperationCosts) / period;
    			}
    			if (measurePredicateCosts) {
    				predicateCosts = (predicateCallCount * singlePredicateCosts) / period;
    			}
    			if (measureAggregationFunctionCosts) {
    				aggregationFunctionCosts = 
    					((aggregationFunctionCallCount * singleAggregationFunctionCosts) / period) +
    					((aggregationFunctionInitCount * singleAggregationFunctionCosts) / period);
    			}
    			if (measureComparatorCosts) {
    				comparatorCosts = (comparatorCallCount * singleComparatorCosts) / period;
    			}
    			if (measureCosts) {    				
    				costs = 0;
    				costs += treeCosts;
    				costs += predicateCosts;
    				costs += aggregationFunctionCosts;
    				costs += comparatorCosts;
    			}				
			}
			
			@Override
			protected boolean addMetaData(Object metaDataIdentifier) {
				if (metaDataIdentifier.equals(TREE_DECORATION)) {
					aggs = new CountingTree<AggEntry<O>>(aggs);
					countTreeOperations = true;
    				treeOperationCount = 0;
    				metaData.add(metaDataIdentifier, new Function<Object,Long>() {					
    					@Override
    					public Long invoke() {
    						return treeOperationCount;
    					}
    				});
    				return true;
    			}
				if (metaDataIdentifier.equals(PREDICATE_DECORATION)) {
					removePredicate = new CountingPredicate<Long>(removePredicate);
					countPredicateCalls = true;
    				predicateCallCount = 0;
    				metaData.add(metaDataIdentifier, new Function<Object,Long>() {					
    					@Override
    					public Long invoke() {
    						return predicateCallCount;
    					}
    				});
    				return true;
    			}
				if (metaDataIdentifier.equals(AGGREGATION_FUNCTION_DECORATION)) {
					aggFunction = new CountingAggregationFunction<I, O>(aggFunction);
					countAggregationFunctionCalls = true;
    				aggregationFunctionCallCount = 0;
    				aggregationFunctionInitCount = 0;
    				metaData.add(metaDataIdentifier, new Function<Object,Long[]>() {					
    					@Override
    					public Long[] invoke() {
    						return new Long[] { aggregationFunctionInitCount ,aggregationFunctionCallCount };
    					}
    				});
    				return true;
    			}
				if (metaDataIdentifier.equals(COMPARATOR_DECORATION)) {
					comparator = new CountingComparator<TemporalObject>(comparator);
					countComparatorCalls = true;
    				comparatorCallCount = 0;
    				metaData.add(metaDataIdentifier, new Function<Object,Long>() {					
    					@Override
    					public Long invoke() {
    						return comparatorCallCount;
    					}
    				});
    				return true;
    			}
				if (metaDataIdentifier.equals(SINGLE_TREE_OPERATION_COSTS)) {
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return singleTreeOperationCosts;
    					}
    				});
    				return true;
    			}
				if (metaDataIdentifier.equals(SINGLE_PREDICATE_COSTS)) {
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return singlePredicateCosts;
    					}
    				});
    				return true;
    			}
				if (metaDataIdentifier.equals(SINGLE_AGGREGATION_FUNCTION_COSTS)) {
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return singleAggregationFunctionCosts;
    					}
    				});
    				return true;
    			}
				if (metaDataIdentifier.equals(SINGLE_COMPARATOR_COSTS)) {
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return singleComparatorCosts;
    					}
    				});
    				return true;
    			}
    			if (metaDataIdentifier.equals(TREE_COSTS)) {
    				include(TREE_DECORATION);
    				treeCosts = 0.0;
    				measureTreeCosts = true;
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return treeCosts;
    					}
    				});
    				return true;
    			}
    			if (metaDataIdentifier.equals(PREDICATE_COSTS)) {
    				include(PREDICATE_DECORATION);
    				predicateCosts = 0.0;
    				measurePredicateCosts = true;
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return predicateCosts;
    					}
    				});
    				return true;
    			}
    			if (metaDataIdentifier.equals(AGGREGATION_FUNCTION_COSTS)) {
    				include(AGGREGATION_FUNCTION_DECORATION);
    				aggregationFunctionCosts = 0.0;
    				measureAggregationFunctionCosts = true;
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return aggregationFunctionCosts;
    					}
    				});
    				return true;
    			}
    			if (metaDataIdentifier.equals(COMPARATOR_COSTS)) {
    				include(COMPARATOR_DECORATION);
    				comparatorCosts = 0.0;
    				measureComparatorCosts = true;
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return comparatorCosts;
    					}
    				});
    				return true;
    			}
    			if (metaDataIdentifier.equals(COST_MEASUREMENT)) {
    				include(TREE_COSTS);
    				include(PREDICATE_COSTS);
    				include(AGGREGATION_FUNCTION_COSTS);
    				include(COMPARATOR_COSTS);
    				costs = 0.0;
    				measureCosts = true;
    				metaData.add(metaDataIdentifier, new Function<Object,Double>() {					
    					@Override
    					public Double invoke() {
    						return costs;
    					}
    				});
    				return true;
    			}
				return super.addMetaData(metaDataIdentifier);
			}
			
			@Override
    		protected boolean removeMetaData(Object metaDataIdentifier) {
    			if (metaDataIdentifier.equals(TREE_DECORATION)) {
    				aggs = ((CountingTree<AggEntry<O>>)aggs).getDecoratedSortedSet();
    				treeOperationCount = 0;
    				countTreeOperations = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(PREDICATE_DECORATION)) {
    				removePredicate = ((CountingPredicate<Long>)removePredicate).getPredicate();
    				predicateCallCount = 0;
    				countPredicateCalls = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(AGGREGATION_FUNCTION_DECORATION)) {
    				aggFunction = ((CountingAggregationFunction<? super I,O>)aggFunction).getAggregationFunction();
    				aggregationFunctionCallCount = 0;
    				aggregationFunctionInitCount = 0;
    				countAggregationFunctionCalls = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(COMPARATOR_DECORATION)) {
    				comparator = ((CountingComparator<TemporalObject>)comparator).getComparator();
    				comparatorCallCount = 0;
    				countComparatorCalls = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(TREE_COSTS)) {
    				exclude(TREE_DECORATION);
    				treeCosts = 0.0;
    				measureTreeCosts = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(PREDICATE_COSTS)) {
    				exclude(PREDICATE_DECORATION);
    				predicateCosts = 0.0;
    				measurePredicateCosts = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(AGGREGATION_FUNCTION_COSTS)) {
    				exclude(AGGREGATION_FUNCTION_DECORATION);
    				aggregationFunctionCosts = 0.0;
    				measureAggregationFunctionCosts = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(COMPARATOR_COSTS)) {
    				exclude(COMPARATOR_DECORATION);
    				comparatorCosts = 0.0;
    				measureComparatorCosts = false;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(SINGLE_TREE_OPERATION_COSTS) ||
    				metaDataIdentifier.equals(SINGLE_PREDICATE_COSTS) ||
    				metaDataIdentifier.equals(SINGLE_AGGREGATION_FUNCTION_COSTS) ||
    				metaDataIdentifier.equals(SINGLE_COMPARATOR_COSTS)) {
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}
    			if (metaDataIdentifier.equals(COST_MEASUREMENT)) {
    				exclude(TREE_COSTS);
    				exclude(PREDICATE_COSTS);
    				exclude(AGGREGATION_FUNCTION_COSTS);
    				exclude(COMPARATOR_COSTS);
    				measureCosts = false;
    				costs = 0.0;
    				metaData.remove(metaDataIdentifier);
    				return true;
    			}

				return super.removeMetaData(metaDataIdentifier);
			}
			
    		public void setSingleTreeOperationCosts(double costs) {
    			singleTreeOperationCosts = costs;
    		}

    		public void setSinglePredicateCosts(double costs) {
    			singlePredicateCosts = costs;
    		}

    		public void setSingleAggregationFunctionCosts(double costs) {
    			singleAggregationFunctionCosts = costs;
    		}

    		public void setSingleComparatorCosts(double costs) {
    			singleComparatorCosts = costs;
    		}
    		
    		public void setCostFactors(double singleListOperationCosts, double singlePredicateCosts, double singleAggregationFunctionCosts, double singleComparatorCosts) {
    			setSingleTreeOperationCosts(singleListOperationCosts);
    			setSinglePredicateCosts(singlePredicateCosts);
    			setSingleAggregationFunctionCosts(singleAggregationFunctionCosts);
    			setSingleComparatorCosts(singleComparatorCosts);
    		}
    		
    	}	
    	
    	protected AggregationFunction<? super I,O> aggFunction;
    	protected SortedSet<AggEntry<O>> aggs;
    	protected Comparator<TemporalObject> comparator;
    	protected Predicate<? super Long> removePredicate;
    	  
    	class AggEntry<X> implements Comparable<AggEntry<X>> {
    		long timeStamp;
    		long nextTimeStamp;
    		X agg;    
    		AggEntry(long timeStamp) {
    			this.timeStamp = timeStamp;
    		}
    		AggEntry(X agg, long timeStamp, long nextTimeStamp) {
    			this.agg = agg;
    			this.timeStamp = timeStamp;    			
    			this.nextTimeStamp = nextTimeStamp;
    		}
			public int compareTo(AggEntry<X> o) {
				if (timeStamp < o.timeStamp)
					return -1;
				if (timeStamp > o.timeStamp)
					return 1;
				return 0;
			}
			@Override
    		public boolean equals(Object o) {
    			return timeStamp == ((AggEntry<X>)o).timeStamp;
    		}
			@Override
			public String toString() {
				return timeStamp + " " + agg;
			}
    	}
    	
    	public TemporalAggregatorSA(AggregationFunction<? super I,O> aggFunction, int objectSize) {
    		super(objectSize);
    		this.aggs =  new TreeSet<AggEntry<O>>();
    		this.aggFunction = aggFunction;
    		this.comparator = TemporalObject.START_TIMESTAMP_COMPARATOR;
    		this.removePredicate = new Predicate<Long>() {
    			public boolean invoke(Long a, Long b) {
    				return a<b;
    			};    			
    		};
    	}
    	
    	public TemporalAggregatorSA(AggregationFunction<? super I,O> aggFunction) {
    		this(aggFunction, MemoryMonitorable.SIZE_UNKNOWN);
    	}
    	    	    	
    	
    	@Override
		public void insert(TemporalObject object) throws IllegalArgumentException {
    		TemporalObject<I> o = (TemporalObject<I>)object;
    		long oStart = object.getStart();
    		long oEnd = object.getEnd();
    		if (aggs.size()==0) {
    			aggs.add(new AggEntry(aggFunction.invoke(null, o.getObject()), oStart, oEnd));
    			aggs.add(new AggEntry(null, oEnd, Long.MAX_VALUE));
    			return;
    		}
    		for (long ts : new long[] { oStart, oEnd }) {
        		AggEntry<O> prevAgg = aggs.headSet(new AggEntry(ts+1)).last();
        		if (prevAgg.timeStamp != ts) {
        			aggs.add(new AggEntry(prevAgg.agg, ts, prevAgg.nextTimeStamp));
        			prevAgg.nextTimeStamp = ts;        			
        		}
    		}
    		SortedSet<AggEntry<O>> update = aggs.subSet(new AggEntry(oStart), new AggEntry(oEnd));
    		for (AggEntry<O> entry : update) 
    			entry.agg = aggFunction.invoke(entry.agg, o.getObject());
    		
    		// determine aggregate size
    		if (checkObjectSize)
    			computeObjectSize(aggs.first()); 		
    	}
    	
    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#expire(java.lang.Object, int)
    	 */
    	@Override
		public Iterator<TemporalObject> expire(final TemporalObject currentStatus, final int ID) {
    		return new AbstractCursor<TemporalObject>() {
    			
    			AggEntry<O> fst;
    			
    			@Override
				public boolean hasNextObject() {
    				if(aggs.size()<2) {
    					aggs.clear();
    					return false;
    				}
    				fst = aggs.first(); 
    				if (fst.agg==null) {
    					aggs.remove(fst);
    					return hasNextObject();
    				}    				
    				return !removePredicate.invoke(currentStatus.getStart(), fst.nextTimeStamp);
    			}
    		

    			@Override
				public TemporalObject nextObject() {
    				aggs.remove(fst);
    				return new TemporalObject(fst.agg, new TimeInterval(fst.timeStamp, fst.nextTimeStamp));
    			}
    		};
    	}
    	
    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#clear()
    	 */
    	@Override
		public void clear() {
    		aggs.clear();
    	}
    	
    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#close()
    	 */
    	@Override
		public void close() {
    		aggs.clear();
    	}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#size()
    	 */
    	@Override
		public int size() {
    		return aggs.size();
    	}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#iterator()
    	 */
    	@Override
		public Iterator<TemporalObject> iterator() throws UnsupportedOperationException {
    		return new AbstractCursor<TemporalObject>() {

    			Iterator<AggEntry<O>> it; 
    			AggEntry<O> fst; 

    			@Override
				protected boolean hasNextObject() {
    				if(aggs.size()<2) {
    					aggs.clear();
    					return false;
    				}
					if (it == null) {
						it = aggs.iterator();
					}
					fst = it.next();
					if (!it.hasNext())
						return false;
					return true;
				}

				@Override
				protected TemporalObject nextObject() {
					return new TemporalObject(fst.agg, new TimeInterval(fst.timeStamp, fst.nextTimeStamp));
				}
    		};
    	}
    	
    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#query(java.lang.Object, int)
    	 */
    	@Override
		public Iterator<TemporalObject> query(TemporalObject o, int ID) throws IllegalArgumentException {
    		return new EmptyCursor<TemporalObject>();
    	}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#query(E[], int[], int)
    	 */
    	@Override
		public Iterator<TemporalObject> query(TemporalObject [] os, int [] IDs, int valid) throws IllegalArgumentException {
    		return new EmptyCursor<TemporalObject>();
    	}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#query(E[], int[])
    	 */
    	@Override
		public Iterator<TemporalObject> query(TemporalObject [] os, int [] IDs) throws IllegalArgumentException {
    		return new EmptyCursor<TemporalObject>();
    	}

    	/* (non-Javadoc)
    	 * @see xxl.core.collections.sweepAreas.AbstractSweepArea#reorganize(java.lang.Object, int)
    	 */
    	@Override
		public void reorganize(TemporalObject currentStatus, int ID) throws IllegalStateException {
    		throw new UnsupportedOperationException("Results are produced during reorganization; use method expire instead.");	
    	}
    	
    	/**
    	 * @return
    	 */
    	public long getMinTimeStamp() {
    		return aggs.first().timeStamp;
    	}

		public void createMetaDataManagement() {
			if (metaDataManagement != null)
				throw new IllegalStateException("An instance of MetaDataManagement already exists.");
			metaDataManagement = new TemporalAggregatorSAMetaDataManagement();
		}
    }
    
	protected TemporalAggregatorSA<I,O> sweepArea;
	
	/**
	 * @param sweepArea
	 */
	public FastTemporalAggregator(TemporalAggregatorSA<I,O> sweepArea) {
		this.sweepArea = sweepArea;		
	}
	
	/** 
	 * Creates a new TemporalAggregator as an internal component of a query graph. <BR>
	 * The subscription to the specified sources is fulfilled immediately. <BR> 
	 * The aggregate function is invoked on elements of <CODE>source</CODE>.
	 * 
	 * @param source This pipe gets subscribed to the specified source.
	 * @param sourceID This pipe uses the given ID for subscription to the source.
	 * @param sweepArea TemporalAggregatorSweepArea managing the elements delivered by the source.
	 */
   	public FastTemporalAggregator(Source<? extends TemporalObject<I>> source, int sourceID, TemporalAggregatorSA<I,O> sweepArea) {
		this(sweepArea);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
   	
	/** 
	 * Creates a new TemporalAggregator as an internal component of a query graph. <BR>
	 * The subscription to the specified sources is fulfilled immediately using a default id. 
	 * The aggregate function is invoked on elements of <CODE>source</CODE>. <BR>
	 * The object size is set to unknown.
	 * 
	 * @param source This pipe gets subscribed to the specified source.
	 * @param sweepArea TemporalDistinctSweepArea managing the elements delivered by the source.
	 */
	public FastTemporalAggregator(Source<? extends TemporalObject<I>> source, TemporalAggregatorSA<I,O> sweepArea) {
		this(source, DEFAULT_ID, sweepArea);
	}
	
	/**
	 * @param source
	 * @param aggFunction
	 */
	public FastTemporalAggregator(Source<? extends TemporalObject<I>> source, AggregationFunction<? super I,O> aggFunction) {
		this(source, DEFAULT_ID, new TemporalAggregatorSA<I,O>(aggFunction));
	}
		
	
	static {
		MetaDataDependencies.addLocalMetaDataDependencies(FastTemporalAggregator.class, SWEEPAREA_PERIODIC_TRIGGER, AbstractPipe.AbstractPipeMetaDataManagement.SYNC_TRIGGER+"_ABSTRACTTIMESTAMPPIPE");
		MetaDataDependencies.addLocalMetaDataDependencies(FastTemporalAggregator.class, SWEEPAREA_COSTS, SWEEPAREA_PERIODIC_TRIGGER);
		MetaDataDependencies.addLocalMetaDataDependencies(FastTemporalAggregator.class, COST_MEASUREMENT, SWEEPAREA_COSTS);
	}

	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#processObject(java.lang.Object, int)
	 */
	@Override
	public void processObject(TemporalObject<I> o, int sourceID) throws IllegalArgumentException {
		sweepArea.insert(o);
		transferIterator(sweepArea.expire(o, 0));
	}

	/**
	 * @param it
	 * @return
	 */
	protected boolean transferIterator(Iterator<? extends TemporalObject> it) {
		boolean ret = it.hasNext();
		while(it.hasNext())
			transfer(it.next());
		return ret;
	}
	
	/**
	 * Calls <CODE>super.close()</CODE> and closes
	 * the sweep area.
	 */
	@Override
	public void close() {
		super.close();
		if (isClosed)
			sweepArea.close();
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#done(int)
	 */
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				transferIterator(sweepArea.iterator());
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}

	/* (non-Javadoc)
	 * @see xxl.core.pipes.memoryManager.MemoryMonitorable#getCurrentMemUsage()
	 */
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractTimeStampPipe#processHeartbeat(long, int)
	 */
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {		
		if (transferIterator(sweepArea.expire(new TemporalObject<I>(null, new TimeInterval(minTS, TimeInterval.INFINITY)), sourceID)))
			return NO_HEARTBEAT;
		return sweepArea.size() > 0 ?  sweepArea.getMinTimeStamp(): minTS;
	}
		
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractTemporalPipe#createMetaDataManagement()
	 */
	@Override
	public void createMetaDataManagement() {
		if (metaDataManagement != null)
			throw new IllegalStateException("An instance of MetaDataManagement already exists.");
		metaDataManagement = new TemporalAggregatorMetaDataManagement();
	}

	
	public static void aamain(String[] args) {
		
		Source<TemporalObject<Integer>> s = new AbstractSource<TemporalObject<Integer>>() {

			int pos = -1;
			int [] values = { 1 , 10  };
			long [] start = { 1  , 20 };
			long [] end =   { 11 , 21 };
			
			public TemporalObject<Integer> next() throws NoSuchElementException {
				if (++pos<values.length)
					return new TemporalObject(new Integer(values[pos]), new TimeInterval(start[pos], end[pos]));
				throw new NoSuchElementException("");
			}
			
		};
		
		final Sum agg = new Sum();
		FastTemporalAggregator<Integer,Number> a = new FastTemporalAggregator<Integer,Number>(s, agg);

		Printer printer = new Printer<TemporalObject<Number>>(a);
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(printer);
		exec.startAllQueries();
		
		
	}
	
	
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 1000;
		final int buckets = 10;
		final int startInc = 20;
		final int intervalSize = 120;
		final long seed = 42;
		final HashMap<Long,List<Integer>> input = new HashMap<Long,List<Integer>>();
		final Long[] hashCodes = new Long[(noOfElements*startInc)+intervalSize];
		for (int i = 0; i < hashCodes.length; i++)
			hashCodes[i] = new Long(i);	
		
		RandomNumber<Integer> r = new RandomNumber<Integer>(
			RandomNumber.DISCRETE, noOfElements, 0
		);		
		
		Source<TemporalObject<Integer>> s = new Mapper<Integer,TemporalObject<Integer>>(r,
			new Function<Integer,TemporalObject<Integer>>() {
				Random random = new Random(seed);
				long start, end;
				long newStart, newEnd;
				@Override
				public TemporalObject<Integer> invoke(Integer o) {
					Integer value;
					newStart = start+random.nextInt(startInc);
					newEnd   = newStart+1+random.nextInt(intervalSize-1);
					if (start == newStart) newEnd = end+random.nextInt(intervalSize-1);
					TemporalObject<Integer> object = new TemporalObject<Integer>(value = o % buckets, 
						new TimeInterval(start = newStart, end = newEnd)
					);
					List<Long> tmp = TemporalObject.getAllSnapshots(object);
					int key;
					List<Integer> l;
					for (int i = 0; i < tmp.size(); i++) {
						key = (int)tmp.get(i).longValue();
						l = input.get(hashCodes[key]);
						if (l == null) l = new LinkedList<Integer>();
						l.add(value);
						input.put(hashCodes[key], l);
					}
					return object;
				}
			}
		);
	
		final Sum agg = new Sum();
		FastTemporalAggregator<Integer,Number> a = new FastTemporalAggregator<Integer,Number>(s, agg);
		
		Pipes.verifyStartTimeStampOrdering(a);
		
		AbstractSink sink = new AbstractSink<TemporalObject<Number>>(a) {
			LinkedList<TemporalObject<Number>> list = new LinkedList<TemporalObject<Number>>();
			
			@Override
			public void processObject(TemporalObject<Number> o, int sourceID) {				
				list.add(o);
			}
			
			@Override
			public void done(int sourceID) {
				super.done(sourceID);
				if (isDone()) {
					Iterator<TemporalObject<Number>> listIt = list.iterator();
					HashMap<Long,List<Number>> result = new HashMap<Long,List<Number>>();
					while (listIt.hasNext()) {
						TemporalObject<Number> next = listIt.next();
						List<Long> tmp = TemporalObject.getAllSnapshots(next);
						int key;
						List<Number> l;
						for (int i = 0; i < tmp.size(); i++) {
							key = (int)tmp.get(i).longValue();
							l = result.get(hashCodes[key]);
							if (l == null) l = new LinkedList<Number>();
							l.add(next.getObject()); 
							result.put(hashCodes[key], l);
						}
					}
					
					List<Integer> in;
					List<Number> out;
					for (int i = 0; i < hashCodes.length; i++) {
						in = input.get(hashCodes[i]);
						out = result.get(hashCodes[i]);	
						if ((in == null && out != null) || (in != null && out == null))
							System.err.println("ERROR: Difference in aggregate at snapshot "+i);
						if (in != null && out != null) {
							if (out.size() > 1)
								System.out.println("ERROR at snapshot "+i);
							Number aggValue = null;
							for (int j = 0; j < in.size(); j++) {
								aggValue = agg.invoke(aggValue, in.get(j));
							}
							if (!aggValue.equals(out.get(0))) {
								System.err.println("ERROR: Difference in aggregate at snapshot "+i);
								System.err.println("proper value: "+aggValue+" current value: "+out.get(0));
							}
						}
						System.out.println("CHECK finished for snapshot: "+i);		
					}
				}
			}
		};
		
		Printer printer = new Printer<TemporalObject<Number>>(a);
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(sink);
		exec.registerQuery(printer);
		exec.startAllQueries();
		
	}
}
