package xxl.core.pipes.operators.mappers;

import xxl.core.functions.Function;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.metaData.DefaultMetaDataHandler;
import xxl.core.pipes.metaData.PeriodicEvaluationMetaDataHandler;
import xxl.core.pipes.operators.AbstractPipe;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;

public class NonTemporal2TemporalMapper<I> extends NonTimeStamped2TimeStampedMapper<I, TemporalObject<I>>{

	/**
	 * Inner class for metadata management. 
	 */
	public class NonTemporal2TemporalMapperMetaDataManagement extends NonTimeStamped2TimeStampedMapperMetaDataManagement {
		
		// measured parameters
		public static final String L = "L"; // average length of time intervals
				
		protected volatile boolean measureL;
		protected volatile Double avgL;
		protected Average avgLFunction;
		
		public static final String L_ESTIMATION = "L_ESTIMATION";
		
		protected Double avgLSet;

		/**
		 * 
		 */
		public NonTemporal2TemporalMapperMetaDataManagement() {
			super();
			this.measureL = false;
			this.avgLSet = Double.NaN;
		}
		
		/**
		 * Used if parameter l is known in advance.
		 * 
		 * @param l
		 */
		public synchronized void setL(double l) {
			this.avgLSet = l;
			refresh(L_ESTIMATION);
		}
		
		/* (non-Javadoc)
		 * @see xxl.core.pipes.operators.AbstractTimeStampPipe.AbstractTimeStampPipeMetaDataManagement#addMetaData(java.lang.Object)
		 */
		@Override
		protected boolean addMetaData(Object metaDataIdentifier) {	
			if (metaDataIdentifier.equals(L)) {
				measureL = true;
				avgL = Double.NaN;
				avgLFunction = new Average();
				metaData.add(metaDataIdentifier, new PeriodicEvaluationMetaDataHandler<Double>(this, metaDataIdentifier, Double.class, 
					new Function<Object,Double>() {			
												
						@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								if (isClosed) 
									return Double.NaN;
								double l = avgL == null ? Double.NaN : avgL;
								avgL = null; // reset average function
								return l;
							}
						}
					}, updatePeriod)
				);
				return true;
			}
			if (metaDataIdentifier.equals(L_ESTIMATION)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Double>(this, metaDataIdentifier,
					new Function<Object,Double>() {
						
					@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								return avgLSet;
							}
						}
					})
				);
				return true;
			}
			
			return super.addMetaData(metaDataIdentifier);
		}
		
		/* (non-Javadoc)
		 * @see xxl.core.pipes.operators.AbstractTimeStampPipe.AbstractTimeStampPipeMetaDataManagement#removeMetaData(java.lang.Object)
		 */
		@Override
		protected boolean removeMetaData(Object metaDataIdentifier) {
			
			if (metaDataIdentifier.equals(L)) {
				measureL = false;
				avgL = Double.NaN;
				avgLFunction = null;
				((PeriodicEvaluationMetaDataHandler)metaData.get(metaDataIdentifier)).close();
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (metaDataIdentifier.equals(L_ESTIMATION)) {
				metaData.remove(metaDataIdentifier);
				return true;
			}
			
			return super.removeMetaData(metaDataIdentifier);
		}
	}
	
	/**
	 * Helps to create a new pipe in a query graph with the given source.
	 * @param source source to connect the pipe with 
	 * @param sourceID ID of the source 
	 */
	public NonTemporal2TemporalMapper(Source<? extends I> source, int sourceID, Function<? super I, ? extends TemporalObject<I>> mapping) {
		super(source, sourceID, mapping);		
	}

	/**
	 * Helps to create a new pipe in a query graph with the given source.
	 * @param source source to connect the pipe with 
	 */
	public NonTemporal2TemporalMapper(Source<? extends I> source, Function<? super I, ? extends TemporalObject<I>> mapping) {
		super(source, mapping);
	}

	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractTimeStampPipe#transfer(xxl.core.pipes.elements.TimeStampedObject)
	 */
	@Override
	public void transfer(TemporalObject<I> o) {
		super.transfer(o);
		NonTemporal2TemporalMapperMetaDataManagement mdm = (NonTemporal2TemporalMapperMetaDataManagement)metaDataManagement;
		if (mdm.measureL)
			synchronized(metaDataManagement) {
				if (mdm.measureL)
					mdm.avgL = (Double)mdm.avgLFunction.invoke(mdm.avgL, new Double(o.getEnd() - o.getStart()));
			}
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractTimeStampPipe#createMetaDataManagement()
	 */
	@Override
	public void createMetaDataManagement() {
		if (metaDataManagement != null)
			throw new IllegalStateException("An instance of MetaDataManagement already exists.");
		metaDataManagement = new NonTemporal2TemporalMapperMetaDataManagement();
	}	
	
	public static void main(String[] args) {		
		final AbstractPipe<Integer,TemporalObject<Integer>> pipe = new NonTemporal2TemporalMapper<Integer>(
							new RandomNumber<Integer>(true, 100), 
							TemporalObject.assignEquidistantTimeInterval(0, 100, Integer.class)
						);			
		
		pipe.getMetaDataManagement().includeAll(
				NonTemporal2TemporalMapper.NonTemporal2TemporalMapperMetaDataManagement.D,
				NonTemporal2TemporalMapper.NonTemporal2TemporalMapperMetaDataManagement.G,
				NonTemporal2TemporalMapper.NonTemporal2TemporalMapperMetaDataManagement.NO_OF_SOURCES				
		);
				
		new Thread(){
			@Override
			public void run() {
				new Printer(pipe).openAllSources();	
				while(!pipe.isClosed()){
					try {Thread.sleep(1000);
					} catch (Exception e) {}
					System.out.println(pipe.getMetaData());
				}					
			}
		}.start();
	}
	
}
