package xxl.core.pipes.operators.mappers;

import xxl.core.functions.Function;
import xxl.core.math.Maths;
import xxl.core.pipes.elements.TimeStampedObject;
import xxl.core.pipes.memoryManager.heartbeat.Heartbeat;
import xxl.core.pipes.memoryManager.heartbeat.Heartbeats;
import xxl.core.pipes.metaData.DefaultMetaDataHandler;
import xxl.core.pipes.metaData.PeriodicEvaluationMetaDataHandler;
import xxl.core.pipes.operators.AbstractPipe;
import xxl.core.pipes.operators.AbstractTimeStampPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sinks.SinkIsDoneException;
import xxl.core.pipes.sources.AbstractTimeStampSource;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.SourceIsClosedException;
import xxl.core.predicates.Predicate;

public class NonTimeStamped2TimeStampedMapper<I, O extends TimeStampedObject<I>> extends AbstractPipe<I, O> implements Heartbeat{

	protected Function<? super I, ? extends O> mapping;
	
	/**
	 * An array containing the latest timestamp delivered by each source. 
	 */
	protected long[] latestTimeStamps;
	
	protected int lastMinIndex;
	
	/**
	 * The minimum of lastTimeStamps
	 */
	protected long minTimeStamp;
	
	/**
	 * The value of the last heartbeat that has been delivered to all subscribed sinks that support heartbeats.
	 */
	protected long lastHB;
	
	/**
	 * The i-th position is true, if the i sink supports heartbeats.
	 */
	protected boolean[] supportsHeartbeat;
	
	/**
	 * Indicates if heartBeats are activated
	 */
	protected volatile boolean activateHeartbeats;
	
	/**
	 * A heartbeat is transfered to all subscribed sinks that support heartbeats, if it has a higher timestamp and this predicates evaluates
	 * to true. Therefore, this predicate allows to control the overhead due to the use of heartbeats.  
	 */
	@SuppressWarnings("unchecked")
	protected Predicate heartbeatPredicate = Heartbeats.everyNMillis(1000);
	
	protected volatile boolean measureD = false;
	protected volatile boolean measureG = false;

	/**
	 * Inner class for metadata management. 
	 */
	public class NonTimeStamped2TimeStampedMapperMetaDataManagement extends AbstractPipeMetaDataManagement {
		
		// measured parameters
		public static final String D = "D"; // average time distance between successive timestamps
		public static final String G = "G"; // time granularity as greatest common divisor of all occurring timestamps
		public static final String D_ESTIMATION = "D_ESTIMATION";
		public static final String G_ESTIMATION = "G_ESTIMATION";
									
		protected long costModelOutputCounter;
		protected long firstOutputTimeStamp;
		protected long lastOutputTimeStamp;
		
		protected long g;
		protected volatile short gState;
		protected double dSet;
		protected long gSet;
	
		/**
		 * 
		 */
		public NonTimeStamped2TimeStampedMapperMetaDataManagement() {
			super();
			this.dSet = Double.NaN;
			this.gSet = -1;
		}
		
		/**
		 * Used if parameter d is known in advance.
		 * 
		 * @param d
		 */
		public synchronized void setD(double d) {
			this.dSet = d;
			refresh(D_ESTIMATION);
		}
		
		/**
		 * Used if parameter g is known in advance.
		 * 
		 * @param g
		 */
		public synchronized void setG(long g) {
			this.gSet = g;
			refresh(G_ESTIMATION);
		}
		
		/* (non-Javadoc)
		 * @see xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement#addMetaData(java.lang.Object)
		 */
		@Override
		protected boolean addMetaData(Object metaDataIdentifier) {
			
			if (metaDataIdentifier.equals(D)) {
				measureD = true;
				costModelOutputCounter = 0;
				firstOutputTimeStamp = Long.MIN_VALUE;
				metaData.add(metaDataIdentifier, new PeriodicEvaluationMetaDataHandler<Double>(this, metaDataIdentifier,
					new Function<Object,Double>() {
						@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								if (isClosed || firstOutputTimeStamp == Long.MIN_VALUE) 
									return Double.NaN;
								double d = costModelOutputCounter == 0 ? Double.NaN : 
									((double)lastOutputTimeStamp - firstOutputTimeStamp)/costModelOutputCounter;
								firstOutputTimeStamp = lastOutputTimeStamp;														
								costModelOutputCounter = 0;
								return d;
							}
						}
					}, updatePeriod)
				);
				return true;
			}
			if (metaDataIdentifier.equals(G)) {
				measureG = true;
				g = 1;
				gState = 0;
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Long>(this, metaDataIdentifier,
					new Function<Object,Long>() {			
						@Override
						public Long invoke() {
							synchronized(metaDataManagement) {
								if (isClosed || (gState < 2 && measureD)) 
									return -1l;
								return g;
							}
						}
					})
				);
				return true;
			}
			if (metaDataIdentifier.equals(D_ESTIMATION)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Double>(this, metaDataIdentifier,
					new Function<Object,Double>() {
						@Override
						public Double invoke() {
							synchronized(metaDataManagement) {
								return dSet;
							}
						}
					})
				);
				return true;
			}
			if (metaDataIdentifier.equals(G_ESTIMATION)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Long>(this, metaDataIdentifier,
					new Function<Object,Long>() {
						@Override
						public Long invoke() {
							synchronized(metaDataManagement) {
								return gSet;
							}
						}
					})
				);
				return true;
			}
			if (super.addMetaData(metaDataIdentifier))
				return true;
			return false;
		}
		
		/* (non-Javadoc)
		 * @see xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement#removeMetaData(java.lang.Object)
		 */
		@Override
		protected boolean removeMetaData(Object metaDataIdentifier) {
			
			if (metaDataIdentifier.equals(D)) {
				measureD = false;
				costModelOutputCounter = 0;
				firstOutputTimeStamp = Long.MIN_VALUE;
				((PeriodicEvaluationMetaDataHandler)metaData.get(metaDataIdentifier)).close();
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (metaDataIdentifier.equals(G)) {
				measureG = false;
				g = 1;
				gState = 0;
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (metaDataIdentifier.equals(D_ESTIMATION) ||
				    metaDataIdentifier.equals(G_ESTIMATION)) {
					metaData.remove(metaDataIdentifier);
					return true;
				}
			if (super.removeMetaData(metaDataIdentifier))
				return true;
			
			return false;
		}
	}
	
	/** 
	 * Helps to create a new pipe in a query graph with the given sources.
	 * @param sources sources to connect the pipe with
	 * @param IDs IDs of the sources
	 */
	public NonTimeStamped2TimeStampedMapper(Source<? extends I>[] sources, int[] sourceIDs, Function<? super I, ? extends O> mapping) {
		super(sources, sourceIDs);
		if(sources instanceof NonTimeStamped2TimeStampedMapper[] || sources instanceof AbstractTimeStampSource[] || sources instanceof AbstractTimeStampPipe[])
			throw new IllegalArgumentException("Nontemporal sources expected!");

		this.mapping = mapping;
		this.activateHeartbeats = false;
		this.lastHB = -1;
		this.minTimeStamp = Long.MAX_VALUE;
		this.lastMinIndex = -1;
	}

	/** 
	 * Helps to create a new pipe in a query graph with the given sources.
	 * @param sources sources to connect the pipe with
	 */
	public NonTimeStamped2TimeStampedMapper(Source<? extends I>[] sources, Function<? super I, ? extends O> mapping) {
		super(sources);
		if(sources instanceof NonTimeStamped2TimeStampedMapper[] || sources instanceof AbstractTimeStampSource[] || sources instanceof AbstractTimeStampPipe[])
			throw new IllegalArgumentException("Nontemporal sources expected!");

		this.mapping = mapping;
		this.activateHeartbeats = false;
		this.lastHB = -1;
		this.minTimeStamp = Long.MAX_VALUE;
		this.lastMinIndex = -1;
	}

	/** 
	 * Helps to create a new pipe in a query graph with the given source.
	 * @param source source to connect the pipe with
	 * @param ID ID of the source
	 */
	public NonTimeStamped2TimeStampedMapper(Source<? extends I> source, int sourceID, Function<? super I, ? extends O> mapping) {
		super(source, sourceID);
		if(source instanceof NonTimeStamped2TimeStampedMapper || source instanceof AbstractTimeStampSource || source instanceof AbstractTimeStampPipe)
			throw new IllegalArgumentException("Nontemporal source expected!");

		this.mapping = mapping;
		this.activateHeartbeats = false;
		this.lastHB = -1;
		this.minTimeStamp = Long.MAX_VALUE;
		this.lastMinIndex = -1;
	}

	/**
	 * Helps to create a new pipe in a query graph with the given source.
	 * @param source source to connect the pipe with
	 */
	public NonTimeStamped2TimeStampedMapper(Source<? extends I> source, Function<? super I, ? extends O> mapping) {
		super(source);
		if(source instanceof NonTimeStamped2TimeStampedMapper || source instanceof AbstractTimeStampSource || source instanceof AbstractTimeStampPipe)
			throw new IllegalArgumentException("Nontemporal source expected!");
			
		this.mapping = mapping;
		this.activateHeartbeats = false;
		this.lastHB = -1;
		this.minTimeStamp = Long.MAX_VALUE;
		this.lastMinIndex = -1;
	}

	/**
	 * Similar to the <CODE>subscribe</CODE> method, this method
	 * establishes a connection between this pipe and a source during runtime. The intention of
	 * both methods is to build up a double-linked query graph, which is completely navigatable. <BR>
	 * The method should be used to store a reference to the given underlying source
	 * together with the ID specified during this pipe's subscription. <BR>
	 * The source and the ID used during subscription are stored in internal arrays.
	 * Their positions in both arrays are equal.
	 *
	 * @param source The source this pipe got subscribed to.
	 * @param sourceID The ID this pipe used during its subscription by the specified source.
	 * @return Returns <CODE>true</CODE>, if the add operation was successful, otherwise <CODE>false</CODE>.
	 * @throws xxl.core.pipes.sinks.SinkIsDoneException If the pipe has already finished its processing.
	 */
	@Override
	public boolean addSource(Source<? extends I> source, int sourceID) throws SinkIsDoneException {
		graph.WLock.lock();
		try {
			boolean ret = super.addSource(source, sourceID);	
			if (ret) {
				if (latestTimeStamps == null) {
					latestTimeStamps = new long[]{Long.MIN_VALUE};
					this.lastMinIndex = 0;
				}
				else {
					long[] newTS = new long[latestTimeStamps.length+1];
					System.arraycopy(latestTimeStamps, 0, newTS, 0, latestTimeStamps.length);
					newTS[latestTimeStamps.length] = Long.MIN_VALUE;
					latestTimeStamps = newTS;
				}
			}
			return ret;
		}
		finally {
			graph.WLock.unlock();
		}
	}

	/**
	 * In conformity to the <CODE>unsubscribe</CODE> method, this
	 * method removes the connection between this pipe and a source on the pipe's side.
	 * Therefore it removes the given source with ID <CODE>sourceID</CODE> from the
	 * internal arrays used to store this pipe's references to its sources.
	 * 
	 * @param source The source this pipe got subscriped to.
	 * @param sourceID The ID this pipe used during its subscription by the specified source.
	 * @return Returns <CODE>true</CODE>, if the remove operation was successful, otherwise <CODE>false</CODE>.
	 * @throws xxl.core.pipes.sinks.SinkIsDoneException If the sink has already finished its processing.
	 */
	@Override
	public boolean removeSource(Source<? extends I> source, int sourceID) throws SinkIsDoneException {
		graph.WLock.lock();
		try {
			int index = -1;
			for (int i = 0; i < sources.length; i++)
				if (sources[i] == source && sourceIDs[i] == sourceID)
					index = i;
			boolean ret = super.removeSource(source, sourceID);	
			if (ret) {
				long[] newTS = new long[latestTimeStamps.length-1];
				System.arraycopy(latestTimeStamps, 0, newTS, 0, index);						
				if (index < latestTimeStamps.length-1) {
					System.arraycopy(latestTimeStamps, index+1, newTS, index, latestTimeStamps.length-1-index);
				}
				latestTimeStamps = newTS;
				if (lastMinIndex == index) 
					lastMinIndex = -1;
				else
					if (lastMinIndex > index)
						lastMinIndex--;
			}
			return ret;
		}
		finally {
			graph.WLock.unlock();
		}
	}

	/**
	 * This method will be called in the <CODE>transfer</CODE> method of each source
	 * this pipe got subscribed to. The element to be transferred will be sent to this pipe's
	 * sink part selecting it as the first paramenter. The second parameter specifies
	 * the ID this sink used during its subscription. Therefore the ID determines from
	 * which source this element is from. <BR>
	 * The element should not be cloned, if it is transferred to multiple sinks, which
	 * have been registered by a source. Due to performance reasons this pipe should rather 
	 * use the same element and pass it to all subscribed sinks based on the assumption
	 * that these sinks will not perform any changes on this element, unless they clone
	 * it themselves. <BR>
	 * The implementation consumes the element and updates metadata information. Furthermore,
	 * it calls the abstract method <CODE>processObject</CODE> in order to process the
	 * incoming object.
	 *
	 * @param o The element streaming in.
	 * @param sourceID One of the IDs this pipe specified during its subscription by the underlying sources. 
	 * @throws java.lang.IllegalArgumentException If the given element or ID is incorrect.
 	 */
	@Override
	public void process(I p, int sourceID) throws IllegalArgumentException {
		processingWLock.lock();
		graph.RLock.lock();
		try {
			O o = mapping.invoke(p);
			long time = o.getTimeStamp();
			int index = Pipes.getSourceIndex(sourceIDs, sourceID);
			if (latestTimeStamps[index] == Long.MIN_VALUE)
				lastMinIndex = -1;
			latestTimeStamps[index] = Math.max(time, latestTimeStamps[index]);
			updateMinTimeStamp(index);				
			super.process(p, sourceID);
			transfer(o);
			lastHB = Math.max(lastHB, minTimeStamp);
		}
		finally {
			graph.RLock.unlock();
			processingWLock.unlock();
		}		
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#close()
	 */
	@Override
	public void close() {
		super.close();
		processingWLock.lock();
		try {
			latestTimeStamps = null;
			supportsHeartbeat = null;
		}
		finally {
			processingWLock.unlock();
		}
	}
		
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#subscribe(xxl.core.pipes.sinks.Sink, int)
	 */
	@Override
	public boolean subscribe(Sink<? super O> sink, int sinkID) throws SourceIsClosedException {
		graph.WLock.lock();
		try {
			boolean ret = super.subscribe(sink, sinkID);
			if (ret) {
				if (supportsHeartbeat == null) {
					supportsHeartbeat = new boolean[]{sink instanceof Heartbeat};
				}
				else {
					boolean[] tmp = new boolean[supportsHeartbeat.length+1];
					System.arraycopy(supportsHeartbeat, 0, tmp, 0, supportsHeartbeat.length);
					tmp[supportsHeartbeat.length] = sink instanceof Heartbeat;
					supportsHeartbeat = tmp;
				}
			}
			return ret;
		}
		finally {
			graph.WLock.unlock();
		}
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#unsubscribe(xxl.core.pipes.sinks.Sink, int)
	 */
	@Override
	public boolean unsubscribe(Sink<? super O> sink, int sinkID) throws SourceIsClosedException {
		graph.WLock.lock();
		try {
			int index = -1;
			for (int i = 0; i < sinks.length; i++)
				if (sinks[i] == sink && sinkIDs[i] == sinkID) 
					index = i;
			boolean ret = super.unsubscribe(sink, sinkID);
			if (ret) {
				boolean[] tmp = new boolean[supportsHeartbeat.length-1];
				System.arraycopy(supportsHeartbeat, 0, tmp, 0, index);						
				if (index < supportsHeartbeat.length-1) {
					System.arraycopy(supportsHeartbeat, index+1, tmp, index, supportsHeartbeat.length-1-index);
				}
				supportsHeartbeat = tmp;
			}
			return ret;
		}
		finally {
			graph.WLock.unlock();
		}
	}
		
	/**
	 * Returns an array containing the latest timestamp delivered by each source.
	 * 
	 * @return array containing the latest timestamp delivered by each source
	 */
	public long[] getLatestTimeStamps() {
		processingRLock.lock();
		try {
			return latestTimeStamps;
		}
		finally {
			processingRLock.unlock();
		}
	}

	/* (non-Javadoc)
	 * @see xxl.core.pipes.memoryManager.heartbeat.Heartbeat#heartbeat(long, int)
	 */
	@SuppressWarnings("unchecked")
	public void heartbeat(long timeStamp, int sourceID) {
		if(!activateHeartbeats) return;
		processingWLock.lock();
		try {
			if(!activateHeartbeats) return;
			if (isOpened && ! isClosed && sinks != null) {
				boolean update = false; 
				if (sourceID != MEMORY_MANAGER) { // heartbeat from source
					int index = Pipes.getSourceIndex(sourceIDs, sourceID);
					if (latestTimeStamps[index] < timeStamp) {				
						latestTimeStamps[index] = timeStamp;
						update = true;
					}
				}
				else { // heartbeat from memory manager
					for (int i = 0; i < latestTimeStamps.length; i++)
						if (latestTimeStamps[i] < timeStamp) {
							latestTimeStamps[i] =  timeStamp;
							update = true;
						}
				}
				if (update)
					updateMinTimeStamp(Pipes.getSourceIndex(sourceIDs, sourceID));	
				if (lastHB < minTimeStamp && heartbeatPredicate.invoke()) {
					long heartbeat = processHeartbeat(minTimeStamp, sourceID); 
					if (heartbeat > 0) 
						transferHeartbeat(heartbeat);
					lastHB = heartbeat > 0 ? heartbeat : minTimeStamp;
				}
			}
		}
		finally {
			processingWLock.unlock();	
		}
	}
	
	// returns the value of the heartbeat that is sent to all sinks.
	// a value <= 0 means that no heartbeat will be sent. 
	// only called inside heartbeat()
	/**
	 * 
	 * @param minTS
	 * @param sourceID
	 * @return
	 */
	protected long processHeartbeat(long minTS, int sourceID) {
		return minTS;
	}
	
	/**
	 * @param timeStamp
	 */
	protected void transferHeartbeat(long timeStamp) {
		graph.RLock.lock();
		try {
			if (isOpened && ! isClosed && sinks != null)
				for (int i = 0; i < sinks.length; i++)
					if (supportsHeartbeat[i])
						((Heartbeat)sinks[i]).heartbeat(timeStamp, sinkIDs[i]);
			}
		finally {
			graph.RLock.unlock();	
		}
	}

	/**
	 * 
	 */
	protected void updateMinTimeStamp(int sourceIndex) {
		if (lastMinIndex < 0 || sourceIndex == lastMinIndex) {
			minTimeStamp = Long.MAX_VALUE;
			long l;
			for (int i = 0; i < latestTimeStamps.length; i++) {
				l = latestTimeStamps[i];
				if (l == Long.MIN_VALUE) {
					minTimeStamp = 0;
					break;
				}
				if (minTimeStamp > l) {
					minTimeStamp = l;
					lastMinIndex = i;
				}
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#updateDoneStatus(int)
	 */
	@Override
	public void updateDoneStatus(int sourceID) {
		processingWLock.lock();
		graph.RLock.lock();
		try {
			super.updateDoneStatus(sourceID);
			int index = Pipes.getSourceIndex(sourceIDs, sourceID);
			latestTimeStamps[index] = Long.MAX_VALUE;
			updateMinTimeStamp(index);
		}
		finally {
			graph.RLock.unlock();
			processingWLock.unlock();
		}
	}

	/* (non-Javadoc)
	 * @see xxl.core.pipes.memoryManager.heartbeat.HeartbeatPipe#setHeartbeatPredicate(xxl.core.predicates.Predicate)
	 */
	@SuppressWarnings("unchecked")
	public void setHeartbeatPredicate(Predicate heartbeatPredicate) {
		processingWLock.lock();
		try {
			this.heartbeatPredicate = heartbeatPredicate;
		}
		finally {
			processingWLock.unlock();	
		}
	}
	
	/**
	 * Activates or deactivates heartbeats for this sink. In order to activate heartbeats
	 * in a query graph top-down or bottom-up use static methods provided in Heartbeats.
	 * 
	 * @param on true activates heartbeats. 
	 * @see xxl.core.pipes.memoryManager.heartbeat.Heartbeats
	 */
	public void activateHeartbeats(boolean on) {
		processingWLock.lock();
		try {
			this.activateHeartbeats = on;
		}
		finally {
			processingWLock.unlock();	
		}
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#transfer(java.lang.Object)
	 */
	@Override
	public void transfer(O o) {
		super.transfer(o);
		if (measureD || measureG) {
			synchronized(metaDataManagement) {
				NonTimeStamped2TimeStampedMapperMetaDataManagement mdm = (NonTimeStamped2TimeStampedMapperMetaDataManagement)metaDataManagement;
				if (measureG) {
					if (mdm.gState > 1) {
						mdm.g = Maths.gcd(mdm.g, o.getTimeStamp());
					}
					else {
						if (mdm.gState == 1) {
							mdm.g = o.getTimeStamp();
							mdm.gState++;
						}
						else mdm.gState++;
					}
				}
				if (measureD) {
					if (mdm.firstOutputTimeStamp == Long.MIN_VALUE)
						mdm.firstOutputTimeStamp = o.getTimeStamp();
					else 
						mdm.costModelOutputCounter++;
					mdm.lastOutputTimeStamp = o.getTimeStamp();
				}
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#createMetaDataManagement()
	 */
	@Override
	public void createMetaDataManagement() {
		if (metaDataManagement != null)
			throw new IllegalStateException("An instance of MetaDataManagement already exists.");
		metaDataManagement = new NonTimeStamped2TimeStampedMapperMetaDataManagement();
	}

	/**
	 * Returns minTimeStamp.
	 * @return minTimeStamp
	 */
	public long getMinTimeStamp() {
		processingRLock.lock();
		try {
			return minTimeStamp;
	    }
		finally {
			processingRLock.unlock();
		}
	}

	/**
	 * Returns the value of the last heartbeat.
	 * @return value of the last heartbeat
	 */
	public long getLastHB() {
		processingRLock.lock();
		try {
			return lastHB;
	    }
		finally {
			processingRLock.unlock();
		}		
	}

	@Override
	public void processObject(I o, int sourceID) throws IllegalArgumentException {
		/* NOP */
	}
}
