/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.mappers;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;

import xxl.core.collections.MapEntry;
import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.collections.sweepAreas.PNSweepArea;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.functions.Function;
import xxl.core.math.functions.AggregationFunction;
import xxl.core.math.statistics.parametric.aggregates.Sum;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.windows.PNTemporalWindow;
import xxl.core.pipes.operators.windows.TemporalWindow;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

/*
 * General implementation permits multiple positive PNObjects with equal timestamps.
 */
public class PNAggregator<I,O> extends AbstractPNPipe<I,O> implements MemoryMonitorable {
	
	protected PNSweepArea<I> sweepArea;
	protected AggregationFunction<? super I,O> aggFunction;
	protected O previousAgg;
	protected O intermediateResult;
	protected long lastTimeStamp;
	protected boolean negativeInput;
	
 	
	public PNAggregator(AggregationFunction<? super I,O> aggFunction, Function<? super PNObject<I>, Integer> hashFunction) {
		super();
		this.aggFunction = aggFunction;
		Function<Object,List<PNObject<I>>> newList = new Function<Object, List<PNObject<I>>>() {
			@Override
			public List<PNObject<I>> invoke() {
				return new LinkedList<PNObject<I>>();
			}
		};
		this.sweepArea = new PNSweepArea<I>(
			new HashSAImplementor<PNObject<I>>(hashFunction, newList, 2, PNObject.VALUE_EQUIVALENCE_PREDICATE) {
				
				public Iterator<PNObject<I>> query(PNObject<I> o, int ID) throws IllegalArgumentException {
					return iterator();	
				}
					
			},
			0,
			PNObject.VALUE_EQUIVALENCE_PREDICATE,
			1
		);
		this.previousAgg = null;
		this.lastTimeStamp = Long.MIN_VALUE;
		this.intermediateResult = null;
		this.negativeInput = false;
	}
	
	public PNAggregator(Source<? extends PNObject<I>> source, int sourceID, AggregationFunction<? super I,O> aggFunction, Function<? super PNObject<I>, Integer> hashFunction) {
		this(aggFunction, hashFunction);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");		
	}
	
	public PNAggregator(Source<? extends PNObject<I>> source, AggregationFunction<? super I,O> aggFunction) {
		this(source, DEFAULT_ID, aggFunction, 
			new Function<PNObject<I>, Integer>() {
				@Override
				public Integer invoke(PNObject<I> pno) {
					return pno.getObject().hashCode();
				}
			});
	}
	
	@Override
	public void processObject(PNObject<I> pno, int sourceID) throws IllegalArgumentException {	
		// the timestamp of the object to process
		long currentTimeStamp = pno.getTimeStamp();
		// the object to process
		I currentObject = pno.getObject();
				
		// check for progress in time
		if (currentTimeStamp > lastTimeStamp) {
			processLastTimeStamp();
			lastTimeStamp = currentTimeStamp;
		}
		
		if (pno.isPositive()) {
			// if only positive inputs: incremental calculation
			if (!negativeInput) {
				// check if incremental calculation in progress ...
				O agg;
				if (intermediateResult != null)
					agg = intermediateResult;
				// ... or start one with last aggregate
				else					
					agg = previousAgg;
				intermediateResult = aggFunction.invoke(agg, currentObject);
			}
			sweepArea.insert(pno);
		}
		else {
			sweepArea.reorganize(pno, sourceID);
			negativeInput = true;
		}
	}
	
	protected void processLastTimeStamp() {
		// progress in time, all tuples for previous timestamp can be released
		// produce negative tuples for old aggregate value of the group
		if (previousAgg != null)
			transfer(
				new PNObject<O>(
					previousAgg, 
					lastTimeStamp,
					false
				)
			); 				
		// determine new aggregate for the group
		O finalResult = null;
		if (negativeInput) {
			// negative tuple arrived for the group => recalculate aggregate
			Iterator<PNObject<I>> it = sweepArea.iterator();
			while (it.hasNext()) 
				finalResult = aggFunction.invoke(finalResult, it.next().getObject());	
		}		
		else
			finalResult = intermediateResult;
		// produce positive tuple for the new aggregate if any
		if (finalResult != null) {
			transfer(
				new PNObject<O>(
					finalResult, 
					lastTimeStamp,
					true
				)
			);
			previousAgg = finalResult;
		}
		else { 
			previousAgg = null;
		}
		
		negativeInput = false;
		intermediateResult = null;	
	}
	
	@Override
	public void close() {
		super.close();
		if (isClosed) {
			sweepArea.close();
		}
	}
	
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				processLastTimeStamp();
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
			
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 50000;
		final int maxValue = 25;
		final int maxTSInc = 20;
		int seed = 42;
		
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxTSInc), noOfElements)
		);
		
		final long windowSize = 15000;
		
		// time interval approach
		TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s1, windowSize);
		TemporalAggregator<Integer,Number> res1 = new TemporalAggregator<Integer,Number>(w1, new Sum());
		
		// positive-negative approach
		TemporalPNPipe<Integer> pnSource = new TemporalPNPipe<Integer>(s1);
		PNTemporalWindow<Integer> pnWindow = new PNTemporalWindow<Integer>(pnSource, windowSize);
		PNAggregator<Integer,Number> pnAgg = new PNAggregator<Integer,Number>(pnWindow, new Sum());
		PNTemporalPipe<Number> res2 = new PNTemporalPipe<Number>(pnAgg);
		
		QueryExecutor exec = new QueryExecutor();
//		exec.registerQuery(new Printer(res1));
//		exec.registerQuery(new Printer(pnAgg));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res1));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(pnAgg));
		exec.registerQuery(Pipes.compareSnapshots(res1, res2, false, false));
		exec.startAllQueries();

	}


}
