/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.mappers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.collections.sweepAreas.ListSA;
import xxl.core.collections.sweepAreas.SweepAreaImplementor;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTimeStampPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;
import xxl.core.util.XXLSystem;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

/**
 * Converts a stream of PNObjects to a stream of TemporalObjects.
 * 
 * @param <I> 
 */
public class PNTemporalPipe<I> extends AbstractTimeStampPipe<I,I, PNObject<I>, TemporalObject<I>> implements MemoryMonitorable {
	
	@SuppressWarnings("hiding")
	public static class PNTemporalPipeSA<I> extends ListSA<PNObject<I>> {
		
		public PNTemporalPipeSA(SweepAreaImplementor<PNObject<I>> impl, int objectSize, List<PNObject<I>> list) {
    		super(impl, DEFAULT_ID, true, ReorganizeModes.LAZY_REORGANIZE, PNObject.VALUE_EQUIVALENCE_PREDICATE, PNObject.TEMPORAL_BEFORE_REORGANIZE, 1, list, objectSize);
    	}
    	
    	public PNTemporalPipeSA(Function<? super PNObject<I>,Integer> hashFunction) {
    		super(new HashSAImplementor<PNObject<I>>(
        			hashFunction,
        			new Function<Object,List<PNObject<I>>>() {
        				@Override
						public List<PNObject<I>> invoke() {
        					return new LinkedList<PNObject<I>>();
        				}
        			}, 1),
        			DEFAULT_ID, true, ReorganizeModes.LAZY_REORGANIZE,
        			PNObject.VALUE_EQUIVALENCE_PREDICATE, 
        			PNObject.TEMPORAL_BEFORE_REORGANIZE,
        			1, new ArrayList<PNObject<I>>(), SIZE_UNKNOWN
        	);
    	}
    	
    	public PNTemporalPipeSA() {
    		this(new Function<PNObject<I>,Integer>() {
    				@Override
					public Integer invoke(PNObject<I> o) {
    					return o.getObject().hashCode();
    				}
    			}
    		);
    	}
    	
    	@Override
		public void insert(PNObject<I> o) throws IllegalArgumentException {
    		if (checkObjectSize)
    			computeObjectSize(o);
    		int index = Collections.binarySearch(list, o, PNObject.TIMESTAMP_COMPARATOR);
    		if (index < 0)
    			list.add(-index-1, o);
    		else
    			list.add(index, o);
    		impl.insert(o);
    	}
    	
    	public long getOldestTimeStamp() {
    		if (list.size() > 0) {
    			return list.get(0).getTimeStamp();
    		}
    		return Long.MIN_VALUE;
    	}
    	
    	public void remove(PNObject<I> object) {
    		if(!impl.remove(object)) // remove from implementor
    			throw new IllegalStateException("Could not remove element from implementor");
    		
    		// remove from list
			int index = Collections.binarySearch(list, object, PNObject.TIMESTAMP_COMPARATOR);
			PNObject cur = list.get(index);
			if (cur.equals(object)) {
				list.remove(index);
				return;
			}
			PNObject suc;
			for (int j = index+1, size = list.size(); j < size; j++) {
				suc = list.get(j);
				if (object.getTimeStamp() < suc.getTimeStamp())
					break;
				if (suc.equals(object)) {
					list.remove(j);
					return;
				}
			}
			PNObject pre;
			for (int j = index-1; j >= 0; j--) {
				pre = list.get(j);
				if (object.getTimeStamp() > pre.getTimeStamp()) 
					break;
				if (pre.equals(object)) {
					list.remove(j);
					return;
				}
			}
    	}
	}
	
	// SweepArea manages only positive elements
	protected PNTemporalPipeSA<I> sweepArea;
	protected DynamicHeap<TemporalObject<I>> heap; // required for order
	protected boolean checkedHeapObjectSize;
	protected int heapObjectSize;
	 
	public PNTemporalPipe(PNTemporalPipeSA<I> sweepArea) {
		this.sweepArea = sweepArea;
		this.heap = new DynamicHeap<TemporalObject<I>>(TemporalObject.START_TIMESTAMP_COMPARATOR);
		this.checkedHeapObjectSize = false;
		this.heapObjectSize = 0;
	}
	
	public PNTemporalPipe(Source<? extends PNObject<I>> source, int sourceID, PNTemporalPipeSA<I> sweepArea) {
		this(sweepArea);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
	
	public PNTemporalPipe(Source<? extends PNObject<I>> source) {
		this(source, DEFAULT_ID, new PNTemporalPipeSA<I>());
	}
	
	
	@Override
	public void processObject(PNObject<I> pno , int sourceID){		
		if (pno.isPositive()) {
			sweepArea.insert(pno);
			return;
		}
		// pno is negative
		Iterator<? extends PNObject<I>> qualifies = sweepArea.query(pno, 0);
		if (qualifies.hasNext()) {
			PNObject<I> next = qualifies.next();
			sweepArea.remove(next);
			if (next.getTimeStamp() != pno.getTimeStamp()) // no coalesce
				heap.enqueue(new TemporalObject<I>(next.getObject(), new TimeInterval(next.getTimeStamp(), pno.getTimeStamp())));
		}		
		transferHeap(sweepArea.getOldestTimeStamp());
	}
	
	protected boolean transferHeap(long timeStamp) {
		boolean ret = false;
		while (!heap.isEmpty() && heap.peek().getStart() <= timeStamp) {
			ret = true;
			transfer(heap.dequeue()); 
		}
		return ret;
	}
	
	@Override
	public void close() {
		super.close();
		if (isClosed) {
			sweepArea.close();
			heap.close();
		}
	}
	
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				while(!heap.isEmpty())
					transfer(heap.dequeue());
				if (sweepArea.size() > 0) 
					throw new IllegalStateException("SweepArea still contains positive elements.");
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	public int getCurrentMemUsage() {
		if (!checkedHeapObjectSize && !heap.isEmpty()) {
			try {	
				heapObjectSize = XXLSystem.getObjectSize(heap.peek());
			} catch (IllegalAccessException e) {
				heapObjectSize = MemoryMonitorable.SIZE_UNKNOWN;
			}
			checkedHeapObjectSize = true;
		}
		return sweepArea.getCurrentMemUsage() + heap.size()*heapObjectSize;
	}
	
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {
		return NO_HEARTBEAT;
	}
	
	
	public static void main(String[] args) {
		
	
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		final int noOfElements = 5000;
		final int maxValue = 7;
		final int startInc = 10;
		final int intervalSize = 75;
		final long seed = 17;
	
	
		RandomNumber<Integer> r = new RandomNumber<Integer>(
			new JavaDiscreteRandomWrapper(new Random(seed), maxValue),
			noOfElements,
			0
		);		
		Source<TemporalObject<Integer>> s = Pipes.decorateWithRandomTimeIntervals(r, startInc, intervalSize, seed);
	
		TemporalPNPipe<Integer> int2PN = new TemporalPNPipe<Integer>(s);
		PNTemporalPipe<Integer> PN2Int = new PNTemporalPipe<Integer>(int2PN);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(PN2Int));
		exec.registerQuery(Pipes.compareSnapshots(s, PN2Int, false, false));
		exec.startAllQueries();
	}
	
}
