/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.mappers;

import xxl.core.collections.queues.DynamicHeap;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTimeStampPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.sources.Source;
import xxl.core.util.XXLSystem;

/**
 * Converts a stream of PNObjects to a stream of TemporalObjects.
 * 
 * @param <I> 
 */
public class TemporalPNPipe<I> extends AbstractTimeStampPipe<I,I, TemporalObject<I>, PNObject<I>> implements MemoryMonitorable {
	
	protected DynamicHeap<PNObject<I>> heap; // required for order; keeps negative elements
	protected boolean checkedHeapObjectSize;
	protected int heapObjectSize;
	 
	public TemporalPNPipe() {
		this.heap = new DynamicHeap<PNObject<I>>(PNObject.TIMESTAMP_COMPARATOR);
		this.checkedHeapObjectSize = false;
		this.heapObjectSize = 0;
	}
		
	public TemporalPNPipe(Source<? extends TemporalObject<I>> source, int sourceID) {
		this();
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
	
	public TemporalPNPipe(Source<? extends TemporalObject<I>> source) {
		this(source, DEFAULT_ID); 
	}
	
	@Override
	public void processObject(TemporalObject<I> o , int sourceID){		
		transferHeap(o.getStart());
		transfer(new PNObject<I>(o.getObject(), o.getStart(), true));
		heap.enqueue(new PNObject<I>(o.getObject(), o.getEnd(), false));
	}
	
	protected boolean transferHeap(long timeStamp) {
		boolean ret = false;
		while (!heap.isEmpty() && heap.peek().getTimeStamp() <= timeStamp) {
			ret = true;
			transfer(heap.dequeue()); 
		}
		return ret;
	}
	
	@Override
	public void close() {
		super.close();
		if (isClosed) {
			heap.close();
		}
	}
	
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				while(!heap.isEmpty())
					transfer(heap.dequeue());
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	public int getCurrentMemUsage() {
		if (!checkedHeapObjectSize && !heap.isEmpty()) {
			try {	
				heapObjectSize = XXLSystem.getObjectSize(heap.peek());
			} catch (IllegalAccessException e) {
				heapObjectSize = MemoryMonitorable.SIZE_UNKNOWN;
			}
			checkedHeapObjectSize = true;
		}
		return heap.size()*heapObjectSize;
	}
	
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {
		return NO_HEARTBEAT;
	}
	
}

