/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.unions;

import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.queues.Heap;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.metaData.DefaultMetaDataHandler;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.joins.TemporalJoin.TemporalJoinMetaDataManagement;
import xxl.core.pipes.operators.mappers.PNTemporalPipe;
import xxl.core.pipes.operators.mappers.TemporalPNPipe;
import xxl.core.pipes.operators.windows.PNTemporalWindow;
import xxl.core.pipes.operators.windows.TemporalWindow;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.util.XXLSystem;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

/**
 * 
 * When merging multiple sources of PNObjects it is important to prefer positive objects to negative ones
 * in the case that both have the same timestamp. See PNObject.TIMESTAMP_COMPARATOR.
 * 
 * @param <I>
 */
public class PNUnion<I> extends AbstractPNPipe<I,I> implements MemoryMonitorable {

	public class PNUnionMetaDataManagement extends AbstractTimeStampPipeMetaDataManagement {
		
		public static final String HEAP_SIZE = "HEAP_SIZE";
		public static final String HEAP_OBJECT_SIZE = "HEAP_OBJECT_SIZE";
		public static final String HEAP_MEM_SIZE = "HEAP_MEM_SIZE";
				
			
		public PNUnionMetaDataManagement() {
			super();
		}
		
		@Override
		protected boolean addMetaData(Object metaDataIdentifier) {
			
			if (metaDataIdentifier.equals(HEAP_SIZE)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Integer>(this, metaDataIdentifier, Integer.class,
					new Function<Object,Integer>() {			
						@Override
						public Integer invoke() {
							if (isClosed || isDone)
								return 0;
							synchronized(metaDataManagement) {
								try {
									return heap.size();
								}
								catch (NullPointerException npe) {
									return 0;
								}
							}
						}
					})
				);
				return true;
			}
			if (metaDataIdentifier.equals(HEAP_OBJECT_SIZE)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Integer>(this, metaDataIdentifier, Integer.class,
					new Function<Object,Integer>() {			
						@Override
						public Integer invoke() {
							synchronized(metaDataManagement) {
								return heapObjectSize;
							}
						}
					})
				);
				return true;
			}
			if (metaDataIdentifier.equals(HEAP_MEM_SIZE)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Integer>(this, metaDataIdentifier, Integer.class,
					new Function<Object,Integer>() {			
						@Override
						public Integer invoke() {
							if (isClosed || isDone)
								return 0;							
							synchronized(metaDataManagement) {
								try {
									return heap.size() * heapObjectSize;
								}
								catch (NullPointerException npe) {
									return 0;
								}
							}
						}
					})
				);
				return true;
			}
			if (super.removeMetaData(metaDataIdentifier))
				return true;
			return false;
		}
		
		@Override
		protected boolean removeMetaData(Object metaDataIdentifier) {
			if (metaDataIdentifier.equals(HEAP_SIZE) ||
				metaDataIdentifier.equals(HEAP_OBJECT_SIZE) ||
				metaDataIdentifier.equals(HEAP_MEM_SIZE)) {
					metaData.remove(metaDataIdentifier);
					return true;
			}
			if (super.removeMetaData(metaDataIdentifier))
				return true;
			return false;
		}
		
	}
	
	protected Heap<PNObject<I>> heap;
	protected boolean checkedHeapObjectSize;
	protected int heapObjectSize;
	
	public PNUnion() {
		super();
		this.heap = new DynamicHeap<PNObject<I>>(PNObject.TIMESTAMP_COMPARATOR);
		this.checkedHeapObjectSize = false;
		this.heapObjectSize = MemoryMonitorable.SIZE_UNKNOWN;
	}
	
	public PNUnion(Source<? extends PNObject<I>>[] sources, int[] sourceIDs) {
		this();
		for (int i = 0; i < sources.length; i++)
			if (!Pipes.connect(sources[i], this, sourceIDs[i]))
				throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");
	}
	
	public PNUnion(Source<? extends PNObject<I>>[] sources) {
		this();
		for (int i = 0; i < sources.length; i++)
			if (!Pipes.connect(sources[i], this, i))
				throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");
	}
	
	@SuppressWarnings("unchecked")
	public PNUnion(Source<? extends PNObject<I>> source1, Source<? extends PNObject<I>> source2) {
		this(new Source[]{source1, source2});
	}
	
	@Override
	public void processObject(PNObject<I> o, int sourceID) throws IllegalArgumentException {
		heap.enqueue(o);
		while(!heap.isEmpty() && heap.peek().getTimeStamp() <= minTimeStamp) 
			transfer(heap.dequeue());
	}
	
	@Override
	public void close() {
		super.close();
		if (isClosed()) 
			heap.close();
	}
	
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				while(!heap.isEmpty())
					transfer(heap.dequeue());
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	public int getCurrentMemUsage() {
		if (!checkedHeapObjectSize && !heap.isEmpty()) {
			try {	
				heapObjectSize = XXLSystem.getObjectSize(heap.peek());
			} catch (IllegalAccessException e) {
				heapObjectSize = MemoryMonitorable.SIZE_UNKNOWN;
			}
			checkedHeapObjectSize = true;
		}
		if (heapObjectSize == SIZE_UNKNOWN)
			return SIZE_UNKNOWN;
		return heap.size() * heapObjectSize;
	}
	
	@Override
	public void createMetaDataManagement() {
		if (metaDataManagement != null)
			throw new IllegalStateException("An instance of MetaDataManagement already exists.");
		metaDataManagement = new PNUnionMetaDataManagement();
	}
	
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 5000;
		final int maxValue = 20;
		final int maxTSInc = 50;
		
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(maxTSInc), noOfElements)
		);
		TimeStampCursorSource<Integer> s2 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(maxTSInc), noOfElements)
		);
		final long windowSize1 = 100;
		final long windowSize2 = 50;
		
		// time interval approach
		TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s1, windowSize1);
		TemporalWindow<Integer> w2 = new TemporalWindow<Integer>(s2, windowSize2);
		TemporalUnion<Integer> res1 = new TemporalUnion<Integer>(w1, w2);
		
		// positive-negative approach
		TemporalPNPipe<Integer> pnSource1 = new TemporalPNPipe<Integer>(s1);
		TemporalPNPipe<Integer> pnSource2 = new TemporalPNPipe<Integer>(s2);
		PNTemporalWindow<Integer> pnw1 = new PNTemporalWindow<Integer>(pnSource1, windowSize1);
		PNTemporalWindow<Integer> pnw2 = new PNTemporalWindow<Integer>(pnSource2, windowSize2);
		PNUnion<Integer> pnUnion = new PNUnion<Integer>(pnw1, pnw2);
		PNTemporalPipe<Integer> res2 = new PNTemporalPipe<Integer>(pnUnion);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res1));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res2));
		exec.registerQuery(Pipes.compareSnapshots(res1, res2, false, false));
		exec.startAllQueries();
	}

}
