/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.windows;

import java.util.Iterator;

import xxl.core.collections.sweepAreas.ImplementorBasedSweepArea;
import xxl.core.collections.sweepAreas.ListSAImplementor;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.Mapper;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sources.Enumerator;
import xxl.core.pipes.sources.Source;
import xxl.core.predicates.Predicate;
import xxl.core.predicates.Predicates;

/**
 * Implementation of a count-based window. The end timestamp of an incoming element is
 * set to the start timestamp of the 'windowSize'-th future incoming element. 
 * This implementation assumes the start timestamps of incoming elements to be unique.
 * @param <E> 
 */
public class CountBasedWindow<E> extends AbstractTemporalPipe<E, E> implements MemoryMonitorable {

	protected CountBasedWindowSA<E> sweepArea;
	
	public static class CountBasedWindowSA<E> extends ImplementorBasedSweepArea<TemporalObject> {
			
		public CountBasedWindowSA(final int N, final ListSAImplementor<TemporalObject> listSAImpl, int objectSize) {
			super(listSAImpl, DEFAULT_ID, true, Predicates.FALSE,
				new Predicate<TemporalObject>() {
					@Override
					public boolean invoke(TemporalObject o1, TemporalObject o2) {
						return listSAImpl.size() == N; // checks if implementor is filled
					}
				},
				1, objectSize
			);
			if (N <= 0)
				throw new IllegalArgumentException("Window size <= 0 is not allowed.");
		}
		
		public CountBasedWindowSA(final int N, final ListSAImplementor<TemporalObject> listSAImpl) {
			this(N, listSAImpl, SIZE_UNKNOWN);
		}
		
		public CountBasedWindowSA(final int N, int objectSize) {
			this(N, new ListSAImplementor<TemporalObject>(), objectSize);
		}
		
		public CountBasedWindowSA(final int N) {
			this(N, new ListSAImplementor<TemporalObject>(), SIZE_UNKNOWN);
		}
		
	}
	
	public CountBasedWindow(CountBasedWindowSA<E> sweepArea) {
		this.sweepArea = sweepArea;
	}
	
	public CountBasedWindow(Source<? extends TemporalObject<E>> source, int sourceID, CountBasedWindowSA<E> sweepArea) {
		this(sweepArea);	
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
	
	public CountBasedWindow(Source<? extends TemporalObject<E>> source, final int N) {
		this(source, DEFAULT_ID, new CountBasedWindowSA<E>(N));
	}
	
	// update stream -> N == 1
	public CountBasedWindow(Source<? extends TemporalObject<E>> source) {
		this(source, 1);
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#processObject(java.lang.Object, int)
	 */
	@Override
	public void processObject(TemporalObject<E> tso, int sourceID) {
		Iterator<TemporalObject> results = sweepArea.expire(tso, sourceID);
		if (results.hasNext()) {
			TemporalObject<E> res = results.next();
			transfer(new TemporalObject<E>(res.getObject(), new TimeInterval(res.getStart(), tso.getStart())));
		}
		sweepArea.insert(tso);
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractTimeStampPipe#close()
	 */
	@Override
	public void close() {
		super.close();
		if (isClosed)
			sweepArea.close();
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#done(int)
	 */
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
			
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				Iterator<TemporalObject> results = sweepArea.iterator();
				while (results.hasNext()) {
					TemporalObject<E> res = results.next();
					transfer(new TemporalObject<E>(res.getObject(), new TimeInterval(res.getStart(), TimeInterval.INFINITY)));
				}
				signalDone();
			}
	    }
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.memoryManager.MemoryMonitorable#getCurrentMemUsage()
	 */
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {
		return NO_HEARTBEAT;
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		Enumerator e = new Enumerator(500, 0);
		Mapper<Integer, TemporalObject<Integer>> m = new Mapper<Integer, TemporalObject<Integer>>(
			e, 
			new Function<Integer, TemporalObject<Integer>>() {
				protected long count = 0;

				@Override
				public TemporalObject<Integer> invoke(Integer o) {
					return new TemporalObject<Integer>(o,
							new TimeInterval(count += 5, count + 10));
				}
			}
		);

		int windowSize = 3;
		CountBasedWindow<Integer> w = new CountBasedWindow<Integer>(m, windowSize); // update semantics

		Pipes.verifyStartTimeStampOrdering(w);
		Printer printer = new Printer<TemporalObject<Integer>>(w);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(printer);
		exec.startQuery(printer);
	}

}
