/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.windows;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import xxl.core.collections.sweepAreas.HashSAImplementor;
import xxl.core.collections.sweepAreas.ListSA;
import xxl.core.collections.sweepAreas.ListSAImplementor;
import xxl.core.collections.sweepAreas.SweepAreaImplementor;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.identities.Coalesce;
import xxl.core.pipes.operators.identities.Coalesce.CoalesceSA;
import xxl.core.pipes.operators.mappers.TemporalMapper;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.AbstractTemporalSink;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;
import xxl.core.predicates.And;
import xxl.core.util.random.JavaDiscreteRandomWrapper;


/**
 * 
 */
public class DStream<E> extends AbstractTemporalPipe<E,E> implements MemoryMonitorable {
	   	
	 			
	public static class CoalesceSensitiveSA<E> extends ListSA<TemporalObject<E>> {
    	
    	public CoalesceSensitiveSA(SweepAreaImplementor<TemporalObject<E>> impl, int objectSize, List<TemporalObject<E>> list) {
    		super(impl, DEFAULT_ID, true, ReorganizeModes.LAZY_REORGANIZE, new And<TemporalObject<E>>(TemporalObject.VALUE_EQUIVALENCE_PREDICATE, Coalesce.CoalesceSA.ADJACENT_INTERVALS), TemporalObject.INTERVAL_OVERLAP_REORGANIZE_NOT_ADJACENT, 1, list, objectSize);
    	}
    	
    	public CoalesceSensitiveSA(Function<? super TemporalObject<E>,Integer> hashFunction) {
    		super(new HashSAImplementor<TemporalObject<E>>(
        			hashFunction,
        			new Function<Object,List<TemporalObject<E>>>() {
        				@Override
						public List<TemporalObject<E>> invoke() {
        					return new LinkedList<TemporalObject<E>>();
        				}
        			}, 1),
        			DEFAULT_ID, true, ReorganizeModes.LAZY_REORGANIZE,
        			new And<TemporalObject<E>>(TemporalObject.VALUE_EQUIVALENCE_PREDICATE, Coalesce.CoalesceSA.ADJACENT_INTERVALS), 
        			TemporalObject.INTERVAL_OVERLAP_REORGANIZE_NOT_ADJACENT, 
        			1, new ArrayList<TemporalObject<E>>(), SIZE_UNKNOWN
        	);
    	}
    	
    	public CoalesceSensitiveSA() {
    		this(new Function<TemporalObject<E>,Integer>() {
    				@Override
					public Integer invoke(TemporalObject<E> o) {
    					return o.getObject().hashCode();
    				}
    			}
    		);
    	}
    	
    	@Override
		public void insert(TemporalObject<E> o) throws IllegalArgumentException {
    		if (checkObjectSize)
    			computeObjectSize(o);
    		int index = Collections.binarySearch(list, o, TemporalObject.END_TIMESTAMP_COMPARATOR);
    		if (index < 0)
    			list.add(-index-1, o);
    		else
    			list.add(index, o);
    		impl.insert(o);
    	}
    	
    	public void remove(TemporalObject<E> object) {
    		if(!impl.remove(object)) // remove from implementor
    			throw new IllegalStateException("Could not remove element from implementor");
    		
    		// remove from list
			int index = Collections.binarySearch(list, object, TemporalObject.END_TIMESTAMP_COMPARATOR);
			TemporalObject cur = list.get(index);
			if (cur.equals(object)) {
				list.remove(index);
				return;
			}
			TemporalObject<E> suc;
			for (int j = index+1, size = list.size(); j < size; j++) {
				suc = list.get(j);
				if (object.getEnd() < suc.getEnd())
					break;
				if (suc.equals(object)) {
					list.remove(j);
					return;
				}
			}
			TemporalObject<E> pre;
			for (int j = index-1; j >= 0; j--) {
				pre = list.get(j);
				if (object.getEnd() > pre.getEnd()) 
					break;
				if (pre.equals(object)) {
					list.remove(j);
					return;
				}
			}
    	}
    	 
    }
	
    protected CoalesceSensitiveSA<E> sweepArea;
 
    public DStream(CoalesceSensitiveSA<E> sweepArea) {		
		this.sweepArea = sweepArea;		
	}
    
   	public DStream(Source<? extends TemporalObject<E>> source, int sourceID, CoalesceSensitiveSA<E> sweepArea) {		
		this(sweepArea);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
   	
   	public DStream(Source<? extends TemporalObject<E>> source) {		
		this(source, DEFAULT_ID, new CoalesceSensitiveSA<E>());
	
	}

	@Override
	public void processObject(TemporalObject<E> o, int sourceID) throws IllegalArgumentException {
		Iterator<? extends TemporalObject<E>> results = sweepArea.expire(o, sourceID);
		while (results.hasNext()) {
			TemporalObject<E> next = results.next();
			transfer(new TemporalObject<E>(next.getObject(), new TimeInterval(next.getEnd()-1, next.getEnd())));
		}
		
		Iterator<? extends TemporalObject<E>> qualifies = sweepArea.query(o, sourceID);
		if (qualifies.hasNext()) {
			sweepArea.remove(qualifies.next());
		}
		sweepArea.insert(o);
	}

	/**
	 * Calls <CODE>super.close()</CODE> and closes
	 * the sweep area.
	 */
	@Override
	public void close() {
		super.close();
		if (isClosed)
			sweepArea.close();
	}
	
	
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				Iterator<? extends TemporalObject<E>> results = sweepArea.iterator();
				while(results.hasNext()) {
					TemporalObject<E> next = results.next();
					transfer(new TemporalObject<E>(next.getObject(), new TimeInterval(next.getEnd()-1, next.getEnd())));
				}
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.memoryManager.MemoryMonitorable#getCurrentMemUsage()
	 */
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractTimeStampPipe#processHeartbeat(long, int)
	 */
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {
		boolean transferred = false;
		Iterator<? extends TemporalObject<E>> results = sweepArea.expire(new TemporalObject<E>(null, new TimeInterval(minTS, TimeInterval.INFINITY)), 0);
		while(results.hasNext()) {
			TemporalObject<E> next = results.next();
			transfer(new TemporalObject<E>(next.getObject(), new TimeInterval(next.getEnd()-1, next.getEnd())));
			transferred = true;
		}
		if (transferred) return NO_HEARTBEAT;
		return minTS;
	}
	
	
	/**
	 * The main method contains some examples to demonstrate the usage
	 * and the functionality of this class.
	 *
	 * @param args array of <tt>String</tt> arguments. It can be used to
	 * 		submit parameters when the main method is called.
	 */
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		final int noOfElements = 7000;
		final int maxValue = 7;
		final int startInc = 10;
		final int intervalSize = 50;
		final long seed = 17;
	
	
		RandomNumber<Integer> r = new RandomNumber<Integer>(
			new JavaDiscreteRandomWrapper(new Random(seed), maxValue),
			noOfElements,
			0
		);		
		Source<TemporalObject<Integer>> s = Pipes.decorateWithRandomTimeIntervals(r, startInc, intervalSize, seed);
	
		// direct implementation
		DStream<Integer> ds = new DStream<Integer>(s);
		
		// indirect implementation
		Coalesce<Integer> c = new Coalesce<Integer>(s, 0, new CoalesceSA<Integer>(new ListSAImplementor<TemporalObject<Integer>>(), Long.MAX_VALUE));
		TemporalMapper<Integer,Integer> m = new TemporalMapper<Integer,Integer>(c, new Function<TemporalObject<Integer>,TemporalObject<Integer>>() {
			@Override
			public TemporalObject<Integer> invoke(TemporalObject<Integer> o) {
				return new TemporalObject<Integer>(o.getObject(), new TimeInterval(o.getEnd()-1, o.getEnd()));
			}
		});
		
		Pipes.verifyStartTimeStampOrdering(ds);
		
		AbstractTemporalSink<Integer> sink = new AbstractTemporalSink<Integer>(new Source[]{ds, m}, new int[]{1, 2}) {
			LinkedList<TemporalObject<Integer>> list1 = new LinkedList<TemporalObject<Integer>>();
			LinkedList<TemporalObject<Integer>> list2 = new LinkedList<TemporalObject<Integer>>();
			
			@Override
			public void processObject(TemporalObject<Integer> o, int sourceID) {				
				if (sourceID == 1) {
					list1.add(o);
					return;
				}
				if (sourceID == 2) {
					list2.add(o);
					return;
				}
			}
			
			@Override
			public void done(int sourceID) {
				super.done(sourceID);
				if (isDone()) {
					processingWLock.lock();
					try {
						Collections.sort(list2, TemporalObject.START_TIMESTAMP_COMPARATOR);
						System.out.println("list1: "+list1);
						System.out.println("list2: "+list2);
						System.out.println("size of list1: "+list1.size());
						System.out.println("size of list2: "+list2.size());
						Iterator it1 = list1.iterator();
						while (it1.hasNext()) {
							Object next1 = it1.next();
							if (!list2.contains(next1)) 
								System.err.println("ERROR: Missing results in indirect implementation: "+next1);
							else 
								list2.remove(next1);
						}
						Iterator it2 = list2.iterator();
						while (it2.hasNext()) 
							System.err.println("ERROR: Missing results in direct implementation: "+it2.next());
						System.out.println("CHECKS FINISHED.");
					}
					finally {
						processingWLock.unlock();
					}
				}
			}
		};
				
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(sink);
		exec.startAllQueries();
	}
}


