/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.windows;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;

import xxl.core.collections.sweepAreas.ListSAImplementor;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.identities.Coalesce;
import xxl.core.pipes.operators.identities.Coalesce.CoalesceSA;
import xxl.core.pipes.operators.mappers.TemporalMapper;
import xxl.core.pipes.operators.windows.DStream.CoalesceSensitiveSA;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.AbstractTemporalSink;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;
import xxl.core.util.random.JavaDiscreteRandomWrapper;


/**
 * 
 */
public class IStream<E> extends AbstractTemporalPipe<E,E> implements MemoryMonitorable {
	   	
	protected CoalesceSensitiveSA<E> sweepArea;
 
    public IStream(CoalesceSensitiveSA<E> sweepArea) {		
		this.sweepArea = sweepArea;		
	}
    
   	public IStream(Source<? extends TemporalObject<E>> source, int sourceID, CoalesceSensitiveSA<E> sweepArea) {		
		this(sweepArea);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
   	
   	public IStream(Source<? extends TemporalObject<E>> source) {		
		this(source, DEFAULT_ID, new CoalesceSensitiveSA<E>());
	
	}

	@Override
	public void processObject(TemporalObject<E> o, int sourceID) throws IllegalArgumentException {		
		sweepArea.reorganize(o, 0);
		Iterator<? extends TemporalObject<E>> qualifies = sweepArea.query(o, 0);
		if (!qualifies.hasNext()) 
			transfer(new TemporalObject<E>(o.getObject(), new TimeInterval(o.getStart(), o.getStart()+1)));
		else {
			sweepArea.remove(qualifies.next());
		}
		
		sweepArea.insert(o);
	}

	/**
	 * Calls <CODE>super.close()</CODE> and closes
	 * the sweep area.
	 */
	@Override
	public void close() {
		super.close();
		if (isClosed)
			sweepArea.close();
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.memoryManager.MemoryMonitorable#getCurrentMemUsage()
	 */
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractTimeStampPipe#processHeartbeat(long, int)
	 */
	@Override
	protected long processHeartbeat(long minTS, int sourceID) {
		sweepArea.reorganize(new TemporalObject<E>(null, new TimeInterval(minTS, TimeInterval.INFINITY)), 0);
		return minTS;
	}
	
	
	/**
	 * The main method contains some examples to demonstrate the usage
	 * and the functionality of this class.
	 *
	 * @param args array of <tt>String</tt> arguments. It can be used to
	 * 		submit parameters when the main method is called.
	 */
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
	
		final int noOfElements = 5000;
		final int maxValue = 5;
		final int startInc = 10;
		final int intervalSize = 70;
		final long seed = 17;
	
	
		RandomNumber<Integer> r = new RandomNumber<Integer>(
			new JavaDiscreteRandomWrapper(new Random(seed), maxValue),
			noOfElements,
			0
		);		
		Source<TemporalObject<Integer>> s = Pipes.decorateWithRandomTimeIntervals(r, startInc, intervalSize, seed);
		
		// direct implementation
		IStream<Integer> is = new IStream<Integer>(s);
		
		// indirect implementation
		Coalesce<Integer> c = new Coalesce<Integer>(s, 0, new CoalesceSA<Integer>(new ListSAImplementor<TemporalObject<Integer>>(), Long.MAX_VALUE));
		TemporalMapper<Integer,Integer> m = new TemporalMapper<Integer,Integer>(c, new Function<TemporalObject<Integer>,TemporalObject<Integer>>() {
			@Override
			public TemporalObject<Integer> invoke(TemporalObject<Integer> o) {
				return new TemporalObject<Integer>(o.getObject(), new TimeInterval(o.getStart(), o.getStart()+1));
			}
		});
		
		Pipes.verifyStartTimeStampOrdering(is);
		Pipes.verifyStartTimeStampOrdering(m);
		
		AbstractTemporalSink<Integer> sink = new AbstractTemporalSink<Integer>(new Source[]{is, m}, new int[]{1, 2}) {
			LinkedList<TemporalObject<Integer>> list1 = new LinkedList<TemporalObject<Integer>>();
			LinkedList<TemporalObject<Integer>> list2 = new LinkedList<TemporalObject<Integer>>();
			
			@Override
			public void processObject(TemporalObject<Integer> o, int sourceID) {				
				if (sourceID == 1) {
					list1.add(o);
					return;
				}
				if (sourceID == 2) {
					list2.add(o);
					return;
				}
			}
			
			@Override
			public void done(int sourceID) {
				super.done(sourceID);
				if (isDone()) {
					processingWLock.lock();
					try {
						System.out.println("list1: "+list1);
						System.out.println("list2: "+list2);
						System.out.println("size of list1: "+list1.size());
						System.out.println("size of list2: "+list2.size());
						Iterator it1 = list1.iterator();
						while (it1.hasNext()) {
							Object next1 = it1.next();
							if (!list2.contains(next1)) 
								System.err.println("ERROR: Missing results in indirect implementation: "+next1);
							else 
								list2.remove(next1);
						}
						Iterator it2 = list2.iterator();
						while (it2.hasNext()) 
							System.err.println("ERROR: Missing results in direct implementation: "+it2.next());
						System.out.println("CHECKS FINISHED.");
					}
					finally {
						processingWLock.unlock();
					}
				}
			}
		};
			
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(sink);
		exec.startAllQueries();
	}
}

