/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.windows;

import java.util.Iterator;

import xxl.core.collections.sweepAreas.ListSAImplementor;
import xxl.core.collections.sweepAreas.PNSweepArea;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.PNTemporalPipe;
import xxl.core.pipes.operators.mappers.TemporalPNPipe;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.predicates.Predicate;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

public class PNCountBasedWindow<I> extends AbstractPNPipe<I,I> implements MemoryMonitorable {

	protected int N;
	protected PNSweepArea<I> sweepArea;
	
	public PNCountBasedWindow(final int N) {
		super();
		this.N = N;
		this.sweepArea = new PNSweepArea<I>(new ListSAImplementor<PNObject<I>>(),0, 
			new Predicate<PNObject>() {
				@Override
				public boolean invoke(PNObject o1, PNObject o2) {
					return sweepArea.size() == N; // checks if implementor is filled
				}
			}, 1);
	}
	
	public PNCountBasedWindow(Source<? extends PNObject<I>> source, int sourceID, int N) {
		this(N);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes.");		
	}
	
	public PNCountBasedWindow(Source<? extends PNObject<I>> source, int N) {
		this(source, DEFAULT_ID, N);
	}
	
	@Override
	public void processObject(PNObject<I> pno, int sourceID) throws IllegalArgumentException {	
		if (pno.isPositive()) {
			Iterator<PNObject<I>> it = sweepArea.query(pno, sourceID);
			if (it.hasNext()) {
				PNObject<I> next = it.next();
				transfer(new PNObject<I>(next.getObject(), pno.getTimeStamp(), false));
				it.remove();
			}			
			sweepArea.insert(pno); 
			transfer(pno);
		}
	}
	
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				Iterator<PNObject<I>> rest = sweepArea.iterator();
				while (rest.hasNext()) {
					transfer(new PNObject<I>(rest.next().getObject(), Long.MAX_VALUE, false));
				}
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}
	
	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 500;
		final int maxValue = 25;
		final int maxTSInc = 75;
		int seed = 12;
		
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxValue), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(seed, maxTSInc), noOfElements)
		);
		
		final int N = 5;
		
		// time interval approach
		CountBasedWindow<Integer> w1 = new CountBasedWindow<Integer>(s1, N);
		
		// positive-negative approach
		TemporalPNPipe<Integer> pnSource = new TemporalPNPipe<Integer>(s1);
		PNCountBasedWindow<Integer> pnWindow = new PNCountBasedWindow<Integer>(pnSource, N);
		PNTemporalPipe<Integer> res2 = new PNTemporalPipe<Integer>(pnWindow);
		
		QueryExecutor exec = new QueryExecutor();
//		exec.registerQuery(new Printer(w1));
//		exec.registerQuery(new Printer(pnWindow));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(w1));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(res2));
		exec.registerQuery(Pipes.compareSnapshots(w1, res2, false, false));
		exec.startAllQueries();

	}
	
}
