/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
						Head of the Database Research Group
						Department of Mathematics and Computer Science
						University of Marburg
						Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

	http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

	xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

/**
 * 
 */
package xxl.core.pipes.operators.windows;

import java.util.Iterator;
import java.util.LinkedList;

import xxl.core.collections.sweepAreas.ImplementorBasedSweepArea;
import xxl.core.collections.sweepAreas.ListSAImplementor;
import xxl.core.cursors.filters.Filter;
import xxl.core.cursors.filters.Remover;
import xxl.core.cursors.filters.WhileTaker;
import xxl.core.cursors.sources.DiscreteRandomNumber;
import xxl.core.cursors.sources.EmptyCursor;
import xxl.core.pipes.elements.PNObject;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractPNPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.mappers.PNTemporalPipe;
import xxl.core.pipes.operators.mappers.TemporalPNPipe;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sources.Source;
import xxl.core.pipes.sources.TimeStampCursorSource;
import xxl.core.predicates.Predicate;
import xxl.core.predicates.Predicates;
import xxl.core.util.random.JavaDiscreteRandomWrapper;

/**
 * Defines the validity of objects. Requires a stream of positive PNObjects and the window size as input. 
 * The case of windowed subqueries is not yet handled.
 * <br />
 * Operator belongs to the Positive-Negative Tuple Approach.  See the following papers for details.
 * <br />
 * Moustafa A. Hammad, Thanaa M. Ghanem, Walid G. Aref, Ahmed K. Elmagarmid, Mohamed F. Mokbel:
 * Efficient Pipelined Execution of Sliding Window Queries over Data Streams, 
 * Purdue University Department of Computer Sciences, CSD TR#03-035, 2003.
 * <br />
 * A. Arasu, S. Babu, and J. Widom. The CQL Continuous Query Language: Semantic Foundations and Query Execution,
 * International Journal on Very Large Data Bases (VLDB Journal)
 */
public class PNTemporalWindow<I> extends AbstractPNPipe<I,I> implements MemoryMonitorable {
		
	public static class PNTemporalWindowSA<I> extends ImplementorBasedSweepArea<PNObject<I>> {
    	
	   	public PNTemporalWindowSA(int objectSize) {
    		super(new ListSAImplementor<PNObject<I>>(new LinkedList<PNObject<I>>(), PNObject.VALUE_EQUIVALENCE_PREDICATE),
    			DEFAULT_ID, true, Predicates.TRUE, PNObject.TEMPORAL_BEFORE_REORGANIZE, 1, objectSize  
    		);
    	}
	   	
		public PNTemporalWindowSA() {
    		this(SIZE_UNKNOWN);
    	}
		
		@Override
		public Iterator<PNObject<I>> expire (PNObject<I> currentStatus, int ID) throws UnsupportedOperationException, IllegalStateException {
			if (ID == this.ID && !selfReorganize)
				return new EmptyCursor<PNObject<I>>();
			return new Remover<PNObject<I>>(
				new WhileTaker<PNObject<I>>(iterator(), 
				removeRightBinds[ID].setRight(currentStatus))
			);
		}
		
	}
	
	
	protected long windowSize;
	protected PNTemporalWindowSA<I> sweepArea;
	
	public PNTemporalWindow(Source<? extends PNObject<I>> source, long windowSize, int objectSize){
		super(source);
		this.windowSize = windowSize;
		sweepArea = new PNTemporalWindowSA<I>(objectSize);
	}
	
	public PNTemporalWindow(Source<? extends PNObject<I>> source, long windowSize){
		this(source, windowSize, SIZE_UNKNOWN);
	}
	
	@Override
	public void processObject(PNObject<I> o, int sourceID)throws IllegalArgumentException {
		if (o.isPositive()) {					
			sweepArea.insert(o.newInverseAndInc(windowSize+1)); // element expires at timestamp + windowSize + 1
					
			Iterator<PNObject<I>> negativeElements = sweepArea.expire(o, sourceID);
			while (negativeElements.hasNext()) {
				transfer(negativeElements.next());
			}
			transfer(o);
		}
	}

	@Override
	public void close() {
		super.close();
		if (isClosed) {
			sweepArea.clear();
			sweepArea.close();
		}
	}
	
	/* (non-Javadoc)
	 * @see xxl.core.pipes.operators.AbstractPipe#done(int)
	 */
	@Override
	public void done(int sourceID) {
		if (isClosed || isDone) return;
		
		processingWLock.lock();
		updateDoneStatus(sourceID);
	    try {
			if (isDone()) {
				Iterator<PNObject<I>> negativeElements = sweepArea.iterator();
				while (negativeElements.hasNext()) {
					transfer(negativeElements.next());
				}
				sweepArea.clear();
				signalDone();
			}
		}
	    finally {
	    	processingWLock.unlock();
	    }
	}
	
	public int getCurrentMemUsage() {
		return sweepArea.getCurrentMemUsage();
	}

	public static void main(String[] args) {
		
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 1000;
		
		TimeStampCursorSource<Integer> s = new TimeStampCursorSource<Integer>(
				0,
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(), noOfElements),
				new DiscreteRandomNumber(new JavaDiscreteRandomWrapper(), noOfElements)
		);
		
		final long windowSize = 1000;
		TemporalWindow<Integer> w1 = new TemporalWindow<Integer>(s, windowSize);
		
		TemporalPNPipe<Integer> int2PN = new TemporalPNPipe<Integer>(s);
		PNTemporalWindow<Integer> w2 = new PNTemporalWindow<Integer>(int2PN, windowSize);
		PNTemporalPipe<Integer> PN2Int = new PNTemporalPipe<Integer>(w2);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(w1));
		exec.registerQuery(Pipes.verifyStartTimeStampOrdering(PN2Int));
		exec.registerQuery(Pipes.compareSnapshots(w1, PN2Int, false, false));
		exec.startAllQueries();
	}
	
}

