/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.operators.windows;

import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeSet;

import xxl.core.collections.MapEntry;
import xxl.core.collections.queues.DynamicHeap;
import xxl.core.collections.queues.Heap;
import xxl.core.collections.sweepAreas.ListSA;
import xxl.core.collections.sweepAreas.ListSAImplementor;
import xxl.core.collections.sweepAreas.SweepArea;
import xxl.core.collections.sweepAreas.SweepAreaImplementor;
import xxl.core.comparators.ComparableComparator;
import xxl.core.comparators.FeatureComparator;
import xxl.core.comparators.LexicographicalComparator;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeInterval;
import xxl.core.pipes.memoryManager.MemoryMonitorable;
import xxl.core.pipes.operators.AbstractTemporalPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.groupers.HashGrouper;
import xxl.core.pipes.operators.groupers.TemporalGrouper;
import xxl.core.pipes.operators.identities.Coalesce;
import xxl.core.pipes.operators.identities.Coalesce.CoalesceSA;
import xxl.core.pipes.operators.mappers.Mapper;
import xxl.core.pipes.operators.unions.TemporalUnion;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.AbstractSink;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sinks.Tester;
import xxl.core.pipes.sources.RandomNumber;
import xxl.core.pipes.sources.Source;
import xxl.core.predicates.Predicate;
import xxl.core.predicates.Predicates;
import xxl.core.util.XXLSystem;

/**
 * 
 */
public class PartitionedWindow<E,K> extends TemporalGrouper<E,K,E> {
		
	public static class PartitionedWindowSA<E> extends ListSA<TemporalObject<E>> {
		
		public PartitionedWindowSA(final SweepAreaImplementor<TemporalObject<E>> impl, int objectSize, List<TemporalObject<E>> list, final int N) {
    		super(impl, DEFAULT_ID, true, ReorganizeModes.LAZY_REORGANIZE,
    			new Predicate<TemporalObject>() {
					@Override
					public boolean invoke(TemporalObject o1, TemporalObject o2) {
						return impl.size() == N; // checks if implementor is filled
					}
				}, 
				Predicates.FALSE,
				1, list, objectSize
			);
    		if (N <= 0)
				throw new IllegalArgumentException("Window size <= 0 is not allowed.");
    	}
		
		public PartitionedWindowSA(final int N) {
    		this(new ListSAImplementor<TemporalObject<E>>(), SIZE_UNKNOWN, new LinkedList<TemporalObject<E>>(), N);
		}
		
		
		public void remove() {
			Iterator<TemporalObject<E>> it = impl.iterator();
			TemporalObject<E> toRemove = it.next();
			it.remove(); // remove from implementor
			list.remove(toRemove); // remove from list
		}
		
		public TemporalObject<E> split(TemporalObject<E> incomingObject) throws IllegalArgumentException, IllegalStateException {
			if (impl.size() == 0)
				throw new IllegalStateException();
						
			Iterator<TemporalObject<E>> it = list.iterator();
			TemporalObject<E> oldestTSO = it.next();
			it.remove(); // delete from list
			TemporalObject<E> oldPart = new TemporalObject<E>(
				oldestTSO.getObject(), 
				new TimeInterval(oldestTSO.getStart(), incomingObject.getStart())
			);
			TemporalObject<E> newPart = new TemporalObject<E>(
				oldestTSO.getObject(), 
				new TimeInterval(incomingObject.getStart(), incomingObject.getStart()+1)
			);
			list.add(newPart); // insert into list
			impl.update(oldestTSO, newPart); // update implementor: requires a linear scan in worst case
			
			if (list.size() != impl.size())
				throw new IllegalStateException();
			return oldPart;
		}
		
	}
	
	public static Comparator<SweepArea<TemporalObject>> SAStartTimeStampComparator = new LexicographicalComparator<SweepArea<TemporalObject>>(		
		new FeatureComparator<TemporalObject, SweepArea<TemporalObject>>(
			TemporalObject.START_TIMESTAMP_COMPARATOR, 
			new Function<SweepArea<TemporalObject>,TemporalObject>() {
				@Override
				public TemporalObject invoke(SweepArea<TemporalObject> sa) {
					return getFirstElement(sa);
				}
			}
		),
		new FeatureComparator<Integer,SweepArea<TemporalObject>>(
			ComparableComparator.INTEGER_COMPARATOR,
			new Function<SweepArea<TemporalObject>,Integer>() {
				@Override
				public Integer invoke(SweepArea<TemporalObject> sa) {
					return sa.hashCode();
				}
			}
		)
	);
	
	public static TemporalObject getFirstElement(SweepArea<TemporalObject> sweepArea) throws IllegalArgumentException {
		Iterator<TemporalObject> it = sweepArea.iterator();
		if (it.hasNext())
			return it.next();
		throw new IllegalArgumentException("Cannot access first element of the SweepArea");
	}
	
	protected SortedSet<SweepArea<TemporalObject>> groupsOrderedByStartTimestamps;
	protected long delta;
	protected int N;
	
	public PartitionedWindow(Function<? super E,? extends K> determineGroup, final int N, long delta) {
		super(determineGroup, 
			new Function<Object,SweepArea<TemporalObject>>() {
				public SweepArea<TemporalObject> invoke() {
					return (SweepArea<TemporalObject>)new PartitionedWindowSA<E>(N);
				}
			}
		);
		this.groupsOrderedByStartTimestamps = new TreeSet<SweepArea<TemporalObject>>(SAStartTimeStampComparator);
		this.N = N;
		this.delta = delta;
	}	
	
	public PartitionedWindow(Source<? extends TemporalObject<E>> source, int sourceID, Function<? super E,? extends K> determineGroup, final int N, long delta) {
		this(determineGroup, N, delta);
		if (!Pipes.connect(source, this, sourceID))
			throw new IllegalArgumentException("Problems occured while establishing the connections between the nodes."); 
	}
	
	public PartitionedWindow(Source<? extends TemporalObject<E>> source, Function<? super E,? extends K> determineGroup, final int N, long delta) {
		this(source, DEFAULT_ID, determineGroup, N, delta);
	}
			
	@Override
	public void processObject(TemporalObject<E> o, int sourceID) throws IllegalArgumentException {
		K groupID = determineGroup.invoke(o.getObject());
		
		SweepArea<TemporalObject> sweepArea;
		if (!groups.containsKey(groupID)) {
			sweepArea = newSweepArea.invoke();
			groups.put(groupID, sweepArea);
		}
		else {
			sweepArea = groups.get(groupID);
			groupsOrderedByStartTimestamps.remove(sweepArea);
		}
		
		Iterator<TemporalObject> it = sweepArea.query(o, sourceID);
		if (it.hasNext()) {
			TemporalObject<E> next = (TemporalObject<E>)it.next();
			TemporalObject<E> newTSO = new TemporalObject<E>(
				next.getObject(), 
				new TimeInterval(next.getStart(), o.getStart())
			);
			heap.enqueue(newTSO);
			((PartitionedWindowSA)sweepArea).remove();
		}
		
		sweepArea.insert(o);
		groupsOrderedByStartTimestamps.add(sweepArea);
	
		if (delta == Long.MAX_VALUE) // no splitting
			transferHeap(groupsOrderedByStartTimestamps.first().iterator().next().getStart());
		transferHeap(addExpiredElementsToHeap(o));
	}
	
	protected long addExpiredElementsToHeap(TemporalObject<E> o) {
		long minTS = groupsOrderedByStartTimestamps.first().iterator().next().getStart();
		if (o.getStart() - minTS > delta) {
			SweepArea sweepArea = groupsOrderedByStartTimestamps.first();
			groupsOrderedByStartTimestamps.remove(sweepArea);
			heap.enqueue(((PartitionedWindowSA)sweepArea).split(o));
			groupsOrderedByStartTimestamps.add(sweepArea);
		}
		return groupsOrderedByStartTimestamps.first().iterator().next().getStart();
	}
					
	protected void addRemainingElementsToHeap() {
		Iterator<SweepArea<TemporalObject>> groupIterator = groupsOrderedByStartTimestamps.iterator();
		while (groupIterator.hasNext()) {
			SweepArea<TemporalObject> sweepArea = groupIterator.next();
			Iterator<TemporalObject> it = sweepArea.iterator();
			while(it.hasNext()) {
				TemporalObject<E> next = (TemporalObject<E>)it.next();
				TemporalObject<E> newTSO = new TemporalObject<E>(
					next.getObject(),
					new TimeInterval(next.getStart(), TimeInterval.INFINITY)
				);
				heap.enqueue(newTSO);
			}
		}
	}
			
	@Override
	public void close() {
		super.close();
		if (isClosed()) {
			groupsOrderedByStartTimestamps.clear();
		}
	}
		
	/**
	 * @param args
	 */
	@SuppressWarnings("unchecked")
	public static void main(String[] args) {
		/*********************************************************************/
		/*                            Example 1                              */
		/*********************************************************************/
		
		final int noOfElements = 2000;
		final int buckets = 10;
		final int startInc = 50;
		final long seed = 42;
		
		RandomNumber<Integer> r = new RandomNumber<Integer>(
			RandomNumber.DISCRETE, noOfElements, 0
		);		
		
		Source<TemporalObject<Integer>> s = new Mapper<Integer,TemporalObject<Integer>>(r,
			new Function<Integer,TemporalObject<Integer>>() {
				Random random = new Random(seed);
				long start = 0;
				@Override
				public TemporalObject<Integer> invoke(Integer o) {
					int inc = 0;
					while (inc == 0)
						inc += random.nextInt(startInc);
					start = start+inc; // distinct start timestamp required for count-based windows
					TemporalObject<Integer> object = new TemporalObject<Integer>(o, 
						new TimeInterval(start, start+1)
					);
					return object;
				}
			}
		);
	
		final Function<Integer,Integer> hashFunction1 = new Function<Integer,Integer>() {
			@Override
			public Integer invoke(Integer o) {
				return o % buckets;
			}
		};
		final Function<TemporalObject<Integer>,Integer> hashFunction2 = new Function<TemporalObject<Integer>,Integer>() {
			@Override
			public Integer invoke(TemporalObject<Integer> o) {
				return o.getObject() % buckets;
			}
		};
		
		// direct implementation
		final int N = 5;
		long delta = 1000; //Long.MAX_VALUE;
		PartitionedWindow<Integer,Integer> pw = new PartitionedWindow<Integer,Integer>(s, hashFunction1, N, delta);
	
		// indirect implementation
		HashGrouper<TemporalObject<Integer>> grouper = new HashGrouper<TemporalObject<Integer>>(s, 0, hashFunction2, buckets);
		
		CountBasedWindow<Integer>[] windows = new CountBasedWindow[buckets];
		for (int i = 0; i < buckets; i++) {
			windows[i] = new CountBasedWindow<Integer>(grouper.getReferenceToGroup(i), N);
		}
		TemporalUnion<Integer> u = new TemporalUnion<Integer>(windows);
	
		// check sorting
		Pipes.verifyStartTimeStampOrdering(pw);
		Pipes.verifyStartTimeStampOrdering(u);
				
		Coalesce<Integer> c1 = new Coalesce<Integer>(pw, 0, new CoalesceSA<Integer>(new ListSAImplementor<TemporalObject<Integer>>(), Long.MAX_VALUE));
		
		AbstractSink sink = new AbstractSink<TemporalObject<MapEntry<Integer,Integer>>>(new Source[]{c1,u}, new int[]{1,2}) {
			LinkedList<TemporalObject<MapEntry<Integer,Integer>>> list1 = new LinkedList<TemporalObject<MapEntry<Integer,Integer>>>();
			LinkedList<TemporalObject<MapEntry<Integer,Integer>>> list2 = new LinkedList<TemporalObject<MapEntry<Integer,Integer>>>();
			
			@Override
			public void processObject(TemporalObject<MapEntry<Integer,Integer>> o, int sourceID) {				
				if (sourceID == 1) {
					//System.out.println(o);
					list1.add(o);
					return;
				}
				if (sourceID == 2) {
					list2.add(o);
					return;
				}
			}
			
			@Override
			public void done(int sourceID) {
				super.done(sourceID);
				if (isDone()) {
					processingWLock.lock();
					try {
						System.out.println("list1: "+list1);
						System.out.println("list2: "+list2);
						Iterator it1 = list1.iterator();
						while (it1.hasNext()) {
							Object next1 = it1.next();
							if (!list2.contains(next1)) 
								System.err.println("ERROR: Missing results in indirect implementation: "+next1);
							else 
								list2.remove(next1);
						}
						Iterator it2 = list2.iterator();
						while (it2.hasNext()) 
							System.err.println("ERROR: Missing results in direct implementation: "+it2.next());
						System.out.println("CHECKS FINISHED.");
					}
					finally {
						processingWLock.unlock();
					}
				}
			}
		};

		Tester t1 = new Tester<TemporalObject<Integer>>(c1);
		Tester t2 = new Tester<TemporalObject<Integer>>(u);
		
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(sink);
		exec.registerQuery(t1);
		exec.registerQuery(t2);
		exec.startAllQueries();		
	}

}