/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.processors;

import java.util.NoSuchElementException;

import xxl.core.collections.MapEntry;
import xxl.core.collections.queues.Heap;
import xxl.core.comparators.FeatureComparator;
import xxl.core.cursors.sources.Enumerator;
import xxl.core.cursors.sources.Repeater;
import xxl.core.functions.Function;
import xxl.core.pipes.elements.TemporalObject;
import xxl.core.pipes.elements.TimeStampedObject;
import xxl.core.pipes.queryGraph.QueryExecutor;
import xxl.core.pipes.sinks.Printer;
import xxl.core.pipes.sources.AbstractSource;
import xxl.core.pipes.sources.CursorSource;
import xxl.core.pipes.sources.TimeStampCursorSource;

/**
 * A thread that controls the registered data sources according to their temporal 
 * ordering (application time).
 *
 * @param <E>
 */
public class TemporalProcessor<E extends TimeStampedObject> extends SourceProcessor {
	
	/**
	 * The heap contains any tuples buffered from the sources whos
	 * timestamp is greater than currentLogicalTime.
	 */
	protected Heap<MapEntry<E,Integer>> heap;		
	protected long physicalDelay;
	protected final long logicalTimeSteps;
	protected long currentLogicalTime;
	protected boolean initPhase = true;
		
	public TemporalProcessor(long physicalDelay, long logicalTimeSteps) {
		super(physicalDelay);
		this.logicalTimeSteps = logicalTimeSteps;
	}

	public TemporalProcessor() {
		this(0, 1);
	}
	
	@Override
	@SuppressWarnings("unchecked")
	public void process() {
		// progress in time
		currentLogicalTime += logicalTimeSteps;
		// process heap - this loop will be executed once per input with at least one tuple before or equal currentLogicalTime
		heap : while (!heap.isEmpty() && heap.peek().getKey().getTimeStamp() <= currentLogicalTime) {
			MapEntry<E,Integer> top = heap.dequeue();
			int i = top.getValue();
			AbstractSource source = sources.get(i);
			E tso = top.getKey();
			try {
				// process input i until first tuple after currentLogicalTime
				do {
					source.transfer(tso);								
					if (source.isClosed()) {
						closedOrDone.set(i);
						checkTermination();
						if (terminate)
							return;
						continue heap;
					}
					tso = (E)source.next();
				}
				while(tso.getTimeStamp() <= currentLogicalTime);
				// save next tuple in heap
				heap.enqueue(new MapEntry<E,Integer>(tso, i));
			}
			catch (NoSuchElementException nsee) {
				source.signalDone();
				closedOrDone.set(i);
				checkTermination();
				if (terminate)
					return;
			}			
		}				
	}
			
	@Override
	protected void onStart() {
		// prohibit further registrations
		initPhase = false;
		// create heap 
		heap = new Heap<MapEntry<E,Integer>>(sources.size(), 
			new FeatureComparator<E, MapEntry<E,Integer>>(TimeStampedObject.START_TIMESTAMP_COMPARATOR,
				new Function<MapEntry<E,Integer>, E>() {
					@Override
					public E invoke(MapEntry<E,Integer> entry) {
						return entry.getKey();
					}
				}
			)
		);		
		// fill heap with one tuple per input
		AbstractSource source;
		for (int i = 0; i < sources.size(); i++) {
			source = sources.get(i);
			if (source.isClosed())
				closedOrDone.set(i);
			else {
				try {
					E a = (E)source.next();
					MapEntry<E,Integer> e = new MapEntry<E,Integer>(a, i);
					
					// MapEntry<E,Integer> e = new MapEntry<E,Integer>((E)source.next(), i);
					heap.enqueue(e);
				}
				catch (NoSuchElementException nsee) {
					source.signalDone();
					closedOrDone.set(i);
				}
			}
		}
		currentLogicalTime = Long.MIN_VALUE;
		// ensure heap condition
		if(!heap.isEmpty())
			currentLogicalTime = heap.peek().getKey().getTimeStamp() - 1;
	}
	
	@Override
	@SuppressWarnings("unchecked")
	public void registerSource(AbstractSource source) {
		super.registerSource(source);
		if (!initPhase) {
			throw new RuntimeException("Cannot add new source after starting.");
		}
	}
	
	public static void main(String[] args)  {
		int step = 1000;
		TemporalProcessor tp = new TemporalProcessor(step, step);
		TimeStampCursorSource<Integer> s1 = new TimeStampCursorSource<Integer>(tp, new Enumerator(10001,10041), new Repeater<Integer>(500));
		TimeStampCursorSource<Integer> s2 = new TimeStampCursorSource<Integer>(tp, new Enumerator(20001,20021), new Repeater<Integer>(1000));
		TimeStampCursorSource<Integer> s3 = new TimeStampCursorSource<Integer>(tp, new Enumerator(30001,30011), new Repeater<Integer>(2000));
		Printer<TemporalObject<Integer>> p1 = new Printer<TemporalObject<Integer>>(s1);
		Printer<TemporalObject<Integer>> p2 = new Printer<TemporalObject<Integer>>(s2);
		Printer<TemporalObject<Integer>> p3 = new Printer<TemporalObject<Integer>>(s3);
		QueryExecutor exec = new QueryExecutor();
		exec.registerQuery(p1);
		exec.registerQuery(p2);
		exec.registerQuery(p3);
		exec.startAllQueries();
	}
	
}
