/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.queryGraph;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import xxl.core.functions.Function;
import xxl.core.pipes.metaData.AggregationMetaDataIdentifier;
import xxl.core.pipes.metaData.ComplexMetaDataIdentifier;
import xxl.core.pipes.metaData.DefaultMetaDataHandler;
import xxl.core.pipes.metaData.MetaDataDependencies;
import xxl.core.pipes.metaData.PeriodicEvaluationMetaDataHandler;
import xxl.core.util.metaData.AbstractMetaDataManagement;
import xxl.core.util.metaData.CompositeMetaData;
import xxl.core.util.metaData.MetaDataException;
import xxl.core.util.metaData.MetaDataHandler;
import xxl.core.util.metaData.MetaDataManagement;

public abstract class AbstractNode implements Node {
		
	public static class NodeMetaDataDependenciesContainer implements ContainsNodeMetaDataDependencies {

		protected ContainsNodeMetaDataDependencies superClassMetaDataDependencies = null;
		protected Map<Object,Object[]> metaDataRequiredLocally;
		
		public NodeMetaDataDependenciesContainer(Class<? extends Node> className) {
			Class searchClass = className;
			while (superClassMetaDataDependencies == null && !searchClass.equals(Object.class)) {
				Class superClass =  searchClass.getSuperclass();
				superClassMetaDataDependencies = GLOBAL_METADATA_DEPENDENCIES.get(superClass);
				searchClass = superClass;
			}
		}
				
		public void addLocalMetaDataDependencies(Object metaDataIdentifier, Object... dependencies) {
			if (metaDataRequiredLocally == null)
				metaDataRequiredLocally = new HashMap<Object,Object[]>();
			if (metaDataRequiredLocally.containsKey(metaDataIdentifier))
				throw new IllegalArgumentException("Only one dependency declaration allowed per class and metaDataIdentifier");				
			metaDataRequiredLocally.put(metaDataIdentifier, dependencies);
		}
		
		public ContainsNodeMetaDataDependencies getSuperClassMetaDataDependencies() {
			return superClassMetaDataDependencies;
		}
		
		public Object[] getLocalMetaDataDependencies(Object metaDataIdentifier) {
			if (metaDataRequiredLocally == null || !metaDataRequiredLocally.containsKey(metaDataIdentifier)) 
				return new Object[0];
			return metaDataRequiredLocally.get(metaDataIdentifier);
		}
		
		public Map<Object,Object[]> getLocalMetaDataDependencies() {
			if (metaDataRequiredLocally == null) 
				return Collections.emptyMap();
			return metaDataRequiredLocally;
		}
		
	}
	
	
	public abstract class AbstractNodeMetaDataManagement extends AbstractMetaDataManagement<Object,Object> {
		
		
		protected Class<? extends Node> enclosingClass;
		protected long updatePeriod;
		
		protected HashMap<Object,Set <Object>> inverseComplexDependencies;
		
		public static final String META_DATA_UPDATE_RATE = "META_DATA_UPDATE_RATE";
		
		@SuppressWarnings("unchecked")
		public AbstractNodeMetaDataManagement() {
			super();
			this.enclosingClass = (Class<? extends Node>)this.getClass().getEnclosingClass();
			this.updatePeriod = 500;
		}
		
		/**
		 * Sets the update period for the metadata items. This method can 
		 * only be called when no periodic updating takes place.
		 * 
		 * @param millis
		 */
		public synchronized void setUpdatePeriod(long millis) {
			if (metaData != null) {
				Iterator md = metaData.fragments();
				while (md.hasNext()) 
					if (md.next() instanceof PeriodicEvaluationMetaDataHandler)
						throw new IllegalStateException("Update period may only be changed when no periodic updating takes place");
			}
			this.updatePeriod = millis;
		}
		
		public synchronized long getUpdatePeriod() {
			return this.updatePeriod;
		}
		
		@SuppressWarnings("unchecked")
		public final synchronized Object[] getLocalDependencies(Object metaDataIdentifier) {
			Object [] staticLocalDependencies = MetaDataDependencies.getLocalMetaDataDependencies(enclosingClass, metaDataIdentifier);
			Set <Object> dynamicLocalDependencies = getDynamicLocalDependencies(metaDataIdentifier);
			if (staticLocalDependencies.length==0)
				return dynamicLocalDependencies.toArray();			
			if (dynamicLocalDependencies.size()==0)
				return staticLocalDependencies;
			for (Object dependency : staticLocalDependencies)
				dynamicLocalDependencies.add(dependency);
			return dynamicLocalDependencies.toArray();
		}

		@SuppressWarnings("unchecked")
		protected synchronized Set<Object> getDynamicLocalDependencies(Object metaDataIdentifier) {
			Set<Object> result = new HashSet<Object>();
			if (metaDataIdentifier instanceof ComplexMetaDataIdentifier) {
				ComplexMetaDataIdentifier complexMetaDataIdentifier = (ComplexMetaDataIdentifier) metaDataIdentifier;
				result.add(complexMetaDataIdentifier.getBasisMetaDataIdentifier());				
			}
			return result;
		}

		@SuppressWarnings("unchecked")
		protected synchronized Set<Object> getInverseDynamicLocalDependencies(Object metaDataIdentifier) {
			Set<Object> result = new HashSet<Object>();
			if (inverseComplexDependencies != null && inverseComplexDependencies.containsKey(metaDataIdentifier))
				result.addAll(inverseComplexDependencies.get(metaDataIdentifier));
			return result;
		}
		
		// inverse function of getLocalDependencies
		@SuppressWarnings("unchecked")
		public final synchronized Object[] affectsLocalMetaData(Object[] changedMDIdentifiers) {			
			// dynamic dependencies
			Set<Object> dynamicResults = new HashSet<Object>();
			for (Object changedMDIdentifier : changedMDIdentifiers) 
				for (Object dependency : getInverseDynamicLocalDependencies(changedMDIdentifier))
					dynamicResults.add(dependency);
            // static dependencies
			Object[] staticResults = MetaDataDependencies.affectsLocalMetaData((Class<? extends Node>)enclosingClass, changedMDIdentifiers);
			
			if (dynamicResults.size()>0) {
				for (Object staticResult : staticResults)
					dynamicResults.add(staticResult);
				return dynamicResults.toArray();
			}
			else
				return staticResults;
		}		
						
		@Override
		public synchronized boolean include(Object metaDataIdentifier) throws MetaDataException {
			Object [] localDependencies = getLocalDependencies(metaDataIdentifier);
			int dependency = 0;
			while (dependency<localDependencies.length && include(localDependencies[dependency])) 
				dependency++;
			boolean ok = super.include(metaDataIdentifier) && dependency==localDependencies.length; 
			if (!ok) {
				for (int undo=0; undo<dependency; undo++)
					exclude(localDependencies[undo]);
				return false;
			}
			return true;
		}
				
		@Override
		public synchronized void exclude(Object metaDataIdentifier) throws MetaDataException {			
			super.exclude(metaDataIdentifier);
			Object [] localDependencies = getLocalDependencies(metaDataIdentifier);
			for (int i=0; i<localDependencies.length; i++)
				exclude(localDependencies[i]);
		}
		
		public void refresh(Object... changedMDIdentifiers) {
			if (metaData != null)
				for (Object md : changedMDIdentifiers)
					if (metaData.contains(md))
						((MetaDataHandler)metaData.get(md)).refresh();
		}
		
		public void globalNotification(Object... changedMDIdentifiers) {
			localNotification(changedMDIdentifiers);
		}
		
		public void localNotification(Object... changedMDIdentifiers) {
			refresh(affectsLocalMetaData(changedMDIdentifiers));		
		}
				
		// helper methods
		// called in node this metadata depends on
		protected boolean manageExternalDependency(boolean create, HashMap<Integer,List<Object>> map, int ID, Object... metaDataIdentifiers) {
			synchronized(this) {
				if (create)
					MetaDataDependencies.addToMap(map, ID, metaDataIdentifiers);
				else
					MetaDataDependencies.removeFromMap(map, ID, metaDataIdentifiers);
			}
			if (create) {
				boolean[] res = includeAll(metaDataIdentifiers);
				for (boolean r : res)
					if (!r) 
						return false;
				return true;
			}
			excludeAll(metaDataIdentifiers);
			return true;
		}
				
		// called in node this metadata depends on
		protected boolean removeAllExternalDependencies(HashMap<Integer,List<Object>> map, int ID) {
			synchronized(this) {
				if (map == null || !map.containsKey(ID))
					return false;
			}
			return manageExternalDependency(false, map, ID, map.get(ID).toArray());
		}
		
		@Override
		protected boolean addMetaData(Object metaDataIdentifier) {
			if (metaDataIdentifier.equals(META_DATA_UPDATE_RATE)) {
				metaData.add(metaDataIdentifier, new DefaultMetaDataHandler<Long>(this, metaDataIdentifier, Long.class,
						new Function<Object,Long>() {
							@Override
							public Long invoke () {
								synchronized(metaDataManagement) {
									return getUpdatePeriod();
								}
							}
						})
					);
				return true;
			}
			if (metaDataIdentifier instanceof ComplexMetaDataIdentifier) {
				ComplexMetaDataIdentifier complexMetaDataIdentifier= (ComplexMetaDataIdentifier) metaDataIdentifier;
				Object basisMetaDataIdentifier = complexMetaDataIdentifier.getBasisMetaDataIdentifier();
				Object instance = complexMetaDataIdentifier.newInstance(this);
				if (instance==null)
					return false;
				metaData.add(metaDataIdentifier, instance);
				if (inverseComplexDependencies==null)
					inverseComplexDependencies = new HashMap<Object,Set<Object>>();
				if (!inverseComplexDependencies.containsKey(basisMetaDataIdentifier))
					inverseComplexDependencies.put(basisMetaDataIdentifier, new HashSet<Object>());
				inverseComplexDependencies.get(basisMetaDataIdentifier).add(complexMetaDataIdentifier);
				return true;
			}
			return false;
		}

		@Override
		protected boolean removeMetaData(Object metaDataIdentifier) {	
			if (metaDataIdentifier.equals(META_DATA_UPDATE_RATE)) {
				metaData.remove(metaDataIdentifier);
				return true;
			}
			if (metaDataIdentifier instanceof ComplexMetaDataIdentifier && metaData.contains(metaDataIdentifier)) {
				metaData.remove(metaDataIdentifier);
				HashSet<Object> toRemove = new HashSet<Object>();
				for (Entry<Object,Set<Object>> e : inverseComplexDependencies.entrySet()) 
					if (e.getValue().contains(metaDataIdentifier))
						toRemove.add(e.getKey());
						//inverseComplexDependencies.remove(e.getKey());
				for (Object e: toRemove)
					inverseComplexDependencies.remove(e);
				if (inverseComplexDependencies.size()==0)
					inverseComplexDependencies = null;
				return true;
			}
			throw new IllegalStateException();
		}
		
	}
	

	/**
	 * The query graph this node belongs to.
	 */
	protected Graph graph;
	
	protected MetaDataManagement<Object,Object> metaDataManagement;
	
	public AbstractNode() {
		this.metaDataManagement = null;
		createMetaDataManagement();
		this.graph = Graph.DEFAULT_INSTANCE;
		this.graph.addNode(this);
	}
		
	public void setGraph(Graph graph) {
		if (this.graph != null)
			this.graph.removeNode(this);
		this.graph = graph;
		this.graph.addNode(this);
	}

	public Graph getGraph() {
		return graph;
	}

	public CompositeMetaData<Object,Object> getMetaData() {
		return metaDataManagement.getMetaData();
	}
	
	public abstract void createMetaDataManagement();
	
	public MetaDataManagement<Object,Object> getMetaDataManagement() {
		return metaDataManagement;
	}
	
}
