/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.queryGraph;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Stack;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.ReadLock;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;

import xxl.core.pipes.metaData.GraphAggregationMetaDataIdentifier;
import xxl.core.pipes.metaData.PeriodicEvaluationMetaDataHandler;
import xxl.core.pipes.operators.Pipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.operators.identities.BufferPipe;
import xxl.core.pipes.operators.identities.TemporalBufferPipe;
import xxl.core.pipes.scheduler.Controllable;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;
import xxl.core.util.metaData.AbstractMetaDataManagement;
import xxl.core.util.metaData.CompositeMetaData;
import xxl.core.util.metaData.MetaDataManageable;
import xxl.core.util.metaData.MetaDataManagement;


/**
 * A query graph consists of multiple possibly shared continuous queries.
 * A physical query is implemented as a tree of nodes, namely: sources, operators, 
 * and sinks. <br> 
 * Modifications of the query graph like subscriptions and plan migration require the graph's write lock.
 * All other operations that access the graph nodes need a graph's read lock.  
 */
public class Graph implements MetaDataManageable<Object,Object> {

    protected volatile boolean sumMetaData = false;
    
    public class GraphMetaDataManagement extends AbstractMetaDataManagement<Object,Object> {

        protected HashSet<Object> basisMetaData;
        
        @Override
        protected boolean addMetaData(Object metaDataIdentifier) {
            if (metaDataIdentifier instanceof GraphAggregationMetaDataIdentifier) {
                GraphAggregationMetaDataIdentifier mdi = (GraphAggregationMetaDataIdentifier) metaDataIdentifier;
                if (!sumMetaData) {
                    sumMetaData = true;
                    basisMetaData = new HashSet<Object>();
                }
                for (Node n : nodes)
                    n.getMetaDataManagement().include(mdi.getBasisMetaDataIdentifier());
                basisMetaData.add(mdi.getBasisMetaDataIdentifier());
                metaData.add(metaDataIdentifier, mdi.newInstance(this));
                return true;
            }
            return false;
        }

        @Override
        protected boolean removeMetaData(Object metaDataIdentifier) {
            if (metaDataIdentifier instanceof GraphAggregationMetaDataIdentifier) {
                GraphAggregationMetaDataIdentifier mdi = (GraphAggregationMetaDataIdentifier) metaDataIdentifier;
                ((PeriodicEvaluationMetaDataHandler)metaData.get(metaDataIdentifier)).close();
                metaData.remove(metaDataIdentifier);
                basisMetaData.remove(mdi.getBasisMetaDataIdentifier());
                for (Node n : nodes) {
                    MetaDataManagement<Object,Object> mdm = n.getMetaDataManagement();
                    Object bmdi = mdi.getBasisMetaDataIdentifier();
                    if (mdm.getMetaData().contains(bmdi))
                        mdm.exclude(bmdi);
                }
                if (basisMetaData.size()==0) {
                    basisMetaData = null;
                    sumMetaData = false;
                }
                return true;
            }
            return false;
        }
        
        public ArrayList<Node> getNodes() {
            return nodes;
        }
        
    }

    public static final Graph DEFAULT_INSTANCE = new Graph();
    
    public final ReentrantReadWriteLock RWLock;
    public final ReadLock RLock;
    public final WriteLock WLock;
    
    protected MetaDataManagement<Object,Object> metaDataManagement;
    
    /**
     * list of all nodes in the graph;
     */
    private ArrayList<Node> nodes;
    
    private ArrayList<Integer> positionPool;
    
    private int nodeCount;
    
    /**
     * map from node to the position in the list
     */
    private HashMap<Node,Integer> nodeToPosition;
    
    /**
     * indicates initial nodes
     */
    private BitSet initialNodes;
    
    /**
     * indicates terminal nodes
     */
    private BitSet terminalNodes;
    
    /**
     * indicates queues
     */
    private BitSet queues;
    
    /**
     * indicates pipes
     */
    private BitSet pipes;
    
    /**
     * operatorpaths
     * an operatorpath is a sequence of connected operators.
     * The first operator is a source of the graph and the last 
     * operator is a sink of the graph.
     * 
     */
    private OperatorPath[] operatorPaths;
    
    /**
     * flag for the operator Paths, indicates if the graph has changed since the operatoprPaths
     * were computed last time
     */
    private boolean changes;
    
    
    /**
     * new Graph instance initializes the lists
     */
    public Graph() {
        this(new ReentrantReadWriteLock());
        nodes = new ArrayList<Node>();
        positionPool = new ArrayList<Integer>();
        nodeToPosition = new HashMap<Node,Integer>();
        initialNodes = new BitSet();
        terminalNodes = new BitSet();
        pipes = new BitSet();
        queues = new BitSet();
        changes = false;
    }
    
    public Graph(ReentrantReadWriteLock graphRWLock) {
        this.RWLock = graphRWLock;
        this.RLock = graphRWLock.readLock();
        this.WLock = graphRWLock.writeLock();
        nodes = new ArrayList<Node>();
        createMetaDataManagement();
    }

    public void createMetaDataManagement() {
        if (metaDataManagement != null)
            throw new IllegalStateException("An instance of MetaDataManagement already exists.");
        metaDataManagement = new GraphMetaDataManagement();
    }

    public CompositeMetaData<Object,Object> getMetaData() {
        return metaDataManagement.getMetaData();
    }

    public MetaDataManagement<Object,Object> getMetaDataManagement() {
        return metaDataManagement;
    }
    

    
    
    /**
     * @param node node to check
     * @return true: node is in graph else false
     */
    public boolean containsNode(Node node) {
        return nodeToPosition.containsKey(node);
    }
    
    /**
     * @param node node to check
     * @return true: pipe is in graph, else false
     */
    public boolean containsPipe(Node node) {
        return containsNode(node) && pipes.get(nodeToPosition.get(node));
    }

    /**
     * @param node  node to check
     * @return true : source is initial source in this graph, else false
     */
    public boolean isInitialSource(Node node) {
        return containsNode(node) && initialNodes.get(nodeToPosition.get(node));
    }

    /**
     * @param node node to check
     * @return true: sink is terminal sink in this graph, else false
     */
    public boolean isTerminalSink(Node node) {
        return containsNode(node) && terminalNodes.get(nodeToPosition.get(node));
    }

    /**
     * @param node node to check
     * @return true: Queue is in this Graph
     */
    public boolean containsQueue(Node node) {
        return containsNode(node) && queues.get(nodeToPosition.get(node));
    }
    
    /**
     * @return number of nodes in this graph
     */
    public synchronized int nodeCount() {
        return nodeCount;
    }
    
    /**
     * @return list of all nodes in this graph
     */
    public Node[] getNodes() {
        Node[] result = new Node[nodeCount()];
        int index = 0;
        for (int i=0;i<nodes.size();i++) 
            if (nodes.get(i)!=null)
                result[index++] = nodes.get(i);
        return result;
    }
    
    /** 
     * @return list of all initial sources
     */
    public Source[] getSources() {
        int nrOfSources = initialNodes.cardinality();
        Source[] sources = new Source[nrOfSources];
        int index = 0;
        int nextPos = -1;
        while (index < nrOfSources) {
            nextPos = initialNodes.nextSetBit(nextPos+1);
            sources[index++] = (Source) nodes.get(nextPos);
        }
        return sources;
    }
    
    /**
     * @return list of all terminal sinks
     */
    public Sink[] getSinks() {
        int nrOfSinks = terminalNodes.cardinality();
        Sink[] sinks = new Sink[nrOfSinks];
        int index = 0;
        int nextPos = -1;
        while (index < nrOfSinks) {
            nextPos = terminalNodes.nextSetBit(nextPos+1);
            sinks[index++] = (Sink) nodes.get(nextPos);
        }
        return sinks;
    }
    
    /**
     * @return list of all pipes in the system
     */
    public Pipe[] getPipes() {
        int nrOfPipes = pipes.cardinality();
        Pipe[] pipRes = new Pipe[nrOfPipes];
        int index = 0;
        int nextPos = -1;
        while (index < nrOfPipes) {
            nextPos = pipes.nextSetBit(nextPos+1);
            pipRes[index++] = (Pipe) nodes.get(nextPos);
        }
        return pipRes;
    }
    
    /**
     * @return list of all queues in the system
     */
    public AbstractBufferPipe[] getQueues() {
        int nrOfQueues = queues.cardinality();
        AbstractBufferPipe[] buffers = new AbstractBufferPipe[nrOfQueues];
        int index = 0;
        int nextPos = -1;
        while (index < nrOfQueues) {
            nextPos = queues.nextSetBit(nextPos+1);
            buffers[index++] = (AbstractBufferPipe) nodes.get(nextPos);
        }
        return buffers;
    }
    
    /**
     * adds node to the Graph and sets the appropriate bits in the BitSet
     * 
     * @param node node to add
     */
    public synchronized void addNode(Node node) {
        if (containsNode(node))
            return;
        int index = -1;
        if (positionPool.size()>0) {
            index = positionPool.remove(0);
            nodes.set(index,node);
        }
        else {
            index = nodes.size();
            nodes.add(node);
        }
        nodeToPosition.put(node, index);
        if (node instanceof Controllable)
            queues.set(index);
        if (node instanceof Pipe) {
            pipes.set(index);
            Pipe pipe = (Pipe) node;    
            boolean isInitial = true;
            boolean isTerminal = true;
            for (int i=0;i<pipe.getNoOfSources();i++) {
                Source source = pipe.getSource(i);
                if (containsNode(source)) 
                    isInitial = false;
                if (isTerminalSink(source))
                    terminalNodes.clear(nodeToPosition.get(pipe.getSource(i)));
            }
            for (int i=0;i<pipe.getNoOfSinks();i++) {
                Sink sink = pipe.getSink(i);
                if (containsNode(sink)) 
                    isTerminal = false;
                if (isInitialSource(sink)) 
                    initialNodes.clear(nodeToPosition.get(pipe.getSink(i)));
            }
            if (isInitial)
                initialNodes.set(index);
            if (isTerminal)
                terminalNodes.set(index);
        }
        else if (node instanceof Source) {
            initialNodes.set(index);
            Source source = (Source) node;
            for (int i=0;i<source.getNoOfSinks();i++) {
                Sink sink = source.getSink(i);
                if (isInitialSource(sink))
                    initialNodes.clear(nodeToPosition.get(sink));
            }
        }
        else if (node instanceof Sink) {
            terminalNodes.set(index);
            Sink sink = (Sink) node;
            for (int i=0;i<sink.getNoOfSources();i++) {
                Source source = sink.getSource(i);
                if (isTerminalSink(source))
                    terminalNodes.clear(nodeToPosition.get(source));
            }
        }
        changes = true;
        nodeCount++;
        if (sumMetaData) {
            synchronized (metaDataManagement) {
                GraphMetaDataManagement mdm = (GraphMetaDataManagement)metaDataManagement;
                for(Object o : mdm.basisMetaData)
                    node.getMetaDataManagement().include(o);
            }
        }
    }

    /**
     * removes node from the Graph and changes lists
     * 
     * @param node
     *            node to remove
     */
    public synchronized void removeNode(Node node) {
        if (node instanceof Sink) {
            Sink sink = (Sink) node;
            for (int i = 0; i < sink.getNoOfSources(); i++) {
                Source source = sink.getSource(i);
                if (containsNode(source)) {
                    boolean test = true;
                    for (int j=0;j<source.getNoOfSinks();j++) {
                        Sink help = source.getSink(j);
                        if (containsNode(help)) {
                            test = false;
                            break;
                        }
                    }
                    if (test)
                        terminalNodes.set(nodeToPosition.get(source));
                }
            }
        }
        if (node instanceof Source) {
            Source source = (Source) node;
            for (int i = 0; i < source.getNoOfSinks(); i++) {
                Sink sink = source.getSink(i);
                if (containsNode(sink)) {
                    boolean test = true;
                    for (int j=0;j<sink.getNoOfSources();j++) {
                        Source help = sink.getSource(j);
                        if (containsNode(help)) {
                            test = false;
                            break;
                        }
                    }
                    if (test)
                        initialNodes.set(nodeToPosition.get(sink));
                }
            }
        }
        int index = nodeToPosition.get(node);
        nodes.set(index, null);
        nodeToPosition.put(node, null);
        queues.clear(index);
        pipes.clear(index);
        terminalNodes.clear(index);
        initialNodes.clear(index);
        positionPool.add(index);
        changes = true;
        nodeCount--;
        if (sumMetaData) {
            synchronized (metaDataManagement) {
                GraphMetaDataManagement mdm = (GraphMetaDataManagement)metaDataManagement;
                for(Object o : mdm.basisMetaData)
                    node.getMetaDataManagement().exclude(o);
            }
        }
    }
    
    /**
     * inserts n nodes in batchmode
     * 
     * @param nodes
     *            list of n nodes to insert
     */
    public void addNodes(Node[] nodes) {
        for (int i = 0; i < nodes.length; i++)
            addNode(nodes[i]);
    }   
    
    /**
     * adds buffer between the two nodes consistency of parameters is checked
     * (hopefully complete) indexOut indexIn source -------- buffer --------
     * Sink
     * 
     * @param opSrc startnode
     * @param opGoal endnode
     * @param indexOut index out of Src
     * @param indexIn index in sink
     * @param atStart insert BP at start
     * @return ? action is succesful
     */
    public synchronized AbstractBufferPipe addBuffer(Source opSrc, Sink opGoal, boolean atStart) {
        // Test auf konsistente Parameter
        BufferPipe bp = new TemporalBufferPipe();
        boolean test = (containsNode(opSrc) || atStart)
                && containsNode(opGoal)
                && Pipes.isConnected(opSrc, opGoal);
        if (!test)
            return null;
        try {
            opSrc.getGraph().WLock.lock();
            int indexIn = opGoal.getSourceID(opSrc);
            int indexOut = opSrc.getSinkID(opGoal);
            Pipes.disconnect(opSrc, opGoal, indexIn);
            Pipes.connect(opSrc, bp, indexOut);
            Pipes.connect(bp, opGoal, indexIn);
            addNode(bp);
            if (operatorPaths!=null) {
                for (int i=0;i<operatorPaths.length;i++) {
                    if (operatorPaths[i].containsNode(opGoal) && (operatorPaths[i].containsNode(opSrc)||atStart)) {
                        operatorPaths[i].addNode(opGoal,bp);
                    }
                }
            }
        } finally {
            opSrc.getGraph().WLock.unlock();
        }
        return bp;
    }
    
    /**
     * removes buffer between the two nodes consistency of parameters is checked
     * (hopefully complete) Assumption: buffer contains no elements indexOut
     * indexIn source -------- buffer -------- Sink
     * 
     * @param opSrc
     *            startnode
     * @param opGoal
     *            endnode
     * @param indexOut
     *            index out of Src
     * @param indexIn
     *            index in sink
     * @return ? action is succesful
     */
    public synchronized AbstractBufferPipe removeBuffer(Source opSrc, AbstractBufferPipe buffer, Sink opGoal) {
        // Test auf konsistente Parameter
        boolean test = containsNode(opSrc)
                && containsNode(opGoal)
                && containsQueue(buffer)
                && Pipes.isConnected(opSrc, buffer)
                && Pipes.isConnected(buffer, opGoal);
        if (!test)
            return null;
        try {
            buffer.lockBuffer();
            opSrc.getGraph().WLock.lock();
            buffer.flush();
            int indexOut = opSrc.getSinkID(buffer);
            int indexIn = opGoal.getSourceID(buffer);
            if (!opGoal.isDone()) {
                Pipes.disconnect(opSrc, buffer, indexOut);
                Pipes.disconnect(buffer, opGoal, indexIn);
                Pipes.connect(opSrc, opGoal, indexIn);
                removeNode(buffer);
                buffer.close();
            
                if (operatorPaths!=null) {
                    for (int i=0;i<operatorPaths.length;i++) {
                        if (operatorPaths[i].containsNode(buffer)) {
                            operatorPaths[i].removeNode(buffer);
                        }
                    }
                }
            }
        } finally {
            buffer.releaseBuffer();
            opSrc.getGraph().WLock.unlock();
        }
        return buffer;
    }

    /**
     * Returns the list of all OperatorPath.
     * 
     * An OperatorPath is a path in the graph which starts at any initial source
     * and ends at any terminal sink
     * 
     * @return list of all operatorpaths
     */
    public synchronized OperatorPath[] getOperatorPaths() {
        if (operatorPaths==null || changes) {
            buildOperatorPaths();
            changes = false;
        }
        return operatorPaths;
    }
    
    /**
     * computes and initiates all operatorPaths of this graph
     */
    private void buildOperatorPaths() {
        ArrayList<ArrayList<Node>> result = new ArrayList<ArrayList<Node>>();
        Stack<Node> nodestack = new Stack<Node>();
        Stack<ArrayList<Node>> paths = new Stack<ArrayList<Node>>();
        Node[] initNd = getSources();
        for (int i = 0; i < initNd.length; i++) {
            ArrayList<Node> graph = new ArrayList<Node>();
            nodestack.push(initNd[i]);
            paths.add(graph);
            while (!nodestack.empty()) {
                Node node = nodestack.pop();
                ArrayList<Node> path = paths.pop();
                path.add(node);
                if (isTerminalSink(node)) 
                    result.add((ArrayList<Node>) path.clone());
                else {
                    Source source = (Source) node;
                    for (int j = 0; j < source.getNoOfSinks(); j++) {
                        Sink sink = source.getSink(j);
                        if (containsNode(sink)) {
                            nodestack.push(sink);
                            paths.push((ArrayList<Node>) path.clone());
                        }
                    }
                }
            }
        }
        operatorPaths = new OperatorPath[result.size()];
        for (int i=0;i<result.size();i++) 
            operatorPaths[i] = new OperatorPath(this,result.get(i));
    }
    
    /**
     * creates a shallow copy of this graph object
     * 
     * @return shallow copy
     */
    public Graph copy() {
        Graph clone = new Graph();
        clone.addNodes(getNodes());
        return clone;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString() {
        String result = "Nodes: ";
        for (int i=0;i<nodes.size();i++) {
            result += "\n\t"
                    + nodes.get(i)
                    + " "
                    + ((initialNodes.get(i)) ? 1:0)
                    + ((pipes.get(i)) ? 1:0)
                    + ((queues.get(i)) ? 1:0)
                    + ((terminalNodes.get(i)) ? 1:0);
        }
        return result;
    }
    
    /**
     * returns a list of all queries in this query graph
     * 
     * a query is a tree with one terminal sink as root and all its descendants
     * 
     * @return list of all query-trees
     */
    public synchronized Graph[] getQueries() {
        Graph[] result = new Graph[terminalNodes.cardinality()];
        Node[] termNodes = getSinks();
        for (int i = 0; i < termNodes.length; i++) {
            Graph query = new Graph();
            Stack<Node> stack = new Stack<Node>();
            stack.push(termNodes[i]);
            query.addNode(termNodes[i]);
            while (!stack.isEmpty()) {
                Sink sink = (Sink) stack.pop();
                for (int j = 0; j < sink.getNoOfSources(); j++) {
                    Source source = sink.getSource(j);
                    if (containsNode(source)) {
                        query.addNode(source);
                        if (!isInitialSource(source))
                            stack.push(source);
                    }
                }
            }
            result[i] = query;
        }
        return result;
    }
    
    public Set<Node> getAdjacentNodes() {
        Set<Node> result = new HashSet<Node>();
        for (Node node : nodes) {
            boolean isSink = containsPipe(node) || isTerminalSink(node);
            boolean isSource = containsPipe(node) || isInitialSource(node);
            if (isSource) {
                Source source = (Source) node;
                for (int i=0;i<source.getNoOfSinks();i++) {
                    Sink sink = source.getSink(i);
                    if (!containsNode(sink))
                        result.add(sink);
                }
            }
            if (isSink) {
                Sink sink = (Sink) node;
                for (int i=0;i<sink.getNoOfSources();i++) {
                    Source source = sink.getSource(i);
                    if (!containsNode(source))
                        result.add(source);
                }
            }
        }
        return result;
    }
    
    /**
     * 
     * @param graph
     * @return
     */
    public boolean equals(Object g) {
        try {
            Graph graph = (Graph) g;
            if (nodes.size()!=graph.nodes.size())
                return false;
            for (Node node : nodes)
                if (!graph.containsNode(node))
                    return false;
            return true;
        }
        catch (ClassCastException cce) {
            return false;
        }
    }
    
    public boolean isSubgraph(Graph graph) {
        if (graph.nodeCount>nodeCount)
            return false;
        for (Node node : graph.nodes) {
            if (!containsNode(node))
                return false;
        }
        return true;
    }
    
    public Graph diffGraph(Graph sub) {
        Graph res = new Graph();
        for (Node node : nodes) {
            if (!sub.containsNode(node))
                res.addNode(node);
        }
        return res;
    }
    
    public Graph intersection(Graph graph) {
        Graph res = new Graph();
        for (Node node : nodes) {
            if (graph.containsNode(node))
                res.addNode(node);
        }
        return res;
    }
    
    public Graph union(Graph graph) {
        Graph res = new Graph();
        for (Node node : nodes)
            res.addNode(node);
        for (Node node : graph.nodes)
            res.addNode(node);
        return res;
    }
    
    public boolean isAdjacent(Node n1, Node n2) {
        if (!containsPipe(n1) && !isInitialSource(n1))
            return false;
        Source source = (Source) n1;
        for (int i=0;i<source.getNoOfSinks();i++) {
            if (source.getSink(i) == n2)
                return true;
        }
        return false;
    }
    
}

