/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.queryGraph;

import java.util.ArrayList;
import java.util.Iterator;

import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;

/**
 * 
 * represents an operatorpath in a query graph. 
 *
 */
public class OperatorPath {

	private Graph graph;
	
	private ArrayList<Node> nodes;

	private Source source;

	private Sink sink;

	private ArrayList<AbstractBufferPipe> queues;
	

	public OperatorPath(Graph g, ArrayList<Node> n) {
		graph = g;
		nodes = n;
		source = (Source) nodes.get(0);
		sink = (Sink) nodes.get(nodes.size()-1);
		queues = new ArrayList<AbstractBufferPipe>();
		for (int i=0;i<nodes.size();i++) {
			if (graph.containsQueue(nodes.get(i)))
				queues.add((AbstractBufferPipe) nodes.get(i));
		}
	}
	
	public Source getSource() {
		return source;
	}

	public Sink getSink() {
		return sink;
	}
	
	public AbstractBufferPipe[] getQueues() { 
		return queues.toArray(new AbstractBufferPipe[queues.size()]);
	}

	public boolean containsNode(Node node) {
		return nodes.contains(node);
	}

	public Graph getGraph() {
		return graph;
	}

	public boolean isSource(Node node) {
		return source.equals(node);
	}

	public boolean isSink(Node node) {
		return sink.equals(node);
	}

	public boolean isControllable(Node node) {
		return queues.contains(node);
	}

	public String toString() {
		StringBuffer res = new StringBuffer("Source:\n\t");
		res.append(getSource() + "\n");
		res.append("Sink:\n\t");
		res.append(getSink() + "\n");
		return res.toString();
	}
    
    /**
     * creates an iterator over this path
     * @param forward <code>true</code>: iterator from source to sink, <code> false </code> iterator from sink to source
     * @return iterator
     */

	public Iterator<Node> iterator(boolean forward) {
		return (forward) ? new Iterator<Node>() {

			private int pos = 0;
			private Node next = null;

			public boolean hasNext() {
				if (next!=null)
					pos = nodes.indexOf(next)+1;
				return pos < nodes.size();
			}

			public Node next() {
				next = nodes.get(pos);
				return next;
			}

			public void remove() {
				if (next!=null) {
					nodes.remove(next);
					pos--;
					if (graph.containsNode(next))
						graph.removeNode(next);
					queues.remove(next);
					source = (Source) nodes.get(0);
					sink = (Sink) nodes.get(nodes.size()-1);
					next = null;
				}
			}

		} : new Iterator<Node>() {

			private int pos = nodes.size()-1;
			private Node next = null;

			public boolean hasNext() {
				return pos >= 0;
			}

			public Node next() {
				next = nodes.get(pos--);
				return next;
			}

			public void remove() {
				if (next!=null) {
					nodes.remove(next);
					pos++;
					if (graph.containsNode(next))
						graph.removeNode(next);
					queues.remove(next);
					source = (Source) nodes.get(0);
					sink = (Sink) nodes.get(nodes.size()-1);
				}
			}
		};
	}
	
	protected synchronized void removeNode(Node node) {
		queues.remove(node);
		nodes.remove(node);
	}
	
	protected synchronized void addNode(Node after, Node node) {
		int index = nodes.indexOf(after);
		nodes.add(((index>0) ? index:0), node);
		if (graph.containsQueue(node)) {
			int qIndex = 0;
			int i;
			for (i=0;qIndex<index && i<queues.size();i++) {
				qIndex = nodes.indexOf(queues.get(i)); 
			}
			queues.add(i, (AbstractBufferPipe) node);
		}
	}
	
	public synchronized int length() {
		return nodes.size();
	}
}

