/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.queryGraph;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.concurrent.locks.ReentrantReadWriteLock.WriteLock;

import xxl.core.collections.queues.StackQueue;
import xxl.core.pipes.operators.Pipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.scheduler.Scheduler;
import xxl.core.pipes.scheduler.Strategy;
import xxl.core.pipes.scheduler.StrategyScheduler;
import xxl.core.pipes.scheduler.TimeSliceScheduler;
import xxl.core.pipes.sinks.AbstractSink;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;

/**
 *
 */
public class QueryExecutor extends Processor {
	
	public enum SchedulerType {
		/**
		 * Timeslice-Scheduler
		 */
		TIMESLICE,
		/**
		 * StrategyScheduler
		 */
		STRATEGY,
		
		/**
		 * Querybased
		 */
		QUERYBASED
	}
    
    
	
	/**
	 * @param st Schedulertype
	 * @param param Strategy if st==Strategy
	 * @return new Scheduler
	 */
	public static Scheduler createScheduler(SchedulerType st,Object param) {
		Scheduler sched = null;
		if (st == SchedulerType.TIMESLICE) 
			sched = new TimeSliceScheduler(true);
		else if (st == SchedulerType.STRATEGY)
			sched = new StrategyScheduler((Strategy) param);
		return sched;
	}
	
    public static final QueryExecutor DEFAULT_INSTANCE = new QueryExecutor();
    
    protected ArrayList<AbstractSink<?>> queries;
    protected ArrayList<Processor> processors;
    
    private boolean schedulerStarted=false;
    
    public QueryExecutor(long period) {
        super(period);
        this.setName("QueryExecutor");
        this.queries = new ArrayList<AbstractSink<?>>();
        this.processors = new ArrayList<Processor>();
    }
    
    public QueryExecutor() {
        this(500);
    }
    
    public synchronized void setQueryScheduler(SchedulerType st,Object param) throws InstantiationException{
    	if (scheduler!=null)
    		throw new InstantiationException("You can't change the scheduler during runtime!!!!");
    	scheduler = createScheduler(st, param);
    }
    
    public synchronized void setQueryScheduler(Scheduler sched) throws InstantiationException {
    	if (scheduler!=null)
    		throw new InstantiationException("You can't change the scheduler during runtime!!!!");
    	scheduler = sched;
    }
            
    public synchronized void registerQuery(AbstractSink<?> query) {
        queries.add(query);
    }
    
    public synchronized void registerProcessor(Processor proc) throws InstantiationException{
    	registerProcessor(proc,true);
    }
    
    public synchronized void registerProcessor(Processor proc,boolean schedule) throws InstantiationException{
        processors.add(proc);
        if (schedule)
        	if (scheduler!=null)
        		proc.setScheduler(scheduler);
    }
    
    public synchronized void startQuery(AbstractSink<?> query) throws IllegalArgumentException {
    	if (getState().equals(State.NEW))
            this.start(); // starting thread
        if (queries.indexOf(query) < 0)
            throw new IllegalArgumentException("Query not registered.");
        query.openAllSources();
    }
    
    public synchronized void startProcessor(Processor proc) {
        if (getState().equals(State.NEW))
            this.start();
        if (processors.indexOf(proc) < 0)
            throw new IllegalArgumentException("Processor not registered.");
        if (scheduler != null && !schedulerStarted && scheduler.isScheduled(proc)) {
        	((Thread) scheduler).start();
        	schedulerStarted = true;
        }
        proc.start();
    }
    
    public synchronized void registerAndStartQuery(AbstractSink<?> query) {
        registerQuery(query);
        startQuery(query);
    }
    
    public synchronized void registerAndStartProcessor(Processor proc,boolean schedule) throws InstantiationException {
        registerProcessor(proc,schedule);
        startProcessor(proc);
    }
    
    public synchronized void startAllQueries() {
        if (queries.size() > 0) {
        	WriteLock graphWLock = Graph.DEFAULT_INSTANCE.WLock; // assumption: all queries belong to the same graph
            graphWLock.lock();
            try {
                for (int i = 0, n = queries.size(); i < n; i++)
                    startQuery(queries.get(i));
            }
            finally {
                graphWLock.unlock();
            }
        }
    }
    
    public synchronized void startAllProcessors() {
        for (int i = 0; i < processors.size(); i++)
                startProcessor(processors.get(i));
    }
    
    public synchronized void stopQuery(AbstractSink<?> query) throws IllegalArgumentException {
        int index = queries.indexOf(query);
        if (index < 0)
            throw new IllegalArgumentException("Query not registered.");
        query.closeAllSources();
    }
    
    public synchronized void deregisterQuery(AbstractSink<?> query) throws IllegalArgumentException {
        int index = queries.indexOf(query);
        if (index < 0)
            throw new IllegalArgumentException("Query not registered.");
        Graph graph = query.graph;
        graph.WLock.lock();
        try{
	        StackQueue<Node> connectedNodes = new StackQueue<Node>();
	        connectedNodes.enqueue(query);
	        
	        while(! connectedNodes.isEmpty()){
	        	Node node = connectedNodes.dequeue();
	        	if(node instanceof Pipe){
	        		Pipe pipe = (Pipe) node;
	        		if(pipe.getNoOfSinks() == 0){
	            		for(int i = pipe.getNoOfSources()-1; i>=0 ;i--){
	            			Source source = pipe.getSource(i);
	            			connectedNodes.enqueue(source);
	            			Pipes.disconnect(source, pipe, source.getSinkID(pipe));	            			
	            		}
	        		}
	        	} else
	        	if( node instanceof Source){
	        		Source source = (Source) node;
	        		if(source.getNoOfSinks() == 0){
	        			source.close();
	        		}
	        	} else 
	        	if( node instanceof Sink){
	        		Sink sink = (Sink) node;
	        		for(int i = sink.getNoOfSources()-1; i>=0 ;i--){
	        			Source source = sink.getSource(i);
	        			Pipes.disconnect(source, sink, source.getSinkID(sink));	        				        			
	        			connectedNodes.enqueue(source);
	        		}
	        		queries.remove(sink);
	        	}
//	        	node.getGraph().removeNode(node);
	        }
        } finally{
        	graph.WLock.unlock();
        }
    }
    
    public synchronized void terminateProcessor(Processor proc) {
        int index = processors.indexOf(proc);
        if (index < 0)
            throw new IllegalArgumentException("Processor not registered.");
        proc.terminate();
        processors.remove(index);   
    }
    
    public synchronized void stopAllQueries() {
        if (queries.size() > 0) {
            WriteLock graphWLock = Graph.DEFAULT_INSTANCE.WLock; // assumption: all queries belong to the same graph
            graphWLock.lock();
            try {
            	Iterator<AbstractSink<?>> it = queries.iterator();
                while (it.hasNext()) 
                	stopQuery(it.next());
            }
            finally {
                graphWLock.unlock();
            }
            queries.clear(); // remove all queries
        }
    }
    
    public synchronized void startAll() {
        startAllQueries();
        startAllProcessors();
    }
    
    public synchronized void stopAllProcessors() {
        for (int i=processors.size()-1;i>-1;i--) {
            terminateProcessor(processors.get(i));
        }
    }
        
    @Override
    protected void onTermination() {
    	stopAllProcessors();
    	stopAllQueries();
    }
    
    @Override
    public synchronized void process() {
        boolean allDone = true;
        for (int i = 0, n = queries.size(); i < n; i++) {
            AbstractSink<?> sink = queries.get(i);
            allDone &= sink.isDone();
        }
        if (allDone) {
            terminate = true;
        }
    }
       
}
