/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.D_ESTIMATION;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.operatorMemoryReleaseCapacity;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import xxl.core.comparators.InverseComparator;
import xxl.core.functions.Function;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.Pipe;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.AbstractSource;
import xxl.core.pipes.sources.Source;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * class represents a virtual operator inside a graph.
 * 
 * There are different methods implemented to create vos.
 * 
 * These are:
 * - Chain-partition (construction of vo is similar to construct a chain) (@see 'Operator scheduling in data stream systems')
 * - Simplified-Segment-partition (construction of vo is similar to construction of simplified segment (@see 'Scheduling strategies
 * for processing continuous queries over streams' )
 * - first-fit-decreasing-heuristic
 * - MINSETCOVER-Heuristic
 * 
 */
public class Partition {

    /**
     * graph, this partition belongs to.
     */
	private Graph graph;

	/**
     * nodes of partition
	 */
	private HashSet<Pipe> nodes;

	/**
	  * new instance 
     * @param graph graph this partition belongs to
	 */
	public Partition(Graph gr) {
		this(gr, new HashSet<Pipe>());
	}

    /**
     * new instance
     * @param graph graph this partition bleongs to
     * @param n nodes of partition
     */
	public Partition(Graph gr, HashSet<Pipe> n) {
		this.graph = gr;
		nodes = n;
	}

    /**
     * adds pipe to partition
     * 
     * @param pipe pipe to add
     */
	public void addNode(Pipe pipe) {
		nodes.add(pipe);
	}

    /**
     * removes node from partition
     * @param pipe pipe to remove
     */
	public void removeNode(Pipe pipe) {
		nodes.remove(pipe);
	}

    /**
     * @return nodes of the partition
     */
	public Pipe[] getNodes() {
		return nodes.toArray(new Pipe[nodes.size()]);
	}

    /**
     * adds all nodes of partition p to this partition
     * 
     * @param p partition to merge
     */
	public void merge(Partition p) {
		nodes.addAll(p.nodes);
	}

    /**
     * @param pipe pipe to check
     * @return ? contains the given pipe
     */
	public boolean containsNode(Pipe pipe) {
		return nodes.contains(pipe);
	}

    /**
     * calculates a graph partitioning based on the chain strategy.
     * each path will be partitioned into chains. The unnecessary bufferpipes will be removed.
     * The resulting graph ist partitioned into a set of virtual operators
     *
     * @param graph considered graph
     * @return list of virtual operators
     * @throws InstantiationException
     */
	public static Partition[] chainPartition(Graph graph)
			throws InstantiationException {
		ArrayList<Node> sdop = new ArrayList<Node>();
		HashMap<Node, Double> nodeToDerivate = new HashMap<Node, Double>();
		System.out.println("Creating ops");
		OperatorPath[] paths = graph.getOperatorPaths();
		System.out.println("finished ops creating");
		for (int ind = 0; ind < paths.length; ind++) {
			OperatorPath path = paths[ind];
			int nrOfPipes = path.length() - path.getQueues().length;
			double[][] steepestDerivates = new double[nrOfPipes][4];
			int index = 0;
			double selectivity = 1d;
			double ttp = 0d;
			Iterator<Node> it = path.iterator(true);
			while (it.hasNext()) {
				Node node = it.next();
				if (!path.isControllable(node)) {
					steepestDerivates[index][1] = ttp;
					steepestDerivates[index][2] = selectivity;
					if (index > 0) {
						calcDerivate(steepestDerivates, index, selectivity, ttp);
					}
					index++;
					MetaDataHandler mdh1 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT);
					MetaDataHandler mdh2 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(INPUT_RATE);
					MetaDataHandler mdh3 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(INPUT_OUTPUT_RATIO);
					double cm = (Double) mdh1.getMetaData();
					double ir = (Double) mdh2.getMetaData();
					double sel = (Double) mdh3.getMetaData();
					ttp += (ir != 0) ? cm / ir : cm;
					selectivity = (sel != 0) ? selectivity * (1 / sel)
							: selectivity;
				}
			}
			// values for last entry
			calcDerivate(steepestDerivates, index, 0.0, ttp);
			int opCount = 0;
			int sdopIndex = 0;
			steepestDerivates[steepestDerivates.length - 1][3] = steepestDerivates.length;
			AbstractBufferPipe bp = null;
			Node prevNode = null;
			it = path.iterator(true);
			while (it.hasNext()) {
				Node node = it.next();
				if (path.isControllable(node))
					bp = (AbstractBufferPipe) node;
				else {
					if (opCount == sdopIndex) {
						if (bp == null && prevNode != null) {
							Source source = (Source) prevNode;
							Sink sink = (Sink) node;
							boolean isDone = (graph.containsPipe(prevNode) && ((Pipe) prevNode)
									.isDone())
									|| (graph.isInitialSource(prevNode)
											&& !graph.containsPipe(prevNode) && ((AbstractSource) prevNode)
											.getProcessor().isTerminated());
							if (!sink.isDone() && !isDone) {
								bp = graph.addBuffer(source, sink, false);
								if (sdop.contains(bp)) {
									double der = nodeToDerivate.get(bp);
									if (der > steepestDerivates[sdopIndex][0])
										nodeToDerivate
												.put(
														bp,
														steepestDerivates[sdopIndex][0]);
								} else {
									sdop.add(bp);
									nodeToDerivate.put(bp,
											steepestDerivates[sdopIndex][0]);
								}
							}
						} else if (bp != null) {
							if (!bp.isFinished()) {
								sdop.add(bp);
								nodeToDerivate.put(bp,
										steepestDerivates[sdopIndex][0]);
							}
						}
						if (graph.containsPipe(node))
							sdopIndex = (int) steepestDerivates[sdopIndex][3];
						bp = null;
					} else {
						if (bp != null) {
							Source source = (Source) prevNode;
							Sink sink = (Sink) node;
							if (!sink.isDone()) {
								graph.removeBuffer(source, bp, sink);
							}
							bp = null;
						}
					}
					if (graph.containsPipe(node))
						opCount++;
					prevNode = node;
					bp = null;
				}
			}
		}
		return getPartitions(graph);
	}

    /**
     * calculates the derivates
     * 
     * @param steepestDerivates array of data
     * @param index endpoint for calculation
     * @param selectivity correlated selectivity of this point
     * @param ttp correlated time to process of this point
     */
	private static void calcDerivate(double[][] steepestDerivates, int index,
			double selectivity, double ttp) {
		for (int i = 0; i < index; i++) {
			double derivate = (selectivity - steepestDerivates[i][2])
					/ (ttp - steepestDerivates[i][1]);
			if (derivate <= steepestDerivates[i][0]) {
				steepestDerivates[i][0] = derivate;
				steepestDerivates[i][3] = index;
			} else if (steepestDerivates[i][3] == 0)
				steepestDerivates[i][3] = i + 1;
		}
	}

    /**
     * calculates a graph partitioning based on the simplified segment strategy.
     * each path will be partitioned into at most two segments. The unnecessary bufferpipes will be removed.
     * The resulting graph ist partitioned into a set of virtual operators
     *
     * 
     * @param graph considered graph
     * @param gamma used similarity factor
     * @return list of virtual operators
     * @throws InstantiationException
     */
	public static Partition[] simplifiedSegmentPartition(Graph graph,
			double gamma) throws InstantiationException {
		Node[] nodes = graph.getPipes();
		for (Node node : nodes) {
			node.getMetaDataManagement().include(COST_MEASUREMENT);
			node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
			node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
		}
		OperatorPath[] operatorPaths = graph.getOperatorPaths();
		ArrayList<ArrayList<Node>> segments = new ArrayList<ArrayList<Node>>();
		HashMap<ArrayList<Node>, Integer> segmentToPath = new HashMap<ArrayList<Node>, Integer>();
		ArrayList<Node> segment = null;
		for (int i = 0; i < operatorPaths.length; i++) {
			segment = new ArrayList<Node>();
			OperatorPath path = operatorPaths[i];
			Iterator<Node> it = path.iterator(true);
			AbstractBufferPipe bp = null;
			boolean secondSegment = false;
			double prev = 0.0;
			Node prevNode = null;
			while (it.hasNext()) {
				Node node = it.next();
				if (path.isControllable(node))
					bp = (AbstractBufferPipe) node;
				else if (graph.containsPipe(node)
						|| !graph.isInitialSource(node)) {
					double mem_rl_cap = 0.0;
					if (!secondSegment)
						mem_rl_cap = operatorMemoryReleaseCapacity.invoke(node);
					if (prevNode != null) {
						if (!secondSegment) {
							if (prevNode != null && (mem_rl_cap / prev) < gamma) {
								secondSegment = true;
								segmentToPath.put(segment, i);
								segments.add(segment);
								segment = new ArrayList<Node>();
							}
						}
						if (bp != null && segment.size() == 0) {
							segment.add(bp);
						} else if (bp != null) {
							Source source = (Source) prevNode;
							Sink sink = (Sink) node;
							if (!sink.isDone()) {
								graph.removeBuffer(source, bp, sink);
							}
						} else if (segment.size() == 0) {
							Source source = (Source) prevNode;
							Sink sink = (Sink) node;
							if (!sink.isDone()) {
								bp = graph.addBuffer(source, sink, false);
								segment.add(bp);
							}
						}
						bp = null;
					}
					segment.add(node);
					prevNode = node;
					prev = mem_rl_cap;
				}
				if (segment.size() == 0 && bp != null) {
					segment.add(bp);
					bp = null;
				}
			}
			segmentToPath.put(segment, i);
			segments.add(segment);

		}
		int lastIndex = -1;
		ArrayList<Node> last = null;
		for (int i = 0; i < segments.size(); i++) {
			segment = segments.get(i);
			int pathIndex = segmentToPath.get(segment);
			// Start eines Pfades
			// bei dem Pfadwechsel geht man davon aus,
			// das man immer wieder einen Pfad neu berechnen muss
			if (!operatorPaths[pathIndex].isControllable(segment.get(0))) {
				if (pathIndex == lastIndex) {
					Source opSrc = (Source) last.get(last.size() - 1);
					Sink opGoal = (Sink) segment.get(0);
					int indexOut = opSrc.getSinkID(opGoal);
					int indexIn = opGoal.getSourceID(opSrc);
					AbstractBufferPipe bp = graph.addBuffer(opSrc, opGoal,false);
//					bp.getMetaDataManagement().include(D_ESTIMATION);
					segment.add(0, bp);
				} else {
					Sink sink = (Sink) segment.get(0);
					for (int j = 0; j < sink.getNoOfSources(); j++) {
						Source src = sink.getSource(j);
						if (!graph.containsQueue(src)) {
							int indexOut = src.getSinkID(sink);
							AbstractBufferPipe bp = graph.addBuffer(src, sink,true);
//							bp.getMetaDataManagement().include(D_ESTIMATION);
							if (j > 0) {
								ArrayList<Node> clone = (ArrayList<Node>) segment
										.clone();
								clone.add(0, src.getSink(indexOut));
								segments.add(clone);
							} else
								segment.add(0, src.getSink(indexOut));
						} else {
							if (j > 0) {
								ArrayList<Node> clone = (ArrayList<Node>) segment
										.clone();
								clone.add(0, sink.getSource(j));
								segments.add(clone);
							} else
								segment.add(0, sink.getSource(j));
						}
					}
				}
			}
			lastIndex = pathIndex;
			last = segment;
		}	
		return getPartitions(graph);
	}

    /**
     * function, which determines for a sink, which sources should be connected directly.
     * uses first fit decreasing heuristic based upon nodes capacity and nodes cost 
     */
	private static Function<Sink,ArrayList<Source>> ff = null;
	
    /**
     * creates a fit function for the given graph
     * 
     * @param graph considered graph
     * @return  function, which determines for a sink, which sources should be connected directly.
     * uses first fit decreasing heuristic based upon nodes capacity and nodes cost 
     */
	public static Function<Sink, ArrayList<Source>> fitFunction(final Graph graph) {
		if (ff==null) 
			ff = new Function<Sink, ArrayList<Source>>() {
			
				HashMap<Sink, Double> costMap;

				public ArrayList<Source> invoke(Sink sink) {
					if (graph.containsNode(sink)) {
					if (costMap == null)
						costMap = new HashMap<Sink, Double>();
					ArrayList<Source> result = new ArrayList<Source>();
					HashMap<Source, Double> cMap = new HashMap<Source, Double>();
					Comparator<Source> comp = new InverseComparator<Source>(
							new HashComparator<Source, Double>(cMap));
					ArrayList<Source> sources = new ArrayList<Source>();
					Double d = 0.0;
					for (int i = 0; i < sink.getNoOfSources(); i++) {
						Source source = sink.getSource(i);
						if (!costMap.containsKey(source)
								&& graph.containsNode(source)) {
							Partition.fitFunction(graph).invoke((Sink) source);
						}
						Double c = (graph.containsNode(source)) ? costMap.get(source) : 0.0;
						double dz = (Double) ((MetaDataHandler) source
								.getMetaDataManagement().getMetaData().get(
										D_ESTIMATION)).getMetaData();
						cMap.put(source, c);
						d += 1 / dz;
						sources.add(source);
					}
					d = 1d / d;
					
					Collections.sort(sources, comp);

					Double cost = (Double) ((MetaDataHandler) sink
							.getMetaDataManagement().getMetaData().get(
									COST_MEASUREMENT)).getMetaData();

					for (Source source : sources) {	
						Double c = cMap.get(source);
						if (cost + c < d) {

							result.add(source);
							cost += c;
						}
					}
					costMap.put(sink, cost);
					return result;
					}
					else
						return new ArrayList<Source>();
				}
		};
		return ff;
	}

    /**
     * function, which calcualtes nodecapacity for a given node
     */
	private static Function<Sink, Double> cncFct = null;

    /**
     * function, which calcualtes nodecapacity for a given vo
     */
	private static Function<Partition, Double> cpcFct = null;

    /**
     * function, which calcualtes nodecapacity for a given graph
     */
	private static Function<Graph, Double> cgcFct = null;

    /**
     * resets functions to use them for a new graph
     * TODO make it better ! sinnvoller machen
     */
	public static void reset() {
		cncFct = null;
		cpcFct = null;
		cgcFct = null;
		ff = null;
	}
	
    /**
     * creates a function for the given graph, which calculates the capacity for a node
     * 
     * @param graph considered graph
     * @return function, which calcualtes nodecapacity for a given node
     */
	public static Function<Sink, Double> calcNodeCapacity(final Graph graph) {
		if (cncFct == null)
			cncFct = new Function<Sink, Double>() {
				private HashMap<Sink, Double> costMap;

				public Double invoke(Sink sink) {
					if (graph.containsQueue(sink)) {
						costMap.put(sink, 0.0);
						return 0.0;
					}
					if (costMap == null)
						costMap = new HashMap<Sink, Double>();
					Double d = 0.0;
					Double cost = 0.0;
					for (int i = 0; i < sink.getNoOfSources(); i++) {
						Source source = sink.getSource(i);	
						double dz = (Double) ((MetaDataHandler) source
									.getMetaDataManagement().getMetaData().get(
											D_ESTIMATION)).getMetaData();
						d += 1d / dz;
						if (!costMap.containsKey(source))
							cncFct.invoke((Sink) source);
						cost += costMap.get(source);
						
					}
					cost += (Double) ((MetaDataHandler) sink
							.getMetaDataManagement().getMetaData().get(
									COST_MEASUREMENT)).getMetaData();
					costMap.put(sink, cost);
					return 1d/d - cost; 
				}
			};
		return cncFct;
	}

    /**
     * creates a function for the given graph, which calculates the capacity for a vo
     * 
     * @param graph considered graph
     * @return function, which calcualtes nodecapacity for a given vo
     */
	public static Function<Partition, Double> calcPartitionCapacity(
			final Graph graph) {
		if (cpcFct == null)
			cpcFct = new Function<Partition, Double>() {
				public Double invoke(Partition partition) {
					double res = 0.0;
					for (Node node : partition.nodes) {
						if(!graph.containsQueue(node)) 
							res += calcNodeCapacity(graph).invoke((Sink) node);
					}						
					return res;
				}
			};
		return cpcFct;
	}

    /**
     * creates a function for the given graph, which calculates the capacity for a graph
     * 
     * @param graph considered graph
     * @return function, which calcualtes nodecapacity for a given graph
     */
	public static Function<Graph, Double> calcGraphCapacity(final Graph graph) {
		if (cgcFct == null)
			cgcFct = new Function<Graph, Double>() {
				public Double invoke(Graph sub) {
					double res = 0.0;
					for (Node node : sub.getNodes())
						res += calcNodeCapacity(graph).invoke((Sink) node);
					return res;
				}
			};
		return cgcFct;
	}

    /**
     * partitions the graph using a given strategy (implemented function: first fit decreasing)
     * @param graph considered graph
     * @param strategyFct function used to determine the set of sources, which are connected directly to a sink
     * @return list of virtual operators
     */
	public static Partition[] buPartition(Graph graph,
			Function<Sink, ArrayList<Source>> strategyFct) {
		Node[] nodes = graph.getNodes();
		for (Node node : nodes) {
			node.getMetaDataManagement().include(D_ESTIMATION);
			node.getMetaDataManagement().include(COST_MEASUREMENT);
		}
		Source[] sources = graph.getSources();
		List<Source> done = new ArrayList<Source>();
		List<Source> todo = new ArrayList<Source>();
		for (Source node : sources) {
			todo.add(node);
		}
		while (!todo.isEmpty()) {
			Source node = todo.remove(0);
			if (!done.contains(node)) {
				done.add(node);
				for (int i = 0; i < node.getNoOfSinks(); i++) {
					if (!todo.contains(node.getSink(i))
							&& graph.containsPipe(node.getSink(i)))
						todo.add((Source) node.getSink(i));
				}
				if (graph.containsPipe(node) && !graph.isInitialSource(node) && !graph.containsQueue(node)) {
					Pipe pipe = (Pipe) node;					
					List directInt = strategyFct.invoke(pipe);
					for (int i = 0; i < pipe.getNoOfSources(); i++) {
						if (!directInt.contains(pipe.getSource(i))) {
							graph.addBuffer(pipe.getSource(i), pipe, false);
							pipe.getSource(i).getMetaDataManagement().include(COST_MEASUREMENT);
							pipe.getSource(i).getMetaDataManagement().include(D_ESTIMATION);
						}
					}					
				}
			}
		}
		return getPartitions(graph);
	}

    /**
     * partitions the graph using the MINSETCOVER-approach
     * @param graph considered graph
     * @return list of virtual operators
     */
	public static Partition[] approxPartition(Graph graph) {
		Node[] allNodes = graph.getNodes();
		int nrNodes = graph.nodeCount() - graph.getQueues().length;
		List<Graph> allSubGraphs = getSubgraphs(graph);
		HashMap<Node, Graph> subgraphMap = new HashMap<Node, Graph>();
		Set<Graph> result = new HashSet<Graph>();
		Graph finished = new Graph();
		while (finished.nodeCount() < nrNodes) {
			int mostBids = -1;
			Graph toAdd = null;
			int done = nrNodes;
			double minCap = Double.MAX_VALUE;
			for (Graph sub : allSubGraphs) {
				Graph temp = graph.intersection(graph.diffGraph(sub));
				if (temp.nodeCount() >= mostBids) {
					int card = temp.nodeCount();
					if (temp.nodeCount() > mostBids) {
						mostBids = card;
						toAdd = sub;
					} else { // optimierung: falls mehrere teilmengen, whle
						// diejenige mit mglichst wenig berschneidung
						temp = sub.intersection(graph);
						if (temp.nodeCount() <= done) {
							if (temp.nodeCount() < done) {
								mostBids = card;
								toAdd = sub;
								done = temp.nodeCount();
							} else if (calcGraphCapacity(graph).invoke(sub) < minCap) {
								mostBids = card;
								toAdd = sub;
								done = temp.nodeCount();
								minCap = calcGraphCapacity(graph).invoke(sub);
							}
						}
					}
				}
			}
			result.add(toAdd);
			for (Node node : toAdd.getNodes())
				subgraphMap.put(node, toAdd);
			finished = finished.union(toAdd);
			ArrayList<Graph> toRemove = new ArrayList<Graph>();
			for (Graph sub : allSubGraphs)
				if (sub.intersection(finished).nodeCount()>0)
					toRemove.add(sub);
			allSubGraphs.removeAll(toRemove);
		}
		for (int i = 0; i < allNodes.length; i++)
			for (int j = i + 1; j < allNodes.length; j++) {
				if (graph.isAdjacent(allNodes[i], allNodes[j])
						&& subgraphMap.get(allNodes[i]) != subgraphMap
								.get(allNodes[j]))
					graph.addBuffer((Source) allNodes[i], (Sink) allNodes[j],
							false);
				else if (graph.isAdjacent(allNodes[j], allNodes[i])
						&& subgraphMap.get(allNodes[i]) != subgraphMap
								.get(allNodes[j]))
					graph.addBuffer((Source) allNodes[j], (Sink) allNodes[i],
							false);

			}
		return getPartitions(graph);
	}

	/**
	 * calculates a list of all connected subgraphs of a subgraph with nonnegative capacity 
     * 
     * !!!! really expensive !!!!!!
	 */
	private static List<Graph> getSubgraphs(Graph graph) {
		Node[] allNodes = graph.getNodes();
		List<Graph> allSubGraphs = new ArrayList<Graph>();
		List<Graph> curStep = new ArrayList<Graph>();
		List<Graph> nextStep = new ArrayList<Graph>();

		// compute all subgraphs with non-negative capacity
		for (Node node : allNodes) {
			Graph graph1 = new Graph();
			Graph graph2 = new Graph();
			graph1.addNode(node);
			graph2.addNode(node);
			nextStep.add(graph1);
			allSubGraphs.add(graph2);
		}
		while (nextStep.size() > 0) {
			curStep = new ArrayList<Graph>(nextStep);
			nextStep.clear();
			for (int i = 0; i < curStep.size(); i++) {
				Graph subgraph = curStep.get(i);
				Set<Node> adjacentNodes = subgraph.getAdjacentNodes();
				for (Node node : adjacentNodes) {
					Graph temp = subgraph.copy();
					temp.addNode(node);
					if (!nextStep.contains(temp) && !allSubGraphs.contains(temp) ) {
						nextStep.add(temp);
						allSubGraphs.add(temp);
					}
				}
			}
		}
		return allSubGraphs;
	}

    /**
     * @return capacity of this vo
     */
	public double getCapacity() {
		return calcPartitionCapacity(graph).invoke(this);
	}

    /**
     * calculates all vos from a given graph
     * 
     * @param graph considered graph
     * @return list of vos
     */
	private static Partition[] getPartitions(Graph graph) {
		try {
			graph.getNodes()[0].getGraph().WLock.lock();
			HashMap<Node, Partition> partMap = new HashMap<Node, Partition>();
			OperatorPath[] paths = graph.getOperatorPaths();
			Partition p = null;
			for (OperatorPath path : paths) {
				Iterator<Node> it = path.iterator(true);
				while (it.hasNext()) {
					Node node = it.next();
					if (partMap.containsKey(node) && p != null) {
						Partition part = partMap.get(node);
						part.merge(p);
						p = part;
					} else if (partMap.containsKey(node)) {
						p = partMap.get(node);
					} else if (graph.containsPipe(node)
							&& !graph.containsQueue(node) && p != null) {
						p.addNode((Pipe) node);
						partMap.put(node, p);
					} else if (graph.containsPipe(node)
							&& !graph.containsQueue(node)) {
						p = new Partition(graph);
						p.addNode((Pipe) node);
						partMap.put(node, p);
					} else if (graph.containsQueue(node)) {
						p = null;
					}
				}
			}
			Set<Partition> set = new HashSet<Partition>(partMap.values());
			return set.toArray(new Partition[set.size()]);
		} finally {
			graph.getNodes()[0].getGraph().WLock.unlock();
		}
	}

    /**
     * @return considered graph
     */
	public Graph getGraph() {
		return graph;
	}

}
