/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler;

import xxl.core.pipes.processors.Processor;

/**
 * A Scheduler for Processors. <code>SchedulableProcessor</code>s and 
 * <code>UrgentProcessor</code>s can be treated separately.
 * @param <T> 
 *
 * @since 1.1
 */
public interface Scheduler<T extends Processor> {
	
	/**
	 * When a Processor's <code>setScheduler<code>-Method is called,
	 * it will call this method to register itself to this scheduler.
	 * 
	 * @param proc the Processor to be scheduled
	 * @see xxl.core.pipes.processors.Processor#setScheduler(Scheduler)
	 */
	public boolean schedule(T proc);
	
	/**
	 * Causes this scheduler to schedule a Processor right after 
	 * another one (optional operation).
	 *
	 * @param proc the Processor to be scheduled
	 * @param predecessor the predecessing Processor
	 *
	 * @throws UnsupportedOperationException if this scheduler doesn't support this method
	 * @throws java.util.NoSuchElementException if predecesssor is not registered to this scheduler
	 */
	public void scheduleAfter(T proc, T predecessor);
	
	/**
	 * Causes this scheduler to stop scheduling a certain Processor.
	 *
	 * @param proc the Processor not to be scheduled anymore
	 */
	public boolean stopScheduling(T proc);

	/**
	 * Called by a Processor when it is started.
	 *
	 * @param proc the started Processor
	 */
	public void reportStart(T proc);
	
	/**
	 * Called by a Processor when it is paused until wakeup, i.e. when it's 
	 * <code>pauseUntilWakeUp<code>-Methode is called or its <code>pause</code>-Method
	 * is called with <code>0</code> as argument.
	 *
	 * @param proc the blocked Processor
	 * @see xxl.core.pipes.processors.Processor#pauseUntilWakeUp()
	 * @see xxl.core.pipes.processors.Processor#pause(long)
	 */
	public void reportBlocked(T proc);

	/**
	 * Called by a Processor when it's <code>pause</code>-Method is called 
	 * with a positive argument.
	 *
	 * @param proc the blocked Processor
	 * @see xxl.core.pipes.processors.Processor#pause(long)
	 */
	public void reportBlocked(T proc, long time);
	
	/**
	 * Called by a Processor when it's woken up by calling its
	 * <code>wakeUp</code>-Method after it had been put to sleep.
	 *
	 * @param proc the now unblocked Processor
	 * @see xxl.core.pipes.processors.Processor#wakeUp()
	 */
	public void reportUnblocked(T proc);

	/**
	 * Called by a Processor when its <code>setSchedulingPriority</code>-Method
	 * has been called.
	 *
	 * @param proc the processor
	 * @see xxl.core.pipes.processors.Processor#setSchedulingPriority(int)
	 */
	public void priorityChange(T proc, int priority);
	
	/**
	 * Called by a Processor directly before it terminates.
	 *
	 * @param proc the terminating Processor
	 */
	public void reportTermination(T proc);
	
	/**
	 * 
	 * @param proc the processor
	 * @return true if this given processor is scheduled by this scheduler
	 */
	public boolean isScheduled(T proc);	
}

