/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.ALL_INPUT_OBJECT_SIZES;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.OUTPUT_OBJECT_SIZE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.D_ESTIMATION;

import java.util.Iterator;

import xxl.core.functions.Function;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.Pipe;
import xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.sources.Source;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * a collection of functions and constants,which are required by some scheduling strategies
 * 
 */
public class SchedulingFunctions {

	/**
	 * size of a window for AggregationMetaDataIdentifier
     * max number of considered measurepoints 
	 */
	public final static int windowSize = 20;
	
	/**
	 * minimal number of considered measurepoints inside a window
	 */
	public final static int minimumResultWindow = 1;
	
	/**
	 * calculates the average of the input size over all input streams 
     * assumption: the input_size of a single stream is constant
	 */
	public static Function<Node, Double> averageInputObjectSize = new Function<Node, Double>() {
		public Double invoke(Node node) {
			if (node.getMetaDataManagement().getMetaData()==null || !node.getMetaDataManagement().getMetaData().contains(ALL_INPUT_OBJECT_SIZES))
				return Double.NaN;
			Long[] inputSizes = (Long[]) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(ALL_INPUT_OBJECT_SIZES)).getMetaData();
			double avgInputSize = 0.0;
			for (int i = 0; i < inputSizes.length; i++)
				avgInputSize += inputSizes[i].doubleValue();
			avgInputSize /= inputSizes.length;
			return avgInputSize;
		}
	};

	/**
	 * calculates the operator processing capacity: 
     * this is the number of tuples, that can be processed within one timeunit
	 * 
	 * CPOi = 1 / COST_MEASUREMENT / (AVG_INPUT_RATE*UpdatePeriod)
	 * 
	 * used MetaData: AVG_INPUT_RATE,COST_MEASUREMENT
	 */
	public static Function<Node, Double> operatorProcessingCapacity = new Function<Node, Double>() {
		public Double invoke(Node node) {
			return 1 / processingTimePerTuple.invoke(node);
		}
	};

	/**
	 * calculates the memory release capacity of an operator 
     * this is the number of memory units that can be relesed within one time unit by one operator
	 * 
	 * required MetaData = requiredMetaData cpoi +
	 * selectivity_measurement,all_input_object_sizes,output_object_size
	 * 
	 * CMOi = CPOi*(INPUT_TUPLE_SIZE-OUTPUT_TUPLE_SIZE*Selectivity)
	 */
	public static Function<Node, Double> operatorMemoryReleaseCapacity = new Function<Node, Double>() {
		public Double invoke(Node node) {
			double cpoi = operatorProcessingCapacity.invoke(node);
			double inputSize = averageInputObjectSize.invoke(node);
			boolean test = (cpoi != Double.NaN
					&& inputSize != Double.NaN
					&& node.getMetaDataManagement().getMetaData()!=null
					&& node.getMetaDataManagement().getMetaData().contains(OUTPUT_OBJECT_SIZE) 
					&& node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO)));
			if (!test)
				return Double.NaN;
			Double selectivity = (Double) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))).getMetaData();
			double outputSize = ((Long) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(OUTPUT_OBJECT_SIZE)).getMetaData()).doubleValue();
			selectivity = (selectivity==null) ? 1.0 : selectivity;
			return cpoi * (inputSize - outputSize * 1/selectivity);
		}
	};

	/**
	 * calculates the processing capacity of an operator path processing
	 * capacity is defined as in operatorProcessingCapacity
	 * 
	 * required MetaData required MetaData cpoi + selectivity_measurement
	 * 
	 */
	public static Function<OperatorPath, Double> operatorPathProcessingCapacity = new Function<OperatorPath, Double>() {
		public Double invoke(OperatorPath op) {
			Iterator<Node> it = op.iterator(true);
			double res = 0.0;
			double sel = 1.0;
			while (it.hasNext()) {
				Node node = it.next();
				if (!op.isControllable(node)) {
					double cpoi = operatorProcessingCapacity.invoke(node);
					if (	node.getMetaDataManagement().getMetaData()==null
							|| !node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))
							|| cpoi == Double.NaN)
						return Double.NaN;
					res += (sel / cpoi);
					Double s = (Double) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))).getMetaData(); 
					sel *=  (s == null || Double.isNaN(s)) ? 1.0 : 1/s;
				}
			}
			return (res == 0.0) ? Double.NaN : 1 / res;
		}
	};

	/**
	 * calculates the memory release capacity of an operator Path
	 * 
	 * requred MetaData requiredMetaData of cpp + required MetaData
	 * avgInputTupleSize
	 */
	public static Function<OperatorPath, Double> operatorPathMemoryReleaseCapacity = new Function<OperatorPath, Double>() {
		public Double invoke(OperatorPath op) {
			double cpp = operatorPathProcessingCapacity.invoke(op);
			Node node = op.getSource();
			double inputSize = averageInputObjectSize.invoke(node);
			if (cpp == Double.NaN || inputSize == Double.NaN) {
				return Double.NaN;
			}
			return cpp * inputSize;
		}
	};
	
	/**
	 * calculates the processing time per tuple
     * 
     * required metadata: costmeasurement + inputrate
	 */
	public static Function<Node,Double> processingTimePerTuple = new Function<Node,Double>() {
		public Double invoke(Node node) {
			if (node.getMetaDataManagement().getMetaData()==null || !node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)) 
					|| !node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT)) 
				return Double.NaN;
			double cm = (Double) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT)).getMetaData();
			Double air = (Double) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE))).getMetaData();
			if (air==null)
				return Double.NaN;
			return (cm / (air*((AbstractPipeMetaDataManagement) node.getMetaDataManagement()).getUpdatePeriod()));
		}
		
	};
	
    /**
     * calculates the capacity of a virtual operator (=partition)
     * 
     * required metadata :required metadata calcNodeCapacity
     */
	public static Function<Partition,Double> calcPartitionCapacity = new Function<Partition,Double>() {
		public Double invoke(Partition partition) {
			double res = 0.0;
			for (Pipe pipe : partition.getNodes()) 
				res += calcNodeCapacity.invoke(pipe);
			return res;
		}
	};
	
    /**
     * calculates the capacity of a whole graph
     * 
     * required metadata :required metadata calcNodeCapacity
     */
	public static Function<Graph,Double> calcGraphCapacity = new Function<Graph,Double>() {
		public Double invoke(Graph graph) {
			double res = 0.0;
			Pipe[] pipes = graph.getPipes();
			for (Pipe pipe : pipes) 
				res += calcNodeCapacity.invoke(pipe);
			return res;
		}
	};
	
    /**
     * calculates the capacity of a single node
     * 
     * required metadata cost_measurement+d_estimstion
     */
	public static Function<Pipe,Double> calcNodeCapacity = new Function<Pipe,Double>() {
		public Double invoke(Pipe pipe) {
			if (!pipe.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT))
				return Double.NaN;
			double d = 0.0;
			double c = 0.0;
			for (int i=0;i<pipe.getNoOfSources();i++) {
				Source source = pipe.getSource(i);
				if (source.getMetaDataManagement().getMetaData().contains(D_ESTIMATION) && source.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT)) {
					d += 1 / (Double) ((MetaDataHandler) source.getMetaDataManagement().getMetaData().get(D_ESTIMATION)).getMetaData();
					c += (Double) ((MetaDataHandler) source.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT)).getMetaData();
				}
				else
					return Double.NaN;
			}
			d = 1/d;
			c += (Double) ((MetaDataHandler) pipe.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT)).getMetaData();
			return d - c;
		}
	};
}