/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler;

import java.util.Comparator;
import java.util.Random;

import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.operators.identities.TemporalBufferPipe;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.scheduler.strategies.BATABUStrategy;
import xxl.core.pipes.scheduler.strategies.BATANaiveStrategy;
import xxl.core.pipes.scheduler.strategies.BATATDStrategy;
import xxl.core.pipes.scheduler.strategies.BATATDWCStrategy;
import xxl.core.pipes.scheduler.strategies.ChainStrategy;
import xxl.core.pipes.scheduler.strategies.ComparatorStrategy;
import xxl.core.pipes.scheduler.strategies.FCFSStrategy;
import xxl.core.pipes.scheduler.strategies.GreedyStrategy;
import xxl.core.pipes.scheduler.strategies.LazyBATANaiveStrategy;
import xxl.core.pipes.scheduler.strategies.MTQStrategy;
import xxl.core.pipes.scheduler.strategies.MinCostStrategy;
import xxl.core.pipes.scheduler.strategies.MinLatencyStrategy;
import xxl.core.pipes.scheduler.strategies.MinMemoryStrategy;
import xxl.core.pipes.scheduler.strategies.MixedStrategy;
import xxl.core.pipes.scheduler.strategies.OptimizedChainStrategy;
import xxl.core.pipes.scheduler.strategies.PathCapacityStrategy;
import xxl.core.pipes.scheduler.strategies.PreRateBasedStrategy;
import xxl.core.pipes.scheduler.strategies.RandomStrategy;
import xxl.core.pipes.scheduler.strategies.RoundRobinStrategy;
import xxl.core.pipes.scheduler.strategies.SegmentStrategy;
import xxl.core.pipes.scheduler.strategies.ShortestPathStrategy;
import xxl.core.pipes.scheduler.strategies.SimplifiedSegmentStrategy;
import xxl.core.pipes.scheduler.strategies.ThreadedShortestPathStrategy;
import xxl.core.pipes.scheduler.strategies.ThresholdStrategy;



/**
 * factory for different types of operator-scheduling strategies
 * 
 */
public class StrategyFactory {

	/**
     * creates a new strategy 
     * @param str StrategyType (@see xxl.core.pipes.scheduler.Strategies)
     * @param graph considered graph
     * @param period updateperiod of processors
     * @return strategy of given type over considered graph
     * @throws InstantiationException
	 */
	public static Strategy createStrategy(Strategies str, Graph graph, long period) throws InstantiationException{
		return createStrategy(str, graph, period, null);
	}
	
    /**
     * creates a new strategy
     * @param str
     * @param str StrategyType (@see xxl.core.pipes.scheduler.Strategies)
     * @param graph considered graph
     * @param period updateperiod of processors
     * @param parameter object array of necessary parameters      
     * @return strategy of given type over considered graph
     * @throws InstantiationException
     */
    public static Strategy createStrategy(Strategies str, Graph graph, long period,
            Object[] parameter) throws InstantiationException {
        if (str == Strategies.ROUNDROBIN)
            return createRoundRobin(graph);
        else if (str == Strategies.FCFS) {
            return createFIFO(graph);
        } else if (str == Strategies.COMPARATOR) {
            if (parameter == null || parameter.length < 1)
                throw new InstantiationException(
                        "It is necessary to assign a Comparator as parameter to use Comparator-Scheduling. There were no parameters assigned");
            if (!(parameter[0] instanceof Comparator))
                throw new InstantiationError(
                        "It is necessary to assign a Comparator as parameter to use Comparator-Scheduling. Wrong parameter-type for this Strategy: "
                                + parameter[0].getClass());
            Comparator c = (Comparator) parameter[0];
            return createComparator(graph, c);
        } else if (str == Strategies.SHORTESTPATH) {
            return createShortestPath(graph);
        } else if (str == Strategies.CHAIN) {
            return createChain(graph, period);
        } else if (str == Strategies.BATANAIVE) {
            return createBATANaive(graph);
        } else if (str == Strategies.TSP) {
            return createTSP(graph, period);
        } else if (str == Strategies.MIXED) {
            if (parameter == null || parameter.length < 1)
                throw new InstantiationException(
                        "It is necessary to assign a double as parameter to use Mixed-Scheduling. There were no parameters assigned");
            if (!(parameter[0] instanceof Double))
                throw new InstantiationError(
                        "It is necessary to assign a double as parameter to use Mixed-Scheduling. Wrong parameter-type for this Strategy: "
                                + parameter[0].getClass());
            double gamma = (Double) parameter[0];
            return createMixed(graph, gamma, period);
        } else if (str == Strategies.MINCOST) {
            return createMinCost(graph);
        } else if (str == Strategies.MINLATENCY) {
            return createMinLatency(graph, period);
        } else if (str == Strategies.MINMEMORY) {
            return createMinMemory(graph, 500);
        } else if (str == Strategies.PATHCAPACITY) {
            return createPathCapacity(graph, period);
        } else if (str == Strategies.GREEDY) {
            return createGreedy(graph, period);
        } else if (str == Strategies.SEGMENT) {
            return createSegment(graph, period);
        } else if (str == Strategies.SIMPLIFIEDSEGMENT) {
            return createSimplifiedSegment(graph, 0.75, period);
        } else if (str == Strategies.OPTIMIZEDCHAIN) {
            return createOptimizedChain(graph, period);
        } else if (str == Strategies.MTQ) {
            return createMostTuplesInQueue(graph);
        } else if (str == Strategies.BATABU) {
        	return createBATABU(graph, period);
        } else if (str == Strategies.BATATD) {
        	return createBATATD(graph, period);
        } else if (str == Strategies.BATATDWC) {
                return createBATATDWC(graph, period);
        } else if (str == Strategies.PREEMPTIVERATEBASED) {
        	return createPreemptiveRatebased(graph, period);
        } else if (str == Strategies.THRESHOLD) {
            if (parameter == null || parameter.length < 2)
                throw new InstantiationException(
                        "It is necessary to assign two parameters to use Threshold-Scheduling. There were no parameters assigned");
            if (!(parameter[0] instanceof Long))
                throw new InstantiationError(
                        "It is necessary to assign a maxQueueSize as parameter to use Threshold-Scheduling. Wrong parameter-type for this Strategy: "
                                + parameter[0].getClass());
            if (!(parameter[1] instanceof Double))
                throw new InstantiationError(
                        "It is necessary to assign parameter beta as parameter to use Threshold-Scheduling. Wrong parameter-type for this Strategy: "
                                + parameter[1].getClass());
            double beta  = (Double) parameter[1];
            long maxMem  = (Long) parameter[0];
            return createThreshold(graph, maxMem, beta, period);
        } else
            return createRandom(graph);
    }

    /**
     * @param graph considered graph
     * @return roundrobin strategy
     */
    private static Strategy createRoundRobin(Graph graph) {
        return new RoundRobinStrategy(graph.getQueues());
    }

    /**
     * @param graph considered graph
     * @return fcfs-strategy
     */
    private static Strategy createFIFO(Graph graph) {
    	if (!checkTemporal(graph))
    		throw new IllegalArgumentException("Query graph not applicable. Use temporal nodes");
        return new FCFSStrategy(graph);
    }

    /**
     * @param graph considered graph
     * @param c used comparator (must be parameter[0]
     * @return comparator strategy
     */
    private static Strategy createComparator(Graph graph,
            Comparator<Controllable> c) {
        return new ComparatorStrategy(graph.getQueues(), c);
    }

    /**
     * @param graph considered graph
     * @return random strategy
     */
    private static Strategy createRandom(Graph graph) {
        return new RandomStrategy(graph.getQueues(), new Random());
    }

    /**
     * @param graph considered graph
     * @return SHORTESTPATH-strategy
     */
    private static Strategy createShortestPath(Graph graph) {
        return new ShortestPathStrategy(graph);
    }

    /**
     * @param graph considered graph
     * @return shortestpath strategy with threads
     */
    private static Strategy createTSP(Graph graph, long period) {
        return new ThreadedShortestPathStrategy(graph, period);
    }

    /**
     * @param graph considered graph
     * @param period updateperiod for processor in ms
     * @return chainstrategy
     */
    private static Strategy createChain(Graph graph, long period) {
        return new ChainStrategy(graph, period);
    }

    /**
     * @param graph considered graph
     * @param gamma gamma (=parameter[0])
     * @param periodupdateperiod for processor in ms
     * @return Mixed-Chain-Strategy
     */
    private static Strategy createMixed(Graph graph, double gamma, long period) {
        return new MixedStrategy(graph, period, gamma);
    }

    /**
     * @param g considered graph
     * @return naive bata strategy
     * @throws IllegalArgumentException
     */
    private static Strategy createBATANaive(Graph g) throws IllegalArgumentException {
    	if (!checkTemporal(g))
    		throw new IllegalArgumentException("Query graph not applicable. Use temporal nodes");
        return new BATANaiveStrategy(g);
    }
    
    /**
     * @param g considered graph
     * @param period updateperiod of processor in ms
     * @return pathbased batastrategy
     * @throws IllegalArgumentException
     */
    private static Strategy createBATABU(Graph g, long period) throws IllegalArgumentException{
    	if (!checkTemporal(g))
    		throw new IllegalArgumentException("Query graph not applicable. Use temporal nodes");
    	return new BATABUStrategy(g,period);
    }
    
    /**
     * @param g considered graph
     * @param period updateperiod of processor in ms
     * @return batastrategy (countermechansim)
     * @throws IllegalArgumentException
     */
    private static Strategy createBATATD(Graph g, long period) throws IllegalArgumentException{
    	if (!checkTemporal(g))
    		throw new IllegalArgumentException("Query graph not applicable. Use temporal nodes");
    	return new BATATDStrategy(g,period);
    }
    
    /**
     * @param g considered graph
     * @param period updateperiod of processor in ms
     * @return batastrategy without counter
     * @throws IllegalArgumentException
     */
    private static Strategy createBATATDWC(Graph g, long period) throws IllegalArgumentException{
    	if (!checkTemporal(g))
    		throw new IllegalArgumentException("Query graph not applicable. Use temporal nodes");
        return new BATATDWCStrategy(g,period);
    }
    
    /**
     * @param g considered graph
     * @param period uodateperiod of processor in ms
     * @return naive bata strategy with threads
     * @throws IllegalArgumentException
     */
    private static Strategy createLazyBataNaive(Graph g, long period) throws IllegalArgumentException{
    	if (!checkTemporal(g))
    		throw new IllegalArgumentException("Query graph not applicable. Use temporal nodes");
    	return new LazyBATANaiveStrategy(g,period);
    }

    /**
     * @param g considered graph
     * @param period updateperiod of processor in ms
     * @return greedy strategy
     */
    private static Strategy createGreedy(Graph g, long period) {
        return new GreedyStrategy(g, period);
    }

    /**
     * @param g considered graph
     * @return mincost strategy
     */
    private static Strategy createMinCost(Graph g) {
        return new MinCostStrategy(g);
    }

    /**
     * @param g considered graph
     * @param period updateperiod of processors in ms
     * @return minlatency strategy
     */
    private static Strategy createMinLatency(Graph g, long period) {
    	return new MinLatencyStrategy(g,period);
    }

    /**
     * @param g considered graph
     * @param period updateperiod of processor in ms
     * @return minmemory strategy
     */
    private static Strategy createMinMemory(Graph g, long period) {
        return new MinMemoryStrategy(g,period);
    }

    /**
     * @param g considered graph
     * @param period update period of processor in ms
     * @return pathcapacity strategy
     */
    private static Strategy createPathCapacity(Graph g, long period) {
        return new PathCapacityStrategy(g, period);
    }

    /**
     * @param g considered graph
     * @param period updateperiod of processor in ms
     * @return segment strategy
     */
    private static Strategy createSegment(Graph g, long period) {
        return new SegmentStrategy(g, period);
    }

    /**
     * @param g considered graph
     * @param gamma similarity factor
     * @param period update period of processor in ms
     * @return simplifed segemtn strategy
     */
    private static Strategy createSimplifiedSegment(Graph g, double gamma,
            long period) {
        return new SimplifiedSegmentStrategy(g, gamma, period);
    }

    /**
     * @param g considered graph
     * @param period update period of processor in ms
     * @return optimized chain strategy
     */
    private static Strategy createOptimizedChain(Graph g, long period) {
        return new OptimizedChainStrategy(g, period);
    }

    /**
     * @param g considered graph
     * @return MTQ strategy
     */
    private static Strategy createMostTuplesInQueue(Graph g) {
        return new MTQStrategy(g);
    }
    
    /**
     * @param g considered graph
     * @param period update period of processor in ms
     * @return preemptive rate based strategy
     */
    private static Strategy createPreemptiveRatebased(Graph g, long period) {
    	return new PreRateBasedStrategy(g,period);
    }
    
    
    /**
     * @param g considered graph
     * @param mqs available memory size for queues parameter[0]
     * @param beta safety-factor parameter[1]
     * @param period update period for processor in ms
     * @return
     */
    private static Strategy createThreshold(Graph g, long mqs, double beta, long period) {
    	return new ThresholdStrategy(g,mqs,beta,period);
    }
   
    /**
     * @param graph considered graph
     * @return ? bufferpipes in graph store temporal objects
     */
    private static boolean checkTemporal(Graph graph) {
    	AbstractBufferPipe[] queues = graph.getQueues();
    	for (AbstractBufferPipe queue : queues)
    		if (!(queue instanceof TemporalBufferPipe))
    			return false;
    	return true;
    }
}