/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler;

import xxl.core.pipes.processors.Processor;

/**
 * A <code>Processor</code> that is used to control several
 * <code>StrategyInput</code>s by means of a <code>Strategy</code>.
 * While running, the <code>ControlProcessor</code> retrieves
 * <code>StrategyInput</code>s by calling the <code>next()</code> method
 * of the <code>Strategy</code>. Then, the <code>performOperation()</code> is called
 * on the retrieved <code>StrategyInput</code>.
 *
 * @see Strategy
 * @since 1.1
 */
public class StrategyProcessor<T extends Controllable> extends Processor {
	
	/**
	 * The <code>Strategy</code> used by this processor.
	 */
	protected Strategy<T> strategy = null;

	protected long updateStrategyPeriod;
	protected long lastStrategyUpdate;
	
	protected T next = null;

	/**
	 * Constructs a new <code>ControlProcessor</code> using the defined
	 * <code>Strategy</code>.
	 *
	 * @param strategy the strategy to be used by the processor.
	 * @param period
	 * @param updateStrategyPeriod
	 */
	public StrategyProcessor(Strategy<T> strategy, long period, long updateStrategyPeriod) {
		super(period);
		this.strategy = strategy;
		this.updateStrategyPeriod = updateStrategyPeriod;
	}

	public synchronized void setSchedulingStrategy(Strategy<T> schedulingStrategy) {
		this.strategy = schedulingStrategy;
	}
	
	public synchronized Strategy<T> getStrategy() {
		return strategy;
	}

	/**
	 * Retrieves the next <code>Controllable</code> by means of the currently used
	 * <code>Strategy</code> and calls <code>execute</code> on it.
	 */
	@Override
	public void process() {
		long now = System.currentTimeMillis();
		boolean test = (now - lastStrategyUpdate > updateStrategyPeriod) || (next!=null && !next.isReady());
		if (test) {
			if (strategy.getNoOfControllables() == 0) {
				terminate();
				return;
			}
			next = strategy.computeNext() ? strategy.next() : null;
			lastStrategyUpdate = now;
		}
		if (next != null) {
			if (next.isFinished()) {
				strategy.deregister(next);
				lastStrategyUpdate = 0; // cause strategy check
				return;
			}
			next.execute();
		}
	}

}

