/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler;

import java.util.HashSet;
import java.util.NoSuchElementException;

import sun.reflect.generics.reflectiveObjects.NotImplementedException;
import xxl.core.pipes.processors.Processor;

/**
 * A implementation of <code>Scheduler</code> using a <code>Strategy</code>.
 * @since 1.1
 */
public class StrategyScheduler<T extends ControllableStrategyProcessor> extends Thread implements Scheduler<T> {
  	
  	/**
  	 * The strategy used by this scheduler
  	 */
  	protected Strategy<T> strategy;
  	
  	/**
  	 * The time each Processor will be scheduled for.
  	 */
  	protected long schedulingTime = 50;
  	
  	private final Object WAIT = new Object();
  	
  	private T activeProc = null;
  	
  	protected HashSet<T> scheduledProcessors;
  	
  	/**
  	 * Constructs a new scheduler that uses the given strategy.
  	 *
  	 * @param strategy the Strategy used by the new constructed scheduler.
  	 */
  	public StrategyScheduler(Strategy<T> strategy) {
  		this.strategy = strategy;
  		this.scheduledProcessors = new HashSet<T>();
  	}
  	  	
  	/**
  	 * Sets the time that each processor will be scheduled for.
  	 *
  	 * @param time the scheduling time in milliseconds
  	 */
  	public void setSchedulingTime(long time) {
  		schedulingTime = time;
  	}
  	
  	/**
  	 * Returns the time that each processor will be scheduled for.
  	 * @return the time that each processor will be scheduled for.
  	 */
  	public long getSchedulingTime() {
  		return schedulingTime;
  	}
  	
	/**
	 * When a SchedulableProcessors <code>setScheduler<code>-Method is called
	 * it will call this method to register itself to this scheduler.
	 * 
	 * @param proc the SchedulableProcessor to be scheduled
	 * @throws IllegalArgumentException if proc does not implement StrategyInput
	 * @see Processor#setScheduler(Scheduler)
	 */
	public synchronized boolean schedule(T proc) throws IllegalArgumentException{
		try {
			if (scheduledProcessors.contains(proc))
	  			return true;
	  		else if (proc.getScheduler()!=null && proc.getScheduler()!=this) 
	  			return false;
			strategy.register(proc);
			scheduledProcessors.add(proc);
		} catch (ClassCastException e) {
			throw new IllegalArgumentException("Argument has to be an instance of ControllableStrategyProcessor.");
		}
		proc.setPriority(1);
		return true;
	}

	/**
	 * As the order of processors is determined by the strategy, 
	 * this method only throws an UnsupportedOperationException.
	 *
	 * @param proc the Processor to be scheduled
	 * @param predecessor the predecessing Processor
	 *
	 * @throws UnsupportedOperationException always
	 */
	public void scheduleAfter(T proc, T predecessor) {
		throw new UnsupportedOperationException("Scheduling order is determined by the strategy.");
	}
	
	/**
	 * Causes this scheduler to stop scheduling a certain Processor.
	 *
	 * @param proc the Processor not to be scheduled anymore
	 * @throws IllegalArgumentException if proc does not implement StrategyInput
	 */
	public synchronized boolean stopScheduling(T proc) throws IllegalArgumentException{
		try {
			if (!scheduledProcessors.contains(proc))
	  			return false;
			strategy.deregister(proc);
			scheduledProcessors.remove(proc);
		} catch (ClassCastException e) {
			throw new IllegalArgumentException("Argument has to be an instance of ControllableStrategyProcessor.");
		}
		
		if (proc == activeProc) 
			synchronized(WAIT) {
				WAIT.notify();
			}
		return true;
	}

	/**
	 * Called by a Processor when it is started.
	 *
	 * @param proc the started Processor
	 */
	public synchronized void reportStart(T proc) {
	}
	
	/**
	 * Called by a Processor when it is paused until wakeup, i.e. when it's 
	 * <code>pauseUntilWakeUp<code>-Methode is called or its <code>pause</code>-Method
	 * is called with <code>0</code> as argument.
	 *
	 * @param proc the blocked Processor
	 * @see Processor#pauseUntilWakeUp()
	 * @see Processor#pause(long)
	 */
	public synchronized void reportBlocked(T proc){
	}

	/**
	 * Called by a Processor when it's <code>pause</code>-Method is called 
	 * with a positive argument.
	 *
	 * @param proc the blocked Processor
	 * @param time
	 * @see Processor#pause(long)
	 */
	public synchronized void reportBlocked(T proc, long time) {
	
	}
	
	/**
	 * Called by a Processor when it's woken up by calling its
	 * <code>wakeUp</code>-Method after it had been put to sleep.
	 *
	 * @param proc the now unblocked Processor
	 * @see Processor#wakeUp()
	 */
	public synchronized void reportUnblocked(T proc) {
	
	}

	/**
	 * Called by a Processor when its <code>setSchedulingPriority</code>-Method
	 * has been called.
	 *
	 * @param proc the processor
	 * @see Processor#setSchedulingPriority(int)
	 */
	public synchronized void priorityChange(T proc,int priority) {
		throw new NotImplementedException();
	}

	/**
	 * Called by a Processor directly before it terminates.
	 *
	 * @param proc the terminating Processor
	 */
	public synchronized void reportTermination(T proc) {
		stopScheduling(proc);
	}
		
	public synchronized boolean isScheduled(T proc) {
		return scheduledProcessors.contains(proc);
	}
	
	/**
	 * In this method, all the scheduling work is done. 
   	 */
  	@Override
	public void run() {
  		while (true) {
  			synchronized(this) {
	  			T proc = null;
				try {
					proc = strategy.next();
				} catch (NoSuchElementException e) {
					e.printStackTrace(System.err);
				}
	  			if (proc != null) {
	  				if (activeProc != null) 
	  					activeProc.setPriority(1);
	  				activeProc = proc;
	  				activeProc.setPriority(7);
	  			}
  			}
  			try {
  				synchronized(WAIT) {
  					WAIT.wait(schedulingTime);
  				}
  			} catch (InterruptedException e) {}
  		}
  	}
}
