/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
 */

package xxl.core.pipes.scheduler;

import static xxl.core.pipes.scheduler.TimeSliceScheduler.MAX_SCHEDULING_PRIORITY;
import static xxl.core.pipes.scheduler.TimeSliceScheduler.MIN_SCHEDULING_PRIORITY;
import static xxl.core.pipes.scheduler.TimeSliceScheduler.NORM_SCHEDULING_PRIORITY;

import java.util.HashMap;
import java.util.NoSuchElementException;

import xxl.core.pipes.processors.CircularProcessorList;
import xxl.core.pipes.processors.Processor;

//erste Version: alle Threads in einer Liste (auch die ungestarteten)
//Vorteil: ordnungserhaltend
/**
 * This scheduler schedules Processors with a priority-based round robin strategy.
 * It simply loops over all the SchedulableProcessors and lets them run for
 * a time proportional to their scheduling priority (in fact, the scheduling 
 * priority in milliseconds) by giving them a high thread priority for that 
 * time.<br>
 * UrgentProcessors are being given an even higher thread priority and then 
 * are left to themselves.
 * @param <T> 
 *
 * @since 1.1
 */
public class TimeSliceScheduler<T extends Processor> extends Thread implements Scheduler<T> {
	
	/**
	 * The maximum value for priorities supported by this scheduler.
	 */
	public final static int MAX_SCHEDULING_PRIORITY = 1000;
	
	/**
	 * The minimum value for priorities supported by this scheduler.
	 */
	public final static int MIN_SCHEDULING_PRIORITY = 10;
	
	/**
	 * The normal value for priorities supported by this scheduler.
	 */
	public final static int NORM_SCHEDULING_PRIORITY = 50;
	
	
	protected CircularProcessorList procList = new CircularProcessorList();
	protected Processor activeProc;
	protected boolean terminatesIfListEmpty;
	
	private HashMap<T,Integer> priorityMap;

	private final Object WAIT = new Object(); 
	
	/**
	 * Constructs a new TimeSliceScheduler.
	 */
	public TimeSliceScheduler(boolean terminates) {
		priorityMap = new HashMap<T,Integer>();
		setPriority(MAX_PRIORITY);
		this.terminatesIfListEmpty = terminates;
	}
	
	/**
	 * Returns the maximum value for priorities supported by this scheduler
	 * @return the maximum value for priorities supported by this scheduler
	 */	
	public int getMaxSchedulingPriority() {
		return MAX_SCHEDULING_PRIORITY;
	}
	
	/**
	 * Returns the mimimum value for priorities supported by this scheduler
	 * @return the mimimum value for priorities supported by this scheduler
	 */	
	public int getMinSchedulingPriority() {
		return MIN_SCHEDULING_PRIORITY;
	}
	
	/**
	 * Returns the normal value for priorities supported by this scheduler
	 * @return the normal value for priorities supported by this scheduler
	 */	
	public int getNormSchedulingPriority() {
		return NORM_SCHEDULING_PRIORITY;
	}
	
	/**
	 * When a Processor's <code>setScheduler<code>-Method is called
	 * it will call this method to register itself to this scheduler.
	 * 
	 * @param proc the Processor to be scheduled
	 * @see Processor#setScheduler(Scheduler)
	 */
	public synchronized boolean schedule(T proc) {
		if (procList.contains(proc)) 
			return true;
		else if (proc.getScheduler()!=null && proc.getScheduler()!=this) 
  			return false;
		else if (!proc.isTerminated()) {
			proc.setPriority(MIN_PRIORITY);
			priorityMap.put(proc, getNormSchedulingPriority());
			procList.add(proc);
		}
		return false;
	}
		
	/**
	 * Causes this scheduler to schedule a Processor right after 
	 * another one.
	 *
	 * @param proc the Processor to be scheduled
	 * @param predecessor the predecessing Processor
	 *
	 * @throws NoSuchElementException if predecessor is not registered to this scheduler
	 */
	public synchronized void scheduleAfter(T proc, T predecessor) {
		if (!procList.contains(predecessor)) 
			throw new NoSuchElementException();
		if (!proc.isTerminated()) {
			if (procList.contains(proc)) {
				// unschn, wenn proc gerade luft, wird danach mit seinem neuen
				// Nachfolger fortgesetzt
				procList.remove(proc);
				procList.addAfter(proc, predecessor);
			}
			else {
				proc.setPriority(MIN_PRIORITY);
				procList.addAfter(proc, predecessor); // may throw NoSuchElementException
			}
		}
	}
	
	/**
	 * Causes this scheduler to stop scheduling a certain Processor.
	 *
	 * @param proc the Processor not to be scheduled anymore
	 */	
	public synchronized boolean stopScheduling(T proc) {
		if (!procList.contains(proc))
  			return false;
		procList.remove(proc);
		if (proc == activeProc) 
			synchronized(WAIT) {
				WAIT.notify();
			}
		return true;
	}
		
	/**
	 * Called by a Processor when it is started.
	 *
	 * @param proc the started Processor
	 */	
	public synchronized void reportStart(T proc) {
		
	}
		
	/**
	 * Called by a Processor when it is paused until wakeup, i.e. when it's 
	 * <code>pauseUntilWakeUp<code>-Methode is called or its <code>pause</code>-Method
	 * is called with <code>0</code> as argument.
	 *
	 * @param proc the blocked Processor
	 * @see Processor#pauseUntilWakeUp()
	 * @see Processor#pause(long)
	 */
	public synchronized void reportBlocked(T proc) {
		procList.setBlocked(proc);
		if (proc == activeProc) 
			synchronized(WAIT) {
				WAIT.notify();
			}
	}
		
	/**
	 * Called by a Processor when it's <code>pause</code>-Method is called 
	 * with a positive argument.
	 *
	 * @param proc the blocked Processor
	 * @param time
	 * @see Processor#pause(long)
	 */
	public synchronized void reportBlocked(T proc, long time) {
		procList.setBlockedUntil(proc, System.currentTimeMillis()+time);
		if (proc == activeProc) 
			synchronized(WAIT) {
				WAIT.notify();
			}
	}
		
	/**
	 * Called by a Processor when it's woken up by calling its
	 * <code>wakeUp</code>-Method after it had been put to sleep.
	 *
	 * @param proc the now unblocked Processor
	 * @see Processor#wakeUp()
	 */
	public synchronized void reportUnblocked(T proc) {
		procList.setUnblocked(proc);
		if (proc == activeProc) 
			synchronized(WAIT) {
				WAIT.notify();
			}
	}
			
	/**
	 * Called by a Processor directly before it terminates.
	 *
	 * @param proc the terminating Processor
	 */
	public synchronized void reportTermination(T proc) {
		stopScheduling(proc);
	}
	
	public synchronized boolean isScheduled(T proc) {
		return procList.contains(proc);
	}
	
	/**
	 * In this method, all the scheduling work is done. 
	 */
	@Override
	public void run() {
		long sleepTime = 100;	
		while (true) {
			synchronized(this) {
				if (terminatesIfListEmpty && procList.isEmpty())
					return;
				if (!procList.isEmpty()) {
					Processor proc = procList.getNextUnblocked();
					if (proc != null) {
//						System.out.println("activeProc: "+proc.getName()+" called "+counterMap.get(proc)+" times!");
						if (activeProc != null) {
							// System.out.println("Decrease priority of "+activeProc.getName());
							activeProc.setPriority(1);
						}
						activeProc = proc;
						if (activeProc != null) {
//							System.out.println("Increase priority of "+activeProc.getName());
							activeProc.setPriority(7);
							sleepTime = priorityMap.get(activeProc);
						}
					}
				}
			}
			try {
				synchronized(WAIT) {
					// System.out.println("Scheduler start waiting: "+sleepTime);
					WAIT.wait(sleepTime);
					// System.out.println("Scheduler stopped waiting");
				}
			} catch (InterruptedException e) {
				System.out.println("Interrupted");
			}
			
		}
	}
	
	/**
	 * Sets the scheduling priority of this Processor.
	 *
	 * @param newSchedulingPriority The new scheduling priority for this processor.
	 * 
	 * @exception IllegalArgumentException If the specified priority is not in the range
	 * 	given by <code>getMinSchedulingPriority()</code> and <code>getMaxSchedulingPriority()</code>
	 *	from the currently used scheduler. If no scheduler is used, all values are accepted.
	 */
	public void setSchedulingPriority(T proc, int newSchedulingPriority) {
		if ((newSchedulingPriority < getMinSchedulingPriority())
			|| (newSchedulingPriority
				> getMaxSchedulingPriority())) {
			throw new IllegalArgumentException(
				"Scheduling priority has to be between "
					+ getMinSchedulingPriority()
					+ " and "
					+ getMaxSchedulingPriority()
					+ ".");
		}
		if (newSchedulingPriority != priorityMap.get(proc)) 
			priorityMap.put(proc, newSchedulingPriority);
		
	}
	
	/**
	 * Returns the scheduling priority of this <code>Processor</code>.
	 *
	 * @return The scheduling priority of this <code>Processor</code>.
	 */
	public int getSchedulingPriority(T proc) {
		return priorityMap.get(proc);
	}

	/**
	 * @see xxl.core.pipes.scheduler.Scheduler#priorityChange(xxl.core.pipes.processors.Processor, int)
	 */
	public void priorityChange(T proc, int priority) {
		setSchedulingPriority(proc, priority);
	}
}
