/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler.strategies;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.Controllable;
import xxl.core.pipes.scheduler.Strategy;

/**
 * The AbstractPathStrategy is the base class for all path-based strategies.
 * This sort of strategy first chooses an operatorpath and then it chooses the bufferpipes 
 * on this path. Different strategies can be used for this. The default strategy to choose 
 * the order of operator execution is the FCFS-Strategy limited on one path. (At this time it 
 * is the only supported path interior strategy. Only nodes from the chosen path are 
 * selected by this strategy)
 *  
 * @param <T> objects from this type can be chosen.
 * 
 */

public abstract class AbstractPathStrategy<T extends AbstractBufferPipe> extends AbstractStrategy<PathStrategy<T>> {
	
	/**
	 * index of the next path to be chosen.
	 */
	private int nextPath;
	
    /**
     * list of path indices, ordered by priority  
     */
    protected ArrayList<Integer> priorities;
    
    /**
     * map: pathindex -> path interior strategy (default FCFS)
     */
    protected HashMap<Integer,PathStrategy<T>> pathToStrategy;
    
    /**
     * map: path interior strategy -> pathindex
     */
    protected HashMap<PathStrategy<T>,Integer> strategyToPath;
    
    /**
     * help variable to assign pathindices
     */
    private int pathCounter = 0;

    /**
     * considered operator paths
     */
    protected OperatorPath[] operatorPaths;
    
    /**
     * new instance: This calculates the paths of graph and registers them to the strategy.
     * If necessary a processor can be started, which can be used to gather information 
     * for this strategy. 
     * 
     * @param graph considered graph
     * @param period updateperiod of the Processor in ms (if one is started)
     */
    public AbstractPathStrategy(Graph graph, long period) {
        operatorPaths = graph.getOperatorPaths();
        controllables = new ArrayList<PathStrategy<T>>();
        pathToStrategy = new HashMap<Integer,PathStrategy<T>>();
        strategyToPath = new HashMap<PathStrategy<T>,Integer>();
        priorities = new ArrayList<Integer>();
        for (OperatorPath path : operatorPaths) 
        	register(new PathStrategy<T>(path));
        startProcessor(graph, period);
    }
    
    /**
     * You can start a strategy-specific processor within this method. 
     * This processor can  periodically  perform any action which is necessary to choose the next path.
     * The processor can use the whole graph. 
     * 
     * @param graph considered graph
     * @param period update period of the processor in ms
     */
    protected abstract void startProcessor(Graph graph, long period);

	/**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
     * 
     * This method chooses the next interior PathStrategy object. The next bufferpipe will be chosen and executed 
     * inside of this object.
	 */
	public boolean computeNext() {
		nextPath = 0;
		while (controllables.size() > 0) {
			int index = -1;
			synchronized (priorities) {
				if (nextPath == controllables.size())
					return false;
				index = priorities.get(nextPath);
			}
            next = pathToStrategy.get(index);
			if (next.isFinished())
				deregister(next);
			if (next.isReady()) {
				return true;
			}
			nextPath++;
		}
		return false;
	}
	
    /**
     * @see xxl.core.pipes.scheduler.AbstractStrategy#register(xxl.core.pipes.scheduler.Controllable)
     */
	public void register(PathStrategy<T> node) {
		super.register(node);
		synchronized(priorities) {
			pathToStrategy.put(pathCounter, node);
			strategyToPath.put(node, pathCounter);
			priorities.add(pathCounter++);
		}
	}
	
    /**
     * @see xxl.core.pipes.scheduler.AbstractStrategy#deregister(xxl.core.pipes.scheduler.Controllable)
     */
	public void deregister(PathStrategy<T> node) {
		super.deregister(node);
		synchronized (priorities) {
			int cnt = strategyToPath.get(node);
			strategyToPath.remove(node);
			pathToStrategy.remove(cnt);
			priorities.remove((Integer) cnt);
		}
	}
}

/**
 * These objects represent path interior strategies used by path-based operator-scheduling
 * strategies. These objects fullfil the interface <code>Controllable</code>. 
 * Arbitrary operator-scheduling strategies are possible. The FCFS-Strategy is the default strategy.
 * 
 * @param <T> Objects from this type can be chosen.
 * 
 */
class PathStrategy<T extends AbstractBufferPipe> implements Controllable {

    /**
     * next object to be chosen
     */
	private T next;
	
	/**
	 * path interior strategy
	 */
	private Strategy<T> internStrategy;
	
    /**
     * path on which this strategy is executed
     */
	private OperatorPath path;

	/**
	 * new instance: default strategy for path interior scheduling is used
	 * 
	 * @param path considered path
	 */
	public PathStrategy(OperatorPath p) {
		this(new OnePathFCFSStrategy<T>(p),p);
	}
	
	
	
	/**
	 * new instance: strategy for path interior scheduling must be chosen
     * 
	 * @param str path interior strategy  
     * @param p considered path
	 */
	public PathStrategy(Strategy<T> str, OperatorPath p) {
		internStrategy = str;
		path = p;
	}
    
    /**
     * @return considered path
     */
    public OperatorPath getPath() {
        return path;
    }

	/**
	 * @see xxl.core.pipes.scheduler.Controllable#execute()
	 */
	public void execute() {
		if (internStrategy.computeNext()) {
			next = internStrategy.next();
			next.execute();
			return;
		}
	}
	
	/**
	 * @see xxl.core.pipes.scheduler.Controllable#isFinished()
     * 
     * returns true, if all path interior controllables are finished
	 */
	public synchronized boolean isFinished() {
		boolean res = true;
		Iterator<T> it = internStrategy.getControllables();
		ArrayList<T> toDeregister = new ArrayList<T>();
		while (it.hasNext()) {
			T c = it.next();
			if (c.isFinished())
				toDeregister.add(c);
			res &= c.isFinished();
		}
		for (T c : toDeregister)
			internStrategy.deregister(c);
		return res;
	}

	/**
	 * @see xxl.core.pipes.scheduler.Controllable#isReady()
	 */
	public boolean isReady() {
		Iterator<T> it = internStrategy.getControllables();
		while (it.hasNext())
			if (it.next().isReady())
				return true;
		return false;
	}

}