/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.ArrayList;
import java.util.HashMap;

import xxl.core.functions.Function;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;


/**
 * This object represents an operator scheduling strategy. The query graph is a tree representing one
 * query. The order of bufferpipes will be determined by a tree-traversal. A function is
 * used to determine this traversal. Another function is used to compute the number of tuples handled 
 * in one pass. A processor can be used to adjust the traversal to a changing environment.
 * 
 * This approach is similar to the one described in 'Operator Scheduling for Data Stream Manager'.
 * There are three possibilites to traverse a query tree described.
 * 
 * -MinCost
 * -MinLatency
 * -MinMemory
 * 
 * These are all implemented within the PIPES subpackage.
 * 
 * Other traversals can be created by extending this class.
 * 
 * @param <T>  objects from this type can be chosen.
 */
public class AuroraStrategy<T extends AbstractBufferPipe> extends AbstractStrategy<T> {

    /**
     * This is a processor, which adjusts the traversal, when the environment changes. 
     * Thus adaptivity is achieved.   
     */
	abstract class Refresh extends Processor {

        
        /**
         * operator paths of the considered query tree
         */
		protected OperatorPath[] paths;

        /**
         * calculated traversal
         */
		protected T[] t;

        /**
         * nodes of the considered graph
         */
		protected Node[] ops;

        /**
         * map: bufferpipe -> downstream(bufferpipe)
         * This is the set of all bufferpipes, which lie on the appropriate operatorpath 
         * from this bufferpipe to the sink 
         */
		protected HashMap<T, ArrayList<T>> downStream;
        
        /**
         * function: query tree -> traversal
         */
        protected Function<Graph,T[]> getTraversals;
        
        /**
         * new instance: a new processor is created, 
         * a set of metadata is registered on the nodes
         * 
         * @param query considered query tree
         * @param period updateperiod of the Processor
         */
		protected Refresh(Graph query, long period) {
			super(period);
            paths = query.getOperatorPaths();
			downStream = new HashMap<T, ArrayList<T>>();
			ops = query.getPipes();
			for (Node node : ops) {
				node.getMetaDataManagement().include(COST_MEASUREMENT);
				node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
				node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
			}
		}

        /**
         * calculates the set of downstream operators for each bufferpipe
         */
		protected void calcDownStream() {
			for (int i = 0; i < paths.length; i++) {
				T[] queues = (T[]) paths[i].getQueues();
				ArrayList<T> help = new ArrayList<T>();
				for (int j = queues.length - 1; j > -1; j--) {
					if (!downStream.containsKey(queues[j])) {
						help.add(0, queues[j]);
                        downStream.put(queues[j],(ArrayList<T>) help.clone());
					} 
                    else 
						help = downStream.get(queues[j]);
				}
				help.clear();
			}
		}
		

        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public void process() {
			if (getTraversals != null) {
				getTraversals.invoke();
				synchronized(traversal) {
					traversal = (T[]) new AbstractBufferPipe[t.length];
					System.arraycopy(t, 0, traversal, 0, t.length);
				}
			}
			if (getNoOfControllables()==0)
				terminate();
		}

        /**
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
		@Override
		public void terminate() {
			for (Node node : ops) {
				if (node.getMetaDataManagement().getMetaData()!=null) {
					if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT))
						node.getMetaDataManagement().exclude(COST_MEASUREMENT);
					if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
						node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
					if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO)))
						node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
				}
			}
			super.terminate();
		}
	}

    /**
     * index of the next controllable of the traversal
     */
	protected int nextControllable;
	
	/**
	 * function: bufferpipe -> integer: calculates the train size. 
     * This is the number of tuples which will be processed in one run.
     * Default function is the batchsize of the bufferpipe.
	 */
	private Function<T,Integer> trainFunction;
	
	/**
     * traversal, which determines the order of the bufferpipes 
	 */
	protected T[] traversal;
	
	/**
     * new instance: sets the considered tree and the train function 
     *  
     * @param query considered query tree
     * @param tf train function
	 */
	public AuroraStrategy(Graph query, Function<T,Integer> tf) {
		super((T[]) query.getQueues());
		nextControllable = 0;
		setTrainFunction(tf);
	}
	
	/**
	 * new instance: sets the considered tree. The default train function is the batchsize
     * of the bufferpipes
     * 
	 * @param query considered query tree
	 */
    public AuroraStrategy(Graph query) {
		this(query, new Function<T,Integer>() {
			public Integer invoke(T arg) {
				return arg.getBatchSize();
			}
		});
    }
    
    /**
     * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
     */
	@Override
	public boolean computeNext() {
		synchronized(traversal) {
			for (int i=0;i<traversal.length;i++) {
				if (nextControllable>=traversal.length) 
					nextControllable = 0;
				next = traversal[nextControllable++];
				if (next!=null) {
					if (next.isFinished()) 
						deregister(next);
					else if (next.isReady()) {
						next.setBatchSize(trainFunction.invoke(next));
						return true;
					}
				}
			}
		}
		return false;
	}
	
	/**
     * sets the function, which calculates the number of tuples processed in one run. 
     * 
     * @param tf function : bufferpipe -> integer( only positive integers)
	 */
	public void setTrainFunction(Function<T,Integer> tf) {
		trainFunction = tf;
	}
}