/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.APPLICATION_TIME_ADVANCEMENT;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import xxl.core.comparators.InverseComparator;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.identities.TemporalBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * This is a path-based strategy. 
 * 
 * A reference point in time is determined, the anchor time.  
 * 
 * For each operator, it is determined the age of the current oldest tuple at each input of an operator. 
 * For each path, the difference between the anchor timestamp and the application timestamp of the tuple 
 * on the related input will be summed.
 * 
 * Paths with a high value are further back in time than pathes with a low value. This value is used 
 * to set the priority of the paths. 
 * 
 * For more information: @see: Master Thesis of Udo Wolske: Eine Evaluation von Strategien fr die Ablaufssteuerung 
 * fr die Datenstrominfrastruktur PIPES
 * 
 * The priorities are calculated periodically. Thus additional costs are avoided.
 */
public class BATABUStrategy extends AbstractPathStrategy<TemporalBufferPipe> {

	/**
     * This processor periodically calculates the priority of the path and adjusts them.
     * 
	 */
	public class Refresh extends Processor {

        /**
         * pipes of the considered graph
         */
		private Node[] ops;
        
        /**
         * considered graph
         */
        private Graph graph;

        /**
         * map: pathIndex -> priority
         * (accumulated amount of time the operators on the path are behind the anchor time) 
         */
		private HashMap<Integer, Double> pathToPriority;

        /**
         * all operators with multiple inputs (choice operators)
         */
		private Set<Node> choices;

        /**
         * comarator of path indices
         */
		private Comparator<Integer> comp;

        /**
         * virtual pipe, is connected to the path segments, which have no choice operators
         */
		private VirtualPipe virtualPipe;

        /**
         * map: node -> long array
         * each array entry represents how much the corresponding input is behind in application time
         */
		private HashMap<Node, long[]> advancementMap;
		
        /**
         * set of sinks of the graph
         */
		private Sink[] sinks;

		/**
		 * new instance
         * 
		 * @param gr considered graph
		 * @param period updateperiod of the Processor in ms
		 */
		public Refresh(Graph gr, long period) {
			super(period);
            graph = gr;
			setName("BATABUStrategy-REFRESH");
			ops = gr.getPipes();
			pathToPriority = new HashMap<Integer, Double>();
			advancementMap = new HashMap<Node, long[]>();
			comp = new InverseComparator<Integer>(
					new HashComparator<Integer, Double>(pathToPriority));
			sinks = gr.getSinks();
			virtualPipe = new VirtualPipe();
			choices = new HashSet<Node>();

			int ID = sinks.length + 1;
			for (Sink sink : sinks) {
				if (gr.containsPipe(sink))
					Pipes.connect((Source) sink, virtualPipe, ID++);
				else
					for (int i = 0; i < sink.getNoOfSources(); i++)
						Pipes.connect(sink.getSource(i), virtualPipe, ID++);
			}

			for (Node node : ops) {
				if (gr.containsPipe(node)) {
					Sink sink = (Sink) node;
					if (sink.getNoOfSources() > 1)
						choices.add(node);
				}
				if (choices.contains(node)) {
					node.getMetaDataManagement().include(
							APPLICATION_TIME_ADVANCEMENT);
				}
			}
			virtualPipe.getMetaDataManagement().include(
					APPLICATION_TIME_ADVANCEMENT);
			choices.add(virtualPipe);
		}

        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public void process() {
			boolean terminated = true;
			for (Sink sink : sinks)
				terminated &= sink.isDone();
			if (terminated)
				terminate();
			else {
				long anchor = 0;
				for (Node choice : choices) {
					Long[] appTimeAdv = (Long[]) ((MetaDataHandler) choice
							.getMetaDataManagement().getMetaData().get(
									APPLICATION_TIME_ADVANCEMENT))
							.getMetaData();
					for (Long l : appTimeAdv) {
						anchor = (anchor < l) ? l : anchor;
					}
					long[] behind = new long[appTimeAdv.length];
					for (int i = 0; i < behind.length; i++)
						behind[i] = anchor - appTimeAdv[i];
					advancementMap.put(choice, behind);
				}
				for (PathStrategy pathStrategy : controllables) {
					OperatorPath path = pathStrategy.getPath();
					double prio = 0;
					Iterator<Node> it = path.iterator(true);
					Node lastNode = null;
					while (it.hasNext()) {
						Node node = it.next();
						if (choices.contains(node) && lastNode != null) {
							long behind = advancementMap.get(node)[Pipes
									.getSourceIndex((Sink) node,
											(Source) lastNode)];
							prio += behind;
							
						}
                        if (graph.containsPipe(node))
                            lastNode = node;
					}
					prio += advancementMap.get(virtualPipe)[Pipes.getSourceIndex(virtualPipe, (Source) lastNode)];
					pathToPriority.put(strategyToPath.get(pathStrategy), prio);
				}
				synchronized (priorities) {
					Collections.sort(priorities, comp);
				}
			}
		}
		
        /**
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
		public void terminate() {
			for (Node node : choices) {
				node.getMetaDataManagement().exclude(APPLICATION_TIME_ADVANCEMENT);
			}
			super.terminate();
		}

	}
	
    /**
     * processor to achive adaptivity
     */
	private Refresh refresh;
	
	

	/**
     * new instance
     * 
     * @param g considered graph
     * @param period updateperiod of the Processor in ms
	 */
	public BATABUStrategy(Graph g, long period) {
		super(g, period);
	}

    /**
     * @see xxl.core.pipes.scheduler.strategies.AbstractPathStrategy#startProcessor(xxl.core.pipes.queryGraph.Graph, long)
     */
	protected void startProcessor(Graph graph, long period) {
		refresh = new Refresh(graph, period);
		refresh.start();
	}
}