/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.APPLICATION_TIME_ADVANCEMENT;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.D_ESTIMATION;
import static xxl.core.pipes.operators.identities.AbstractBufferPipe.AbstractBufferPipeMetaDataManagement.BUFFER_SIZE;
import static xxl.core.pipes.operators.identities.AbstractBufferPipe.AbstractBufferPipeMetaDataManagement.FIRST_ELEMENT_START;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import xxl.core.comparators.LexicographicalComparator;
import xxl.core.pipes.operators.AbstractPipe;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.operators.identities.TemporalBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * This strategy chooses the operator, which is furthest back in application time.
 * The inputs will be read relative to the rate of streams.
 * 
 * To avoid costly operations the calculations will be done periodically and priorities will be assigned.
 * 
 * To achieve an balanced operator a ticket mechanism is used. The number of tickets are based on the average 
 * rates of the streams
 * (the stream parameter d)
 * 
 * For more information: @see: Master Thesis of Udo Wolske: Eine Evaluation von Strategien fr die Ablaufssteuerung 
 * fr die Datenstrominfrastruktur PIPES
 */
public class BATATDStrategy extends AbstractStrategy<TemporalBufferPipe> {
	
    /**
     * periodically calculates how far back in application time the operators are 
     * and assigns an appropriate number of tickets.
     */
	class Refresh extends Processor {

        /**
         * pipes of the considered graph
         */
		private Node[] vertices;
        
        /**
         * set of choice operators (operators with multiple inputs)
         */
		private Set<Node> choices;
        
        /**
         * map: node -> long array
         * each array entry represents how much the corresponding input is behind application time
         * in regards to an anchortime
         */
		private HashMap<Node, Long[]> advancementMap;
        
        /**
         * map: bufferpipe -> number of tickets
         */
		private HashMap<TemporalBufferPipe,Long> cntMap;
        
        /**
         * operatorpaths of the considered graph
         */
		private OperatorPath[] paths;
        
        /**
         * map: bufferpipe -> priority 
         * (how much the given operator is behind application time in regards to an anchortime.)
         */
		private HashMap<TemporalBufferPipe, Long> prioMap;
        
        /**
         * map: bufferpipe -> position on a path
         */
		private HashMap<TemporalBufferPipe,Integer> posMap;
        
        /**
         * comparator of bufferpipes
         */
		private Comparator<TemporalBufferPipe> comp;
        
        /**
         * virtual pipe, is connected to the path segments, which have no choice operators
         */
		private VirtualPipe virtualPipe;
        
        /**
         * map: node -> a relative global position in application time
         */
		private HashMap<Node, Long> maxTs;
        
        /**
         * set of sinks of the graph
         */
		private Sink[] sinks;
        
        /**
         * considered graph
         */
        private Graph graph;

		/**
		 * new instance
         * 
		 * @param graph considered graph
		 * @param period updateperiod of the processor in ms
		 */
		public Refresh(Graph gr, long period) {
			super(period);
			setName("BATATDStrategy.Refresh");
            this.graph = gr;
			vertices = graph.getNodes();
			sinks = graph.getSinks();
			advancementMap = new HashMap<Node, Long[]>();
			maxTs = new HashMap<Node, Long>();
			virtualPipe = new VirtualPipe();
			int ID = sinks.length + 1;
			for (Sink sink : sinks) {
				if (graph.containsPipe(sink))
					Pipes.connect((Source) sink, virtualPipe, ID++);
				else
					for (int i = 0; i < sink.getNoOfSources(); i++)
						Pipes.connect(sink.getSource(i), virtualPipe, ID++);
			}
			paths = graph.getOperatorPaths();
			choices = new HashSet<Node>();
            choices.add(virtualPipe);
			prioMap = new HashMap<TemporalBufferPipe, Long>();
			posMap = new HashMap<TemporalBufferPipe,Integer>();
			cntMap = new HashMap<TemporalBufferPipe,Long>();
			HashSet<Sink> realChoiceSet = new HashSet<Sink>();
			Comparator<TemporalBufferPipe>[] ca = new Comparator[2];
			ca[0] = new HashComparator<TemporalBufferPipe,Long>(prioMap);
			ca[1] = new HashComparator<TemporalBufferPipe,Integer>(posMap);
			comp = new LexicographicalComparator<TemporalBufferPipe>(ca);
			for (Node node : vertices) {
				if (graph.containsPipe(node) || graph.isTerminalSink(node)) {
					Sink sink = (Sink) node;
					if (sink.getNoOfSources() > 1)
						choices.add(sink);
				}
				node.getMetaDataManagement().include(D_ESTIMATION);
				if (!graph.containsQueue(node)) {
					node.getMetaDataManagement().include(
							APPLICATION_TIME_ADVANCEMENT);
				} else
					node.getMetaDataManagement().include(BUFFER_SIZE);
			}
			virtualPipe.getMetaDataManagement().include(APPLICATION_TIME_ADVANCEMENT);
			for (OperatorPath path : paths) {
				Node lastChoice = virtualPipe;
				int maxIndex = path.getQueues().length;
				Iterator<Node> it = path.iterator(false);
				while (it.hasNext()) {
					Node node = it.next();
					if (choices.contains(node)) 
						lastChoice = node;
					if (path.isControllable(node)) {
						realChoiceSet.add((Sink) lastChoice);
						posMap.put((TemporalBufferPipe) node, maxIndex--);
					}
				}
			}
                
			choices.retainAll(realChoiceSet);
			
			// if the virtual pipe is not necessary, remove it
			if (!choices.contains(virtualPipe)) {
				ID = sinks.length + 1;
				for (Sink sink : sinks) {
					if (graph.containsPipe(sink)) {
						Pipes.disconnect((Source) sink, virtualPipe, ID++);
					} else
						for (int i = 0; i < sink.getNoOfSources(); i++)
							Pipes.disconnect(sink.getSource(i), virtualPipe,
									ID++);
				}
				virtualPipe = null;
			}
		}

        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public void process() {
			boolean terminated = true;
			for (Sink sink : sinks)
				terminated &= sink.isDone();
			if (terminated)
				terminate();
			else {
				for (Node choice : choices) {
					Long[] appTimeAdv = (Long[]) ((MetaDataHandler) choice.getMetaDataManagement().getMetaData().get(APPLICATION_TIME_ADVANCEMENT)).getMetaData();
					long max = 0;
					for (Long l : appTimeAdv) {
						max = (max < l) ? l : max;
					}
					Long[] behind = new Long[appTimeAdv.length];
					for (int i = 0; i < behind.length; i++) {
						behind[i] = max - appTimeAdv[i];
					}
					advancementMap.put(choice, behind);
					maxTs.put(choice, max);
				}
				for (OperatorPath path : paths) {
					Node lastChoice = virtualPipe;
					double d_est = 0.0;
					//double sel = 1.0;
					int index1 = -1;
					Iterator<Node> it = path.iterator(false);
					while (it.hasNext()) {
						Node node = it.next();
						// lege die Prioritt fest => lokaler relativer
						// Ausgleich, den diese BP am Knoten lastChoice bewirken
						// wrde
						if (path.isControllable(node)) {
							if (index1 == -1 && lastChoice != null) {
								index1 = Pipes.getSourceIndex((Sink) lastChoice,(Source) node);
								d_est = (Double) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(D_ESTIMATION)).getMetaData();
							}
							long max =  maxTs.get(lastChoice);
							Long behind = advancementMap.get(lastChoice)[index1];
							long min = max - behind;
							if (prioMap.get(node)==null || (prioMap.get(node)!=null && min < prioMap.get(node))) {
								prioMap.put((TemporalBufferPipe) node,min);
								long counter = (long)  (((behind / d_est)/((AbstractBufferPipe) node).getBatchSize())/* *sel*/);
								counter = (counter==0) ? 1 : counter;
								cntMap.put((TemporalBufferPipe) node, counter);
							}
						} 
						else {
							if (choices.contains(node)) {
								lastChoice = node;
								index1 = -1;
							}
							else if (lastChoice != null && graph.containsPipe(node)  && index1 == -1) { 
								index1 = Pipes.getSourceIndex((Sink) lastChoice,(Source) node);
								d_est = (Double) ((MetaDataHandler) node.getMetaDataManagement().getMetaData().get(D_ESTIMATION)).getMetaData();
							}
						}
					}
				}
				Collections.sort(controllables, comp);
				synchronized (queues) {
					queues.clear();
					queues.addAll(controllables);
					for (int i=0;i<queues.size();i++) 
						counterMap[i] = cntMap.get(queues.get(i));
				}
                prioMap.clear();
                advancementMap.clear();
			}
		}

        /**
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
		public void terminate() {
			for (Node node : vertices) {
				if (node.getMetaDataManagement().getMetaData().contains(
						D_ESTIMATION))
					node.getMetaDataManagement().exclude(D_ESTIMATION);
				if (node.getMetaDataManagement().getMetaData().contains(
						APPLICATION_TIME_ADVANCEMENT))
					node.getMetaDataManagement().exclude(
							APPLICATION_TIME_ADVANCEMENT);
				if (node.getMetaDataManagement().getMetaData().contains(
						BUFFER_SIZE))
					node.getMetaDataManagement().exclude(BUFFER_SIZE);
				if (node.getMetaDataManagement().getMetaData().contains(
						FIRST_ELEMENT_START))
					node.getMetaDataManagement().exclude(FIRST_ELEMENT_START);
			}
			super.terminate();
		}
	}
	
    /**
     * bufferpipes of the considered graph
     */
	private ArrayList<TemporalBufferPipe> queues;
    
    /**
     * chosen bufferpipe
     */
	private int index;
    
    /**
     * ticketcounter
     */
	private long count = 0;
    
    /**
     * Refresh-processor
     */
	private Refresh refresh;
    
    /**
     * map: index-> number of tickets for corresponding bufferpipe in queues
     */
	private long[] counterMap;

	/**
	 * new instance
     * 
	 * @param graph considered graph
	 * @param period updateperiod of reresh processor in ms
	 */
	public BATATDStrategy(Graph graph, long period) {
        AbstractBufferPipe[] nodes = graph.getQueues();
        for (int i = 0; i < nodes.length; i++)
            register((TemporalBufferPipe) nodes[i]);
		queues = new ArrayList<TemporalBufferPipe>();
		queues.addAll(controllables);
		counterMap = new long[queues.size()];
		refresh = new Refresh(graph, period);
		refresh.start();
	}
	
    /**
     * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
     */
	@Override
	public boolean computeNext() {
		//TODO synchronisierung von nodes im Strategieprozessor
		synchronized (queues) {
			synchronized (counterMap) {
				for (int i=0;i<queues.size();i++) {
					if (count<=0) {
						index = (index+1)%queues.size();
						next = queues.get(index);
						count = counterMap[index];
					}
					if (next.isFinished()) {
						count = 0;
						deregister(next);
					}
					else if (next.isReady()) {
						count--;
						return true;
					}
					else {
						count = 0;
					}
				}
			}
		}
		return false;
	}
}


/**
 * Virtual pipe.
 * Is used to simulate the advancement for pathsegments which leave the graph and
 * have no choice operators. If it is needed, it will be added to the graph.
 */
class VirtualPipe extends AbstractPipe {

    /**
     * @see xxl.core.pipes.operators.AbstractPipe#processObject(java.lang.Object, int)
     * 
     * does nothing
     */
	@Override
	public void processObject(Object o, int sourceID)
			throws IllegalArgumentException {
	}

}
