/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.APPLICATION_TIME_ADVANCEMENT;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import xxl.core.comparators.InverseComparator;
import xxl.core.comparators.LexicographicalComparator;
import xxl.core.pipes.operators.Pipes;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.operators.identities.TemporalBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * This strategy chooses the operatorpath segment, which is furthest back in application time.
 * 
 * To avoid costly operations the calculations will be done periodically and priorities will be assigned.
 * 
 * For more information: @see: Master Thesis of Udo Wolske: Eine Evaluation von Strategien fr die Ablaufssteuerung 
 * fr die Datenstrominfrastruktur PIPES
 */

public class BATATDWCStrategy extends AbstractStrategy<TemporalBufferPipe> {
    
    /**
     * calculates periodically how much back in application time the operators are and assigns an appropriate number of tickets.
     */
    class Refresh extends Processor {
    	
        /**
         * pipes of the considered graph
         */
        private Node[] vertices;
        
        /**
         * considered graph
         */
        private Graph graph;
        
        /**
         * pipes of the considered graph
         */
        private Set<Node> choices;
        
        /**
         * map: node -> long array
         * each array entry represents how much the corresponding input is back in Applicationtime
         */
        private HashMap<Node, Long[]> advancementMap;
        
        /**
         * operatorpaths of the considered graph
         */
        private OperatorPath[] paths;
        
        /**
         * map: bufferpipe -> priority (how much the given operator is back in application time in regards to an anchortime.)
         */
        private HashMap<TemporalBufferPipe, Long> prioMap;
        
        /**
         * map: bufferpipe -> position on a path
         */
        private HashMap<TemporalBufferPipe,Integer> posMap;
        
        /**
         * comparator of bufferpipes
         */
        private Comparator<TemporalBufferPipe> comp;
        
        /**
         * virtual pipe, is connected to the path segments, which have no choice operators
         */
        private VirtualPipe virtualPipe;
        
        
        /**
         * set of sinks of the graph
         */
        private Sink[] sinks;

        /**
         * new instance
         * 
         * @param graph considered graph
         * @param period updateperiod of the processor in ms
         */
        public Refresh(Graph graph, long period) {
            super(period);
            this.graph = graph;
            setName("BATATDStrategy.Refresh");
            vertices = graph.getNodes();
            sinks = graph.getSinks();
            advancementMap = new HashMap<Node, Long[]>();
            virtualPipe = new VirtualPipe();
            // Baue virtuelle Pipe an
            int ID = sinks.length + 1;
            for (Sink sink : sinks) {
                if (graph.containsPipe(sink))
                    Pipes.connect((Source) sink, virtualPipe, ID++);
                else
                    for (int i = 0; i < sink.getNoOfSources(); i++)
                        Pipes.connect(sink.getSource(i), virtualPipe, ID++);
            }
            paths = graph.getOperatorPaths();
            choices = new HashSet<Node>();
            prioMap = new HashMap<TemporalBufferPipe, Long>();
            posMap = new HashMap<TemporalBufferPipe,Integer>();
            HashSet<Sink> realChoiceSet = new HashSet<Sink>();
            Comparator<TemporalBufferPipe>[] ca = new Comparator[2];
            ca[0] = new HashComparator<TemporalBufferPipe,Long>(prioMap);
            ca[1] = new InverseComparator<TemporalBufferPipe>(new HashComparator<TemporalBufferPipe,Integer>(posMap));
            comp = new LexicographicalComparator<TemporalBufferPipe>(ca);
            // choice operators
            choices.add(virtualPipe);
            for (Node node : vertices) {
                if (graph.containsPipe(node) || graph.isTerminalSink(node)) {
                    Sink sink = (Sink) node;
                    if (sink.getNoOfSources() > 1)
                        choices.add(sink);
                }
                if (!graph.containsQueue(node)) {
                    node.getMetaDataManagement().include(APPLICATION_TIME_ADVANCEMENT);
                } 
            }
            virtualPipe.getMetaDataManagement().include(APPLICATION_TIME_ADVANCEMENT);
            // operatorpaths
            for (OperatorPath path : paths) {
                Node lastChoice = virtualPipe;
                int maxIndex = path.getQueues().length;
                Iterator<Node> it = path.iterator(false);
                while (it.hasNext()) {
                    Node node = it.next();
                    if (choices.contains(node)) 
                        lastChoice = node;
                    if (path.isControllable(node)) {
                        realChoiceSet.add((Sink) lastChoice);
                        posMap.put((TemporalBufferPipe) node, maxIndex--);
                        lastChoice = null;
                    }
                }
            }
            choices.retainAll(realChoiceSet);
            
            // if the virtualpipe is not needed
            if (!choices.contains(virtualPipe)) {
                ID = sinks.length + 1;
                for (Sink sink : sinks) {
                    if (graph.containsPipe(sink)) {
                        Pipes.disconnect((Source) sink, virtualPipe, ID++);
                    } else
                        for (int i = 0; i < sink.getNoOfSources(); i++)
                            Pipes.disconnect(sink.getSource(i), virtualPipe,
                                    ID++);
                }
                virtualPipe = null;
            }
        }

        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
        @Override
        public void process() {
        	// Thread terminated
            boolean terminated = true;
            for (Sink sink : sinks)
                terminated &= sink.isDone();
            if (terminated)
                terminate();
            else {
                // calc how much the operator is back in time (global)
                for (Node choice : choices) {
                    Long[] appTimeAdv = (Long[]) ((MetaDataHandler) choice.getMetaDataManagement().getMetaData().get(APPLICATION_TIME_ADVANCEMENT)).getMetaData();
                    Long[] behind = new Long[appTimeAdv.length];
                    for (int i = 0; i < behind.length; i++) {
                        behind[i] = appTimeAdv[i];
                    }
                    advancementMap.put(choice, behind);
                }
                for (OperatorPath path : paths) {
                    Node lastChoice = virtualPipe;
                    int index1 = -1;
                    Iterator<Node> it = path.iterator(false);
                    while (it.hasNext()) {
                        Node node = it.next();
                        if (path.isControllable(node)) {
                            if (index1 == -1 && lastChoice != null) {
                                index1 = Pipes.getSourceIndex((Sink) lastChoice,(Source) node);
                            }
                            long min = advancementMap.get(lastChoice)[index1];
                            if (prioMap.get(node)==null || (prioMap.get(node)!=null && min < prioMap.get(node))) {
                                prioMap.put((TemporalBufferPipe) node,min);
                            }
                        } 
                        else {
                            if (choices.contains(node)) {
                                lastChoice = node;
                                index1 = -1;
                            }
                            else if (lastChoice != null && graph.containsPipe(node)&& index1 == -1) { 
                                index1 = Pipes.getSourceIndex((Sink) lastChoice,(Source) node);
                            }
                        }
                    }
                }
                Collections.sort(controllables, comp);
                synchronized (queues) {
                    queues.clear();
                    queues.addAll(controllables);
                }
                prioMap.clear();
                advancementMap.clear();
            }
        }

        /**
         * 
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
        public void terminate() {
            for (Node node : vertices) {
                if (node.getMetaDataManagement().getMetaData().contains(APPLICATION_TIME_ADVANCEMENT))
                    node.getMetaDataManagement().exclude(APPLICATION_TIME_ADVANCEMENT);
            }
            super.terminate();
        }
    }
    
    /**
     * bufferpipes of the considered graph
     */
    private ArrayList<TemporalBufferPipe> queues;
    
    /**
     * Refresh-processor
     */
    private Refresh refresh;

    /**
     * new instance
     * 
     * @param graph considered graph
     * @param period updateperiod of reresh processor in ms
     */
    public BATATDWCStrategy(Graph graph, long period) {
        AbstractBufferPipe[] nodes = graph.getQueues();
        for (int i = 0; i < nodes.length; i++)
            register((TemporalBufferPipe) nodes[i]);
        queues = new ArrayList<TemporalBufferPipe>();
        queues.addAll(controllables);
        refresh = new Refresh(graph, period);
        refresh.start();
    }
    
    /**
     * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
     */
    @Override
    public boolean computeNext() {
        // TODO synchronisierung von nodes im Strategieprozessor
        synchronized (queues) {
            for (int i = 0; i < queues.size(); i++) {
                next = queues.get(i);
                if (next.isFinished()) {
                    deregister(next);
                } else if (next.isReady()) {
                    return true;
                }
            }
        }
        return false;
    }
}