/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * Implementation of the chain strategy proposed in 'Operator scheduling in data stream systems'.
 * 
 * For more information: @see 'Operator scheduling in data stream systems'
 * 
 * This strategy has a proven property: it minimizes the memory requirement for queues in the system.
 */
public class ChainStrategy extends AbstractStrategy<AbstractBufferPipe> {

    /**
     * calculates periodically the lower envelopes of the operatorpaths of the considered graphs.
     * It determines the chains and the priorities of the chains.
     * 
     * The calculations are based on the most recent gathered meta data.
     * 
     */
	class Refresh extends Processor {

        /**
         * nodes of the considered graph
         */
        protected Node[] n;
        
        /**
         * map: sdop -> derivate of the chain
         */
        protected HashMap<int[],Double> nodeToDerivate;
        
        /**
         * contains all coordinates of the sdops
         * ind.get(i)[0] = index of the path, which contains the bufferpipe
         * ind.get(i)[1] = position of the bufferpipe inside the path (only bufferpipes are considered here) 
         */
        protected ArrayList<int[]> ind;
        
        /**
         * list of sdops (steepest derivative operator points)
         */
        protected ArrayList<AbstractBufferPipe> sdop;
        
		/**
         * new instance
         *  
         * @param graph considered graph
         * @param period update period for refresh processor in ms
		 */
        public Refresh(Graph graph, long period) {
			super(period);
            nodeToDerivate = new HashMap<int[],Double>();
            ind = new ArrayList<int[]>();
            sdop = new ArrayList<AbstractBufferPipe>();
			this.setName("Chain.Refresh");
			n = graph.getPipes();
			for (int i=0;i<n.length;i++) {
				n[i].getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
				n[i].getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
				n[i].getMetaDataManagement().include(COST_MEASUREMENT);
			}
		}

        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public synchronized void process() {
			boolean finished = true;
			for (int i = 0; i < controllables.size(); i++) {
				if (!controllables.get(i).isFinished()) {
					finished = false;
                    break;
                }
			}
			if (finished)
				terminate();
			else {
				ind.clear();
				sdop.clear();
				nodeToDerivate.clear();
				calculateLowerEnvelopes();
				HashComparator<int[],Double> comp = new HashComparator<int[],Double>(nodeToDerivate);
	            Collections.sort(ind, comp);	            
				synchronized (indices) {
					indices.clear();
					indices.addAll(ind);
				}
				synchronized (SDOPs) {
					SDOPs.clear();
					SDOPs.addAll(sdop);
				}
			}
		}

        /**
         * calculates the lower envelopes of the considered graph
         */
		protected void calculateLowerEnvelopes() {
			for (int i = 0; i < paths.length; i++) {
				buildSegments(i);
			}
		}

       /**
        * constructs the chains for one operatorpath
        * 
        * @param ind index of the considered operatorpath
        */
		private void buildSegments(int ind) {
			// DoubleArray: [0] = current steepest derivate of this point
			// [1] = time to process for this point
			// [2] = input output ratio for this point
			// [3] = index of the next sdop
			// chains
			double[][] steepestDerivates = new double[operatorPaths[ind].length][4];
			OperatorPath path = paths[ind];
			int index = 0;
			double selectivity = 1d;
			double ttp = 0d;
			Iterator<Node> it = path.iterator(true);
			while (it.hasNext()) {
				Node node = it.next();
				if (path.isControllable(node)) {
					steepestDerivates[index][1] = ttp;
					steepestDerivates[index][2] = selectivity;
					if (index > 0) {
						calcDerivate(steepestDerivates, index, selectivity, ttp);
					}
					index++;
				} else {
					if (node.getMetaDataManagement().getMetaData()!=null) {
						if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT)) {
							MetaDataHandler mdh1 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT);
							MetaDataHandler mdh2 = null;
							if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
								mdh2 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
							double cm = (Double) mdh1.getMetaData();
							Double ir = 0.0;
							if (mdh2!=null)
								ir = (Double) mdh2.getMetaData();
							if (cm==0)
								cm = 1d;
							ttp += (ir!=null && ir!=0) ? cm/ir : cm;
						} 
						else 
							ttp+=1;
						if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))) {
							MetaDataHandler mdh = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
							Double sel = (Double) mdh.getMetaData();
							selectivity = (sel==null || Double.isNaN(sel) || sel==0) ? selectivity : selectivity * (1 / sel);
						}
					}
				}
			}
			// values for last entry
			calcDerivate(steepestDerivates,index,0.0,ttp);
			fillLists(ind, steepestDerivates);
		}


		/**
		 * calculates the derivates for the operator points  
		 * 
		 * @param steepestDerivates array with data
		 * @param index endpoint for calculation
		 * @param selectivity correlated selectivity of this point
		 * @param ttp correlated time to process of this point
		 */
		private void calcDerivate(double[][] steepestDerivates, int index, double selectivity, double ttp) {
			for (int i = 0; i < index; i++) {
				double derivate = (selectivity - steepestDerivates[i][2]) / (ttp - steepestDerivates[i][1]);
				if (derivate <= steepestDerivates[i][0]) {
					steepestDerivates[i][0] = derivate;
					steepestDerivates[i][3] = index;
				} else if (steepestDerivates[i][3] == 0)
					steepestDerivates[i][3] = i + 1;
			}
		}

        /**
         * fills the sdop and ind lists for one operatorpath 
         *
         * @param index index of the operator path
         * @param steepestDerivates array of data
         */
		protected void fillLists(int index, double[][] steepestDerivates) {
			steepestDerivates[steepestDerivates.length - 1][3] = steepestDerivates.length;
			for (int i = 0; i < steepestDerivates.length;) {
                int[] entry = new int[] { index, i };
				ind.add(entry);
				sdop.add(operatorPaths[index][i]);
                nodeToDerivate.put(entry, steepestDerivates[i][0]);
				i = (int) steepestDerivates[i][3];
			}
		}

        /**
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
        @Override
		public void terminate() {
			for (int i = 0; i < n.length; i++) {
				if (n[i].getMetaDataManagement().getMetaData()!=null) {
					if (n[i].getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,20,1,INPUT_OUTPUT_RATIO)))
						n[i].getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
					if (n[i].getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
						n[i].getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
					if (n[i].getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT))
						n[i].getMetaDataManagement().exclude(COST_MEASUREMENT);
				}
			}
			super.terminate();
		}
	}

    /**
     * operatorpaths of the considered graph
     */
	protected OperatorPath[] paths;

    /**
     * controllables of the operatorpaths of the considered graphs
     */
	protected AbstractBufferPipe[][] operatorPaths;

	/**
     * refresh processor 
	 */
	protected Refresh refresh;

    /**
     * index of the current path
     */
	private int pathIndex = 0;

	
    /**
     * index of the next operator on the chosen path
     */
	private int operatorIndex = 0;

    /**
     * list of all sdops
     */
	protected List<AbstractBufferPipe> SDOPs = Collections.synchronizedList(new ArrayList<AbstractBufferPipe>());

    /**
     * list of coordinates of sdops
     */
	protected List<int[]> indices = Collections.synchronizedList(new ArrayList<int[]>());

	/**
     * new instance
     * 
     * @param graph considered graph
     * @param period update period of the refresh processor in ms
	 */
	public ChainStrategy(Graph graph,long period) {
		paths = graph.getOperatorPaths();
		operatorPaths = new AbstractBufferPipe[paths.length][];
		for (int i = 0; i < paths.length; i++)
			operatorPaths[i] = paths[i].getQueues();
		AbstractBufferPipe[] queues = graph.getQueues();
		for (int i = 0; i < queues.length; i++)
			register(queues[i]);
		startProcessor(graph,period);
	} 
	
    
    /**
     * initiates and starts the refresh processor
     * 
     * @param graph considered graph
     * @param period update period of the refresh processor in ms
     */
    protected void startProcessor(Graph graph, long period) {
        refresh = new Refresh(graph,period);
        refresh.start();
    }

	/**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
	 */
	public boolean computeNext() {
        int index = 0;
		while (controllables.size() > 0) {
            if (index>=indices.size()) {
                index = 0;
                return false;
            }
            next = operatorPaths[pathIndex][operatorIndex++];
			if (next==null || operatorIndex == 0 || SDOPs.contains(next)) {
				int[] nextIndex = indices.get(index++);
                pathIndex = nextIndex[0];
                operatorIndex = nextIndex[1];
                next = operatorPaths[pathIndex][operatorIndex++];       
            }
            if (operatorIndex==operatorPaths[pathIndex].length)
                operatorIndex=0;
            if (next.isFinished())
                deregister(next);
            if (next.isReady()) {
                return true;
            }
		}
		return false;
	}
}