/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.identities.AbstractBufferPipe.AbstractBufferPipeMetaDataManagement.FIRST_ELEMENT_START;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Set;

import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.operators.identities.BufferPipe;
import xxl.core.pipes.operators.identities.TemporalBufferPipe;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * Implementation of the FCFS-Strategy.
 * 
 * Chooses the oldest tuple in the system (applicationtime) 
 * and processes it as far as possible on all related operatorpaths.
 * 
 */
public class FCFSStrategy extends AbstractStrategy<TemporalBufferPipe> {

    /**
     * considered graph
     */
    private Graph graph;

    /**
     * queues on the operatorpaths
     */
    private TemporalBufferPipe[][] operatorPaths;

    /**
     * index of the current path
     */
    private int pathIndex;
    
    /**
     * next considered path
     */
    private int nextPath;
    
    /**
     * indices of all paths, which are already considered processing the current tuple
     */
    private ArrayList<Integer> consideredPaths;
    
    /**
     * queues already considered
     */
    private ArrayList<TemporalBufferPipe> consideredQueues;

    /**
     * index of the bufferpipe on the current path
     */
    private int operatorIndex;

    /**
     * starting point of the paths
     */
    private HashMap<TemporalBufferPipe, ArrayList<Integer>> pathStart;

    /**
     * new instance
     * 
     * @param g considered graph
     */
    public FCFSStrategy(Graph g) {
        graph = g;
        OperatorPath[] paths = graph.getOperatorPaths();
        operatorPaths = new TemporalBufferPipe[paths.length][];
        pathStart = new HashMap<TemporalBufferPipe, ArrayList<Integer>>();
        consideredQueues = new ArrayList<TemporalBufferPipe>();
        nextPath = 0;
        // uses the property: the tuples in a bufferpipe are always ordered by timestamp (at)
        for (int i = 0; i < paths.length; i++) {
            AbstractBufferPipe[] abps = paths[i].getQueues();
            operatorPaths[i] = new TemporalBufferPipe[abps.length];
            for (int j=0;j<abps.length;j++) 
               operatorPaths[i][j] = (TemporalBufferPipe) abps[j];
            ArrayList<Integer> pathIndices = pathStart.get(operatorPaths[i][0]);
            if (pathIndices==null) {
            	pathIndices = new ArrayList<Integer>();
            }
            pathIndices.add(i);
            pathStart.put(operatorPaths[i][0], pathIndices);
        }
        AbstractBufferPipe[] queues = graph.getQueues();
        for (int i = 0; i < queues.length; i++) {
            register((TemporalBufferPipe) queues[i]);
        }
    }

    /**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#deregister(xxl.core.pipes.scheduler.Controllable)
	 */
    public void deregister(TemporalBufferPipe node) {
        super.deregister(node);
        if (pathStart.get(node) != null) {
            node.getMetaDataManagement().exclude(FIRST_ELEMENT_START);
            ArrayList<Integer> pathIndices = pathStart.get(node);
            for (int index : pathIndices) {
            	for (int j = 1; j < operatorPaths[index].length; j++) {
            		((BufferPipe) operatorPaths[index][j]).flush();
            		if (operatorPaths[index][j].isFinished()) {
            			deregister(operatorPaths[index][j]);
            		}
            	}
            }
        }
    }

    /**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#register(xxl.core.pipes.scheduler.Controllable)
	 */
    public void register(TemporalBufferPipe node) {
        super.register(node);
        if (pathStart.get(node) != null) 
        	node.getMetaDataManagement().include(FIRST_ELEMENT_START);
    }
   
    /**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
	 */
    @Override
    public boolean computeNext() {
        while (controllables.size() > 0) {
            if (next == null) {
                long min = Long.MAX_VALUE;
                Set<TemporalBufferPipe> startNodes = pathStart.keySet();
                for (TemporalBufferPipe node : startNodes) {
                	long v = Long.MAX_VALUE;
                	MetaDataHandler mdh = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(FIRST_ELEMENT_START);
            		v = (Long) mdh.getMetaData();
                	next = (v < min) ? node : next;
                	pathIndex = (v < min) ? pathStart.get(node).get(0) : pathIndex;
                	min = (v < min) ? v : min;
                	nextPath = 1;
                	consideredQueues.clear();
                	consideredPaths = pathStart.get(next);
                	operatorIndex = 1;
                }
                if (next==null)
                	return false;
            }
            else if (operatorPaths[pathIndex].length == operatorIndex)  {
            	if (consideredPaths.size()>1)
            		searchNext();
            	else 
            		next = null;
            }
            else {
                next = operatorPaths[pathIndex][operatorIndex++];
                consideredQueues.add(next);
            }
            if (next!=null) {
            	if (next.isFinished()) {
            		deregister(next);
            		next = null;
            	}
            	else if (next.isReady()) {
            		return true;
            	}
            }
            else { 
            	if (consideredPaths.size()>1)
            		searchNext();
            }
        }
        return false;
    }

    /**
     *  if a tuple has been processed completely on a path, the next path must be chosen
     *  this is achieved in this method.
     *
     */
	private void searchNext() {
		boolean nextFound = false;
		while (!nextFound) {
			if (nextPath == consideredPaths.size()) {
				next = null;
				nextFound = true;
			}
			else {
				pathIndex = consideredPaths.get(nextPath++);
				for (int i=0;i<operatorPaths[pathIndex].length;i++) {
					if (!consideredQueues.contains(operatorPaths[pathIndex][i])) {
						operatorIndex = i;
						next = operatorPaths[pathIndex][operatorIndex++];
						nextFound = true;
						consideredQueues.add(next);
					}
				}
			}
		}
	}
}