/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;

import xxl.core.comparators.InverseComparator;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * Greedy strategy as proposed in 'Operator scheduling in data stream systems'
 * 
 * it chooses the operator (bufferpipe) with the highest memory release capacity per timeunit
 * 
 * This value will be calculated periodically for each operator and the priorities will be adjusted.
 *
 */
public class GreedyStrategy extends AbstractStrategy<AbstractBufferPipe> {

	 /**
      * Refresh processor - periodically calculates the memory release capacity 
      * per time unit for each operator
      * 
	  */
	class Refresh extends Processor {

		/**
		 *  pipes of the considered graph
		 */
		private Node[] n;
		
		/**
		 *  list of controllables
		 */
		ArrayList<AbstractBufferPipe> cont = new ArrayList<AbstractBufferPipe>();
		
        /**
         * map: bufferpipe -> priority (mmr/tu of related vo)
         */
        protected HashMap<AbstractBufferPipe,Double> nodeToPriority;
        
        /**
         * copmarator of bufferpipes
         */
        private Comparator<AbstractBufferPipe> comp;
          
		/**
         * new instance
         *  
         *  @param graph considered graph
         * @param period update period of refresh processor in ms
		 */
        public Refresh(Graph graph, long period) {
			super(period);
            nodeToPriority = new HashMap<AbstractBufferPipe,Double>();
			this.setName("MetadataGreedy.Refresh");
			n = graph.getPipes();
			cont = new ArrayList<AbstractBufferPipe>();
			AbstractBufferPipe[] queues = graph.getQueues();
			for (int i=0;i<queues.length;i++) 
				cont.add(queues[i]);
			for (int i=0;i<n.length;i++) {
				n[i].getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
				n[i].getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
				n[i].getMetaDataManagement().include(COST_MEASUREMENT);
			}
			comp = new InverseComparator<AbstractBufferPipe>(new HashComparator<AbstractBufferPipe,Double>(nodeToPriority));
		}


        /**
         * 
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public synchronized void process() {
			boolean finished = true;
			for (int i = 0; i < controllables.size(); i++) {
				if (!controllables.get(i).isFinished()) {
					finished = false;
                    break;
                }
			}
			if (finished)
				terminate();
			else {
				cont.clear();
				nodeToPriority.clear();
				calculatePriorities();
	            Collections.sort(cont, comp);
				synchronized (controllables) {
					controllables.clear();
					controllables.addAll(cont);
				}
			}
		}
		
		 /**
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
        @Override
		public void terminate() {
			for (int i = 0; i < n.length; i++) {
				if (n[i].getMetaDataManagement().getMetaData()!=null) {
					if (n[i].getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO)))
						n[i].getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
					if (n[i].getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
						n[i].getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
					if (n[i].getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT))
						n[i].getMetaDataManagement().exclude(COST_MEASUREMENT);
				}
			}
			super.terminate();
		}

        /**
         * calculates path-based mmr/tu for each virtual operator and assigns priority to each bufferpipe 
         *
         */
		private void calculatePriorities() {
			for (int i=0;i<paths.length;i++) {
				OperatorPath path = paths[i];
				double selectivity = 1d;
				double ttp = 0d;
				Iterator<Node> it = path.iterator(true);
				AbstractBufferPipe toSet = null;
				while (it.hasNext()) {
					Node node = it.next();
					if (!path.isControllable(node)) {
						if (node.getMetaDataManagement().getMetaData() != null) {
							if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT)) {
								MetaDataHandler mdh1 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT);
								MetaDataHandler mdh2 = null;
								if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
									mdh2 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
								double cm = (Double) mdh1.getMetaData();
								Double ir = 0.0;
								if (mdh2!=null)
									ir = (Double) mdh2.getMetaData();
								if (cm==0)
									cm = 1;	
								ttp += (ir!=null && ir!=0) ? cm/ir : cm;
							} 
							else 
								ttp+=1;
							if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))) {
								MetaDataHandler mdh = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
								Double sel = (Double) mdh.getMetaData();
								selectivity = (sel == null || Double.isNaN(sel) || sel==0) ? selectivity : selectivity * (1 / sel);
							}
						}
					}
					else if (controllables.contains(node)){
						cont.add((AbstractBufferPipe) node);
						calcPriority(selectivity, ttp, toSet);
						toSet = (AbstractBufferPipe) node;
					}
				}
				calcPriority(selectivity,ttp,toSet);
			}
		}

		/**
         * calculates priority of a single controllable 
         * @param selectivity selectivity of related vo
         * @param ttp time to process of related vo
         * @param toSet considered bufferpipe
		 */
		private void calcPriority(double selectivity, double ttp, AbstractBufferPipe toSet) {
			double val = (1-selectivity)/ttp;
			selectivity = 1.0;
			ttp = 0.0;
			if (toSet==null) {
				if (nodeToPriority.get(toSet)!=null) {
					double old = nodeToPriority.get(toSet);
					if (val<old)
						nodeToPriority.put(toSet, val);
				}
				else
					nodeToPriority.put(toSet,val);
			}
		}
	}
	
	/**
	 * operatorpaths of considered graphs
	 */
	private OperatorPath[] paths;

	/**
	 * Refresh-Thread
	 */
	private Refresh refresh;
	
	/**
	 * new instance
	 * @param graph considered graph
	 * @param updatePeriod update period of refresh processor in ms
	 */
	public GreedyStrategy(Graph graph, long updatePeriod) {
		super(graph.getQueues());
		refresh = new Refresh(graph,updatePeriod);
		paths = graph.getOperatorPaths();
		refresh.start();
	}
	
	/**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
	 */
	@Override
	public boolean computeNext() {
		synchronized(controllables) {
			for (int i=0;i<controllables.size();i++) {
				next = controllables.get(i);
				if (next.isFinished())
					deregister(next);
				if (next.isReady()) {
					return true;		
				}
			}
		}
		return false;
	}

}
