/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.identities.AbstractBufferPipe.AbstractBufferPipeMetaDataManagement.FIRST_ELEMENT_START;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;

import xxl.core.comparators.InverseComparator;
import xxl.core.pipes.operators.identities.TemporalBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * This strategy chooses the bufferpipe with the tuple with the oldest application timestamp. This bufferpipe will 
 * transfer this tuple to the following operators. The application time of the operator will be brought forward.
 * 
 *  The execution order will be calculated periodically. (lazy)
 * 
 * For more information: @see: Master Thesis of Udo Wolske: Eine Evaluation von Strategien fr die Ablaufssteuerung 
 * fr die Datenstrominfrastruktur PIPES
 */
public class LazyBATANaiveStrategy extends AbstractStrategy<TemporalBufferPipe> {

    /**
     *periodically  calculates the execution order of bufferpipes
     * 
     * 
     */
	class Refresh extends Processor {
		
        /**
         * map: bufferpipe -> timestamp
         */
		private HashMap<TemporalBufferPipe,Long> prioMap;
        
        /**
         * comparator of bufferpipes
         */
		private Comparator<TemporalBufferPipe> comp; 
		/**
		 * refresh-processor
         * 
		 * @param period update period of the processor
		 */
		public Refresh(long period) {
			super(period);
			for (TemporalBufferPipe queue : controllables)
				queue.getMetaDataManagement().include(FIRST_ELEMENT_START);
			prioMap = new HashMap<TemporalBufferPipe,Long>();
			comp = new HashComparator<TemporalBufferPipe,Long>(prioMap);
		}
		
        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public void process() {
			if (controllables.size()==0)
				terminate();
			else {
				synchronized(controllables) {
					for (TemporalBufferPipe queue : controllables) {
						if (queue.getMetaDataManagement().getMetaData().contains(FIRST_ELEMENT_START)) {
							MetaDataHandler mdh = (MetaDataHandler) ((Node) queue).getMetaDataManagement().getMetaData().get(FIRST_ELEMENT_START);
							Long v = (Long) mdh.getMetaData();
							prioMap.put(queue , v);
						}
					}
					Collections.sort(controllables, comp);
				}
			}
		}
		
        /**
         * 
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
		public void terminate() {
			for (TemporalBufferPipe queue : controllables) {
				queue.getMetaDataManagement().exclude(FIRST_ELEMENT_START);
			}
			super.terminate();
		}
		
	}
	
    /**
     * refresh-processor
     */
	private Refresh refresh;
	
	/**
	 * new instance
     * 
	 * @param graph considered graph
	 * @param period update period of refresh processor in ms
	 */
	public LazyBATANaiveStrategy(Graph graph, long period) {
		super((TemporalBufferPipe[]) graph.getQueues());
		refresh = new Refresh(period);
		refresh.start();
	}
	
    /**
     * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
     */
	@Override
	public boolean computeNext() {
		synchronized (controllables) {
			for (int i = 0;i<controllables.size();i++) {
				next = controllables.get(i);
				if (next.isFinished())
					deregister(next);
				else if (next.isReady())
					return true;
			}
		}
		return false;
	}
}