/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;

import xxl.core.functions.Function;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.SchedulingFunctions;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * Implementation of the MinLatency-Traversal-Strategy proposed in 'Operator scheduling in a data stream manager'
 * 
 * calculates a query-tree-traversal, which induces minimal latency per tuple
 * the traversal will be adjusted periodically.
 */
public class MinLatencyStrategy extends AuroraStrategy<AbstractBufferPipe> {
	
	/**
     * calculates and adjusts traversal periodically
	 */
	class MinLatencyRefresh extends Refresh {

        /**
         * new instance
         * @param query considered query tree
         * @param period update period of refresh processor in ms
         */
		public MinLatencyRefresh(Graph query, long period) {
			super(query, period);
            calcDownStream();
			setName("Aurora.MinLatency.Refresh");
			getTraversals = new Function<Graph, AbstractBufferPipe[]>() {
				public AbstractBufferPipe[] invoke() {
					HashMap<AbstractBufferPipe, Double> sel = new HashMap<AbstractBufferPipe, Double>();
					HashMap<AbstractBufferPipe, Double> outputCost = new HashMap<AbstractBufferPipe, Double>();
					ArrayList<AbstractBufferPipe> help = new ArrayList<AbstractBufferPipe>(); 
					for (int i = 0; i < paths.length; i++) {
						OperatorPath path = paths[i];
						Node queue = null; 
						AbstractBufferPipe[] queues = path.getQueues();
						for (int k = queues.length - 1; k > -1; k--) {
							if (sel.containsKey(queues[k]))
								queue = (Node) queues[k];
							else
								break;
						}
						Iterator<Node> it = path.iterator(false);
						double selectivity = (sel.containsKey(queue)) ? sel.get(queue) : 1.0;
						double output = (outputCost.containsKey(queue)) ? outputCost.get(queue) : 0.0;
						double cost = 0.0;
						while (it.hasNext()) {
							Node node = it.next();
							if (queue!=null && node!=queue) {
								continue;
							}
							else if (node==queue) {
								queue = null;
								continue;
							}
							if (node.getMetaDataManagement().getMetaData()!=null) {
								double c = SchedulingFunctions.processingTimePerTuple.invoke(node);
								if (c == Double.NaN) {
									cost += 1 ;
								}
								else 
									cost += c;
								if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))) {
									MetaDataHandler mdh = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
									Double s = (Double) mdh.getMetaData();
									s = (s==null) ? 1 : 1/s;
									selectivity *= s;  
								}
							}
							if (path.isControllable(node)) {
								sel.put((AbstractBufferPipe) node, selectivity);
								output += (cost / selectivity);
								outputCost.put((AbstractBufferPipe) node, output);
								selectivity = 1.0;
								cost = 0.0;
							}
						}
						HashComparator<AbstractBufferPipe, Double> comp = new HashComparator<AbstractBufferPipe, Double>(outputCost);
						Arrays.sort(queues, 0, queues.length, comp);
						for (int j = 0; j < queues.length; j++)
							help.addAll(downStream.get(queues[j]));
					}
					AbstractBufferPipe[] result = new AbstractBufferPipe[help.size()];
					for (int j = 0; j < help.size(); j++)
						result[j] = help.get(j);
					return result;
				}
			};
			traversal = t = getTraversals.invoke();
		}
	}
	
	private MinLatencyRefresh refresh;
	
	/**
	 * new instance
	 * @param g considered query tree
	 * @param period updateperiod of processor in ms
	 */
	public MinLatencyStrategy(Graph g, long period){
		super(g);
		refresh = new MinLatencyRefresh(g,period);
		refresh.start();
	}

}
