/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.ALL_INPUT_OBJECT_SIZES;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.scheduler.SchedulingFunctions.averageInputObjectSize;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;

import xxl.core.comparators.InverseComparator;
import xxl.core.functions.Function;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.SchedulingFunctions;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * Implementation of the MinMemory-Traversal-Strategy proposed in 'Operator scheduling in a data stream manager'
 * 
 * calculates a query-tree-traversal, which induces a maximal memory release per time unit
 * the traversal will be adjusted periodically.
 */
public class MinMemoryStrategy extends AuroraStrategy<AbstractBufferPipe> {
	
    /**
     * calculates and adjusts traversal periodically
     */
	class MinMemoryRefresh extends Refresh {

	    /**
         * queues of the considered graphs 
	     */
        private AbstractBufferPipe[] queues;
        
        /**
         * map: bufferpipe -> input tuple size
         */
        private HashMap<AbstractBufferPipe,Double> controllableToTupleSize;
        
        /**
         * new instance
         * 
         * @param query considered graph
         * @param period updateperiod of the processor in ms
         */
		public MinMemoryRefresh(Graph query, long period) {
			super(query, period);
			controllableToTupleSize = new HashMap<AbstractBufferPipe,Double>();
			setName("Aurora.MinMemory.Refresh");
            queues = query.getQueues();
            for (int i=0;i<queues.length;i++) 
                queues[i].getMetaDataManagement().include(ALL_INPUT_OBJECT_SIZES);
            calcDownStream();
            // minmemory traversal
			getTraversals = new Function<Graph, AbstractBufferPipe[]>() {
				public AbstractBufferPipe[] invoke() {
					double cost = 0.0;
					double sel = 1.0;
					HashMap<AbstractBufferPipe, Double> memRR = new HashMap<AbstractBufferPipe, Double>();
					ArrayList<AbstractBufferPipe> help = new ArrayList<AbstractBufferPipe>();
					for (int j = 0; j < paths.length; j++) {
						OperatorPath path = paths[j];
						Node queue = null;
						AbstractBufferPipe[] queues = path.getQueues();
						Iterator<Node> it = path.iterator(false);
						for (int k = queues.length - 1; k > -1; k--)
							if (memRR.containsKey(queues[k]))
								queue = (Node) queues[k];
							else
								break;
						while (it.hasNext()) {
							Node node = it.next();
							if (queue!=null && node!=queue) {
								continue;
							}
							else if (node==queue) {
								queue = null;
								continue;
							}
							if (node.getMetaDataManagement().getMetaData()!=null) {
								double c = SchedulingFunctions.processingTimePerTuple.invoke(node);
								if (c == Double.NaN) {
									cost += 1 ;
								}
								else 
									cost += c;
								if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))) {
									MetaDataHandler mdh = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
									Double selectivity  = (Double) mdh.getMetaData();
									selectivity = (selectivity==null) ? 1.0 : selectivity;
									sel *= selectivity;
								}
							}
							if (path.isControllable(node)) {
                                double tSize = controllableToTupleSize.containsKey(node) ? controllableToTupleSize.get(node) : 0;
                            	if (tSize==0) { 
                            		if (node.getMetaDataManagement().getMetaData().contains(ALL_INPUT_OBJECT_SIZES)) {
                            			tSize = averageInputObjectSize.invoke(node);
                            		}
                            		if (tSize==0)
                            			tSize = 1;
                                }
								if (!memRR.containsKey(node)) {
									double memory_rr = (tSize * (1 - sel)) / cost;
									memRR.put((AbstractBufferPipe) node, memory_rr);
									sel = 1.0;
									cost = 0.0;
								}
							}
						}
					}
					Comparator<AbstractBufferPipe> comp = new InverseComparator(new HashComparator<AbstractBufferPipe, Double>(memRR));
					Arrays.sort(queues, 0, queues.length, comp);
					for (int j = 0; j < queues.length; j++) {
						ArrayList<AbstractBufferPipe> ds = downStream.get(queues[j]);
						help.add(queues[j]);
						if (ds.size()>1) {
							if (j<queues.length-1)
								for (int i=1;i<ds.size();i++) {
									if (memRR.get(ds.get(i))>=memRR.get(queues[j+1]))
										help.add(ds.get(i));
									else 
										break;
								}
						}	
					}
					AbstractBufferPipe[] result = new AbstractBufferPipe[help.size()];
					for (int j = 0; j < help.size(); j++)
						result[j] = help.get(j);
					return result;
				}
			};
			traversal = t = getTraversals.invoke();
		}
        
        /**
         * @see xxl.core.pipes.scheduler.strategies.AuroraStrategy.Refresh#terminate()
         */
        public void terminate() {
            for (AbstractBufferPipe abp : queues) {
                abp.getMetaDataManagement().exclude(ALL_INPUT_OBJECT_SIZES);
            }
            super.terminate();
        }
	}
	
    /**
     * refresh-processor
     */
	private MinMemoryRefresh refresh;
	
	/**
	 * new instance
     * 
	 * @param q considered query tree
	 * @param period update period of refresh processor
	 */
	public MinMemoryStrategy (Graph q, long period) {
		super(q);
		refresh = new MinMemoryRefresh(q,period);
		refresh.start();
	}

}
