/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.Pipe;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.SchedulingFunctions;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.AbstractSource;
import xxl.core.pipes.sources.Source;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * Implementation of the chain strategy proposed in 'Operator scheduling in data stream systems.
 * 
 * This strategy has a proven property: it minimizes the memory requirement for queues in the system.
 * 
 * Therefor it calculates the lower envelopes of the operatorpaths of the considered query graph. 
 * 
 * It determines chains of consecutive bufferpipes which will be executed in fcfs-manner. 
 * 
 * The priority of the chains is the negative gradient of the straight line determined by two points on 
 * the lower envelope.  The priority ofthe chain will be assigned to the first bufferpipe of the chain 
 * (the steepest derivate operator points (sdops))
 * 
 * all bufferpipes within the chain are useless and will be removed.
 * 
 * TODO : inherit from chain-strategy
 * 
 */
public class OptimizedChainStrategy extends AbstractStrategy<AbstractBufferPipe> {

    /**
     * calculates periodically the lower envelopes of the operatorpaths of the considered graphs.
     * It determines the chains and the priorities of the chains.
     * 
     * The calculations are based on the most recent gathered meta data.
     * 
     */
	class Refresh extends Processor {

       /**
        * pipes of the considered graph
        */
        private Node[] n;
        
        /**
         * map: sdop -> derivate of the chain
         */
        protected HashMap<AbstractBufferPipe,Double> nodeToDerivate;
       
        /**
         * list of sdops
         */
        private ArrayList<AbstractBufferPipe> sdop;
        
        /**
         * considered graph
         */
        private Graph graph;
        
		/**
         * new instance
         *  
         *  @param gr considered graph
         * @param period update period of processor in ms
		 */
        public Refresh(Graph gr, long period) {
			super(period);
			graph = gr;
            nodeToDerivate = new HashMap<AbstractBufferPipe,Double>();
            sdop = new ArrayList<AbstractBufferPipe>();
			this.setName("Chain.Refresh");
			n = graph.getPipes();
			for (Node node : n) {
				node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_OUTPUT_RATIO));
				node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_RATE));
				node.getMetaDataManagement().include(COST_MEASUREMENT);
			}
		}

        /**
         * 
         * 
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public synchronized void process() {
			boolean finished = true;
			for (int i = 0; i < controllables.size(); i++) {
				if (!controllables.get(i).isFinished()) {
					finished = false;
                    break;
                }
			}
			if (finished)
				terminate();
			else {
				sdop.clear();
				nodeToDerivate.clear();
				calculateLowerEnvelopes();
				HashComparator<AbstractBufferPipe,Double> comp = new HashComparator<AbstractBufferPipe,Double>(nodeToDerivate);
	            Collections.sort(sdop, comp);
	            for (int i=0;i<controllables.size();i++)
	            	if (!sdop.contains(controllables.get(i)))
	            		deregister(controllables.get(i--));
				synchronized (SDOPs) {
					SDOPs.clear();
					SDOPs.addAll(sdop);
				}
			}
		}

         /**
         * calculates the lower envelopes of the considered graph
         */
		private void calculateLowerEnvelopes() {
			for (int i = 0; i < paths.length; i++) {
				buildSegments(i);
			}
		}

          /**
            * constructs the chains for one operatorpath
            * 
            * @param ind index of the considered operatorpath
            */
		private void buildSegments(int ind) {
		        //        DoubleArray: [0] = current steepest derivate of this point
            // [1] = time to process for this point
            // [2] = input output ratio for this point
            // [3] = index of the next sdop
            // chains
			OperatorPath path = paths[ind];
			int nrOfPipes = path.length() - path.getQueues().length;
			double[][] steepestDerivates = new double[nrOfPipes][4];
			int index = 0;
			double selectivity = 1d;
			double ttp = 0d;
			Iterator<Node> it = path.iterator(true);
			while (it.hasNext()) {
				Node node = it.next();
				if (!path.isControllable(node)) {
					steepestDerivates[index][1] = ttp;
					steepestDerivates[index][2] = selectivity;
					if (index > 0) {
						calcDerivate(steepestDerivates, index, selectivity, ttp);
					}
					index++;
					if (node.getMetaDataManagement().getMetaData()!=null) {
						if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT)) {
							MetaDataHandler mdh1 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(COST_MEASUREMENT);
							MetaDataHandler mdh2 = null;
							if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_RATE)))
								mdh2 = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_RATE));
							double cm = (Double) mdh1.getMetaData();
							Double ir = 0.0;
							if (mdh2!=null)
								ir = (Double) mdh2.getMetaData();
							if (cm==0)
								cm = 1;
							ttp += (ir!=null && ir!=0) ? cm/ir : cm;
						} 
						else 
							ttp+=1;
						if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_OUTPUT_RATIO))) {
							MetaDataHandler mdh = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_OUTPUT_RATIO));
							Double sel = (Double) mdh.getMetaData();
							selectivity = (sel == null || Double.isNaN(sel) || sel==0) ? selectivity : selectivity * (1 / sel);
						}
					}
				}
			}
			// Werte fr letzten Eintrag
			calcDerivate(steepestDerivates,index,0.0,ttp);
			fillLists(ind, steepestDerivates);
		}

		/**
		 * Berechnet alle strksten Steigungen bis zu dem Punkt index 
		 * 
		 * @param steepestDerivates Array mit gesammelten Daten
		 * @param index Punkt, bis zu dem gerechnet wird
		 * @param selectivity Selektivitt in diesem Punkt
		 * @param ttp TimeToProcess bis zum nchsten Punkt
		 */
		private void calcDerivate(double[][] steepestDerivates, int index, double selectivity, double ttp) {
			for (int i = 0; i < index; i++) {
				double derivate = (selectivity - steepestDerivates[i][2]) / (ttp - steepestDerivates[i][1]);
				if (derivate <= steepestDerivates[i][0]) {
					steepestDerivates[i][0] = derivate;
					steepestDerivates[i][3] = index;
				} else if (steepestDerivates[i][3] == 0)
					steepestDerivates[i][3] = i + 1;
			}
		}

        /**
         * calculates the derivates for the points 
         * 
         * @param steepestDerivates array with data
         * @param index endpoint for calculation
         * @param selectivity correlated selectivity of this point
         * @param ttp correlated time to process of this point
         */
		protected void fillLists(int index, double[][] steepestDerivates) {
			int opCount = 0;
			int sdopIndex = 0;
			steepestDerivates[steepestDerivates.length - 1][3] = steepestDerivates.length;
			OperatorPath path = paths[index];
			AbstractBufferPipe bp = null;
			Node prevNode = null;
			Iterator<Node> it = path.iterator(true);
			while (it.hasNext()) {
				Node node = it.next();
				if (path.isControllable(node)) 
					bp = (AbstractBufferPipe) node;
				else {
					if (opCount==sdopIndex) {
						if (bp==null && prevNode!=null) {
							Source source = (Source) prevNode;
							Sink sink = (Sink) node;
							boolean isDone = 
								(graph.containsPipe(prevNode) && ((Pipe) prevNode).isDone())
								|| (graph.isInitialSource(prevNode) && !graph.containsPipe(prevNode) && ((AbstractSource) prevNode).getProcessor().isTerminated());
							if (!sink.isDone() && !isDone) {
								bp = graph.addBuffer(source, sink, false);
								register(bp);
								if (sdop.contains(bp)) {
									double der = nodeToDerivate.get(bp);
									if (der>steepestDerivates[sdopIndex][0])
										nodeToDerivate.put(bp, steepestDerivates[sdopIndex][0]);
								}
								else {
									sdop.add(bp);
									nodeToDerivate.put(bp, steepestDerivates[sdopIndex][0]);
								}
							}
						}
						else if (bp!=null) {
							if (!bp.isFinished()) {
								sdop.add(bp);
								nodeToDerivate.put(bp, steepestDerivates[sdopIndex][0]);
							}
						}
						if (graph.containsPipe(node))
							sdopIndex = (int) steepestDerivates[sdopIndex][3];
						bp = null;
					}
					else {
						if (bp!=null) {
							Source source = (Source) prevNode;
							Sink sink = (Sink) node;
							if (!sink.isDone()) {
								graph.removeBuffer(source, bp,sink);
							}
							bp = null;
						}
					}
					if (graph.containsPipe(node))
						opCount++;
					prevNode = node;
					bp = null;
				}	
			}
		}
		
        /**
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
        @Override
		public void terminate() {
			for (Node node : n) {
				if (node.getMetaDataManagement().getMetaData()!=null) {
					if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_OUTPUT_RATIO)))
						node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_OUTPUT_RATIO));
					if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_RATE)))
						node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,SchedulingFunctions.windowSize,SchedulingFunctions.minimumResultWindow,INPUT_RATE));
					if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT))
						node.getMetaDataManagement().exclude(COST_MEASUREMENT);
				}
			}
			super.terminate();
		}
	}
	

    /**
     * operatorpaths of the graph
     */
	protected OperatorPath[] paths;

    /**
     * refresh-processor
     */
	protected Refresh refresh;

	/**
     * list of all sdops 
	 */
	protected List<AbstractBufferPipe> SDOPs = Collections.synchronizedList(new ArrayList<AbstractBufferPipe>());

	/**
     * New instance
     * 
     * @param graph considered graph
     * @param period updateperiod of refresh processor in ms
	 */
	public OptimizedChainStrategy(Graph graph,long period) {
		paths = graph.getOperatorPaths();
		AbstractBufferPipe[] queues = graph.getQueues();
		SDOPs = new ArrayList<AbstractBufferPipe>();
		for (int i = 0; i < queues.length; i++)
			register(queues[i]);
		startProcessor(graph,period);
	}
    
    /**
     * initiates and starts the refresh processor
     * 
     * @param graph considered graph
     * @param period update period of the refresh processor in ms
     */
    protected void startProcessor(Graph graph, long period) {
        refresh = new Refresh(graph,period);
        refresh.start();
    }

	/**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
	 */
	public boolean computeNext() {
//		System.out.println("nodes");
//		for (int i=0;i<nodes.size();i++) {
//			BufferPipe bp = (BufferPipe) nodes.get(i);
//			System.out.println(bp+" "+bp.getBufferSize());
//		}
//		System.out.println();
//		for (int i=0;i<SDOPs.size();i++) {
//			BufferPipe bp = (BufferPipe) SDOPs.get(i);
//			System.out.println(bp+" "+bp.getBufferSize());
//		}
		for (int i=0;i<SDOPs.size();i++) {
			next = SDOPs.get(i);
			if (next.isFinished()) {
				deregister(next);
			}
			else if (next.isReady()) {
				return true;
			}
		}
		return false;
	}
}