/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.operatorPathProcessingCapacity;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;

import xxl.core.comparators.InverseComparator;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.scheduler.HashComparator;

/**
 * Implementation of the path capacity strategy proposed in 'Scheduling strategies
 * for processing continuous queries over streams'
 * 
 * path based strategy: For each path the pathcapacity is calculated as priority.
 * The chosen path will be executed in fcfs manner. 
 * 
 * The priorities will be adjusted periodically
 */
public class PathCapacityStrategy extends AbstractPathStrategy<AbstractBufferPipe> {

	
    /**
     * for each path the pathcapacity will be calculated and assigned as priority.
     */
    class Refresh extends Processor {

        /**
         * nodes of the considered graph
         */
        private Node[] nodes;

        /**
         * map: pathindex -> pathcapacity
         */
        private HashMap<Integer, Double> pathCapacities;

        /**
         * comparator of paths
         */
        private Comparator<Integer> comp;

        /**
         * new instance
         * 
         * @param graph considered graph
         * @param period updateperiod of considered graph
         */
        public Refresh(Graph graph, long period) {
            super(period);
            setName("PathCapacity.Refresh");
            pathCapacities = new HashMap<Integer, Double>();
            comp = new InverseComparator<Integer>(new HashComparator<Integer, Double>(pathCapacities));
            nodes = graph.getPipes();
            for (Node node : nodes) {
                node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
                node.getMetaDataManagement().include(COST_MEASUREMENT);
                node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
            }
        }
        
        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
        @Override
        public void process() {
            boolean notTerminate = false;
            for (int i = 0; i < controllables.size(); i++) {
                notTerminate |= controllables.get(i).isReady(); 
            }
            if (!notTerminate) {
                terminate();
                return;
            }
            for (PathStrategy<AbstractBufferPipe> pathStrategy : controllables) {
                
                // pathcapacity is implemented as function
            	pathCapacities.put(
            			strategyToPath.get(pathStrategy), 
            			operatorPathProcessingCapacity.invoke(pathStrategy.getPath()));
            }
            synchronized (priorities) {
            	Collections.sort(priorities, comp);
            }
        }

        /**
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
        @Override
        public void terminate() {
            for (Node node : nodes) {
            	if (node.getMetaDataManagement().getMetaData()!=null) {
            		if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
            			node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
            		if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT))
            			node.getMetaDataManagement().exclude(COST_MEASUREMENT);
            		if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO)))
            			node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
            	}
            }
            super.terminate();
        }
    }

    /**
     * refresh - processor
     */
   private Refresh refresh;

    
   /**
    * new instance 
    * 
    * @param graph considered graph
    * @param period updateperiod for the refresh processor in ms
    */
    public PathCapacityStrategy(Graph graph, long period) {
		super(graph, period);
	}
    
    /**
     * @see xxl.core.pipes.scheduler.strategies.AbstractPathStrategy#startProcessor(xxl.core.pipes.scheduler.Graph, long)
     */
	protected void startProcessor(Graph graph, long period) {
		refresh = new Refresh(graph, period);
        refresh.start();
	}


}
