/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import xxl.core.comparators.InverseComparator;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.SchedulingFunctions;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * preemptive ratebased scheduling strategy as proposed in ' preemptive ratebased operator scheduling  in a dsms'
 * 
 * calculates periodically the induced output rate of an operator and assigns this as its priority.
 * 
 * 
 */
public class PreRateBasedStrategy extends AbstractStrategy<AbstractBufferPipe> {

	/**
	 * calculates periodically the induced output rate of an operator and assigns this as its priority.
	 * 
	 *
	 */
	public class Refresh extends Processor {

        /**
         * operatorpaths of considered graph
         */
		private OperatorPath[] paths;

        /**
         * nodes of considered graph
         */
		private Node[] vertices;

        /**
         * map: node -> output rate
         */
		private HashMap<AbstractBufferPipe, Double> nodeToPriority;
        
        /**
         * map: node -> ior
         */
		private HashMap<Node,Double> nodeToSel;
        
        /**
         * map: node -> or
         */
		private HashMap<Node,Double> nodeToOr;

        /**
         * priority comparator
         */
		private Comparator<AbstractBufferPipe> comp;

		/**
		 * new instance
         * 
		 * @param graph considered graph
		 * @param period updateperiod of processor in ms
		 */
		public Refresh(Graph graph, long period) {
			super(period);
			paths = graph.getOperatorPaths();
			vertices = graph.getNodes();
			nodeToPriority = new HashMap<AbstractBufferPipe, Double>();
			nodeToOr = new HashMap<Node,Double>();
			nodeToSel = new HashMap<Node,Double>();
			comp = new InverseComparator<AbstractBufferPipe>(
					new HashComparator<AbstractBufferPipe, Double>(nodeToPriority));
			for (Node vertex : vertices) {
				if (graph.containsPipe(vertex)) {
					vertex.getMetaDataManagement().include(COST_MEASUREMENT);
					vertex.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
					vertex.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
				}
			}
		}

        /**
         * 
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public void process() {
			Set<Node> test = new HashSet<Node>();
			for (OperatorPath path : paths) {
				double sel = 1d;
				double cost = 0d;
				Iterator<Node> it = path.iterator(false);
				Node lastContr = null;
				while (it.hasNext()) {
					Node node = it.next();
					if (!test.contains(node)) {
						test.add(node);
						if (path.isControllable(node)) {
							if (lastContr != null) {
								double or = sel / cost;
								nodeToPriority.put((AbstractBufferPipe) lastContr, or);
							}
							lastContr = node;
						} else {
							if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT)) {
								double c = SchedulingFunctions.processingTimePerTuple.invoke(node);
								cost += c * sel;
								nodeToOr.put(node, cost);
							}
							if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO))) {
								MetaDataHandler mdh = (MetaDataHandler) node.getMetaDataManagement().getMetaData().get(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
								Double s = (Double) mdh.getMetaData();
								s = (s==null) ? 1 : s;
								sel *= s; 
								nodeToSel.put(node, sel);
							}
						}
					}
					else {
						if (nodeToSel.containsKey(node))
							sel = nodeToSel.get(node);
						if (nodeToOr.containsKey(node))
							cost = nodeToOr.get(node);
						
					}
				}
			}
			synchronized(controllables) {
				Collections.sort(controllables,comp);
			}
		}

        /**
         * 
         * @see xxl.core.pipes.processors.Processor#terminate()
         */
		public void terminate() {
			for (Node vertex : vertices) {
				if (vertex.getMetaDataManagement().getMetaData().contains(
						COST_MEASUREMENT))
					vertex.getMetaDataManagement().exclude(COST_MEASUREMENT);
				if (vertex.getMetaDataManagement().getMetaData().contains(
						new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO)))
					vertex.getMetaDataManagement().exclude(
							new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
				if (vertex.getMetaDataManagement().getMetaData().contains(
						new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
					vertex.getMetaDataManagement().exclude(
							new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
			}
			super.terminate();
		}
	}
	
    /**
     * refresh-processor
     */
	private Refresh refresh;
	
	/**
	 * new instance
     * 
     * 
	 * @param graph considered graph
	 * @param period updateperiod of processor in ms
	 */
	public PreRateBasedStrategy(Graph graph, long period) {
		super(graph.getQueues());
		refresh = new Refresh(graph,period);
		refresh.start();
	}

    /**
     * 
     * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
     */
	@Override
	public boolean computeNext() {
		synchronized(controllables) {
			for (int i=0;i<controllables.size();i++) {
				next = controllables.get(i);
				if (next.isFinished())
					deregister(next);
				else if (next.isReady())
					return true;
			}
		}
		return false;
	}
	
}