/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.ALL_INPUT_OBJECT_SIZES;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_OUTPUT_RATIO;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.INPUT_RATE;
import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.OUTPUT_OBJECT_SIZE;
import static xxl.core.pipes.operators.CostModelMetaDataIdentifiers.COST_MEASUREMENT;
import static xxl.core.pipes.scheduler.SchedulingFunctions.minimumResultWindow;
import static xxl.core.pipes.scheduler.SchedulingFunctions.operatorMemoryReleaseCapacity;
import static xxl.core.pipes.scheduler.SchedulingFunctions.operatorPathMemoryReleaseCapacity;
import static xxl.core.pipes.scheduler.SchedulingFunctions.windowSize;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;

import xxl.core.comparators.InverseComparator;
import xxl.core.math.statistics.parametric.aggregates.Average;
import xxl.core.pipes.metaData.AggregationOverWindowMetaDataIdentifier;
import xxl.core.pipes.operators.Pipe;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.scheduler.HashComparator;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;

/**
 * Implementation of segment strategy proposed in 'Scheduling strategies
 * for processing continuous queries over streams'
 * 
 * each operatorpath will be partitioned into a set of disjunct segments, which will be scheduled.
 * Each segment will be processed in FCFS-order.
 * 
 * Each segment has the segmentmemoryrelease capacity as priority.
 * 
 * segment-partition will be recalculated periodically.
 * 
 * Segments have bufferpipes at the start only. Interior bufferpipes will be removed.
 * 
 * assumption:  No chain of bufferpipes exists.
 * 
 */
public class SegmentStrategy extends AbstractStrategy<AbstractBufferPipe> {

	/**
     * refresh - processor, adjusts segments periodically
     * 
	 */
	class Refresh extends Processor {

        /**
         * considered graph
         */
		protected Graph graph;

        /**
         * operatorpaths of the considered graph
         */
		protected OperatorPath[] operatorPaths;

        /**
         * map: segment -> pathindex
         */
		protected HashMap<ArrayList<Node>, Integer> segmentToPath;

        /**
         * has partitin changed since last calculation?
         * if not, do nothing
         */
		protected boolean changes = false;

        /**
         * list of segments
         */
		protected ArrayList<OperatorPath> partition = null;
		
        /**
         * list of deleted buffers, will be deregistered at the end
         */
		protected ArrayList<AbstractBufferPipe> deletedBuffer;

		/**
		 * new instance
		 * 
		 * @param g considered graph
		 * @param updatePeriod updateperiod of processor in ms
		 */
		public Refresh(Graph g, long updatePeriod) {
			super(updatePeriod);
			setName("SegmentStrategy.Refresh");
			graph = g;
			Node[] n = graph.getPipes();
			segmentToPath = new HashMap<ArrayList<Node>, Integer>();
			partition = new ArrayList<OperatorPath>();
			deletedBuffer = new ArrayList<AbstractBufferPipe>();
			operatorPaths = graph.getOperatorPaths();
			for (int i = 0; i < n.length; i++) {
				Node node = n[i];
				if (!graph.containsQueue(node))  {
					node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
					node.getMetaDataManagement().include(COST_MEASUREMENT);
					node.getMetaDataManagement().include(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
					node.getMetaDataManagement().include(ALL_INPUT_OBJECT_SIZES);
					node.getMetaDataManagement().include(OUTPUT_OBJECT_SIZE);
				}
			}
			ArrayList<ArrayList<Node>> s = buildSegments();
			s = prune(s);
			segments = partition = completeSegments(s);
			for (int i=0;i<deletedBuffer.size();i++) {
				deregister(deletedBuffer.get(i));
			}
		}

		/**
		 * @see xxl.core.pipes.processors.Processor#process()
		 */
		@Override
		public void process() {
			if (getNoOfControllables() == 0)
				terminate();
			ArrayList<ArrayList<Node>> s = buildSegments();
			if (changes) {
				s = prune(s);
				partition = completeSegments(s);
				synchronized (segments) {
					segments = partition;
				}
				for (int i=0;i<deletedBuffer.size();i++) {
					deregister(deletedBuffer.get(i));
				}
				deletedBuffer.clear();
			}
		}

        /**
         * builds segments, unnecessary bufferpipes will be removed
         * 
         * @return list of all conbstructed segments
         */
		protected ArrayList<ArrayList<Node>> buildSegments() {
			changes = false;
			ArrayList<ArrayList<Node>> segments = new ArrayList<ArrayList<Node>>();
			ArrayList<Node> segment = new ArrayList<Node>();
			for (int i = 0; i < operatorPaths.length; i++) {
				OperatorPath path = operatorPaths[i];
				Iterator<Node> it = path.iterator(true);
				AbstractBufferPipe bp = null;
				double prev = 0.0;
				Node prevNode = null;
				while (it.hasNext()) {
					Node node = it.next();
					if (path.isControllable(node))
						bp = (AbstractBufferPipe) node;
					else if (graph.containsPipe(node) && !((Pipe) node).isDone()) {
						double mem_rl_cap = operatorMemoryReleaseCapacity.invoke(node);
						if (prev <= mem_rl_cap && prevNode != null) {
							if (bp != null) {
								Source source = (Source) prevNode;
								Sink sink = (Sink) node;
								changes = true;
								// flush bufferpipe to avoid implicit load shedding
								graph.removeBuffer(source, bp, sink);
								deletedBuffer.add(bp);
							}
							segment.add(node);
						} else {
							if (segment.size() > 0) {
								segmentToPath.put(segment, i);
								segments.add(segment);
								segment = new ArrayList<Node>();
							}
							if (bp != null)
								segment.add(bp);
							segment.add(node);
							prevNode = node;
							prev = mem_rl_cap;
						}
						bp = null;
					}
				}
				segmentToPath.put(segment, i);
				segments.add(segment);
				if (bp != null) {
					segment = new ArrayList<Node>();
					segment.add(bp);
					segmentToPath.put(segment, i);
					segments.add(segment);
				}
			}
			return segments;
		}

        /**
         * prune segments that are not necessary
         * @param toPrune list of segments
         * @return cleaned list of segments
         */
		protected ArrayList<ArrayList<Node>> prune(
				ArrayList<ArrayList<Node>> toPrune) {
			ArrayList<ArrayList<Node>> segments = new ArrayList<ArrayList<Node>>();
			segments.add(toPrune.get(0));
			for (int i = 1; i < toPrune.size(); i++) {
				ArrayList<Node> segment = toPrune.get(i);
				boolean insert = false;
				for (int j = 0; j < segments.size(); j++) {
					if (segment.containsAll(segments.get(j))) {
						insert = true;
						segments.remove(j--);
					}
				}
				if (!insert) {
					insert = true;
					for (int j = 0; j < segments.size(); j++) {
						if (segments.get(j).containsAll(segment)) {
							insert = false;
							break;
						}
					}
				}
				if (insert)
					segments.add(segment);
			}
			return segments;
		}

        /**
         * finish segments, necessary bufferpipes will be added (bufferpipes at the start of a segment)
         * 
         * @param segments list of segments
         * @return complete list of segments
         */
		protected ArrayList<OperatorPath> completeSegments(
				ArrayList<ArrayList<Node>> segments) {
			ArrayList<Node> last = null;
			int lastIndex = -1;
			for (int i = 0; i < segments.size(); i++) {
				ArrayList<Node> segment = segments.get(i);
				int pathIndex = segmentToPath.get(segment);

				if (!graph.containsQueue(segment.get(0))) {
					if (pathIndex == lastIndex) {
						Source opSrc = (Source) last.get(last.size() - 1);
						Sink opGoal = (Sink) segment.get(0);
						if (!opGoal.isDone()) {
							AbstractBufferPipe bp = graph.addBuffer(opSrc, opGoal,false);
							if (bp!=null) {
								segment.add(0, bp);
								register(bp);
							}
						}
					} else {
						Sink sink = (Sink) segment.get(0);
						for (int j = 0; j < sink.getNoOfSources(); j++) {
							Source src = sink.getSource(j);
							if (!graph.containsQueue(src)) {
								int indexOut = src.getSinkID(sink);
								AbstractBufferPipe bp = graph.addBuffer(src, sink,true);
								register(bp);
								if (j > 0) {
									ArrayList<Node> clone = (ArrayList<Node>) segment
											.clone();
									clone.add(0, src.getSink(indexOut));
									segments.add(clone);
								} else
									segment.add(0, src.getSink(indexOut));
							} else {
								if (j > 0) {
									ArrayList<Node> clone = (ArrayList<Node>) segment
											.clone();
									clone.add(0, sink.getSource(j));
									segments.add(clone);
								} else
									segment.add(0, sink.getSource(j));
							}
						}
					}
				}
				lastIndex = pathIndex;
				last = segment;
			}
			ArrayList<OperatorPath> result = new ArrayList<OperatorPath>();
			HashMap<OperatorPath, Double> segmentToMRC = new HashMap<OperatorPath, Double>();
			for (int i = 0; i < segments.size(); i++) {
				ArrayList<Node> segment = segments.get(i);
				OperatorPath path = new OperatorPath(graph, segment);
				result.add(path);
				double pmrc = operatorPathMemoryReleaseCapacity.invoke(path);
				segmentToMRC.put(path, pmrc);
			}
			Comparator<OperatorPath> comp = new InverseComparator(new HashComparator<OperatorPath, Double>(
					segmentToMRC));
			Collections.sort(result, comp);
			return result;
		}

		/**
		 * @see xxl.core.pipes.processors.Processor#terminate()
		 */
		@Override
		public void terminate() {
			Node[] n = graph.getPipes();
			for (int i = 0; i < n.length; i++) {
				Node node = n[i];
				if (!graph.containsQueue(node))
					cleanupNode(node);
			}
			super.terminate();
		}

        /**
         * 
         * remove metadata from node
         * 
         * @param node node with metadata
         */
		protected void cleanupNode(Node node) {
			if (node.getMetaDataManagement().getMetaData()!=null) {
				if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE)))
					node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_RATE));
				if (node.getMetaDataManagement().getMetaData().contains(COST_MEASUREMENT))
					node.getMetaDataManagement().exclude(COST_MEASUREMENT);
				if (node.getMetaDataManagement().getMetaData().contains(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO)))
					node.getMetaDataManagement().exclude(new AggregationOverWindowMetaDataIdentifier(Average.class,windowSize,minimumResultWindow,INPUT_OUTPUT_RATIO));
				if (node.getMetaDataManagement().getMetaData().contains(ALL_INPUT_OBJECT_SIZES))
					node.getMetaDataManagement().exclude(ALL_INPUT_OBJECT_SIZES);
				if (node.getMetaDataManagement().getMetaData().contains(OUTPUT_OBJECT_SIZE))
					node.getMetaDataManagement().exclude(OUTPUT_OBJECT_SIZE);
			}
		}
	}

    /**
     * new instance
     * 
     * @param g considered graph
     * @param period updateperiod of processor in ms
     */
	public SegmentStrategy(Graph g, long period) {
		super(g.getQueues());
		segments = new ArrayList<OperatorPath>();
		startProcessor(g, period);
		
	}

    /**
     * starts and initiates refresh processor
     * 
     * @param g considered graph
     * @param period updateperiod of processor in ms
     */
	protected void startProcessor(Graph g, long period) {
		refresh = new Refresh(g, period);
		refresh.start();
	}

	/**
     * current segments 
	 */
	protected ArrayList<OperatorPath> segments;

	/**
	 * refresh-threads
	 */
	protected Refresh refresh;

	/**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
	 */
	@Override
	public boolean computeNext() {
		synchronized (segments) {
			for (int i = 0; i < segments.size(); i++) {
				next = (AbstractBufferPipe) segments.get(i).getSource();
				if (next.isFinished())
					deregister(next);
				else if (next.isReady()) {
					return true;
				}
			}
		}
		return false;
	}
}
