/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.scheduler.SchedulingFunctions.operatorMemoryReleaseCapacity;

import java.util.ArrayList;
import java.util.Iterator;

import xxl.core.pipes.operators.Pipe;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.queryGraph.Node;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.sinks.Sink;
import xxl.core.pipes.sources.Source;

/**
 * Implementation of simplified segment strategy proposed in 'Schedulling strategies
 * for processing continuous queries over streams'
 * 
 * each operatorpath will be partitioned into two disjunct segments, which will be scheduled.
 * Each segment will be processed in FCFS-order.
 * 
 * The decision is based on the memory 
 * release capacity of following  operators in a segment and a similarity factor gamma.  
 * 
 * Each segment has the segmentmemoryrelease capacity as priority.
 * 
 * segment-partition will be recalculated periodically.
 * 
 * Segments have bufferpipes at the start only. Interior bufferpipes will be removed.
 * 
 * assumption: No chain of bufferpipes exist.
 * 
 */
public class SimplifiedSegmentStrategy extends SegmentStrategy {

    /**
     * refresh - processor, adjusts segments periodically
     * 
     */
	class Refresh extends SegmentStrategy.Refresh {
	
		/**
		 * new instance 
		 * 
		 * @param gr considered graph
		 * @param updatePeriod updateperiod of processor in ms
		 */
		public Refresh(Graph gr, long updatePeriod) {
			super(gr,updatePeriod);
			setName("SimplifiedSegment.Refresh");
		}
		
		/**
         * 
         * @see xxl.core.pipes.scheduler.strategies.SegmentStrategy.Refresh#buildSegments()
		 */
		protected ArrayList<ArrayList<Node>> buildSegments() {
			changes = false;
			ArrayList<ArrayList<Node>> segments = new ArrayList<ArrayList<Node>>();
			ArrayList<Node> segment = null;
			for (int i = 0; i < operatorPaths.length; i++) {
				segment = new ArrayList<Node>();
				OperatorPath path = operatorPaths[i];
				Iterator<Node> it = path.iterator(true);
				AbstractBufferPipe bp = null;
				boolean secondSegment = false;
				double prev = 0.0;
				Node prevNode = null;
				while (it.hasNext()) {
					Node node = it.next();
					if (path.isControllable(node))
						bp = (AbstractBufferPipe) node;
					else if ((graph.containsPipe(node) && !((Pipe) node).isDone()) || !graph.isInitialSource(node)) {
						double mem_rl_cap = 0.0;
						if (!secondSegment)
							mem_rl_cap = operatorMemoryReleaseCapacity.invoke(node);
						if (prevNode!=null) {
							if (!secondSegment) {
								if (prevNode != null && (mem_rl_cap/prev) < gamma) {
									secondSegment = true;
									segmentToPath.put(segment, i);
									segments.add(segment);
									segment = new ArrayList<Node>();
								}
							}
							if (bp != null && segment.size()==0) { 
								segment.add(bp);
							}
							else if (bp != null) {
								Source source = (Source) prevNode;
								Sink sink = (Sink) node;
								if (!sink.isDone()) {
									graph.removeBuffer(source, bp, sink);
                                    changes = true;
									deletedBuffer.add(bp);
									cleanupNode(bp);
								}
							}
							else if (segment.size()==0) {
								Source source = (Source) prevNode;
								Sink sink = (Sink) node;
								if (!sink.isDone()) {
									bp = graph.addBuffer(source, sink, false);
									register(bp);
									segment.add(bp);
								}
							}
							bp = null;
						}
						segment.add(node);
						prevNode = node;
						prev = mem_rl_cap;
					}
					if (segment.size()==0 && bp!=null) {
						segment.add(bp);
						bp = null;
					}
				}
				segmentToPath.put(segment, i);
				segments.add(segment);
			}
			return segments;
		}
		
        /**
         * 
         * @see xxl.core.pipes.scheduler.strategies.SegmentStrategy.Refresh#process()
         */
		@Override
		public void process() {
			if (getNoOfControllables()==0)
				terminate();
			else {
				ArrayList<ArrayList<Node>> s = buildSegments();
				if (changes) {
					partition = completeSegments(s);
					synchronized (segments) {
						segments.clear();
						for (int i = 0; i < partition.size(); i++) {
							segments.add(partition.get(i));
						}
					}
					for (int i=0;i<deletedBuffer.size();i++) {
						deregister(deletedBuffer.get(i));
					}
					deletedBuffer.clear();
				}
			}
		}
        
        /**
         *  sets segments and deregisgters unnecessary bufferpipes
         *
         */
        protected void build() {
            segments = partition = completeSegments(buildSegments());
            for (int i=0;i<deletedBuffer.size();i++) {
                deregister(deletedBuffer.get(i));
            }
            deletedBuffer.clear();
        }
	}
    

	/**
     * similarity factor 
	 */
	private double gamma;

    /**
     * new instance
     * @param gr considere graph
     * @param ga similarity factor gamma
     * @param period updateperiod of processor in ms
     */
	public SimplifiedSegmentStrategy(Graph gr, double ga, long period) {
		super(gr,period);
		gamma = ga;
	}

    /**
     * 
     * @see xxl.core.pipes.scheduler.strategies.SegmentStrategy#startProcessor(xxl.core.pipes.queryGraph.Graph, long)
     */
	protected void startProcessor(Graph g, long period) {
		refresh = new Refresh(g, period);
		refresh.start();
	}
	
	/**
	 * terminate refresh processor
	 *
	 */
	public void terminateRefresh() {
		refresh.terminate();
	}
}
