/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/
package xxl.core.pipes.scheduler.strategies;

import java.util.ArrayList;
import java.util.HashMap;

import xxl.core.functions.Function;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.queryGraph.OperatorPath;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.Controllable;
import xxl.core.pipes.scheduler.HashComparator;

/**
 * implementation of shortest path approach as described in master thesis of Udo Wolske
 * For more information: @see: Master Thesis of Udo Wolske: Eine Evaluation von Strategien fr die Ablaufssteuerung 
 * fr die Datenstrominfrastruktur PIPES 
 * 
 * sets static priorities for each bufferpipe.: The priority is the shortest distance from a source of the graph 
 * to this bufferpipe
 * 
 * a processor actualizes periodically a heap, which includes all bufferpipes containing tuples
 */
public class ThreadedShortestPathStrategy extends AbstractStrategy<AbstractBufferPipe> {
	
    /**
     * refresh processor, adjusting heap
     * @author Udo Wolske
     */
	class Priority extends Processor {
		
        /**
         * list of queues
         */
		private ArrayList<AbstractBufferPipe> controllables;
		
        /**
         * function: bufferpipe -> priority
         */
		private Function<AbstractBufferPipe, Integer> assignPriority = new Function<AbstractBufferPipe, Integer>() {

            /**
             * @see xxl.core.functions.Function#invoke(java.lang.Object)
             */
            @Override
			public Integer invoke(AbstractBufferPipe cont) {
				int res = Integer.MAX_VALUE;
				OperatorPath[] paths = graph.getOperatorPaths();
				for (int i = 0; i < paths.length; i++) {
					if (paths[i].isControllable(cont)) {
						Controllable[] c = paths[i].getQueues();
						for (int j = 0; j < c.length; j++) {
							if (c[j] == cont) {
								if (j < res)
									res = j;
								break;
							}
						}
					}
				}
				return res;
			}
		};

				
        /**
         * heap
         */
		private CollectionHeap<AbstractBufferPipe> h;
		
		/**
         * map : bufferpipe -> priority 
		 */
		private HashMap<AbstractBufferPipe,Integer> priorities;
        
        /**
         * comparator
         */
        private HashComparator<AbstractBufferPipe,Integer> comp;
		
        /**
         * new instance
         * 
         * @param period updateperiod of processor in ms
         */
		public Priority(long period) {
			super(period);
			setName("ShortestPathStrategy.Priority");
			controllables = new ArrayList<AbstractBufferPipe>();
			priorities = new HashMap<AbstractBufferPipe,Integer>();
            comp = new HashComparator<AbstractBufferPipe,Integer>(priorities);
			h = new CollectionHeap<AbstractBufferPipe>(comp);
		}
		
        
        /**
         * registers bufferpipe to processor
         * @param node bufferpipe to register
         */
		public void register(AbstractBufferPipe node) {
			controllables.add(node);
			int prio = assignPriority.invoke(node);
			priorities.put(node, prio);
		}
		
        /**
         * deregisters bufferpipe from processor
         * @param node bufferpipe to deregister
         */
		public void deregister(AbstractBufferPipe node) {
			controllables.remove(node);
			priorities.remove(node);
		}
		
        /**
         * @see xxl.core.pipes.processors.Processor#process()
         */
		@Override
		public synchronized void process() {
            if (controllables.size()==0)
                terminate();
            h.clear();
			for (int i=0;i<controllables.size();i++) {
				if (controllables.get(i).isReady()) {
					if (!h.contains(controllables.get(i)))
						h.enqueue(controllables.get(i));
				}
                else {
                    if (controllables.get(i).isFinished()) {
                        ThreadedShortestPathStrategy.this.deregister(controllables.get(i--));
                    }
                }
			}
		}
		
		/** 
         * @return next ready element
		 */
		public synchronized AbstractBufferPipe dequeue() {
			return h.dequeue();
		}
		
		/**
		 * @return ? no ready element
		 */
		public boolean isEmpty() {
			return h.isEmpty();
		}
		
	}
	
    /**
     * considered graph
     */
	private Graph graph;

	/**
     * refresh thread with heap 
	 */
	private Priority heap;


	/**
     * new instance 
     * @param g considered graph
     * @param period updateperiod of processor in ms
	 */
	public ThreadedShortestPathStrategy(Graph g, long period) {
        this.graph = g;
        heap = new Priority(period);
        AbstractBufferPipe[] nodes = g.getQueues();
        for (int i = 0; i < nodes.length; i++)
            register(nodes[i]);
        heap.start();
	}

	/**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
	 */
    @Override
	public boolean computeNext() {
		while (controllables.size() > 0) {
			if (heap.isEmpty()) {
				return false;
            }
			next = heap.dequeue();
			if (!next.isFinished()) {
				return true;
			}
		}
		return false;
	}

	/** 
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#register(xxl.core.pipes.scheduler.Controllable)
	 */
    @Override
	public void register(AbstractBufferPipe node) {
		super.register(node);
		heap.register(node);
	}

	/**
	 * @see xxl.core.pipes.scheduler.AbstractStrategy#deregister(xxl.core.pipes.scheduler.Controllable)
	 */
    @Override
	public void deregister(AbstractBufferPipe node) {
		super.deregister(node);
		heap.deregister(node);
	}
}
