/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.pipes.scheduler.strategies;

import static xxl.core.pipes.operators.AbstractPipe.AbstractPipeMetaDataManagement.ALL_INPUT_OBJECT_SIZES;
import static xxl.core.pipes.operators.identities.AbstractBufferPipe.AbstractBufferPipeMetaDataManagement.BUFFER_SIZE;
import xxl.core.functions.Function;
import xxl.core.pipes.operators.identities.AbstractBufferPipe;
import xxl.core.pipes.processors.Processor;
import xxl.core.pipes.scheduler.AbstractStrategy;
import xxl.core.pipes.scheduler.Controllable;
import xxl.core.pipes.queryGraph.Graph;
import xxl.core.pipes.scheduler.Strategy;
import xxl.core.util.metaData.MetaDataHandler;

/**
 * Implementation of threshold strategy as proposed in 'Scheduling strategies
 * for processing continuous queries over streams'
 * 
 * hybrid strategy: uses two strategies
 *  1.): strategy, that induces a good tuple latency (e.g. pathcapacity)
 *  2.): strategy, that minimizes the queue-memory requirement (e.g. chain)
 *  
 *  default strategy strategy 1
 *  minimizes the memory requirement for queues, 
 *  
 *  when the memreq increases and crosses threshold t_max use strategy 2.
 *  
 *  if strategy 2 is used:
 *  when memreq decreases and crosses threshold t_min
 *  use strategy 1.
 *  
 *  attention: only to use with strategies, which don't remove or add buffers 
 *  between operators to avoid nasty side effects.
 * 
 */
public class ThresholdStrategy<T extends AbstractBufferPipe> extends AbstractStrategy<Controllable> {

	/**
     * refresh processor
     * 
     * determines periodically, which strategy is to be used
     * 
	 */
	class Refresh extends Processor {

		/**
		 * new instance
         * 
		 * @param updatePeriod update period of processor in ms
		 */
		public Refresh(long updatePeriod) {
			super(updatePeriod);
		}
		
        /**
         * 
         * @see xxl.core.pipes.processors.Processor#process()
         */
		public void process() {
			if (getNoOfControllables()==0)
				terminate();
			long ql = 0;
			for (Controllable cont : controllables) {
                AbstractBufferPipe abp = (AbstractBufferPipe) cont;
				int bs = (Integer) ((MetaDataHandler) abp.getMetaDataManagement().getMetaData().get(BUFFER_SIZE)).getMetaData();
				long ios = ((Long[]) ((MetaDataHandler) abp.getMetaDataManagement().getMetaData().get(ALL_INPUT_OBJECT_SIZES)).getMetaData())[0];
				ql += ios*bs;
			}
			synchronized(contingencyMode) {
				if (!contingencyMode)
					contingencyMode = (ql>maxThreshold.invoke()) ? true : false;
				else
					contingencyMode = (ql<minThreshold.invoke()) ? false : true;
			}
		}
	}
	
    /**
     * default strategy (good tuple latency)
     */
	private Strategy<AbstractBufferPipe> defaultStrategy;
    
    /**
     * contingency strategy (minimizes memreq)
     */
	private Strategy<AbstractBufferPipe> contingencyStrategy;
	
    /**
     * which strategy is used
     */
	private Boolean contingencyMode = false;
	
    /**
     * available memory
     */
	private long globalMemory;
    
    /**
     * current memory requirement
     */
	private long queueMemory;
	
    /**
     * refresh processor
     */
	private Refresh refresh;
	
    /**
     * safety-value @see 'Scheduling strategies
     * for processing continuous queries over streams'
     */
	private double beta=0.75;
	
    /**
     * how much memory is used in % 
     */
	private Function<Object,Double> alpha = new Function<Object,Double>() {
		public Double invoke() {
			return (double)queueMemory/(double)globalMemory;
		}
	};
	
    /**
     * t_max
     */
	private Function<Object,Double> maxThreshold = new Function<Object,Double>() {
		public Double invoke() {
			double a = (1 + alpha.invoke()/2)*globalMemory;
			double b = beta * globalMemory;
			return Math.min(a, b);
		}
	};
	
    /**
     * t_min
     */
	private Function<Object,Double> minThreshold = new Function<Object,Double>() {
		public Double invoke() {
			double a = queueMemory;
			double b = beta * maxThreshold.invoke();
			return Math.min(a,b);
		}
	};
		
	/**
	 * new instance
     * 
     * 
	 * @param graph considered graph
	 * @param maxQueueMemory maximal available memory
	 * @param beta beta-value
	 * @param updatePeriod updateperiod for processor in ms
	 */
	public ThresholdStrategy(Graph graph,long maxQueueMemory, double beta, long updatePeriod) {
		this(graph,new PathCapacityStrategy(graph,updatePeriod), new ChainStrategy(graph,updatePeriod),maxQueueMemory,beta,updatePeriod);
	}
	
	/**
	 * new instance
     * 
	 * @param gr considered graph
	 * @param def default strategy
	 * @param con contingency strategy
	 * @param maxQueueMemory naximal available memory
	 * @param beta beta-value
	 * @param updatePeriod updateperiod of processor in ms
	 */
	public ThresholdStrategy(Graph gr, Strategy def, Strategy<AbstractBufferPipe> con, long maxQueueMemory, double beta,long updatePeriod) {
		super(gr.getQueues());
		defaultStrategy = def;
		contingencyStrategy = con;
		contingencyMode = false;
		globalMemory = maxQueueMemory;
		this.beta = beta;
		refresh = new Refresh(updatePeriod);
		refresh.start();
	}

	/**
     * sets beta 
     * @param b new beta
	 */
	public void setBeta(double b) {
		beta = b;
	}
	
    /**
     * 
     * @see xxl.core.pipes.scheduler.AbstractStrategy#register(xxl.core.pipes.scheduler.Controllable)
     */
	public void register(Controllable node) {
		super.register(node);
        AbstractBufferPipe abp = (AbstractBufferPipe) node;
		abp.getMetaDataManagement().include(BUFFER_SIZE);
		abp.getMetaDataManagement().include(ALL_INPUT_OBJECT_SIZES);
	}
	
    /**
     * 
     * @see xxl.core.pipes.scheduler.AbstractStrategy#deregister(xxl.core.pipes.scheduler.Controllable)
     */
	public void deregister(Controllable node) {
        if (controllables.contains(node)) {
            super.deregister(node);
            AbstractBufferPipe abp = (AbstractBufferPipe) node;
            abp.getMetaDataManagement().exclude(BUFFER_SIZE);
            abp.getMetaDataManagement().exclude(ALL_INPUT_OBJECT_SIZES);
        }
	}
	
    /**
     * 
     * @see xxl.core.pipes.scheduler.AbstractStrategy#computeNext()
     */
	@Override
	public boolean computeNext() {
		while (controllables.size()>0) {
			boolean res = false;
			synchronized (contingencyMode) {
				if (contingencyMode) {
					res = contingencyStrategy.computeNext();
					if (res)
						next = contingencyStrategy.next();
				} else {
					res = defaultStrategy.computeNext();
					if (res)
						next = defaultStrategy.next();
				}
			}
			if (res)
				return true;
			else if (next!=null && next.isFinished())
				deregister(next);
		}
		return false;
	}

}
