/* XXL: The eXtensible and fleXible Library for data processing

Copyright (C) 2000-2006 Prof. Dr. Bernhard Seeger
                        Head of the Database Research Group
                        Department of Mathematics and Computer Science
                        University of Marburg
                        Germany

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
USA

    http://www.xxl-library.de

bugs, requests for enhancements: request@xxl-library.de

If you want to be informed on new versions of XXL you can
subscribe to our mailing-list. Send an email to

    xxl-request@lists.uni-marburg.de

without subject and the word "subscribe" in the message body.
*/

package xxl.core.relational.query.operators;

import java.text.FieldPosition;
import java.text.Format;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Map.Entry;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import xxl.core.cursors.Cursor;
import xxl.core.cursors.Cursors;
import xxl.core.cursors.sources.ArrayCursor;
import xxl.core.cursors.sources.SingleObjectCursor;
import xxl.core.functions.DecoratorFunction;
import xxl.core.functions.Function;
import xxl.core.functions.Switch;
import xxl.core.relational.QueryConverter;
import xxl.core.relational.QueryTranslator;
import xxl.core.relational.query.Node;
import xxl.core.relational.query.Nodes;
import xxl.core.util.metaData.CompositeMetaData;
import xxl.core.util.metaData.MetaDataException;
import xxl.core.util.metaData.MetaDataProvider;
import xxl.core.util.metaData.TransparentCompositeMetaData;

/**
 * This class provides static methods for creating <i>ready to use</i>
 * instances of an {@link Node equivalence node} in a directed acyclic graph.
 * Beside these methods, it contains constants for identifying local metadata
 * fragments inside an equivalence node's global metadata, methods for
 * accessing them and local metadata factories for updating them.
 * 
 * @see Node
 */
public class Equivalences {
	
	// format for action maps
	
	/**
	 * This class provides a format that is able to format instances of type
	 * {@link Map} mapping an {@link Action action} to a {@link List list} of
	 * indices (the indices of the nodes where the action is performed).
	 */
	public static class ActionMapFormat extends Format {
		
		/**
		 * A default instance of the (stateless) action map format. It can be
		 * (re)used to avoid multiple instances of this class.
		 */
		public static final ActionMapFormat DEFAULT_INSTANCE = new ActionMapFormat();

		/**
		 * The number format that is used for parsing integers.
		 */
		protected NumberFormat numberFormat;
		
		/**
		 * Creates a new action map format that is able to format instances of
		 * type {@link Map} mapping an {@link Action action} to a
		 * {@link List list} of indices (the indices of the nodes where the
		 * action is performed).
		 */
		public ActionMapFormat() {
			numberFormat = NumberFormat.getIntegerInstance(Locale.GERMANY);
			numberFormat.setGroupingUsed(false);
		}
		
		/**
		 * Formats an action map and appends the resulting text to a given
		 * string buffer. If the <code>pos</code> argument identifies a field
		 * used by the format, then its indices are set to the beginning and
		 * end of the first such field encountered.
		 * 
		 * @param <T> the type of the list storing the indices.
		 * @param actions the action map to format.
		 * @param toAppendTo where the text is to be appended.
		 * @param pos a <code>FieldPosition</code> identifying a field in the
		 *        formatted text.
		 * @return the string buffer passed in as <code>toAppendTo</code>, with
		 *         formatted text appended.
		 * @throws NullPointerException if <code>toAppendTo</code> or
		 *         <code>pos</code> is null.
		 */
		public <T extends List<Integer>> StringBuffer formatActionMap(Map<Action, T> actions, StringBuffer toAppendTo, FieldPosition pos) {
			StringBuffer stringBuffer = new StringBuffer();
			if (actions == null)
				stringBuffer.append("null");
			else {
				stringBuffer.append('[');
				Iterator<Entry<Action, T>> entries = actions.entrySet().iterator();
				while (entries.hasNext()) {
					Entry<Action, T> entry = entries.next();
					stringBuffer.append(entry.getKey().toString()).append(' ');
					if (entry.getValue() == null)
						stringBuffer.append("null");
					else {
						stringBuffer.append('[');
						Iterator<Integer> indices = entry.getValue().iterator();
						while (indices.hasNext()) {
							stringBuffer.append(numberFormat.format(indices.next()));
							if (indices.hasNext())
								stringBuffer.append(", ");
						}
						stringBuffer.append(']');
					}
					if (entries.hasNext())
						stringBuffer.append(", ");
				}
				stringBuffer.append(']');
			}
			pos.setBeginIndex(toAppendTo.length());
			pos.setEndIndex(toAppendTo.length() + stringBuffer.length());
			return toAppendTo.append(stringBuffer);
		}

		/**
		 * Formats an action map to produce a string. This is equivalent to
		 * <blockquote>
		 * {@link #formatActionMap(Map, StringBuffer, FieldPosition) formatActionMap}<code>(actions, new StringBuffer(), new FieldPosition(0)).toString();</code>
		 * </blockquote>
		 *
		 * @param actions the action map to format.
		 * @return formatted string.
		 */
		public String formatActionMap(Map<Action, ? extends List<Integer>> actions) {
			return formatActionMap(actions, new StringBuffer(0), new FieldPosition(0)).toString();
		}

		/**
		 * Formats an action map and appends the resulting text to a given
		 * string buffer. If the <code>pos</code> argument identifies a field
		 * used by the format, then its indices are set to the beginning and
		 * end of the first such field encountered.
		 * 
		 * @param object the action map to format.
		 * @param toAppendTo where the text is to be appended.
		 * @param pos a <code>FieldPosition</code> identifying a field in the
		 *        formatted text.
		 * @return the string buffer passed in as <code>toAppendTo</code>, with
		 *         formatted text appended.
		 * @throws NullPointerException if <code>toAppendTo</code> or
		 *         <code>pos</code> is null.
		 * @throws IllegalArgumentException if the format cannot format the
		 *         given object.
		 */
		@Override
		public StringBuffer format(Object object, StringBuffer toAppendTo, FieldPosition pos) {
			try {
				return formatActionMap((Map<Action, ? extends List<Integer>>)object, toAppendTo, pos);
			}
			catch (ClassCastException cce) {
				throw new IllegalArgumentException("the specified object is not of the expected type; " + cce.getMessage());
			}
		}
		
		/**
		 * Parses text from a string to produce an action map.
		 * 
		 * <p>The method attempts to parse text starting at the index given by
		 * <code>pos</code>. If parsing succeeds, then the index of
		 * <code>pos</code> is updated to the index after the last character
		 * used (parsing does not necessarily use all characters up to the end
		 * of the string), and the parsed object is returned. The updated
		 * <code>pos</code> can be used to indicate the starting point for the
		 * next call to this method. If an error occurs, then the index of
		 * <code>pos</code> is not changed, the error index of <code>pos</code>
		 * is set to the index of the character where the error occurred, and
		 * null is returned.
		 * 
		 * @param source a <code>String</code>, part of which should be parsed.
		 * @param pos a <code>ParsePosition</code> object with index and error
		 *        index information as described above.
		 * @return an action map parsed from the string. In case of error,
		 *         returns null.
		 * @throws NullPointerException if <code>pos</code> is null.
		 */
		@SuppressWarnings("null")
		public Map<Action, ? extends List<Integer>> parseActionMap(String source, ParsePosition pos) {
			Map<Action, List<Integer>> actions = null;
			Action action = null;
			List<Integer> indices = null;
			Number number = null;
			
			int state = 1;
			int index = pos.getIndex();
			ParsePosition parsePosition;
			
			while (state == 8 || state == 9 || state != 10 && index < source.length())
				switch (state) {
					case 1:
						if (source.charAt(index) == '[') {
							actions = new HashMap<Action, List<Integer>>();
							state = 2;
							index++;
						}
						else
							if (source.regionMatches(index, "null", 0, 4)) {
								actions = null;
								state = 8;
								index += 4;
							}
							else
								state = 9;
						break;
					case 2:
						action = null;
						for (Action value : Action.values()) {
							String actionString = value.toString();
							if (source.regionMatches(index, actionString, 0, actionString.length())) {
								action = value;
								state = 3;
								index += actionString.length();
								break;
							}
						}
						if (action == null)
							if (source.charAt(index) == ']') {
								state = 8;
								index++;
							}
							else
								state = 9;
						break;
					case 3:
						if (source.charAt(index) == ' ') {
							state = 4;
							index++;
						}
						else
							state = 9;
						break;
					case 4:
						if (source.charAt(index) == '[') {
							indices = new ArrayList<Integer>();
							state = 5;
							index++;
						}
						else
							if (source.regionMatches(index, "null", 0, 4)) {
								indices = null;
								state = 7;
								index += 4;
							}
							else
								state = 9;
						break;
					case 5:
						if ((number = numberFormat.parse(source, parsePosition = new ParsePosition(index))) != null) {
							state = 6;
							index = parsePosition.getIndex();
						}
						else
							if (source.charAt(index) == ']') {
								state = 7;
								index ++;
							}
							else
								state = 9;
						break;
					case 6:
						indices.add(number == null ? null : number.intValue());
						if (source.regionMatches(index, ", ", 0, 2)) {
							state = 5;
							index += 2;
						}
						else
							if (source.charAt(index) == ']') {
								state = 7;
								index++;
							}
							else
								state = 9;
						break;
					case 7:
						actions.put(action, indices);
						if (source.regionMatches(index, ", ", 0, 2)) {
							state = 2;
							index += 2;
						}
						else
							if (source.charAt(index) == ']') {
								state = 8;
								index++;
							}
							else
								state = 9;
						break;
					case 8:
						state = 10;
						pos.setIndex(index);
						break;
					case 9:
						state = 10;
						actions = null;
						pos.setErrorIndex(index);
						break;
				}
			
			return actions;
		}
		
		/**
		 * Parses text from the beginning of the given string to produce an
		 * action map. The method may not use the entire text of the given
		 * string.
		 * 
		 * @param source a <code>String</code> whose beginning should be
		 *        parsed.
		 * @return an action map parsed from the string.
		 * @throws ParseException if the beginning of the specified string
		 *         cannot be parsed.
		 */
		public Map<Action, ? extends List<Integer>> parseActionMap(String source) throws ParseException {
			ParsePosition pos = new ParsePosition(0);
			Map<Action, ? extends List<Integer>>actions = parseActionMap(source, pos);
			if (pos.getIndex() == 0)
				throw new ParseException("ActionMapFormat.parseActionMap(String) failed", pos.getErrorIndex());
			return actions;
		}

		/**
		 * Parses text from a string to produce an object.
		 * 
		 * <p>The method attempts to parse text starting at the index given by
		 * <code>pos</code>. If parsing succeeds, then the index of
		 * <code>pos</code> is updated to the index after the last character
		 * used (parsing does not necessarily use all characters up to the end
		 * of the string), and the parsed object is returned. The updated
		 * <code>pos</code> can be used to indicate the starting point for the
		 * next call to this method. If an error occurs, then the index of
		 * <code>pos</code> is not changed, the error index of <code>pos</code>
		 * is set to the index of the character where the error occurred, and
		 * null is returned.
		 * 
		 * @param source a <code>String</code>, part of which should be parsed.
		 * @param pos a <code>ParsePosition</code> object with index and error
		 *        index information as described above.
		 * @return an <code>Object</code> parsed from the string. In case of
		 *         error, returns null.
		 * @throws NullPointerException if <code>pos</code> is null.
		 */
		@Override
		public Object parseObject(String source, ParsePosition pos) {
			return parseActionMap(source, pos);
		}
		
	}
	
	// stateful XML I/O functions
	
	/**
	 * A job aware implementation of a factory method that can be used to
	 * transform the XML representation of an equivalence node into the node
	 * itself. The factory method decorates another factory (that is
	 * responsible for the proper transformation of the equivalence node) and
	 * additionally reads job information from XML and publishs them in the
	 * equivalence node's global metadata.
	 */
	public static class JobAwareXmlReader extends DecoratorFunction<Object, Node> {
		
		/**
		 * A mapping between the read job numbers and the first equivalence
		 * nodes of the mentioned job with index <tt>0</tt>.
		 */
		protected HashMap<Integer, Node> jobs;
		
		/**
		 * Creates a new job aware implementation of a XML reader for
		 * equivalence nodes. The specified factory is used for the proper
		 * transformation of the equivalence node, while the returned factory
		 * additionally reads the job information from XML and publishs them in
		 * the equivalence node's global metadata.
		 * 
		 * @param function the factory that is used for the proper
		 *        transformation of the equivalence node.
		 */
		public JobAwareXmlReader(Function<Object, ? extends Node> function) {
			super(function);
			jobs = new HashMap<Integer, Node>();
		}
		
		/**
		 * Returns the mapping between the read job numbers and the first
		 * equivalence nodes of the mentioned job with index <tt>0</tt>.
		 * 
		 * @return the mapping between the read job numbers and the first
		 *         equivalence nodes of the mentioned job with index
		 *         <tt>0</tt>.
		 */
		public HashMap<Integer, Node> getJobs() {
			return jobs;
		}
		
		/**
		 * Reads the job information from XML and publishs them in the returned
		 * equivalence node's global metadata.
		 * 
		 * @param args the argument list to the factory method containing the
		 *        constant {@link QueryConverter#READ}, the calling instance of
		 *        the query {@link QueryConverter converter} and a DOM element
		 *        representing the equivalence node to be read.
		 * @return the read equivalence node enriched by job information.
		 */
		@Override
		public Node invoke(List<? extends Object> args) {
			Node equivalence = super.invoke(args);
			
			Element element = (Element)args.get(2);
			
			CompositeMetaData<Object, Object> globalMetaData = equivalence.getMetaData();
			Cursor<Element> jobs = QueryConverter.getChildren(element, QueryConverter.JOB_ELEMENT);
			
			while (jobs.hasNext()) {
				Element job = jobs.next();
				Integer jobNumber = Integer.valueOf(job.getAttribute(QueryConverter.ID_ATTRIBUTE));
				String jobIdentifier = Equivalences.JOB_PREFIX + jobNumber;
				if (job.hasAttribute(QueryConverter.INDEX_ATTRIBUTE)) {
					Integer index = Integer.valueOf(job.getAttribute(QueryConverter.INDEX_ATTRIBUTE));
					globalMetaData.add(jobIdentifier, index);
					if (index == 0 && !getJobs().containsKey(jobNumber))
						getJobs().put(jobNumber, equivalence);
				}
				if (job.hasAttribute(QueryConverter.CHILD_INDICES_ATTRIBUTE))
					try {
						globalMetaData.add(jobIdentifier + Equivalences.CHILD_ACTIONS_SUFFIX, ActionMapFormat.DEFAULT_INSTANCE.parseActionMap(job.getAttribute(QueryConverter.CHILD_INDICES_ATTRIBUTE)));
					}
					catch (ParseException pe) {
						// the metadata provided by the equivalence node has the wrong format
					}
				if (job.hasAttribute(QueryConverter.PARENT_INDICES_ATTRIBUTE))
					try {
						globalMetaData.add(jobIdentifier + Equivalences.PARENT_ACTIONS_SUFFIX, ActionMapFormat.DEFAULT_INSTANCE.parseActionMap(job.getAttribute(QueryConverter.PARENT_INDICES_ATTRIBUTE)));
					}
					catch (ParseException pe) {
						// the metadata provided by the equivalence node has the wrong format
					}
			}
			
			return equivalence;
		}
	};
	
	/**
	 * A job aware implementation of a factory method that can be used to
	 * transform an equivalence node into its XML representation. The factory
	 * method decorates another factory (that is responsible for the proper
	 * transformation of the equivalence node) and additionally writes job
	 * information from the equivalence node's global metadata to XML.
	 */
	public static class JobAwareXmlWriter extends DecoratorFunction<Object, Node> {
		
		/**
		 * The number of the job whose information should be written to XML.
		 */
		protected Integer jobNumber = null;
		
		/**
		 * Creates a new job aware implementation of a XML writer for
		 * equivalence nodes. The specified factory is used for the proper
		 * transformation of the equivalence node, while the returned factory
		 * additionally writes job information from the equivalence node's
		 * global metadata to XML.
		 * 
		 * @param function the factory that is used for the proper
		 *        transformation of the equivalence node.
		 */
		public JobAwareXmlWriter(Function<Object, ? extends Node> function) {
			super(function);
		}
		
		/**
		 * Returns the number of the job whose information could be written to
		 * XML using this XML writer.
		 * 
		 * @return the number of the job whose information could be written to
		 *         XML using this XML writer.
		 */
		public Integer getJobNumber() {
			return jobNumber;
		}

		/**
		 * Sets the number of the job whose information should be written to
		 * XML using this XML writer to the given one.
		 * 
		 * @param jobNumber the number of the job whose information should be
		 *        written to XML using this XML writer.
		 */
		public void setJobNumber(Integer jobNumber) {
			this.jobNumber = jobNumber;
		}

		/**
		 * Writes the job information from the specified equivalence node's
		 * global metadata to XML.
		 * 
		 * @param args the argument list to the factory method containing the
		 *        constant {@link QueryConverter#WRITE}, the calling instance
		 *        of the query {@link QueryConverter converter}, a DOM
		 *        document, a DOM element and the equivalence node whose XML
		 *        representation should be written to the given DOM document
		 *        using the specified DOM element as parent for the new XML
		 *        content.
		 * @return <tt>null</tt> is returned after writing the specfied
		 *         equivalence node to XML.
		 */
		@Override
		public Node invoke(List<? extends Object> args) {
			super.invoke(args);
			
			Document document = (Document)args.get(2);
			Element element = (Element)args.get(3);
			Node equivalence = (Node)args.get(4);
			
			if (jobNumber != null) {
				CompositeMetaData<Object, Object> globalMetaData = equivalence.getMetaData();
				
				Element childElement = document.createElement(QueryConverter.JOB_ELEMENT);
				
				childElement.setAttribute(QueryConverter.ID_ATTRIBUTE, jobNumber.toString());
				String jobIdentifier = JOB_PREFIX + jobNumber;
				if (globalMetaData.contains(jobIdentifier))
					childElement.setAttribute(QueryConverter.INDEX_ATTRIBUTE, globalMetaData.get(jobIdentifier).toString());
				String childIndicesIdentifier = jobIdentifier + CHILD_ACTIONS_SUFFIX;
				if (globalMetaData.contains(childIndicesIdentifier))
					childElement.setAttribute(QueryConverter.CHILD_INDICES_ATTRIBUTE, ActionMapFormat.DEFAULT_INSTANCE.formatActionMap((Map<Action, ? extends List<Integer>>)globalMetaData.get(childIndicesIdentifier)));
				String parentIndicesIdentifier = jobIdentifier + PARENT_ACTIONS_SUFFIX;
				if (globalMetaData.contains(parentIndicesIdentifier))
					childElement.setAttribute(QueryConverter.PARENT_INDICES_ATTRIBUTE, ActionMapFormat.DEFAULT_INSTANCE.formatActionMap((Map<Action, ? extends List<Integer>>)globalMetaData.get(parentIndicesIdentifier)));
				
				element.appendChild(childElement);
			}
			
			return null;
		}
	}
	
	// metadata identifier and metadata fragment constants
	
	/**
	 * This constant can be used to denote that a node represents an
	 * equivalence node in the directed acyclic graph.
	 */
	public static final String OPERATOR_TYPE = "EQUIVALENCE";
	
	/**
	 * This constant is used to identify an equivalance node's choice, i.e.,
	 * the index of the child node that represents the equivalence node's query
	 * at this time, inside its global metadata.
	 */
	public static final String CHOICE = "EQUIVALENCE->CHOICE";
	
	/**
	 * This constant is used to identify the physical representation of an
	 * equivalance node, i.e., the physical representation of the equivalence
	 * node's choice, inside its global metadata.
	 */
	public static final String PHYSICAL_REPRESENTATION = "EQUIVALENCE->PHYSICAL_REPRESENTATION";
	
	/**
	 * This constant is used to identify the additional utilization of an
	 * equivalance node, i.e., the number of nodes that are part of an
	 * equivalence node's query that cannot be reused.
	 */
	public static final String ADDITIONAL_UTILIZATION = "EQUIVALENCE->ADDITIONAL_UTILIZATION";

	/**
	 * A <i>all-signatures</i> strategy that simple returns the equivalence
	 * node's chosen child node.
	 */
	public static final Function<Node, Iterator<Node>> SIGNATURE_OF_CHOICE = new Function<Node, Iterator<Node>>() {
		@Override
		public Iterator<Node> invoke(Node equivalence) {
			return new SingleObjectCursor<Node>(equivalence.getChild(getChoice(equivalence)));
		}
	};

	/**
	 * This constant can be used to identify a job's local metadata inside the
	 * equivalance node's global metadata. The ID identifying the described job
	 * just has to be appended to the specified prefix.
	 */
	public static final String JOB_PREFIX = "JOB#";
	
	/**
	 * This constant can be used to identify local metadata about changes made
	 * to this equivalence node's children during the specified job inside the
	 * equivalance node's global metadata. The ID identifying the described job
	 * just has to be put between the prefix specified above and this suffix.
	 */
	public static final String CHILD_ACTIONS_SUFFIX = "->CHILD_ACTIONS";
	
	/**
	 * This constant can be used to identify local metadata about changes made
	 * to this equivalence node's parents during the specified job inside the
	 * equivalance node's global metadata. The ID identifying the described job
	 * just has to be put between the prefix specified above and this suffix.
	 */
	public static final String PARENT_ACTIONS_SUFFIX = "->PARENT_ACTIONS";
	
	/**
	 * This constant can be used to identify local metadata about the state of
	 * the specified job inside the equivalance node's global metadata. The ID
	 * identifying the described job just has to be put between the prefix
	 * specified above and this suffix.
	 */
	public static final String STATE_SUFFIX = "->STATE";
	
	/**
	 * This constant can be used to denote that the specified job is finished.
	 */
	public static final String FINISHED = "FINISHED";
	
	/**
	 * An enumeration providing the possible actions of a job. It can be used
	 * within local metadata identified by strings composed of
	 * {@link #JOB_PREFIX}, a job number and
	 * {@link #CHILD_ACTIONS_SUFFIX} or rather {@link #PARENT_ACTIONS_SUFFIX}.
	 */
	public static enum Action {
		
		/**
		 * This constant can be used to denote that a connection between this
		 * equivalence node and a <i>child</i> or <i>parent</i> opertaor node
		 * is inserted during the specified job.
		 */
		INSERT,
		
		/**
		 * This constant can be used to denote that a connection between this
		 * equivalence node and a <i>child</i> or <i>parent</i> operator node
		 * is removed during the specified job.
		 */
		REMOVE,
		
		/**
		 * This constant can be used to denote that a connection between this
		 * equivalence node and a <i>child</i> or <i>parent</i> operator node
		 * is replaced during the specified job.
		 */
		REPLACE
		
	}
	
	// metadata fragment factories
	
	/**
	 * A local metadata factory that can be used to create/update the metadata
	 * fragment describing the choice of an equivalence node.
	 */
	public static final Function<Object, Integer> CHOICE_FACTORY = new Function<Object, Integer>() {
		@Override
		public Integer invoke(Object identifier, final Object operator) {
			Node equivalence = (Node)operator;
			TransparentCompositeMetaData<Object, Object> globalMetaData = (TransparentCompositeMetaData<Object, Object>)(equivalence).getMetaData();
			int choice = getChoice(equivalence);
			
			globalMetaData.setCompositeMetaData(equivalence.getChild(choice).getMetaData());
			
			return choice;
		}
	};

	// static 'constructors'
	
	/**
	 * Creates a new equivalence node gathering the equivalent queries
	 * represented by the given operator nodes (the node having the specified
	 * index is the chosen one) and providing the required local metadata
	 * fragments and registers factory methods for this metadata fragments.
	 * 
	 * @param operators an iteration over the operator nodes representing the
	 *        equivalent queries gathered by the returned equivalence node.
	 * @param choice the choice of the returned equivalence node, i.e., the
	 *        index of the equivalence node's child that is actually chosen to
	 *        represent the query.
	 * @return a new equivalence node gathering the equivalent queries
	 *         represented by the given operator nodes (the node having the
	 *         specified index is the chosen one) and providing the required
	 *         local metadata fragments and registers factory methods for this
	 *         metadata fragments.
	 */
	public static final Node newEquivalence(Iterator<? extends Node> operators, int choice) {
		Node equivalence = new Node(new TransparentCompositeMetaData<Object, Object>(), new Switch<Object, Object>(), new HashMap<Object, Class<?>>());
		CompositeMetaData<Object, Object> globalMetaData = equivalence.getMetaData();
		Switch<Object, Object> globalMetaDataFactory = equivalence.getGlobalMetaDataFactory();
		Map<Object, Class<?>> signature = equivalence.getSignature();
		
		globalMetaDataFactory.put(equivalence, Nodes.GLOBAL_METADATA_FACTORY);
		globalMetaData.put(Nodes.TYPE, Operators.NODE_TYPE);
		globalMetaData.put(Nodes.NUMBER_OF_CHILDREN, Nodes.VARIABLE);
		globalMetaData.put(Nodes.NUMBER_OF_PARENTS, Nodes.VARIABLE);
		globalMetaData.put(Nodes.ALL_SIGNATURES_STRATEGY, SIGNATURE_OF_CHOICE);
		globalMetaData.put(Operators.TYPE, OPERATOR_TYPE);
		globalMetaData.put(CHOICE, choice);
		globalMetaDataFactory.put(CHOICE, CHOICE_FACTORY);
		
		signature.put(Nodes.TYPE, String.class);
		signature.put(Nodes.NUMBER_OF_CHILDREN, Integer.class);
		signature.put(Nodes.NUMBER_OF_PARENTS, Integer.class);
		signature.put(Operators.TYPE, String.class);
		
		Node operator;
		for (int i = 0; operators.hasNext(); i++) {
			if (Nodes.getType(operator = operators.next()) != Operators.NODE_TYPE)
				throw new IllegalArgumentException("only operators can be used as child nodes of an equivalence operator");
			
			equivalence.addChild(operator);
			if (i == choice)
				((TransparentCompositeMetaData<Object, Object>)globalMetaData).setCompositeMetaData(operator.getMetaData());
			
			for (int j = 0; j < operator.getActualNumberOfChildren(); j++) {
				Node child = operator.getChild(j);
				if (Operators.getType(child) != OPERATOR_TYPE) {
					Node[] parents = Cursors.toFittingArray(child.getParents(), new Node[child.getActualNumberOfParents()]);
					Node newChild = newEquivalence(child);
					for (Node parent : parents)
						parent.replaceChild(child, newChild);
				}
			}
		}
		
		return equivalence;
	}
	
	/**
	 * Creates a new equivalence node gathering the equivalent queries
	 * represented by the given operator nodes (the node having the specified
	 * index is the chosen one) and providing the required local metadata
	 * fragments and registers factory methods for this metadata fragments.
	 * 
	 * @param choice the choice of the returned equivalence node, i.e., the
	 *        index of the equivalence node's child that is actually chosen to
	 *        represent the query.
	 * @param operators the operator nodes representing the equivalent queries
	 *        gathered by the returned equivalence node.
	 * @return a new equivalence node gathering the equivalent queries
	 *         represented by the given operator nodes (the node having the
	 *         specified index is the chosen one) and providing the required
	 *         local metadata fragments and registers factory methods for this
	 *         metadata fragments.
	 */
	public static final Node newEquivalence(int choice, Node... operators) {
		return newEquivalence(new ArrayCursor<Node>(operators), choice);
	}
	
	/**
	 * Creates a new equivalence node representing the query defined by the
	 * given operator node and providing the required local metadata fragments
	 * and registers factory methods for this metadata fragments.
	 * 
	 * @param operator the operator node defining the query that should be
	 *        represented by the returned equivalence node.
	 * @return a new equivalence node representing the query defined by the
	 *         given operator node and providing the required local metadata
	 *         fragments and registers factory methods for this metadata
	 *         fragments.
	 */
	public static final Node newEquivalence(Node operator) {
		return newEquivalence(new SingleObjectCursor<Node>(operator), 0);
	}
	
	// metadata fragment accessors
	
	/**
	 * Returns the choice of the given equivalence node, i.e., the index of the
	 * child node that represents the equivalence node's query at this time.
	 * 
	 * @param equivalence the equivalence node whose choice should be returned.
	 * @return the choice of the given equivalence node, i.e., the index of the
	 *         child node that represents the equivalence node's query at this
	 *         time.
	 */
	public static final int getChoice(Node equivalence) {
		return (Integer)equivalence.getMetaData().get(CHOICE);
	}
	
	/**
	 * Returns the physical representation of the given equivalence node, i.e.,
	 * the physical representation of the given equivalence node's choice.
	 * 
	 * @param equivalence the equivalence node whose physical representation
	 *        should be returned.
	 * @return the physical representation of the given equivalence node, i.e.,
	 *         the physical representation of the given equivalence node's
	 *         choice.
	 */
	@SuppressWarnings("unchecked")
	public static final MetaDataProvider<CompositeMetaData<Object, Object>> getPhysicalRepresentation(Node equivalence) {
		return (MetaDataProvider<CompositeMetaData<Object, Object>>)equivalence.getMetaData().get(PHYSICAL_REPRESENTATION);
	}

	/**
	 * Returns the additional utilization of the given equivalence node, i.e.,
	 * the number of nodes that are part of the query represented by the given
	 * equivalence node that cannot be reused.
	 * 
	 * @param equivalence the equivalence node whose additional utilization
	 *        should be returned.
	 * @return the additional utilization of the given equivalence node, i.e.,
	 *         the number of nodes that are part of the query represented by
	 *         the given equivalence node that cannot be reused.
	 */
	public static final int getAdditionalUtilization(Node equivalence) {
		return (Integer)equivalence.getMetaData().get(ADDITIONAL_UTILIZATION);
	}
	
	// XML I/O
	
	/**
	 * A factory method that can be used to transform the XML representation of
	 * an equivalence node into the node itself.
	 */
	public static final Function<Object, Node> XML_READER = new Function<Object, Node>() {
		@Override
		public Node invoke(List<? extends Object> args) {
			QueryConverter queryConverter = (QueryConverter)args.get(1);
			Element element = (Element)args.get(2);
			
			return newEquivalence(
				queryConverter.readChildren(element, QueryConverter.OPERATOR_ELEMENT),
				Integer.valueOf(element.getAttribute(QueryConverter.CHOICE_ATTRIBUTE))
			);
		}
	};

	/**
	 * A factory method that can be used to transform an equivalence node into
	 * its XML representation.
	 */
	public static final Function<Object, Node> XML_WRITER = new Function<Object, Node>() {
		@Override
		public Node invoke(List<? extends Object> args) {
			QueryConverter queryConverter = (QueryConverter)args.get(1);
			Document document = (Document)args.get(2);
			Element element = (Element)args.get(3);
			Node equivalence = (Node)args.get(4);
				
			element.setAttribute(QueryConverter.CHOICE_ATTRIBUTE, Integer.toString(getChoice(equivalence)));
			queryConverter.writeChildren(equivalence.getChildren(), document, element);
			
			return null;
		}
	};
	
	// query translation
	
	/**
	 * A factory method that can be used to translate (remove) an equivalence
	 * node and replace it by the implementation of the chosen child node.
	 */
	public static final Function<Object, MetaDataProvider<CompositeMetaData<Object, Object>>> TRANSLATION_FUNCTION = new Function<Object, MetaDataProvider<CompositeMetaData<Object, Object>>>() {
		@Override
		public MetaDataProvider<CompositeMetaData<Object, Object>> invoke(List<? extends Object> args) {
			QueryTranslator queryTranslator = (QueryTranslator)args.get(0);
			Node node = (Node)args.get(1);
			
			if (node.getMetaData().contains(PHYSICAL_REPRESENTATION))
				return getPhysicalRepresentation(node);
			
			Cursor<Node> children = node.getChildren();
			
			int choice = getChoice(node);
			
			try {
				MetaDataProvider<CompositeMetaData<Object, Object>> physicalRepresentation = queryTranslator.translate(Cursors.nth(children, choice));
				
				node.getMetaData().put(PHYSICAL_REPRESENTATION, physicalRepresentation);
				
				return physicalRepresentation;
			}
			catch (NoSuchElementException nsee) {
				throw new MetaDataException("the chosen query plan (operator at index " + choice + ") does not exist");
			}
		}
	};
		
	// private constructor

	/**
	 * The default constructor has private access in order to ensure
	 * non-instantiability.
	 */
	private Equivalences() {
		// private access in order to ensure non-instantiability
	}
	
}
