package weka.classifiers.rules;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import weka.classifiers.Classifier;
import weka.classifiers.rules.rise.SimpleConjunctiveRule;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.NoSupportForMissingValuesException;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.UnsupportedAttributeTypeException;
import weka.core.UnsupportedClassTypeException;
import weka.core.Utils;
import weka.core.fuzzy.FuzzySet;
import weka.core.fuzzy.norm.Frank;
import weka.core.fuzzy.norm.TNorm;

/**
 * @author beringer
 *
 */
public class FuzzyCN2 extends Classifier implements OptionHandler {

	public static double CONFIDENCE_0_9 = 2.706;

	public static double CONFIDENCE_0_95 = 3.841;

	public static double CONFIDENCE_0_99 = 6.635;

	public static class FuzzyRule extends SimpleConjunctiveRule {

		protected FuzzySet[][] fuzzySets;

		protected TNorm tnorm;

		protected FuzzyRule() {
		}

		public FuzzyRule(Instances instances, FuzzySet[][] fuzzySets,
				TNorm tnorm, boolean store) {
			super(instances, store);
			this.fuzzySets = fuzzySets;
			this.tnorm = tnorm;
		}

		public FuzzyRule(FuzzyRule rule, TNorm tnorm) {
			super(rule);
			this.tnorm = tnorm;
			fuzzySets = rule.fuzzySets;
		}

		public void computeDistribution(Instances data) {
			resetDistribution();
			int size = data.numInstances();
			for (int i = size - 1; i >= 0; i--) {
				Instance inst = data.instance(i);
				double cover = fuzzyCovers(inst);
				if (cover > 0.0) {
					addClass((int) inst.classValue(), cover * inst.weight());
					if (storeCovered) coveredExamples.addElement(inst);
				}
			}
		}

		public boolean covers(Instance instance) {
			return fuzzyCovers(instance) >= 0.5;
		}

		public double fuzzyCovers(Instance instance) {
			double fuzzy = 1.0;
			for (int i = predCount - 1; i >= 0; i--) {
				int attr = usedAttr[i];
				if (values[attr] != -1) {
					if (fuzzySets[attr] != null) {
						double val = fuzzySets[attr][values[attr]]
								.getFuzzyValue(instance.value(attr));
						fuzzy = tnorm.invoke(fuzzy, val);
					} else if ((((int) instance.value(attr)) != values[attr]))
							return 0.0;
				} else if (bounds[attr] != null) {
					if ((Utils.gr(bounds[attr][0], instance.value(attr)) || Utils
							.gr(instance.value(attr), bounds[attr][1])))
							return 0.0;
				}
			}
			return fuzzy;
		}

		public void computeDistribution(FastVector data) {
			resetDistribution();
			int size = data.size();
			for (int i = size - 1; i >= 0; i--) {
				Instance inst = (Instance) data.elementAt(i);
				double cover = fuzzyCovers(inst);
				if (cover > 0.0) {
					addClass((int) inst.classValue(), cover * inst.weight());
					if (storeCovered) coveredExamples.addElement(inst);
				}
			}
		}

		public String toString() {
			StringBuffer sb = new StringBuffer();
			int pos = 0;
			for (int i = 0; i < predCount; i++) {
				while (values[pos] == -1 && bounds[i] == null)
					pos++;
				if (values[pos] != -1) {
					if (fuzzySets[pos] != null) {
						sb.append(instances.attribute(pos).name() + "="
								+ values[pos] + " ");
					} else {
						sb.append(instances.attribute(pos).name() + "="
								+ instances.attribute(pos).value(values[pos])
								+ " ");
					}
					if (i + 1 < predCount) sb.append("& ");
					pos++;
				}
			}
			if (predCount == 0) {
				sb.append("true ");
			}
			sb.append("=>" + instances.classAttribute().value(classValue));
			return sb.toString();
		}
	}

	protected double convidence = CONFIDENCE_0_95;

	protected long seed = 1;

	protected boolean ordered = false;

	protected int beamCount = 2;

	protected List unusedAttribute;

	protected boolean[] usedAttribute;

	protected Random random;

	protected int classAttr;

	protected FastVector ruleset;

	protected FastVector classruleset;

	protected FuzzyRule defaultRule;

	protected double[] orderedClasses;

	protected int[] classOrder;

	protected double dataCount;

	protected Instances allData;

	protected FuzzySet[][] fuzzySets;

	protected FastVector data;

	protected int attrCount;

	protected int classCount;

	protected FuzzyRule[] beamArray;

	protected TNorm tnorm;

	protected TNorm[] alternativeTNorm;

	public boolean isUsedAttribute(int attribute) {
		if (unusedAttribute == null)
			return true;
		else
			return (unusedAttribute.indexOf(new Integer(attribute)) >= 0);
	}

	/**
	 * Generates a classifier. Must initialize all fields of the classifier
	 * that are not being set via options (ie. multiple calls of buildClassifier
	 * must always lead to the same result). Must not change the dataset
	 * in any way.
	 *
	 * @param dataDFT set of instances serving as training data 
	 * @exception Exception if the classifier has not been 
	 * generated successfully
	 */
	public void buildClassifier(Instances instances) throws Exception {
		if (instances.numInstances() == 0)
				throw new Exception(" No instances with a class value!");

		if (instances.checkForStringAttributes())
				throw new UnsupportedAttributeTypeException(
						" Cannot handle string attributes!");

		if (!instances.classAttribute().isNominal())
				throw new UnsupportedClassTypeException(
						" Only nominal class, please.");

		Enumeration enumAtt = instances.enumerateAttributes();
		while (enumAtt.hasMoreElements()) {
			Attribute attr = (Attribute) enumAtt.nextElement();
			Enumeration enum1 = instances.enumerateInstances();
			while (enum1.hasMoreElements()) {
				if (((Instance) enum1.nextElement()).isMissing(attr)) { throw new NoSupportForMissingValuesException(
						"FuzzyCN2: no missing values, please."); }
			}
		}

		alternativeTNorm = new TNorm[] { new Frank(0.01), new Frank(0.5),
				new Frank(0.99), new Frank(2), new Frank(10), new Frank(100),
				new Frank(10000), new Frank(1000000)};


		allData = instances;

		defaultRule = getRule(allData, fuzzySets);
		defaultRule.computeDistribution(allData);
		defaultRule.computeClass();

		orderedClasses = defaultRule.getDistribution();

		classOrder = Utils.sort(orderedClasses);

		dataCount = allData.numInstances();

		if (allData == null)
				throw new Exception(" Unable to randomize the class orders.");

		allData.deleteWithMissingClass();
		if (allData.numInstances() == 0)
				throw new Exception(" No instances with a class value!");

		ruleset = new FastVector();
		classruleset = new FastVector();

		classCount = allData.numClasses();
		attrCount = allData.numAttributes();
		classAttr = allData.classIndex();
		beamArray = new FuzzyRule[beamCount];

		computeUsedAttributes(allData);

		//CN2 erzeugt Regeln von der seltensten Klasse bis zur hufigsten (Reihenfolge!!!)
		for (int cl = 0; cl < allData.numClasses() - 1; cl++) {
			if (!ordered || cl == 0) data = instanceVector(allData);
			addRulesetForOneClass(classOrder[cl]);
			ruleset.appendElements(classruleset);
		}
		if (!ordered) data = instanceVector(allData);
		if (!ordered) {
			addRulesetForOneClass(classOrder[allData.numClasses() - 1]);
			ruleset.appendElements(classruleset);
		}

		//optimizeRules();
	}

	public void setFuzzySets(FuzzySet[][] fuzzySets) {
		this.fuzzySets = fuzzySets;
	}

	public void setTNorm(TNorm tnorm) {
		this.tnorm = tnorm;
	}

	protected FastVector instanceVector(Instances data) {
		int size = data.numInstances();
		FastVector result = new FastVector(size);

		for (int i = size - 1; i >= 0; i--) {
			Instance inst = data.instance(i);
			inst.setWeight(1.0);
			result.addElement(inst);
		}
		return result;
	}

	protected void computeUsedAttributes(Instances inst) {
		usedAttribute = new boolean[inst.numAttributes()];

		for (int i = 0; i < inst.numAttributes(); i++) {
			if ((i != classAttr)
					&& (inst.attribute(i).isNominal() || fuzzySets[i] != null)
					&& isUsedAttribute(i)) usedAttribute[i] = true;
		}
	}

	protected void addRulesetForOneClass(int classIndex) {
		FuzzyRule bestRule = null;
		classruleset.removeAllElements();
		do {
			//System.out.println(data.size());
			bestRule = findBestRule(classIndex);
			if (bestRule != null) {
				classruleset.addElement(bestRule);
				//System.out.println(bestRule);
				//Alle untersttzten datenstze lschen
				/*int size = data.size();
				 int pos = 0;
				 while (pos < size) {
				 Instance inst = (Instance) data.elementAt(pos);
				 double cover = bestRule.fuzzyCovers(inst);
				 if (cover > 0.0) {
				 double w = inst.weight();
				 if (w - cover < 0.001) {
				 data.removeElementAt(pos);
				 size--;
				 } else {
				 inst.setWeight(w - cover);
				 pos++;
				 }

				 } else
				 pos++;
				 }*/
			}
		} while (bestRule != null && data.size() > 0);
	}

	protected FuzzyRule getRule(Instances data, FuzzySet[][] fuzzySets) {
		return new FuzzyRule(data, fuzzySets, tnorm, true);
	}

	protected FuzzyRule getRule(FuzzyRule rule) {
		return new FuzzyRule(rule, tnorm);
	}

	protected FuzzyRule getEmptyRule(int classIndex) {
		FuzzyRule empty = getRule(allData, fuzzySets);
		empty.setClass(classIndex);
		if (!ordered) {
			for (int i = 0; i < classCount; i++)
				empty.addClass(i, orderedClasses[i]);
			empty.addCovered(data);
		} else {
			empty.computeDistribution(data);
			//if (data.numInstances()==0) empty.addClass(classIndex);
		}
		return empty;
	}

	/**
	 * @param dataDFT
	 * @return
	 */
	protected FuzzyRule findBestRule(int classIndex) {
		FuzzyRule empty = getEmptyRule(classIndex);
		FastVector star = new FastVector();
		star.addElement(empty);
		FuzzyRule best = null;
		int predCount = 0;
		double bestCriteria = -Double.MAX_VALUE;
		while (star.size() > 0) {
			for (int i = 0; i < beamCount; i++)
				beamArray[i] = null;
			Enumeration enum1 = star.elements();
			while (enum1.hasMoreElements()) {
				FuzzyRule rule = (FuzzyRule) enum1.nextElement();
				if (rule.getPredicateCount() == predCount) {
					Iterator it = specialize(rule);
					while (it.hasNext()) {
						FuzzyRule newRule = (FuzzyRule) it.next();
						newRule.computeDistribution(rule.getCovered());
						//newRule.computeDistribution(data);
						double crit = computeCriteria(newRule);
						if (newRule.getCoverage() > 0 && canSpecialize(newRule)) {
							if (newRule.frequentClass() == classIndex
									&& crit > bestCriteria
									&& canBest(newRule,classIndex)) {
								best = newRule;
								bestCriteria = computeCriteria(newRule);
							}
							if (!significance(newRule, classIndex)) {
								int pos = 0;
								while (pos < beamCount
										&& beamArray[pos] != null
										&& computeCriteria(beamArray[pos]) > crit)
									pos++;
								if (pos < beamCount) {
									if (beamArray[pos] != null)
											System.arraycopy(beamArray, pos,
													beamArray, pos + 1,
													beamCount - pos - 1);
									beamArray[pos] = newRule;
								}
							}
						}
					}
				}
			}
			star.removeAllElements();
			int pos = 0;
			while (pos < beamCount && beamArray[pos] != null) {
				star.addElement(beamArray[pos]);
				pos++;
			}
			predCount++;
		}
		return best;
	}

	protected boolean canBest(FuzzyRule rule, int classIndex) {
		if (orderedClasses[classIndex] / dataCount < rule.getDistribution()[classIndex]
				/ rule.getDistributionSum()) return true;
		for (int i=0;i<attrCount;i++){
			if (usedAttribute[i]){
				if (rule.getNominalValue(i)==-1) return false;
			}
		}
		return true;
	}

	protected boolean canSpecialize(FuzzyRule rule) {
		Enumeration enum1 = classruleset.elements();
		while (enum1.hasMoreElements()) {
			FuzzyRule rule2 = (FuzzyRule) enum1.nextElement();
			if (rule2.contains(rule)) return false;
		}
		return true;
	}

	protected double computeCriteria(FuzzyRule rule) {
		return (rule.getDistribution()[rule.getClassValue()] + 1)
				/ (rule.getDistributionSum() + classCount);
	}

	protected boolean significance(FuzzyRule rule, int classIndex) {
		double[] distribution = rule.getDistribution();
		double sum = rule.getDistributionSum();
		int cl = classIndex;
		double p = distribution[cl];
		if (p <= 0) return false;
		if (sum - p <= 0) return false;
		if (p < sum / 2) return false;

		if (orderedClasses[cl] / dataCount >= p / sum) return false;

		double val = 2 * (p * log2(p * dataCount / (orderedClasses[cl] * sum)) + (sum - p)
				* log2((sum - p) * dataCount
						/ ((dataCount - orderedClasses[cl]) * sum)));
		return val >= convidence;
		//return true;
	}

	protected double log2(double d) {
		double v = Math.log(d);
		return v; /*/ log2*/
	}

	protected Iterator specialize(final FuzzyRule rule) {
		return new Iterator() {

			FuzzyRule next = null;

			int attr = -1;

			int val = -1;

			int valCount = -1;

			public void remove() {
			}

			public boolean hasNext() {
				if (next == null)
					return computeNext();
				else
					return true;
			}

			public Object next() {
				if (next == null) computeNext();
				Object val = next;
				next = null;
				return val;
			}

			private boolean computeNext() {
				if (attr >= attrCount) return false;

				if (valCount < 0) {
					nextCondition();
					if (attr >= attrCount) return false;
				}

				next = getRule(rule);
				next.setNominalValue(attr, val);

				nextCondition();

				return true;
			}

			private void nextCondition() {
				val++;
				if (val >= valCount) {
					attr++;
					while (attr < attrCount
							&& ((!usedAttribute[attr]) || rule.getNominalValue(attr) >= 0))
						attr++;
					val = 0;
					if (attr < attrCount) {
						if (fuzzySets[attr] != null)
							valCount = fuzzySets[attr].length;
						else {
							valCount = allData.numDistinctValues(attr);
						}
					}
				}
			}

		};
	}

	/**
	 * Predicts the class memberships for a given instance. If
	 * an instance is unclassified, the returned array elements
	 * must be all zero. If the class is numeric, the array
	 * must consist of only one element, which contains the
	 * predicted value. Note that a classifier MUST implement
	 * either this or classifyInstance().
	 *
	 * @param instance the instance to be classified
	 * @return an array containing the estimated membership 
	 * probabilities of the test instance in each class 
	 * or the numeric prediction
	 * @exception Exception if distribution could not be 
	 * computed successfully
	 */
	public double[] distributionForInstance(Instance instance) throws Exception {
		double[] distribution = new double[classCount];
		double sum = 0;
		try {
			boolean found = false;
			int lastClass = -1;
			for (int i = 0; i < ruleset.size(); i++) {
				FuzzyRule rule = (FuzzyRule) ruleset.elementAt(i);
				//if (found && rule.getClassIndex()==lastClass) continue;

				double cover = rule.fuzzyCovers(instance);
				if (cover > 0.0 /*&& lastClass!=rule.getClassIndex()*/) {
					if (ordered) return rule.getRelativeDistribution();
					found = true;
					lastClass = rule.getClassValue();
					addDistribution(distribution, rule.getDistribution(), cover);
					sum += rule.getDistributionSum() * cover;
				}
			}
			if (!found) {
				return defaultRule.getRelativeDistribution();
			} else {
				for (int i = 0; i < classCount; i++) {
					distribution[i] /= sum;
				}
				return distribution;
			}
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}

		return defaultRule.getRelativeDistribution();
	}

	public void addDistribution(double[] dist, double[] newDist, double weight) {
		for (int i = 0; i < classCount; i++) {
			dist[i] += newDist[i] * weight;
		}
	}

	/**
	 * Returns an enumeration of all the available options..
	 *
	 * @return an enumeration of all available options.
	 */
	public Enumeration listOptions() {
		Vector newVector = new Vector(3);
		newVector.addElement(new Option("\tSet the number of rules\n"
				+ "\tused for beam search\n" + "\t(default 2)", "B", 1,
				"-B <number of rules>"));

		newVector.addElement(new Option("\tSet whether class ordered\n"
				+ "\trules are used (Default: false)", "O", 0, "-O"));

		

		newVector.addElement(new Option("\tThe seed of randomization\n"
				+ "\t(Default: 1)", "S", 1, "-S <seed>"));

		return newVector.elements();
	}

	/**
	 * Sets the OptionHandler's options using the given list. All options
	 * will be set (or reset) during this call (i.e. incremental setting
	 * of options is not possible).
	 *
	 * @param options the list of options as an array of strings
	 * @exception Exception if an option is not supported
	 */
	public void setOptions(String[] options) throws Exception {
		String numBeamString = Utils.getOption('B', options);
		if (numBeamString.length() != 0)
			beamCount = Integer.parseInt(numBeamString);
		else
			beamCount = 2;

		String seedString = Utils.getOption('S', options);
		if (seedString.length() != 0)
			seed = Integer.parseInt(seedString);
		else
			seed = 1;

		ordered = Utils.getFlag('O', options);
	}

	/**
	 * Gets the current option settings for the OptionHandler.
	 *
	 * @return the list of current option settings as an array of strings
	 */
	public String[] getOptions() {
		String[] options = new String[11];
		int current = 0;
		options[current++] = "-B";
		options[current++] = "" + beamCount;
		options[current++] = "-S";
		options[current++] = "" + seed;


		if (!ordered) options[current++] = "-O";

		while (current < options.length)
			options[current++] = "";

		return options;
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		Enumeration enum1 = ruleset.elements();
		while (enum1.hasMoreElements()) {
			FuzzyRule rule = (FuzzyRule) enum1.nextElement();
			sb.append(rule.toString());
			sb.append("\n");
		}
		sb.append(defaultRule.toString());
		sb.append("\n");
		return sb.toString();
	}

	public void optimizeRules() {
		for (int i = 0; i < ruleset.size(); i++) {
			optimizeRule((FuzzyRule) ruleset.elementAt(i));
		}
	}

	public void optimizeRule(FuzzyRule rule) {
		FastVector data = rule.getCovered();
		double bestVal = computeCriteria(rule);
		TNorm best = rule.tnorm;
		System.out.println();
		for (int i = 0; i < alternativeTNorm.length; i++) {

			rule.tnorm = alternativeTNorm[i];
			rule.resetCoveredExamples(data.size());
			rule.computeDistribution(data);
			double val = computeCriteria(rule);
			if (val > bestVal) {
				System.out.println("better one: " + bestVal + " -> " + val);
				bestVal = val;
				best = alternativeTNorm[i];
			} else
				System.out.println(bestVal + " -> " + val);
		}
		rule.tnorm = best;
		rule.resetCoveredExamples(2);
		rule.computeDistribution(data);
	}

}